// SPDX-License-Identifier: MIT
// Copyright (C) 2018-present iced project and contributors

package com.github.icedland.iced.x86;

import com.github.icedland.iced.x86.info.OpCodeInfo;
import com.github.icedland.iced.x86.internal.InstrInfoTable;
import com.github.icedland.iced.x86.internal.MnemonicUtilsData;
import com.github.icedland.iced.x86.internal.info.CpuidFeatureInternalData;
import com.github.icedland.iced.x86.internal.info.InfoFlags2;

/**
 * x86 instruction code
 */
public final class Code {
	private Code() {
	}

	/**
	 * DO NOT USE: INTERNAL API
	 *
	 * @deprecated Not part of the public API
	 */
	@Deprecated
	public static boolean ignoresSegment(int code) {
		switch (code) {
		// GENERATOR-BEGIN: IgnoresSegmentTable
		// ⚠️This was generated by GENERATOR!🦹‍♂️
		case Code.LEA_R16_M:
		case Code.LEA_R32_M:
		case Code.LEA_R64_M:
		case Code.BNDCL_BND_RM32:
		case Code.BNDCL_BND_RM64:
		case Code.BNDCU_BND_RM32:
		case Code.BNDCU_BND_RM64:
		case Code.BNDMK_BND_M32:
		case Code.BNDMK_BND_M64:
		case Code.BNDCN_BND_RM32:
		case Code.BNDCN_BND_RM64:
			return true;
		// GENERATOR-END: IgnoresSegmentTable
		default:
			return false;
		}
	}

	/**
	 * DO NOT USE: INTERNAL API
	 *
	 * @deprecated Not part of the public API
	 */
	@Deprecated
	public static boolean ignoresIndex(int code) {
		switch (code) {
		// GENERATOR-BEGIN: IgnoresIndexTable
		// ⚠️This was generated by GENERATOR!🦹‍♂️
		case Code.BNDLDX_BND_MIB:
		case Code.BNDSTX_MIB_BND:
			return true;
		// GENERATOR-END: IgnoresIndexTable
		default:
			return false;
		}
	}

	static boolean isTileStrideIndex(int code) {
		switch (code) {
		// GENERATOR-BEGIN: TileStrideIndexTable
		// ⚠️This was generated by GENERATOR!🦹‍♂️
		case Code.VEX_TILELOADDT1_TMM_SIBMEM:
		case Code.VEX_TILESTORED_SIBMEM_TMM:
		case Code.VEX_TILELOADD_TMM_SIBMEM:
			return true;
		// GENERATOR-END: TileStrideIndexTable
		default:
			return false;
		}
	}

	/**
	 * Gets an {@link OpCodeInfo}
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	@SuppressWarnings("deprecation")
	public static OpCodeInfo toOpCode(int code) {
		return com.github.icedland.iced.x86.info.InternalOpCodeInfos.opCodes[code];
	}

	/**
	 * Gets the encoding (an {@link EncodingKind} enum variant), eg.<!-- --> Legacy, 3DNow!, VEX, EVEX, XOP
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static int encoding(int code) {
		int[] data = InstrInfoTable.data;
		int index = code * 2 + 1;
		return (data[index] >>> InfoFlags2.ENCODING_SHIFT) & InfoFlags2.ENCODING_MASK;
	}

	/**
	 * Gets the CPU or CPUID feature flags (an array of {@link CpuidFeature} enum variants)
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static int[] cpuidFeatures(int code) {
		int[] data = InstrInfoTable.data;
		int index = code * 2 + 1;
		int cpuidFeature = (data[index] >>> InfoFlags2.CPUID_FEATURE_INTERNAL_SHIFT)
				& InfoFlags2.CPUID_FEATURE_INTERNAL_MASK;
		return CpuidFeatureInternalData.toCpuidFeatures[cpuidFeature];
	}

	/**
	 * Gets control flow info (a {@link FlowControl} enum variant)
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static int flowControl(int code) {
		int[] data = InstrInfoTable.data;
		int index = code * 2 + 1;
		return (data[index] >>> InfoFlags2.FLOW_CONTROL_SHIFT) & InfoFlags2.FLOW_CONTROL_MASK;
	}

	/**
	 * Checks if it's a privileged instruction (all CPL=0 instructions (except {@code VMCALL}) and IOPL instructions {@code IN},
	 * {@code INS}, {@code OUT}, {@code OUTS}, {@code CLI}, {@code STI})
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isPrivileged(int code) {
		int[] data = InstrInfoTable.data;
		int index = code * 2 + 1;
		return (data[index] & InfoFlags2.PRIVILEGED) != 0;
	}

	/**
	 * Checks if this is an instruction that implicitly uses the stack pointer ({@code SP}/{@code ESP}/{@code RSP}), eg.<!-- -->
	 * {@code CALL}, {@code PUSH}, {@code POP}, {@code RET}, etc.
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 *
	 * @see Instruction#getStackPointerIncrement()
	 */
	public static boolean isStackInstruction(int code) {
		int[] data = InstrInfoTable.data;
		int index = code * 2 + 1;
		return (data[index] & InfoFlags2.STACK_INSTRUCTION) != 0;
	}

	/**
	 * Checks if it's an instruction that saves or restores too many registers (eg.<!-- --> {@code FXRSTOR}, {@code XSAVE}, etc).
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isSaveRestoreInstruction(int code) {
		int[] data = InstrInfoTable.data;
		int index = code * 2 + 1;
		return (data[index] & InfoFlags2.SAVE_RESTORE) != 0;
	}

	/**
	 * Checks if it's a {@code Jcc NEAR} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isJccNear(int code) {
		return Integer.compareUnsigned(code - Code.JO_REL16, Code.JG_REL32_64 - Code.JO_REL16) <= 0;
	}

	/**
	 * Checks if it's a {@code Jcc SHORT} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isJccShort(int code) {
		return Integer.compareUnsigned(code - Code.JO_REL8_16, Code.JG_REL8_64 - Code.JO_REL8_16) <= 0;
	}

	/**
	 * Checks if it's a {@code JMP SHORT} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isJmpShort(int code) {
		return Integer.compareUnsigned(code - Code.JMP_REL8_16, Code.JMP_REL8_64 - Code.JMP_REL8_16) <= 0;
	}

	/**
	 * Checks if it's a {@code JMP NEAR} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isJmpNear(int code) {
		return Integer.compareUnsigned(code - Code.JMP_REL16, Code.JMP_REL32_64 - Code.JMP_REL16) <= 0;
	}

	/**
	 * Checks if it's a {@code JMP SHORT} or a {@code JMP NEAR} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isJmpShortOrNear(int code) {
		return Integer.compareUnsigned(code - Code.JMP_REL8_16, Code.JMP_REL8_64 - Code.JMP_REL8_16) <= 0 ||
				Integer.compareUnsigned(code - Code.JMP_REL16, Code.JMP_REL32_64 - Code.JMP_REL16) <= 0;
	}

	/**
	 * Checks if it's a {@code JMP FAR} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isJmpFar(int code) {
		return Integer.compareUnsigned(code - Code.JMP_PTR1616, Code.JMP_PTR1632 - Code.JMP_PTR1616) <= 0;
	}

	/**
	 * Checks if it's a {@code CALL NEAR} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isCallNear(int code) {
		return Integer.compareUnsigned(code - Code.CALL_REL16, Code.CALL_REL32_64 - Code.CALL_REL16) <= 0;
	}

	/**
	 * Checks if it's a {@code CALL FAR} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isCallFar(int code) {
		return Integer.compareUnsigned(code - Code.CALL_PTR1616, Code.CALL_PTR1632 - Code.CALL_PTR1616) <= 0;
	}

	/**
	 * Checks if it's a {@code JMP NEAR reg/[mem]} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isJmpNearIndirect(int code) {
		return Integer.compareUnsigned(code - Code.JMP_RM16, Code.JMP_RM64 - Code.JMP_RM16) <= 0;
	}

	/**
	 * Checks if it's a {@code JMP FAR [mem]} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isJmpFarIndirect(int code) {
		return Integer.compareUnsigned(code - Code.JMP_M1616, Code.JMP_M1664 - Code.JMP_M1616) <= 0;
	}

	/**
	 * Checks if it's a {@code CALL NEAR reg/[mem]} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isCallNearIndirect(int code) {
		return Integer.compareUnsigned(code - Code.CALL_RM16, Code.CALL_RM64 - Code.CALL_RM16) <= 0;
	}

	/**
	 * Checks if it's a {@code CALL FAR [mem]} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isCallFarIndirect(int code) {
		return Integer.compareUnsigned(code - Code.CALL_M1616, Code.CALL_M1664 - Code.CALL_M1616) <= 0;
	}

	/**
	 * Checks if it's a {@code JKccD SHORT} or {@code JKccD NEAR} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isJkccShortOrNear(int code) {
		return code == Code.VEX_KNC_JKZD_KR_REL32_64 || code == Code.VEX_KNC_JKNZD_KR_REL32_64 ||
				code == Code.VEX_KNC_JKZD_KR_REL8_64 || code == Code.VEX_KNC_JKNZD_KR_REL8_64;
	}

	/**
	 * Checks if it's a {@code JKccD NEAR} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isJkccNear(int code) {
		return code == Code.VEX_KNC_JKZD_KR_REL32_64 || code == Code.VEX_KNC_JKNZD_KR_REL32_64;
	}

	/**
	 * Checks if it's a {@code JKccD SHORT} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isJkccShort(int code) {
		return code == Code.VEX_KNC_JKZD_KR_REL8_64 || code == Code.VEX_KNC_JKNZD_KR_REL8_64;
	}

	/**
	 * Gets the condition code (a {@link ConditionCode} enum variant) if it's {@code Jcc}, {@code SETcc}, {@code CMOVcc}, {@code CMPccXADD},
	 * {@code LOOPcc} else {@link ConditionCode#NONE} is returned
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static int conditionCode(int code) {
		int t;

		if (Integer.compareUnsigned(t = code - Code.JO_REL16, Code.JG_REL32_64 - Code.JO_REL16) <= 0 ||
				Integer.compareUnsigned(t = code - Code.JO_REL8_16, Code.JG_REL8_64 - Code.JO_REL8_16) <= 0 ||
				Integer.compareUnsigned(t = code - Code.CMOVO_R16_RM16, Code.CMOVG_R64_RM64 - Code.CMOVO_R16_RM16) <= 0) {
			return t / 3 + ConditionCode.O;
		}

		t = code - Code.SETO_RM8;
		if (Integer.compareUnsigned(t, Code.SETG_RM8 - Code.SETO_RM8) <= 0)
			return t + ConditionCode.O;

		t = code - Code.LOOPNE_REL8_16_CX;
		if (Integer.compareUnsigned(t, Code.LOOPNE_REL8_64_RCX - Code.LOOPNE_REL8_16_CX) <= 0) {
			return ConditionCode.NE;
		}

		t = code - Code.LOOPE_REL8_16_CX;
		if (Integer.compareUnsigned(t, Code.LOOPE_REL8_64_RCX - Code.LOOPE_REL8_16_CX) <= 0) {
			return ConditionCode.E;
		}

		t = code - Code.VEX_CMPOXADD_M32_R32_R32;
		if (Integer.compareUnsigned(t, Code.VEX_CMPNLEXADD_M64_R64_R64 - Code.VEX_CMPOXADD_M32_R32_R32) <= 0) {
			return t / 2 + ConditionCode.O;
		}

		switch (code) {
		case Code.VEX_KNC_JKZD_KR_REL8_64:
		case Code.VEX_KNC_JKZD_KR_REL32_64:
			return ConditionCode.E;
		case Code.VEX_KNC_JKNZD_KR_REL8_64:
		case Code.VEX_KNC_JKNZD_KR_REL32_64:
			return ConditionCode.NE;
		default:
			break;
		}

		return ConditionCode.NONE;
	}

	/**
	 * Checks if it's a string instruction such as {@code MOVS}, {@code LODS}, {@code STOS}, etc.
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isStringInstruction(int code) {
		switch (code) {
		// GENERATOR-BEGIN: IsStringOpTable
		// ⚠️This was generated by GENERATOR!🦹‍♂️
		case Code.INSB_M8_DX:
		case Code.INSW_M16_DX:
		case Code.INSD_M32_DX:
		case Code.OUTSB_DX_M8:
		case Code.OUTSW_DX_M16:
		case Code.OUTSD_DX_M32:
		case Code.MOVSB_M8_M8:
		case Code.MOVSW_M16_M16:
		case Code.MOVSD_M32_M32:
		case Code.MOVSQ_M64_M64:
		case Code.CMPSB_M8_M8:
		case Code.CMPSW_M16_M16:
		case Code.CMPSD_M32_M32:
		case Code.CMPSQ_M64_M64:
		case Code.STOSB_M8_AL:
		case Code.STOSW_M16_AX:
		case Code.STOSD_M32_EAX:
		case Code.STOSQ_M64_RAX:
		case Code.LODSB_AL_M8:
		case Code.LODSW_AX_M16:
		case Code.LODSD_EAX_M32:
		case Code.LODSQ_RAX_M64:
		case Code.SCASB_AL_M8:
		case Code.SCASW_AX_M16:
		case Code.SCASD_EAX_M32:
		case Code.SCASQ_RAX_M64:
			return true;
		// GENERATOR-END: IsStringOpTable
		default:
			return false;
		}
	}

	/**
	 * Checks if it's a {@code JCXZ SHORT}, {@code JECXZ SHORT} or {@code JRCXZ SHORT} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isJcxShort(int code) {
		return Integer.compareUnsigned(code - Code.JCXZ_REL8_16, Code.JRCXZ_REL8_64 - Code.JCXZ_REL8_16) <= 0;
	}

	/**
	 * Checks if it's a {@code LOOPcc SHORT} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isLoopcc(int code) {
		return Integer.compareUnsigned(code - Code.LOOPNE_REL8_16_CX, Code.LOOPE_REL8_64_RCX - Code.LOOPNE_REL8_16_CX) <= 0;
	}

	/**
	 * Checks if it's a {@code LOOP SHORT} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isLoop(int code) {
		return Integer.compareUnsigned(code - Code.LOOP_REL8_16_CX, Code.LOOP_REL8_64_RCX - Code.LOOP_REL8_16_CX) <= 0;
	}

	/**
	 * Negates the condition code, eg.<!-- --> {@code JE} -&gt; {@code JNE}. Can be used if it's {@code Jcc}, {@code SETcc},
	 * {@code CMOVcc}, {@code CMPccXADD}, {@code LOOPcc} and returns the original value if it's none of those instructions.
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static int negateConditionCode(int code) {
		int t;

		if (Integer.compareUnsigned(t = code - Code.JO_REL16, Code.JG_REL32_64 - Code.JO_REL16) <= 0 ||
				Integer.compareUnsigned(t = code - Code.JO_REL8_16, Code.JG_REL8_64 - Code.JO_REL8_16) <= 0 ||
				Integer.compareUnsigned(t = code - Code.CMOVO_R16_RM16, Code.CMOVG_R64_RM64 - Code.CMOVO_R16_RM16) <= 0) {
			if (((t / 3) & 1) != 0)
				return code - 3;
			return code + 3;
		}

		t = code - Code.SETO_RM8;
		if (Integer.compareUnsigned(t, Code.SETG_RM8 - Code.SETO_RM8) <= 0)
			return (t ^ 1) + Code.SETO_RM8;

		t = code - Code.LOOPNE_REL8_16_CX;
		if (Integer.compareUnsigned(t, Code.LOOPE_REL8_64_RCX - Code.LOOPNE_REL8_16_CX) <= 0) {
			return Code.LOOPNE_REL8_16_CX + (t + 7) % 14;
		}

		t = code - Code.VEX_CMPOXADD_M32_R32_R32;
		if (Integer.compareUnsigned(t, Code.VEX_CMPNLEXADD_M64_R64_R64 - Code.VEX_CMPOXADD_M32_R32_R32) <= 0) {
			if ((t & 2) != 0)
				return code - 2;
			return code + 2;
		}

		switch (code) {
		case Code.VEX_KNC_JKZD_KR_REL8_64:
			return Code.VEX_KNC_JKNZD_KR_REL8_64;
		case Code.VEX_KNC_JKNZD_KR_REL8_64:
			return Code.VEX_KNC_JKZD_KR_REL8_64;
		case Code.VEX_KNC_JKZD_KR_REL32_64:
			return Code.VEX_KNC_JKNZD_KR_REL32_64;
		case Code.VEX_KNC_JKNZD_KR_REL32_64:
			return Code.VEX_KNC_JKZD_KR_REL32_64;
		default:
			break;
		}

		return code;
	}

	/**
	 * Converts {@code Jcc/JMP NEAR} to {@code Jcc/JMP SHORT}. Returns the input if it's not a {@code Jcc/JMP NEAR} instruction.
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static int toShortBranch(int code) {
		int t;

		t = code - Code.JO_REL16;
		if (Integer.compareUnsigned(t, Code.JG_REL32_64 - Code.JO_REL16) <= 0)
			return t + Code.JO_REL8_16;

		t = code - Code.JMP_REL16;
		if (Integer.compareUnsigned(t, Code.JMP_REL32_64 - Code.JMP_REL16) <= 0)
			return t + Code.JMP_REL8_16;

		switch (code) {
		case Code.VEX_KNC_JKZD_KR_REL32_64:
			return Code.VEX_KNC_JKZD_KR_REL8_64;
		case Code.VEX_KNC_JKNZD_KR_REL32_64:
			return Code.VEX_KNC_JKNZD_KR_REL8_64;
		default:
			break;
		}

		return code;
	}

	/**
	 * Converts {@code Jcc/JMP SHORT} to {@code Jcc/JMP NEAR}. Returns the input if it's not a {@code Jcc/JMP SHORT} instruction.
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static int toNearBranch(int code) {
		int t;

		t = code - Code.JO_REL8_16;
		if (Integer.compareUnsigned(t, Code.JG_REL8_64 - Code.JO_REL8_16) <= 0)
			return t + Code.JO_REL16;

		t = code - Code.JMP_REL8_16;
		if (Integer.compareUnsigned(t, Code.JMP_REL8_64 - Code.JMP_REL8_16) <= 0)
			return t + Code.JMP_REL16;

		switch (code) {
		case Code.VEX_KNC_JKZD_KR_REL8_64:
			return Code.VEX_KNC_JKZD_KR_REL32_64;
		case Code.VEX_KNC_JKNZD_KR_REL8_64:
			return Code.VEX_KNC_JKNZD_KR_REL32_64;
		default:
			break;
		}

		return code;
	}

	/**
	 * Checks if it's a {@code Jcc SHORT} or {@code Jcc NEAR} instruction
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static boolean isJccShortOrNear(int code) {
		return Integer.compareUnsigned(code - Code.JO_REL8_16, Code.JG_REL8_64 - Code.JO_REL8_16) <= 0 ||
				Integer.compareUnsigned(code - Code.JO_REL16, Code.JG_REL32_64 - Code.JO_REL16) <= 0;
	}

	/**
	 * Gets the mnemonic (a {@link Mnemonic} enum variant)
	 *
	 * @param code Code value (a {@link Code} enum variant)
	 */
	public static int mnemonic(int code) {
		return MnemonicUtilsData.toMnemonic[code] & 0xFFFF;
	}

	// GENERATOR-BEGIN: Variants
	// ⚠️This was generated by GENERATOR!🦹‍♂️
	/**
	 * It's an invalid instruction, eg.<!-- --> it's a new unknown instruction, garbage or there's not enough bytes to decode the instruction etc.<!-- -->
	 */
	public static final int INVALID = 0;
	/**
	 * A {@code db}/{@code .byte} asm directive that can store 1-16 bytes
	 */
	public static final int DECLAREBYTE = 1;
	/**
	 * A {@code dw}/{@code .word} asm directive that can store 1-8 words
	 */
	public static final int DECLAREWORD = 2;
	/**
	 * A {@code dd}/{@code .int} asm directive that can store 1-4 dwords
	 */
	public static final int DECLAREDWORD = 3;
	/**
	 * A {@code dq}/{@code .quad} asm directive that can store 1-2 qwords
	 */
	public static final int DECLAREQWORD = 4;
	/**
	 * {@code ADD r/m8, r8}<p>
	 * {@code 00 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADD_RM8_R8 = 5;
	/**
	 * {@code ADD r/m16, r16}<p>
	 * {@code o16 01 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADD_RM16_R16 = 6;
	/**
	 * {@code ADD r/m32, r32}<p>
	 * {@code o32 01 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADD_RM32_R32 = 7;
	/**
	 * {@code ADD r/m64, r64}<p>
	 * {@code o64 01 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ADD_RM64_R64 = 8;
	/**
	 * {@code ADD r8, r/m8}<p>
	 * {@code 02 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADD_R8_RM8 = 9;
	/**
	 * {@code ADD r16, r/m16}<p>
	 * {@code o16 03 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADD_R16_RM16 = 10;
	/**
	 * {@code ADD r32, r/m32}<p>
	 * {@code o32 03 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADD_R32_RM32 = 11;
	/**
	 * {@code ADD r64, r/m64}<p>
	 * {@code o64 03 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ADD_R64_RM64 = 12;
	/**
	 * {@code ADD AL, imm8}<p>
	 * {@code 04 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADD_AL_IMM8 = 13;
	/**
	 * {@code ADD AX, imm16}<p>
	 * {@code o16 05 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADD_AX_IMM16 = 14;
	/**
	 * {@code ADD EAX, imm32}<p>
	 * {@code o32 05 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADD_EAX_IMM32 = 15;
	/**
	 * {@code ADD RAX, imm32}<p>
	 * {@code o64 05 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ADD_RAX_IMM32 = 16;
	/**
	 * {@code PUSH ES}<p>
	 * {@code o16 06}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSHW_ES = 17;
	/**
	 * {@code PUSH ES}<p>
	 * {@code o32 06}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSHD_ES = 18;
	/**
	 * {@code POP ES}<p>
	 * {@code o16 07}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int POPW_ES = 19;
	/**
	 * {@code POP ES}<p>
	 * {@code o32 07}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int POPD_ES = 20;
	/**
	 * {@code OR r/m8, r8}<p>
	 * {@code 08 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OR_RM8_R8 = 21;
	/**
	 * {@code OR r/m16, r16}<p>
	 * {@code o16 09 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OR_RM16_R16 = 22;
	/**
	 * {@code OR r/m32, r32}<p>
	 * {@code o32 09 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OR_RM32_R32 = 23;
	/**
	 * {@code OR r/m64, r64}<p>
	 * {@code o64 09 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int OR_RM64_R64 = 24;
	/**
	 * {@code OR r8, r/m8}<p>
	 * {@code 0A /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OR_R8_RM8 = 25;
	/**
	 * {@code OR r16, r/m16}<p>
	 * {@code o16 0B /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OR_R16_RM16 = 26;
	/**
	 * {@code OR r32, r/m32}<p>
	 * {@code o32 0B /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OR_R32_RM32 = 27;
	/**
	 * {@code OR r64, r/m64}<p>
	 * {@code o64 0B /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int OR_R64_RM64 = 28;
	/**
	 * {@code OR AL, imm8}<p>
	 * {@code 0C ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OR_AL_IMM8 = 29;
	/**
	 * {@code OR AX, imm16}<p>
	 * {@code o16 0D iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OR_AX_IMM16 = 30;
	/**
	 * {@code OR EAX, imm32}<p>
	 * {@code o32 0D id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OR_EAX_IMM32 = 31;
	/**
	 * {@code OR RAX, imm32}<p>
	 * {@code o64 0D id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int OR_RAX_IMM32 = 32;
	/**
	 * {@code PUSH CS}<p>
	 * {@code o16 0E}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSHW_CS = 33;
	/**
	 * {@code PUSH CS}<p>
	 * {@code o32 0E}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSHD_CS = 34;
	/**
	 * {@code POP CS}<p>
	 * {@code o16 0F}<p>
	 * {@code 8086}<p>
	 * {@code 16-bit}
	 */
	public static final int POPW_CS = 35;
	/**
	 * {@code ADC r/m8, r8}<p>
	 * {@code 10 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADC_RM8_R8 = 36;
	/**
	 * {@code ADC r/m16, r16}<p>
	 * {@code o16 11 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADC_RM16_R16 = 37;
	/**
	 * {@code ADC r/m32, r32}<p>
	 * {@code o32 11 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADC_RM32_R32 = 38;
	/**
	 * {@code ADC r/m64, r64}<p>
	 * {@code o64 11 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ADC_RM64_R64 = 39;
	/**
	 * {@code ADC r8, r/m8}<p>
	 * {@code 12 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADC_R8_RM8 = 40;
	/**
	 * {@code ADC r16, r/m16}<p>
	 * {@code o16 13 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADC_R16_RM16 = 41;
	/**
	 * {@code ADC r32, r/m32}<p>
	 * {@code o32 13 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADC_R32_RM32 = 42;
	/**
	 * {@code ADC r64, r/m64}<p>
	 * {@code o64 13 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ADC_R64_RM64 = 43;
	/**
	 * {@code ADC AL, imm8}<p>
	 * {@code 14 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADC_AL_IMM8 = 44;
	/**
	 * {@code ADC AX, imm16}<p>
	 * {@code o16 15 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADC_AX_IMM16 = 45;
	/**
	 * {@code ADC EAX, imm32}<p>
	 * {@code o32 15 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADC_EAX_IMM32 = 46;
	/**
	 * {@code ADC RAX, imm32}<p>
	 * {@code o64 15 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ADC_RAX_IMM32 = 47;
	/**
	 * {@code PUSH SS}<p>
	 * {@code o16 16}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSHW_SS = 48;
	/**
	 * {@code PUSH SS}<p>
	 * {@code o32 16}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSHD_SS = 49;
	/**
	 * {@code POP SS}<p>
	 * {@code o16 17}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int POPW_SS = 50;
	/**
	 * {@code POP SS}<p>
	 * {@code o32 17}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int POPD_SS = 51;
	/**
	 * {@code SBB r/m8, r8}<p>
	 * {@code 18 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SBB_RM8_R8 = 52;
	/**
	 * {@code SBB r/m16, r16}<p>
	 * {@code o16 19 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SBB_RM16_R16 = 53;
	/**
	 * {@code SBB r/m32, r32}<p>
	 * {@code o32 19 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SBB_RM32_R32 = 54;
	/**
	 * {@code SBB r/m64, r64}<p>
	 * {@code o64 19 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SBB_RM64_R64 = 55;
	/**
	 * {@code SBB r8, r/m8}<p>
	 * {@code 1A /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SBB_R8_RM8 = 56;
	/**
	 * {@code SBB r16, r/m16}<p>
	 * {@code o16 1B /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SBB_R16_RM16 = 57;
	/**
	 * {@code SBB r32, r/m32}<p>
	 * {@code o32 1B /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SBB_R32_RM32 = 58;
	/**
	 * {@code SBB r64, r/m64}<p>
	 * {@code o64 1B /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SBB_R64_RM64 = 59;
	/**
	 * {@code SBB AL, imm8}<p>
	 * {@code 1C ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SBB_AL_IMM8 = 60;
	/**
	 * {@code SBB AX, imm16}<p>
	 * {@code o16 1D iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SBB_AX_IMM16 = 61;
	/**
	 * {@code SBB EAX, imm32}<p>
	 * {@code o32 1D id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SBB_EAX_IMM32 = 62;
	/**
	 * {@code SBB RAX, imm32}<p>
	 * {@code o64 1D id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SBB_RAX_IMM32 = 63;
	/**
	 * {@code PUSH DS}<p>
	 * {@code o16 1E}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSHW_DS = 64;
	/**
	 * {@code PUSH DS}<p>
	 * {@code o32 1E}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSHD_DS = 65;
	/**
	 * {@code POP DS}<p>
	 * {@code o16 1F}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int POPW_DS = 66;
	/**
	 * {@code POP DS}<p>
	 * {@code o32 1F}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int POPD_DS = 67;
	/**
	 * {@code AND r/m8, r8}<p>
	 * {@code 20 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AND_RM8_R8 = 68;
	/**
	 * {@code AND r/m16, r16}<p>
	 * {@code o16 21 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AND_RM16_R16 = 69;
	/**
	 * {@code AND r/m32, r32}<p>
	 * {@code o32 21 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AND_RM32_R32 = 70;
	/**
	 * {@code AND r/m64, r64}<p>
	 * {@code o64 21 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int AND_RM64_R64 = 71;
	/**
	 * {@code AND r8, r/m8}<p>
	 * {@code 22 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AND_R8_RM8 = 72;
	/**
	 * {@code AND r16, r/m16}<p>
	 * {@code o16 23 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AND_R16_RM16 = 73;
	/**
	 * {@code AND r32, r/m32}<p>
	 * {@code o32 23 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AND_R32_RM32 = 74;
	/**
	 * {@code AND r64, r/m64}<p>
	 * {@code o64 23 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int AND_R64_RM64 = 75;
	/**
	 * {@code AND AL, imm8}<p>
	 * {@code 24 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AND_AL_IMM8 = 76;
	/**
	 * {@code AND AX, imm16}<p>
	 * {@code o16 25 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AND_AX_IMM16 = 77;
	/**
	 * {@code AND EAX, imm32}<p>
	 * {@code o32 25 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AND_EAX_IMM32 = 78;
	/**
	 * {@code AND RAX, imm32}<p>
	 * {@code o64 25 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int AND_RAX_IMM32 = 79;
	/**
	 * {@code DAA}<p>
	 * {@code 27}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int DAA = 80;
	/**
	 * {@code SUB r/m8, r8}<p>
	 * {@code 28 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUB_RM8_R8 = 81;
	/**
	 * {@code SUB r/m16, r16}<p>
	 * {@code o16 29 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUB_RM16_R16 = 82;
	/**
	 * {@code SUB r/m32, r32}<p>
	 * {@code o32 29 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUB_RM32_R32 = 83;
	/**
	 * {@code SUB r/m64, r64}<p>
	 * {@code o64 29 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SUB_RM64_R64 = 84;
	/**
	 * {@code SUB r8, r/m8}<p>
	 * {@code 2A /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUB_R8_RM8 = 85;
	/**
	 * {@code SUB r16, r/m16}<p>
	 * {@code o16 2B /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUB_R16_RM16 = 86;
	/**
	 * {@code SUB r32, r/m32}<p>
	 * {@code o32 2B /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUB_R32_RM32 = 87;
	/**
	 * {@code SUB r64, r/m64}<p>
	 * {@code o64 2B /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SUB_R64_RM64 = 88;
	/**
	 * {@code SUB AL, imm8}<p>
	 * {@code 2C ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUB_AL_IMM8 = 89;
	/**
	 * {@code SUB AX, imm16}<p>
	 * {@code o16 2D iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUB_AX_IMM16 = 90;
	/**
	 * {@code SUB EAX, imm32}<p>
	 * {@code o32 2D id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUB_EAX_IMM32 = 91;
	/**
	 * {@code SUB RAX, imm32}<p>
	 * {@code o64 2D id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SUB_RAX_IMM32 = 92;
	/**
	 * {@code DAS}<p>
	 * {@code 2F}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int DAS = 93;
	/**
	 * {@code XOR r/m8, r8}<p>
	 * {@code 30 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOR_RM8_R8 = 94;
	/**
	 * {@code XOR r/m16, r16}<p>
	 * {@code o16 31 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOR_RM16_R16 = 95;
	/**
	 * {@code XOR r/m32, r32}<p>
	 * {@code o32 31 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOR_RM32_R32 = 96;
	/**
	 * {@code XOR r/m64, r64}<p>
	 * {@code o64 31 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int XOR_RM64_R64 = 97;
	/**
	 * {@code XOR r8, r/m8}<p>
	 * {@code 32 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOR_R8_RM8 = 98;
	/**
	 * {@code XOR r16, r/m16}<p>
	 * {@code o16 33 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOR_R16_RM16 = 99;
	/**
	 * {@code XOR r32, r/m32}<p>
	 * {@code o32 33 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOR_R32_RM32 = 100;
	/**
	 * {@code XOR r64, r/m64}<p>
	 * {@code o64 33 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int XOR_R64_RM64 = 101;
	/**
	 * {@code XOR AL, imm8}<p>
	 * {@code 34 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOR_AL_IMM8 = 102;
	/**
	 * {@code XOR AX, imm16}<p>
	 * {@code o16 35 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOR_AX_IMM16 = 103;
	/**
	 * {@code XOR EAX, imm32}<p>
	 * {@code o32 35 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOR_EAX_IMM32 = 104;
	/**
	 * {@code XOR RAX, imm32}<p>
	 * {@code o64 35 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int XOR_RAX_IMM32 = 105;
	/**
	 * {@code AAA}<p>
	 * {@code 37}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int AAA = 106;
	/**
	 * {@code CMP r/m8, r8}<p>
	 * {@code 38 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMP_RM8_R8 = 107;
	/**
	 * {@code CMP r/m16, r16}<p>
	 * {@code o16 39 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMP_RM16_R16 = 108;
	/**
	 * {@code CMP r/m32, r32}<p>
	 * {@code o32 39 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMP_RM32_R32 = 109;
	/**
	 * {@code CMP r/m64, r64}<p>
	 * {@code o64 39 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int CMP_RM64_R64 = 110;
	/**
	 * {@code CMP r8, r/m8}<p>
	 * {@code 3A /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMP_R8_RM8 = 111;
	/**
	 * {@code CMP r16, r/m16}<p>
	 * {@code o16 3B /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMP_R16_RM16 = 112;
	/**
	 * {@code CMP r32, r/m32}<p>
	 * {@code o32 3B /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMP_R32_RM32 = 113;
	/**
	 * {@code CMP r64, r/m64}<p>
	 * {@code o64 3B /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int CMP_R64_RM64 = 114;
	/**
	 * {@code CMP AL, imm8}<p>
	 * {@code 3C ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMP_AL_IMM8 = 115;
	/**
	 * {@code CMP AX, imm16}<p>
	 * {@code o16 3D iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMP_AX_IMM16 = 116;
	/**
	 * {@code CMP EAX, imm32}<p>
	 * {@code o32 3D id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMP_EAX_IMM32 = 117;
	/**
	 * {@code CMP RAX, imm32}<p>
	 * {@code o64 3D id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int CMP_RAX_IMM32 = 118;
	/**
	 * {@code AAS}<p>
	 * {@code 3F}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int AAS = 119;
	/**
	 * {@code INC r16}<p>
	 * {@code o16 40+rw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int INC_R16 = 120;
	/**
	 * {@code INC r32}<p>
	 * {@code o32 40+rd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int INC_R32 = 121;
	/**
	 * {@code DEC r16}<p>
	 * {@code o16 48+rw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int DEC_R16 = 122;
	/**
	 * {@code DEC r32}<p>
	 * {@code o32 48+rd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int DEC_R32 = 123;
	/**
	 * {@code PUSH r16}<p>
	 * {@code o16 50+rw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUSH_R16 = 124;
	/**
	 * {@code PUSH r32}<p>
	 * {@code o32 50+rd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSH_R32 = 125;
	/**
	 * {@code PUSH r64}<p>
	 * {@code o64 50+ro}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int PUSH_R64 = 126;
	/**
	 * {@code POP r16}<p>
	 * {@code o16 58+rw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int POP_R16 = 127;
	/**
	 * {@code POP r32}<p>
	 * {@code o32 58+rd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int POP_R32 = 128;
	/**
	 * {@code POP r64}<p>
	 * {@code o64 58+ro}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int POP_R64 = 129;
	/**
	 * {@code PUSHA}<p>
	 * {@code o16 60}<p>
	 * {@code 186+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSHAW = 130;
	/**
	 * {@code PUSHAD}<p>
	 * {@code o32 60}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSHAD = 131;
	/**
	 * {@code POPA}<p>
	 * {@code o16 61}<p>
	 * {@code 186+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int POPAW = 132;
	/**
	 * {@code POPAD}<p>
	 * {@code o32 61}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int POPAD = 133;
	/**
	 * {@code BOUND r16, m16&amp;16}<p>
	 * {@code o16 62 /r}<p>
	 * {@code 186+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int BOUND_R16_M1616 = 134;
	/**
	 * {@code BOUND r32, m32&amp;32}<p>
	 * {@code o32 62 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int BOUND_R32_M3232 = 135;
	/**
	 * {@code ARPL r/m16, r16}<p>
	 * {@code o16 63 /r}<p>
	 * {@code 286+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int ARPL_RM16_R16 = 136;
	/**
	 * {@code ARPL r32/m16, r32}<p>
	 * {@code o32 63 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int ARPL_R32M16_R32 = 137;
	/**
	 * {@code MOVSXD r16, r/m16}<p>
	 * {@code o16 63 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVSXD_R16_RM16 = 138;
	/**
	 * {@code MOVSXD r32, r/m32}<p>
	 * {@code o32 63 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVSXD_R32_RM32 = 139;
	/**
	 * {@code MOVSXD r64, r/m32}<p>
	 * {@code o64 63 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVSXD_R64_RM32 = 140;
	/**
	 * {@code PUSH imm16}<p>
	 * {@code o16 68 iw}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUSH_IMM16 = 141;
	/**
	 * {@code PUSH imm32}<p>
	 * {@code o32 68 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSHD_IMM32 = 142;
	/**
	 * {@code PUSH imm32}<p>
	 * {@code o64 68 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int PUSHQ_IMM32 = 143;
	/**
	 * {@code IMUL r16, r/m16, imm16}<p>
	 * {@code o16 69 /r iw}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IMUL_R16_RM16_IMM16 = 144;
	/**
	 * {@code IMUL r32, r/m32, imm32}<p>
	 * {@code o32 69 /r id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IMUL_R32_RM32_IMM32 = 145;
	/**
	 * {@code IMUL r64, r/m64, imm32}<p>
	 * {@code o64 69 /r id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int IMUL_R64_RM64_IMM32 = 146;
	/**
	 * {@code PUSH imm8}<p>
	 * {@code o16 6A ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUSHW_IMM8 = 147;
	/**
	 * {@code PUSH imm8}<p>
	 * {@code o32 6A ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSHD_IMM8 = 148;
	/**
	 * {@code PUSH imm8}<p>
	 * {@code o64 6A ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int PUSHQ_IMM8 = 149;
	/**
	 * {@code IMUL r16, r/m16, imm8}<p>
	 * {@code o16 6B /r ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IMUL_R16_RM16_IMM8 = 150;
	/**
	 * {@code IMUL r32, r/m32, imm8}<p>
	 * {@code o32 6B /r ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IMUL_R32_RM32_IMM8 = 151;
	/**
	 * {@code IMUL r64, r/m64, imm8}<p>
	 * {@code o64 6B /r ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int IMUL_R64_RM64_IMM8 = 152;
	/**
	 * {@code INSB}<p>
	 * {@code 6C}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INSB_M8_DX = 153;
	/**
	 * {@code INSW}<p>
	 * {@code o16 6D}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INSW_M16_DX = 154;
	/**
	 * {@code INSD}<p>
	 * {@code o32 6D}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INSD_M32_DX = 155;
	/**
	 * {@code OUTSB}<p>
	 * {@code 6E}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OUTSB_DX_M8 = 156;
	/**
	 * {@code OUTSW}<p>
	 * {@code o16 6F}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OUTSW_DX_M16 = 157;
	/**
	 * {@code OUTSD}<p>
	 * {@code o32 6F}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OUTSD_DX_M32 = 158;
	/**
	 * {@code JO rel8}<p>
	 * {@code o16 70 cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JO_REL8_16 = 159;
	/**
	 * {@code JO rel8}<p>
	 * {@code o32 70 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JO_REL8_32 = 160;
	/**
	 * {@code JO rel8}<p>
	 * {@code o64 70 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JO_REL8_64 = 161;
	/**
	 * {@code JNO rel8}<p>
	 * {@code o16 71 cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JNO_REL8_16 = 162;
	/**
	 * {@code JNO rel8}<p>
	 * {@code o32 71 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JNO_REL8_32 = 163;
	/**
	 * {@code JNO rel8}<p>
	 * {@code o64 71 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JNO_REL8_64 = 164;
	/**
	 * {@code JB rel8}<p>
	 * {@code o16 72 cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JB_REL8_16 = 165;
	/**
	 * {@code JB rel8}<p>
	 * {@code o32 72 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JB_REL8_32 = 166;
	/**
	 * {@code JB rel8}<p>
	 * {@code o64 72 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JB_REL8_64 = 167;
	/**
	 * {@code JAE rel8}<p>
	 * {@code o16 73 cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JAE_REL8_16 = 168;
	/**
	 * {@code JAE rel8}<p>
	 * {@code o32 73 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JAE_REL8_32 = 169;
	/**
	 * {@code JAE rel8}<p>
	 * {@code o64 73 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JAE_REL8_64 = 170;
	/**
	 * {@code JE rel8}<p>
	 * {@code o16 74 cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JE_REL8_16 = 171;
	/**
	 * {@code JE rel8}<p>
	 * {@code o32 74 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JE_REL8_32 = 172;
	/**
	 * {@code JE rel8}<p>
	 * {@code o64 74 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JE_REL8_64 = 173;
	/**
	 * {@code JNE rel8}<p>
	 * {@code o16 75 cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JNE_REL8_16 = 174;
	/**
	 * {@code JNE rel8}<p>
	 * {@code o32 75 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JNE_REL8_32 = 175;
	/**
	 * {@code JNE rel8}<p>
	 * {@code o64 75 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JNE_REL8_64 = 176;
	/**
	 * {@code JBE rel8}<p>
	 * {@code o16 76 cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JBE_REL8_16 = 177;
	/**
	 * {@code JBE rel8}<p>
	 * {@code o32 76 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JBE_REL8_32 = 178;
	/**
	 * {@code JBE rel8}<p>
	 * {@code o64 76 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JBE_REL8_64 = 179;
	/**
	 * {@code JA rel8}<p>
	 * {@code o16 77 cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JA_REL8_16 = 180;
	/**
	 * {@code JA rel8}<p>
	 * {@code o32 77 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JA_REL8_32 = 181;
	/**
	 * {@code JA rel8}<p>
	 * {@code o64 77 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JA_REL8_64 = 182;
	/**
	 * {@code JS rel8}<p>
	 * {@code o16 78 cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JS_REL8_16 = 183;
	/**
	 * {@code JS rel8}<p>
	 * {@code o32 78 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JS_REL8_32 = 184;
	/**
	 * {@code JS rel8}<p>
	 * {@code o64 78 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JS_REL8_64 = 185;
	/**
	 * {@code JNS rel8}<p>
	 * {@code o16 79 cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JNS_REL8_16 = 186;
	/**
	 * {@code JNS rel8}<p>
	 * {@code o32 79 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JNS_REL8_32 = 187;
	/**
	 * {@code JNS rel8}<p>
	 * {@code o64 79 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JNS_REL8_64 = 188;
	/**
	 * {@code JP rel8}<p>
	 * {@code o16 7A cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JP_REL8_16 = 189;
	/**
	 * {@code JP rel8}<p>
	 * {@code o32 7A cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JP_REL8_32 = 190;
	/**
	 * {@code JP rel8}<p>
	 * {@code o64 7A cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JP_REL8_64 = 191;
	/**
	 * {@code JNP rel8}<p>
	 * {@code o16 7B cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JNP_REL8_16 = 192;
	/**
	 * {@code JNP rel8}<p>
	 * {@code o32 7B cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JNP_REL8_32 = 193;
	/**
	 * {@code JNP rel8}<p>
	 * {@code o64 7B cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JNP_REL8_64 = 194;
	/**
	 * {@code JL rel8}<p>
	 * {@code o16 7C cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JL_REL8_16 = 195;
	/**
	 * {@code JL rel8}<p>
	 * {@code o32 7C cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JL_REL8_32 = 196;
	/**
	 * {@code JL rel8}<p>
	 * {@code o64 7C cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JL_REL8_64 = 197;
	/**
	 * {@code JGE rel8}<p>
	 * {@code o16 7D cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JGE_REL8_16 = 198;
	/**
	 * {@code JGE rel8}<p>
	 * {@code o32 7D cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JGE_REL8_32 = 199;
	/**
	 * {@code JGE rel8}<p>
	 * {@code o64 7D cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JGE_REL8_64 = 200;
	/**
	 * {@code JLE rel8}<p>
	 * {@code o16 7E cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JLE_REL8_16 = 201;
	/**
	 * {@code JLE rel8}<p>
	 * {@code o32 7E cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JLE_REL8_32 = 202;
	/**
	 * {@code JLE rel8}<p>
	 * {@code o64 7E cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JLE_REL8_64 = 203;
	/**
	 * {@code JG rel8}<p>
	 * {@code o16 7F cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JG_REL8_16 = 204;
	/**
	 * {@code JG rel8}<p>
	 * {@code o32 7F cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JG_REL8_32 = 205;
	/**
	 * {@code JG rel8}<p>
	 * {@code o64 7F cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JG_REL8_64 = 206;
	/**
	 * {@code ADD r/m8, imm8}<p>
	 * {@code 80 /0 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADD_RM8_IMM8 = 207;
	/**
	 * {@code OR r/m8, imm8}<p>
	 * {@code 80 /1 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OR_RM8_IMM8 = 208;
	/**
	 * {@code ADC r/m8, imm8}<p>
	 * {@code 80 /2 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADC_RM8_IMM8 = 209;
	/**
	 * {@code SBB r/m8, imm8}<p>
	 * {@code 80 /3 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SBB_RM8_IMM8 = 210;
	/**
	 * {@code AND r/m8, imm8}<p>
	 * {@code 80 /4 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AND_RM8_IMM8 = 211;
	/**
	 * {@code SUB r/m8, imm8}<p>
	 * {@code 80 /5 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUB_RM8_IMM8 = 212;
	/**
	 * {@code XOR r/m8, imm8}<p>
	 * {@code 80 /6 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOR_RM8_IMM8 = 213;
	/**
	 * {@code CMP r/m8, imm8}<p>
	 * {@code 80 /7 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMP_RM8_IMM8 = 214;
	/**
	 * {@code ADD r/m16, imm16}<p>
	 * {@code o16 81 /0 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADD_RM16_IMM16 = 215;
	/**
	 * {@code ADD r/m32, imm32}<p>
	 * {@code o32 81 /0 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADD_RM32_IMM32 = 216;
	/**
	 * {@code ADD r/m64, imm32}<p>
	 * {@code o64 81 /0 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ADD_RM64_IMM32 = 217;
	/**
	 * {@code OR r/m16, imm16}<p>
	 * {@code o16 81 /1 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OR_RM16_IMM16 = 218;
	/**
	 * {@code OR r/m32, imm32}<p>
	 * {@code o32 81 /1 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OR_RM32_IMM32 = 219;
	/**
	 * {@code OR r/m64, imm32}<p>
	 * {@code o64 81 /1 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int OR_RM64_IMM32 = 220;
	/**
	 * {@code ADC r/m16, imm16}<p>
	 * {@code o16 81 /2 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADC_RM16_IMM16 = 221;
	/**
	 * {@code ADC r/m32, imm32}<p>
	 * {@code o32 81 /2 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADC_RM32_IMM32 = 222;
	/**
	 * {@code ADC r/m64, imm32}<p>
	 * {@code o64 81 /2 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ADC_RM64_IMM32 = 223;
	/**
	 * {@code SBB r/m16, imm16}<p>
	 * {@code o16 81 /3 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SBB_RM16_IMM16 = 224;
	/**
	 * {@code SBB r/m32, imm32}<p>
	 * {@code o32 81 /3 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SBB_RM32_IMM32 = 225;
	/**
	 * {@code SBB r/m64, imm32}<p>
	 * {@code o64 81 /3 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SBB_RM64_IMM32 = 226;
	/**
	 * {@code AND r/m16, imm16}<p>
	 * {@code o16 81 /4 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AND_RM16_IMM16 = 227;
	/**
	 * {@code AND r/m32, imm32}<p>
	 * {@code o32 81 /4 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AND_RM32_IMM32 = 228;
	/**
	 * {@code AND r/m64, imm32}<p>
	 * {@code o64 81 /4 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int AND_RM64_IMM32 = 229;
	/**
	 * {@code SUB r/m16, imm16}<p>
	 * {@code o16 81 /5 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUB_RM16_IMM16 = 230;
	/**
	 * {@code SUB r/m32, imm32}<p>
	 * {@code o32 81 /5 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUB_RM32_IMM32 = 231;
	/**
	 * {@code SUB r/m64, imm32}<p>
	 * {@code o64 81 /5 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SUB_RM64_IMM32 = 232;
	/**
	 * {@code XOR r/m16, imm16}<p>
	 * {@code o16 81 /6 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOR_RM16_IMM16 = 233;
	/**
	 * {@code XOR r/m32, imm32}<p>
	 * {@code o32 81 /6 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOR_RM32_IMM32 = 234;
	/**
	 * {@code XOR r/m64, imm32}<p>
	 * {@code o64 81 /6 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int XOR_RM64_IMM32 = 235;
	/**
	 * {@code CMP r/m16, imm16}<p>
	 * {@code o16 81 /7 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMP_RM16_IMM16 = 236;
	/**
	 * {@code CMP r/m32, imm32}<p>
	 * {@code o32 81 /7 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMP_RM32_IMM32 = 237;
	/**
	 * {@code CMP r/m64, imm32}<p>
	 * {@code o64 81 /7 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int CMP_RM64_IMM32 = 238;
	/**
	 * {@code ADD r/m8, imm8}<p>
	 * {@code 82 /0 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int ADD_RM8_IMM8_82 = 239;
	/**
	 * {@code OR r/m8, imm8}<p>
	 * {@code 82 /1 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int OR_RM8_IMM8_82 = 240;
	/**
	 * {@code ADC r/m8, imm8}<p>
	 * {@code 82 /2 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int ADC_RM8_IMM8_82 = 241;
	/**
	 * {@code SBB r/m8, imm8}<p>
	 * {@code 82 /3 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int SBB_RM8_IMM8_82 = 242;
	/**
	 * {@code AND r/m8, imm8}<p>
	 * {@code 82 /4 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int AND_RM8_IMM8_82 = 243;
	/**
	 * {@code SUB r/m8, imm8}<p>
	 * {@code 82 /5 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int SUB_RM8_IMM8_82 = 244;
	/**
	 * {@code XOR r/m8, imm8}<p>
	 * {@code 82 /6 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int XOR_RM8_IMM8_82 = 245;
	/**
	 * {@code CMP r/m8, imm8}<p>
	 * {@code 82 /7 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CMP_RM8_IMM8_82 = 246;
	/**
	 * {@code ADD r/m16, imm8}<p>
	 * {@code o16 83 /0 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADD_RM16_IMM8 = 247;
	/**
	 * {@code ADD r/m32, imm8}<p>
	 * {@code o32 83 /0 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADD_RM32_IMM8 = 248;
	/**
	 * {@code ADD r/m64, imm8}<p>
	 * {@code o64 83 /0 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ADD_RM64_IMM8 = 249;
	/**
	 * {@code OR r/m16, imm8}<p>
	 * {@code o16 83 /1 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OR_RM16_IMM8 = 250;
	/**
	 * {@code OR r/m32, imm8}<p>
	 * {@code o32 83 /1 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OR_RM32_IMM8 = 251;
	/**
	 * {@code OR r/m64, imm8}<p>
	 * {@code o64 83 /1 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int OR_RM64_IMM8 = 252;
	/**
	 * {@code ADC r/m16, imm8}<p>
	 * {@code o16 83 /2 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADC_RM16_IMM8 = 253;
	/**
	 * {@code ADC r/m32, imm8}<p>
	 * {@code o32 83 /2 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADC_RM32_IMM8 = 254;
	/**
	 * {@code ADC r/m64, imm8}<p>
	 * {@code o64 83 /2 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ADC_RM64_IMM8 = 255;
	/**
	 * {@code SBB r/m16, imm8}<p>
	 * {@code o16 83 /3 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SBB_RM16_IMM8 = 256;
	/**
	 * {@code SBB r/m32, imm8}<p>
	 * {@code o32 83 /3 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SBB_RM32_IMM8 = 257;
	/**
	 * {@code SBB r/m64, imm8}<p>
	 * {@code o64 83 /3 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SBB_RM64_IMM8 = 258;
	/**
	 * {@code AND r/m16, imm8}<p>
	 * {@code o16 83 /4 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AND_RM16_IMM8 = 259;
	/**
	 * {@code AND r/m32, imm8}<p>
	 * {@code o32 83 /4 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AND_RM32_IMM8 = 260;
	/**
	 * {@code AND r/m64, imm8}<p>
	 * {@code o64 83 /4 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int AND_RM64_IMM8 = 261;
	/**
	 * {@code SUB r/m16, imm8}<p>
	 * {@code o16 83 /5 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUB_RM16_IMM8 = 262;
	/**
	 * {@code SUB r/m32, imm8}<p>
	 * {@code o32 83 /5 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUB_RM32_IMM8 = 263;
	/**
	 * {@code SUB r/m64, imm8}<p>
	 * {@code o64 83 /5 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SUB_RM64_IMM8 = 264;
	/**
	 * {@code XOR r/m16, imm8}<p>
	 * {@code o16 83 /6 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOR_RM16_IMM8 = 265;
	/**
	 * {@code XOR r/m32, imm8}<p>
	 * {@code o32 83 /6 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOR_RM32_IMM8 = 266;
	/**
	 * {@code XOR r/m64, imm8}<p>
	 * {@code o64 83 /6 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int XOR_RM64_IMM8 = 267;
	/**
	 * {@code CMP r/m16, imm8}<p>
	 * {@code o16 83 /7 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMP_RM16_IMM8 = 268;
	/**
	 * {@code CMP r/m32, imm8}<p>
	 * {@code o32 83 /7 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMP_RM32_IMM8 = 269;
	/**
	 * {@code CMP r/m64, imm8}<p>
	 * {@code o64 83 /7 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int CMP_RM64_IMM8 = 270;
	/**
	 * {@code TEST r/m8, r8}<p>
	 * {@code 84 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TEST_RM8_R8 = 271;
	/**
	 * {@code TEST r/m16, r16}<p>
	 * {@code o16 85 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TEST_RM16_R16 = 272;
	/**
	 * {@code TEST r/m32, r32}<p>
	 * {@code o32 85 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TEST_RM32_R32 = 273;
	/**
	 * {@code TEST r/m64, r64}<p>
	 * {@code o64 85 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int TEST_RM64_R64 = 274;
	/**
	 * {@code XCHG r/m8, r8}<p>
	 * {@code 86 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XCHG_RM8_R8 = 275;
	/**
	 * {@code XCHG r/m16, r16}<p>
	 * {@code o16 87 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XCHG_RM16_R16 = 276;
	/**
	 * {@code XCHG r/m32, r32}<p>
	 * {@code o32 87 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XCHG_RM32_R32 = 277;
	/**
	 * {@code XCHG r/m64, r64}<p>
	 * {@code o64 87 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int XCHG_RM64_R64 = 278;
	/**
	 * {@code MOV r/m8, r8}<p>
	 * {@code 88 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_RM8_R8 = 279;
	/**
	 * {@code MOV r/m16, r16}<p>
	 * {@code o16 89 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_RM16_R16 = 280;
	/**
	 * {@code MOV r/m32, r32}<p>
	 * {@code o32 89 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_RM32_R32 = 281;
	/**
	 * {@code MOV r/m64, r64}<p>
	 * {@code o64 89 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOV_RM64_R64 = 282;
	/**
	 * {@code MOV r8, r/m8}<p>
	 * {@code 8A /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_R8_RM8 = 283;
	/**
	 * {@code MOV r16, r/m16}<p>
	 * {@code o16 8B /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_R16_RM16 = 284;
	/**
	 * {@code MOV r32, r/m32}<p>
	 * {@code o32 8B /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_R32_RM32 = 285;
	/**
	 * {@code MOV r64, r/m64}<p>
	 * {@code o64 8B /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOV_R64_RM64 = 286;
	/**
	 * {@code MOV r/m16, Sreg}<p>
	 * {@code o16 8C /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_RM16_SREG = 287;
	/**
	 * {@code MOV r32/m16, Sreg}<p>
	 * {@code o32 8C /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_R32M16_SREG = 288;
	/**
	 * {@code MOV r64/m16, Sreg}<p>
	 * {@code o64 8C /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOV_R64M16_SREG = 289;
	/**
	 * {@code LEA r16, m}<p>
	 * {@code o16 8D /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LEA_R16_M = 290;
	/**
	 * {@code LEA r32, m}<p>
	 * {@code o32 8D /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LEA_R32_M = 291;
	/**
	 * {@code LEA r64, m}<p>
	 * {@code o64 8D /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LEA_R64_M = 292;
	/**
	 * {@code MOV Sreg, r/m16}<p>
	 * {@code o16 8E /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_SREG_RM16 = 293;
	/**
	 * {@code MOV Sreg, r32/m16}<p>
	 * {@code o32 8E /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_SREG_R32M16 = 294;
	/**
	 * {@code MOV Sreg, r64/m16}<p>
	 * {@code o64 8E /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOV_SREG_R64M16 = 295;
	/**
	 * {@code POP r/m16}<p>
	 * {@code o16 8F /0}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int POP_RM16 = 296;
	/**
	 * {@code POP r/m32}<p>
	 * {@code o32 8F /0}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int POP_RM32 = 297;
	/**
	 * {@code POP r/m64}<p>
	 * {@code o64 8F /0}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int POP_RM64 = 298;
	/**
	 * {@code NOP}<p>
	 * {@code o16 90}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int NOPW = 299;
	/**
	 * {@code NOP}<p>
	 * {@code o32 90}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int NOPD = 300;
	/**
	 * {@code NOP}<p>
	 * {@code o64 90}<p>
	 * {@code 8086+}<p>
	 * {@code 64-bit}
	 */
	public static final int NOPQ = 301;
	/**
	 * {@code XCHG r16, AX}<p>
	 * {@code o16 90+rw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XCHG_R16_AX = 302;
	/**
	 * {@code XCHG r32, EAX}<p>
	 * {@code o32 90+rd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XCHG_R32_EAX = 303;
	/**
	 * {@code XCHG r64, RAX}<p>
	 * {@code o64 90+ro}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int XCHG_R64_RAX = 304;
	/**
	 * {@code PAUSE}<p>
	 * {@code F3 90}<p>
	 * {@code Pentium 4 or later}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PAUSE = 305;
	/**
	 * {@code CBW}<p>
	 * {@code o16 98}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CBW = 306;
	/**
	 * {@code CWDE}<p>
	 * {@code o32 98}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CWDE = 307;
	/**
	 * {@code CDQE}<p>
	 * {@code o64 98}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int CDQE = 308;
	/**
	 * {@code CWD}<p>
	 * {@code o16 99}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CWD = 309;
	/**
	 * {@code CDQ}<p>
	 * {@code o32 99}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CDQ = 310;
	/**
	 * {@code CQO}<p>
	 * {@code o64 99}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int CQO = 311;
	/**
	 * {@code CALL ptr16:16}<p>
	 * {@code o16 9A cd}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CALL_PTR1616 = 312;
	/**
	 * {@code CALL ptr16:32}<p>
	 * {@code o32 9A cp}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CALL_PTR1632 = 313;
	/**
	 * {@code WAIT}<p>
	 * {@code 9B}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int WAIT = 314;
	/**
	 * {@code PUSHF}<p>
	 * {@code o16 9C}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUSHFW = 315;
	/**
	 * {@code PUSHFD}<p>
	 * {@code o32 9C}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSHFD = 316;
	/**
	 * {@code PUSHFQ}<p>
	 * {@code o64 9C}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int PUSHFQ = 317;
	/**
	 * {@code POPF}<p>
	 * {@code o16 9D}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int POPFW = 318;
	/**
	 * {@code POPFD}<p>
	 * {@code o32 9D}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int POPFD = 319;
	/**
	 * {@code POPFQ}<p>
	 * {@code o64 9D}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int POPFQ = 320;
	/**
	 * {@code SAHF}<p>
	 * {@code 9E}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAHF = 321;
	/**
	 * {@code LAHF}<p>
	 * {@code 9F}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LAHF = 322;
	/**
	 * {@code MOV AL, moffs8}<p>
	 * {@code A0 mo}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_AL_MOFFS8 = 323;
	/**
	 * {@code MOV AX, moffs16}<p>
	 * {@code o16 A1 mo}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_AX_MOFFS16 = 324;
	/**
	 * {@code MOV EAX, moffs32}<p>
	 * {@code o32 A1 mo}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_EAX_MOFFS32 = 325;
	/**
	 * {@code MOV RAX, moffs64}<p>
	 * {@code o64 A1 mo}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOV_RAX_MOFFS64 = 326;
	/**
	 * {@code MOV moffs8, AL}<p>
	 * {@code A2 mo}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_MOFFS8_AL = 327;
	/**
	 * {@code MOV moffs16, AX}<p>
	 * {@code o16 A3 mo}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_MOFFS16_AX = 328;
	/**
	 * {@code MOV moffs32, EAX}<p>
	 * {@code o32 A3 mo}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_MOFFS32_EAX = 329;
	/**
	 * {@code MOV moffs64, RAX}<p>
	 * {@code o64 A3 mo}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOV_MOFFS64_RAX = 330;
	/**
	 * {@code MOVSB}<p>
	 * {@code A4}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVSB_M8_M8 = 331;
	/**
	 * {@code MOVSW}<p>
	 * {@code o16 A5}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVSW_M16_M16 = 332;
	/**
	 * {@code MOVSD}<p>
	 * {@code o32 A5}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVSD_M32_M32 = 333;
	/**
	 * {@code MOVSQ}<p>
	 * {@code o64 A5}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVSQ_M64_M64 = 334;
	/**
	 * {@code CMPSB}<p>
	 * {@code A6}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMPSB_M8_M8 = 335;
	/**
	 * {@code CMPSW}<p>
	 * {@code o16 A7}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMPSW_M16_M16 = 336;
	/**
	 * {@code CMPSD}<p>
	 * {@code o32 A7}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMPSD_M32_M32 = 337;
	/**
	 * {@code CMPSQ}<p>
	 * {@code o64 A7}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int CMPSQ_M64_M64 = 338;
	/**
	 * {@code TEST AL, imm8}<p>
	 * {@code A8 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TEST_AL_IMM8 = 339;
	/**
	 * {@code TEST AX, imm16}<p>
	 * {@code o16 A9 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TEST_AX_IMM16 = 340;
	/**
	 * {@code TEST EAX, imm32}<p>
	 * {@code o32 A9 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TEST_EAX_IMM32 = 341;
	/**
	 * {@code TEST RAX, imm32}<p>
	 * {@code o64 A9 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int TEST_RAX_IMM32 = 342;
	/**
	 * {@code STOSB}<p>
	 * {@code AA}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int STOSB_M8_AL = 343;
	/**
	 * {@code STOSW}<p>
	 * {@code o16 AB}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int STOSW_M16_AX = 344;
	/**
	 * {@code STOSD}<p>
	 * {@code o32 AB}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int STOSD_M32_EAX = 345;
	/**
	 * {@code STOSQ}<p>
	 * {@code o64 AB}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int STOSQ_M64_RAX = 346;
	/**
	 * {@code LODSB}<p>
	 * {@code AC}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LODSB_AL_M8 = 347;
	/**
	 * {@code LODSW}<p>
	 * {@code o16 AD}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LODSW_AX_M16 = 348;
	/**
	 * {@code LODSD}<p>
	 * {@code o32 AD}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LODSD_EAX_M32 = 349;
	/**
	 * {@code LODSQ}<p>
	 * {@code o64 AD}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LODSQ_RAX_M64 = 350;
	/**
	 * {@code SCASB}<p>
	 * {@code AE}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SCASB_AL_M8 = 351;
	/**
	 * {@code SCASW}<p>
	 * {@code o16 AF}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SCASW_AX_M16 = 352;
	/**
	 * {@code SCASD}<p>
	 * {@code o32 AF}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SCASD_EAX_M32 = 353;
	/**
	 * {@code SCASQ}<p>
	 * {@code o64 AF}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SCASQ_RAX_M64 = 354;
	/**
	 * {@code MOV r8, imm8}<p>
	 * {@code B0+rb ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_R8_IMM8 = 355;
	/**
	 * {@code MOV r16, imm16}<p>
	 * {@code o16 B8+rw iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_R16_IMM16 = 356;
	/**
	 * {@code MOV r32, imm32}<p>
	 * {@code o32 B8+rd id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_R32_IMM32 = 357;
	/**
	 * {@code MOV r64, imm64}<p>
	 * {@code o64 B8+ro io}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOV_R64_IMM64 = 358;
	/**
	 * {@code ROL r/m8, imm8}<p>
	 * {@code C0 /0 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROL_RM8_IMM8 = 359;
	/**
	 * {@code ROR r/m8, imm8}<p>
	 * {@code C0 /1 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROR_RM8_IMM8 = 360;
	/**
	 * {@code RCL r/m8, imm8}<p>
	 * {@code C0 /2 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCL_RM8_IMM8 = 361;
	/**
	 * {@code RCR r/m8, imm8}<p>
	 * {@code C0 /3 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCR_RM8_IMM8 = 362;
	/**
	 * {@code SHL r/m8, imm8}<p>
	 * {@code C0 /4 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHL_RM8_IMM8 = 363;
	/**
	 * {@code SHR r/m8, imm8}<p>
	 * {@code C0 /5 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHR_RM8_IMM8 = 364;
	/**
	 * {@code SAL r/m8, imm8}<p>
	 * {@code C0 /6 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAL_RM8_IMM8 = 365;
	/**
	 * {@code SAR r/m8, imm8}<p>
	 * {@code C0 /7 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAR_RM8_IMM8 = 366;
	/**
	 * {@code ROL r/m16, imm8}<p>
	 * {@code o16 C1 /0 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROL_RM16_IMM8 = 367;
	/**
	 * {@code ROL r/m32, imm8}<p>
	 * {@code o32 C1 /0 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROL_RM32_IMM8 = 368;
	/**
	 * {@code ROL r/m64, imm8}<p>
	 * {@code o64 C1 /0 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ROL_RM64_IMM8 = 369;
	/**
	 * {@code ROR r/m16, imm8}<p>
	 * {@code o16 C1 /1 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROR_RM16_IMM8 = 370;
	/**
	 * {@code ROR r/m32, imm8}<p>
	 * {@code o32 C1 /1 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROR_RM32_IMM8 = 371;
	/**
	 * {@code ROR r/m64, imm8}<p>
	 * {@code o64 C1 /1 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ROR_RM64_IMM8 = 372;
	/**
	 * {@code RCL r/m16, imm8}<p>
	 * {@code o16 C1 /2 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCL_RM16_IMM8 = 373;
	/**
	 * {@code RCL r/m32, imm8}<p>
	 * {@code o32 C1 /2 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCL_RM32_IMM8 = 374;
	/**
	 * {@code RCL r/m64, imm8}<p>
	 * {@code o64 C1 /2 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int RCL_RM64_IMM8 = 375;
	/**
	 * {@code RCR r/m16, imm8}<p>
	 * {@code o16 C1 /3 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCR_RM16_IMM8 = 376;
	/**
	 * {@code RCR r/m32, imm8}<p>
	 * {@code o32 C1 /3 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCR_RM32_IMM8 = 377;
	/**
	 * {@code RCR r/m64, imm8}<p>
	 * {@code o64 C1 /3 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int RCR_RM64_IMM8 = 378;
	/**
	 * {@code SHL r/m16, imm8}<p>
	 * {@code o16 C1 /4 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHL_RM16_IMM8 = 379;
	/**
	 * {@code SHL r/m32, imm8}<p>
	 * {@code o32 C1 /4 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHL_RM32_IMM8 = 380;
	/**
	 * {@code SHL r/m64, imm8}<p>
	 * {@code o64 C1 /4 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SHL_RM64_IMM8 = 381;
	/**
	 * {@code SHR r/m16, imm8}<p>
	 * {@code o16 C1 /5 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHR_RM16_IMM8 = 382;
	/**
	 * {@code SHR r/m32, imm8}<p>
	 * {@code o32 C1 /5 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHR_RM32_IMM8 = 383;
	/**
	 * {@code SHR r/m64, imm8}<p>
	 * {@code o64 C1 /5 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SHR_RM64_IMM8 = 384;
	/**
	 * {@code SAL r/m16, imm8}<p>
	 * {@code o16 C1 /6 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAL_RM16_IMM8 = 385;
	/**
	 * {@code SAL r/m32, imm8}<p>
	 * {@code o32 C1 /6 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAL_RM32_IMM8 = 386;
	/**
	 * {@code SAL r/m64, imm8}<p>
	 * {@code o64 C1 /6 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SAL_RM64_IMM8 = 387;
	/**
	 * {@code SAR r/m16, imm8}<p>
	 * {@code o16 C1 /7 ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAR_RM16_IMM8 = 388;
	/**
	 * {@code SAR r/m32, imm8}<p>
	 * {@code o32 C1 /7 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAR_RM32_IMM8 = 389;
	/**
	 * {@code SAR r/m64, imm8}<p>
	 * {@code o64 C1 /7 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SAR_RM64_IMM8 = 390;
	/**
	 * {@code RET imm16}<p>
	 * {@code o16 C2 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RETNW_IMM16 = 391;
	/**
	 * {@code RET imm16}<p>
	 * {@code o32 C2 iw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int RETND_IMM16 = 392;
	/**
	 * {@code RET imm16}<p>
	 * {@code o64 C2 iw}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int RETNQ_IMM16 = 393;
	/**
	 * {@code RET}<p>
	 * {@code o16 C3}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RETNW = 394;
	/**
	 * {@code RET}<p>
	 * {@code o32 C3}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int RETND = 395;
	/**
	 * {@code RET}<p>
	 * {@code o64 C3}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int RETNQ = 396;
	/**
	 * {@code LES r16, m16:16}<p>
	 * {@code o16 C4 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LES_R16_M1616 = 397;
	/**
	 * {@code LES r32, m16:32}<p>
	 * {@code o32 C4 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LES_R32_M1632 = 398;
	/**
	 * {@code LDS r16, m16:16}<p>
	 * {@code o16 C5 /r}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LDS_R16_M1616 = 399;
	/**
	 * {@code LDS r32, m16:32}<p>
	 * {@code o32 C5 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LDS_R32_M1632 = 400;
	/**
	 * {@code MOV r/m8, imm8}<p>
	 * {@code C6 /0 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_RM8_IMM8 = 401;
	/**
	 * {@code XABORT imm8}<p>
	 * {@code C6 F8 ib}<p>
	 * {@code RTM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XABORT_IMM8 = 402;
	/**
	 * {@code MOV r/m16, imm16}<p>
	 * {@code o16 C7 /0 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_RM16_IMM16 = 403;
	/**
	 * {@code MOV r/m32, imm32}<p>
	 * {@code o32 C7 /0 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOV_RM32_IMM32 = 404;
	/**
	 * {@code MOV r/m64, imm32}<p>
	 * {@code o64 C7 /0 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOV_RM64_IMM32 = 405;
	/**
	 * {@code XBEGIN rel16}<p>
	 * {@code o16 C7 F8 cw}<p>
	 * {@code RTM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XBEGIN_REL16 = 406;
	/**
	 * {@code XBEGIN rel32}<p>
	 * {@code o32 C7 F8 cd}<p>
	 * {@code RTM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XBEGIN_REL32 = 407;
	/**
	 * {@code ENTER imm16, imm8}<p>
	 * {@code o16 C8 iw ib}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ENTERW_IMM16_IMM8 = 408;
	/**
	 * {@code ENTER imm16, imm8}<p>
	 * {@code o32 C8 iw ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int ENTERD_IMM16_IMM8 = 409;
	/**
	 * {@code ENTER imm16, imm8}<p>
	 * {@code o64 C8 iw ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ENTERQ_IMM16_IMM8 = 410;
	/**
	 * {@code LEAVE}<p>
	 * {@code o16 C9}<p>
	 * {@code 186+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LEAVEW = 411;
	/**
	 * {@code LEAVE}<p>
	 * {@code o32 C9}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LEAVED = 412;
	/**
	 * {@code LEAVE}<p>
	 * {@code o64 C9}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LEAVEQ = 413;
	/**
	 * {@code RETF imm16}<p>
	 * {@code o16 CA iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RETFW_IMM16 = 414;
	/**
	 * {@code RETF imm16}<p>
	 * {@code o32 CA iw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RETFD_IMM16 = 415;
	/**
	 * {@code RETF imm16}<p>
	 * {@code o64 CA iw}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int RETFQ_IMM16 = 416;
	/**
	 * {@code RETF}<p>
	 * {@code o16 CB}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RETFW = 417;
	/**
	 * {@code RETF}<p>
	 * {@code o32 CB}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RETFD = 418;
	/**
	 * {@code RETF}<p>
	 * {@code o64 CB}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int RETFQ = 419;
	/**
	 * {@code INT3}<p>
	 * {@code CC}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INT3 = 420;
	/**
	 * {@code INT imm8}<p>
	 * {@code CD ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INT_IMM8 = 421;
	/**
	 * {@code INTO}<p>
	 * {@code CE}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int INTO = 422;
	/**
	 * {@code IRET}<p>
	 * {@code o16 CF}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IRETW = 423;
	/**
	 * {@code IRETD}<p>
	 * {@code o32 CF}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IRETD = 424;
	/**
	 * {@code IRETQ}<p>
	 * {@code o64 CF}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int IRETQ = 425;
	/**
	 * {@code ROL r/m8, 1}<p>
	 * {@code D0 /0}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROL_RM8_1 = 426;
	/**
	 * {@code ROR r/m8, 1}<p>
	 * {@code D0 /1}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROR_RM8_1 = 427;
	/**
	 * {@code RCL r/m8, 1}<p>
	 * {@code D0 /2}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCL_RM8_1 = 428;
	/**
	 * {@code RCR r/m8, 1}<p>
	 * {@code D0 /3}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCR_RM8_1 = 429;
	/**
	 * {@code SHL r/m8, 1}<p>
	 * {@code D0 /4}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHL_RM8_1 = 430;
	/**
	 * {@code SHR r/m8, 1}<p>
	 * {@code D0 /5}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHR_RM8_1 = 431;
	/**
	 * {@code SAL r/m8, 1}<p>
	 * {@code D0 /6}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAL_RM8_1 = 432;
	/**
	 * {@code SAR r/m8, 1}<p>
	 * {@code D0 /7}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAR_RM8_1 = 433;
	/**
	 * {@code ROL r/m16, 1}<p>
	 * {@code o16 D1 /0}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROL_RM16_1 = 434;
	/**
	 * {@code ROL r/m32, 1}<p>
	 * {@code o32 D1 /0}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROL_RM32_1 = 435;
	/**
	 * {@code ROL r/m64, 1}<p>
	 * {@code o64 D1 /0}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ROL_RM64_1 = 436;
	/**
	 * {@code ROR r/m16, 1}<p>
	 * {@code o16 D1 /1}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROR_RM16_1 = 437;
	/**
	 * {@code ROR r/m32, 1}<p>
	 * {@code o32 D1 /1}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROR_RM32_1 = 438;
	/**
	 * {@code ROR r/m64, 1}<p>
	 * {@code o64 D1 /1}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ROR_RM64_1 = 439;
	/**
	 * {@code RCL r/m16, 1}<p>
	 * {@code o16 D1 /2}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCL_RM16_1 = 440;
	/**
	 * {@code RCL r/m32, 1}<p>
	 * {@code o32 D1 /2}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCL_RM32_1 = 441;
	/**
	 * {@code RCL r/m64, 1}<p>
	 * {@code o64 D1 /2}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int RCL_RM64_1 = 442;
	/**
	 * {@code RCR r/m16, 1}<p>
	 * {@code o16 D1 /3}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCR_RM16_1 = 443;
	/**
	 * {@code RCR r/m32, 1}<p>
	 * {@code o32 D1 /3}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCR_RM32_1 = 444;
	/**
	 * {@code RCR r/m64, 1}<p>
	 * {@code o64 D1 /3}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int RCR_RM64_1 = 445;
	/**
	 * {@code SHL r/m16, 1}<p>
	 * {@code o16 D1 /4}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHL_RM16_1 = 446;
	/**
	 * {@code SHL r/m32, 1}<p>
	 * {@code o32 D1 /4}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHL_RM32_1 = 447;
	/**
	 * {@code SHL r/m64, 1}<p>
	 * {@code o64 D1 /4}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SHL_RM64_1 = 448;
	/**
	 * {@code SHR r/m16, 1}<p>
	 * {@code o16 D1 /5}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHR_RM16_1 = 449;
	/**
	 * {@code SHR r/m32, 1}<p>
	 * {@code o32 D1 /5}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHR_RM32_1 = 450;
	/**
	 * {@code SHR r/m64, 1}<p>
	 * {@code o64 D1 /5}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SHR_RM64_1 = 451;
	/**
	 * {@code SAL r/m16, 1}<p>
	 * {@code o16 D1 /6}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAL_RM16_1 = 452;
	/**
	 * {@code SAL r/m32, 1}<p>
	 * {@code o32 D1 /6}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAL_RM32_1 = 453;
	/**
	 * {@code SAL r/m64, 1}<p>
	 * {@code o64 D1 /6}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SAL_RM64_1 = 454;
	/**
	 * {@code SAR r/m16, 1}<p>
	 * {@code o16 D1 /7}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAR_RM16_1 = 455;
	/**
	 * {@code SAR r/m32, 1}<p>
	 * {@code o32 D1 /7}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAR_RM32_1 = 456;
	/**
	 * {@code SAR r/m64, 1}<p>
	 * {@code o64 D1 /7}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SAR_RM64_1 = 457;
	/**
	 * {@code ROL r/m8, CL}<p>
	 * {@code D2 /0}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROL_RM8_CL = 458;
	/**
	 * {@code ROR r/m8, CL}<p>
	 * {@code D2 /1}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROR_RM8_CL = 459;
	/**
	 * {@code RCL r/m8, CL}<p>
	 * {@code D2 /2}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCL_RM8_CL = 460;
	/**
	 * {@code RCR r/m8, CL}<p>
	 * {@code D2 /3}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCR_RM8_CL = 461;
	/**
	 * {@code SHL r/m8, CL}<p>
	 * {@code D2 /4}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHL_RM8_CL = 462;
	/**
	 * {@code SHR r/m8, CL}<p>
	 * {@code D2 /5}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHR_RM8_CL = 463;
	/**
	 * {@code SAL r/m8, CL}<p>
	 * {@code D2 /6}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAL_RM8_CL = 464;
	/**
	 * {@code SAR r/m8, CL}<p>
	 * {@code D2 /7}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAR_RM8_CL = 465;
	/**
	 * {@code ROL r/m16, CL}<p>
	 * {@code o16 D3 /0}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROL_RM16_CL = 466;
	/**
	 * {@code ROL r/m32, CL}<p>
	 * {@code o32 D3 /0}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROL_RM32_CL = 467;
	/**
	 * {@code ROL r/m64, CL}<p>
	 * {@code o64 D3 /0}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ROL_RM64_CL = 468;
	/**
	 * {@code ROR r/m16, CL}<p>
	 * {@code o16 D3 /1}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROR_RM16_CL = 469;
	/**
	 * {@code ROR r/m32, CL}<p>
	 * {@code o32 D3 /1}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROR_RM32_CL = 470;
	/**
	 * {@code ROR r/m64, CL}<p>
	 * {@code o64 D3 /1}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int ROR_RM64_CL = 471;
	/**
	 * {@code RCL r/m16, CL}<p>
	 * {@code o16 D3 /2}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCL_RM16_CL = 472;
	/**
	 * {@code RCL r/m32, CL}<p>
	 * {@code o32 D3 /2}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCL_RM32_CL = 473;
	/**
	 * {@code RCL r/m64, CL}<p>
	 * {@code o64 D3 /2}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int RCL_RM64_CL = 474;
	/**
	 * {@code RCR r/m16, CL}<p>
	 * {@code o16 D3 /3}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCR_RM16_CL = 475;
	/**
	 * {@code RCR r/m32, CL}<p>
	 * {@code o32 D3 /3}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCR_RM32_CL = 476;
	/**
	 * {@code RCR r/m64, CL}<p>
	 * {@code o64 D3 /3}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int RCR_RM64_CL = 477;
	/**
	 * {@code SHL r/m16, CL}<p>
	 * {@code o16 D3 /4}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHL_RM16_CL = 478;
	/**
	 * {@code SHL r/m32, CL}<p>
	 * {@code o32 D3 /4}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHL_RM32_CL = 479;
	/**
	 * {@code SHL r/m64, CL}<p>
	 * {@code o64 D3 /4}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SHL_RM64_CL = 480;
	/**
	 * {@code SHR r/m16, CL}<p>
	 * {@code o16 D3 /5}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHR_RM16_CL = 481;
	/**
	 * {@code SHR r/m32, CL}<p>
	 * {@code o32 D3 /5}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHR_RM32_CL = 482;
	/**
	 * {@code SHR r/m64, CL}<p>
	 * {@code o64 D3 /5}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SHR_RM64_CL = 483;
	/**
	 * {@code SAL r/m16, CL}<p>
	 * {@code o16 D3 /6}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAL_RM16_CL = 484;
	/**
	 * {@code SAL r/m32, CL}<p>
	 * {@code o32 D3 /6}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAL_RM32_CL = 485;
	/**
	 * {@code SAL r/m64, CL}<p>
	 * {@code o64 D3 /6}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SAL_RM64_CL = 486;
	/**
	 * {@code SAR r/m16, CL}<p>
	 * {@code o16 D3 /7}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAR_RM16_CL = 487;
	/**
	 * {@code SAR r/m32, CL}<p>
	 * {@code o32 D3 /7}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAR_RM32_CL = 488;
	/**
	 * {@code SAR r/m64, CL}<p>
	 * {@code o64 D3 /7}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SAR_RM64_CL = 489;
	/**
	 * {@code AAM imm8}<p>
	 * {@code D4 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int AAM_IMM8 = 490;
	/**
	 * {@code AAD imm8}<p>
	 * {@code D5 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int AAD_IMM8 = 491;
	/**
	 * {@code SALC}<p>
	 * {@code D6}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int SALC = 492;
	/**
	 * {@code XLATB}<p>
	 * {@code D7}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XLAT_M8 = 493;
	/**
	 * {@code FADD m32fp}<p>
	 * {@code D8 /0}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FADD_M32FP = 494;
	/**
	 * {@code FMUL m32fp}<p>
	 * {@code D8 /1}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FMUL_M32FP = 495;
	/**
	 * {@code FCOM m32fp}<p>
	 * {@code D8 /2}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCOM_M32FP = 496;
	/**
	 * {@code FCOMP m32fp}<p>
	 * {@code D8 /3}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCOMP_M32FP = 497;
	/**
	 * {@code FSUB m32fp}<p>
	 * {@code D8 /4}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSUB_M32FP = 498;
	/**
	 * {@code FSUBR m32fp}<p>
	 * {@code D8 /5}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSUBR_M32FP = 499;
	/**
	 * {@code FDIV m32fp}<p>
	 * {@code D8 /6}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FDIV_M32FP = 500;
	/**
	 * {@code FDIVR m32fp}<p>
	 * {@code D8 /7}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FDIVR_M32FP = 501;
	/**
	 * {@code FADD ST(0), ST(i)}<p>
	 * {@code D8 C0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FADD_ST0_STI = 502;
	/**
	 * {@code FMUL ST(0), ST(i)}<p>
	 * {@code D8 C8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FMUL_ST0_STI = 503;
	/**
	 * {@code FCOM ST(i)}<p>
	 * {@code D8 D0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCOM_ST0_STI = 504;
	/**
	 * {@code FCOMP ST(i)}<p>
	 * {@code D8 D8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCOMP_ST0_STI = 505;
	/**
	 * {@code FSUB ST(0), ST(i)}<p>
	 * {@code D8 E0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSUB_ST0_STI = 506;
	/**
	 * {@code FSUBR ST(0), ST(i)}<p>
	 * {@code D8 E8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSUBR_ST0_STI = 507;
	/**
	 * {@code FDIV ST(0), ST(i)}<p>
	 * {@code D8 F0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FDIV_ST0_STI = 508;
	/**
	 * {@code FDIVR ST(0), ST(i)}<p>
	 * {@code D8 F8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FDIVR_ST0_STI = 509;
	/**
	 * {@code FLD m32fp}<p>
	 * {@code D9 /0}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FLD_M32FP = 510;
	/**
	 * {@code FST m32fp}<p>
	 * {@code D9 /2}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FST_M32FP = 511;
	/**
	 * {@code FSTP m32fp}<p>
	 * {@code D9 /3}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSTP_M32FP = 512;
	/**
	 * {@code FLDENV m14byte}<p>
	 * {@code o16 D9 /4}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FLDENV_M14BYTE = 513;
	/**
	 * {@code FLDENV m28byte}<p>
	 * {@code o32 D9 /4}<p>
	 * {@code 387+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FLDENV_M28BYTE = 514;
	/**
	 * {@code FLDCW m2byte}<p>
	 * {@code D9 /5}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FLDCW_M2BYTE = 515;
	/**
	 * {@code FNSTENV m14byte}<p>
	 * {@code o16 D9 /6}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FNSTENV_M14BYTE = 516;
	/**
	 * {@code FSTENV m14byte}<p>
	 * {@code 9B o16 D9 /6}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSTENV_M14BYTE = 517;
	/**
	 * {@code FNSTENV m28byte}<p>
	 * {@code o32 D9 /6}<p>
	 * {@code 387+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FNSTENV_M28BYTE = 518;
	/**
	 * {@code FSTENV m28byte}<p>
	 * {@code 9B o32 D9 /6}<p>
	 * {@code 387+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSTENV_M28BYTE = 519;
	/**
	 * {@code FNSTCW m2byte}<p>
	 * {@code D9 /7}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FNSTCW_M2BYTE = 520;
	/**
	 * {@code FSTCW m2byte}<p>
	 * {@code 9B D9 /7}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSTCW_M2BYTE = 521;
	/**
	 * {@code FLD ST(i)}<p>
	 * {@code D9 C0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FLD_STI = 522;
	/**
	 * {@code FXCH ST(i)}<p>
	 * {@code D9 C8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FXCH_ST0_STI = 523;
	/**
	 * {@code FNOP}<p>
	 * {@code D9 D0}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FNOP = 524;
	/**
	 * {@code FSTPNCE ST(i)}<p>
	 * {@code D9 D8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSTPNCE_STI = 525;
	/**
	 * {@code FCHS}<p>
	 * {@code D9 E0}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCHS = 526;
	/**
	 * {@code FABS}<p>
	 * {@code D9 E1}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FABS = 527;
	/**
	 * {@code FTST}<p>
	 * {@code D9 E4}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FTST = 528;
	/**
	 * {@code FXAM}<p>
	 * {@code D9 E5}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FXAM = 529;
	/**
	 * {@code FLD1}<p>
	 * {@code D9 E8}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FLD1 = 530;
	/**
	 * {@code FLDL2T}<p>
	 * {@code D9 E9}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FLDL2T = 531;
	/**
	 * {@code FLDL2E}<p>
	 * {@code D9 EA}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FLDL2E = 532;
	/**
	 * {@code FLDPI}<p>
	 * {@code D9 EB}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FLDPI = 533;
	/**
	 * {@code FLDLG2}<p>
	 * {@code D9 EC}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FLDLG2 = 534;
	/**
	 * {@code FLDLN2}<p>
	 * {@code D9 ED}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FLDLN2 = 535;
	/**
	 * {@code FLDZ}<p>
	 * {@code D9 EE}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FLDZ = 536;
	/**
	 * {@code F2XM1}<p>
	 * {@code D9 F0}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int F2XM1 = 537;
	/**
	 * {@code FYL2X}<p>
	 * {@code D9 F1}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FYL2X = 538;
	/**
	 * {@code FPTAN}<p>
	 * {@code D9 F2}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FPTAN = 539;
	/**
	 * {@code FPATAN}<p>
	 * {@code D9 F3}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FPATAN = 540;
	/**
	 * {@code FXTRACT}<p>
	 * {@code D9 F4}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FXTRACT = 541;
	/**
	 * {@code FPREM1}<p>
	 * {@code D9 F5}<p>
	 * {@code 387+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FPREM1 = 542;
	/**
	 * {@code FDECSTP}<p>
	 * {@code D9 F6}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FDECSTP = 543;
	/**
	 * {@code FINCSTP}<p>
	 * {@code D9 F7}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FINCSTP = 544;
	/**
	 * {@code FPREM}<p>
	 * {@code D9 F8}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FPREM = 545;
	/**
	 * {@code FYL2XP1}<p>
	 * {@code D9 F9}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FYL2XP1 = 546;
	/**
	 * {@code FSQRT}<p>
	 * {@code D9 FA}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSQRT = 547;
	/**
	 * {@code FSINCOS}<p>
	 * {@code D9 FB}<p>
	 * {@code 387+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSINCOS = 548;
	/**
	 * {@code FRNDINT}<p>
	 * {@code D9 FC}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FRNDINT = 549;
	/**
	 * {@code FSCALE}<p>
	 * {@code D9 FD}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSCALE = 550;
	/**
	 * {@code FSIN}<p>
	 * {@code D9 FE}<p>
	 * {@code 387+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSIN = 551;
	/**
	 * {@code FCOS}<p>
	 * {@code D9 FF}<p>
	 * {@code 387+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCOS = 552;
	/**
	 * {@code FIADD m32int}<p>
	 * {@code DA /0}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FIADD_M32INT = 553;
	/**
	 * {@code FIMUL m32int}<p>
	 * {@code DA /1}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FIMUL_M32INT = 554;
	/**
	 * {@code FICOM m32int}<p>
	 * {@code DA /2}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FICOM_M32INT = 555;
	/**
	 * {@code FICOMP m32int}<p>
	 * {@code DA /3}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FICOMP_M32INT = 556;
	/**
	 * {@code FISUB m32int}<p>
	 * {@code DA /4}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FISUB_M32INT = 557;
	/**
	 * {@code FISUBR m32int}<p>
	 * {@code DA /5}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FISUBR_M32INT = 558;
	/**
	 * {@code FIDIV m32int}<p>
	 * {@code DA /6}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FIDIV_M32INT = 559;
	/**
	 * {@code FIDIVR m32int}<p>
	 * {@code DA /7}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FIDIVR_M32INT = 560;
	/**
	 * {@code FCMOVB ST(0), ST(i)}<p>
	 * {@code DA C0+i}<p>
	 * {@code 8087+ and CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCMOVB_ST0_STI = 561;
	/**
	 * {@code FCMOVE ST(0), ST(i)}<p>
	 * {@code DA C8+i}<p>
	 * {@code 8087+ and CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCMOVE_ST0_STI = 562;
	/**
	 * {@code FCMOVBE ST(0), ST(i)}<p>
	 * {@code DA D0+i}<p>
	 * {@code 8087+ and CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCMOVBE_ST0_STI = 563;
	/**
	 * {@code FCMOVU ST(0), ST(i)}<p>
	 * {@code DA D8+i}<p>
	 * {@code 8087+ and CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCMOVU_ST0_STI = 564;
	/**
	 * {@code FUCOMPP}<p>
	 * {@code DA E9}<p>
	 * {@code 387+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FUCOMPP = 565;
	/**
	 * {@code FILD m32int}<p>
	 * {@code DB /0}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FILD_M32INT = 566;
	/**
	 * {@code FISTTP m32int}<p>
	 * {@code DB /1}<p>
	 * {@code 8087+ and SSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FISTTP_M32INT = 567;
	/**
	 * {@code FIST m32int}<p>
	 * {@code DB /2}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FIST_M32INT = 568;
	/**
	 * {@code FISTP m32int}<p>
	 * {@code DB /3}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FISTP_M32INT = 569;
	/**
	 * {@code FLD m80fp}<p>
	 * {@code DB /5}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FLD_M80FP = 570;
	/**
	 * {@code FSTP m80fp}<p>
	 * {@code DB /7}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSTP_M80FP = 571;
	/**
	 * {@code FCMOVNB ST(0), ST(i)}<p>
	 * {@code DB C0+i}<p>
	 * {@code 8087+ and CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCMOVNB_ST0_STI = 572;
	/**
	 * {@code FCMOVNE ST(0), ST(i)}<p>
	 * {@code DB C8+i}<p>
	 * {@code 8087+ and CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCMOVNE_ST0_STI = 573;
	/**
	 * {@code FCMOVNBE ST(0), ST(i)}<p>
	 * {@code DB D0+i}<p>
	 * {@code 8087+ and CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCMOVNBE_ST0_STI = 574;
	/**
	 * {@code FCMOVNU ST(0), ST(i)}<p>
	 * {@code DB D8+i}<p>
	 * {@code 8087+ and CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCMOVNU_ST0_STI = 575;
	/**
	 * {@code FNENI}<p>
	 * {@code DB E0}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FNENI = 576;
	/**
	 * {@code FENI}<p>
	 * {@code 9B DB E0}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FENI = 577;
	/**
	 * {@code FNDISI}<p>
	 * {@code DB E1}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FNDISI = 578;
	/**
	 * {@code FDISI}<p>
	 * {@code 9B DB E1}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FDISI = 579;
	/**
	 * {@code FNCLEX}<p>
	 * {@code DB E2}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FNCLEX = 580;
	/**
	 * {@code FCLEX}<p>
	 * {@code 9B DB E2}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCLEX = 581;
	/**
	 * {@code FNINIT}<p>
	 * {@code DB E3}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FNINIT = 582;
	/**
	 * {@code FINIT}<p>
	 * {@code 9B DB E3}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FINIT = 583;
	/**
	 * {@code FNSETPM}<p>
	 * {@code DB E4}<p>
	 * {@code 287+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FNSETPM = 584;
	/**
	 * {@code FSETPM}<p>
	 * {@code 9B DB E4}<p>
	 * {@code 287+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSETPM = 585;
	/**
	 * {@code FRSTPM}<p>
	 * {@code DB E5}<p>
	 * {@code 287 XL}<p>
	 * {@code 16/32-bit}
	 */
	public static final int FRSTPM = 586;
	/**
	 * {@code FUCOMI ST, ST(i)}<p>
	 * {@code DB E8+i}<p>
	 * {@code 8087+ and CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FUCOMI_ST0_STI = 587;
	/**
	 * {@code FCOMI ST, ST(i)}<p>
	 * {@code DB F0+i}<p>
	 * {@code 8087+ and CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCOMI_ST0_STI = 588;
	/**
	 * {@code FADD m64fp}<p>
	 * {@code DC /0}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FADD_M64FP = 589;
	/**
	 * {@code FMUL m64fp}<p>
	 * {@code DC /1}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FMUL_M64FP = 590;
	/**
	 * {@code FCOM m64fp}<p>
	 * {@code DC /2}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCOM_M64FP = 591;
	/**
	 * {@code FCOMP m64fp}<p>
	 * {@code DC /3}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCOMP_M64FP = 592;
	/**
	 * {@code FSUB m64fp}<p>
	 * {@code DC /4}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSUB_M64FP = 593;
	/**
	 * {@code FSUBR m64fp}<p>
	 * {@code DC /5}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSUBR_M64FP = 594;
	/**
	 * {@code FDIV m64fp}<p>
	 * {@code DC /6}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FDIV_M64FP = 595;
	/**
	 * {@code FDIVR m64fp}<p>
	 * {@code DC /7}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FDIVR_M64FP = 596;
	/**
	 * {@code FADD ST(i), ST(0)}<p>
	 * {@code DC C0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FADD_STI_ST0 = 597;
	/**
	 * {@code FMUL ST(i), ST(0)}<p>
	 * {@code DC C8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FMUL_STI_ST0 = 598;
	/**
	 * {@code FCOM ST(i)}<p>
	 * {@code DC D0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCOM_ST0_STI_DCD0 = 599;
	/**
	 * {@code FCOMP ST(i)}<p>
	 * {@code DC D8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCOMP_ST0_STI_DCD8 = 600;
	/**
	 * {@code FSUBR ST(i), ST(0)}<p>
	 * {@code DC E0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSUBR_STI_ST0 = 601;
	/**
	 * {@code FSUB ST(i), ST(0)}<p>
	 * {@code DC E8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSUB_STI_ST0 = 602;
	/**
	 * {@code FDIVR ST(i), ST(0)}<p>
	 * {@code DC F0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FDIVR_STI_ST0 = 603;
	/**
	 * {@code FDIV ST(i), ST(0)}<p>
	 * {@code DC F8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FDIV_STI_ST0 = 604;
	/**
	 * {@code FLD m64fp}<p>
	 * {@code DD /0}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FLD_M64FP = 605;
	/**
	 * {@code FISTTP m64int}<p>
	 * {@code DD /1}<p>
	 * {@code 8087+ and SSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FISTTP_M64INT = 606;
	/**
	 * {@code FST m64fp}<p>
	 * {@code DD /2}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FST_M64FP = 607;
	/**
	 * {@code FSTP m64fp}<p>
	 * {@code DD /3}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSTP_M64FP = 608;
	/**
	 * {@code FRSTOR m94byte}<p>
	 * {@code o16 DD /4}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FRSTOR_M94BYTE = 609;
	/**
	 * {@code FRSTOR m108byte}<p>
	 * {@code o32 DD /4}<p>
	 * {@code 387+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FRSTOR_M108BYTE = 610;
	/**
	 * {@code FNSAVE m94byte}<p>
	 * {@code o16 DD /6}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FNSAVE_M94BYTE = 611;
	/**
	 * {@code FSAVE m94byte}<p>
	 * {@code 9B o16 DD /6}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSAVE_M94BYTE = 612;
	/**
	 * {@code FNSAVE m108byte}<p>
	 * {@code o32 DD /6}<p>
	 * {@code 387+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FNSAVE_M108BYTE = 613;
	/**
	 * {@code FSAVE m108byte}<p>
	 * {@code 9B o32 DD /6}<p>
	 * {@code 387+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSAVE_M108BYTE = 614;
	/**
	 * {@code FNSTSW m2byte}<p>
	 * {@code DD /7}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FNSTSW_M2BYTE = 615;
	/**
	 * {@code FSTSW m2byte}<p>
	 * {@code 9B DD /7}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSTSW_M2BYTE = 616;
	/**
	 * {@code FFREE ST(i)}<p>
	 * {@code DD C0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FFREE_STI = 617;
	/**
	 * {@code FXCH ST(i)}<p>
	 * {@code DD C8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FXCH_ST0_STI_DDC8 = 618;
	/**
	 * {@code FST ST(i)}<p>
	 * {@code DD D0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FST_STI = 619;
	/**
	 * {@code FSTP ST(i)}<p>
	 * {@code DD D8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSTP_STI = 620;
	/**
	 * {@code FUCOM ST(i)}<p>
	 * {@code DD E0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FUCOM_ST0_STI = 621;
	/**
	 * {@code FUCOMP ST(i)}<p>
	 * {@code DD E8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FUCOMP_ST0_STI = 622;
	/**
	 * {@code FIADD m16int}<p>
	 * {@code DE /0}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FIADD_M16INT = 623;
	/**
	 * {@code FIMUL m16int}<p>
	 * {@code DE /1}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FIMUL_M16INT = 624;
	/**
	 * {@code FICOM m16int}<p>
	 * {@code DE /2}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FICOM_M16INT = 625;
	/**
	 * {@code FICOMP m16int}<p>
	 * {@code DE /3}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FICOMP_M16INT = 626;
	/**
	 * {@code FISUB m16int}<p>
	 * {@code DE /4}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FISUB_M16INT = 627;
	/**
	 * {@code FISUBR m16int}<p>
	 * {@code DE /5}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FISUBR_M16INT = 628;
	/**
	 * {@code FIDIV m16int}<p>
	 * {@code DE /6}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FIDIV_M16INT = 629;
	/**
	 * {@code FIDIVR m16int}<p>
	 * {@code DE /7}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FIDIVR_M16INT = 630;
	/**
	 * {@code FADDP ST(i), ST(0)}<p>
	 * {@code DE C0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FADDP_STI_ST0 = 631;
	/**
	 * {@code FMULP ST(i), ST(0)}<p>
	 * {@code DE C8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FMULP_STI_ST0 = 632;
	/**
	 * {@code FCOMP ST(i)}<p>
	 * {@code DE D0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCOMP_ST0_STI_DED0 = 633;
	/**
	 * {@code FCOMPP}<p>
	 * {@code DE D9}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCOMPP = 634;
	/**
	 * {@code FSUBRP ST(i), ST(0)}<p>
	 * {@code DE E0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSUBRP_STI_ST0 = 635;
	/**
	 * {@code FSUBP ST(i), ST(0)}<p>
	 * {@code DE E8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSUBP_STI_ST0 = 636;
	/**
	 * {@code FDIVRP ST(i), ST(0)}<p>
	 * {@code DE F0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FDIVRP_STI_ST0 = 637;
	/**
	 * {@code FDIVP ST(i), ST(0)}<p>
	 * {@code DE F8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FDIVP_STI_ST0 = 638;
	/**
	 * {@code FILD m16int}<p>
	 * {@code DF /0}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FILD_M16INT = 639;
	/**
	 * {@code FISTTP m16int}<p>
	 * {@code DF /1}<p>
	 * {@code 8087+ and SSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FISTTP_M16INT = 640;
	/**
	 * {@code FIST m16int}<p>
	 * {@code DF /2}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FIST_M16INT = 641;
	/**
	 * {@code FISTP m16int}<p>
	 * {@code DF /3}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FISTP_M16INT = 642;
	/**
	 * {@code FBLD m80bcd}<p>
	 * {@code DF /4}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FBLD_M80BCD = 643;
	/**
	 * {@code FILD m64int}<p>
	 * {@code DF /5}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FILD_M64INT = 644;
	/**
	 * {@code FBSTP m80bcd}<p>
	 * {@code DF /6}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FBSTP_M80BCD = 645;
	/**
	 * {@code FISTP m64int}<p>
	 * {@code DF /7}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FISTP_M64INT = 646;
	/**
	 * {@code FFREEP ST(i)}<p>
	 * {@code DF C0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FFREEP_STI = 647;
	/**
	 * {@code FXCH ST(i)}<p>
	 * {@code DF C8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FXCH_ST0_STI_DFC8 = 648;
	/**
	 * {@code FSTP ST(i)}<p>
	 * {@code DF D0+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSTP_STI_DFD0 = 649;
	/**
	 * {@code FSTP ST(i)}<p>
	 * {@code DF D8+i}<p>
	 * {@code 8087+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSTP_STI_DFD8 = 650;
	/**
	 * {@code FNSTSW AX}<p>
	 * {@code DF E0}<p>
	 * {@code 287+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FNSTSW_AX = 651;
	/**
	 * {@code FSTSW AX}<p>
	 * {@code 9B DF E0}<p>
	 * {@code 287+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FSTSW_AX = 652;
	/**
	 * {@code FSTDW AX}<p>
	 * {@code 9B DF E1}<p>
	 * {@code 387 SL}<p>
	 * {@code 16/32-bit}
	 */
	public static final int FSTDW_AX = 653;
	/**
	 * {@code FSTSG AX}<p>
	 * {@code 9B DF E2}<p>
	 * {@code 387 SL}<p>
	 * {@code 16/32-bit}
	 */
	public static final int FSTSG_AX = 654;
	/**
	 * {@code FUCOMIP ST, ST(i)}<p>
	 * {@code DF E8+i}<p>
	 * {@code 8087+ and CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FUCOMIP_ST0_STI = 655;
	/**
	 * {@code FCOMIP ST, ST(i)}<p>
	 * {@code DF F0+i}<p>
	 * {@code 8087+ and CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FCOMIP_ST0_STI = 656;
	/**
	 * {@code LOOPNE rel8}<p>
	 * {@code a16 o16 E0 cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LOOPNE_REL8_16_CX = 657;
	/**
	 * {@code LOOPNE rel8}<p>
	 * {@code a16 o32 E0 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LOOPNE_REL8_32_CX = 658;
	/**
	 * {@code LOOPNE rel8}<p>
	 * {@code a32 o16 E0 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LOOPNE_REL8_16_ECX = 659;
	/**
	 * {@code LOOPNE rel8}<p>
	 * {@code a32 o32 E0 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LOOPNE_REL8_32_ECX = 660;
	/**
	 * {@code LOOPNE rel8}<p>
	 * {@code a32 o64 E0 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LOOPNE_REL8_64_ECX = 661;
	/**
	 * {@code LOOPNE rel8}<p>
	 * {@code a64 o16 E0 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LOOPNE_REL8_16_RCX = 662;
	/**
	 * {@code LOOPNE rel8}<p>
	 * {@code a64 o64 E0 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LOOPNE_REL8_64_RCX = 663;
	/**
	 * {@code LOOPE rel8}<p>
	 * {@code a16 o16 E1 cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LOOPE_REL8_16_CX = 664;
	/**
	 * {@code LOOPE rel8}<p>
	 * {@code a16 o32 E1 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LOOPE_REL8_32_CX = 665;
	/**
	 * {@code LOOPE rel8}<p>
	 * {@code a32 o16 E1 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LOOPE_REL8_16_ECX = 666;
	/**
	 * {@code LOOPE rel8}<p>
	 * {@code a32 o32 E1 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LOOPE_REL8_32_ECX = 667;
	/**
	 * {@code LOOPE rel8}<p>
	 * {@code a32 o64 E1 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LOOPE_REL8_64_ECX = 668;
	/**
	 * {@code LOOPE rel8}<p>
	 * {@code a64 o16 E1 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LOOPE_REL8_16_RCX = 669;
	/**
	 * {@code LOOPE rel8}<p>
	 * {@code a64 o64 E1 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LOOPE_REL8_64_RCX = 670;
	/**
	 * {@code LOOP rel8}<p>
	 * {@code a16 o16 E2 cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LOOP_REL8_16_CX = 671;
	/**
	 * {@code LOOP rel8}<p>
	 * {@code a16 o32 E2 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LOOP_REL8_32_CX = 672;
	/**
	 * {@code LOOP rel8}<p>
	 * {@code a32 o16 E2 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LOOP_REL8_16_ECX = 673;
	/**
	 * {@code LOOP rel8}<p>
	 * {@code a32 o32 E2 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LOOP_REL8_32_ECX = 674;
	/**
	 * {@code LOOP rel8}<p>
	 * {@code a32 o64 E2 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LOOP_REL8_64_ECX = 675;
	/**
	 * {@code LOOP rel8}<p>
	 * {@code a64 o16 E2 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LOOP_REL8_16_RCX = 676;
	/**
	 * {@code LOOP rel8}<p>
	 * {@code a64 o64 E2 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LOOP_REL8_64_RCX = 677;
	/**
	 * {@code JCXZ rel8}<p>
	 * {@code a16 o16 E3 cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JCXZ_REL8_16 = 678;
	/**
	 * {@code JCXZ rel8}<p>
	 * {@code a16 o32 E3 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JCXZ_REL8_32 = 679;
	/**
	 * {@code JECXZ rel8}<p>
	 * {@code a32 o16 E3 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JECXZ_REL8_16 = 680;
	/**
	 * {@code JECXZ rel8}<p>
	 * {@code a32 o32 E3 cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JECXZ_REL8_32 = 681;
	/**
	 * {@code JECXZ rel8}<p>
	 * {@code a32 o64 E3 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JECXZ_REL8_64 = 682;
	/**
	 * {@code JRCXZ rel8}<p>
	 * {@code a64 o16 E3 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JRCXZ_REL8_16 = 683;
	/**
	 * {@code JRCXZ rel8}<p>
	 * {@code a64 o64 E3 cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JRCXZ_REL8_64 = 684;
	/**
	 * {@code IN AL, imm8}<p>
	 * {@code E4 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IN_AL_IMM8 = 685;
	/**
	 * {@code IN AX, imm8}<p>
	 * {@code o16 E5 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IN_AX_IMM8 = 686;
	/**
	 * {@code IN EAX, imm8}<p>
	 * {@code o32 E5 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IN_EAX_IMM8 = 687;
	/**
	 * {@code OUT imm8, AL}<p>
	 * {@code E6 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OUT_IMM8_AL = 688;
	/**
	 * {@code OUT imm8, AX}<p>
	 * {@code o16 E7 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OUT_IMM8_AX = 689;
	/**
	 * {@code OUT imm8, EAX}<p>
	 * {@code o32 E7 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OUT_IMM8_EAX = 690;
	/**
	 * {@code CALL rel16}<p>
	 * {@code o16 E8 cw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CALL_REL16 = 691;
	/**
	 * {@code CALL rel32}<p>
	 * {@code o32 E8 cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CALL_REL32_32 = 692;
	/**
	 * {@code CALL rel32}<p>
	 * {@code o64 E8 cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int CALL_REL32_64 = 693;
	/**
	 * {@code JMP rel16}<p>
	 * {@code o16 E9 cw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JMP_REL16 = 694;
	/**
	 * {@code JMP rel32}<p>
	 * {@code o32 E9 cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JMP_REL32_32 = 695;
	/**
	 * {@code JMP rel32}<p>
	 * {@code o64 E9 cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JMP_REL32_64 = 696;
	/**
	 * {@code JMP ptr16:16}<p>
	 * {@code o16 EA cd}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JMP_PTR1616 = 697;
	/**
	 * {@code JMP ptr16:32}<p>
	 * {@code o32 EA cp}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JMP_PTR1632 = 698;
	/**
	 * {@code JMP rel8}<p>
	 * {@code o16 EB cb}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JMP_REL8_16 = 699;
	/**
	 * {@code JMP rel8}<p>
	 * {@code o32 EB cb}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JMP_REL8_32 = 700;
	/**
	 * {@code JMP rel8}<p>
	 * {@code o64 EB cb}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JMP_REL8_64 = 701;
	/**
	 * {@code IN AL, DX}<p>
	 * {@code EC}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IN_AL_DX = 702;
	/**
	 * {@code IN AX, DX}<p>
	 * {@code o16 ED}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IN_AX_DX = 703;
	/**
	 * {@code IN EAX, DX}<p>
	 * {@code o32 ED}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IN_EAX_DX = 704;
	/**
	 * {@code OUT DX, AL}<p>
	 * {@code EE}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OUT_DX_AL = 705;
	/**
	 * {@code OUT DX, AX}<p>
	 * {@code o16 EF}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OUT_DX_AX = 706;
	/**
	 * {@code OUT DX, EAX}<p>
	 * {@code o32 EF}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int OUT_DX_EAX = 707;
	/**
	 * {@code INT1}<p>
	 * {@code F1}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INT1 = 708;
	/**
	 * {@code HLT}<p>
	 * {@code F4}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int HLT = 709;
	/**
	 * {@code CMC}<p>
	 * {@code F5}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMC = 710;
	/**
	 * {@code TEST r/m8, imm8}<p>
	 * {@code F6 /0 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TEST_RM8_IMM8 = 711;
	/**
	 * {@code TEST r/m8, imm8}<p>
	 * {@code F6 /1 ib}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TEST_RM8_IMM8_F6R1 = 712;
	/**
	 * {@code NOT r/m8}<p>
	 * {@code F6 /2}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int NOT_RM8 = 713;
	/**
	 * {@code NEG r/m8}<p>
	 * {@code F6 /3}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int NEG_RM8 = 714;
	/**
	 * {@code MUL r/m8}<p>
	 * {@code F6 /4}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MUL_RM8 = 715;
	/**
	 * {@code IMUL r/m8}<p>
	 * {@code F6 /5}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IMUL_RM8 = 716;
	/**
	 * {@code DIV r/m8}<p>
	 * {@code F6 /6}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int DIV_RM8 = 717;
	/**
	 * {@code IDIV r/m8}<p>
	 * {@code F6 /7}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IDIV_RM8 = 718;
	/**
	 * {@code TEST r/m16, imm16}<p>
	 * {@code o16 F7 /0 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TEST_RM16_IMM16 = 719;
	/**
	 * {@code TEST r/m32, imm32}<p>
	 * {@code o32 F7 /0 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TEST_RM32_IMM32 = 720;
	/**
	 * {@code TEST r/m64, imm32}<p>
	 * {@code o64 F7 /0 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int TEST_RM64_IMM32 = 721;
	/**
	 * {@code TEST r/m16, imm16}<p>
	 * {@code o16 F7 /1 iw}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TEST_RM16_IMM16_F7R1 = 722;
	/**
	 * {@code TEST r/m32, imm32}<p>
	 * {@code o32 F7 /1 id}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TEST_RM32_IMM32_F7R1 = 723;
	/**
	 * {@code TEST r/m64, imm32}<p>
	 * {@code o64 F7 /1 id}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int TEST_RM64_IMM32_F7R1 = 724;
	/**
	 * {@code NOT r/m16}<p>
	 * {@code o16 F7 /2}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int NOT_RM16 = 725;
	/**
	 * {@code NOT r/m32}<p>
	 * {@code o32 F7 /2}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int NOT_RM32 = 726;
	/**
	 * {@code NOT r/m64}<p>
	 * {@code o64 F7 /2}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int NOT_RM64 = 727;
	/**
	 * {@code NEG r/m16}<p>
	 * {@code o16 F7 /3}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int NEG_RM16 = 728;
	/**
	 * {@code NEG r/m32}<p>
	 * {@code o32 F7 /3}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int NEG_RM32 = 729;
	/**
	 * {@code NEG r/m64}<p>
	 * {@code o64 F7 /3}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int NEG_RM64 = 730;
	/**
	 * {@code MUL r/m16}<p>
	 * {@code o16 F7 /4}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MUL_RM16 = 731;
	/**
	 * {@code MUL r/m32}<p>
	 * {@code o32 F7 /4}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MUL_RM32 = 732;
	/**
	 * {@code MUL r/m64}<p>
	 * {@code o64 F7 /4}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MUL_RM64 = 733;
	/**
	 * {@code IMUL r/m16}<p>
	 * {@code o16 F7 /5}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IMUL_RM16 = 734;
	/**
	 * {@code IMUL r/m32}<p>
	 * {@code o32 F7 /5}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IMUL_RM32 = 735;
	/**
	 * {@code IMUL r/m64}<p>
	 * {@code o64 F7 /5}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int IMUL_RM64 = 736;
	/**
	 * {@code DIV r/m16}<p>
	 * {@code o16 F7 /6}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int DIV_RM16 = 737;
	/**
	 * {@code DIV r/m32}<p>
	 * {@code o32 F7 /6}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int DIV_RM32 = 738;
	/**
	 * {@code DIV r/m64}<p>
	 * {@code o64 F7 /6}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int DIV_RM64 = 739;
	/**
	 * {@code IDIV r/m16}<p>
	 * {@code o16 F7 /7}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IDIV_RM16 = 740;
	/**
	 * {@code IDIV r/m32}<p>
	 * {@code o32 F7 /7}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IDIV_RM32 = 741;
	/**
	 * {@code IDIV r/m64}<p>
	 * {@code o64 F7 /7}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int IDIV_RM64 = 742;
	/**
	 * {@code CLC}<p>
	 * {@code F8}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CLC = 743;
	/**
	 * {@code STC}<p>
	 * {@code F9}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int STC = 744;
	/**
	 * {@code CLI}<p>
	 * {@code FA}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CLI = 745;
	/**
	 * {@code STI}<p>
	 * {@code FB}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int STI = 746;
	/**
	 * {@code CLD}<p>
	 * {@code FC}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CLD = 747;
	/**
	 * {@code STD}<p>
	 * {@code FD}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int STD = 748;
	/**
	 * {@code INC r/m8}<p>
	 * {@code FE /0}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INC_RM8 = 749;
	/**
	 * {@code DEC r/m8}<p>
	 * {@code FE /1}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int DEC_RM8 = 750;
	/**
	 * {@code INC r/m16}<p>
	 * {@code o16 FF /0}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INC_RM16 = 751;
	/**
	 * {@code INC r/m32}<p>
	 * {@code o32 FF /0}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INC_RM32 = 752;
	/**
	 * {@code INC r/m64}<p>
	 * {@code o64 FF /0}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int INC_RM64 = 753;
	/**
	 * {@code DEC r/m16}<p>
	 * {@code o16 FF /1}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int DEC_RM16 = 754;
	/**
	 * {@code DEC r/m32}<p>
	 * {@code o32 FF /1}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int DEC_RM32 = 755;
	/**
	 * {@code DEC r/m64}<p>
	 * {@code o64 FF /1}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int DEC_RM64 = 756;
	/**
	 * {@code CALL r/m16}<p>
	 * {@code o16 FF /2}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CALL_RM16 = 757;
	/**
	 * {@code CALL r/m32}<p>
	 * {@code o32 FF /2}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CALL_RM32 = 758;
	/**
	 * {@code CALL r/m64}<p>
	 * {@code o64 FF /2}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int CALL_RM64 = 759;
	/**
	 * {@code CALL m16:16}<p>
	 * {@code o16 FF /3}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CALL_M1616 = 760;
	/**
	 * {@code CALL m16:32}<p>
	 * {@code o32 FF /3}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CALL_M1632 = 761;
	/**
	 * {@code CALL m16:64}<p>
	 * {@code o64 FF /3}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int CALL_M1664 = 762;
	/**
	 * {@code JMP r/m16}<p>
	 * {@code o16 FF /4}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JMP_RM16 = 763;
	/**
	 * {@code JMP r/m32}<p>
	 * {@code o32 FF /4}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JMP_RM32 = 764;
	/**
	 * {@code JMP r/m64}<p>
	 * {@code o64 FF /4}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JMP_RM64 = 765;
	/**
	 * {@code JMP m16:16}<p>
	 * {@code o16 FF /5}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JMP_M1616 = 766;
	/**
	 * {@code JMP m16:32}<p>
	 * {@code o32 FF /5}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JMP_M1632 = 767;
	/**
	 * {@code JMP m16:64}<p>
	 * {@code o64 FF /5}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JMP_M1664 = 768;
	/**
	 * {@code PUSH r/m16}<p>
	 * {@code o16 FF /6}<p>
	 * {@code 8086+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUSH_RM16 = 769;
	/**
	 * {@code PUSH r/m32}<p>
	 * {@code o32 FF /6}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSH_RM32 = 770;
	/**
	 * {@code PUSH r/m64}<p>
	 * {@code o64 FF /6}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int PUSH_RM64 = 771;
	/**
	 * {@code SLDT r/m16}<p>
	 * {@code o16 0F 00 /0}<p>
	 * {@code 286+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SLDT_RM16 = 772;
	/**
	 * {@code SLDT r32/m16}<p>
	 * {@code o32 0F 00 /0}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SLDT_R32M16 = 773;
	/**
	 * {@code SLDT r64/m16}<p>
	 * {@code o64 0F 00 /0}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SLDT_R64M16 = 774;
	/**
	 * {@code STR r/m16}<p>
	 * {@code o16 0F 00 /1}<p>
	 * {@code 286+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int STR_RM16 = 775;
	/**
	 * {@code STR r32/m16}<p>
	 * {@code o32 0F 00 /1}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int STR_R32M16 = 776;
	/**
	 * {@code STR r64/m16}<p>
	 * {@code o64 0F 00 /1}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int STR_R64M16 = 777;
	/**
	 * {@code LLDT r/m16}<p>
	 * {@code o16 0F 00 /2}<p>
	 * {@code 286+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LLDT_RM16 = 778;
	/**
	 * {@code LLDT r32/m16}<p>
	 * {@code o32 0F 00 /2}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LLDT_R32M16 = 779;
	/**
	 * {@code LLDT r64/m16}<p>
	 * {@code o64 0F 00 /2}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LLDT_R64M16 = 780;
	/**
	 * {@code LTR r/m16}<p>
	 * {@code o16 0F 00 /3}<p>
	 * {@code 286+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LTR_RM16 = 781;
	/**
	 * {@code LTR r32/m16}<p>
	 * {@code o32 0F 00 /3}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LTR_R32M16 = 782;
	/**
	 * {@code LTR r64/m16}<p>
	 * {@code o64 0F 00 /3}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LTR_R64M16 = 783;
	/**
	 * {@code VERR r/m16}<p>
	 * {@code o16 0F 00 /4}<p>
	 * {@code 286+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VERR_RM16 = 784;
	/**
	 * {@code VERR r32/m16}<p>
	 * {@code o32 0F 00 /4}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VERR_R32M16 = 785;
	/**
	 * {@code VERR r64/m16}<p>
	 * {@code o64 0F 00 /4}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int VERR_R64M16 = 786;
	/**
	 * {@code VERW r/m16}<p>
	 * {@code o16 0F 00 /5}<p>
	 * {@code 286+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VERW_RM16 = 787;
	/**
	 * {@code VERW r32/m16}<p>
	 * {@code o32 0F 00 /5}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VERW_R32M16 = 788;
	/**
	 * {@code VERW r64/m16}<p>
	 * {@code o64 0F 00 /5}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int VERW_R64M16 = 789;
	/**
	 * {@code JMPE r/m16}<p>
	 * {@code o16 0F 00 /6}<p>
	 * {@code IA-64}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JMPE_RM16 = 790;
	/**
	 * {@code JMPE r/m32}<p>
	 * {@code o32 0F 00 /6}<p>
	 * {@code IA-64}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JMPE_RM32 = 791;
	/**
	 * {@code SGDT m}<p>
	 * {@code o16 0F 01 /0}<p>
	 * {@code 286+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int SGDT_M1632_16 = 792;
	/**
	 * {@code SGDT m}<p>
	 * {@code o32 0F 01 /0}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int SGDT_M1632 = 793;
	/**
	 * {@code SGDT m}<p>
	 * {@code 0F 01 /0}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SGDT_M1664 = 794;
	/**
	 * {@code SIDT m}<p>
	 * {@code o16 0F 01 /1}<p>
	 * {@code 286+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int SIDT_M1632_16 = 795;
	/**
	 * {@code SIDT m}<p>
	 * {@code o32 0F 01 /1}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int SIDT_M1632 = 796;
	/**
	 * {@code SIDT m}<p>
	 * {@code 0F 01 /1}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SIDT_M1664 = 797;
	/**
	 * {@code LGDT m16&amp;32}<p>
	 * {@code o16 0F 01 /2}<p>
	 * {@code 286+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LGDT_M1632_16 = 798;
	/**
	 * {@code LGDT m16&amp;32}<p>
	 * {@code o32 0F 01 /2}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LGDT_M1632 = 799;
	/**
	 * {@code LGDT m16&amp;64}<p>
	 * {@code 0F 01 /2}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LGDT_M1664 = 800;
	/**
	 * {@code LIDT m16&amp;32}<p>
	 * {@code o16 0F 01 /3}<p>
	 * {@code 286+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LIDT_M1632_16 = 801;
	/**
	 * {@code LIDT m16&amp;32}<p>
	 * {@code o32 0F 01 /3}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LIDT_M1632 = 802;
	/**
	 * {@code LIDT m16&amp;64}<p>
	 * {@code 0F 01 /3}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LIDT_M1664 = 803;
	/**
	 * {@code SMSW r/m16}<p>
	 * {@code o16 0F 01 /4}<p>
	 * {@code 286+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SMSW_RM16 = 804;
	/**
	 * {@code SMSW r32/m16}<p>
	 * {@code o32 0F 01 /4}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SMSW_R32M16 = 805;
	/**
	 * {@code SMSW r64/m16}<p>
	 * {@code o64 0F 01 /4}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SMSW_R64M16 = 806;
	/**
	 * {@code RSTORSSP m64}<p>
	 * {@code F3 0F 01 /5}<p>
	 * {@code CET_SS}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RSTORSSP_M64 = 807;
	/**
	 * {@code LMSW r/m16}<p>
	 * {@code o16 0F 01 /6}<p>
	 * {@code 286+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LMSW_RM16 = 808;
	/**
	 * {@code LMSW r32/m16}<p>
	 * {@code o32 0F 01 /6}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LMSW_R32M16 = 809;
	/**
	 * {@code LMSW r64/m16}<p>
	 * {@code o64 0F 01 /6}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LMSW_R64M16 = 810;
	/**
	 * {@code INVLPG m}<p>
	 * {@code 0F 01 /7}<p>
	 * {@code 486+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INVLPG_M = 811;
	/**
	 * {@code ENCLV}<p>
	 * {@code NP 0F 01 C0}<p>
	 * {@code OSS}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ENCLV = 812;
	/**
	 * {@code VMCALL}<p>
	 * {@code NP 0F 01 C1}<p>
	 * {@code VMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VMCALL = 813;
	/**
	 * {@code VMLAUNCH}<p>
	 * {@code NP 0F 01 C2}<p>
	 * {@code VMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VMLAUNCH = 814;
	/**
	 * {@code VMRESUME}<p>
	 * {@code NP 0F 01 C3}<p>
	 * {@code VMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VMRESUME = 815;
	/**
	 * {@code VMXOFF}<p>
	 * {@code NP 0F 01 C4}<p>
	 * {@code VMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VMXOFF = 816;
	/**
	 * {@code PCONFIG}<p>
	 * {@code NP 0F 01 C5}<p>
	 * {@code PCONFIG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCONFIG = 817;
	/**
	 * {@code MONITOR}<p>
	 * {@code a16 NP 0F 01 C8}<p>
	 * {@code MONITOR}<p>
	 * {@code 16/32-bit}
	 */
	public static final int MONITORW = 818;
	/**
	 * {@code MONITOR}<p>
	 * {@code a32 NP 0F 01 C8}<p>
	 * {@code MONITOR}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MONITORD = 819;
	/**
	 * {@code MONITOR}<p>
	 * {@code a64 NP 0F 01 C8}<p>
	 * {@code MONITOR}<p>
	 * {@code 64-bit}
	 */
	public static final int MONITORQ = 820;
	/**
	 * {@code MWAIT}<p>
	 * {@code NP 0F 01 C9}<p>
	 * {@code MONITOR}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MWAIT = 821;
	/**
	 * {@code CLAC}<p>
	 * {@code NP 0F 01 CA}<p>
	 * {@code SMAP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CLAC = 822;
	/**
	 * {@code STAC}<p>
	 * {@code NP 0F 01 CB}<p>
	 * {@code SMAP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int STAC = 823;
	/**
	 * {@code ENCLS}<p>
	 * {@code NP 0F 01 CF}<p>
	 * {@code SGX1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ENCLS = 824;
	/**
	 * {@code XGETBV}<p>
	 * {@code NP 0F 01 D0}<p>
	 * {@code XSAVE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XGETBV = 825;
	/**
	 * {@code XSETBV}<p>
	 * {@code NP 0F 01 D1}<p>
	 * {@code XSAVE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XSETBV = 826;
	/**
	 * {@code VMFUNC}<p>
	 * {@code NP 0F 01 D4}<p>
	 * {@code VMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VMFUNC = 827;
	/**
	 * {@code XEND}<p>
	 * {@code NP 0F 01 D5}<p>
	 * {@code RTM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XEND = 828;
	/**
	 * {@code XTEST}<p>
	 * {@code NP 0F 01 D6}<p>
	 * {@code HLE or RTM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XTEST = 829;
	/**
	 * {@code ENCLU}<p>
	 * {@code NP 0F 01 D7}<p>
	 * {@code SGX1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ENCLU = 830;
	/**
	 * {@code VMRUN}<p>
	 * {@code a16 0F 01 D8}<p>
	 * {@code SVM}<p>
	 * {@code 16/32-bit}
	 */
	public static final int VMRUNW = 831;
	/**
	 * {@code VMRUN}<p>
	 * {@code a32 0F 01 D8}<p>
	 * {@code SVM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VMRUND = 832;
	/**
	 * {@code VMRUN}<p>
	 * {@code a64 0F 01 D8}<p>
	 * {@code SVM}<p>
	 * {@code 64-bit}
	 */
	public static final int VMRUNQ = 833;
	/**
	 * {@code VMMCALL}<p>
	 * {@code 0F 01 D9}<p>
	 * {@code SVM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VMMCALL = 834;
	/**
	 * {@code VMLOAD}<p>
	 * {@code a16 0F 01 DA}<p>
	 * {@code SVM}<p>
	 * {@code 16/32-bit}
	 */
	public static final int VMLOADW = 835;
	/**
	 * {@code VMLOAD}<p>
	 * {@code a32 0F 01 DA}<p>
	 * {@code SVM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VMLOADD = 836;
	/**
	 * {@code VMLOAD}<p>
	 * {@code a64 0F 01 DA}<p>
	 * {@code SVM}<p>
	 * {@code 64-bit}
	 */
	public static final int VMLOADQ = 837;
	/**
	 * {@code VMSAVE}<p>
	 * {@code a16 0F 01 DB}<p>
	 * {@code SVM}<p>
	 * {@code 16/32-bit}
	 */
	public static final int VMSAVEW = 838;
	/**
	 * {@code VMSAVE}<p>
	 * {@code a32 0F 01 DB}<p>
	 * {@code SVM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VMSAVED = 839;
	/**
	 * {@code VMSAVE}<p>
	 * {@code a64 0F 01 DB}<p>
	 * {@code SVM}<p>
	 * {@code 64-bit}
	 */
	public static final int VMSAVEQ = 840;
	/**
	 * {@code STGI}<p>
	 * {@code 0F 01 DC}<p>
	 * {@code SKINIT or SVM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int STGI = 841;
	/**
	 * {@code CLGI}<p>
	 * {@code 0F 01 DD}<p>
	 * {@code SVM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CLGI = 842;
	/**
	 * {@code SKINIT}<p>
	 * {@code 0F 01 DE}<p>
	 * {@code SKINIT or SVM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SKINIT = 843;
	/**
	 * {@code INVLPGA}<p>
	 * {@code a16 0F 01 DF}<p>
	 * {@code SVM}<p>
	 * {@code 16/32-bit}
	 */
	public static final int INVLPGAW = 844;
	/**
	 * {@code INVLPGA}<p>
	 * {@code a32 0F 01 DF}<p>
	 * {@code SVM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INVLPGAD = 845;
	/**
	 * {@code INVLPGA}<p>
	 * {@code a64 0F 01 DF}<p>
	 * {@code SVM}<p>
	 * {@code 64-bit}
	 */
	public static final int INVLPGAQ = 846;
	/**
	 * {@code SETSSBSY}<p>
	 * {@code F3 0F 01 E8}<p>
	 * {@code CET_SS}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETSSBSY = 847;
	/**
	 * {@code SAVEPREVSSP}<p>
	 * {@code F3 0F 01 EA}<p>
	 * {@code CET_SS}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SAVEPREVSSP = 848;
	/**
	 * {@code RDPKRU}<p>
	 * {@code NP 0F 01 EE}<p>
	 * {@code PKU}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RDPKRU = 849;
	/**
	 * {@code WRPKRU}<p>
	 * {@code NP 0F 01 EF}<p>
	 * {@code PKU}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int WRPKRU = 850;
	/**
	 * {@code SWAPGS}<p>
	 * {@code 0F 01 F8}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SWAPGS = 851;
	/**
	 * {@code RDTSCP}<p>
	 * {@code 0F 01 F9}<p>
	 * {@code RDTSCP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RDTSCP = 852;
	/**
	 * {@code MONITORX}<p>
	 * {@code a16 NP 0F 01 FA}<p>
	 * {@code MONITORX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int MONITORXW = 853;
	/**
	 * {@code MONITORX}<p>
	 * {@code a32 NP 0F 01 FA}<p>
	 * {@code MONITORX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MONITORXD = 854;
	/**
	 * {@code MONITORX}<p>
	 * {@code a64 NP 0F 01 FA}<p>
	 * {@code MONITORX}<p>
	 * {@code 64-bit}
	 */
	public static final int MONITORXQ = 855;
	/**
	 * {@code MCOMMIT}<p>
	 * {@code F3 0F 01 FA}<p>
	 * {@code MCOMMIT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MCOMMIT = 856;
	/**
	 * {@code MWAITX}<p>
	 * {@code NP 0F 01 FB}<p>
	 * {@code MONITORX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MWAITX = 857;
	/**
	 * {@code CLZERO}<p>
	 * {@code a16 0F 01 FC}<p>
	 * {@code CLZERO}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CLZEROW = 858;
	/**
	 * {@code CLZERO}<p>
	 * {@code a32 0F 01 FC}<p>
	 * {@code CLZERO}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CLZEROD = 859;
	/**
	 * {@code CLZERO}<p>
	 * {@code a64 0F 01 FC}<p>
	 * {@code CLZERO}<p>
	 * {@code 64-bit}
	 */
	public static final int CLZEROQ = 860;
	/**
	 * {@code RDPRU}<p>
	 * {@code NP 0F 01 FD}<p>
	 * {@code RDPRU}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RDPRU = 861;
	/**
	 * {@code LAR r16, r/m16}<p>
	 * {@code o16 0F 02 /r}<p>
	 * {@code 286+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LAR_R16_RM16 = 862;
	/**
	 * {@code LAR r32, r32/m16}<p>
	 * {@code o32 0F 02 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LAR_R32_R32M16 = 863;
	/**
	 * {@code LAR r64, r64/m16}<p>
	 * {@code o64 0F 02 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LAR_R64_R64M16 = 864;
	/**
	 * {@code LSL r16, r/m16}<p>
	 * {@code o16 0F 03 /r}<p>
	 * {@code 286+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LSL_R16_RM16 = 865;
	/**
	 * {@code LSL r32, r32/m16}<p>
	 * {@code o32 0F 03 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LSL_R32_R32M16 = 866;
	/**
	 * {@code LSL r64, r64/m16}<p>
	 * {@code o64 0F 03 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LSL_R64_R64M16 = 867;
	/**
	 * {@code STOREALL}<p>
	 * {@code 0F 04}<p>
	 * {@code 286}<p>
	 * {@code 16/32-bit}
	 */
	public static final int STOREALL = 868;
	/**
	 * {@code LOADALL}<p>
	 * {@code 0F 05}<p>
	 * {@code 286}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LOADALL286 = 869;
	/**
	 * {@code SYSCALL}<p>
	 * {@code 0F 05}<p>
	 * {@code SYSCALL}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SYSCALL = 870;
	/**
	 * {@code CLTS}<p>
	 * {@code 0F 06}<p>
	 * {@code 286+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CLTS = 871;
	/**
	 * {@code LOADALL}<p>
	 * {@code 0F 07}<p>
	 * {@code 386}<p>
	 * {@code 16/32-bit}
	 */
	public static final int LOADALL386 = 872;
	/**
	 * {@code SYSRET}<p>
	 * {@code 0F 07}<p>
	 * {@code SYSCALL}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SYSRETD = 873;
	/**
	 * {@code SYSRETQ}<p>
	 * {@code o64 0F 07}<p>
	 * {@code SYSCALL}<p>
	 * {@code 64-bit}
	 */
	public static final int SYSRETQ = 874;
	/**
	 * {@code INVD}<p>
	 * {@code 0F 08}<p>
	 * {@code 486+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INVD = 875;
	/**
	 * {@code WBINVD}<p>
	 * {@code 0F 09}<p>
	 * {@code 486+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int WBINVD = 876;
	/**
	 * {@code WBNOINVD}<p>
	 * {@code F3 0F 09}<p>
	 * {@code WBNOINVD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int WBNOINVD = 877;
	/**
	 * {@code CL1INVMB}<p>
	 * {@code 0F 0A}<p>
	 * {@code CL1INVMB}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CL1INVMB = 878;
	/**
	 * {@code UD2}<p>
	 * {@code 0F 0B}<p>
	 * {@code 286+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int UD2 = 879;
	/**
	 * {@code RESERVEDNOP r/m16, r16}<p>
	 * {@code o16 0F 0D /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM16_R16_0F0D = 880;
	/**
	 * {@code RESERVEDNOP r/m32, r32}<p>
	 * {@code o32 0F 0D /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM32_R32_0F0D = 881;
	/**
	 * {@code RESERVEDNOP r/m64, r64}<p>
	 * {@code o64 0F 0D /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 64-bit}
	 */
	public static final int RESERVEDNOP_RM64_R64_0F0D = 882;
	/**
	 * {@code PREFETCH m8}<p>
	 * {@code 0F 0D /0}<p>
	 * {@code PREFETCHW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PREFETCH_M8 = 883;
	/**
	 * {@code PREFETCHW m8}<p>
	 * {@code 0F 0D /1}<p>
	 * {@code PREFETCHW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PREFETCHW_M8 = 884;
	/**
	 * {@code PREFETCHWT1 m8}<p>
	 * {@code 0F 0D /2}<p>
	 * {@code PREFETCHWT1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PREFETCHWT1_M8 = 885;
	/**
	 * {@code FEMMS}<p>
	 * {@code 0F 0E}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FEMMS = 886;
	/**
	 * {@code UMOV r/m8, r8}<p>
	 * {@code 0F 10 /r}<p>
	 * {@code 386/486}<p>
	 * {@code 16/32-bit}
	 */
	public static final int UMOV_RM8_R8 = 887;
	/**
	 * {@code UMOV r/m16, r16}<p>
	 * {@code o16 0F 11 /r}<p>
	 * {@code 386/486}<p>
	 * {@code 16/32-bit}
	 */
	public static final int UMOV_RM16_R16 = 888;
	/**
	 * {@code UMOV r/m32, r32}<p>
	 * {@code o32 0F 11 /r}<p>
	 * {@code 386/486}<p>
	 * {@code 16/32-bit}
	 */
	public static final int UMOV_RM32_R32 = 889;
	/**
	 * {@code UMOV r8, r/m8}<p>
	 * {@code 0F 12 /r}<p>
	 * {@code 386/486}<p>
	 * {@code 16/32-bit}
	 */
	public static final int UMOV_R8_RM8 = 890;
	/**
	 * {@code UMOV r16, r/m16}<p>
	 * {@code o16 0F 13 /r}<p>
	 * {@code 386/486}<p>
	 * {@code 16/32-bit}
	 */
	public static final int UMOV_R16_RM16 = 891;
	/**
	 * {@code UMOV r32, r/m32}<p>
	 * {@code o32 0F 13 /r}<p>
	 * {@code 386/486}<p>
	 * {@code 16/32-bit}
	 */
	public static final int UMOV_R32_RM32 = 892;
	/**
	 * {@code MOVUPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 10 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVUPS_XMM_XMMM128 = 893;
	/**
	 * {@code VMOVUPS xmm1, xmm2/m128}<p>
	 * {@code VEX.128.0F.WIG 10 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVUPS_XMM_XMMM128 = 894;
	/**
	 * {@code VMOVUPS ymm1, ymm2/m256}<p>
	 * {@code VEX.256.0F.WIG 10 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVUPS_YMM_YMMM256 = 895;
	/**
	 * {@code VMOVUPS xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.0F.W0 10 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVUPS_XMM_K1Z_XMMM128 = 896;
	/**
	 * {@code VMOVUPS ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.0F.W0 10 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVUPS_YMM_K1Z_YMMM256 = 897;
	/**
	 * {@code VMOVUPS zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.0F.W0 10 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVUPS_ZMM_K1Z_ZMMM512 = 898;
	/**
	 * {@code MOVUPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 10 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVUPD_XMM_XMMM128 = 899;
	/**
	 * {@code VMOVUPD xmm1, xmm2/m128}<p>
	 * {@code VEX.128.66.0F.WIG 10 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVUPD_XMM_XMMM128 = 900;
	/**
	 * {@code VMOVUPD ymm1, ymm2/m256}<p>
	 * {@code VEX.256.66.0F.WIG 10 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVUPD_YMM_YMMM256 = 901;
	/**
	 * {@code VMOVUPD xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.66.0F.W1 10 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVUPD_XMM_K1Z_XMMM128 = 902;
	/**
	 * {@code VMOVUPD ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.66.0F.W1 10 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVUPD_YMM_K1Z_YMMM256 = 903;
	/**
	 * {@code VMOVUPD zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.66.0F.W1 10 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVUPD_ZMM_K1Z_ZMMM512 = 904;
	/**
	 * {@code MOVSS xmm1, xmm2/m32}<p>
	 * {@code F3 0F 10 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVSS_XMM_XMMM32 = 905;
	/**
	 * {@code VMOVSS xmm1, xmm2, xmm3}<p>
	 * {@code VEX.LIG.F3.0F.WIG 10 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVSS_XMM_XMM_XMM = 906;
	/**
	 * {@code VMOVSS xmm1, m32}<p>
	 * {@code VEX.LIG.F3.0F.WIG 10 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVSS_XMM_M32 = 907;
	/**
	 * {@code VMOVSS xmm1 {k1}{z}, xmm2, xmm3}<p>
	 * {@code EVEX.LIG.F3.0F.W0 10 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSS_XMM_K1Z_XMM_XMM = 908;
	/**
	 * {@code VMOVSS xmm1 {k1}{z}, m32}<p>
	 * {@code EVEX.LIG.F3.0F.W0 10 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSS_XMM_K1Z_M32 = 909;
	/**
	 * {@code MOVSD xmm1, xmm2/m64}<p>
	 * {@code F2 0F 10 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVSD_XMM_XMMM64 = 910;
	/**
	 * {@code VMOVSD xmm1, xmm2, xmm3}<p>
	 * {@code VEX.LIG.F2.0F.WIG 10 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVSD_XMM_XMM_XMM = 911;
	/**
	 * {@code VMOVSD xmm1, m64}<p>
	 * {@code VEX.LIG.F2.0F.WIG 10 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVSD_XMM_M64 = 912;
	/**
	 * {@code VMOVSD xmm1 {k1}{z}, xmm2, xmm3}<p>
	 * {@code EVEX.LIG.F2.0F.W1 10 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSD_XMM_K1Z_XMM_XMM = 913;
	/**
	 * {@code VMOVSD xmm1 {k1}{z}, m64}<p>
	 * {@code EVEX.LIG.F2.0F.W1 10 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSD_XMM_K1Z_M64 = 914;
	/**
	 * {@code MOVUPS xmm2/m128, xmm1}<p>
	 * {@code NP 0F 11 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVUPS_XMMM128_XMM = 915;
	/**
	 * {@code VMOVUPS xmm2/m128, xmm1}<p>
	 * {@code VEX.128.0F.WIG 11 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVUPS_XMMM128_XMM = 916;
	/**
	 * {@code VMOVUPS ymm2/m256, ymm1}<p>
	 * {@code VEX.256.0F.WIG 11 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVUPS_YMMM256_YMM = 917;
	/**
	 * {@code VMOVUPS xmm2/m128 {k1}{z}, xmm1}<p>
	 * {@code EVEX.128.0F.W0 11 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVUPS_XMMM128_K1Z_XMM = 918;
	/**
	 * {@code VMOVUPS ymm2/m256 {k1}{z}, ymm1}<p>
	 * {@code EVEX.256.0F.W0 11 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVUPS_YMMM256_K1Z_YMM = 919;
	/**
	 * {@code VMOVUPS zmm2/m512 {k1}{z}, zmm1}<p>
	 * {@code EVEX.512.0F.W0 11 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVUPS_ZMMM512_K1Z_ZMM = 920;
	/**
	 * {@code MOVUPD xmm2/m128, xmm1}<p>
	 * {@code 66 0F 11 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVUPD_XMMM128_XMM = 921;
	/**
	 * {@code VMOVUPD xmm2/m128, xmm1}<p>
	 * {@code VEX.128.66.0F.WIG 11 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVUPD_XMMM128_XMM = 922;
	/**
	 * {@code VMOVUPD ymm2/m256, ymm1}<p>
	 * {@code VEX.256.66.0F.WIG 11 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVUPD_YMMM256_YMM = 923;
	/**
	 * {@code VMOVUPD xmm2/m128 {k1}{z}, xmm1}<p>
	 * {@code EVEX.128.66.0F.W1 11 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVUPD_XMMM128_K1Z_XMM = 924;
	/**
	 * {@code VMOVUPD ymm2/m256 {k1}{z}, ymm1}<p>
	 * {@code EVEX.256.66.0F.W1 11 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVUPD_YMMM256_K1Z_YMM = 925;
	/**
	 * {@code VMOVUPD zmm2/m512 {k1}{z}, zmm1}<p>
	 * {@code EVEX.512.66.0F.W1 11 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVUPD_ZMMM512_K1Z_ZMM = 926;
	/**
	 * {@code MOVSS xmm2/m32, xmm1}<p>
	 * {@code F3 0F 11 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVSS_XMMM32_XMM = 927;
	/**
	 * {@code VMOVSS xmm1, xmm2, xmm3}<p>
	 * {@code VEX.LIG.F3.0F.WIG 11 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVSS_XMM_XMM_XMM_0F11 = 928;
	/**
	 * {@code VMOVSS m32, xmm1}<p>
	 * {@code VEX.LIG.F3.0F.WIG 11 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVSS_M32_XMM = 929;
	/**
	 * {@code VMOVSS xmm1 {k1}{z}, xmm2, xmm3}<p>
	 * {@code EVEX.LIG.F3.0F.W0 11 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSS_XMM_K1Z_XMM_XMM_0F11 = 930;
	/**
	 * {@code VMOVSS m32 {k1}, xmm1}<p>
	 * {@code EVEX.LIG.F3.0F.W0 11 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSS_M32_K1_XMM = 931;
	/**
	 * {@code MOVSD xmm1/m64, xmm2}<p>
	 * {@code F2 0F 11 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVSD_XMMM64_XMM = 932;
	/**
	 * {@code VMOVSD xmm1, xmm2, xmm3}<p>
	 * {@code VEX.LIG.F2.0F.WIG 11 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVSD_XMM_XMM_XMM_0F11 = 933;
	/**
	 * {@code VMOVSD m64, xmm1}<p>
	 * {@code VEX.LIG.F2.0F.WIG 11 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVSD_M64_XMM = 934;
	/**
	 * {@code VMOVSD xmm1 {k1}{z}, xmm2, xmm3}<p>
	 * {@code EVEX.LIG.F2.0F.W1 11 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSD_XMM_K1Z_XMM_XMM_0F11 = 935;
	/**
	 * {@code VMOVSD m64 {k1}, xmm1}<p>
	 * {@code EVEX.LIG.F2.0F.W1 11 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSD_M64_K1_XMM = 936;
	/**
	 * {@code MOVHLPS xmm1, xmm2}<p>
	 * {@code NP 0F 12 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVHLPS_XMM_XMM = 937;
	/**
	 * {@code MOVLPS xmm1, m64}<p>
	 * {@code NP 0F 12 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVLPS_XMM_M64 = 938;
	/**
	 * {@code VMOVHLPS xmm1, xmm2, xmm3}<p>
	 * {@code VEX.128.0F.WIG 12 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVHLPS_XMM_XMM_XMM = 939;
	/**
	 * {@code VMOVLPS xmm2, xmm1, m64}<p>
	 * {@code VEX.128.0F.WIG 12 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVLPS_XMM_XMM_M64 = 940;
	/**
	 * {@code VMOVHLPS xmm1, xmm2, xmm3}<p>
	 * {@code EVEX.128.0F.W0 12 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVHLPS_XMM_XMM_XMM = 941;
	/**
	 * {@code VMOVLPS xmm2, xmm1, m64}<p>
	 * {@code EVEX.128.0F.W0 12 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVLPS_XMM_XMM_M64 = 942;
	/**
	 * {@code MOVLPD xmm1, m64}<p>
	 * {@code 66 0F 12 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVLPD_XMM_M64 = 943;
	/**
	 * {@code VMOVLPD xmm2, xmm1, m64}<p>
	 * {@code VEX.128.66.0F.WIG 12 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVLPD_XMM_XMM_M64 = 944;
	/**
	 * {@code VMOVLPD xmm2, xmm1, m64}<p>
	 * {@code EVEX.128.66.0F.W1 12 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVLPD_XMM_XMM_M64 = 945;
	/**
	 * {@code MOVSLDUP xmm1, xmm2/m128}<p>
	 * {@code F3 0F 12 /r}<p>
	 * {@code SSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVSLDUP_XMM_XMMM128 = 946;
	/**
	 * {@code VMOVSLDUP xmm1, xmm2/m128}<p>
	 * {@code VEX.128.F3.0F.WIG 12 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVSLDUP_XMM_XMMM128 = 947;
	/**
	 * {@code VMOVSLDUP ymm1, ymm2/m256}<p>
	 * {@code VEX.256.F3.0F.WIG 12 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVSLDUP_YMM_YMMM256 = 948;
	/**
	 * {@code VMOVSLDUP xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.F3.0F.W0 12 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSLDUP_XMM_K1Z_XMMM128 = 949;
	/**
	 * {@code VMOVSLDUP ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.F3.0F.W0 12 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSLDUP_YMM_K1Z_YMMM256 = 950;
	/**
	 * {@code VMOVSLDUP zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.F3.0F.W0 12 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSLDUP_ZMM_K1Z_ZMMM512 = 951;
	/**
	 * {@code MOVDDUP xmm1, xmm2/m64}<p>
	 * {@code F2 0F 12 /r}<p>
	 * {@code SSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVDDUP_XMM_XMMM64 = 952;
	/**
	 * {@code VMOVDDUP xmm1, xmm2/m64}<p>
	 * {@code VEX.128.F2.0F.WIG 12 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVDDUP_XMM_XMMM64 = 953;
	/**
	 * {@code VMOVDDUP ymm1, ymm2/m256}<p>
	 * {@code VEX.256.F2.0F.WIG 12 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVDDUP_YMM_YMMM256 = 954;
	/**
	 * {@code VMOVDDUP xmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.128.F2.0F.W1 12 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDDUP_XMM_K1Z_XMMM64 = 955;
	/**
	 * {@code VMOVDDUP ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.F2.0F.W1 12 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDDUP_YMM_K1Z_YMMM256 = 956;
	/**
	 * {@code VMOVDDUP zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.F2.0F.W1 12 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDDUP_ZMM_K1Z_ZMMM512 = 957;
	/**
	 * {@code MOVLPS m64, xmm1}<p>
	 * {@code NP 0F 13 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVLPS_M64_XMM = 958;
	/**
	 * {@code VMOVLPS m64, xmm1}<p>
	 * {@code VEX.128.0F.WIG 13 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVLPS_M64_XMM = 959;
	/**
	 * {@code VMOVLPS m64, xmm1}<p>
	 * {@code EVEX.128.0F.W0 13 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVLPS_M64_XMM = 960;
	/**
	 * {@code MOVLPD m64, xmm1}<p>
	 * {@code 66 0F 13 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVLPD_M64_XMM = 961;
	/**
	 * {@code VMOVLPD m64, xmm1}<p>
	 * {@code VEX.128.66.0F.WIG 13 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVLPD_M64_XMM = 962;
	/**
	 * {@code VMOVLPD m64, xmm1}<p>
	 * {@code EVEX.128.66.0F.W1 13 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVLPD_M64_XMM = 963;
	/**
	 * {@code UNPCKLPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 14 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int UNPCKLPS_XMM_XMMM128 = 964;
	/**
	 * {@code VUNPCKLPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.0F.WIG 14 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VUNPCKLPS_XMM_XMM_XMMM128 = 965;
	/**
	 * {@code VUNPCKLPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.0F.WIG 14 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VUNPCKLPS_YMM_YMM_YMMM256 = 966;
	/**
	 * {@code VUNPCKLPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 14 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VUNPCKLPS_XMM_K1Z_XMM_XMMM128B32 = 967;
	/**
	 * {@code VUNPCKLPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 14 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VUNPCKLPS_YMM_K1Z_YMM_YMMM256B32 = 968;
	/**
	 * {@code VUNPCKLPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.0F.W0 14 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VUNPCKLPS_ZMM_K1Z_ZMM_ZMMM512B32 = 969;
	/**
	 * {@code UNPCKLPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 14 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int UNPCKLPD_XMM_XMMM128 = 970;
	/**
	 * {@code VUNPCKLPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 14 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VUNPCKLPD_XMM_XMM_XMMM128 = 971;
	/**
	 * {@code VUNPCKLPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 14 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VUNPCKLPD_YMM_YMM_YMMM256 = 972;
	/**
	 * {@code VUNPCKLPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 14 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VUNPCKLPD_XMM_K1Z_XMM_XMMM128B64 = 973;
	/**
	 * {@code VUNPCKLPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 14 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VUNPCKLPD_YMM_K1Z_YMM_YMMM256B64 = 974;
	/**
	 * {@code VUNPCKLPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F.W1 14 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VUNPCKLPD_ZMM_K1Z_ZMM_ZMMM512B64 = 975;
	/**
	 * {@code UNPCKHPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 15 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int UNPCKHPS_XMM_XMMM128 = 976;
	/**
	 * {@code VUNPCKHPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.0F.WIG 15 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VUNPCKHPS_XMM_XMM_XMMM128 = 977;
	/**
	 * {@code VUNPCKHPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.0F.WIG 15 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VUNPCKHPS_YMM_YMM_YMMM256 = 978;
	/**
	 * {@code VUNPCKHPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 15 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VUNPCKHPS_XMM_K1Z_XMM_XMMM128B32 = 979;
	/**
	 * {@code VUNPCKHPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 15 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VUNPCKHPS_YMM_K1Z_YMM_YMMM256B32 = 980;
	/**
	 * {@code VUNPCKHPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.0F.W0 15 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VUNPCKHPS_ZMM_K1Z_ZMM_ZMMM512B32 = 981;
	/**
	 * {@code UNPCKHPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 15 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int UNPCKHPD_XMM_XMMM128 = 982;
	/**
	 * {@code VUNPCKHPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 15 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VUNPCKHPD_XMM_XMM_XMMM128 = 983;
	/**
	 * {@code VUNPCKHPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 15 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VUNPCKHPD_YMM_YMM_YMMM256 = 984;
	/**
	 * {@code VUNPCKHPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 15 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VUNPCKHPD_XMM_K1Z_XMM_XMMM128B64 = 985;
	/**
	 * {@code VUNPCKHPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 15 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VUNPCKHPD_YMM_K1Z_YMM_YMMM256B64 = 986;
	/**
	 * {@code VUNPCKHPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F.W1 15 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VUNPCKHPD_ZMM_K1Z_ZMM_ZMMM512B64 = 987;
	/**
	 * {@code MOVLHPS xmm1, xmm2}<p>
	 * {@code NP 0F 16 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVLHPS_XMM_XMM = 988;
	/**
	 * {@code VMOVLHPS xmm1, xmm2, xmm3}<p>
	 * {@code VEX.128.0F.WIG 16 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVLHPS_XMM_XMM_XMM = 989;
	/**
	 * {@code VMOVLHPS xmm1, xmm2, xmm3}<p>
	 * {@code EVEX.128.0F.W0 16 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVLHPS_XMM_XMM_XMM = 990;
	/**
	 * {@code MOVHPS xmm1, m64}<p>
	 * {@code NP 0F 16 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVHPS_XMM_M64 = 991;
	/**
	 * {@code VMOVHPS xmm2, xmm1, m64}<p>
	 * {@code VEX.128.0F.WIG 16 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVHPS_XMM_XMM_M64 = 992;
	/**
	 * {@code VMOVHPS xmm2, xmm1, m64}<p>
	 * {@code EVEX.128.0F.W0 16 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVHPS_XMM_XMM_M64 = 993;
	/**
	 * {@code MOVHPD xmm1, m64}<p>
	 * {@code 66 0F 16 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVHPD_XMM_M64 = 994;
	/**
	 * {@code VMOVHPD xmm2, xmm1, m64}<p>
	 * {@code VEX.128.66.0F.WIG 16 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVHPD_XMM_XMM_M64 = 995;
	/**
	 * {@code VMOVHPD xmm2, xmm1, m64}<p>
	 * {@code EVEX.128.66.0F.W1 16 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVHPD_XMM_XMM_M64 = 996;
	/**
	 * {@code MOVSHDUP xmm1, xmm2/m128}<p>
	 * {@code F3 0F 16 /r}<p>
	 * {@code SSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVSHDUP_XMM_XMMM128 = 997;
	/**
	 * {@code VMOVSHDUP xmm1, xmm2/m128}<p>
	 * {@code VEX.128.F3.0F.WIG 16 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVSHDUP_XMM_XMMM128 = 998;
	/**
	 * {@code VMOVSHDUP ymm1, ymm2/m256}<p>
	 * {@code VEX.256.F3.0F.WIG 16 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVSHDUP_YMM_YMMM256 = 999;
	/**
	 * {@code VMOVSHDUP xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.F3.0F.W0 16 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSHDUP_XMM_K1Z_XMMM128 = 1000;
	/**
	 * {@code VMOVSHDUP ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.F3.0F.W0 16 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSHDUP_YMM_K1Z_YMMM256 = 1001;
	/**
	 * {@code VMOVSHDUP zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.F3.0F.W0 16 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSHDUP_ZMM_K1Z_ZMMM512 = 1002;
	/**
	 * {@code MOVHPS m64, xmm1}<p>
	 * {@code NP 0F 17 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVHPS_M64_XMM = 1003;
	/**
	 * {@code VMOVHPS m64, xmm1}<p>
	 * {@code VEX.128.0F.WIG 17 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVHPS_M64_XMM = 1004;
	/**
	 * {@code VMOVHPS m64, xmm1}<p>
	 * {@code EVEX.128.0F.W0 17 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVHPS_M64_XMM = 1005;
	/**
	 * {@code MOVHPD m64, xmm1}<p>
	 * {@code 66 0F 17 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVHPD_M64_XMM = 1006;
	/**
	 * {@code VMOVHPD m64, xmm1}<p>
	 * {@code VEX.128.66.0F.WIG 17 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVHPD_M64_XMM = 1007;
	/**
	 * {@code VMOVHPD m64, xmm1}<p>
	 * {@code EVEX.128.66.0F.W1 17 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVHPD_M64_XMM = 1008;
	/**
	 * {@code RESERVEDNOP r/m16, r16}<p>
	 * {@code o16 0F 18 /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM16_R16_0F18 = 1009;
	/**
	 * {@code RESERVEDNOP r/m32, r32}<p>
	 * {@code o32 0F 18 /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM32_R32_0F18 = 1010;
	/**
	 * {@code RESERVEDNOP r/m64, r64}<p>
	 * {@code o64 0F 18 /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 64-bit}
	 */
	public static final int RESERVEDNOP_RM64_R64_0F18 = 1011;
	/**
	 * {@code RESERVEDNOP r/m16, r16}<p>
	 * {@code o16 0F 19 /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM16_R16_0F19 = 1012;
	/**
	 * {@code RESERVEDNOP r/m32, r32}<p>
	 * {@code o32 0F 19 /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM32_R32_0F19 = 1013;
	/**
	 * {@code RESERVEDNOP r/m64, r64}<p>
	 * {@code o64 0F 19 /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 64-bit}
	 */
	public static final int RESERVEDNOP_RM64_R64_0F19 = 1014;
	/**
	 * {@code RESERVEDNOP r/m16, r16}<p>
	 * {@code o16 0F 1A /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM16_R16_0F1A = 1015;
	/**
	 * {@code RESERVEDNOP r/m32, r32}<p>
	 * {@code o32 0F 1A /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM32_R32_0F1A = 1016;
	/**
	 * {@code RESERVEDNOP r/m64, r64}<p>
	 * {@code o64 0F 1A /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 64-bit}
	 */
	public static final int RESERVEDNOP_RM64_R64_0F1A = 1017;
	/**
	 * {@code RESERVEDNOP r/m16, r16}<p>
	 * {@code o16 0F 1B /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM16_R16_0F1B = 1018;
	/**
	 * {@code RESERVEDNOP r/m32, r32}<p>
	 * {@code o32 0F 1B /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM32_R32_0F1B = 1019;
	/**
	 * {@code RESERVEDNOP r/m64, r64}<p>
	 * {@code o64 0F 1B /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 64-bit}
	 */
	public static final int RESERVEDNOP_RM64_R64_0F1B = 1020;
	/**
	 * {@code RESERVEDNOP r/m16, r16}<p>
	 * {@code o16 0F 1C /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM16_R16_0F1C = 1021;
	/**
	 * {@code RESERVEDNOP r/m32, r32}<p>
	 * {@code o32 0F 1C /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM32_R32_0F1C = 1022;
	/**
	 * {@code RESERVEDNOP r/m64, r64}<p>
	 * {@code o64 0F 1C /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 64-bit}
	 */
	public static final int RESERVEDNOP_RM64_R64_0F1C = 1023;
	/**
	 * {@code RESERVEDNOP r/m16, r16}<p>
	 * {@code o16 0F 1D /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM16_R16_0F1D = 1024;
	/**
	 * {@code RESERVEDNOP r/m32, r32}<p>
	 * {@code o32 0F 1D /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM32_R32_0F1D = 1025;
	/**
	 * {@code RESERVEDNOP r/m64, r64}<p>
	 * {@code o64 0F 1D /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 64-bit}
	 */
	public static final int RESERVEDNOP_RM64_R64_0F1D = 1026;
	/**
	 * {@code RESERVEDNOP r/m16, r16}<p>
	 * {@code o16 0F 1E /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM16_R16_0F1E = 1027;
	/**
	 * {@code RESERVEDNOP r/m32, r32}<p>
	 * {@code o32 0F 1E /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM32_R32_0F1E = 1028;
	/**
	 * {@code RESERVEDNOP r/m64, r64}<p>
	 * {@code o64 0F 1E /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 64-bit}
	 */
	public static final int RESERVEDNOP_RM64_R64_0F1E = 1029;
	/**
	 * {@code RESERVEDNOP r/m16, r16}<p>
	 * {@code o16 0F 1F /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM16_R16_0F1F = 1030;
	/**
	 * {@code RESERVEDNOP r/m32, r32}<p>
	 * {@code o32 0F 1F /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RESERVEDNOP_RM32_R32_0F1F = 1031;
	/**
	 * {@code RESERVEDNOP r/m64, r64}<p>
	 * {@code o64 0F 1F /r}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 64-bit}
	 */
	public static final int RESERVEDNOP_RM64_R64_0F1F = 1032;
	/**
	 * {@code PREFETCHNTA m8}<p>
	 * {@code 0F 18 /0}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PREFETCHNTA_M8 = 1033;
	/**
	 * {@code PREFETCHT0 m8}<p>
	 * {@code 0F 18 /1}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PREFETCHT0_M8 = 1034;
	/**
	 * {@code PREFETCHT1 m8}<p>
	 * {@code 0F 18 /2}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PREFETCHT1_M8 = 1035;
	/**
	 * {@code PREFETCHT2 m8}<p>
	 * {@code 0F 18 /3}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PREFETCHT2_M8 = 1036;
	/**
	 * {@code BNDLDX bnd, mib}<p>
	 * {@code NP 0F 1A /r}<p>
	 * {@code MPX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BNDLDX_BND_MIB = 1037;
	/**
	 * {@code BNDMOV bnd1, bnd2/m64}<p>
	 * {@code 66 0F 1A /r}<p>
	 * {@code MPX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int BNDMOV_BND_BNDM64 = 1038;
	/**
	 * {@code BNDMOV bnd1, bnd2/m128}<p>
	 * {@code 66 0F 1A /r}<p>
	 * {@code MPX}<p>
	 * {@code 64-bit}
	 */
	public static final int BNDMOV_BND_BNDM128 = 1039;
	/**
	 * {@code BNDCL bnd, r/m32}<p>
	 * {@code F3 0F 1A /r}<p>
	 * {@code MPX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int BNDCL_BND_RM32 = 1040;
	/**
	 * {@code BNDCL bnd, r/m64}<p>
	 * {@code F3 0F 1A /r}<p>
	 * {@code MPX}<p>
	 * {@code 64-bit}
	 */
	public static final int BNDCL_BND_RM64 = 1041;
	/**
	 * {@code BNDCU bnd, r/m32}<p>
	 * {@code F2 0F 1A /r}<p>
	 * {@code MPX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int BNDCU_BND_RM32 = 1042;
	/**
	 * {@code BNDCU bnd, r/m64}<p>
	 * {@code F2 0F 1A /r}<p>
	 * {@code MPX}<p>
	 * {@code 64-bit}
	 */
	public static final int BNDCU_BND_RM64 = 1043;
	/**
	 * {@code BNDSTX mib, bnd}<p>
	 * {@code NP 0F 1B /r}<p>
	 * {@code MPX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BNDSTX_MIB_BND = 1044;
	/**
	 * {@code BNDMOV bnd1/m64, bnd2}<p>
	 * {@code 66 0F 1B /r}<p>
	 * {@code MPX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int BNDMOV_BNDM64_BND = 1045;
	/**
	 * {@code BNDMOV bnd1/m128, bnd2}<p>
	 * {@code 66 0F 1B /r}<p>
	 * {@code MPX}<p>
	 * {@code 64-bit}
	 */
	public static final int BNDMOV_BNDM128_BND = 1046;
	/**
	 * {@code BNDMK bnd, m32}<p>
	 * {@code F3 0F 1B /r}<p>
	 * {@code MPX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int BNDMK_BND_M32 = 1047;
	/**
	 * {@code BNDMK bnd, m64}<p>
	 * {@code F3 0F 1B /r}<p>
	 * {@code MPX}<p>
	 * {@code 64-bit}
	 */
	public static final int BNDMK_BND_M64 = 1048;
	/**
	 * {@code BNDCN bnd, r/m32}<p>
	 * {@code F2 0F 1B /r}<p>
	 * {@code MPX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int BNDCN_BND_RM32 = 1049;
	/**
	 * {@code BNDCN bnd, r/m64}<p>
	 * {@code F2 0F 1B /r}<p>
	 * {@code MPX}<p>
	 * {@code 64-bit}
	 */
	public static final int BNDCN_BND_RM64 = 1050;
	/**
	 * {@code CLDEMOTE m8}<p>
	 * {@code NP 0F 1C /0}<p>
	 * {@code CLDEMOTE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CLDEMOTE_M8 = 1051;
	/**
	 * {@code RDSSPD r32}<p>
	 * {@code F3 0F 1E /1}<p>
	 * {@code CET_SS}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RDSSPD_R32 = 1052;
	/**
	 * {@code RDSSPQ r64}<p>
	 * {@code F3 o64 0F 1E /1}<p>
	 * {@code CET_SS}<p>
	 * {@code 64-bit}
	 */
	public static final int RDSSPQ_R64 = 1053;
	/**
	 * {@code ENDBR64}<p>
	 * {@code F3 0F 1E FA}<p>
	 * {@code CET_IBT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ENDBR64 = 1054;
	/**
	 * {@code ENDBR32}<p>
	 * {@code F3 0F 1E FB}<p>
	 * {@code CET_IBT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ENDBR32 = 1055;
	/**
	 * {@code NOP r/m16}<p>
	 * {@code o16 0F 1F /0}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int NOP_RM16 = 1056;
	/**
	 * {@code NOP r/m32}<p>
	 * {@code o32 0F 1F /0}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int NOP_RM32 = 1057;
	/**
	 * {@code NOP r/m64}<p>
	 * {@code o64 0F 1F /0}<p>
	 * {@code CPUID.01H.EAX[Bits 11:8] = 0110B or 1111B}<p>
	 * {@code 64-bit}
	 */
	public static final int NOP_RM64 = 1058;
	/**
	 * {@code MOV r32, cr}<p>
	 * {@code 0F 20 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int MOV_R32_CR = 1059;
	/**
	 * {@code MOV r64, cr}<p>
	 * {@code 0F 20 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOV_R64_CR = 1060;
	/**
	 * {@code MOV r32, dr}<p>
	 * {@code 0F 21 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int MOV_R32_DR = 1061;
	/**
	 * {@code MOV r64, dr}<p>
	 * {@code 0F 21 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOV_R64_DR = 1062;
	/**
	 * {@code MOV cr, r32}<p>
	 * {@code 0F 22 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int MOV_CR_R32 = 1063;
	/**
	 * {@code MOV cr, r64}<p>
	 * {@code 0F 22 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOV_CR_R64 = 1064;
	/**
	 * {@code MOV dr, r32}<p>
	 * {@code 0F 23 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int MOV_DR_R32 = 1065;
	/**
	 * {@code MOV dr, r64}<p>
	 * {@code 0F 23 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOV_DR_R64 = 1066;
	/**
	 * {@code MOV r32, tr}<p>
	 * {@code 0F 24 /r}<p>
	 * {@code 386/486/Cyrix/Geode}<p>
	 * {@code 16/32-bit}
	 */
	public static final int MOV_R32_TR = 1067;
	/**
	 * {@code MOV tr, r32}<p>
	 * {@code 0F 26 /r}<p>
	 * {@code 386/486/Cyrix/Geode}<p>
	 * {@code 16/32-bit}
	 */
	public static final int MOV_TR_R32 = 1068;
	/**
	 * {@code MOVAPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 28 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVAPS_XMM_XMMM128 = 1069;
	/**
	 * {@code VMOVAPS xmm1, xmm2/m128}<p>
	 * {@code VEX.128.0F.WIG 28 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVAPS_XMM_XMMM128 = 1070;
	/**
	 * {@code VMOVAPS ymm1, ymm2/m256}<p>
	 * {@code VEX.256.0F.WIG 28 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVAPS_YMM_YMMM256 = 1071;
	/**
	 * {@code VMOVAPS xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.0F.W0 28 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVAPS_XMM_K1Z_XMMM128 = 1072;
	/**
	 * {@code VMOVAPS ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.0F.W0 28 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVAPS_YMM_K1Z_YMMM256 = 1073;
	/**
	 * {@code VMOVAPS zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.0F.W0 28 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVAPS_ZMM_K1Z_ZMMM512 = 1074;
	/**
	 * {@code MOVAPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 28 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVAPD_XMM_XMMM128 = 1075;
	/**
	 * {@code VMOVAPD xmm1, xmm2/m128}<p>
	 * {@code VEX.128.66.0F.WIG 28 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVAPD_XMM_XMMM128 = 1076;
	/**
	 * {@code VMOVAPD ymm1, ymm2/m256}<p>
	 * {@code VEX.256.66.0F.WIG 28 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVAPD_YMM_YMMM256 = 1077;
	/**
	 * {@code VMOVAPD xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.66.0F.W1 28 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVAPD_XMM_K1Z_XMMM128 = 1078;
	/**
	 * {@code VMOVAPD ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.66.0F.W1 28 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVAPD_YMM_K1Z_YMMM256 = 1079;
	/**
	 * {@code VMOVAPD zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.66.0F.W1 28 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVAPD_ZMM_K1Z_ZMMM512 = 1080;
	/**
	 * {@code MOVAPS xmm2/m128, xmm1}<p>
	 * {@code NP 0F 29 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVAPS_XMMM128_XMM = 1081;
	/**
	 * {@code VMOVAPS xmm2/m128, xmm1}<p>
	 * {@code VEX.128.0F.WIG 29 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVAPS_XMMM128_XMM = 1082;
	/**
	 * {@code VMOVAPS ymm2/m256, ymm1}<p>
	 * {@code VEX.256.0F.WIG 29 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVAPS_YMMM256_YMM = 1083;
	/**
	 * {@code VMOVAPS xmm2/m128 {k1}{z}, xmm1}<p>
	 * {@code EVEX.128.0F.W0 29 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVAPS_XMMM128_K1Z_XMM = 1084;
	/**
	 * {@code VMOVAPS ymm2/m256 {k1}{z}, ymm1}<p>
	 * {@code EVEX.256.0F.W0 29 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVAPS_YMMM256_K1Z_YMM = 1085;
	/**
	 * {@code VMOVAPS zmm2/m512 {k1}{z}, zmm1}<p>
	 * {@code EVEX.512.0F.W0 29 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVAPS_ZMMM512_K1Z_ZMM = 1086;
	/**
	 * {@code MOVAPD xmm2/m128, xmm1}<p>
	 * {@code 66 0F 29 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVAPD_XMMM128_XMM = 1087;
	/**
	 * {@code VMOVAPD xmm2/m128, xmm1}<p>
	 * {@code VEX.128.66.0F.WIG 29 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVAPD_XMMM128_XMM = 1088;
	/**
	 * {@code VMOVAPD ymm2/m256, ymm1}<p>
	 * {@code VEX.256.66.0F.WIG 29 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVAPD_YMMM256_YMM = 1089;
	/**
	 * {@code VMOVAPD xmm2/m128 {k1}{z}, xmm1}<p>
	 * {@code EVEX.128.66.0F.W1 29 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVAPD_XMMM128_K1Z_XMM = 1090;
	/**
	 * {@code VMOVAPD ymm2/m256 {k1}{z}, ymm1}<p>
	 * {@code EVEX.256.66.0F.W1 29 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVAPD_YMMM256_K1Z_YMM = 1091;
	/**
	 * {@code VMOVAPD zmm2/m512 {k1}{z}, zmm1}<p>
	 * {@code EVEX.512.66.0F.W1 29 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVAPD_ZMMM512_K1Z_ZMM = 1092;
	/**
	 * {@code CVTPI2PS xmm, mm/m64}<p>
	 * {@code NP 0F 2A /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTPI2PS_XMM_MMM64 = 1093;
	/**
	 * {@code CVTPI2PD xmm, mm/m64}<p>
	 * {@code 66 0F 2A /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTPI2PD_XMM_MMM64 = 1094;
	/**
	 * {@code CVTSI2SS xmm1, r/m32}<p>
	 * {@code F3 0F 2A /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTSI2SS_XMM_RM32 = 1095;
	/**
	 * {@code CVTSI2SS xmm1, r/m64}<p>
	 * {@code F3 o64 0F 2A /r}<p>
	 * {@code SSE}<p>
	 * {@code 64-bit}
	 */
	public static final int CVTSI2SS_XMM_RM64 = 1096;
	/**
	 * {@code VCVTSI2SS xmm1, xmm2, r/m32}<p>
	 * {@code VEX.LIG.F3.0F.W0 2A /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTSI2SS_XMM_XMM_RM32 = 1097;
	/**
	 * {@code VCVTSI2SS xmm1, xmm2, r/m64}<p>
	 * {@code VEX.LIG.F3.0F.W1 2A /r}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VCVTSI2SS_XMM_XMM_RM64 = 1098;
	/**
	 * {@code VCVTSI2SS xmm1, xmm2, r/m32{er}}<p>
	 * {@code EVEX.LIG.F3.0F.W0 2A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTSI2SS_XMM_XMM_RM32_ER = 1099;
	/**
	 * {@code VCVTSI2SS xmm1, xmm2, r/m64{er}}<p>
	 * {@code EVEX.LIG.F3.0F.W1 2A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTSI2SS_XMM_XMM_RM64_ER = 1100;
	/**
	 * {@code CVTSI2SD xmm1, r/m32}<p>
	 * {@code F2 0F 2A /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTSI2SD_XMM_RM32 = 1101;
	/**
	 * {@code CVTSI2SD xmm1, r/m64}<p>
	 * {@code F2 o64 0F 2A /r}<p>
	 * {@code SSE2}<p>
	 * {@code 64-bit}
	 */
	public static final int CVTSI2SD_XMM_RM64 = 1102;
	/**
	 * {@code VCVTSI2SD xmm1, xmm2, r/m32}<p>
	 * {@code VEX.LIG.F2.0F.W0 2A /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTSI2SD_XMM_XMM_RM32 = 1103;
	/**
	 * {@code VCVTSI2SD xmm1, xmm2, r/m64}<p>
	 * {@code VEX.LIG.F2.0F.W1 2A /r}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VCVTSI2SD_XMM_XMM_RM64 = 1104;
	/**
	 * {@code VCVTSI2SD xmm1, xmm2, r/m32{er}}<p>
	 * {@code EVEX.LIG.F2.0F.W0 2A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTSI2SD_XMM_XMM_RM32_ER = 1105;
	/**
	 * {@code VCVTSI2SD xmm1, xmm2, r/m64{er}}<p>
	 * {@code EVEX.LIG.F2.0F.W1 2A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTSI2SD_XMM_XMM_RM64_ER = 1106;
	/**
	 * {@code MOVNTPS m128, xmm1}<p>
	 * {@code NP 0F 2B /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVNTPS_M128_XMM = 1107;
	/**
	 * {@code VMOVNTPS m128, xmm1}<p>
	 * {@code VEX.128.0F.WIG 2B /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVNTPS_M128_XMM = 1108;
	/**
	 * {@code VMOVNTPS m256, ymm1}<p>
	 * {@code VEX.256.0F.WIG 2B /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVNTPS_M256_YMM = 1109;
	/**
	 * {@code VMOVNTPS m128, xmm1}<p>
	 * {@code EVEX.128.0F.W0 2B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVNTPS_M128_XMM = 1110;
	/**
	 * {@code VMOVNTPS m256, ymm1}<p>
	 * {@code EVEX.256.0F.W0 2B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVNTPS_M256_YMM = 1111;
	/**
	 * {@code VMOVNTPS m512, zmm1}<p>
	 * {@code EVEX.512.0F.W0 2B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVNTPS_M512_ZMM = 1112;
	/**
	 * {@code MOVNTPD m128, xmm1}<p>
	 * {@code 66 0F 2B /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVNTPD_M128_XMM = 1113;
	/**
	 * {@code VMOVNTPD m128, xmm1}<p>
	 * {@code VEX.128.66.0F.WIG 2B /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVNTPD_M128_XMM = 1114;
	/**
	 * {@code VMOVNTPD m256, ymm1}<p>
	 * {@code VEX.256.66.0F.WIG 2B /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVNTPD_M256_YMM = 1115;
	/**
	 * {@code VMOVNTPD m128, xmm1}<p>
	 * {@code EVEX.128.66.0F.W1 2B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVNTPD_M128_XMM = 1116;
	/**
	 * {@code VMOVNTPD m256, ymm1}<p>
	 * {@code EVEX.256.66.0F.W1 2B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVNTPD_M256_YMM = 1117;
	/**
	 * {@code VMOVNTPD m512, zmm1}<p>
	 * {@code EVEX.512.66.0F.W1 2B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVNTPD_M512_ZMM = 1118;
	/**
	 * {@code MOVNTSS m32, xmm1}<p>
	 * {@code F3 0F 2B /r}<p>
	 * {@code SSE4A}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVNTSS_M32_XMM = 1119;
	/**
	 * {@code MOVNTSD m64, xmm1}<p>
	 * {@code F2 0F 2B /r}<p>
	 * {@code SSE4A}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVNTSD_M64_XMM = 1120;
	/**
	 * {@code CVTTPS2PI mm, xmm/m64}<p>
	 * {@code NP 0F 2C /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTTPS2PI_MM_XMMM64 = 1121;
	/**
	 * {@code CVTTPD2PI mm, xmm/m128}<p>
	 * {@code 66 0F 2C /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTTPD2PI_MM_XMMM128 = 1122;
	/**
	 * {@code CVTTSS2SI r32, xmm1/m32}<p>
	 * {@code F3 0F 2C /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTTSS2SI_R32_XMMM32 = 1123;
	/**
	 * {@code CVTTSS2SI r64, xmm1/m32}<p>
	 * {@code F3 o64 0F 2C /r}<p>
	 * {@code SSE}<p>
	 * {@code 64-bit}
	 */
	public static final int CVTTSS2SI_R64_XMMM32 = 1124;
	/**
	 * {@code VCVTTSS2SI r32, xmm1/m32}<p>
	 * {@code VEX.LIG.F3.0F.W0 2C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTTSS2SI_R32_XMMM32 = 1125;
	/**
	 * {@code VCVTTSS2SI r64, xmm1/m32}<p>
	 * {@code VEX.LIG.F3.0F.W1 2C /r}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VCVTTSS2SI_R64_XMMM32 = 1126;
	/**
	 * {@code VCVTTSS2SI r32, xmm1/m32{sae}}<p>
	 * {@code EVEX.LIG.F3.0F.W0 2C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTSS2SI_R32_XMMM32_SAE = 1127;
	/**
	 * {@code VCVTTSS2SI r64, xmm1/m32{sae}}<p>
	 * {@code EVEX.LIG.F3.0F.W1 2C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTTSS2SI_R64_XMMM32_SAE = 1128;
	/**
	 * {@code CVTTSD2SI r32, xmm1/m64}<p>
	 * {@code F2 0F 2C /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTTSD2SI_R32_XMMM64 = 1129;
	/**
	 * {@code CVTTSD2SI r64, xmm1/m64}<p>
	 * {@code F2 o64 0F 2C /r}<p>
	 * {@code SSE2}<p>
	 * {@code 64-bit}
	 */
	public static final int CVTTSD2SI_R64_XMMM64 = 1130;
	/**
	 * {@code VCVTTSD2SI r32, xmm1/m64}<p>
	 * {@code VEX.LIG.F2.0F.W0 2C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTTSD2SI_R32_XMMM64 = 1131;
	/**
	 * {@code VCVTTSD2SI r64, xmm1/m64}<p>
	 * {@code VEX.LIG.F2.0F.W1 2C /r}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VCVTTSD2SI_R64_XMMM64 = 1132;
	/**
	 * {@code VCVTTSD2SI r32, xmm1/m64{sae}}<p>
	 * {@code EVEX.LIG.F2.0F.W0 2C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTSD2SI_R32_XMMM64_SAE = 1133;
	/**
	 * {@code VCVTTSD2SI r64, xmm1/m64{sae}}<p>
	 * {@code EVEX.LIG.F2.0F.W1 2C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTTSD2SI_R64_XMMM64_SAE = 1134;
	/**
	 * {@code CVTPS2PI mm, xmm/m64}<p>
	 * {@code NP 0F 2D /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTPS2PI_MM_XMMM64 = 1135;
	/**
	 * {@code CVTPD2PI mm, xmm/m128}<p>
	 * {@code 66 0F 2D /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTPD2PI_MM_XMMM128 = 1136;
	/**
	 * {@code CVTSS2SI r32, xmm1/m32}<p>
	 * {@code F3 0F 2D /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTSS2SI_R32_XMMM32 = 1137;
	/**
	 * {@code CVTSS2SI r64, xmm1/m32}<p>
	 * {@code F3 o64 0F 2D /r}<p>
	 * {@code SSE}<p>
	 * {@code 64-bit}
	 */
	public static final int CVTSS2SI_R64_XMMM32 = 1138;
	/**
	 * {@code VCVTSS2SI r32, xmm1/m32}<p>
	 * {@code VEX.LIG.F3.0F.W0 2D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTSS2SI_R32_XMMM32 = 1139;
	/**
	 * {@code VCVTSS2SI r64, xmm1/m32}<p>
	 * {@code VEX.LIG.F3.0F.W1 2D /r}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VCVTSS2SI_R64_XMMM32 = 1140;
	/**
	 * {@code VCVTSS2SI r32, xmm1/m32{er}}<p>
	 * {@code EVEX.LIG.F3.0F.W0 2D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTSS2SI_R32_XMMM32_ER = 1141;
	/**
	 * {@code VCVTSS2SI r64, xmm1/m32{er}}<p>
	 * {@code EVEX.LIG.F3.0F.W1 2D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTSS2SI_R64_XMMM32_ER = 1142;
	/**
	 * {@code CVTSD2SI r32, xmm1/m64}<p>
	 * {@code F2 0F 2D /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTSD2SI_R32_XMMM64 = 1143;
	/**
	 * {@code CVTSD2SI r64, xmm1/m64}<p>
	 * {@code F2 o64 0F 2D /r}<p>
	 * {@code SSE2}<p>
	 * {@code 64-bit}
	 */
	public static final int CVTSD2SI_R64_XMMM64 = 1144;
	/**
	 * {@code VCVTSD2SI r32, xmm1/m64}<p>
	 * {@code VEX.LIG.F2.0F.W0 2D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTSD2SI_R32_XMMM64 = 1145;
	/**
	 * {@code VCVTSD2SI r64, xmm1/m64}<p>
	 * {@code VEX.LIG.F2.0F.W1 2D /r}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VCVTSD2SI_R64_XMMM64 = 1146;
	/**
	 * {@code VCVTSD2SI r32, xmm1/m64{er}}<p>
	 * {@code EVEX.LIG.F2.0F.W0 2D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTSD2SI_R32_XMMM64_ER = 1147;
	/**
	 * {@code VCVTSD2SI r64, xmm1/m64{er}}<p>
	 * {@code EVEX.LIG.F2.0F.W1 2D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTSD2SI_R64_XMMM64_ER = 1148;
	/**
	 * {@code UCOMISS xmm1, xmm2/m32}<p>
	 * {@code NP 0F 2E /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int UCOMISS_XMM_XMMM32 = 1149;
	/**
	 * {@code VUCOMISS xmm1, xmm2/m32}<p>
	 * {@code VEX.LIG.0F.WIG 2E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VUCOMISS_XMM_XMMM32 = 1150;
	/**
	 * {@code VUCOMISS xmm1, xmm2/m32{sae}}<p>
	 * {@code EVEX.LIG.0F.W0 2E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VUCOMISS_XMM_XMMM32_SAE = 1151;
	/**
	 * {@code UCOMISD xmm1, xmm2/m64}<p>
	 * {@code 66 0F 2E /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int UCOMISD_XMM_XMMM64 = 1152;
	/**
	 * {@code VUCOMISD xmm1, xmm2/m64}<p>
	 * {@code VEX.LIG.66.0F.WIG 2E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VUCOMISD_XMM_XMMM64 = 1153;
	/**
	 * {@code VUCOMISD xmm1, xmm2/m64{sae}}<p>
	 * {@code EVEX.LIG.66.0F.W1 2E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VUCOMISD_XMM_XMMM64_SAE = 1154;
	/**
	 * {@code COMISS xmm1, xmm2/m32}<p>
	 * {@code NP 0F 2F /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int COMISS_XMM_XMMM32 = 1155;
	/**
	 * {@code COMISD xmm1, xmm2/m64}<p>
	 * {@code 66 0F 2F /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int COMISD_XMM_XMMM64 = 1156;
	/**
	 * {@code VCOMISS xmm1, xmm2/m32}<p>
	 * {@code VEX.LIG.0F.WIG 2F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCOMISS_XMM_XMMM32 = 1157;
	/**
	 * {@code VCOMISD xmm1, xmm2/m64}<p>
	 * {@code VEX.LIG.66.0F.WIG 2F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCOMISD_XMM_XMMM64 = 1158;
	/**
	 * {@code VCOMISS xmm1, xmm2/m32{sae}}<p>
	 * {@code EVEX.LIG.0F.W0 2F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCOMISS_XMM_XMMM32_SAE = 1159;
	/**
	 * {@code VCOMISD xmm1, xmm2/m64{sae}}<p>
	 * {@code EVEX.LIG.66.0F.W1 2F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCOMISD_XMM_XMMM64_SAE = 1160;
	/**
	 * {@code WRMSR}<p>
	 * {@code 0F 30}<p>
	 * {@code MSR}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int WRMSR = 1161;
	/**
	 * {@code RDTSC}<p>
	 * {@code 0F 31}<p>
	 * {@code TSC}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RDTSC = 1162;
	/**
	 * {@code RDMSR}<p>
	 * {@code 0F 32}<p>
	 * {@code MSR}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RDMSR = 1163;
	/**
	 * {@code RDPMC}<p>
	 * {@code 0F 33}<p>
	 * {@code Pentium MMX or later, or Pentium Pro or later}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RDPMC = 1164;
	/**
	 * {@code SYSENTER}<p>
	 * {@code 0F 34}<p>
	 * {@code SEP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SYSENTER = 1165;
	/**
	 * {@code SYSEXIT}<p>
	 * {@code 0F 35}<p>
	 * {@code SEP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SYSEXITD = 1166;
	/**
	 * {@code SYSEXITQ}<p>
	 * {@code o64 0F 35}<p>
	 * {@code SEP}<p>
	 * {@code 64-bit}
	 */
	public static final int SYSEXITQ = 1167;
	/**
	 * {@code GETSEC}<p>
	 * {@code NP 0F 37}<p>
	 * {@code SMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int GETSECD = 1168;
	/**
	 * {@code CMOVO r16, r/m16}<p>
	 * {@code o16 0F 40 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVO_R16_RM16 = 1169;
	/**
	 * {@code CMOVO r32, r/m32}<p>
	 * {@code o32 0F 40 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVO_R32_RM32 = 1170;
	/**
	 * {@code CMOVO r64, r/m64}<p>
	 * {@code o64 0F 40 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVO_R64_RM64 = 1171;
	/**
	 * {@code CMOVNO r16, r/m16}<p>
	 * {@code o16 0F 41 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVNO_R16_RM16 = 1172;
	/**
	 * {@code CMOVNO r32, r/m32}<p>
	 * {@code o32 0F 41 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVNO_R32_RM32 = 1173;
	/**
	 * {@code CMOVNO r64, r/m64}<p>
	 * {@code o64 0F 41 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVNO_R64_RM64 = 1174;
	/**
	 * {@code CMOVB r16, r/m16}<p>
	 * {@code o16 0F 42 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVB_R16_RM16 = 1175;
	/**
	 * {@code CMOVB r32, r/m32}<p>
	 * {@code o32 0F 42 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVB_R32_RM32 = 1176;
	/**
	 * {@code CMOVB r64, r/m64}<p>
	 * {@code o64 0F 42 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVB_R64_RM64 = 1177;
	/**
	 * {@code CMOVAE r16, r/m16}<p>
	 * {@code o16 0F 43 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVAE_R16_RM16 = 1178;
	/**
	 * {@code CMOVAE r32, r/m32}<p>
	 * {@code o32 0F 43 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVAE_R32_RM32 = 1179;
	/**
	 * {@code CMOVAE r64, r/m64}<p>
	 * {@code o64 0F 43 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVAE_R64_RM64 = 1180;
	/**
	 * {@code CMOVE r16, r/m16}<p>
	 * {@code o16 0F 44 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVE_R16_RM16 = 1181;
	/**
	 * {@code CMOVE r32, r/m32}<p>
	 * {@code o32 0F 44 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVE_R32_RM32 = 1182;
	/**
	 * {@code CMOVE r64, r/m64}<p>
	 * {@code o64 0F 44 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVE_R64_RM64 = 1183;
	/**
	 * {@code CMOVNE r16, r/m16}<p>
	 * {@code o16 0F 45 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVNE_R16_RM16 = 1184;
	/**
	 * {@code CMOVNE r32, r/m32}<p>
	 * {@code o32 0F 45 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVNE_R32_RM32 = 1185;
	/**
	 * {@code CMOVNE r64, r/m64}<p>
	 * {@code o64 0F 45 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVNE_R64_RM64 = 1186;
	/**
	 * {@code CMOVBE r16, r/m16}<p>
	 * {@code o16 0F 46 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVBE_R16_RM16 = 1187;
	/**
	 * {@code CMOVBE r32, r/m32}<p>
	 * {@code o32 0F 46 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVBE_R32_RM32 = 1188;
	/**
	 * {@code CMOVBE r64, r/m64}<p>
	 * {@code o64 0F 46 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVBE_R64_RM64 = 1189;
	/**
	 * {@code CMOVA r16, r/m16}<p>
	 * {@code o16 0F 47 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVA_R16_RM16 = 1190;
	/**
	 * {@code CMOVA r32, r/m32}<p>
	 * {@code o32 0F 47 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVA_R32_RM32 = 1191;
	/**
	 * {@code CMOVA r64, r/m64}<p>
	 * {@code o64 0F 47 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVA_R64_RM64 = 1192;
	/**
	 * {@code CMOVS r16, r/m16}<p>
	 * {@code o16 0F 48 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVS_R16_RM16 = 1193;
	/**
	 * {@code CMOVS r32, r/m32}<p>
	 * {@code o32 0F 48 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVS_R32_RM32 = 1194;
	/**
	 * {@code CMOVS r64, r/m64}<p>
	 * {@code o64 0F 48 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVS_R64_RM64 = 1195;
	/**
	 * {@code CMOVNS r16, r/m16}<p>
	 * {@code o16 0F 49 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVNS_R16_RM16 = 1196;
	/**
	 * {@code CMOVNS r32, r/m32}<p>
	 * {@code o32 0F 49 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVNS_R32_RM32 = 1197;
	/**
	 * {@code CMOVNS r64, r/m64}<p>
	 * {@code o64 0F 49 /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVNS_R64_RM64 = 1198;
	/**
	 * {@code CMOVP r16, r/m16}<p>
	 * {@code o16 0F 4A /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVP_R16_RM16 = 1199;
	/**
	 * {@code CMOVP r32, r/m32}<p>
	 * {@code o32 0F 4A /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVP_R32_RM32 = 1200;
	/**
	 * {@code CMOVP r64, r/m64}<p>
	 * {@code o64 0F 4A /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVP_R64_RM64 = 1201;
	/**
	 * {@code CMOVNP r16, r/m16}<p>
	 * {@code o16 0F 4B /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVNP_R16_RM16 = 1202;
	/**
	 * {@code CMOVNP r32, r/m32}<p>
	 * {@code o32 0F 4B /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVNP_R32_RM32 = 1203;
	/**
	 * {@code CMOVNP r64, r/m64}<p>
	 * {@code o64 0F 4B /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVNP_R64_RM64 = 1204;
	/**
	 * {@code CMOVL r16, r/m16}<p>
	 * {@code o16 0F 4C /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVL_R16_RM16 = 1205;
	/**
	 * {@code CMOVL r32, r/m32}<p>
	 * {@code o32 0F 4C /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVL_R32_RM32 = 1206;
	/**
	 * {@code CMOVL r64, r/m64}<p>
	 * {@code o64 0F 4C /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVL_R64_RM64 = 1207;
	/**
	 * {@code CMOVGE r16, r/m16}<p>
	 * {@code o16 0F 4D /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVGE_R16_RM16 = 1208;
	/**
	 * {@code CMOVGE r32, r/m32}<p>
	 * {@code o32 0F 4D /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVGE_R32_RM32 = 1209;
	/**
	 * {@code CMOVGE r64, r/m64}<p>
	 * {@code o64 0F 4D /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVGE_R64_RM64 = 1210;
	/**
	 * {@code CMOVLE r16, r/m16}<p>
	 * {@code o16 0F 4E /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVLE_R16_RM16 = 1211;
	/**
	 * {@code CMOVLE r32, r/m32}<p>
	 * {@code o32 0F 4E /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVLE_R32_RM32 = 1212;
	/**
	 * {@code CMOVLE r64, r/m64}<p>
	 * {@code o64 0F 4E /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVLE_R64_RM64 = 1213;
	/**
	 * {@code CMOVG r16, r/m16}<p>
	 * {@code o16 0F 4F /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVG_R16_RM16 = 1214;
	/**
	 * {@code CMOVG r32, r/m32}<p>
	 * {@code o32 0F 4F /r}<p>
	 * {@code CMOV}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMOVG_R32_RM32 = 1215;
	/**
	 * {@code CMOVG r64, r/m64}<p>
	 * {@code o64 0F 4F /r}<p>
	 * {@code CMOV}<p>
	 * {@code 64-bit}
	 */
	public static final int CMOVG_R64_RM64 = 1216;
	/**
	 * {@code KANDW k1, k2, k3}<p>
	 * {@code VEX.L1.0F.W0 41 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KANDW_KR_KR_KR = 1217;
	/**
	 * {@code KANDQ k1, k2, k3}<p>
	 * {@code VEX.L1.0F.W1 41 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KANDQ_KR_KR_KR = 1218;
	/**
	 * {@code KANDB k1, k2, k3}<p>
	 * {@code VEX.L1.66.0F.W0 41 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KANDB_KR_KR_KR = 1219;
	/**
	 * {@code KANDD k1, k2, k3}<p>
	 * {@code VEX.L1.66.0F.W1 41 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KANDD_KR_KR_KR = 1220;
	/**
	 * {@code KANDNW k1, k2, k3}<p>
	 * {@code VEX.L1.0F.W0 42 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KANDNW_KR_KR_KR = 1221;
	/**
	 * {@code KANDNQ k1, k2, k3}<p>
	 * {@code VEX.L1.0F.W1 42 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KANDNQ_KR_KR_KR = 1222;
	/**
	 * {@code KANDNB k1, k2, k3}<p>
	 * {@code VEX.L1.66.0F.W0 42 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KANDNB_KR_KR_KR = 1223;
	/**
	 * {@code KANDND k1, k2, k3}<p>
	 * {@code VEX.L1.66.0F.W1 42 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KANDND_KR_KR_KR = 1224;
	/**
	 * {@code KNOTW k1, k2}<p>
	 * {@code VEX.L0.0F.W0 44 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KNOTW_KR_KR = 1225;
	/**
	 * {@code KNOTQ k1, k2}<p>
	 * {@code VEX.L0.0F.W1 44 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KNOTQ_KR_KR = 1226;
	/**
	 * {@code KNOTB k1, k2}<p>
	 * {@code VEX.L0.66.0F.W0 44 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KNOTB_KR_KR = 1227;
	/**
	 * {@code KNOTD k1, k2}<p>
	 * {@code VEX.L0.66.0F.W1 44 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KNOTD_KR_KR = 1228;
	/**
	 * {@code KORW k1, k2, k3}<p>
	 * {@code VEX.L1.0F.W0 45 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KORW_KR_KR_KR = 1229;
	/**
	 * {@code KORQ k1, k2, k3}<p>
	 * {@code VEX.L1.0F.W1 45 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KORQ_KR_KR_KR = 1230;
	/**
	 * {@code KORB k1, k2, k3}<p>
	 * {@code VEX.L1.66.0F.W0 45 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KORB_KR_KR_KR = 1231;
	/**
	 * {@code KORD k1, k2, k3}<p>
	 * {@code VEX.L1.66.0F.W1 45 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KORD_KR_KR_KR = 1232;
	/**
	 * {@code KXNORW k1, k2, k3}<p>
	 * {@code VEX.L1.0F.W0 46 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KXNORW_KR_KR_KR = 1233;
	/**
	 * {@code KXNORQ k1, k2, k3}<p>
	 * {@code VEX.L1.0F.W1 46 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KXNORQ_KR_KR_KR = 1234;
	/**
	 * {@code KXNORB k1, k2, k3}<p>
	 * {@code VEX.L1.66.0F.W0 46 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KXNORB_KR_KR_KR = 1235;
	/**
	 * {@code KXNORD k1, k2, k3}<p>
	 * {@code VEX.L1.66.0F.W1 46 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KXNORD_KR_KR_KR = 1236;
	/**
	 * {@code KXORW k1, k2, k3}<p>
	 * {@code VEX.L1.0F.W0 47 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KXORW_KR_KR_KR = 1237;
	/**
	 * {@code KXORQ k1, k2, k3}<p>
	 * {@code VEX.L1.0F.W1 47 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KXORQ_KR_KR_KR = 1238;
	/**
	 * {@code KXORB k1, k2, k3}<p>
	 * {@code VEX.L1.66.0F.W0 47 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KXORB_KR_KR_KR = 1239;
	/**
	 * {@code KXORD k1, k2, k3}<p>
	 * {@code VEX.L1.66.0F.W1 47 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KXORD_KR_KR_KR = 1240;
	/**
	 * {@code KADDW k1, k2, k3}<p>
	 * {@code VEX.L1.0F.W0 4A /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KADDW_KR_KR_KR = 1241;
	/**
	 * {@code KADDQ k1, k2, k3}<p>
	 * {@code VEX.L1.0F.W1 4A /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KADDQ_KR_KR_KR = 1242;
	/**
	 * {@code KADDB k1, k2, k3}<p>
	 * {@code VEX.L1.66.0F.W0 4A /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KADDB_KR_KR_KR = 1243;
	/**
	 * {@code KADDD k1, k2, k3}<p>
	 * {@code VEX.L1.66.0F.W1 4A /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KADDD_KR_KR_KR = 1244;
	/**
	 * {@code KUNPCKWD k1, k2, k3}<p>
	 * {@code VEX.L1.0F.W0 4B /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KUNPCKWD_KR_KR_KR = 1245;
	/**
	 * {@code KUNPCKDQ k1, k2, k3}<p>
	 * {@code VEX.L1.0F.W1 4B /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KUNPCKDQ_KR_KR_KR = 1246;
	/**
	 * {@code KUNPCKBW k1, k2, k3}<p>
	 * {@code VEX.L1.66.0F.W0 4B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KUNPCKBW_KR_KR_KR = 1247;
	/**
	 * {@code MOVMSKPS r32, xmm}<p>
	 * {@code NP 0F 50 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVMSKPS_R32_XMM = 1248;
	/**
	 * {@code MOVMSKPS r64, xmm}<p>
	 * {@code NP o64 0F 50 /r}<p>
	 * {@code SSE}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVMSKPS_R64_XMM = 1249;
	/**
	 * {@code VMOVMSKPS r32, xmm2}<p>
	 * {@code VEX.128.0F.W0 50 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVMSKPS_R32_XMM = 1250;
	/**
	 * {@code VMOVMSKPS r64, xmm2}<p>
	 * {@code VEX.128.0F.W1 50 /r}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VMOVMSKPS_R64_XMM = 1251;
	/**
	 * {@code VMOVMSKPS r32, ymm2}<p>
	 * {@code VEX.256.0F.W0 50 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVMSKPS_R32_YMM = 1252;
	/**
	 * {@code VMOVMSKPS r64, ymm2}<p>
	 * {@code VEX.256.0F.W1 50 /r}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VMOVMSKPS_R64_YMM = 1253;
	/**
	 * {@code MOVMSKPD r32, xmm}<p>
	 * {@code 66 0F 50 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVMSKPD_R32_XMM = 1254;
	/**
	 * {@code MOVMSKPD r64, xmm}<p>
	 * {@code 66 o64 0F 50 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVMSKPD_R64_XMM = 1255;
	/**
	 * {@code VMOVMSKPD r32, xmm2}<p>
	 * {@code VEX.128.66.0F.W0 50 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVMSKPD_R32_XMM = 1256;
	/**
	 * {@code VMOVMSKPD r64, xmm2}<p>
	 * {@code VEX.128.66.0F.W1 50 /r}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VMOVMSKPD_R64_XMM = 1257;
	/**
	 * {@code VMOVMSKPD r32, ymm2}<p>
	 * {@code VEX.256.66.0F.W0 50 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVMSKPD_R32_YMM = 1258;
	/**
	 * {@code VMOVMSKPD r64, ymm2}<p>
	 * {@code VEX.256.66.0F.W1 50 /r}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VMOVMSKPD_R64_YMM = 1259;
	/**
	 * {@code SQRTPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 51 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SQRTPS_XMM_XMMM128 = 1260;
	/**
	 * {@code VSQRTPS xmm1, xmm2/m128}<p>
	 * {@code VEX.128.0F.WIG 51 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSQRTPS_XMM_XMMM128 = 1261;
	/**
	 * {@code VSQRTPS ymm1, ymm2/m256}<p>
	 * {@code VEX.256.0F.WIG 51 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSQRTPS_YMM_YMMM256 = 1262;
	/**
	 * {@code VSQRTPS xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 51 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSQRTPS_XMM_K1Z_XMMM128B32 = 1263;
	/**
	 * {@code VSQRTPS ymm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 51 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSQRTPS_YMM_K1Z_YMMM256B32 = 1264;
	/**
	 * {@code VSQRTPS zmm1 {k1}{z}, zmm2/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.0F.W0 51 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSQRTPS_ZMM_K1Z_ZMMM512B32_ER = 1265;
	/**
	 * {@code SQRTPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 51 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SQRTPD_XMM_XMMM128 = 1266;
	/**
	 * {@code VSQRTPD xmm1, xmm2/m128}<p>
	 * {@code VEX.128.66.0F.WIG 51 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSQRTPD_XMM_XMMM128 = 1267;
	/**
	 * {@code VSQRTPD ymm1, ymm2/m256}<p>
	 * {@code VEX.256.66.0F.WIG 51 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSQRTPD_YMM_YMMM256 = 1268;
	/**
	 * {@code VSQRTPD xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 51 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSQRTPD_XMM_K1Z_XMMM128B64 = 1269;
	/**
	 * {@code VSQRTPD ymm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 51 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSQRTPD_YMM_K1Z_YMMM256B64 = 1270;
	/**
	 * {@code VSQRTPD zmm1 {k1}{z}, zmm2/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F.W1 51 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSQRTPD_ZMM_K1Z_ZMMM512B64_ER = 1271;
	/**
	 * {@code SQRTSS xmm1, xmm2/m32}<p>
	 * {@code F3 0F 51 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SQRTSS_XMM_XMMM32 = 1272;
	/**
	 * {@code VSQRTSS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.F3.0F.WIG 51 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSQRTSS_XMM_XMM_XMMM32 = 1273;
	/**
	 * {@code VSQRTSS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.F3.0F.W0 51 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSQRTSS_XMM_K1Z_XMM_XMMM32_ER = 1274;
	/**
	 * {@code SQRTSD xmm1, xmm2/m64}<p>
	 * {@code F2 0F 51 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SQRTSD_XMM_XMMM64 = 1275;
	/**
	 * {@code VSQRTSD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.F2.0F.WIG 51 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSQRTSD_XMM_XMM_XMMM64 = 1276;
	/**
	 * {@code VSQRTSD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.F2.0F.W1 51 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSQRTSD_XMM_K1Z_XMM_XMMM64_ER = 1277;
	/**
	 * {@code RSQRTPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 52 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RSQRTPS_XMM_XMMM128 = 1278;
	/**
	 * {@code VRSQRTPS xmm1, xmm2/m128}<p>
	 * {@code VEX.128.0F.WIG 52 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VRSQRTPS_XMM_XMMM128 = 1279;
	/**
	 * {@code VRSQRTPS ymm1, ymm2/m256}<p>
	 * {@code VEX.256.0F.WIG 52 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VRSQRTPS_YMM_YMMM256 = 1280;
	/**
	 * {@code RSQRTSS xmm1, xmm2/m32}<p>
	 * {@code F3 0F 52 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RSQRTSS_XMM_XMMM32 = 1281;
	/**
	 * {@code VRSQRTSS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.F3.0F.WIG 52 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VRSQRTSS_XMM_XMM_XMMM32 = 1282;
	/**
	 * {@code RCPPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 53 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCPPS_XMM_XMMM128 = 1283;
	/**
	 * {@code VRCPPS xmm1, xmm2/m128}<p>
	 * {@code VEX.128.0F.WIG 53 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VRCPPS_XMM_XMMM128 = 1284;
	/**
	 * {@code VRCPPS ymm1, ymm2/m256}<p>
	 * {@code VEX.256.0F.WIG 53 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VRCPPS_YMM_YMMM256 = 1285;
	/**
	 * {@code RCPSS xmm1, xmm2/m32}<p>
	 * {@code F3 0F 53 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RCPSS_XMM_XMMM32 = 1286;
	/**
	 * {@code VRCPSS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.F3.0F.WIG 53 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VRCPSS_XMM_XMM_XMMM32 = 1287;
	/**
	 * {@code ANDPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 54 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ANDPS_XMM_XMMM128 = 1288;
	/**
	 * {@code VANDPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.0F.WIG 54 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VANDPS_XMM_XMM_XMMM128 = 1289;
	/**
	 * {@code VANDPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.0F.WIG 54 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VANDPS_YMM_YMM_YMMM256 = 1290;
	/**
	 * {@code VANDPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 54 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VANDPS_XMM_K1Z_XMM_XMMM128B32 = 1291;
	/**
	 * {@code VANDPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 54 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VANDPS_YMM_K1Z_YMM_YMMM256B32 = 1292;
	/**
	 * {@code VANDPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.0F.W0 54 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VANDPS_ZMM_K1Z_ZMM_ZMMM512B32 = 1293;
	/**
	 * {@code ANDPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 54 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ANDPD_XMM_XMMM128 = 1294;
	/**
	 * {@code VANDPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 54 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VANDPD_XMM_XMM_XMMM128 = 1295;
	/**
	 * {@code VANDPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 54 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VANDPD_YMM_YMM_YMMM256 = 1296;
	/**
	 * {@code VANDPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 54 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VANDPD_XMM_K1Z_XMM_XMMM128B64 = 1297;
	/**
	 * {@code VANDPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 54 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VANDPD_YMM_K1Z_YMM_YMMM256B64 = 1298;
	/**
	 * {@code VANDPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F.W1 54 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VANDPD_ZMM_K1Z_ZMM_ZMMM512B64 = 1299;
	/**
	 * {@code ANDNPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 55 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ANDNPS_XMM_XMMM128 = 1300;
	/**
	 * {@code VANDNPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.0F.WIG 55 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VANDNPS_XMM_XMM_XMMM128 = 1301;
	/**
	 * {@code VANDNPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.0F.WIG 55 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VANDNPS_YMM_YMM_YMMM256 = 1302;
	/**
	 * {@code VANDNPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 55 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VANDNPS_XMM_K1Z_XMM_XMMM128B32 = 1303;
	/**
	 * {@code VANDNPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 55 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VANDNPS_YMM_K1Z_YMM_YMMM256B32 = 1304;
	/**
	 * {@code VANDNPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.0F.W0 55 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VANDNPS_ZMM_K1Z_ZMM_ZMMM512B32 = 1305;
	/**
	 * {@code ANDNPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 55 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ANDNPD_XMM_XMMM128 = 1306;
	/**
	 * {@code VANDNPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 55 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VANDNPD_XMM_XMM_XMMM128 = 1307;
	/**
	 * {@code VANDNPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 55 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VANDNPD_YMM_YMM_YMMM256 = 1308;
	/**
	 * {@code VANDNPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 55 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VANDNPD_XMM_K1Z_XMM_XMMM128B64 = 1309;
	/**
	 * {@code VANDNPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 55 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VANDNPD_YMM_K1Z_YMM_YMMM256B64 = 1310;
	/**
	 * {@code VANDNPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F.W1 55 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VANDNPD_ZMM_K1Z_ZMM_ZMMM512B64 = 1311;
	/**
	 * {@code ORPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 56 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ORPS_XMM_XMMM128 = 1312;
	/**
	 * {@code VORPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.0F.WIG 56 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VORPS_XMM_XMM_XMMM128 = 1313;
	/**
	 * {@code VORPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.0F.WIG 56 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VORPS_YMM_YMM_YMMM256 = 1314;
	/**
	 * {@code VORPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 56 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VORPS_XMM_K1Z_XMM_XMMM128B32 = 1315;
	/**
	 * {@code VORPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 56 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VORPS_YMM_K1Z_YMM_YMMM256B32 = 1316;
	/**
	 * {@code VORPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.0F.W0 56 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VORPS_ZMM_K1Z_ZMM_ZMMM512B32 = 1317;
	/**
	 * {@code ORPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 56 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ORPD_XMM_XMMM128 = 1318;
	/**
	 * {@code VORPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 56 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VORPD_XMM_XMM_XMMM128 = 1319;
	/**
	 * {@code VORPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 56 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VORPD_YMM_YMM_YMMM256 = 1320;
	/**
	 * {@code VORPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 56 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VORPD_XMM_K1Z_XMM_XMMM128B64 = 1321;
	/**
	 * {@code VORPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 56 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VORPD_YMM_K1Z_YMM_YMMM256B64 = 1322;
	/**
	 * {@code VORPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F.W1 56 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VORPD_ZMM_K1Z_ZMM_ZMMM512B64 = 1323;
	/**
	 * {@code XORPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 57 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XORPS_XMM_XMMM128 = 1324;
	/**
	 * {@code VXORPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.0F.WIG 57 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VXORPS_XMM_XMM_XMMM128 = 1325;
	/**
	 * {@code VXORPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.0F.WIG 57 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VXORPS_YMM_YMM_YMMM256 = 1326;
	/**
	 * {@code VXORPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 57 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VXORPS_XMM_K1Z_XMM_XMMM128B32 = 1327;
	/**
	 * {@code VXORPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 57 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VXORPS_YMM_K1Z_YMM_YMMM256B32 = 1328;
	/**
	 * {@code VXORPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.0F.W0 57 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VXORPS_ZMM_K1Z_ZMM_ZMMM512B32 = 1329;
	/**
	 * {@code XORPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 57 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XORPD_XMM_XMMM128 = 1330;
	/**
	 * {@code VXORPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 57 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VXORPD_XMM_XMM_XMMM128 = 1331;
	/**
	 * {@code VXORPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 57 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VXORPD_YMM_YMM_YMMM256 = 1332;
	/**
	 * {@code VXORPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 57 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VXORPD_XMM_K1Z_XMM_XMMM128B64 = 1333;
	/**
	 * {@code VXORPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 57 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VXORPD_YMM_K1Z_YMM_YMMM256B64 = 1334;
	/**
	 * {@code VXORPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F.W1 57 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VXORPD_ZMM_K1Z_ZMM_ZMMM512B64 = 1335;
	/**
	 * {@code ADDPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 58 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADDPS_XMM_XMMM128 = 1336;
	/**
	 * {@code VADDPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.0F.WIG 58 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VADDPS_XMM_XMM_XMMM128 = 1337;
	/**
	 * {@code VADDPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.0F.WIG 58 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VADDPS_YMM_YMM_YMMM256 = 1338;
	/**
	 * {@code VADDPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 58 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VADDPS_XMM_K1Z_XMM_XMMM128B32 = 1339;
	/**
	 * {@code VADDPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 58 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VADDPS_YMM_K1Z_YMM_YMMM256B32 = 1340;
	/**
	 * {@code VADDPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.0F.W0 58 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VADDPS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 1341;
	/**
	 * {@code ADDPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 58 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADDPD_XMM_XMMM128 = 1342;
	/**
	 * {@code VADDPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 58 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VADDPD_XMM_XMM_XMMM128 = 1343;
	/**
	 * {@code VADDPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 58 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VADDPD_YMM_YMM_YMMM256 = 1344;
	/**
	 * {@code VADDPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 58 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VADDPD_XMM_K1Z_XMM_XMMM128B64 = 1345;
	/**
	 * {@code VADDPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 58 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VADDPD_YMM_K1Z_YMM_YMMM256B64 = 1346;
	/**
	 * {@code VADDPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F.W1 58 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VADDPD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 1347;
	/**
	 * {@code ADDSS xmm1, xmm2/m32}<p>
	 * {@code F3 0F 58 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADDSS_XMM_XMMM32 = 1348;
	/**
	 * {@code VADDSS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.F3.0F.WIG 58 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VADDSS_XMM_XMM_XMMM32 = 1349;
	/**
	 * {@code VADDSS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.F3.0F.W0 58 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VADDSS_XMM_K1Z_XMM_XMMM32_ER = 1350;
	/**
	 * {@code ADDSD xmm1, xmm2/m64}<p>
	 * {@code F2 0F 58 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADDSD_XMM_XMMM64 = 1351;
	/**
	 * {@code VADDSD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.F2.0F.WIG 58 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VADDSD_XMM_XMM_XMMM64 = 1352;
	/**
	 * {@code VADDSD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.F2.0F.W1 58 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VADDSD_XMM_K1Z_XMM_XMMM64_ER = 1353;
	/**
	 * {@code MULPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 59 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MULPS_XMM_XMMM128 = 1354;
	/**
	 * {@code VMULPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.0F.WIG 59 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMULPS_XMM_XMM_XMMM128 = 1355;
	/**
	 * {@code VMULPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.0F.WIG 59 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMULPS_YMM_YMM_YMMM256 = 1356;
	/**
	 * {@code VMULPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 59 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMULPS_XMM_K1Z_XMM_XMMM128B32 = 1357;
	/**
	 * {@code VMULPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 59 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMULPS_YMM_K1Z_YMM_YMMM256B32 = 1358;
	/**
	 * {@code VMULPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.0F.W0 59 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMULPS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 1359;
	/**
	 * {@code MULPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 59 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MULPD_XMM_XMMM128 = 1360;
	/**
	 * {@code VMULPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 59 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMULPD_XMM_XMM_XMMM128 = 1361;
	/**
	 * {@code VMULPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 59 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMULPD_YMM_YMM_YMMM256 = 1362;
	/**
	 * {@code VMULPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 59 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMULPD_XMM_K1Z_XMM_XMMM128B64 = 1363;
	/**
	 * {@code VMULPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 59 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMULPD_YMM_K1Z_YMM_YMMM256B64 = 1364;
	/**
	 * {@code VMULPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F.W1 59 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMULPD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 1365;
	/**
	 * {@code MULSS xmm1, xmm2/m32}<p>
	 * {@code F3 0F 59 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MULSS_XMM_XMMM32 = 1366;
	/**
	 * {@code VMULSS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.F3.0F.WIG 59 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMULSS_XMM_XMM_XMMM32 = 1367;
	/**
	 * {@code VMULSS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.F3.0F.W0 59 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMULSS_XMM_K1Z_XMM_XMMM32_ER = 1368;
	/**
	 * {@code MULSD xmm1, xmm2/m64}<p>
	 * {@code F2 0F 59 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MULSD_XMM_XMMM64 = 1369;
	/**
	 * {@code VMULSD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.F2.0F.WIG 59 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMULSD_XMM_XMM_XMMM64 = 1370;
	/**
	 * {@code VMULSD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.F2.0F.W1 59 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMULSD_XMM_K1Z_XMM_XMMM64_ER = 1371;
	/**
	 * {@code CVTPS2PD xmm1, xmm2/m64}<p>
	 * {@code NP 0F 5A /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTPS2PD_XMM_XMMM64 = 1372;
	/**
	 * {@code VCVTPS2PD xmm1, xmm2/m64}<p>
	 * {@code VEX.128.0F.WIG 5A /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTPS2PD_XMM_XMMM64 = 1373;
	/**
	 * {@code VCVTPS2PD ymm1, xmm2/m128}<p>
	 * {@code VEX.256.0F.WIG 5A /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTPS2PD_YMM_XMMM128 = 1374;
	/**
	 * {@code VCVTPS2PD xmm1 {k1}{z}, xmm2/m64/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 5A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2PD_XMM_K1Z_XMMM64B32 = 1375;
	/**
	 * {@code VCVTPS2PD ymm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 5A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2PD_YMM_K1Z_XMMM128B32 = 1376;
	/**
	 * {@code VCVTPS2PD zmm1 {k1}{z}, ymm2/m256/m32bcst{sae}}<p>
	 * {@code EVEX.512.0F.W0 5A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2PD_ZMM_K1Z_YMMM256B32_SAE = 1377;
	/**
	 * {@code CVTPD2PS xmm1, xmm2/m128}<p>
	 * {@code 66 0F 5A /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTPD2PS_XMM_XMMM128 = 1378;
	/**
	 * {@code VCVTPD2PS xmm1, xmm2/m128}<p>
	 * {@code VEX.128.66.0F.WIG 5A /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTPD2PS_XMM_XMMM128 = 1379;
	/**
	 * {@code VCVTPD2PS xmm1, ymm2/m256}<p>
	 * {@code VEX.256.66.0F.WIG 5A /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTPD2PS_XMM_YMMM256 = 1380;
	/**
	 * {@code VCVTPD2PS xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 5A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2PS_XMM_K1Z_XMMM128B64 = 1381;
	/**
	 * {@code VCVTPD2PS xmm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 5A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2PS_XMM_K1Z_YMMM256B64 = 1382;
	/**
	 * {@code VCVTPD2PS ymm1 {k1}{z}, zmm2/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F.W1 5A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2PS_YMM_K1Z_ZMMM512B64_ER = 1383;
	/**
	 * {@code CVTSS2SD xmm1, xmm2/m32}<p>
	 * {@code F3 0F 5A /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTSS2SD_XMM_XMMM32 = 1384;
	/**
	 * {@code VCVTSS2SD xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.F3.0F.WIG 5A /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTSS2SD_XMM_XMM_XMMM32 = 1385;
	/**
	 * {@code VCVTSS2SD xmm1 {k1}{z}, xmm2, xmm3/m32{sae}}<p>
	 * {@code EVEX.LIG.F3.0F.W0 5A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTSS2SD_XMM_K1Z_XMM_XMMM32_SAE = 1386;
	/**
	 * {@code CVTSD2SS xmm1, xmm2/m64}<p>
	 * {@code F2 0F 5A /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTSD2SS_XMM_XMMM64 = 1387;
	/**
	 * {@code VCVTSD2SS xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.F2.0F.WIG 5A /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTSD2SS_XMM_XMM_XMMM64 = 1388;
	/**
	 * {@code VCVTSD2SS xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.F2.0F.W1 5A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTSD2SS_XMM_K1Z_XMM_XMMM64_ER = 1389;
	/**
	 * {@code CVTDQ2PS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 5B /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTDQ2PS_XMM_XMMM128 = 1390;
	/**
	 * {@code VCVTDQ2PS xmm1, xmm2/m128}<p>
	 * {@code VEX.128.0F.WIG 5B /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTDQ2PS_XMM_XMMM128 = 1391;
	/**
	 * {@code VCVTDQ2PS ymm1, ymm2/m256}<p>
	 * {@code VEX.256.0F.WIG 5B /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTDQ2PS_YMM_YMMM256 = 1392;
	/**
	 * {@code VCVTDQ2PS xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 5B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTDQ2PS_XMM_K1Z_XMMM128B32 = 1393;
	/**
	 * {@code VCVTDQ2PS ymm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 5B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTDQ2PS_YMM_K1Z_YMMM256B32 = 1394;
	/**
	 * {@code VCVTDQ2PS zmm1 {k1}{z}, zmm2/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.0F.W0 5B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTDQ2PS_ZMM_K1Z_ZMMM512B32_ER = 1395;
	/**
	 * {@code VCVTQQ2PS xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.0F.W1 5B /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTQQ2PS_XMM_K1Z_XMMM128B64 = 1396;
	/**
	 * {@code VCVTQQ2PS xmm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.0F.W1 5B /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTQQ2PS_XMM_K1Z_YMMM256B64 = 1397;
	/**
	 * {@code VCVTQQ2PS ymm1 {k1}{z}, zmm2/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.0F.W1 5B /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTQQ2PS_YMM_K1Z_ZMMM512B64_ER = 1398;
	/**
	 * {@code CVTPS2DQ xmm1, xmm2/m128}<p>
	 * {@code 66 0F 5B /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTPS2DQ_XMM_XMMM128 = 1399;
	/**
	 * {@code VCVTPS2DQ xmm1, xmm2/m128}<p>
	 * {@code VEX.128.66.0F.WIG 5B /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTPS2DQ_XMM_XMMM128 = 1400;
	/**
	 * {@code VCVTPS2DQ ymm1, ymm2/m256}<p>
	 * {@code VEX.256.66.0F.WIG 5B /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTPS2DQ_YMM_YMMM256 = 1401;
	/**
	 * {@code VCVTPS2DQ xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 5B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2DQ_XMM_K1Z_XMMM128B32 = 1402;
	/**
	 * {@code VCVTPS2DQ ymm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 5B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2DQ_YMM_K1Z_YMMM256B32 = 1403;
	/**
	 * {@code VCVTPS2DQ zmm1 {k1}{z}, zmm2/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F.W0 5B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2DQ_ZMM_K1Z_ZMMM512B32_ER = 1404;
	/**
	 * {@code CVTTPS2DQ xmm1, xmm2/m128}<p>
	 * {@code F3 0F 5B /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTTPS2DQ_XMM_XMMM128 = 1405;
	/**
	 * {@code VCVTTPS2DQ xmm1, xmm2/m128}<p>
	 * {@code VEX.128.F3.0F.WIG 5B /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTTPS2DQ_XMM_XMMM128 = 1406;
	/**
	 * {@code VCVTTPS2DQ ymm1, ymm2/m256}<p>
	 * {@code VEX.256.F3.0F.WIG 5B /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTTPS2DQ_YMM_YMMM256 = 1407;
	/**
	 * {@code VCVTTPS2DQ xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.F3.0F.W0 5B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPS2DQ_XMM_K1Z_XMMM128B32 = 1408;
	/**
	 * {@code VCVTTPS2DQ ymm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.F3.0F.W0 5B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPS2DQ_YMM_K1Z_YMMM256B32 = 1409;
	/**
	 * {@code VCVTTPS2DQ zmm1 {k1}{z}, zmm2/m512/m32bcst{sae}}<p>
	 * {@code EVEX.512.F3.0F.W0 5B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPS2DQ_ZMM_K1Z_ZMMM512B32_SAE = 1410;
	/**
	 * {@code SUBPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 5C /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUBPS_XMM_XMMM128 = 1411;
	/**
	 * {@code VSUBPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.0F.WIG 5C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSUBPS_XMM_XMM_XMMM128 = 1412;
	/**
	 * {@code VSUBPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.0F.WIG 5C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSUBPS_YMM_YMM_YMMM256 = 1413;
	/**
	 * {@code VSUBPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 5C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSUBPS_XMM_K1Z_XMM_XMMM128B32 = 1414;
	/**
	 * {@code VSUBPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 5C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSUBPS_YMM_K1Z_YMM_YMMM256B32 = 1415;
	/**
	 * {@code VSUBPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.0F.W0 5C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSUBPS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 1416;
	/**
	 * {@code SUBPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 5C /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUBPD_XMM_XMMM128 = 1417;
	/**
	 * {@code VSUBPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 5C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSUBPD_XMM_XMM_XMMM128 = 1418;
	/**
	 * {@code VSUBPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 5C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSUBPD_YMM_YMM_YMMM256 = 1419;
	/**
	 * {@code VSUBPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 5C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSUBPD_XMM_K1Z_XMM_XMMM128B64 = 1420;
	/**
	 * {@code VSUBPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 5C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSUBPD_YMM_K1Z_YMM_YMMM256B64 = 1421;
	/**
	 * {@code VSUBPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F.W1 5C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSUBPD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 1422;
	/**
	 * {@code SUBSS xmm1, xmm2/m32}<p>
	 * {@code F3 0F 5C /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUBSS_XMM_XMMM32 = 1423;
	/**
	 * {@code VSUBSS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.F3.0F.WIG 5C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSUBSS_XMM_XMM_XMMM32 = 1424;
	/**
	 * {@code VSUBSS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.F3.0F.W0 5C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSUBSS_XMM_K1Z_XMM_XMMM32_ER = 1425;
	/**
	 * {@code SUBSD xmm1, xmm2/m64}<p>
	 * {@code F2 0F 5C /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SUBSD_XMM_XMMM64 = 1426;
	/**
	 * {@code VSUBSD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.F2.0F.WIG 5C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSUBSD_XMM_XMM_XMMM64 = 1427;
	/**
	 * {@code VSUBSD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.F2.0F.W1 5C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSUBSD_XMM_K1Z_XMM_XMMM64_ER = 1428;
	/**
	 * {@code MINPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 5D /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MINPS_XMM_XMMM128 = 1429;
	/**
	 * {@code VMINPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.0F.WIG 5D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMINPS_XMM_XMM_XMMM128 = 1430;
	/**
	 * {@code VMINPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.0F.WIG 5D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMINPS_YMM_YMM_YMMM256 = 1431;
	/**
	 * {@code VMINPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 5D /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMINPS_XMM_K1Z_XMM_XMMM128B32 = 1432;
	/**
	 * {@code VMINPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 5D /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMINPS_YMM_K1Z_YMM_YMMM256B32 = 1433;
	/**
	 * {@code VMINPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{sae}}<p>
	 * {@code EVEX.512.0F.W0 5D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMINPS_ZMM_K1Z_ZMM_ZMMM512B32_SAE = 1434;
	/**
	 * {@code MINPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 5D /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MINPD_XMM_XMMM128 = 1435;
	/**
	 * {@code VMINPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 5D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMINPD_XMM_XMM_XMMM128 = 1436;
	/**
	 * {@code VMINPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 5D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMINPD_YMM_YMM_YMMM256 = 1437;
	/**
	 * {@code VMINPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 5D /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMINPD_XMM_K1Z_XMM_XMMM128B64 = 1438;
	/**
	 * {@code VMINPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 5D /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMINPD_YMM_K1Z_YMM_YMMM256B64 = 1439;
	/**
	 * {@code VMINPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{sae}}<p>
	 * {@code EVEX.512.66.0F.W1 5D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMINPD_ZMM_K1Z_ZMM_ZMMM512B64_SAE = 1440;
	/**
	 * {@code MINSS xmm1, xmm2/m32}<p>
	 * {@code F3 0F 5D /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MINSS_XMM_XMMM32 = 1441;
	/**
	 * {@code VMINSS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.F3.0F.WIG 5D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMINSS_XMM_XMM_XMMM32 = 1442;
	/**
	 * {@code VMINSS xmm1 {k1}{z}, xmm2, xmm3/m32{sae}}<p>
	 * {@code EVEX.LIG.F3.0F.W0 5D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMINSS_XMM_K1Z_XMM_XMMM32_SAE = 1443;
	/**
	 * {@code MINSD xmm1, xmm2/m64}<p>
	 * {@code F2 0F 5D /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MINSD_XMM_XMMM64 = 1444;
	/**
	 * {@code VMINSD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.F2.0F.WIG 5D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMINSD_XMM_XMM_XMMM64 = 1445;
	/**
	 * {@code VMINSD xmm1 {k1}{z}, xmm2, xmm3/m64{sae}}<p>
	 * {@code EVEX.LIG.F2.0F.W1 5D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMINSD_XMM_K1Z_XMM_XMMM64_SAE = 1446;
	/**
	 * {@code DIVPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 5E /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int DIVPS_XMM_XMMM128 = 1447;
	/**
	 * {@code VDIVPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.0F.WIG 5E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VDIVPS_XMM_XMM_XMMM128 = 1448;
	/**
	 * {@code VDIVPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.0F.WIG 5E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VDIVPS_YMM_YMM_YMMM256 = 1449;
	/**
	 * {@code VDIVPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 5E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDIVPS_XMM_K1Z_XMM_XMMM128B32 = 1450;
	/**
	 * {@code VDIVPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 5E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDIVPS_YMM_K1Z_YMM_YMMM256B32 = 1451;
	/**
	 * {@code VDIVPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.0F.W0 5E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDIVPS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 1452;
	/**
	 * {@code DIVPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 5E /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int DIVPD_XMM_XMMM128 = 1453;
	/**
	 * {@code VDIVPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 5E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VDIVPD_XMM_XMM_XMMM128 = 1454;
	/**
	 * {@code VDIVPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 5E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VDIVPD_YMM_YMM_YMMM256 = 1455;
	/**
	 * {@code VDIVPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 5E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDIVPD_XMM_K1Z_XMM_XMMM128B64 = 1456;
	/**
	 * {@code VDIVPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 5E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDIVPD_YMM_K1Z_YMM_YMMM256B64 = 1457;
	/**
	 * {@code VDIVPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F.W1 5E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDIVPD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 1458;
	/**
	 * {@code DIVSS xmm1, xmm2/m32}<p>
	 * {@code F3 0F 5E /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int DIVSS_XMM_XMMM32 = 1459;
	/**
	 * {@code VDIVSS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.F3.0F.WIG 5E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VDIVSS_XMM_XMM_XMMM32 = 1460;
	/**
	 * {@code VDIVSS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.F3.0F.W0 5E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDIVSS_XMM_K1Z_XMM_XMMM32_ER = 1461;
	/**
	 * {@code DIVSD xmm1, xmm2/m64}<p>
	 * {@code F2 0F 5E /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int DIVSD_XMM_XMMM64 = 1462;
	/**
	 * {@code VDIVSD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.F2.0F.WIG 5E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VDIVSD_XMM_XMM_XMMM64 = 1463;
	/**
	 * {@code VDIVSD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.F2.0F.W1 5E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDIVSD_XMM_K1Z_XMM_XMMM64_ER = 1464;
	/**
	 * {@code MAXPS xmm1, xmm2/m128}<p>
	 * {@code NP 0F 5F /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MAXPS_XMM_XMMM128 = 1465;
	/**
	 * {@code VMAXPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.0F.WIG 5F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMAXPS_XMM_XMM_XMMM128 = 1466;
	/**
	 * {@code VMAXPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.0F.WIG 5F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMAXPS_YMM_YMM_YMMM256 = 1467;
	/**
	 * {@code VMAXPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 5F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMAXPS_XMM_K1Z_XMM_XMMM128B32 = 1468;
	/**
	 * {@code VMAXPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 5F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMAXPS_YMM_K1Z_YMM_YMMM256B32 = 1469;
	/**
	 * {@code VMAXPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{sae}}<p>
	 * {@code EVEX.512.0F.W0 5F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMAXPS_ZMM_K1Z_ZMM_ZMMM512B32_SAE = 1470;
	/**
	 * {@code MAXPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 5F /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MAXPD_XMM_XMMM128 = 1471;
	/**
	 * {@code VMAXPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 5F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMAXPD_XMM_XMM_XMMM128 = 1472;
	/**
	 * {@code VMAXPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 5F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMAXPD_YMM_YMM_YMMM256 = 1473;
	/**
	 * {@code VMAXPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 5F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMAXPD_XMM_K1Z_XMM_XMMM128B64 = 1474;
	/**
	 * {@code VMAXPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 5F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMAXPD_YMM_K1Z_YMM_YMMM256B64 = 1475;
	/**
	 * {@code VMAXPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{sae}}<p>
	 * {@code EVEX.512.66.0F.W1 5F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMAXPD_ZMM_K1Z_ZMM_ZMMM512B64_SAE = 1476;
	/**
	 * {@code MAXSS xmm1, xmm2/m32}<p>
	 * {@code F3 0F 5F /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MAXSS_XMM_XMMM32 = 1477;
	/**
	 * {@code VMAXSS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.F3.0F.WIG 5F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMAXSS_XMM_XMM_XMMM32 = 1478;
	/**
	 * {@code VMAXSS xmm1 {k1}{z}, xmm2, xmm3/m32{sae}}<p>
	 * {@code EVEX.LIG.F3.0F.W0 5F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMAXSS_XMM_K1Z_XMM_XMMM32_SAE = 1479;
	/**
	 * {@code MAXSD xmm1, xmm2/m64}<p>
	 * {@code F2 0F 5F /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MAXSD_XMM_XMMM64 = 1480;
	/**
	 * {@code VMAXSD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.F2.0F.WIG 5F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMAXSD_XMM_XMM_XMMM64 = 1481;
	/**
	 * {@code VMAXSD xmm1 {k1}{z}, xmm2, xmm3/m64{sae}}<p>
	 * {@code EVEX.LIG.F2.0F.W1 5F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMAXSD_XMM_K1Z_XMM_XMMM64_SAE = 1482;
	/**
	 * {@code PUNPCKLBW mm, mm/m32}<p>
	 * {@code NP 0F 60 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUNPCKLBW_MM_MMM32 = 1483;
	/**
	 * {@code PUNPCKLBW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 60 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUNPCKLBW_XMM_XMMM128 = 1484;
	/**
	 * {@code VPUNPCKLBW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 60 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKLBW_XMM_XMM_XMMM128 = 1485;
	/**
	 * {@code VPUNPCKLBW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 60 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKLBW_YMM_YMM_YMMM256 = 1486;
	/**
	 * {@code VPUNPCKLBW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG 60 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKLBW_XMM_K1Z_XMM_XMMM128 = 1487;
	/**
	 * {@code VPUNPCKLBW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG 60 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKLBW_YMM_K1Z_YMM_YMMM256 = 1488;
	/**
	 * {@code VPUNPCKLBW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG 60 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKLBW_ZMM_K1Z_ZMM_ZMMM512 = 1489;
	/**
	 * {@code PUNPCKLWD mm, mm/m32}<p>
	 * {@code NP 0F 61 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUNPCKLWD_MM_MMM32 = 1490;
	/**
	 * {@code PUNPCKLWD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 61 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUNPCKLWD_XMM_XMMM128 = 1491;
	/**
	 * {@code VPUNPCKLWD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 61 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKLWD_XMM_XMM_XMMM128 = 1492;
	/**
	 * {@code VPUNPCKLWD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 61 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKLWD_YMM_YMM_YMMM256 = 1493;
	/**
	 * {@code VPUNPCKLWD xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG 61 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKLWD_XMM_K1Z_XMM_XMMM128 = 1494;
	/**
	 * {@code VPUNPCKLWD ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG 61 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKLWD_YMM_K1Z_YMM_YMMM256 = 1495;
	/**
	 * {@code VPUNPCKLWD zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG 61 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKLWD_ZMM_K1Z_ZMM_ZMMM512 = 1496;
	/**
	 * {@code PUNPCKLDQ mm, mm/m32}<p>
	 * {@code NP 0F 62 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUNPCKLDQ_MM_MMM32 = 1497;
	/**
	 * {@code PUNPCKLDQ xmm1, xmm2/m128}<p>
	 * {@code 66 0F 62 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUNPCKLDQ_XMM_XMMM128 = 1498;
	/**
	 * {@code VPUNPCKLDQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 62 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKLDQ_XMM_XMM_XMMM128 = 1499;
	/**
	 * {@code VPUNPCKLDQ ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 62 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKLDQ_YMM_YMM_YMMM256 = 1500;
	/**
	 * {@code VPUNPCKLDQ xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 62 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKLDQ_XMM_K1Z_XMM_XMMM128B32 = 1501;
	/**
	 * {@code VPUNPCKLDQ ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 62 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKLDQ_YMM_K1Z_YMM_YMMM256B32 = 1502;
	/**
	 * {@code VPUNPCKLDQ zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F.W0 62 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKLDQ_ZMM_K1Z_ZMM_ZMMM512B32 = 1503;
	/**
	 * {@code PACKSSWB mm1, mm2/m64}<p>
	 * {@code NP 0F 63 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PACKSSWB_MM_MMM64 = 1504;
	/**
	 * {@code PACKSSWB xmm1, xmm2/m128}<p>
	 * {@code 66 0F 63 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PACKSSWB_XMM_XMMM128 = 1505;
	/**
	 * {@code VPACKSSWB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 63 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPACKSSWB_XMM_XMM_XMMM128 = 1506;
	/**
	 * {@code VPACKSSWB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 63 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPACKSSWB_YMM_YMM_YMMM256 = 1507;
	/**
	 * {@code VPACKSSWB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG 63 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPACKSSWB_XMM_K1Z_XMM_XMMM128 = 1508;
	/**
	 * {@code VPACKSSWB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG 63 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPACKSSWB_YMM_K1Z_YMM_YMMM256 = 1509;
	/**
	 * {@code VPACKSSWB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG 63 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPACKSSWB_ZMM_K1Z_ZMM_ZMMM512 = 1510;
	/**
	 * {@code PCMPGTB mm, mm/m64}<p>
	 * {@code NP 0F 64 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPGTB_MM_MMM64 = 1511;
	/**
	 * {@code PCMPGTB xmm1, xmm2/m128}<p>
	 * {@code 66 0F 64 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPGTB_XMM_XMMM128 = 1512;
	/**
	 * {@code VPCMPGTB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 64 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPGTB_XMM_XMM_XMMM128 = 1513;
	/**
	 * {@code VPCMPGTB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 64 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPGTB_YMM_YMM_YMMM256 = 1514;
	/**
	 * {@code VPCMPGTB k1 {k2}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG 64 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPGTB_KR_K1_XMM_XMMM128 = 1515;
	/**
	 * {@code VPCMPGTB k1 {k2}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG 64 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPGTB_KR_K1_YMM_YMMM256 = 1516;
	/**
	 * {@code VPCMPGTB k1 {k2}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG 64 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPGTB_KR_K1_ZMM_ZMMM512 = 1517;
	/**
	 * {@code PCMPGTW mm, mm/m64}<p>
	 * {@code NP 0F 65 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPGTW_MM_MMM64 = 1518;
	/**
	 * {@code PCMPGTW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 65 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPGTW_XMM_XMMM128 = 1519;
	/**
	 * {@code VPCMPGTW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 65 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPGTW_XMM_XMM_XMMM128 = 1520;
	/**
	 * {@code VPCMPGTW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 65 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPGTW_YMM_YMM_YMMM256 = 1521;
	/**
	 * {@code VPCMPGTW k1 {k2}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG 65 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPGTW_KR_K1_XMM_XMMM128 = 1522;
	/**
	 * {@code VPCMPGTW k1 {k2}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG 65 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPGTW_KR_K1_YMM_YMMM256 = 1523;
	/**
	 * {@code VPCMPGTW k1 {k2}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG 65 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPGTW_KR_K1_ZMM_ZMMM512 = 1524;
	/**
	 * {@code PCMPGTD mm, mm/m64}<p>
	 * {@code NP 0F 66 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPGTD_MM_MMM64 = 1525;
	/**
	 * {@code PCMPGTD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 66 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPGTD_XMM_XMMM128 = 1526;
	/**
	 * {@code VPCMPGTD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 66 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPGTD_XMM_XMM_XMMM128 = 1527;
	/**
	 * {@code VPCMPGTD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 66 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPGTD_YMM_YMM_YMMM256 = 1528;
	/**
	 * {@code VPCMPGTD k1 {k2}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 66 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPGTD_KR_K1_XMM_XMMM128B32 = 1529;
	/**
	 * {@code VPCMPGTD k1 {k2}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 66 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPGTD_KR_K1_YMM_YMMM256B32 = 1530;
	/**
	 * {@code VPCMPGTD k1 {k2}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F.W0 66 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPGTD_KR_K1_ZMM_ZMMM512B32 = 1531;
	/**
	 * {@code PACKUSWB mm, mm/m64}<p>
	 * {@code NP 0F 67 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PACKUSWB_MM_MMM64 = 1532;
	/**
	 * {@code PACKUSWB xmm1, xmm2/m128}<p>
	 * {@code 66 0F 67 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PACKUSWB_XMM_XMMM128 = 1533;
	/**
	 * {@code VPACKUSWB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 67 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPACKUSWB_XMM_XMM_XMMM128 = 1534;
	/**
	 * {@code VPACKUSWB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 67 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPACKUSWB_YMM_YMM_YMMM256 = 1535;
	/**
	 * {@code VPACKUSWB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG 67 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPACKUSWB_XMM_K1Z_XMM_XMMM128 = 1536;
	/**
	 * {@code VPACKUSWB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG 67 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPACKUSWB_YMM_K1Z_YMM_YMMM256 = 1537;
	/**
	 * {@code VPACKUSWB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG 67 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPACKUSWB_ZMM_K1Z_ZMM_ZMMM512 = 1538;
	/**
	 * {@code PUNPCKHBW mm, mm/m64}<p>
	 * {@code NP 0F 68 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUNPCKHBW_MM_MMM64 = 1539;
	/**
	 * {@code PUNPCKHBW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 68 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUNPCKHBW_XMM_XMMM128 = 1540;
	/**
	 * {@code VPUNPCKHBW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 68 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKHBW_XMM_XMM_XMMM128 = 1541;
	/**
	 * {@code VPUNPCKHBW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 68 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKHBW_YMM_YMM_YMMM256 = 1542;
	/**
	 * {@code VPUNPCKHBW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG 68 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKHBW_XMM_K1Z_XMM_XMMM128 = 1543;
	/**
	 * {@code VPUNPCKHBW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG 68 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKHBW_YMM_K1Z_YMM_YMMM256 = 1544;
	/**
	 * {@code VPUNPCKHBW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG 68 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKHBW_ZMM_K1Z_ZMM_ZMMM512 = 1545;
	/**
	 * {@code PUNPCKHWD mm, mm/m64}<p>
	 * {@code NP 0F 69 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUNPCKHWD_MM_MMM64 = 1546;
	/**
	 * {@code PUNPCKHWD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 69 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUNPCKHWD_XMM_XMMM128 = 1547;
	/**
	 * {@code VPUNPCKHWD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 69 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKHWD_XMM_XMM_XMMM128 = 1548;
	/**
	 * {@code VPUNPCKHWD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 69 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKHWD_YMM_YMM_YMMM256 = 1549;
	/**
	 * {@code VPUNPCKHWD xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG 69 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKHWD_XMM_K1Z_XMM_XMMM128 = 1550;
	/**
	 * {@code VPUNPCKHWD ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG 69 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKHWD_YMM_K1Z_YMM_YMMM256 = 1551;
	/**
	 * {@code VPUNPCKHWD zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG 69 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKHWD_ZMM_K1Z_ZMM_ZMMM512 = 1552;
	/**
	 * {@code PUNPCKHDQ mm, mm/m64}<p>
	 * {@code NP 0F 6A /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUNPCKHDQ_MM_MMM64 = 1553;
	/**
	 * {@code PUNPCKHDQ xmm1, xmm2/m128}<p>
	 * {@code 66 0F 6A /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUNPCKHDQ_XMM_XMMM128 = 1554;
	/**
	 * {@code VPUNPCKHDQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 6A /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKHDQ_XMM_XMM_XMMM128 = 1555;
	/**
	 * {@code VPUNPCKHDQ ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 6A /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKHDQ_YMM_YMM_YMMM256 = 1556;
	/**
	 * {@code VPUNPCKHDQ xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 6A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKHDQ_XMM_K1Z_XMM_XMMM128B32 = 1557;
	/**
	 * {@code VPUNPCKHDQ ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 6A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKHDQ_YMM_K1Z_YMM_YMMM256B32 = 1558;
	/**
	 * {@code VPUNPCKHDQ zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F.W0 6A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKHDQ_ZMM_K1Z_ZMM_ZMMM512B32 = 1559;
	/**
	 * {@code PACKSSDW mm1, mm2/m64}<p>
	 * {@code NP 0F 6B /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PACKSSDW_MM_MMM64 = 1560;
	/**
	 * {@code PACKSSDW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 6B /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PACKSSDW_XMM_XMMM128 = 1561;
	/**
	 * {@code VPACKSSDW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 6B /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPACKSSDW_XMM_XMM_XMMM128 = 1562;
	/**
	 * {@code VPACKSSDW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 6B /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPACKSSDW_YMM_YMM_YMMM256 = 1563;
	/**
	 * {@code VPACKSSDW xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 6B /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPACKSSDW_XMM_K1Z_XMM_XMMM128B32 = 1564;
	/**
	 * {@code VPACKSSDW ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 6B /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPACKSSDW_YMM_K1Z_YMM_YMMM256B32 = 1565;
	/**
	 * {@code VPACKSSDW zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F.W0 6B /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPACKSSDW_ZMM_K1Z_ZMM_ZMMM512B32 = 1566;
	/**
	 * {@code PUNPCKLQDQ xmm1, xmm2/m128}<p>
	 * {@code 66 0F 6C /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUNPCKLQDQ_XMM_XMMM128 = 1567;
	/**
	 * {@code VPUNPCKLQDQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 6C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKLQDQ_XMM_XMM_XMMM128 = 1568;
	/**
	 * {@code VPUNPCKLQDQ ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 6C /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKLQDQ_YMM_YMM_YMMM256 = 1569;
	/**
	 * {@code VPUNPCKLQDQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 6C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKLQDQ_XMM_K1Z_XMM_XMMM128B64 = 1570;
	/**
	 * {@code VPUNPCKLQDQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 6C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKLQDQ_YMM_K1Z_YMM_YMMM256B64 = 1571;
	/**
	 * {@code VPUNPCKLQDQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F.W1 6C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKLQDQ_ZMM_K1Z_ZMM_ZMMM512B64 = 1572;
	/**
	 * {@code PUNPCKHQDQ xmm1, xmm2/m128}<p>
	 * {@code 66 0F 6D /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUNPCKHQDQ_XMM_XMMM128 = 1573;
	/**
	 * {@code VPUNPCKHQDQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 6D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKHQDQ_XMM_XMM_XMMM128 = 1574;
	/**
	 * {@code VPUNPCKHQDQ ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 6D /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPUNPCKHQDQ_YMM_YMM_YMMM256 = 1575;
	/**
	 * {@code VPUNPCKHQDQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 6D /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKHQDQ_XMM_K1Z_XMM_XMMM128B64 = 1576;
	/**
	 * {@code VPUNPCKHQDQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 6D /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKHQDQ_YMM_K1Z_YMM_YMMM256B64 = 1577;
	/**
	 * {@code VPUNPCKHQDQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F.W1 6D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPUNPCKHQDQ_ZMM_K1Z_ZMM_ZMMM512B64 = 1578;
	/**
	 * {@code MOVD mm, r/m32}<p>
	 * {@code NP 0F 6E /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVD_MM_RM32 = 1579;
	/**
	 * {@code MOVQ mm, r/m64}<p>
	 * {@code NP o64 0F 6E /r}<p>
	 * {@code MMX}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVQ_MM_RM64 = 1580;
	/**
	 * {@code MOVD xmm, r/m32}<p>
	 * {@code 66 0F 6E /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVD_XMM_RM32 = 1581;
	/**
	 * {@code MOVQ xmm, r/m64}<p>
	 * {@code 66 o64 0F 6E /r}<p>
	 * {@code SSE2}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVQ_XMM_RM64 = 1582;
	/**
	 * {@code VMOVD xmm1, r/m32}<p>
	 * {@code VEX.128.66.0F.W0 6E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVD_XMM_RM32 = 1583;
	/**
	 * {@code VMOVQ xmm1, r/m64}<p>
	 * {@code VEX.128.66.0F.W1 6E /r}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VMOVQ_XMM_RM64 = 1584;
	/**
	 * {@code VMOVD xmm1, r/m32}<p>
	 * {@code EVEX.128.66.0F.W0 6E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVD_XMM_RM32 = 1585;
	/**
	 * {@code VMOVQ xmm1, r/m64}<p>
	 * {@code EVEX.128.66.0F.W1 6E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VMOVQ_XMM_RM64 = 1586;
	/**
	 * {@code MOVQ mm, mm/m64}<p>
	 * {@code NP 0F 6F /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVQ_MM_MMM64 = 1587;
	/**
	 * {@code MOVDQA xmm1, xmm2/m128}<p>
	 * {@code 66 0F 6F /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVDQA_XMM_XMMM128 = 1588;
	/**
	 * {@code VMOVDQA xmm1, xmm2/m128}<p>
	 * {@code VEX.128.66.0F.WIG 6F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVDQA_XMM_XMMM128 = 1589;
	/**
	 * {@code VMOVDQA ymm1, ymm2/m256}<p>
	 * {@code VEX.256.66.0F.WIG 6F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVDQA_YMM_YMMM256 = 1590;
	/**
	 * {@code VMOVDQA32 xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.66.0F.W0 6F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQA32_XMM_K1Z_XMMM128 = 1591;
	/**
	 * {@code VMOVDQA32 ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.66.0F.W0 6F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQA32_YMM_K1Z_YMMM256 = 1592;
	/**
	 * {@code VMOVDQA32 zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.66.0F.W0 6F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQA32_ZMM_K1Z_ZMMM512 = 1593;
	/**
	 * {@code VMOVDQA64 xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.66.0F.W1 6F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQA64_XMM_K1Z_XMMM128 = 1594;
	/**
	 * {@code VMOVDQA64 ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.66.0F.W1 6F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQA64_YMM_K1Z_YMMM256 = 1595;
	/**
	 * {@code VMOVDQA64 zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.66.0F.W1 6F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQA64_ZMM_K1Z_ZMMM512 = 1596;
	/**
	 * {@code MOVDQU xmm1, xmm2/m128}<p>
	 * {@code F3 0F 6F /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVDQU_XMM_XMMM128 = 1597;
	/**
	 * {@code VMOVDQU xmm1, xmm2/m128}<p>
	 * {@code VEX.128.F3.0F.WIG 6F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVDQU_XMM_XMMM128 = 1598;
	/**
	 * {@code VMOVDQU ymm1, ymm2/m256}<p>
	 * {@code VEX.256.F3.0F.WIG 6F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVDQU_YMM_YMMM256 = 1599;
	/**
	 * {@code VMOVDQU32 xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.F3.0F.W0 6F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU32_XMM_K1Z_XMMM128 = 1600;
	/**
	 * {@code VMOVDQU32 ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.F3.0F.W0 6F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU32_YMM_K1Z_YMMM256 = 1601;
	/**
	 * {@code VMOVDQU32 zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.F3.0F.W0 6F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU32_ZMM_K1Z_ZMMM512 = 1602;
	/**
	 * {@code VMOVDQU64 xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.F3.0F.W1 6F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU64_XMM_K1Z_XMMM128 = 1603;
	/**
	 * {@code VMOVDQU64 ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.F3.0F.W1 6F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU64_YMM_K1Z_YMMM256 = 1604;
	/**
	 * {@code VMOVDQU64 zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.F3.0F.W1 6F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU64_ZMM_K1Z_ZMMM512 = 1605;
	/**
	 * {@code VMOVDQU8 xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.F2.0F.W0 6F /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU8_XMM_K1Z_XMMM128 = 1606;
	/**
	 * {@code VMOVDQU8 ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.F2.0F.W0 6F /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU8_YMM_K1Z_YMMM256 = 1607;
	/**
	 * {@code VMOVDQU8 zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.F2.0F.W0 6F /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU8_ZMM_K1Z_ZMMM512 = 1608;
	/**
	 * {@code VMOVDQU16 xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.F2.0F.W1 6F /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU16_XMM_K1Z_XMMM128 = 1609;
	/**
	 * {@code VMOVDQU16 ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.F2.0F.W1 6F /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU16_YMM_K1Z_YMMM256 = 1610;
	/**
	 * {@code VMOVDQU16 zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.F2.0F.W1 6F /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU16_ZMM_K1Z_ZMMM512 = 1611;
	/**
	 * {@code PSHUFW mm1, mm2/m64, imm8}<p>
	 * {@code NP 0F 70 /r ib}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSHUFW_MM_MMM64_IMM8 = 1612;
	/**
	 * {@code PSHUFD xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 70 /r ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSHUFD_XMM_XMMM128_IMM8 = 1613;
	/**
	 * {@code VPSHUFD xmm1, xmm2/m128, imm8}<p>
	 * {@code VEX.128.66.0F.WIG 70 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSHUFD_XMM_XMMM128_IMM8 = 1614;
	/**
	 * {@code VPSHUFD ymm1, ymm2/m256, imm8}<p>
	 * {@code VEX.256.66.0F.WIG 70 /r ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSHUFD_YMM_YMMM256_IMM8 = 1615;
	/**
	 * {@code VPSHUFD xmm1 {k1}{z}, xmm2/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F.W0 70 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHUFD_XMM_K1Z_XMMM128B32_IMM8 = 1616;
	/**
	 * {@code VPSHUFD ymm1 {k1}{z}, ymm2/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F.W0 70 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHUFD_YMM_K1Z_YMMM256B32_IMM8 = 1617;
	/**
	 * {@code VPSHUFD zmm1 {k1}{z}, zmm2/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F.W0 70 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHUFD_ZMM_K1Z_ZMMM512B32_IMM8 = 1618;
	/**
	 * {@code PSHUFHW xmm1, xmm2/m128, imm8}<p>
	 * {@code F3 0F 70 /r ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSHUFHW_XMM_XMMM128_IMM8 = 1619;
	/**
	 * {@code VPSHUFHW xmm1, xmm2/m128, imm8}<p>
	 * {@code VEX.128.F3.0F.WIG 70 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSHUFHW_XMM_XMMM128_IMM8 = 1620;
	/**
	 * {@code VPSHUFHW ymm1, ymm2/m256, imm8}<p>
	 * {@code VEX.256.F3.0F.WIG 70 /r ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSHUFHW_YMM_YMMM256_IMM8 = 1621;
	/**
	 * {@code VPSHUFHW xmm1 {k1}{z}, xmm2/m128, imm8}<p>
	 * {@code EVEX.128.F3.0F.WIG 70 /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHUFHW_XMM_K1Z_XMMM128_IMM8 = 1622;
	/**
	 * {@code VPSHUFHW ymm1 {k1}{z}, ymm2/m256, imm8}<p>
	 * {@code EVEX.256.F3.0F.WIG 70 /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHUFHW_YMM_K1Z_YMMM256_IMM8 = 1623;
	/**
	 * {@code VPSHUFHW zmm1 {k1}{z}, zmm2/m512, imm8}<p>
	 * {@code EVEX.512.F3.0F.WIG 70 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHUFHW_ZMM_K1Z_ZMMM512_IMM8 = 1624;
	/**
	 * {@code PSHUFLW xmm1, xmm2/m128, imm8}<p>
	 * {@code F2 0F 70 /r ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSHUFLW_XMM_XMMM128_IMM8 = 1625;
	/**
	 * {@code VPSHUFLW xmm1, xmm2/m128, imm8}<p>
	 * {@code VEX.128.F2.0F.WIG 70 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSHUFLW_XMM_XMMM128_IMM8 = 1626;
	/**
	 * {@code VPSHUFLW ymm1, ymm2/m256, imm8}<p>
	 * {@code VEX.256.F2.0F.WIG 70 /r ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSHUFLW_YMM_YMMM256_IMM8 = 1627;
	/**
	 * {@code VPSHUFLW xmm1 {k1}{z}, xmm2/m128, imm8}<p>
	 * {@code EVEX.128.F2.0F.WIG 70 /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHUFLW_XMM_K1Z_XMMM128_IMM8 = 1628;
	/**
	 * {@code VPSHUFLW ymm1 {k1}{z}, ymm2/m256, imm8}<p>
	 * {@code EVEX.256.F2.0F.WIG 70 /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHUFLW_YMM_K1Z_YMMM256_IMM8 = 1629;
	/**
	 * {@code VPSHUFLW zmm1 {k1}{z}, zmm2/m512, imm8}<p>
	 * {@code EVEX.512.F2.0F.WIG 70 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHUFLW_ZMM_K1Z_ZMMM512_IMM8 = 1630;
	/**
	 * {@code PSRLW mm, imm8}<p>
	 * {@code NP 0F 71 /2 ib}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRLW_MM_IMM8 = 1631;
	/**
	 * {@code PSRLW xmm1, imm8}<p>
	 * {@code 66 0F 71 /2 ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRLW_XMM_IMM8 = 1632;
	/**
	 * {@code VPSRLW xmm1, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F.WIG 71 /2 ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLW_XMM_XMM_IMM8 = 1633;
	/**
	 * {@code VPSRLW ymm1, ymm2, imm8}<p>
	 * {@code VEX.256.66.0F.WIG 71 /2 ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLW_YMM_YMM_IMM8 = 1634;
	/**
	 * {@code VPSRLW xmm1 {k1}{z}, xmm2/m128, imm8}<p>
	 * {@code EVEX.128.66.0F.WIG 71 /2 ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLW_XMM_K1Z_XMMM128_IMM8 = 1635;
	/**
	 * {@code VPSRLW ymm1 {k1}{z}, ymm2/m256, imm8}<p>
	 * {@code EVEX.256.66.0F.WIG 71 /2 ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLW_YMM_K1Z_YMMM256_IMM8 = 1636;
	/**
	 * {@code VPSRLW zmm1 {k1}{z}, zmm2/m512, imm8}<p>
	 * {@code EVEX.512.66.0F.WIG 71 /2 ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLW_ZMM_K1Z_ZMMM512_IMM8 = 1637;
	/**
	 * {@code PSRAW mm, imm8}<p>
	 * {@code NP 0F 71 /4 ib}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRAW_MM_IMM8 = 1638;
	/**
	 * {@code PSRAW xmm1, imm8}<p>
	 * {@code 66 0F 71 /4 ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRAW_XMM_IMM8 = 1639;
	/**
	 * {@code VPSRAW xmm1, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F.WIG 71 /4 ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRAW_XMM_XMM_IMM8 = 1640;
	/**
	 * {@code VPSRAW ymm1, ymm2, imm8}<p>
	 * {@code VEX.256.66.0F.WIG 71 /4 ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRAW_YMM_YMM_IMM8 = 1641;
	/**
	 * {@code VPSRAW xmm1 {k1}{z}, xmm2/m128, imm8}<p>
	 * {@code EVEX.128.66.0F.WIG 71 /4 ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAW_XMM_K1Z_XMMM128_IMM8 = 1642;
	/**
	 * {@code VPSRAW ymm1 {k1}{z}, ymm2/m256, imm8}<p>
	 * {@code EVEX.256.66.0F.WIG 71 /4 ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAW_YMM_K1Z_YMMM256_IMM8 = 1643;
	/**
	 * {@code VPSRAW zmm1 {k1}{z}, zmm2/m512, imm8}<p>
	 * {@code EVEX.512.66.0F.WIG 71 /4 ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAW_ZMM_K1Z_ZMMM512_IMM8 = 1644;
	/**
	 * {@code PSLLW mm1, imm8}<p>
	 * {@code NP 0F 71 /6 ib}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSLLW_MM_IMM8 = 1645;
	/**
	 * {@code PSLLW xmm1, imm8}<p>
	 * {@code 66 0F 71 /6 ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSLLW_XMM_IMM8 = 1646;
	/**
	 * {@code VPSLLW xmm1, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F.WIG 71 /6 ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLW_XMM_XMM_IMM8 = 1647;
	/**
	 * {@code VPSLLW ymm1, ymm2, imm8}<p>
	 * {@code VEX.256.66.0F.WIG 71 /6 ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLW_YMM_YMM_IMM8 = 1648;
	/**
	 * {@code VPSLLW xmm1 {k1}{z}, xmm2/m128, imm8}<p>
	 * {@code EVEX.128.66.0F.WIG 71 /6 ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLW_XMM_K1Z_XMMM128_IMM8 = 1649;
	/**
	 * {@code VPSLLW ymm1 {k1}{z}, ymm2/m256, imm8}<p>
	 * {@code EVEX.256.66.0F.WIG 71 /6 ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLW_YMM_K1Z_YMMM256_IMM8 = 1650;
	/**
	 * {@code VPSLLW zmm1 {k1}{z}, zmm2/m512, imm8}<p>
	 * {@code EVEX.512.66.0F.WIG 71 /6 ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLW_ZMM_K1Z_ZMMM512_IMM8 = 1651;
	/**
	 * {@code VPRORD xmm1 {k1}{z}, xmm2/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F.W0 72 /0 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPRORD_XMM_K1Z_XMMM128B32_IMM8 = 1652;
	/**
	 * {@code VPRORD ymm1 {k1}{z}, ymm2/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F.W0 72 /0 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPRORD_YMM_K1Z_YMMM256B32_IMM8 = 1653;
	/**
	 * {@code VPRORD zmm1 {k1}{z}, zmm2/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F.W0 72 /0 ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPRORD_ZMM_K1Z_ZMMM512B32_IMM8 = 1654;
	/**
	 * {@code VPRORQ xmm1 {k1}{z}, xmm2/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F.W1 72 /0 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPRORQ_XMM_K1Z_XMMM128B64_IMM8 = 1655;
	/**
	 * {@code VPRORQ ymm1 {k1}{z}, ymm2/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F.W1 72 /0 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPRORQ_YMM_K1Z_YMMM256B64_IMM8 = 1656;
	/**
	 * {@code VPRORQ zmm1 {k1}{z}, zmm2/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F.W1 72 /0 ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPRORQ_ZMM_K1Z_ZMMM512B64_IMM8 = 1657;
	/**
	 * {@code VPROLD xmm1 {k1}{z}, xmm2/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F.W0 72 /1 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPROLD_XMM_K1Z_XMMM128B32_IMM8 = 1658;
	/**
	 * {@code VPROLD ymm1 {k1}{z}, ymm2/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F.W0 72 /1 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPROLD_YMM_K1Z_YMMM256B32_IMM8 = 1659;
	/**
	 * {@code VPROLD zmm1 {k1}{z}, zmm2/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F.W0 72 /1 ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPROLD_ZMM_K1Z_ZMMM512B32_IMM8 = 1660;
	/**
	 * {@code VPROLQ xmm1 {k1}{z}, xmm2/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F.W1 72 /1 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPROLQ_XMM_K1Z_XMMM128B64_IMM8 = 1661;
	/**
	 * {@code VPROLQ ymm1 {k1}{z}, ymm2/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F.W1 72 /1 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPROLQ_YMM_K1Z_YMMM256B64_IMM8 = 1662;
	/**
	 * {@code VPROLQ zmm1 {k1}{z}, zmm2/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F.W1 72 /1 ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPROLQ_ZMM_K1Z_ZMMM512B64_IMM8 = 1663;
	/**
	 * {@code PSRLD mm, imm8}<p>
	 * {@code NP 0F 72 /2 ib}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRLD_MM_IMM8 = 1664;
	/**
	 * {@code PSRLD xmm1, imm8}<p>
	 * {@code 66 0F 72 /2 ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRLD_XMM_IMM8 = 1665;
	/**
	 * {@code VPSRLD xmm1, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F.WIG 72 /2 ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLD_XMM_XMM_IMM8 = 1666;
	/**
	 * {@code VPSRLD ymm1, ymm2, imm8}<p>
	 * {@code VEX.256.66.0F.WIG 72 /2 ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLD_YMM_YMM_IMM8 = 1667;
	/**
	 * {@code VPSRLD xmm1 {k1}{z}, xmm2/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F.W0 72 /2 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLD_XMM_K1Z_XMMM128B32_IMM8 = 1668;
	/**
	 * {@code VPSRLD ymm1 {k1}{z}, ymm2/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F.W0 72 /2 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLD_YMM_K1Z_YMMM256B32_IMM8 = 1669;
	/**
	 * {@code VPSRLD zmm1 {k1}{z}, zmm2/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F.W0 72 /2 ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLD_ZMM_K1Z_ZMMM512B32_IMM8 = 1670;
	/**
	 * {@code PSRAD mm, imm8}<p>
	 * {@code NP 0F 72 /4 ib}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRAD_MM_IMM8 = 1671;
	/**
	 * {@code PSRAD xmm1, imm8}<p>
	 * {@code 66 0F 72 /4 ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRAD_XMM_IMM8 = 1672;
	/**
	 * {@code VPSRAD xmm1, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F.WIG 72 /4 ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRAD_XMM_XMM_IMM8 = 1673;
	/**
	 * {@code VPSRAD ymm1, ymm2, imm8}<p>
	 * {@code VEX.256.66.0F.WIG 72 /4 ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRAD_YMM_YMM_IMM8 = 1674;
	/**
	 * {@code VPSRAD xmm1 {k1}{z}, xmm2/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F.W0 72 /4 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAD_XMM_K1Z_XMMM128B32_IMM8 = 1675;
	/**
	 * {@code VPSRAD ymm1 {k1}{z}, ymm2/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F.W0 72 /4 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAD_YMM_K1Z_YMMM256B32_IMM8 = 1676;
	/**
	 * {@code VPSRAD zmm1 {k1}{z}, zmm2/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F.W0 72 /4 ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAD_ZMM_K1Z_ZMMM512B32_IMM8 = 1677;
	/**
	 * {@code VPSRAQ xmm1 {k1}{z}, xmm2/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F.W1 72 /4 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAQ_XMM_K1Z_XMMM128B64_IMM8 = 1678;
	/**
	 * {@code VPSRAQ ymm1 {k1}{z}, ymm2/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F.W1 72 /4 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAQ_YMM_K1Z_YMMM256B64_IMM8 = 1679;
	/**
	 * {@code VPSRAQ zmm1 {k1}{z}, zmm2/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F.W1 72 /4 ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAQ_ZMM_K1Z_ZMMM512B64_IMM8 = 1680;
	/**
	 * {@code PSLLD mm, imm8}<p>
	 * {@code NP 0F 72 /6 ib}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSLLD_MM_IMM8 = 1681;
	/**
	 * {@code PSLLD xmm1, imm8}<p>
	 * {@code 66 0F 72 /6 ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSLLD_XMM_IMM8 = 1682;
	/**
	 * {@code VPSLLD xmm1, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F.WIG 72 /6 ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLD_XMM_XMM_IMM8 = 1683;
	/**
	 * {@code VPSLLD ymm1, ymm2, imm8}<p>
	 * {@code VEX.256.66.0F.WIG 72 /6 ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLD_YMM_YMM_IMM8 = 1684;
	/**
	 * {@code VPSLLD xmm1 {k1}{z}, xmm2/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F.W0 72 /6 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLD_XMM_K1Z_XMMM128B32_IMM8 = 1685;
	/**
	 * {@code VPSLLD ymm1 {k1}{z}, ymm2/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F.W0 72 /6 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLD_YMM_K1Z_YMMM256B32_IMM8 = 1686;
	/**
	 * {@code VPSLLD zmm1 {k1}{z}, zmm2/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F.W0 72 /6 ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLD_ZMM_K1Z_ZMMM512B32_IMM8 = 1687;
	/**
	 * {@code PSRLQ mm, imm8}<p>
	 * {@code NP 0F 73 /2 ib}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRLQ_MM_IMM8 = 1688;
	/**
	 * {@code PSRLQ xmm1, imm8}<p>
	 * {@code 66 0F 73 /2 ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRLQ_XMM_IMM8 = 1689;
	/**
	 * {@code VPSRLQ xmm1, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F.WIG 73 /2 ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLQ_XMM_XMM_IMM8 = 1690;
	/**
	 * {@code VPSRLQ ymm1, ymm2, imm8}<p>
	 * {@code VEX.256.66.0F.WIG 73 /2 ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLQ_YMM_YMM_IMM8 = 1691;
	/**
	 * {@code VPSRLQ xmm1 {k1}{z}, xmm2/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F.W1 73 /2 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLQ_XMM_K1Z_XMMM128B64_IMM8 = 1692;
	/**
	 * {@code VPSRLQ ymm1 {k1}{z}, ymm2/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F.W1 73 /2 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLQ_YMM_K1Z_YMMM256B64_IMM8 = 1693;
	/**
	 * {@code VPSRLQ zmm1 {k1}{z}, zmm2/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F.W1 73 /2 ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLQ_ZMM_K1Z_ZMMM512B64_IMM8 = 1694;
	/**
	 * {@code PSRLDQ xmm1, imm8}<p>
	 * {@code 66 0F 73 /3 ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRLDQ_XMM_IMM8 = 1695;
	/**
	 * {@code VPSRLDQ xmm1, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F.WIG 73 /3 ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLDQ_XMM_XMM_IMM8 = 1696;
	/**
	 * {@code VPSRLDQ ymm1, ymm2, imm8}<p>
	 * {@code VEX.256.66.0F.WIG 73 /3 ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLDQ_YMM_YMM_IMM8 = 1697;
	/**
	 * {@code VPSRLDQ xmm1, xmm2/m128, imm8}<p>
	 * {@code EVEX.128.66.0F.WIG 73 /3 ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLDQ_XMM_XMMM128_IMM8 = 1698;
	/**
	 * {@code VPSRLDQ ymm1, ymm2/m256, imm8}<p>
	 * {@code EVEX.256.66.0F.WIG 73 /3 ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLDQ_YMM_YMMM256_IMM8 = 1699;
	/**
	 * {@code VPSRLDQ zmm1, zmm2/m512, imm8}<p>
	 * {@code EVEX.512.66.0F.WIG 73 /3 ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLDQ_ZMM_ZMMM512_IMM8 = 1700;
	/**
	 * {@code PSLLQ mm, imm8}<p>
	 * {@code NP 0F 73 /6 ib}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSLLQ_MM_IMM8 = 1701;
	/**
	 * {@code PSLLQ xmm1, imm8}<p>
	 * {@code 66 0F 73 /6 ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSLLQ_XMM_IMM8 = 1702;
	/**
	 * {@code VPSLLQ xmm1, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F.WIG 73 /6 ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLQ_XMM_XMM_IMM8 = 1703;
	/**
	 * {@code VPSLLQ ymm1, ymm2, imm8}<p>
	 * {@code VEX.256.66.0F.WIG 73 /6 ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLQ_YMM_YMM_IMM8 = 1704;
	/**
	 * {@code VPSLLQ xmm1 {k1}{z}, xmm2/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F.W1 73 /6 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLQ_XMM_K1Z_XMMM128B64_IMM8 = 1705;
	/**
	 * {@code VPSLLQ ymm1 {k1}{z}, ymm2/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F.W1 73 /6 ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLQ_YMM_K1Z_YMMM256B64_IMM8 = 1706;
	/**
	 * {@code VPSLLQ zmm1 {k1}{z}, zmm2/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F.W1 73 /6 ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLQ_ZMM_K1Z_ZMMM512B64_IMM8 = 1707;
	/**
	 * {@code PSLLDQ xmm1, imm8}<p>
	 * {@code 66 0F 73 /7 ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSLLDQ_XMM_IMM8 = 1708;
	/**
	 * {@code VPSLLDQ xmm1, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F.WIG 73 /7 ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLDQ_XMM_XMM_IMM8 = 1709;
	/**
	 * {@code VPSLLDQ ymm1, ymm2, imm8}<p>
	 * {@code VEX.256.66.0F.WIG 73 /7 ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLDQ_YMM_YMM_IMM8 = 1710;
	/**
	 * {@code VPSLLDQ xmm1, xmm2/m128, imm8}<p>
	 * {@code EVEX.128.66.0F.WIG 73 /7 ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLDQ_XMM_XMMM128_IMM8 = 1711;
	/**
	 * {@code VPSLLDQ ymm1, ymm2/m256, imm8}<p>
	 * {@code EVEX.256.66.0F.WIG 73 /7 ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLDQ_YMM_YMMM256_IMM8 = 1712;
	/**
	 * {@code VPSLLDQ zmm1, zmm2/m512, imm8}<p>
	 * {@code EVEX.512.66.0F.WIG 73 /7 ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLDQ_ZMM_ZMMM512_IMM8 = 1713;
	/**
	 * {@code PCMPEQB mm, mm/m64}<p>
	 * {@code NP 0F 74 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPEQB_MM_MMM64 = 1714;
	/**
	 * {@code PCMPEQB xmm1, xmm2/m128}<p>
	 * {@code 66 0F 74 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPEQB_XMM_XMMM128 = 1715;
	/**
	 * {@code VPCMPEQB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 74 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPEQB_XMM_XMM_XMMM128 = 1716;
	/**
	 * {@code VPCMPEQB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 74 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPEQB_YMM_YMM_YMMM256 = 1717;
	/**
	 * {@code VPCMPEQB k1 {k2}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG 74 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPEQB_KR_K1_XMM_XMMM128 = 1718;
	/**
	 * {@code VPCMPEQB k1 {k2}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG 74 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPEQB_KR_K1_YMM_YMMM256 = 1719;
	/**
	 * {@code VPCMPEQB k1 {k2}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG 74 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPEQB_KR_K1_ZMM_ZMMM512 = 1720;
	/**
	 * {@code PCMPEQW mm, mm/m64}<p>
	 * {@code NP 0F 75 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPEQW_MM_MMM64 = 1721;
	/**
	 * {@code PCMPEQW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 75 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPEQW_XMM_XMMM128 = 1722;
	/**
	 * {@code VPCMPEQW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 75 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPEQW_XMM_XMM_XMMM128 = 1723;
	/**
	 * {@code VPCMPEQW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 75 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPEQW_YMM_YMM_YMMM256 = 1724;
	/**
	 * {@code VPCMPEQW k1 {k2}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG 75 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPEQW_KR_K1_XMM_XMMM128 = 1725;
	/**
	 * {@code VPCMPEQW k1 {k2}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG 75 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPEQW_KR_K1_YMM_YMMM256 = 1726;
	/**
	 * {@code VPCMPEQW k1 {k2}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG 75 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPEQW_KR_K1_ZMM_ZMMM512 = 1727;
	/**
	 * {@code PCMPEQD mm, mm/m64}<p>
	 * {@code NP 0F 76 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPEQD_MM_MMM64 = 1728;
	/**
	 * {@code PCMPEQD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 76 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPEQD_XMM_XMMM128 = 1729;
	/**
	 * {@code VPCMPEQD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 76 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPEQD_XMM_XMM_XMMM128 = 1730;
	/**
	 * {@code VPCMPEQD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 76 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPEQD_YMM_YMM_YMMM256 = 1731;
	/**
	 * {@code VPCMPEQD k1 {k2}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 76 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPEQD_KR_K1_XMM_XMMM128B32 = 1732;
	/**
	 * {@code VPCMPEQD k1 {k2}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 76 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPEQD_KR_K1_YMM_YMMM256B32 = 1733;
	/**
	 * {@code VPCMPEQD k1 {k2}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F.W0 76 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPEQD_KR_K1_ZMM_ZMMM512B32 = 1734;
	/**
	 * {@code EMMS}<p>
	 * {@code NP 0F 77}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EMMS = 1735;
	/**
	 * {@code VZEROUPPER}<p>
	 * {@code VEX.128.0F.WIG 77}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VZEROUPPER = 1736;
	/**
	 * {@code VZEROALL}<p>
	 * {@code VEX.256.0F.WIG 77}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VZEROALL = 1737;
	/**
	 * {@code VMREAD r/m32, r32}<p>
	 * {@code NP 0F 78 /r}<p>
	 * {@code VMX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int VMREAD_RM32_R32 = 1738;
	/**
	 * {@code VMREAD r/m64, r64}<p>
	 * {@code NP 0F 78 /r}<p>
	 * {@code VMX}<p>
	 * {@code 64-bit}
	 */
	public static final int VMREAD_RM64_R64 = 1739;
	/**
	 * {@code VCVTTPS2UDQ xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 78 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPS2UDQ_XMM_K1Z_XMMM128B32 = 1740;
	/**
	 * {@code VCVTTPS2UDQ ymm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 78 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPS2UDQ_YMM_K1Z_YMMM256B32 = 1741;
	/**
	 * {@code VCVTTPS2UDQ zmm1 {k1}{z}, zmm2/m512/m32bcst{sae}}<p>
	 * {@code EVEX.512.0F.W0 78 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPS2UDQ_ZMM_K1Z_ZMMM512B32_SAE = 1742;
	/**
	 * {@code VCVTTPD2UDQ xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.0F.W1 78 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPD2UDQ_XMM_K1Z_XMMM128B64 = 1743;
	/**
	 * {@code VCVTTPD2UDQ xmm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.0F.W1 78 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPD2UDQ_XMM_K1Z_YMMM256B64 = 1744;
	/**
	 * {@code VCVTTPD2UDQ ymm1 {k1}{z}, zmm2/m512/m64bcst{sae}}<p>
	 * {@code EVEX.512.0F.W1 78 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPD2UDQ_YMM_K1Z_ZMMM512B64_SAE = 1745;
	/**
	 * {@code EXTRQ xmm1, imm8, imm8}<p>
	 * {@code 66 0F 78 /0 ib ib}<p>
	 * {@code SSE4A}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EXTRQ_XMM_IMM8_IMM8 = 1746;
	/**
	 * {@code VCVTTPS2UQQ xmm1 {k1}{z}, xmm2/m64/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 78 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPS2UQQ_XMM_K1Z_XMMM64B32 = 1747;
	/**
	 * {@code VCVTTPS2UQQ ymm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 78 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPS2UQQ_YMM_K1Z_XMMM128B32 = 1748;
	/**
	 * {@code VCVTTPS2UQQ zmm1 {k1}{z}, ymm2/m256/m32bcst{sae}}<p>
	 * {@code EVEX.512.66.0F.W0 78 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPS2UQQ_ZMM_K1Z_YMMM256B32_SAE = 1749;
	/**
	 * {@code VCVTTPD2UQQ xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 78 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPD2UQQ_XMM_K1Z_XMMM128B64 = 1750;
	/**
	 * {@code VCVTTPD2UQQ ymm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 78 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPD2UQQ_YMM_K1Z_YMMM256B64 = 1751;
	/**
	 * {@code VCVTTPD2UQQ zmm1 {k1}{z}, zmm2/m512/m64bcst{sae}}<p>
	 * {@code EVEX.512.66.0F.W1 78 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPD2UQQ_ZMM_K1Z_ZMMM512B64_SAE = 1752;
	/**
	 * {@code VCVTTSS2USI r32, xmm1/m32{sae}}<p>
	 * {@code EVEX.LIG.F3.0F.W0 78 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTSS2USI_R32_XMMM32_SAE = 1753;
	/**
	 * {@code VCVTTSS2USI r64, xmm1/m32{sae}}<p>
	 * {@code EVEX.LIG.F3.0F.W1 78 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTTSS2USI_R64_XMMM32_SAE = 1754;
	/**
	 * {@code INSERTQ xmm1, xmm2, imm8, imm8}<p>
	 * {@code F2 0F 78 /r ib ib}<p>
	 * {@code SSE4A}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INSERTQ_XMM_XMM_IMM8_IMM8 = 1755;
	/**
	 * {@code VCVTTSD2USI r32, xmm1/m64{sae}}<p>
	 * {@code EVEX.LIG.F2.0F.W0 78 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTSD2USI_R32_XMMM64_SAE = 1756;
	/**
	 * {@code VCVTTSD2USI r64, xmm1/m64{sae}}<p>
	 * {@code EVEX.LIG.F2.0F.W1 78 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTTSD2USI_R64_XMMM64_SAE = 1757;
	/**
	 * {@code VMWRITE r32, r/m32}<p>
	 * {@code NP 0F 79 /r}<p>
	 * {@code VMX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int VMWRITE_R32_RM32 = 1758;
	/**
	 * {@code VMWRITE r64, r/m64}<p>
	 * {@code NP 0F 79 /r}<p>
	 * {@code VMX}<p>
	 * {@code 64-bit}
	 */
	public static final int VMWRITE_R64_RM64 = 1759;
	/**
	 * {@code VCVTPS2UDQ xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.0F.W0 79 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2UDQ_XMM_K1Z_XMMM128B32 = 1760;
	/**
	 * {@code VCVTPS2UDQ ymm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.0F.W0 79 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2UDQ_YMM_K1Z_YMMM256B32 = 1761;
	/**
	 * {@code VCVTPS2UDQ zmm1 {k1}{z}, zmm2/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.0F.W0 79 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2UDQ_ZMM_K1Z_ZMMM512B32_ER = 1762;
	/**
	 * {@code VCVTPD2UDQ xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.0F.W1 79 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2UDQ_XMM_K1Z_XMMM128B64 = 1763;
	/**
	 * {@code VCVTPD2UDQ xmm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.0F.W1 79 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2UDQ_XMM_K1Z_YMMM256B64 = 1764;
	/**
	 * {@code VCVTPD2UDQ ymm1 {k1}{z}, zmm2/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.0F.W1 79 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2UDQ_YMM_K1Z_ZMMM512B64_ER = 1765;
	/**
	 * {@code EXTRQ xmm1, xmm2}<p>
	 * {@code 66 0F 79 /r}<p>
	 * {@code SSE4A}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EXTRQ_XMM_XMM = 1766;
	/**
	 * {@code VCVTPS2UQQ xmm1 {k1}{z}, xmm2/m64/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 79 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2UQQ_XMM_K1Z_XMMM64B32 = 1767;
	/**
	 * {@code VCVTPS2UQQ ymm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 79 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2UQQ_YMM_K1Z_XMMM128B32 = 1768;
	/**
	 * {@code VCVTPS2UQQ zmm1 {k1}{z}, ymm2/m256/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F.W0 79 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2UQQ_ZMM_K1Z_YMMM256B32_ER = 1769;
	/**
	 * {@code VCVTPD2UQQ xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 79 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2UQQ_XMM_K1Z_XMMM128B64 = 1770;
	/**
	 * {@code VCVTPD2UQQ ymm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 79 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2UQQ_YMM_K1Z_YMMM256B64 = 1771;
	/**
	 * {@code VCVTPD2UQQ zmm1 {k1}{z}, zmm2/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F.W1 79 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2UQQ_ZMM_K1Z_ZMMM512B64_ER = 1772;
	/**
	 * {@code VCVTSS2USI r32, xmm1/m32{er}}<p>
	 * {@code EVEX.LIG.F3.0F.W0 79 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTSS2USI_R32_XMMM32_ER = 1773;
	/**
	 * {@code VCVTSS2USI r64, xmm1/m32{er}}<p>
	 * {@code EVEX.LIG.F3.0F.W1 79 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTSS2USI_R64_XMMM32_ER = 1774;
	/**
	 * {@code INSERTQ xmm1, xmm2}<p>
	 * {@code F2 0F 79 /r}<p>
	 * {@code SSE4A}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INSERTQ_XMM_XMM = 1775;
	/**
	 * {@code VCVTSD2USI r32, xmm1/m64{er}}<p>
	 * {@code EVEX.LIG.F2.0F.W0 79 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTSD2USI_R32_XMMM64_ER = 1776;
	/**
	 * {@code VCVTSD2USI r64, xmm1/m64{er}}<p>
	 * {@code EVEX.LIG.F2.0F.W1 79 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTSD2USI_R64_XMMM64_ER = 1777;
	/**
	 * {@code VCVTTPS2QQ xmm1 {k1}{z}, xmm2/m64/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 7A /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPS2QQ_XMM_K1Z_XMMM64B32 = 1778;
	/**
	 * {@code VCVTTPS2QQ ymm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 7A /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPS2QQ_YMM_K1Z_XMMM128B32 = 1779;
	/**
	 * {@code VCVTTPS2QQ zmm1 {k1}{z}, ymm2/m256/m32bcst{sae}}<p>
	 * {@code EVEX.512.66.0F.W0 7A /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPS2QQ_ZMM_K1Z_YMMM256B32_SAE = 1780;
	/**
	 * {@code VCVTTPD2QQ xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 7A /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPD2QQ_XMM_K1Z_XMMM128B64 = 1781;
	/**
	 * {@code VCVTTPD2QQ ymm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 7A /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPD2QQ_YMM_K1Z_YMMM256B64 = 1782;
	/**
	 * {@code VCVTTPD2QQ zmm1 {k1}{z}, zmm2/m512/m64bcst{sae}}<p>
	 * {@code EVEX.512.66.0F.W1 7A /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPD2QQ_ZMM_K1Z_ZMMM512B64_SAE = 1783;
	/**
	 * {@code VCVTUDQ2PD xmm1 {k1}{z}, xmm2/m64/m32bcst}<p>
	 * {@code EVEX.128.F3.0F.W0 7A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUDQ2PD_XMM_K1Z_XMMM64B32 = 1784;
	/**
	 * {@code VCVTUDQ2PD ymm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.256.F3.0F.W0 7A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUDQ2PD_YMM_K1Z_XMMM128B32 = 1785;
	/**
	 * {@code VCVTUDQ2PD zmm1 {k1}{z}, ymm2/m256/m32bcst{er}}<p>
	 * {@code EVEX.512.F3.0F.W0 7A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUDQ2PD_ZMM_K1Z_YMMM256B32_ER = 1786;
	/**
	 * {@code VCVTUQQ2PD xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.F3.0F.W1 7A /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUQQ2PD_XMM_K1Z_XMMM128B64 = 1787;
	/**
	 * {@code VCVTUQQ2PD ymm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.F3.0F.W1 7A /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUQQ2PD_YMM_K1Z_YMMM256B64 = 1788;
	/**
	 * {@code VCVTUQQ2PD zmm1 {k1}{z}, zmm2/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.F3.0F.W1 7A /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUQQ2PD_ZMM_K1Z_ZMMM512B64_ER = 1789;
	/**
	 * {@code VCVTUDQ2PS xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.F2.0F.W0 7A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUDQ2PS_XMM_K1Z_XMMM128B32 = 1790;
	/**
	 * {@code VCVTUDQ2PS ymm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.F2.0F.W0 7A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUDQ2PS_YMM_K1Z_YMMM256B32 = 1791;
	/**
	 * {@code VCVTUDQ2PS zmm1 {k1}{z}, zmm2/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.F2.0F.W0 7A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUDQ2PS_ZMM_K1Z_ZMMM512B32_ER = 1792;
	/**
	 * {@code VCVTUQQ2PS xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.F2.0F.W1 7A /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUQQ2PS_XMM_K1Z_XMMM128B64 = 1793;
	/**
	 * {@code VCVTUQQ2PS xmm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.F2.0F.W1 7A /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUQQ2PS_XMM_K1Z_YMMM256B64 = 1794;
	/**
	 * {@code VCVTUQQ2PS ymm1 {k1}{z}, zmm2/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.F2.0F.W1 7A /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUQQ2PS_YMM_K1Z_ZMMM512B64_ER = 1795;
	/**
	 * {@code VCVTPS2QQ xmm1 {k1}{z}, xmm2/m64/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 7B /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2QQ_XMM_K1Z_XMMM64B32 = 1796;
	/**
	 * {@code VCVTPS2QQ ymm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 7B /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2QQ_YMM_K1Z_XMMM128B32 = 1797;
	/**
	 * {@code VCVTPS2QQ zmm1 {k1}{z}, ymm2/m256/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F.W0 7B /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2QQ_ZMM_K1Z_YMMM256B32_ER = 1798;
	/**
	 * {@code VCVTPD2QQ xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 7B /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2QQ_XMM_K1Z_XMMM128B64 = 1799;
	/**
	 * {@code VCVTPD2QQ ymm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 7B /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2QQ_YMM_K1Z_YMMM256B64 = 1800;
	/**
	 * {@code VCVTPD2QQ zmm1 {k1}{z}, zmm2/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F.W1 7B /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2QQ_ZMM_K1Z_ZMMM512B64_ER = 1801;
	/**
	 * {@code VCVTUSI2SS xmm1, xmm2, r/m32{er}}<p>
	 * {@code EVEX.LIG.F3.0F.W0 7B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUSI2SS_XMM_XMM_RM32_ER = 1802;
	/**
	 * {@code VCVTUSI2SS xmm1, xmm2, r/m64{er}}<p>
	 * {@code EVEX.LIG.F3.0F.W1 7B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTUSI2SS_XMM_XMM_RM64_ER = 1803;
	/**
	 * {@code VCVTUSI2SD xmm1, xmm2, r/m32{er}}<p>
	 * {@code EVEX.LIG.F2.0F.W0 7B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUSI2SD_XMM_XMM_RM32_ER = 1804;
	/**
	 * {@code VCVTUSI2SD xmm1, xmm2, r/m64{er}}<p>
	 * {@code EVEX.LIG.F2.0F.W1 7B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTUSI2SD_XMM_XMM_RM64_ER = 1805;
	/**
	 * {@code HADDPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 7C /r}<p>
	 * {@code SSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int HADDPD_XMM_XMMM128 = 1806;
	/**
	 * {@code VHADDPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 7C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VHADDPD_XMM_XMM_XMMM128 = 1807;
	/**
	 * {@code VHADDPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 7C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VHADDPD_YMM_YMM_YMMM256 = 1808;
	/**
	 * {@code HADDPS xmm1, xmm2/m128}<p>
	 * {@code F2 0F 7C /r}<p>
	 * {@code SSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int HADDPS_XMM_XMMM128 = 1809;
	/**
	 * {@code VHADDPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.F2.0F.WIG 7C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VHADDPS_XMM_XMM_XMMM128 = 1810;
	/**
	 * {@code VHADDPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.F2.0F.WIG 7C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VHADDPS_YMM_YMM_YMMM256 = 1811;
	/**
	 * {@code HSUBPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 7D /r}<p>
	 * {@code SSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int HSUBPD_XMM_XMMM128 = 1812;
	/**
	 * {@code VHSUBPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG 7D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VHSUBPD_XMM_XMM_XMMM128 = 1813;
	/**
	 * {@code VHSUBPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG 7D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VHSUBPD_YMM_YMM_YMMM256 = 1814;
	/**
	 * {@code HSUBPS xmm1, xmm2/m128}<p>
	 * {@code F2 0F 7D /r}<p>
	 * {@code SSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int HSUBPS_XMM_XMMM128 = 1815;
	/**
	 * {@code VHSUBPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.F2.0F.WIG 7D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VHSUBPS_XMM_XMM_XMMM128 = 1816;
	/**
	 * {@code VHSUBPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.F2.0F.WIG 7D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VHSUBPS_YMM_YMM_YMMM256 = 1817;
	/**
	 * {@code MOVD r/m32, mm}<p>
	 * {@code NP 0F 7E /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVD_RM32_MM = 1818;
	/**
	 * {@code MOVQ r/m64, mm}<p>
	 * {@code NP o64 0F 7E /r}<p>
	 * {@code MMX}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVQ_RM64_MM = 1819;
	/**
	 * {@code MOVD r/m32, xmm}<p>
	 * {@code 66 0F 7E /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVD_RM32_XMM = 1820;
	/**
	 * {@code MOVQ r/m64, xmm}<p>
	 * {@code 66 o64 0F 7E /r}<p>
	 * {@code SSE2}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVQ_RM64_XMM = 1821;
	/**
	 * {@code VMOVD r/m32, xmm1}<p>
	 * {@code VEX.128.66.0F.W0 7E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVD_RM32_XMM = 1822;
	/**
	 * {@code VMOVQ r/m64, xmm1}<p>
	 * {@code VEX.128.66.0F.W1 7E /r}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VMOVQ_RM64_XMM = 1823;
	/**
	 * {@code VMOVD r/m32, xmm1}<p>
	 * {@code EVEX.128.66.0F.W0 7E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVD_RM32_XMM = 1824;
	/**
	 * {@code VMOVQ r/m64, xmm1}<p>
	 * {@code EVEX.128.66.0F.W1 7E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VMOVQ_RM64_XMM = 1825;
	/**
	 * {@code MOVQ xmm1, xmm2/m64}<p>
	 * {@code F3 0F 7E /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVQ_XMM_XMMM64 = 1826;
	/**
	 * {@code VMOVQ xmm1, xmm2/m64}<p>
	 * {@code VEX.128.F3.0F.WIG 7E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVQ_XMM_XMMM64 = 1827;
	/**
	 * {@code VMOVQ xmm1, xmm2/m64}<p>
	 * {@code EVEX.128.F3.0F.W1 7E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVQ_XMM_XMMM64 = 1828;
	/**
	 * {@code MOVQ mm/m64, mm}<p>
	 * {@code NP 0F 7F /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVQ_MMM64_MM = 1829;
	/**
	 * {@code MOVDQA xmm2/m128, xmm1}<p>
	 * {@code 66 0F 7F /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVDQA_XMMM128_XMM = 1830;
	/**
	 * {@code VMOVDQA xmm2/m128, xmm1}<p>
	 * {@code VEX.128.66.0F.WIG 7F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVDQA_XMMM128_XMM = 1831;
	/**
	 * {@code VMOVDQA ymm2/m256, ymm1}<p>
	 * {@code VEX.256.66.0F.WIG 7F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVDQA_YMMM256_YMM = 1832;
	/**
	 * {@code VMOVDQA32 xmm2/m128 {k1}{z}, xmm1}<p>
	 * {@code EVEX.128.66.0F.W0 7F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQA32_XMMM128_K1Z_XMM = 1833;
	/**
	 * {@code VMOVDQA32 ymm2/m256 {k1}{z}, ymm1}<p>
	 * {@code EVEX.256.66.0F.W0 7F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQA32_YMMM256_K1Z_YMM = 1834;
	/**
	 * {@code VMOVDQA32 zmm2/m512 {k1}{z}, zmm1}<p>
	 * {@code EVEX.512.66.0F.W0 7F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQA32_ZMMM512_K1Z_ZMM = 1835;
	/**
	 * {@code VMOVDQA64 xmm2/m128 {k1}{z}, xmm1}<p>
	 * {@code EVEX.128.66.0F.W1 7F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQA64_XMMM128_K1Z_XMM = 1836;
	/**
	 * {@code VMOVDQA64 ymm2/m256 {k1}{z}, ymm1}<p>
	 * {@code EVEX.256.66.0F.W1 7F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQA64_YMMM256_K1Z_YMM = 1837;
	/**
	 * {@code VMOVDQA64 zmm2/m512 {k1}{z}, zmm1}<p>
	 * {@code EVEX.512.66.0F.W1 7F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQA64_ZMMM512_K1Z_ZMM = 1838;
	/**
	 * {@code MOVDQU xmm2/m128, xmm1}<p>
	 * {@code F3 0F 7F /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVDQU_XMMM128_XMM = 1839;
	/**
	 * {@code VMOVDQU xmm2/m128, xmm1}<p>
	 * {@code VEX.128.F3.0F.WIG 7F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVDQU_XMMM128_XMM = 1840;
	/**
	 * {@code VMOVDQU ymm2/m256, ymm1}<p>
	 * {@code VEX.256.F3.0F.WIG 7F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVDQU_YMMM256_YMM = 1841;
	/**
	 * {@code VMOVDQU32 xmm2/m128 {k1}{z}, xmm1}<p>
	 * {@code EVEX.128.F3.0F.W0 7F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU32_XMMM128_K1Z_XMM = 1842;
	/**
	 * {@code VMOVDQU32 ymm2/m256 {k1}{z}, ymm1}<p>
	 * {@code EVEX.256.F3.0F.W0 7F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU32_YMMM256_K1Z_YMM = 1843;
	/**
	 * {@code VMOVDQU32 zmm2/m512 {k1}{z}, zmm1}<p>
	 * {@code EVEX.512.F3.0F.W0 7F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU32_ZMMM512_K1Z_ZMM = 1844;
	/**
	 * {@code VMOVDQU64 xmm2/m128 {k1}{z}, xmm1}<p>
	 * {@code EVEX.128.F3.0F.W1 7F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU64_XMMM128_K1Z_XMM = 1845;
	/**
	 * {@code VMOVDQU64 ymm2/m256 {k1}{z}, ymm1}<p>
	 * {@code EVEX.256.F3.0F.W1 7F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU64_YMMM256_K1Z_YMM = 1846;
	/**
	 * {@code VMOVDQU64 zmm2/m512 {k1}{z}, zmm1}<p>
	 * {@code EVEX.512.F3.0F.W1 7F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU64_ZMMM512_K1Z_ZMM = 1847;
	/**
	 * {@code VMOVDQU8 xmm2/m128 {k1}{z}, xmm1}<p>
	 * {@code EVEX.128.F2.0F.W0 7F /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU8_XMMM128_K1Z_XMM = 1848;
	/**
	 * {@code VMOVDQU8 ymm2/m256 {k1}{z}, ymm1}<p>
	 * {@code EVEX.256.F2.0F.W0 7F /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU8_YMMM256_K1Z_YMM = 1849;
	/**
	 * {@code VMOVDQU8 zmm2/m512 {k1}{z}, zmm1}<p>
	 * {@code EVEX.512.F2.0F.W0 7F /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU8_ZMMM512_K1Z_ZMM = 1850;
	/**
	 * {@code VMOVDQU16 xmm2/m128 {k1}{z}, xmm1}<p>
	 * {@code EVEX.128.F2.0F.W1 7F /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU16_XMMM128_K1Z_XMM = 1851;
	/**
	 * {@code VMOVDQU16 ymm2/m256 {k1}{z}, ymm1}<p>
	 * {@code EVEX.256.F2.0F.W1 7F /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU16_YMMM256_K1Z_YMM = 1852;
	/**
	 * {@code VMOVDQU16 zmm2/m512 {k1}{z}, zmm1}<p>
	 * {@code EVEX.512.F2.0F.W1 7F /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVDQU16_ZMMM512_K1Z_ZMM = 1853;
	/**
	 * {@code JO rel16}<p>
	 * {@code o16 0F 80 cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JO_REL16 = 1854;
	/**
	 * {@code JO rel32}<p>
	 * {@code o32 0F 80 cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JO_REL32_32 = 1855;
	/**
	 * {@code JO rel32}<p>
	 * {@code o64 0F 80 cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JO_REL32_64 = 1856;
	/**
	 * {@code JNO rel16}<p>
	 * {@code o16 0F 81 cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JNO_REL16 = 1857;
	/**
	 * {@code JNO rel32}<p>
	 * {@code o32 0F 81 cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JNO_REL32_32 = 1858;
	/**
	 * {@code JNO rel32}<p>
	 * {@code o64 0F 81 cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JNO_REL32_64 = 1859;
	/**
	 * {@code JB rel16}<p>
	 * {@code o16 0F 82 cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JB_REL16 = 1860;
	/**
	 * {@code JB rel32}<p>
	 * {@code o32 0F 82 cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JB_REL32_32 = 1861;
	/**
	 * {@code JB rel32}<p>
	 * {@code o64 0F 82 cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JB_REL32_64 = 1862;
	/**
	 * {@code JAE rel16}<p>
	 * {@code o16 0F 83 cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JAE_REL16 = 1863;
	/**
	 * {@code JAE rel32}<p>
	 * {@code o32 0F 83 cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JAE_REL32_32 = 1864;
	/**
	 * {@code JAE rel32}<p>
	 * {@code o64 0F 83 cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JAE_REL32_64 = 1865;
	/**
	 * {@code JE rel16}<p>
	 * {@code o16 0F 84 cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JE_REL16 = 1866;
	/**
	 * {@code JE rel32}<p>
	 * {@code o32 0F 84 cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JE_REL32_32 = 1867;
	/**
	 * {@code JE rel32}<p>
	 * {@code o64 0F 84 cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JE_REL32_64 = 1868;
	/**
	 * {@code JNE rel16}<p>
	 * {@code o16 0F 85 cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JNE_REL16 = 1869;
	/**
	 * {@code JNE rel32}<p>
	 * {@code o32 0F 85 cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JNE_REL32_32 = 1870;
	/**
	 * {@code JNE rel32}<p>
	 * {@code o64 0F 85 cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JNE_REL32_64 = 1871;
	/**
	 * {@code JBE rel16}<p>
	 * {@code o16 0F 86 cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JBE_REL16 = 1872;
	/**
	 * {@code JBE rel32}<p>
	 * {@code o32 0F 86 cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JBE_REL32_32 = 1873;
	/**
	 * {@code JBE rel32}<p>
	 * {@code o64 0F 86 cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JBE_REL32_64 = 1874;
	/**
	 * {@code JA rel16}<p>
	 * {@code o16 0F 87 cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JA_REL16 = 1875;
	/**
	 * {@code JA rel32}<p>
	 * {@code o32 0F 87 cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JA_REL32_32 = 1876;
	/**
	 * {@code JA rel32}<p>
	 * {@code o64 0F 87 cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JA_REL32_64 = 1877;
	/**
	 * {@code JS rel16}<p>
	 * {@code o16 0F 88 cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JS_REL16 = 1878;
	/**
	 * {@code JS rel32}<p>
	 * {@code o32 0F 88 cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JS_REL32_32 = 1879;
	/**
	 * {@code JS rel32}<p>
	 * {@code o64 0F 88 cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JS_REL32_64 = 1880;
	/**
	 * {@code JNS rel16}<p>
	 * {@code o16 0F 89 cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JNS_REL16 = 1881;
	/**
	 * {@code JNS rel32}<p>
	 * {@code o32 0F 89 cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JNS_REL32_32 = 1882;
	/**
	 * {@code JNS rel32}<p>
	 * {@code o64 0F 89 cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JNS_REL32_64 = 1883;
	/**
	 * {@code JP rel16}<p>
	 * {@code o16 0F 8A cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JP_REL16 = 1884;
	/**
	 * {@code JP rel32}<p>
	 * {@code o32 0F 8A cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JP_REL32_32 = 1885;
	/**
	 * {@code JP rel32}<p>
	 * {@code o64 0F 8A cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JP_REL32_64 = 1886;
	/**
	 * {@code JNP rel16}<p>
	 * {@code o16 0F 8B cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JNP_REL16 = 1887;
	/**
	 * {@code JNP rel32}<p>
	 * {@code o32 0F 8B cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JNP_REL32_32 = 1888;
	/**
	 * {@code JNP rel32}<p>
	 * {@code o64 0F 8B cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JNP_REL32_64 = 1889;
	/**
	 * {@code JL rel16}<p>
	 * {@code o16 0F 8C cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JL_REL16 = 1890;
	/**
	 * {@code JL rel32}<p>
	 * {@code o32 0F 8C cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JL_REL32_32 = 1891;
	/**
	 * {@code JL rel32}<p>
	 * {@code o64 0F 8C cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JL_REL32_64 = 1892;
	/**
	 * {@code JGE rel16}<p>
	 * {@code o16 0F 8D cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JGE_REL16 = 1893;
	/**
	 * {@code JGE rel32}<p>
	 * {@code o32 0F 8D cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JGE_REL32_32 = 1894;
	/**
	 * {@code JGE rel32}<p>
	 * {@code o64 0F 8D cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JGE_REL32_64 = 1895;
	/**
	 * {@code JLE rel16}<p>
	 * {@code o16 0F 8E cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JLE_REL16 = 1896;
	/**
	 * {@code JLE rel32}<p>
	 * {@code o32 0F 8E cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JLE_REL32_32 = 1897;
	/**
	 * {@code JLE rel32}<p>
	 * {@code o64 0F 8E cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JLE_REL32_64 = 1898;
	/**
	 * {@code JG rel16}<p>
	 * {@code o16 0F 8F cw}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int JG_REL16 = 1899;
	/**
	 * {@code JG rel32}<p>
	 * {@code o32 0F 8F cd}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JG_REL32_32 = 1900;
	/**
	 * {@code JG rel32}<p>
	 * {@code o64 0F 8F cd}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int JG_REL32_64 = 1901;
	/**
	 * {@code SETO r/m8}<p>
	 * {@code 0F 90 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETO_RM8 = 1902;
	/**
	 * {@code SETNO r/m8}<p>
	 * {@code 0F 91 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETNO_RM8 = 1903;
	/**
	 * {@code SETB r/m8}<p>
	 * {@code 0F 92 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETB_RM8 = 1904;
	/**
	 * {@code SETAE r/m8}<p>
	 * {@code 0F 93 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETAE_RM8 = 1905;
	/**
	 * {@code SETE r/m8}<p>
	 * {@code 0F 94 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETE_RM8 = 1906;
	/**
	 * {@code SETNE r/m8}<p>
	 * {@code 0F 95 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETNE_RM8 = 1907;
	/**
	 * {@code SETBE r/m8}<p>
	 * {@code 0F 96 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETBE_RM8 = 1908;
	/**
	 * {@code SETA r/m8}<p>
	 * {@code 0F 97 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETA_RM8 = 1909;
	/**
	 * {@code SETS r/m8}<p>
	 * {@code 0F 98 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETS_RM8 = 1910;
	/**
	 * {@code SETNS r/m8}<p>
	 * {@code 0F 99 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETNS_RM8 = 1911;
	/**
	 * {@code SETP r/m8}<p>
	 * {@code 0F 9A /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETP_RM8 = 1912;
	/**
	 * {@code SETNP r/m8}<p>
	 * {@code 0F 9B /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETNP_RM8 = 1913;
	/**
	 * {@code SETL r/m8}<p>
	 * {@code 0F 9C /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETL_RM8 = 1914;
	/**
	 * {@code SETGE r/m8}<p>
	 * {@code 0F 9D /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETGE_RM8 = 1915;
	/**
	 * {@code SETLE r/m8}<p>
	 * {@code 0F 9E /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETLE_RM8 = 1916;
	/**
	 * {@code SETG r/m8}<p>
	 * {@code 0F 9F /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SETG_RM8 = 1917;
	/**
	 * {@code KMOVW k1, k2/m16}<p>
	 * {@code VEX.L0.0F.W0 90 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KMOVW_KR_KM16 = 1918;
	/**
	 * {@code KMOVQ k1, k2/m64}<p>
	 * {@code VEX.L0.0F.W1 90 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KMOVQ_KR_KM64 = 1919;
	/**
	 * {@code KMOVB k1, k2/m8}<p>
	 * {@code VEX.L0.66.0F.W0 90 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KMOVB_KR_KM8 = 1920;
	/**
	 * {@code KMOVD k1, k2/m32}<p>
	 * {@code VEX.L0.66.0F.W1 90 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KMOVD_KR_KM32 = 1921;
	/**
	 * {@code KMOVW m16, k1}<p>
	 * {@code VEX.L0.0F.W0 91 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KMOVW_M16_KR = 1922;
	/**
	 * {@code KMOVQ m64, k1}<p>
	 * {@code VEX.L0.0F.W1 91 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KMOVQ_M64_KR = 1923;
	/**
	 * {@code KMOVB m8, k1}<p>
	 * {@code VEX.L0.66.0F.W0 91 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KMOVB_M8_KR = 1924;
	/**
	 * {@code KMOVD m32, k1}<p>
	 * {@code VEX.L0.66.0F.W1 91 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KMOVD_M32_KR = 1925;
	/**
	 * {@code KMOVW k1, r32}<p>
	 * {@code VEX.L0.0F.W0 92 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KMOVW_KR_R32 = 1926;
	/**
	 * {@code KMOVB k1, r32}<p>
	 * {@code VEX.L0.66.0F.W0 92 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KMOVB_KR_R32 = 1927;
	/**
	 * {@code KMOVD k1, r32}<p>
	 * {@code VEX.L0.F2.0F.W0 92 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KMOVD_KR_R32 = 1928;
	/**
	 * {@code KMOVQ k1, r64}<p>
	 * {@code VEX.L0.F2.0F.W1 92 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KMOVQ_KR_R64 = 1929;
	/**
	 * {@code KMOVW r32, k1}<p>
	 * {@code VEX.L0.0F.W0 93 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KMOVW_R32_KR = 1930;
	/**
	 * {@code KMOVB r32, k1}<p>
	 * {@code VEX.L0.66.0F.W0 93 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KMOVB_R32_KR = 1931;
	/**
	 * {@code KMOVD r32, k1}<p>
	 * {@code VEX.L0.F2.0F.W0 93 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KMOVD_R32_KR = 1932;
	/**
	 * {@code KMOVQ r64, k1}<p>
	 * {@code VEX.L0.F2.0F.W1 93 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KMOVQ_R64_KR = 1933;
	/**
	 * {@code KORTESTW k1, k2}<p>
	 * {@code VEX.L0.0F.W0 98 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KORTESTW_KR_KR = 1934;
	/**
	 * {@code KORTESTQ k1, k2}<p>
	 * {@code VEX.L0.0F.W1 98 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KORTESTQ_KR_KR = 1935;
	/**
	 * {@code KORTESTB k1, k2}<p>
	 * {@code VEX.L0.66.0F.W0 98 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KORTESTB_KR_KR = 1936;
	/**
	 * {@code KORTESTD k1, k2}<p>
	 * {@code VEX.L0.66.0F.W1 98 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KORTESTD_KR_KR = 1937;
	/**
	 * {@code KTESTW k1, k2}<p>
	 * {@code VEX.L0.0F.W0 99 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KTESTW_KR_KR = 1938;
	/**
	 * {@code KTESTQ k1, k2}<p>
	 * {@code VEX.L0.0F.W1 99 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KTESTQ_KR_KR = 1939;
	/**
	 * {@code KTESTB k1, k2}<p>
	 * {@code VEX.L0.66.0F.W0 99 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KTESTB_KR_KR = 1940;
	/**
	 * {@code KTESTD k1, k2}<p>
	 * {@code VEX.L0.66.0F.W1 99 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KTESTD_KR_KR = 1941;
	/**
	 * {@code PUSH FS}<p>
	 * {@code o16 0F A0}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUSHW_FS = 1942;
	/**
	 * {@code PUSH FS}<p>
	 * {@code o32 0F A0}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSHD_FS = 1943;
	/**
	 * {@code PUSH FS}<p>
	 * {@code o64 0F A0}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int PUSHQ_FS = 1944;
	/**
	 * {@code POP FS}<p>
	 * {@code o16 0F A1}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int POPW_FS = 1945;
	/**
	 * {@code POP FS}<p>
	 * {@code o32 0F A1}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int POPD_FS = 1946;
	/**
	 * {@code POP FS}<p>
	 * {@code o64 0F A1}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int POPQ_FS = 1947;
	/**
	 * {@code CPUID}<p>
	 * {@code 0F A2}<p>
	 * {@code CPUID}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CPUID = 1948;
	/**
	 * {@code BT r/m16, r16}<p>
	 * {@code o16 0F A3 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BT_RM16_R16 = 1949;
	/**
	 * {@code BT r/m32, r32}<p>
	 * {@code o32 0F A3 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BT_RM32_R32 = 1950;
	/**
	 * {@code BT r/m64, r64}<p>
	 * {@code o64 0F A3 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int BT_RM64_R64 = 1951;
	/**
	 * {@code SHLD r/m16, r16, imm8}<p>
	 * {@code o16 0F A4 /r ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHLD_RM16_R16_IMM8 = 1952;
	/**
	 * {@code SHLD r/m32, r32, imm8}<p>
	 * {@code o32 0F A4 /r ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHLD_RM32_R32_IMM8 = 1953;
	/**
	 * {@code SHLD r/m64, r64, imm8}<p>
	 * {@code o64 0F A4 /r ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SHLD_RM64_R64_IMM8 = 1954;
	/**
	 * {@code SHLD r/m16, r16, CL}<p>
	 * {@code o16 0F A5 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHLD_RM16_R16_CL = 1955;
	/**
	 * {@code SHLD r/m32, r32, CL}<p>
	 * {@code o32 0F A5 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHLD_RM32_R32_CL = 1956;
	/**
	 * {@code SHLD r/m64, r64, CL}<p>
	 * {@code o64 0F A5 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SHLD_RM64_R64_CL = 1957;
	/**
	 * {@code MONTMUL}<p>
	 * {@code a16 F3 0F A6 C0}<p>
	 * {@code PADLOCK_PMM}<p>
	 * {@code 16/32-bit}
	 */
	public static final int MONTMUL_16 = 1958;
	/**
	 * {@code MONTMUL}<p>
	 * {@code a32 F3 0F A6 C0}<p>
	 * {@code PADLOCK_PMM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MONTMUL_32 = 1959;
	/**
	 * {@code MONTMUL}<p>
	 * {@code a64 F3 0F A6 C0}<p>
	 * {@code PADLOCK_PMM}<p>
	 * {@code 64-bit}
	 */
	public static final int MONTMUL_64 = 1960;
	/**
	 * {@code XSHA1}<p>
	 * {@code a16 F3 0F A6 C8}<p>
	 * {@code PADLOCK_PHE}<p>
	 * {@code 16/32-bit}
	 */
	public static final int XSHA1_16 = 1961;
	/**
	 * {@code XSHA1}<p>
	 * {@code a32 F3 0F A6 C8}<p>
	 * {@code PADLOCK_PHE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XSHA1_32 = 1962;
	/**
	 * {@code XSHA1}<p>
	 * {@code a64 F3 0F A6 C8}<p>
	 * {@code PADLOCK_PHE}<p>
	 * {@code 64-bit}
	 */
	public static final int XSHA1_64 = 1963;
	/**
	 * {@code XSHA256}<p>
	 * {@code a16 F3 0F A6 D0}<p>
	 * {@code PADLOCK_PHE}<p>
	 * {@code 16/32-bit}
	 */
	public static final int XSHA256_16 = 1964;
	/**
	 * {@code XSHA256}<p>
	 * {@code a32 F3 0F A6 D0}<p>
	 * {@code PADLOCK_PHE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XSHA256_32 = 1965;
	/**
	 * {@code XSHA256}<p>
	 * {@code a64 F3 0F A6 D0}<p>
	 * {@code PADLOCK_PHE}<p>
	 * {@code 64-bit}
	 */
	public static final int XSHA256_64 = 1966;
	/**
	 * {@code XBTS r16, r/m16}<p>
	 * {@code o16 0F A6 /r}<p>
	 * {@code 386 A0}<p>
	 * {@code 16/32-bit}
	 */
	public static final int XBTS_R16_RM16 = 1967;
	/**
	 * {@code XBTS r32, r/m32}<p>
	 * {@code o32 0F A6 /r}<p>
	 * {@code 386 A0}<p>
	 * {@code 16/32-bit}
	 */
	public static final int XBTS_R32_RM32 = 1968;
	/**
	 * {@code XSTORE}<p>
	 * {@code a16 0F A7 C0}<p>
	 * {@code PADLOCK_RNG}<p>
	 * {@code 16/32-bit}
	 */
	public static final int XSTORE_16 = 1969;
	/**
	 * {@code XSTORE}<p>
	 * {@code a32 0F A7 C0}<p>
	 * {@code PADLOCK_RNG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XSTORE_32 = 1970;
	/**
	 * {@code XSTORE}<p>
	 * {@code a64 0F A7 C0}<p>
	 * {@code PADLOCK_RNG}<p>
	 * {@code 64-bit}
	 */
	public static final int XSTORE_64 = 1971;
	/**
	 * {@code XCRYPTECB}<p>
	 * {@code a16 F3 0F A7 C8}<p>
	 * {@code PADLOCK_ACE}<p>
	 * {@code 16/32-bit}
	 */
	public static final int XCRYPTECB_16 = 1972;
	/**
	 * {@code XCRYPTECB}<p>
	 * {@code a32 F3 0F A7 C8}<p>
	 * {@code PADLOCK_ACE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XCRYPTECB_32 = 1973;
	/**
	 * {@code XCRYPTECB}<p>
	 * {@code a64 F3 0F A7 C8}<p>
	 * {@code PADLOCK_ACE}<p>
	 * {@code 64-bit}
	 */
	public static final int XCRYPTECB_64 = 1974;
	/**
	 * {@code XCRYPTCBC}<p>
	 * {@code a16 F3 0F A7 D0}<p>
	 * {@code PADLOCK_ACE}<p>
	 * {@code 16/32-bit}
	 */
	public static final int XCRYPTCBC_16 = 1975;
	/**
	 * {@code XCRYPTCBC}<p>
	 * {@code a32 F3 0F A7 D0}<p>
	 * {@code PADLOCK_ACE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XCRYPTCBC_32 = 1976;
	/**
	 * {@code XCRYPTCBC}<p>
	 * {@code a64 F3 0F A7 D0}<p>
	 * {@code PADLOCK_ACE}<p>
	 * {@code 64-bit}
	 */
	public static final int XCRYPTCBC_64 = 1977;
	/**
	 * {@code XCRYPTCTR}<p>
	 * {@code a16 F3 0F A7 D8}<p>
	 * {@code PADLOCK_ACE}<p>
	 * {@code 16/32-bit}
	 */
	public static final int XCRYPTCTR_16 = 1978;
	/**
	 * {@code XCRYPTCTR}<p>
	 * {@code a32 F3 0F A7 D8}<p>
	 * {@code PADLOCK_ACE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XCRYPTCTR_32 = 1979;
	/**
	 * {@code XCRYPTCTR}<p>
	 * {@code a64 F3 0F A7 D8}<p>
	 * {@code PADLOCK_ACE}<p>
	 * {@code 64-bit}
	 */
	public static final int XCRYPTCTR_64 = 1980;
	/**
	 * {@code XCRYPTCFB}<p>
	 * {@code a16 F3 0F A7 E0}<p>
	 * {@code PADLOCK_ACE}<p>
	 * {@code 16/32-bit}
	 */
	public static final int XCRYPTCFB_16 = 1981;
	/**
	 * {@code XCRYPTCFB}<p>
	 * {@code a32 F3 0F A7 E0}<p>
	 * {@code PADLOCK_ACE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XCRYPTCFB_32 = 1982;
	/**
	 * {@code XCRYPTCFB}<p>
	 * {@code a64 F3 0F A7 E0}<p>
	 * {@code PADLOCK_ACE}<p>
	 * {@code 64-bit}
	 */
	public static final int XCRYPTCFB_64 = 1983;
	/**
	 * {@code XCRYPTOFB}<p>
	 * {@code a16 F3 0F A7 E8}<p>
	 * {@code PADLOCK_ACE}<p>
	 * {@code 16/32-bit}
	 */
	public static final int XCRYPTOFB_16 = 1984;
	/**
	 * {@code XCRYPTOFB}<p>
	 * {@code a32 F3 0F A7 E8}<p>
	 * {@code PADLOCK_ACE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XCRYPTOFB_32 = 1985;
	/**
	 * {@code XCRYPTOFB}<p>
	 * {@code a64 F3 0F A7 E8}<p>
	 * {@code PADLOCK_ACE}<p>
	 * {@code 64-bit}
	 */
	public static final int XCRYPTOFB_64 = 1986;
	/**
	 * {@code IBTS r/m16, r16}<p>
	 * {@code o16 0F A7 /r}<p>
	 * {@code 386 A0}<p>
	 * {@code 16/32-bit}
	 */
	public static final int IBTS_RM16_R16 = 1987;
	/**
	 * {@code IBTS r/m32, r32}<p>
	 * {@code o32 0F A7 /r}<p>
	 * {@code 386 A0}<p>
	 * {@code 16/32-bit}
	 */
	public static final int IBTS_RM32_R32 = 1988;
	/**
	 * {@code CMPXCHG r/m8, r8}<p>
	 * {@code 0F A6 /r}<p>
	 * {@code 486 A}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CMPXCHG486_RM8_R8 = 1989;
	/**
	 * {@code CMPXCHG r/m16, r16}<p>
	 * {@code o16 0F A7 /r}<p>
	 * {@code 486 A}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CMPXCHG486_RM16_R16 = 1990;
	/**
	 * {@code CMPXCHG r/m32, r32}<p>
	 * {@code o32 0F A7 /r}<p>
	 * {@code 486 A}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CMPXCHG486_RM32_R32 = 1991;
	/**
	 * {@code PUSH GS}<p>
	 * {@code o16 0F A8}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PUSHW_GS = 1992;
	/**
	 * {@code PUSH GS}<p>
	 * {@code o32 0F A8}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PUSHD_GS = 1993;
	/**
	 * {@code PUSH GS}<p>
	 * {@code o64 0F A8}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int PUSHQ_GS = 1994;
	/**
	 * {@code POP GS}<p>
	 * {@code o16 0F A9}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int POPW_GS = 1995;
	/**
	 * {@code POP GS}<p>
	 * {@code o32 0F A9}<p>
	 * {@code 386+}<p>
	 * {@code 16/32-bit}
	 */
	public static final int POPD_GS = 1996;
	/**
	 * {@code POP GS}<p>
	 * {@code o64 0F A9}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int POPQ_GS = 1997;
	/**
	 * {@code RSM}<p>
	 * {@code 0F AA}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RSM = 1998;
	/**
	 * {@code BTS r/m16, r16}<p>
	 * {@code o16 0F AB /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BTS_RM16_R16 = 1999;
	/**
	 * {@code BTS r/m32, r32}<p>
	 * {@code o32 0F AB /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BTS_RM32_R32 = 2000;
	/**
	 * {@code BTS r/m64, r64}<p>
	 * {@code o64 0F AB /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int BTS_RM64_R64 = 2001;
	/**
	 * {@code SHRD r/m16, r16, imm8}<p>
	 * {@code o16 0F AC /r ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHRD_RM16_R16_IMM8 = 2002;
	/**
	 * {@code SHRD r/m32, r32, imm8}<p>
	 * {@code o32 0F AC /r ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHRD_RM32_R32_IMM8 = 2003;
	/**
	 * {@code SHRD r/m64, r64, imm8}<p>
	 * {@code o64 0F AC /r ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SHRD_RM64_R64_IMM8 = 2004;
	/**
	 * {@code SHRD r/m16, r16, CL}<p>
	 * {@code o16 0F AD /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHRD_RM16_R16_CL = 2005;
	/**
	 * {@code SHRD r/m32, r32, CL}<p>
	 * {@code o32 0F AD /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHRD_RM32_R32_CL = 2006;
	/**
	 * {@code SHRD r/m64, r64, CL}<p>
	 * {@code o64 0F AD /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int SHRD_RM64_R64_CL = 2007;
	/**
	 * {@code FXSAVE m512byte}<p>
	 * {@code NP 0F AE /0}<p>
	 * {@code FXSR}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FXSAVE_M512BYTE = 2008;
	/**
	 * {@code FXSAVE64 m512byte}<p>
	 * {@code NP o64 0F AE /0}<p>
	 * {@code FXSR}<p>
	 * {@code 64-bit}
	 */
	public static final int FXSAVE64_M512BYTE = 2009;
	/**
	 * {@code RDFSBASE r32}<p>
	 * {@code F3 0F AE /0}<p>
	 * {@code FSGSBASE}<p>
	 * {@code 64-bit}
	 */
	public static final int RDFSBASE_R32 = 2010;
	/**
	 * {@code RDFSBASE r64}<p>
	 * {@code F3 o64 0F AE /0}<p>
	 * {@code FSGSBASE}<p>
	 * {@code 64-bit}
	 */
	public static final int RDFSBASE_R64 = 2011;
	/**
	 * {@code FXRSTOR m512byte}<p>
	 * {@code NP 0F AE /1}<p>
	 * {@code FXSR}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int FXRSTOR_M512BYTE = 2012;
	/**
	 * {@code FXRSTOR64 m512byte}<p>
	 * {@code NP o64 0F AE /1}<p>
	 * {@code FXSR}<p>
	 * {@code 64-bit}
	 */
	public static final int FXRSTOR64_M512BYTE = 2013;
	/**
	 * {@code RDGSBASE r32}<p>
	 * {@code F3 0F AE /1}<p>
	 * {@code FSGSBASE}<p>
	 * {@code 64-bit}
	 */
	public static final int RDGSBASE_R32 = 2014;
	/**
	 * {@code RDGSBASE r64}<p>
	 * {@code F3 o64 0F AE /1}<p>
	 * {@code FSGSBASE}<p>
	 * {@code 64-bit}
	 */
	public static final int RDGSBASE_R64 = 2015;
	/**
	 * {@code LDMXCSR m32}<p>
	 * {@code NP 0F AE /2}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LDMXCSR_M32 = 2016;
	/**
	 * {@code WRFSBASE r32}<p>
	 * {@code F3 0F AE /2}<p>
	 * {@code FSGSBASE}<p>
	 * {@code 64-bit}
	 */
	public static final int WRFSBASE_R32 = 2017;
	/**
	 * {@code WRFSBASE r64}<p>
	 * {@code F3 o64 0F AE /2}<p>
	 * {@code FSGSBASE}<p>
	 * {@code 64-bit}
	 */
	public static final int WRFSBASE_R64 = 2018;
	/**
	 * {@code VLDMXCSR m32}<p>
	 * {@code VEX.LZ.0F.WIG AE /2}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VLDMXCSR_M32 = 2019;
	/**
	 * {@code STMXCSR m32}<p>
	 * {@code NP 0F AE /3}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int STMXCSR_M32 = 2020;
	/**
	 * {@code WRGSBASE r32}<p>
	 * {@code F3 0F AE /3}<p>
	 * {@code FSGSBASE}<p>
	 * {@code 64-bit}
	 */
	public static final int WRGSBASE_R32 = 2021;
	/**
	 * {@code WRGSBASE r64}<p>
	 * {@code F3 o64 0F AE /3}<p>
	 * {@code FSGSBASE}<p>
	 * {@code 64-bit}
	 */
	public static final int WRGSBASE_R64 = 2022;
	/**
	 * {@code VSTMXCSR m32}<p>
	 * {@code VEX.LZ.0F.WIG AE /3}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSTMXCSR_M32 = 2023;
	/**
	 * {@code XSAVE mem}<p>
	 * {@code NP 0F AE /4}<p>
	 * {@code XSAVE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XSAVE_MEM = 2024;
	/**
	 * {@code XSAVE64 mem}<p>
	 * {@code NP o64 0F AE /4}<p>
	 * {@code XSAVE}<p>
	 * {@code 64-bit}
	 */
	public static final int XSAVE64_MEM = 2025;
	/**
	 * {@code PTWRITE r/m32}<p>
	 * {@code F3 0F AE /4}<p>
	 * {@code PTWRITE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PTWRITE_RM32 = 2026;
	/**
	 * {@code PTWRITE r/m64}<p>
	 * {@code F3 o64 0F AE /4}<p>
	 * {@code PTWRITE}<p>
	 * {@code 64-bit}
	 */
	public static final int PTWRITE_RM64 = 2027;
	/**
	 * {@code XRSTOR mem}<p>
	 * {@code NP 0F AE /5}<p>
	 * {@code XSAVE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XRSTOR_MEM = 2028;
	/**
	 * {@code XRSTOR64 mem}<p>
	 * {@code NP o64 0F AE /5}<p>
	 * {@code XSAVE}<p>
	 * {@code 64-bit}
	 */
	public static final int XRSTOR64_MEM = 2029;
	/**
	 * {@code INCSSPD r32}<p>
	 * {@code F3 0F AE /5}<p>
	 * {@code CET_SS}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INCSSPD_R32 = 2030;
	/**
	 * {@code INCSSPQ r64}<p>
	 * {@code F3 o64 0F AE /5}<p>
	 * {@code CET_SS}<p>
	 * {@code 64-bit}
	 */
	public static final int INCSSPQ_R64 = 2031;
	/**
	 * {@code XSAVEOPT mem}<p>
	 * {@code NP 0F AE /6}<p>
	 * {@code XSAVEOPT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XSAVEOPT_MEM = 2032;
	/**
	 * {@code XSAVEOPT64 mem}<p>
	 * {@code NP o64 0F AE /6}<p>
	 * {@code XSAVEOPT}<p>
	 * {@code 64-bit}
	 */
	public static final int XSAVEOPT64_MEM = 2033;
	/**
	 * {@code CLWB m8}<p>
	 * {@code 66 0F AE /6}<p>
	 * {@code CLWB}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CLWB_M8 = 2034;
	/**
	 * {@code TPAUSE r32, &lt;edx&gt;, &lt;eax&gt;}<p>
	 * {@code 66 0F AE /6}<p>
	 * {@code WAITPKG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TPAUSE_R32 = 2035;
	/**
	 * {@code TPAUSE r64, &lt;edx&gt;, &lt;eax&gt;}<p>
	 * {@code 66 o64 0F AE /6}<p>
	 * {@code WAITPKG}<p>
	 * {@code 64-bit}
	 */
	public static final int TPAUSE_R64 = 2036;
	/**
	 * {@code CLRSSBSY m64}<p>
	 * {@code F3 0F AE /6}<p>
	 * {@code CET_SS}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CLRSSBSY_M64 = 2037;
	/**
	 * {@code UMONITOR r16}<p>
	 * {@code a16 F3 0F AE /6}<p>
	 * {@code WAITPKG}<p>
	 * {@code 16/32-bit}
	 */
	public static final int UMONITOR_R16 = 2038;
	/**
	 * {@code UMONITOR r32}<p>
	 * {@code a32 F3 0F AE /6}<p>
	 * {@code WAITPKG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int UMONITOR_R32 = 2039;
	/**
	 * {@code UMONITOR r64}<p>
	 * {@code a64 F3 0F AE /6}<p>
	 * {@code WAITPKG}<p>
	 * {@code 64-bit}
	 */
	public static final int UMONITOR_R64 = 2040;
	/**
	 * {@code UMWAIT r32, &lt;edx&gt;, &lt;eax&gt;}<p>
	 * {@code F2 0F AE /6}<p>
	 * {@code WAITPKG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int UMWAIT_R32 = 2041;
	/**
	 * {@code UMWAIT r64, &lt;edx&gt;, &lt;eax&gt;}<p>
	 * {@code F2 o64 0F AE /6}<p>
	 * {@code WAITPKG}<p>
	 * {@code 64-bit}
	 */
	public static final int UMWAIT_R64 = 2042;
	/**
	 * {@code CLFLUSH m8}<p>
	 * {@code NP 0F AE /7}<p>
	 * {@code CLFSH}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CLFLUSH_M8 = 2043;
	/**
	 * {@code CLFLUSHOPT m8}<p>
	 * {@code 66 0F AE /7}<p>
	 * {@code CLFLUSHOPT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CLFLUSHOPT_M8 = 2044;
	/**
	 * {@code LFENCE}<p>
	 * {@code NP 0F AE E8}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LFENCE = 2045;
	/**
	 * {@code LFENCE}<p>
	 * {@code NP 0F AE E9}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LFENCE_E9 = 2046;
	/**
	 * {@code LFENCE}<p>
	 * {@code NP 0F AE EA}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LFENCE_EA = 2047;
	/**
	 * {@code LFENCE}<p>
	 * {@code NP 0F AE EB}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LFENCE_EB = 2048;
	/**
	 * {@code LFENCE}<p>
	 * {@code NP 0F AE EC}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LFENCE_EC = 2049;
	/**
	 * {@code LFENCE}<p>
	 * {@code NP 0F AE ED}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LFENCE_ED = 2050;
	/**
	 * {@code LFENCE}<p>
	 * {@code NP 0F AE EE}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LFENCE_EE = 2051;
	/**
	 * {@code LFENCE}<p>
	 * {@code NP 0F AE EF}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LFENCE_EF = 2052;
	/**
	 * {@code MFENCE}<p>
	 * {@code NP 0F AE F0}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MFENCE = 2053;
	/**
	 * {@code MFENCE}<p>
	 * {@code NP 0F AE F1}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MFENCE_F1 = 2054;
	/**
	 * {@code MFENCE}<p>
	 * {@code NP 0F AE F2}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MFENCE_F2 = 2055;
	/**
	 * {@code MFENCE}<p>
	 * {@code NP 0F AE F3}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MFENCE_F3 = 2056;
	/**
	 * {@code MFENCE}<p>
	 * {@code NP 0F AE F4}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MFENCE_F4 = 2057;
	/**
	 * {@code MFENCE}<p>
	 * {@code NP 0F AE F5}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MFENCE_F5 = 2058;
	/**
	 * {@code MFENCE}<p>
	 * {@code NP 0F AE F6}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MFENCE_F6 = 2059;
	/**
	 * {@code MFENCE}<p>
	 * {@code NP 0F AE F7}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MFENCE_F7 = 2060;
	/**
	 * {@code SFENCE}<p>
	 * {@code NP 0F AE F8}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SFENCE = 2061;
	/**
	 * {@code SFENCE}<p>
	 * {@code NP 0F AE F9}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SFENCE_F9 = 2062;
	/**
	 * {@code SFENCE}<p>
	 * {@code NP 0F AE FA}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SFENCE_FA = 2063;
	/**
	 * {@code SFENCE}<p>
	 * {@code NP 0F AE FB}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SFENCE_FB = 2064;
	/**
	 * {@code SFENCE}<p>
	 * {@code NP 0F AE FC}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SFENCE_FC = 2065;
	/**
	 * {@code SFENCE}<p>
	 * {@code NP 0F AE FD}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SFENCE_FD = 2066;
	/**
	 * {@code SFENCE}<p>
	 * {@code NP 0F AE FE}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SFENCE_FE = 2067;
	/**
	 * {@code SFENCE}<p>
	 * {@code NP 0F AE FF}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SFENCE_FF = 2068;
	/**
	 * {@code PCOMMIT}<p>
	 * {@code 66 0F AE F8}<p>
	 * {@code PCOMMIT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCOMMIT = 2069;
	/**
	 * {@code IMUL r16, r/m16}<p>
	 * {@code o16 0F AF /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IMUL_R16_RM16 = 2070;
	/**
	 * {@code IMUL r32, r/m32}<p>
	 * {@code o32 0F AF /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int IMUL_R32_RM32 = 2071;
	/**
	 * {@code IMUL r64, r/m64}<p>
	 * {@code o64 0F AF /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int IMUL_R64_RM64 = 2072;
	/**
	 * {@code CMPXCHG r/m8, r8}<p>
	 * {@code 0F B0 /r}<p>
	 * {@code 486+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMPXCHG_RM8_R8 = 2073;
	/**
	 * {@code CMPXCHG r/m16, r16}<p>
	 * {@code o16 0F B1 /r}<p>
	 * {@code 486+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMPXCHG_RM16_R16 = 2074;
	/**
	 * {@code CMPXCHG r/m32, r32}<p>
	 * {@code o32 0F B1 /r}<p>
	 * {@code 486+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMPXCHG_RM32_R32 = 2075;
	/**
	 * {@code CMPXCHG r/m64, r64}<p>
	 * {@code o64 0F B1 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int CMPXCHG_RM64_R64 = 2076;
	/**
	 * {@code LSS r16, m16:16}<p>
	 * {@code o16 0F B2 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LSS_R16_M1616 = 2077;
	/**
	 * {@code LSS r32, m16:32}<p>
	 * {@code o32 0F B2 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LSS_R32_M1632 = 2078;
	/**
	 * {@code LSS r64, m16:64}<p>
	 * {@code o64 0F B2 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LSS_R64_M1664 = 2079;
	/**
	 * {@code BTR r/m16, r16}<p>
	 * {@code o16 0F B3 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BTR_RM16_R16 = 2080;
	/**
	 * {@code BTR r/m32, r32}<p>
	 * {@code o32 0F B3 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BTR_RM32_R32 = 2081;
	/**
	 * {@code BTR r/m64, r64}<p>
	 * {@code o64 0F B3 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int BTR_RM64_R64 = 2082;
	/**
	 * {@code LFS r16, m16:16}<p>
	 * {@code o16 0F B4 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LFS_R16_M1616 = 2083;
	/**
	 * {@code LFS r32, m16:32}<p>
	 * {@code o32 0F B4 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LFS_R32_M1632 = 2084;
	/**
	 * {@code LFS r64, m16:64}<p>
	 * {@code o64 0F B4 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LFS_R64_M1664 = 2085;
	/**
	 * {@code LGS r16, m16:16}<p>
	 * {@code o16 0F B5 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LGS_R16_M1616 = 2086;
	/**
	 * {@code LGS r32, m16:32}<p>
	 * {@code o32 0F B5 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LGS_R32_M1632 = 2087;
	/**
	 * {@code LGS r64, m16:64}<p>
	 * {@code o64 0F B5 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int LGS_R64_M1664 = 2088;
	/**
	 * {@code MOVZX r16, r/m8}<p>
	 * {@code o16 0F B6 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVZX_R16_RM8 = 2089;
	/**
	 * {@code MOVZX r32, r/m8}<p>
	 * {@code o32 0F B6 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVZX_R32_RM8 = 2090;
	/**
	 * {@code MOVZX r64, r/m8}<p>
	 * {@code o64 0F B6 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVZX_R64_RM8 = 2091;
	/**
	 * {@code MOVZX r16, r/m16}<p>
	 * {@code o16 0F B7 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVZX_R16_RM16 = 2092;
	/**
	 * {@code MOVZX r32, r/m16}<p>
	 * {@code o32 0F B7 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVZX_R32_RM16 = 2093;
	/**
	 * {@code MOVZX r64, r/m16}<p>
	 * {@code o64 0F B7 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVZX_R64_RM16 = 2094;
	/**
	 * {@code JMPE disp16}<p>
	 * {@code o16 0F B8 cw}<p>
	 * {@code IA-64}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JMPE_DISP16 = 2095;
	/**
	 * {@code JMPE disp32}<p>
	 * {@code o32 0F B8 cd}<p>
	 * {@code IA-64}<p>
	 * {@code 16/32-bit}
	 */
	public static final int JMPE_DISP32 = 2096;
	/**
	 * {@code POPCNT r16, r/m16}<p>
	 * {@code o16 F3 0F B8 /r}<p>
	 * {@code POPCNT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int POPCNT_R16_RM16 = 2097;
	/**
	 * {@code POPCNT r32, r/m32}<p>
	 * {@code o32 F3 0F B8 /r}<p>
	 * {@code POPCNT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int POPCNT_R32_RM32 = 2098;
	/**
	 * {@code POPCNT r64, r/m64}<p>
	 * {@code F3 o64 0F B8 /r}<p>
	 * {@code POPCNT}<p>
	 * {@code 64-bit}
	 */
	public static final int POPCNT_R64_RM64 = 2099;
	/**
	 * {@code UD1 r16, r/m16}<p>
	 * {@code o16 0F B9 /r}<p>
	 * {@code 286+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int UD1_R16_RM16 = 2100;
	/**
	 * {@code UD1 r32, r/m32}<p>
	 * {@code o32 0F B9 /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int UD1_R32_RM32 = 2101;
	/**
	 * {@code UD1 r64, r/m64}<p>
	 * {@code o64 0F B9 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int UD1_R64_RM64 = 2102;
	/**
	 * {@code BT r/m16, imm8}<p>
	 * {@code o16 0F BA /4 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BT_RM16_IMM8 = 2103;
	/**
	 * {@code BT r/m32, imm8}<p>
	 * {@code o32 0F BA /4 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BT_RM32_IMM8 = 2104;
	/**
	 * {@code BT r/m64, imm8}<p>
	 * {@code o64 0F BA /4 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int BT_RM64_IMM8 = 2105;
	/**
	 * {@code BTS r/m16, imm8}<p>
	 * {@code o16 0F BA /5 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BTS_RM16_IMM8 = 2106;
	/**
	 * {@code BTS r/m32, imm8}<p>
	 * {@code o32 0F BA /5 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BTS_RM32_IMM8 = 2107;
	/**
	 * {@code BTS r/m64, imm8}<p>
	 * {@code o64 0F BA /5 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int BTS_RM64_IMM8 = 2108;
	/**
	 * {@code BTR r/m16, imm8}<p>
	 * {@code o16 0F BA /6 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BTR_RM16_IMM8 = 2109;
	/**
	 * {@code BTR r/m32, imm8}<p>
	 * {@code o32 0F BA /6 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BTR_RM32_IMM8 = 2110;
	/**
	 * {@code BTR r/m64, imm8}<p>
	 * {@code o64 0F BA /6 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int BTR_RM64_IMM8 = 2111;
	/**
	 * {@code BTC r/m16, imm8}<p>
	 * {@code o16 0F BA /7 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BTC_RM16_IMM8 = 2112;
	/**
	 * {@code BTC r/m32, imm8}<p>
	 * {@code o32 0F BA /7 ib}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BTC_RM32_IMM8 = 2113;
	/**
	 * {@code BTC r/m64, imm8}<p>
	 * {@code o64 0F BA /7 ib}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int BTC_RM64_IMM8 = 2114;
	/**
	 * {@code BTC r/m16, r16}<p>
	 * {@code o16 0F BB /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BTC_RM16_R16 = 2115;
	/**
	 * {@code BTC r/m32, r32}<p>
	 * {@code o32 0F BB /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BTC_RM32_R32 = 2116;
	/**
	 * {@code BTC r/m64, r64}<p>
	 * {@code o64 0F BB /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int BTC_RM64_R64 = 2117;
	/**
	 * {@code BSF r16, r/m16}<p>
	 * {@code o16 0F BC /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BSF_R16_RM16 = 2118;
	/**
	 * {@code BSF r32, r/m32}<p>
	 * {@code o32 0F BC /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BSF_R32_RM32 = 2119;
	/**
	 * {@code BSF r64, r/m64}<p>
	 * {@code o64 0F BC /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int BSF_R64_RM64 = 2120;
	/**
	 * {@code TZCNT r16, r/m16}<p>
	 * {@code o16 F3 0F BC /r}<p>
	 * {@code BMI1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TZCNT_R16_RM16 = 2121;
	/**
	 * {@code TZCNT r32, r/m32}<p>
	 * {@code o32 F3 0F BC /r}<p>
	 * {@code BMI1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TZCNT_R32_RM32 = 2122;
	/**
	 * {@code TZCNT r64, r/m64}<p>
	 * {@code F3 o64 0F BC /r}<p>
	 * {@code BMI1}<p>
	 * {@code 64-bit}
	 */
	public static final int TZCNT_R64_RM64 = 2123;
	/**
	 * {@code BSR r16, r/m16}<p>
	 * {@code o16 0F BD /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BSR_R16_RM16 = 2124;
	/**
	 * {@code BSR r32, r/m32}<p>
	 * {@code o32 0F BD /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BSR_R32_RM32 = 2125;
	/**
	 * {@code BSR r64, r/m64}<p>
	 * {@code o64 0F BD /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int BSR_R64_RM64 = 2126;
	/**
	 * {@code LZCNT r16, r/m16}<p>
	 * {@code o16 F3 0F BD /r}<p>
	 * {@code LZCNT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LZCNT_R16_RM16 = 2127;
	/**
	 * {@code LZCNT r32, r/m32}<p>
	 * {@code o32 F3 0F BD /r}<p>
	 * {@code LZCNT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LZCNT_R32_RM32 = 2128;
	/**
	 * {@code LZCNT r64, r/m64}<p>
	 * {@code F3 o64 0F BD /r}<p>
	 * {@code LZCNT}<p>
	 * {@code 64-bit}
	 */
	public static final int LZCNT_R64_RM64 = 2129;
	/**
	 * {@code MOVSX r16, r/m8}<p>
	 * {@code o16 0F BE /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVSX_R16_RM8 = 2130;
	/**
	 * {@code MOVSX r32, r/m8}<p>
	 * {@code o32 0F BE /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVSX_R32_RM8 = 2131;
	/**
	 * {@code MOVSX r64, r/m8}<p>
	 * {@code o64 0F BE /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVSX_R64_RM8 = 2132;
	/**
	 * {@code MOVSX r16, r/m16}<p>
	 * {@code o16 0F BF /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVSX_R16_RM16 = 2133;
	/**
	 * {@code MOVSX r32, r/m16}<p>
	 * {@code o32 0F BF /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVSX_R32_RM16 = 2134;
	/**
	 * {@code MOVSX r64, r/m16}<p>
	 * {@code o64 0F BF /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVSX_R64_RM16 = 2135;
	/**
	 * {@code XADD r/m8, r8}<p>
	 * {@code 0F C0 /r}<p>
	 * {@code 486+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XADD_RM8_R8 = 2136;
	/**
	 * {@code XADD r/m16, r16}<p>
	 * {@code o16 0F C1 /r}<p>
	 * {@code 486+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XADD_RM16_R16 = 2137;
	/**
	 * {@code XADD r/m32, r32}<p>
	 * {@code o32 0F C1 /r}<p>
	 * {@code 486+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XADD_RM32_R32 = 2138;
	/**
	 * {@code XADD r/m64, r64}<p>
	 * {@code o64 0F C1 /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int XADD_RM64_R64 = 2139;
	/**
	 * {@code CMPPS xmm1, xmm2/m128, imm8}<p>
	 * {@code NP 0F C2 /r ib}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMPPS_XMM_XMMM128_IMM8 = 2140;
	/**
	 * {@code VCMPPS xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code VEX.128.0F.WIG C2 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCMPPS_XMM_XMM_XMMM128_IMM8 = 2141;
	/**
	 * {@code VCMPPS ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.0F.WIG C2 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCMPPS_YMM_YMM_YMMM256_IMM8 = 2142;
	/**
	 * {@code VCMPPS k1 {k2}, xmm2, xmm3/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.0F.W0 C2 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCMPPS_KR_K1_XMM_XMMM128B32_IMM8 = 2143;
	/**
	 * {@code VCMPPS k1 {k2}, ymm2, ymm3/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.0F.W0 C2 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCMPPS_KR_K1_YMM_YMMM256B32_IMM8 = 2144;
	/**
	 * {@code VCMPPS k1 {k2}, zmm2, zmm3/m512/m32bcst{sae}, imm8}<p>
	 * {@code EVEX.512.0F.W0 C2 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCMPPS_KR_K1_ZMM_ZMMM512B32_IMM8_SAE = 2145;
	/**
	 * {@code CMPPD xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F C2 /r ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMPPD_XMM_XMMM128_IMM8 = 2146;
	/**
	 * {@code VCMPPD xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code VEX.128.66.0F.WIG C2 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCMPPD_XMM_XMM_XMMM128_IMM8 = 2147;
	/**
	 * {@code VCMPPD ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.66.0F.WIG C2 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCMPPD_YMM_YMM_YMMM256_IMM8 = 2148;
	/**
	 * {@code VCMPPD k1 {k2}, xmm2, xmm3/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F.W1 C2 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCMPPD_KR_K1_XMM_XMMM128B64_IMM8 = 2149;
	/**
	 * {@code VCMPPD k1 {k2}, ymm2, ymm3/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F.W1 C2 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCMPPD_KR_K1_YMM_YMMM256B64_IMM8 = 2150;
	/**
	 * {@code VCMPPD k1 {k2}, zmm2, zmm3/m512/m64bcst{sae}, imm8}<p>
	 * {@code EVEX.512.66.0F.W1 C2 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCMPPD_KR_K1_ZMM_ZMMM512B64_IMM8_SAE = 2151;
	/**
	 * {@code CMPSS xmm1, xmm2/m32, imm8}<p>
	 * {@code F3 0F C2 /r ib}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMPSS_XMM_XMMM32_IMM8 = 2152;
	/**
	 * {@code VCMPSS xmm1, xmm2, xmm3/m32, imm8}<p>
	 * {@code VEX.LIG.F3.0F.WIG C2 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCMPSS_XMM_XMM_XMMM32_IMM8 = 2153;
	/**
	 * {@code VCMPSS k1 {k2}, xmm2, xmm3/m32{sae}, imm8}<p>
	 * {@code EVEX.LIG.F3.0F.W0 C2 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCMPSS_KR_K1_XMM_XMMM32_IMM8_SAE = 2154;
	/**
	 * {@code CMPSD xmm1, xmm2/m64, imm8}<p>
	 * {@code F2 0F C2 /r ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMPSD_XMM_XMMM64_IMM8 = 2155;
	/**
	 * {@code VCMPSD xmm1, xmm2, xmm3/m64, imm8}<p>
	 * {@code VEX.LIG.F2.0F.WIG C2 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCMPSD_XMM_XMM_XMMM64_IMM8 = 2156;
	/**
	 * {@code VCMPSD k1 {k2}, xmm2, xmm3/m64{sae}, imm8}<p>
	 * {@code EVEX.LIG.F2.0F.W1 C2 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCMPSD_KR_K1_XMM_XMMM64_IMM8_SAE = 2157;
	/**
	 * {@code MOVNTI m32, r32}<p>
	 * {@code NP 0F C3 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVNTI_M32_R32 = 2158;
	/**
	 * {@code MOVNTI m64, r64}<p>
	 * {@code NP o64 0F C3 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVNTI_M64_R64 = 2159;
	/**
	 * {@code PINSRW mm, r32/m16, imm8}<p>
	 * {@code NP 0F C4 /r ib}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PINSRW_MM_R32M16_IMM8 = 2160;
	/**
	 * {@code PINSRW mm, r64/m16, imm8}<p>
	 * {@code NP o64 0F C4 /r ib}<p>
	 * {@code SSE}<p>
	 * {@code 64-bit}
	 */
	public static final int PINSRW_MM_R64M16_IMM8 = 2161;
	/**
	 * {@code PINSRW xmm, r32/m16, imm8}<p>
	 * {@code 66 0F C4 /r ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PINSRW_XMM_R32M16_IMM8 = 2162;
	/**
	 * {@code PINSRW xmm, r64/m16, imm8}<p>
	 * {@code 66 o64 0F C4 /r ib}<p>
	 * {@code SSE2}<p>
	 * {@code 64-bit}
	 */
	public static final int PINSRW_XMM_R64M16_IMM8 = 2163;
	/**
	 * {@code VPINSRW xmm1, xmm2, r32/m16, imm8}<p>
	 * {@code VEX.128.66.0F.W0 C4 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPINSRW_XMM_XMM_R32M16_IMM8 = 2164;
	/**
	 * {@code VPINSRW xmm1, xmm2, r64/m16, imm8}<p>
	 * {@code VEX.128.66.0F.W1 C4 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VPINSRW_XMM_XMM_R64M16_IMM8 = 2165;
	/**
	 * {@code VPINSRW xmm1, xmm2, r32/m16, imm8}<p>
	 * {@code EVEX.128.66.0F.W0 C4 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPINSRW_XMM_XMM_R32M16_IMM8 = 2166;
	/**
	 * {@code VPINSRW xmm1, xmm2, r64/m16, imm8}<p>
	 * {@code EVEX.128.66.0F.W1 C4 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VPINSRW_XMM_XMM_R64M16_IMM8 = 2167;
	/**
	 * {@code PEXTRW r32, mm, imm8}<p>
	 * {@code NP 0F C5 /r ib}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PEXTRW_R32_MM_IMM8 = 2168;
	/**
	 * {@code PEXTRW r64, mm, imm8}<p>
	 * {@code NP o64 0F C5 /r ib}<p>
	 * {@code SSE}<p>
	 * {@code 64-bit}
	 */
	public static final int PEXTRW_R64_MM_IMM8 = 2169;
	/**
	 * {@code PEXTRW r32, xmm, imm8}<p>
	 * {@code 66 0F C5 /r ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PEXTRW_R32_XMM_IMM8 = 2170;
	/**
	 * {@code PEXTRW r64, xmm, imm8}<p>
	 * {@code 66 o64 0F C5 /r ib}<p>
	 * {@code SSE2}<p>
	 * {@code 64-bit}
	 */
	public static final int PEXTRW_R64_XMM_IMM8 = 2171;
	/**
	 * {@code VPEXTRW r32, xmm1, imm8}<p>
	 * {@code VEX.128.66.0F.W0 C5 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPEXTRW_R32_XMM_IMM8 = 2172;
	/**
	 * {@code VPEXTRW r64, xmm1, imm8}<p>
	 * {@code VEX.128.66.0F.W1 C5 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VPEXTRW_R64_XMM_IMM8 = 2173;
	/**
	 * {@code VPEXTRW r32, xmm1, imm8}<p>
	 * {@code EVEX.128.66.0F.W0 C5 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXTRW_R32_XMM_IMM8 = 2174;
	/**
	 * {@code VPEXTRW r64, xmm1, imm8}<p>
	 * {@code EVEX.128.66.0F.W1 C5 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VPEXTRW_R64_XMM_IMM8 = 2175;
	/**
	 * {@code SHUFPS xmm1, xmm2/m128, imm8}<p>
	 * {@code NP 0F C6 /r ib}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHUFPS_XMM_XMMM128_IMM8 = 2176;
	/**
	 * {@code VSHUFPS xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code VEX.128.0F.WIG C6 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSHUFPS_XMM_XMM_XMMM128_IMM8 = 2177;
	/**
	 * {@code VSHUFPS ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.0F.WIG C6 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSHUFPS_YMM_YMM_YMMM256_IMM8 = 2178;
	/**
	 * {@code VSHUFPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.0F.W0 C6 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSHUFPS_XMM_K1Z_XMM_XMMM128B32_IMM8 = 2179;
	/**
	 * {@code VSHUFPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.0F.W0 C6 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSHUFPS_YMM_K1Z_YMM_YMMM256B32_IMM8 = 2180;
	/**
	 * {@code VSHUFPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.0F.W0 C6 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSHUFPS_ZMM_K1Z_ZMM_ZMMM512B32_IMM8 = 2181;
	/**
	 * {@code SHUFPD xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F C6 /r ib}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHUFPD_XMM_XMMM128_IMM8 = 2182;
	/**
	 * {@code VSHUFPD xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code VEX.128.66.0F.WIG C6 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSHUFPD_XMM_XMM_XMMM128_IMM8 = 2183;
	/**
	 * {@code VSHUFPD ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.66.0F.WIG C6 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VSHUFPD_YMM_YMM_YMMM256_IMM8 = 2184;
	/**
	 * {@code VSHUFPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F.W1 C6 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSHUFPD_XMM_K1Z_XMM_XMMM128B64_IMM8 = 2185;
	/**
	 * {@code VSHUFPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F.W1 C6 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSHUFPD_YMM_K1Z_YMM_YMMM256B64_IMM8 = 2186;
	/**
	 * {@code VSHUFPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F.W1 C6 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSHUFPD_ZMM_K1Z_ZMM_ZMMM512B64_IMM8 = 2187;
	/**
	 * {@code CMPXCHG8B m64}<p>
	 * {@code 0F C7 /1}<p>
	 * {@code CX8}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CMPXCHG8B_M64 = 2188;
	/**
	 * {@code CMPXCHG16B m128}<p>
	 * {@code o64 0F C7 /1}<p>
	 * {@code CMPXCHG16B}<p>
	 * {@code 64-bit}
	 */
	public static final int CMPXCHG16B_M128 = 2189;
	/**
	 * {@code XRSTORS mem}<p>
	 * {@code NP 0F C7 /3}<p>
	 * {@code XSAVES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XRSTORS_MEM = 2190;
	/**
	 * {@code XRSTORS64 mem}<p>
	 * {@code NP o64 0F C7 /3}<p>
	 * {@code XSAVES}<p>
	 * {@code 64-bit}
	 */
	public static final int XRSTORS64_MEM = 2191;
	/**
	 * {@code XSAVEC mem}<p>
	 * {@code NP 0F C7 /4}<p>
	 * {@code XSAVEC}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XSAVEC_MEM = 2192;
	/**
	 * {@code XSAVEC64 mem}<p>
	 * {@code NP o64 0F C7 /4}<p>
	 * {@code XSAVEC}<p>
	 * {@code 64-bit}
	 */
	public static final int XSAVEC64_MEM = 2193;
	/**
	 * {@code XSAVES mem}<p>
	 * {@code NP 0F C7 /5}<p>
	 * {@code XSAVES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XSAVES_MEM = 2194;
	/**
	 * {@code XSAVES64 mem}<p>
	 * {@code NP o64 0F C7 /5}<p>
	 * {@code XSAVES}<p>
	 * {@code 64-bit}
	 */
	public static final int XSAVES64_MEM = 2195;
	/**
	 * {@code VMPTRLD m64}<p>
	 * {@code NP 0F C7 /6}<p>
	 * {@code VMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VMPTRLD_M64 = 2196;
	/**
	 * {@code VMCLEAR m64}<p>
	 * {@code 66 0F C7 /6}<p>
	 * {@code VMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VMCLEAR_M64 = 2197;
	/**
	 * {@code VMXON m64}<p>
	 * {@code F3 0F C7 /6}<p>
	 * {@code VMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VMXON_M64 = 2198;
	/**
	 * {@code RDRAND r16}<p>
	 * {@code o16 0F C7 /6}<p>
	 * {@code RDRAND}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RDRAND_R16 = 2199;
	/**
	 * {@code RDRAND r32}<p>
	 * {@code o32 0F C7 /6}<p>
	 * {@code RDRAND}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RDRAND_R32 = 2200;
	/**
	 * {@code RDRAND r64}<p>
	 * {@code o64 0F C7 /6}<p>
	 * {@code RDRAND}<p>
	 * {@code 64-bit}
	 */
	public static final int RDRAND_R64 = 2201;
	/**
	 * {@code VMPTRST m64}<p>
	 * {@code NP 0F C7 /7}<p>
	 * {@code VMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VMPTRST_M64 = 2202;
	/**
	 * {@code RDSEED r16}<p>
	 * {@code o16 0F C7 /7}<p>
	 * {@code RDSEED}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RDSEED_R16 = 2203;
	/**
	 * {@code RDSEED r32}<p>
	 * {@code o32 0F C7 /7}<p>
	 * {@code RDSEED}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RDSEED_R32 = 2204;
	/**
	 * {@code RDSEED r64}<p>
	 * {@code o64 0F C7 /7}<p>
	 * {@code RDSEED}<p>
	 * {@code 64-bit}
	 */
	public static final int RDSEED_R64 = 2205;
	/**
	 * {@code RDPID r32}<p>
	 * {@code F3 0F C7 /7}<p>
	 * {@code RDPID}<p>
	 * {@code 16/32-bit}
	 */
	public static final int RDPID_R32 = 2206;
	/**
	 * {@code RDPID r64}<p>
	 * {@code F3 0F C7 /7}<p>
	 * {@code RDPID}<p>
	 * {@code 64-bit}
	 */
	public static final int RDPID_R64 = 2207;
	/**
	 * {@code BSWAP r16}<p>
	 * {@code o16 0F C8+rw}<p>
	 * {@code 486+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BSWAP_R16 = 2208;
	/**
	 * {@code BSWAP r32}<p>
	 * {@code o32 0F C8+rd}<p>
	 * {@code 486+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BSWAP_R32 = 2209;
	/**
	 * {@code BSWAP r64}<p>
	 * {@code o64 0F C8+ro}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int BSWAP_R64 = 2210;
	/**
	 * {@code ADDSUBPD xmm1, xmm2/m128}<p>
	 * {@code 66 0F D0 /r}<p>
	 * {@code SSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADDSUBPD_XMM_XMMM128 = 2211;
	/**
	 * {@code VADDSUBPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG D0 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VADDSUBPD_XMM_XMM_XMMM128 = 2212;
	/**
	 * {@code VADDSUBPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG D0 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VADDSUBPD_YMM_YMM_YMMM256 = 2213;
	/**
	 * {@code ADDSUBPS xmm1, xmm2/m128}<p>
	 * {@code F2 0F D0 /r}<p>
	 * {@code SSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADDSUBPS_XMM_XMMM128 = 2214;
	/**
	 * {@code VADDSUBPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.F2.0F.WIG D0 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VADDSUBPS_XMM_XMM_XMMM128 = 2215;
	/**
	 * {@code VADDSUBPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.F2.0F.WIG D0 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VADDSUBPS_YMM_YMM_YMMM256 = 2216;
	/**
	 * {@code PSRLW mm, mm/m64}<p>
	 * {@code NP 0F D1 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRLW_MM_MMM64 = 2217;
	/**
	 * {@code PSRLW xmm1, xmm2/m128}<p>
	 * {@code 66 0F D1 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRLW_XMM_XMMM128 = 2218;
	/**
	 * {@code VPSRLW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG D1 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLW_XMM_XMM_XMMM128 = 2219;
	/**
	 * {@code VPSRLW ymm1, ymm2, xmm3/m128}<p>
	 * {@code VEX.256.66.0F.WIG D1 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLW_YMM_YMM_XMMM128 = 2220;
	/**
	 * {@code VPSRLW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG D1 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLW_XMM_K1Z_XMM_XMMM128 = 2221;
	/**
	 * {@code VPSRLW ymm1 {k1}{z}, ymm2, xmm3/m128}<p>
	 * {@code EVEX.256.66.0F.WIG D1 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLW_YMM_K1Z_YMM_XMMM128 = 2222;
	/**
	 * {@code VPSRLW zmm1 {k1}{z}, zmm2, xmm3/m128}<p>
	 * {@code EVEX.512.66.0F.WIG D1 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLW_ZMM_K1Z_ZMM_XMMM128 = 2223;
	/**
	 * {@code PSRLD mm, mm/m64}<p>
	 * {@code NP 0F D2 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRLD_MM_MMM64 = 2224;
	/**
	 * {@code PSRLD xmm1, xmm2/m128}<p>
	 * {@code 66 0F D2 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRLD_XMM_XMMM128 = 2225;
	/**
	 * {@code VPSRLD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG D2 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLD_XMM_XMM_XMMM128 = 2226;
	/**
	 * {@code VPSRLD ymm1, ymm2, xmm3/m128}<p>
	 * {@code VEX.256.66.0F.WIG D2 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLD_YMM_YMM_XMMM128 = 2227;
	/**
	 * {@code VPSRLD xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.W0 D2 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLD_XMM_K1Z_XMM_XMMM128 = 2228;
	/**
	 * {@code VPSRLD ymm1 {k1}{z}, ymm2, xmm3/m128}<p>
	 * {@code EVEX.256.66.0F.W0 D2 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLD_YMM_K1Z_YMM_XMMM128 = 2229;
	/**
	 * {@code VPSRLD zmm1 {k1}{z}, zmm2, xmm3/m128}<p>
	 * {@code EVEX.512.66.0F.W0 D2 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLD_ZMM_K1Z_ZMM_XMMM128 = 2230;
	/**
	 * {@code PSRLQ mm, mm/m64}<p>
	 * {@code NP 0F D3 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRLQ_MM_MMM64 = 2231;
	/**
	 * {@code PSRLQ xmm1, xmm2/m128}<p>
	 * {@code 66 0F D3 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRLQ_XMM_XMMM128 = 2232;
	/**
	 * {@code VPSRLQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG D3 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLQ_XMM_XMM_XMMM128 = 2233;
	/**
	 * {@code VPSRLQ ymm1, ymm2, xmm3/m128}<p>
	 * {@code VEX.256.66.0F.WIG D3 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLQ_YMM_YMM_XMMM128 = 2234;
	/**
	 * {@code VPSRLQ xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.W1 D3 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLQ_XMM_K1Z_XMM_XMMM128 = 2235;
	/**
	 * {@code VPSRLQ ymm1 {k1}{z}, ymm2, xmm3/m128}<p>
	 * {@code EVEX.256.66.0F.W1 D3 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLQ_YMM_K1Z_YMM_XMMM128 = 2236;
	/**
	 * {@code VPSRLQ zmm1 {k1}{z}, zmm2, xmm3/m128}<p>
	 * {@code EVEX.512.66.0F.W1 D3 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLQ_ZMM_K1Z_ZMM_XMMM128 = 2237;
	/**
	 * {@code PADDQ mm, mm/m64}<p>
	 * {@code NP 0F D4 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDQ_MM_MMM64 = 2238;
	/**
	 * {@code PADDQ xmm1, xmm2/m128}<p>
	 * {@code 66 0F D4 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDQ_XMM_XMMM128 = 2239;
	/**
	 * {@code VPADDQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG D4 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDQ_XMM_XMM_XMMM128 = 2240;
	/**
	 * {@code VPADDQ ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG D4 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDQ_YMM_YMM_YMMM256 = 2241;
	/**
	 * {@code VPADDQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 D4 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDQ_XMM_K1Z_XMM_XMMM128B64 = 2242;
	/**
	 * {@code VPADDQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 D4 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDQ_YMM_K1Z_YMM_YMMM256B64 = 2243;
	/**
	 * {@code VPADDQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F.W1 D4 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDQ_ZMM_K1Z_ZMM_ZMMM512B64 = 2244;
	/**
	 * {@code PMULLW mm, mm/m64}<p>
	 * {@code NP 0F D5 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMULLW_MM_MMM64 = 2245;
	/**
	 * {@code PMULLW xmm1, xmm2/m128}<p>
	 * {@code 66 0F D5 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMULLW_XMM_XMMM128 = 2246;
	/**
	 * {@code VPMULLW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG D5 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMULLW_XMM_XMM_XMMM128 = 2247;
	/**
	 * {@code VPMULLW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG D5 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMULLW_YMM_YMM_YMMM256 = 2248;
	/**
	 * {@code VPMULLW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG D5 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULLW_XMM_K1Z_XMM_XMMM128 = 2249;
	/**
	 * {@code VPMULLW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG D5 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULLW_YMM_K1Z_YMM_YMMM256 = 2250;
	/**
	 * {@code VPMULLW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG D5 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULLW_ZMM_K1Z_ZMM_ZMMM512 = 2251;
	/**
	 * {@code MOVQ xmm2/m64, xmm1}<p>
	 * {@code 66 0F D6 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVQ_XMMM64_XMM = 2252;
	/**
	 * {@code VMOVQ xmm1/m64, xmm2}<p>
	 * {@code VEX.128.66.0F.WIG D6 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVQ_XMMM64_XMM = 2253;
	/**
	 * {@code VMOVQ xmm1/m64, xmm2}<p>
	 * {@code EVEX.128.66.0F.W1 D6 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVQ_XMMM64_XMM = 2254;
	/**
	 * {@code MOVQ2DQ xmm, mm}<p>
	 * {@code F3 0F D6 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVQ2DQ_XMM_MM = 2255;
	/**
	 * {@code MOVDQ2Q mm, xmm}<p>
	 * {@code F2 0F D6 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVDQ2Q_MM_XMM = 2256;
	/**
	 * {@code PMOVMSKB r32, mm}<p>
	 * {@code NP 0F D7 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMOVMSKB_R32_MM = 2257;
	/**
	 * {@code PMOVMSKB r64, mm}<p>
	 * {@code NP o64 0F D7 /r}<p>
	 * {@code SSE}<p>
	 * {@code 64-bit}
	 */
	public static final int PMOVMSKB_R64_MM = 2258;
	/**
	 * {@code PMOVMSKB r32, xmm}<p>
	 * {@code 66 0F D7 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMOVMSKB_R32_XMM = 2259;
	/**
	 * {@code PMOVMSKB r64, xmm}<p>
	 * {@code 66 o64 0F D7 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 64-bit}
	 */
	public static final int PMOVMSKB_R64_XMM = 2260;
	/**
	 * {@code VPMOVMSKB r32, xmm1}<p>
	 * {@code VEX.128.66.0F.W0 D7 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVMSKB_R32_XMM = 2261;
	/**
	 * {@code VPMOVMSKB r64, xmm1}<p>
	 * {@code VEX.128.66.0F.W1 D7 /r}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VPMOVMSKB_R64_XMM = 2262;
	/**
	 * {@code VPMOVMSKB r32, ymm1}<p>
	 * {@code VEX.256.66.0F.W0 D7 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVMSKB_R32_YMM = 2263;
	/**
	 * {@code VPMOVMSKB r64, ymm1}<p>
	 * {@code VEX.256.66.0F.W1 D7 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VPMOVMSKB_R64_YMM = 2264;
	/**
	 * {@code PSUBUSB mm, mm/m64}<p>
	 * {@code NP 0F D8 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBUSB_MM_MMM64 = 2265;
	/**
	 * {@code PSUBUSB xmm1, xmm2/m128}<p>
	 * {@code 66 0F D8 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBUSB_XMM_XMMM128 = 2266;
	/**
	 * {@code VPSUBUSB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG D8 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBUSB_XMM_XMM_XMMM128 = 2267;
	/**
	 * {@code VPSUBUSB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG D8 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBUSB_YMM_YMM_YMMM256 = 2268;
	/**
	 * {@code VPSUBUSB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG D8 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBUSB_XMM_K1Z_XMM_XMMM128 = 2269;
	/**
	 * {@code VPSUBUSB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG D8 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBUSB_YMM_K1Z_YMM_YMMM256 = 2270;
	/**
	 * {@code VPSUBUSB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG D8 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBUSB_ZMM_K1Z_ZMM_ZMMM512 = 2271;
	/**
	 * {@code PSUBUSW mm, mm/m64}<p>
	 * {@code NP 0F D9 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBUSW_MM_MMM64 = 2272;
	/**
	 * {@code PSUBUSW xmm1, xmm2/m128}<p>
	 * {@code 66 0F D9 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBUSW_XMM_XMMM128 = 2273;
	/**
	 * {@code VPSUBUSW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG D9 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBUSW_XMM_XMM_XMMM128 = 2274;
	/**
	 * {@code VPSUBUSW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG D9 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBUSW_YMM_YMM_YMMM256 = 2275;
	/**
	 * {@code VPSUBUSW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG D9 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBUSW_XMM_K1Z_XMM_XMMM128 = 2276;
	/**
	 * {@code VPSUBUSW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG D9 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBUSW_YMM_K1Z_YMM_YMMM256 = 2277;
	/**
	 * {@code VPSUBUSW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG D9 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBUSW_ZMM_K1Z_ZMM_ZMMM512 = 2278;
	/**
	 * {@code PMINUB mm1, mm2/m64}<p>
	 * {@code NP 0F DA /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMINUB_MM_MMM64 = 2279;
	/**
	 * {@code PMINUB xmm1, xmm2/m128}<p>
	 * {@code 66 0F DA /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMINUB_XMM_XMMM128 = 2280;
	/**
	 * {@code VPMINUB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG DA /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMINUB_XMM_XMM_XMMM128 = 2281;
	/**
	 * {@code VPMINUB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG DA /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMINUB_YMM_YMM_YMMM256 = 2282;
	/**
	 * {@code VPMINUB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG DA /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINUB_XMM_K1Z_XMM_XMMM128 = 2283;
	/**
	 * {@code VPMINUB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG DA /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINUB_YMM_K1Z_YMM_YMMM256 = 2284;
	/**
	 * {@code VPMINUB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG DA /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINUB_ZMM_K1Z_ZMM_ZMMM512 = 2285;
	/**
	 * {@code PAND mm, mm/m64}<p>
	 * {@code NP 0F DB /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PAND_MM_MMM64 = 2286;
	/**
	 * {@code PAND xmm1, xmm2/m128}<p>
	 * {@code 66 0F DB /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PAND_XMM_XMMM128 = 2287;
	/**
	 * {@code VPAND xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG DB /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPAND_XMM_XMM_XMMM128 = 2288;
	/**
	 * {@code VPAND ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG DB /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPAND_YMM_YMM_YMMM256 = 2289;
	/**
	 * {@code VPANDD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 DB /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPANDD_XMM_K1Z_XMM_XMMM128B32 = 2290;
	/**
	 * {@code VPANDD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 DB /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPANDD_YMM_K1Z_YMM_YMMM256B32 = 2291;
	/**
	 * {@code VPANDD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F.W0 DB /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPANDD_ZMM_K1Z_ZMM_ZMMM512B32 = 2292;
	/**
	 * {@code VPANDQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 DB /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPANDQ_XMM_K1Z_XMM_XMMM128B64 = 2293;
	/**
	 * {@code VPANDQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 DB /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPANDQ_YMM_K1Z_YMM_YMMM256B64 = 2294;
	/**
	 * {@code VPANDQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F.W1 DB /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPANDQ_ZMM_K1Z_ZMM_ZMMM512B64 = 2295;
	/**
	 * {@code PADDUSB mm, mm/m64}<p>
	 * {@code NP 0F DC /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDUSB_MM_MMM64 = 2296;
	/**
	 * {@code PADDUSB xmm1, xmm2/m128}<p>
	 * {@code 66 0F DC /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDUSB_XMM_XMMM128 = 2297;
	/**
	 * {@code VPADDUSB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG DC /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDUSB_XMM_XMM_XMMM128 = 2298;
	/**
	 * {@code VPADDUSB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG DC /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDUSB_YMM_YMM_YMMM256 = 2299;
	/**
	 * {@code VPADDUSB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG DC /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDUSB_XMM_K1Z_XMM_XMMM128 = 2300;
	/**
	 * {@code VPADDUSB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG DC /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDUSB_YMM_K1Z_YMM_YMMM256 = 2301;
	/**
	 * {@code VPADDUSB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG DC /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDUSB_ZMM_K1Z_ZMM_ZMMM512 = 2302;
	/**
	 * {@code PADDUSW mm, mm/m64}<p>
	 * {@code NP 0F DD /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDUSW_MM_MMM64 = 2303;
	/**
	 * {@code PADDUSW xmm1, xmm2/m128}<p>
	 * {@code 66 0F DD /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDUSW_XMM_XMMM128 = 2304;
	/**
	 * {@code VPADDUSW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG DD /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDUSW_XMM_XMM_XMMM128 = 2305;
	/**
	 * {@code VPADDUSW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG DD /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDUSW_YMM_YMM_YMMM256 = 2306;
	/**
	 * {@code VPADDUSW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG DD /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDUSW_XMM_K1Z_XMM_XMMM128 = 2307;
	/**
	 * {@code VPADDUSW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG DD /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDUSW_YMM_K1Z_YMM_YMMM256 = 2308;
	/**
	 * {@code VPADDUSW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG DD /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDUSW_ZMM_K1Z_ZMM_ZMMM512 = 2309;
	/**
	 * {@code PMAXUB mm1, mm2/m64}<p>
	 * {@code NP 0F DE /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMAXUB_MM_MMM64 = 2310;
	/**
	 * {@code PMAXUB xmm1, xmm2/m128}<p>
	 * {@code 66 0F DE /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMAXUB_XMM_XMMM128 = 2311;
	/**
	 * {@code VPMAXUB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG DE /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMAXUB_XMM_XMM_XMMM128 = 2312;
	/**
	 * {@code VPMAXUB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG DE /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMAXUB_YMM_YMM_YMMM256 = 2313;
	/**
	 * {@code VPMAXUB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG DE /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXUB_XMM_K1Z_XMM_XMMM128 = 2314;
	/**
	 * {@code VPMAXUB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG DE /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXUB_YMM_K1Z_YMM_YMMM256 = 2315;
	/**
	 * {@code VPMAXUB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG DE /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXUB_ZMM_K1Z_ZMM_ZMMM512 = 2316;
	/**
	 * {@code PANDN mm, mm/m64}<p>
	 * {@code NP 0F DF /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PANDN_MM_MMM64 = 2317;
	/**
	 * {@code PANDN xmm1, xmm2/m128}<p>
	 * {@code 66 0F DF /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PANDN_XMM_XMMM128 = 2318;
	/**
	 * {@code VPANDN xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG DF /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPANDN_XMM_XMM_XMMM128 = 2319;
	/**
	 * {@code VPANDN ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG DF /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPANDN_YMM_YMM_YMMM256 = 2320;
	/**
	 * {@code VPANDND xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 DF /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPANDND_XMM_K1Z_XMM_XMMM128B32 = 2321;
	/**
	 * {@code VPANDND ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 DF /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPANDND_YMM_K1Z_YMM_YMMM256B32 = 2322;
	/**
	 * {@code VPANDND zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F.W0 DF /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPANDND_ZMM_K1Z_ZMM_ZMMM512B32 = 2323;
	/**
	 * {@code VPANDNQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 DF /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPANDNQ_XMM_K1Z_XMM_XMMM128B64 = 2324;
	/**
	 * {@code VPANDNQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 DF /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPANDNQ_YMM_K1Z_YMM_YMMM256B64 = 2325;
	/**
	 * {@code VPANDNQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F.W1 DF /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPANDNQ_ZMM_K1Z_ZMM_ZMMM512B64 = 2326;
	/**
	 * {@code PAVGB mm1, mm2/m64}<p>
	 * {@code NP 0F E0 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PAVGB_MM_MMM64 = 2327;
	/**
	 * {@code PAVGB xmm1, xmm2/m128}<p>
	 * {@code 66 0F E0 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PAVGB_XMM_XMMM128 = 2328;
	/**
	 * {@code VPAVGB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG E0 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPAVGB_XMM_XMM_XMMM128 = 2329;
	/**
	 * {@code VPAVGB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG E0 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPAVGB_YMM_YMM_YMMM256 = 2330;
	/**
	 * {@code VPAVGB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG E0 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPAVGB_XMM_K1Z_XMM_XMMM128 = 2331;
	/**
	 * {@code VPAVGB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG E0 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPAVGB_YMM_K1Z_YMM_YMMM256 = 2332;
	/**
	 * {@code VPAVGB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG E0 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPAVGB_ZMM_K1Z_ZMM_ZMMM512 = 2333;
	/**
	 * {@code PSRAW mm, mm/m64}<p>
	 * {@code NP 0F E1 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRAW_MM_MMM64 = 2334;
	/**
	 * {@code PSRAW xmm1, xmm2/m128}<p>
	 * {@code 66 0F E1 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRAW_XMM_XMMM128 = 2335;
	/**
	 * {@code VPSRAW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG E1 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRAW_XMM_XMM_XMMM128 = 2336;
	/**
	 * {@code VPSRAW ymm1, ymm2, xmm3/m128}<p>
	 * {@code VEX.256.66.0F.WIG E1 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRAW_YMM_YMM_XMMM128 = 2337;
	/**
	 * {@code VPSRAW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG E1 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAW_XMM_K1Z_XMM_XMMM128 = 2338;
	/**
	 * {@code VPSRAW ymm1 {k1}{z}, ymm2, xmm3/m128}<p>
	 * {@code EVEX.256.66.0F.WIG E1 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAW_YMM_K1Z_YMM_XMMM128 = 2339;
	/**
	 * {@code VPSRAW zmm1 {k1}{z}, zmm2, xmm3/m128}<p>
	 * {@code EVEX.512.66.0F.WIG E1 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAW_ZMM_K1Z_ZMM_XMMM128 = 2340;
	/**
	 * {@code PSRAD mm, mm/m64}<p>
	 * {@code NP 0F E2 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRAD_MM_MMM64 = 2341;
	/**
	 * {@code PSRAD xmm1, xmm2/m128}<p>
	 * {@code 66 0F E2 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSRAD_XMM_XMMM128 = 2342;
	/**
	 * {@code VPSRAD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG E2 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRAD_XMM_XMM_XMMM128 = 2343;
	/**
	 * {@code VPSRAD ymm1, ymm2, xmm3/m128}<p>
	 * {@code VEX.256.66.0F.WIG E2 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRAD_YMM_YMM_XMMM128 = 2344;
	/**
	 * {@code VPSRAD xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.W0 E2 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAD_XMM_K1Z_XMM_XMMM128 = 2345;
	/**
	 * {@code VPSRAD ymm1 {k1}{z}, ymm2, xmm3/m128}<p>
	 * {@code EVEX.256.66.0F.W0 E2 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAD_YMM_K1Z_YMM_XMMM128 = 2346;
	/**
	 * {@code VPSRAD zmm1 {k1}{z}, zmm2, xmm3/m128}<p>
	 * {@code EVEX.512.66.0F.W0 E2 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAD_ZMM_K1Z_ZMM_XMMM128 = 2347;
	/**
	 * {@code VPSRAQ xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.W1 E2 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAQ_XMM_K1Z_XMM_XMMM128 = 2348;
	/**
	 * {@code VPSRAQ ymm1 {k1}{z}, ymm2, xmm3/m128}<p>
	 * {@code EVEX.256.66.0F.W1 E2 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAQ_YMM_K1Z_YMM_XMMM128 = 2349;
	/**
	 * {@code VPSRAQ zmm1 {k1}{z}, zmm2, xmm3/m128}<p>
	 * {@code EVEX.512.66.0F.W1 E2 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAQ_ZMM_K1Z_ZMM_XMMM128 = 2350;
	/**
	 * {@code PAVGW mm1, mm2/m64}<p>
	 * {@code NP 0F E3 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PAVGW_MM_MMM64 = 2351;
	/**
	 * {@code PAVGW xmm1, xmm2/m128}<p>
	 * {@code 66 0F E3 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PAVGW_XMM_XMMM128 = 2352;
	/**
	 * {@code VPAVGW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG E3 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPAVGW_XMM_XMM_XMMM128 = 2353;
	/**
	 * {@code VPAVGW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG E3 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPAVGW_YMM_YMM_YMMM256 = 2354;
	/**
	 * {@code VPAVGW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG E3 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPAVGW_XMM_K1Z_XMM_XMMM128 = 2355;
	/**
	 * {@code VPAVGW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG E3 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPAVGW_YMM_K1Z_YMM_YMMM256 = 2356;
	/**
	 * {@code VPAVGW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG E3 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPAVGW_ZMM_K1Z_ZMM_ZMMM512 = 2357;
	/**
	 * {@code PMULHUW mm1, mm2/m64}<p>
	 * {@code NP 0F E4 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMULHUW_MM_MMM64 = 2358;
	/**
	 * {@code PMULHUW xmm1, xmm2/m128}<p>
	 * {@code 66 0F E4 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMULHUW_XMM_XMMM128 = 2359;
	/**
	 * {@code VPMULHUW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG E4 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMULHUW_XMM_XMM_XMMM128 = 2360;
	/**
	 * {@code VPMULHUW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG E4 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMULHUW_YMM_YMM_YMMM256 = 2361;
	/**
	 * {@code VPMULHUW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG E4 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULHUW_XMM_K1Z_XMM_XMMM128 = 2362;
	/**
	 * {@code VPMULHUW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG E4 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULHUW_YMM_K1Z_YMM_YMMM256 = 2363;
	/**
	 * {@code VPMULHUW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG E4 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULHUW_ZMM_K1Z_ZMM_ZMMM512 = 2364;
	/**
	 * {@code PMULHW mm, mm/m64}<p>
	 * {@code NP 0F E5 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMULHW_MM_MMM64 = 2365;
	/**
	 * {@code PMULHW xmm1, xmm2/m128}<p>
	 * {@code 66 0F E5 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMULHW_XMM_XMMM128 = 2366;
	/**
	 * {@code VPMULHW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG E5 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMULHW_XMM_XMM_XMMM128 = 2367;
	/**
	 * {@code VPMULHW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG E5 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMULHW_YMM_YMM_YMMM256 = 2368;
	/**
	 * {@code VPMULHW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG E5 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULHW_XMM_K1Z_XMM_XMMM128 = 2369;
	/**
	 * {@code VPMULHW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG E5 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULHW_YMM_K1Z_YMM_YMMM256 = 2370;
	/**
	 * {@code VPMULHW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG E5 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULHW_ZMM_K1Z_ZMM_ZMMM512 = 2371;
	/**
	 * {@code CVTTPD2DQ xmm1, xmm2/m128}<p>
	 * {@code 66 0F E6 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTTPD2DQ_XMM_XMMM128 = 2372;
	/**
	 * {@code VCVTTPD2DQ xmm1, xmm2/m128}<p>
	 * {@code VEX.128.66.0F.WIG E6 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTTPD2DQ_XMM_XMMM128 = 2373;
	/**
	 * {@code VCVTTPD2DQ xmm1, ymm2/m256}<p>
	 * {@code VEX.256.66.0F.WIG E6 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTTPD2DQ_XMM_YMMM256 = 2374;
	/**
	 * {@code VCVTTPD2DQ xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 E6 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPD2DQ_XMM_K1Z_XMMM128B64 = 2375;
	/**
	 * {@code VCVTTPD2DQ xmm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 E6 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPD2DQ_XMM_K1Z_YMMM256B64 = 2376;
	/**
	 * {@code VCVTTPD2DQ ymm1 {k1}{z}, zmm2/m512/m64bcst{sae}}<p>
	 * {@code EVEX.512.66.0F.W1 E6 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPD2DQ_YMM_K1Z_ZMMM512B64_SAE = 2377;
	/**
	 * {@code CVTDQ2PD xmm1, xmm2/m64}<p>
	 * {@code F3 0F E6 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTDQ2PD_XMM_XMMM64 = 2378;
	/**
	 * {@code VCVTDQ2PD xmm1, xmm2/m64}<p>
	 * {@code VEX.128.F3.0F.WIG E6 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTDQ2PD_XMM_XMMM64 = 2379;
	/**
	 * {@code VCVTDQ2PD ymm1, xmm2/m128}<p>
	 * {@code VEX.256.F3.0F.WIG E6 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTDQ2PD_YMM_XMMM128 = 2380;
	/**
	 * {@code VCVTDQ2PD xmm1 {k1}{z}, xmm2/m64/m32bcst}<p>
	 * {@code EVEX.128.F3.0F.W0 E6 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTDQ2PD_XMM_K1Z_XMMM64B32 = 2381;
	/**
	 * {@code VCVTDQ2PD ymm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.256.F3.0F.W0 E6 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTDQ2PD_YMM_K1Z_XMMM128B32 = 2382;
	/**
	 * {@code VCVTDQ2PD zmm1 {k1}{z}, ymm2/m256/m32bcst{er}}<p>
	 * {@code EVEX.512.F3.0F.W0 E6 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTDQ2PD_ZMM_K1Z_YMMM256B32_ER = 2383;
	/**
	 * {@code VCVTQQ2PD xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.F3.0F.W1 E6 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTQQ2PD_XMM_K1Z_XMMM128B64 = 2384;
	/**
	 * {@code VCVTQQ2PD ymm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.F3.0F.W1 E6 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTQQ2PD_YMM_K1Z_YMMM256B64 = 2385;
	/**
	 * {@code VCVTQQ2PD zmm1 {k1}{z}, zmm2/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.F3.0F.W1 E6 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTQQ2PD_ZMM_K1Z_ZMMM512B64_ER = 2386;
	/**
	 * {@code CVTPD2DQ xmm1, xmm2/m128}<p>
	 * {@code F2 0F E6 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CVTPD2DQ_XMM_XMMM128 = 2387;
	/**
	 * {@code VCVTPD2DQ xmm1, xmm2/m128}<p>
	 * {@code VEX.128.F2.0F.WIG E6 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTPD2DQ_XMM_XMMM128 = 2388;
	/**
	 * {@code VCVTPD2DQ xmm1, ymm2/m256}<p>
	 * {@code VEX.256.F2.0F.WIG E6 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTPD2DQ_XMM_YMMM256 = 2389;
	/**
	 * {@code VCVTPD2DQ xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.F2.0F.W1 E6 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2DQ_XMM_K1Z_XMMM128B64 = 2390;
	/**
	 * {@code VCVTPD2DQ xmm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.F2.0F.W1 E6 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2DQ_XMM_K1Z_YMMM256B64 = 2391;
	/**
	 * {@code VCVTPD2DQ ymm1 {k1}{z}, zmm2/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.F2.0F.W1 E6 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2DQ_YMM_K1Z_ZMMM512B64_ER = 2392;
	/**
	 * {@code MOVNTQ m64, mm}<p>
	 * {@code NP 0F E7 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVNTQ_M64_MM = 2393;
	/**
	 * {@code MOVNTDQ m128, xmm1}<p>
	 * {@code 66 0F E7 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVNTDQ_M128_XMM = 2394;
	/**
	 * {@code VMOVNTDQ m128, xmm1}<p>
	 * {@code VEX.128.66.0F.WIG E7 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVNTDQ_M128_XMM = 2395;
	/**
	 * {@code VMOVNTDQ m256, ymm1}<p>
	 * {@code VEX.256.66.0F.WIG E7 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVNTDQ_M256_YMM = 2396;
	/**
	 * {@code VMOVNTDQ m128, xmm1}<p>
	 * {@code EVEX.128.66.0F.W0 E7 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVNTDQ_M128_XMM = 2397;
	/**
	 * {@code VMOVNTDQ m256, ymm1}<p>
	 * {@code EVEX.256.66.0F.W0 E7 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVNTDQ_M256_YMM = 2398;
	/**
	 * {@code VMOVNTDQ m512, zmm1}<p>
	 * {@code EVEX.512.66.0F.W0 E7 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVNTDQ_M512_ZMM = 2399;
	/**
	 * {@code PSUBSB mm, mm/m64}<p>
	 * {@code NP 0F E8 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBSB_MM_MMM64 = 2400;
	/**
	 * {@code PSUBSB xmm1, xmm2/m128}<p>
	 * {@code 66 0F E8 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBSB_XMM_XMMM128 = 2401;
	/**
	 * {@code VPSUBSB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG E8 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBSB_XMM_XMM_XMMM128 = 2402;
	/**
	 * {@code VPSUBSB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG E8 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBSB_YMM_YMM_YMMM256 = 2403;
	/**
	 * {@code VPSUBSB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG E8 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBSB_XMM_K1Z_XMM_XMMM128 = 2404;
	/**
	 * {@code VPSUBSB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG E8 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBSB_YMM_K1Z_YMM_YMMM256 = 2405;
	/**
	 * {@code VPSUBSB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG E8 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBSB_ZMM_K1Z_ZMM_ZMMM512 = 2406;
	/**
	 * {@code PSUBSW mm, mm/m64}<p>
	 * {@code NP 0F E9 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBSW_MM_MMM64 = 2407;
	/**
	 * {@code PSUBSW xmm1, xmm2/m128}<p>
	 * {@code 66 0F E9 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBSW_XMM_XMMM128 = 2408;
	/**
	 * {@code VPSUBSW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG E9 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBSW_XMM_XMM_XMMM128 = 2409;
	/**
	 * {@code VPSUBSW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG E9 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBSW_YMM_YMM_YMMM256 = 2410;
	/**
	 * {@code VPSUBSW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG E9 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBSW_XMM_K1Z_XMM_XMMM128 = 2411;
	/**
	 * {@code VPSUBSW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG E9 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBSW_YMM_K1Z_YMM_YMMM256 = 2412;
	/**
	 * {@code VPSUBSW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG E9 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBSW_ZMM_K1Z_ZMM_ZMMM512 = 2413;
	/**
	 * {@code PMINSW mm1, mm2/m64}<p>
	 * {@code NP 0F EA /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMINSW_MM_MMM64 = 2414;
	/**
	 * {@code PMINSW xmm1, xmm2/m128}<p>
	 * {@code 66 0F EA /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMINSW_XMM_XMMM128 = 2415;
	/**
	 * {@code VPMINSW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG EA /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMINSW_XMM_XMM_XMMM128 = 2416;
	/**
	 * {@code VPMINSW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG EA /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMINSW_YMM_YMM_YMMM256 = 2417;
	/**
	 * {@code VPMINSW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG EA /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINSW_XMM_K1Z_XMM_XMMM128 = 2418;
	/**
	 * {@code VPMINSW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG EA /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINSW_YMM_K1Z_YMM_YMMM256 = 2419;
	/**
	 * {@code VPMINSW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG EA /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINSW_ZMM_K1Z_ZMM_ZMMM512 = 2420;
	/**
	 * {@code POR mm, mm/m64}<p>
	 * {@code NP 0F EB /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int POR_MM_MMM64 = 2421;
	/**
	 * {@code POR xmm1, xmm2/m128}<p>
	 * {@code 66 0F EB /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int POR_XMM_XMMM128 = 2422;
	/**
	 * {@code VPOR xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG EB /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPOR_XMM_XMM_XMMM128 = 2423;
	/**
	 * {@code VPOR ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG EB /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPOR_YMM_YMM_YMMM256 = 2424;
	/**
	 * {@code VPORD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 EB /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPORD_XMM_K1Z_XMM_XMMM128B32 = 2425;
	/**
	 * {@code VPORD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 EB /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPORD_YMM_K1Z_YMM_YMMM256B32 = 2426;
	/**
	 * {@code VPORD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F.W0 EB /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPORD_ZMM_K1Z_ZMM_ZMMM512B32 = 2427;
	/**
	 * {@code VPORQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 EB /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPORQ_XMM_K1Z_XMM_XMMM128B64 = 2428;
	/**
	 * {@code VPORQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 EB /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPORQ_YMM_K1Z_YMM_YMMM256B64 = 2429;
	/**
	 * {@code VPORQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F.W1 EB /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPORQ_ZMM_K1Z_ZMM_ZMMM512B64 = 2430;
	/**
	 * {@code PADDSB mm, mm/m64}<p>
	 * {@code NP 0F EC /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDSB_MM_MMM64 = 2431;
	/**
	 * {@code PADDSB xmm1, xmm2/m128}<p>
	 * {@code 66 0F EC /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDSB_XMM_XMMM128 = 2432;
	/**
	 * {@code VPADDSB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG EC /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDSB_XMM_XMM_XMMM128 = 2433;
	/**
	 * {@code VPADDSB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG EC /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDSB_YMM_YMM_YMMM256 = 2434;
	/**
	 * {@code VPADDSB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG EC /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDSB_XMM_K1Z_XMM_XMMM128 = 2435;
	/**
	 * {@code VPADDSB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG EC /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDSB_YMM_K1Z_YMM_YMMM256 = 2436;
	/**
	 * {@code VPADDSB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG EC /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDSB_ZMM_K1Z_ZMM_ZMMM512 = 2437;
	/**
	 * {@code PADDSW mm, mm/m64}<p>
	 * {@code NP 0F ED /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDSW_MM_MMM64 = 2438;
	/**
	 * {@code PADDSW xmm1, xmm2/m128}<p>
	 * {@code 66 0F ED /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDSW_XMM_XMMM128 = 2439;
	/**
	 * {@code VPADDSW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG ED /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDSW_XMM_XMM_XMMM128 = 2440;
	/**
	 * {@code VPADDSW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG ED /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDSW_YMM_YMM_YMMM256 = 2441;
	/**
	 * {@code VPADDSW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG ED /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDSW_XMM_K1Z_XMM_XMMM128 = 2442;
	/**
	 * {@code VPADDSW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG ED /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDSW_YMM_K1Z_YMM_YMMM256 = 2443;
	/**
	 * {@code VPADDSW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG ED /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDSW_ZMM_K1Z_ZMM_ZMMM512 = 2444;
	/**
	 * {@code PMAXSW mm1, mm2/m64}<p>
	 * {@code NP 0F EE /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMAXSW_MM_MMM64 = 2445;
	/**
	 * {@code PMAXSW xmm1, xmm2/m128}<p>
	 * {@code 66 0F EE /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMAXSW_XMM_XMMM128 = 2446;
	/**
	 * {@code VPMAXSW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG EE /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMAXSW_XMM_XMM_XMMM128 = 2447;
	/**
	 * {@code VPMAXSW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG EE /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMAXSW_YMM_YMM_YMMM256 = 2448;
	/**
	 * {@code VPMAXSW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG EE /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXSW_XMM_K1Z_XMM_XMMM128 = 2449;
	/**
	 * {@code VPMAXSW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG EE /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXSW_YMM_K1Z_YMM_YMMM256 = 2450;
	/**
	 * {@code VPMAXSW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG EE /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXSW_ZMM_K1Z_ZMM_ZMMM512 = 2451;
	/**
	 * {@code PXOR mm, mm/m64}<p>
	 * {@code NP 0F EF /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PXOR_MM_MMM64 = 2452;
	/**
	 * {@code PXOR xmm1, xmm2/m128}<p>
	 * {@code 66 0F EF /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PXOR_XMM_XMMM128 = 2453;
	/**
	 * {@code VPXOR xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG EF /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPXOR_XMM_XMM_XMMM128 = 2454;
	/**
	 * {@code VPXOR ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG EF /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPXOR_YMM_YMM_YMMM256 = 2455;
	/**
	 * {@code VPXORD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 EF /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPXORD_XMM_K1Z_XMM_XMMM128B32 = 2456;
	/**
	 * {@code VPXORD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 EF /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPXORD_YMM_K1Z_YMM_YMMM256B32 = 2457;
	/**
	 * {@code VPXORD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F.W0 EF /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPXORD_ZMM_K1Z_ZMM_ZMMM512B32 = 2458;
	/**
	 * {@code VPXORQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 EF /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPXORQ_XMM_K1Z_XMM_XMMM128B64 = 2459;
	/**
	 * {@code VPXORQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 EF /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPXORQ_YMM_K1Z_YMM_YMMM256B64 = 2460;
	/**
	 * {@code VPXORQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F.W1 EF /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPXORQ_ZMM_K1Z_ZMM_ZMMM512B64 = 2461;
	/**
	 * {@code LDDQU xmm1, m128}<p>
	 * {@code F2 0F F0 /r}<p>
	 * {@code SSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LDDQU_XMM_M128 = 2462;
	/**
	 * {@code VLDDQU xmm1, m128}<p>
	 * {@code VEX.128.F2.0F.WIG F0 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VLDDQU_XMM_M128 = 2463;
	/**
	 * {@code VLDDQU ymm1, m256}<p>
	 * {@code VEX.256.F2.0F.WIG F0 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VLDDQU_YMM_M256 = 2464;
	/**
	 * {@code PSLLW mm, mm/m64}<p>
	 * {@code NP 0F F1 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSLLW_MM_MMM64 = 2465;
	/**
	 * {@code PSLLW xmm1, xmm2/m128}<p>
	 * {@code 66 0F F1 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSLLW_XMM_XMMM128 = 2466;
	/**
	 * {@code VPSLLW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG F1 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLW_XMM_XMM_XMMM128 = 2467;
	/**
	 * {@code VPSLLW ymm1, ymm2, xmm3/m128}<p>
	 * {@code VEX.256.66.0F.WIG F1 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLW_YMM_YMM_XMMM128 = 2468;
	/**
	 * {@code VPSLLW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG F1 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLW_XMM_K1Z_XMM_XMMM128 = 2469;
	/**
	 * {@code VPSLLW ymm1 {k1}{z}, ymm2, xmm3/m128}<p>
	 * {@code EVEX.256.66.0F.WIG F1 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLW_YMM_K1Z_YMM_XMMM128 = 2470;
	/**
	 * {@code VPSLLW zmm1 {k1}{z}, zmm2, xmm3/m128}<p>
	 * {@code EVEX.512.66.0F.WIG F1 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLW_ZMM_K1Z_ZMM_XMMM128 = 2471;
	/**
	 * {@code PSLLD mm, mm/m64}<p>
	 * {@code NP 0F F2 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSLLD_MM_MMM64 = 2472;
	/**
	 * {@code PSLLD xmm1, xmm2/m128}<p>
	 * {@code 66 0F F2 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSLLD_XMM_XMMM128 = 2473;
	/**
	 * {@code VPSLLD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG F2 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLD_XMM_XMM_XMMM128 = 2474;
	/**
	 * {@code VPSLLD ymm1, ymm2, xmm3/m128}<p>
	 * {@code VEX.256.66.0F.WIG F2 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLD_YMM_YMM_XMMM128 = 2475;
	/**
	 * {@code VPSLLD xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.W0 F2 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLD_XMM_K1Z_XMM_XMMM128 = 2476;
	/**
	 * {@code VPSLLD ymm1 {k1}{z}, ymm2, xmm3/m128}<p>
	 * {@code EVEX.256.66.0F.W0 F2 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLD_YMM_K1Z_YMM_XMMM128 = 2477;
	/**
	 * {@code VPSLLD zmm1 {k1}{z}, zmm2, xmm3/m128}<p>
	 * {@code EVEX.512.66.0F.W0 F2 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLD_ZMM_K1Z_ZMM_XMMM128 = 2478;
	/**
	 * {@code PSLLQ mm, mm/m64}<p>
	 * {@code NP 0F F3 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSLLQ_MM_MMM64 = 2479;
	/**
	 * {@code PSLLQ xmm1, xmm2/m128}<p>
	 * {@code 66 0F F3 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSLLQ_XMM_XMMM128 = 2480;
	/**
	 * {@code VPSLLQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG F3 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLQ_XMM_XMM_XMMM128 = 2481;
	/**
	 * {@code VPSLLQ ymm1, ymm2, xmm3/m128}<p>
	 * {@code VEX.256.66.0F.WIG F3 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLQ_YMM_YMM_XMMM128 = 2482;
	/**
	 * {@code VPSLLQ xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.W1 F3 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLQ_XMM_K1Z_XMM_XMMM128 = 2483;
	/**
	 * {@code VPSLLQ ymm1 {k1}{z}, ymm2, xmm3/m128}<p>
	 * {@code EVEX.256.66.0F.W1 F3 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLQ_YMM_K1Z_YMM_XMMM128 = 2484;
	/**
	 * {@code VPSLLQ zmm1 {k1}{z}, zmm2, xmm3/m128}<p>
	 * {@code EVEX.512.66.0F.W1 F3 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLQ_ZMM_K1Z_ZMM_XMMM128 = 2485;
	/**
	 * {@code PMULUDQ mm1, mm2/m64}<p>
	 * {@code NP 0F F4 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMULUDQ_MM_MMM64 = 2486;
	/**
	 * {@code PMULUDQ xmm1, xmm2/m128}<p>
	 * {@code 66 0F F4 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMULUDQ_XMM_XMMM128 = 2487;
	/**
	 * {@code VPMULUDQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG F4 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMULUDQ_XMM_XMM_XMMM128 = 2488;
	/**
	 * {@code VPMULUDQ ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG F4 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMULUDQ_YMM_YMM_YMMM256 = 2489;
	/**
	 * {@code VPMULUDQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 F4 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULUDQ_XMM_K1Z_XMM_XMMM128B64 = 2490;
	/**
	 * {@code VPMULUDQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 F4 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULUDQ_YMM_K1Z_YMM_YMMM256B64 = 2491;
	/**
	 * {@code VPMULUDQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F.W1 F4 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULUDQ_ZMM_K1Z_ZMM_ZMMM512B64 = 2492;
	/**
	 * {@code PMADDWD mm, mm/m64}<p>
	 * {@code NP 0F F5 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMADDWD_MM_MMM64 = 2493;
	/**
	 * {@code PMADDWD xmm1, xmm2/m128}<p>
	 * {@code 66 0F F5 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMADDWD_XMM_XMMM128 = 2494;
	/**
	 * {@code VPMADDWD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG F5 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMADDWD_XMM_XMM_XMMM128 = 2495;
	/**
	 * {@code VPMADDWD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG F5 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMADDWD_YMM_YMM_YMMM256 = 2496;
	/**
	 * {@code VPMADDWD xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG F5 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMADDWD_XMM_K1Z_XMM_XMMM128 = 2497;
	/**
	 * {@code VPMADDWD ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG F5 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMADDWD_YMM_K1Z_YMM_YMMM256 = 2498;
	/**
	 * {@code VPMADDWD zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG F5 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMADDWD_ZMM_K1Z_ZMM_ZMMM512 = 2499;
	/**
	 * {@code PSADBW mm1, mm2/m64}<p>
	 * {@code NP 0F F6 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSADBW_MM_MMM64 = 2500;
	/**
	 * {@code PSADBW xmm1, xmm2/m128}<p>
	 * {@code 66 0F F6 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSADBW_XMM_XMMM128 = 2501;
	/**
	 * {@code VPSADBW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG F6 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSADBW_XMM_XMM_XMMM128 = 2502;
	/**
	 * {@code VPSADBW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG F6 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSADBW_YMM_YMM_YMMM256 = 2503;
	/**
	 * {@code VPSADBW xmm1, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG F6 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSADBW_XMM_XMM_XMMM128 = 2504;
	/**
	 * {@code VPSADBW ymm1, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG F6 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSADBW_YMM_YMM_YMMM256 = 2505;
	/**
	 * {@code VPSADBW zmm1, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG F6 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSADBW_ZMM_ZMM_ZMMM512 = 2506;
	/**
	 * {@code MASKMOVQ mm1, mm2}<p>
	 * {@code NP 0F F7 /r}<p>
	 * {@code SSE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MASKMOVQ_RDI_MM_MM = 2507;
	/**
	 * {@code MASKMOVDQU xmm1, xmm2}<p>
	 * {@code 66 0F F7 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MASKMOVDQU_RDI_XMM_XMM = 2508;
	/**
	 * {@code VMASKMOVDQU xmm1, xmm2}<p>
	 * {@code VEX.128.66.0F.WIG F7 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMASKMOVDQU_RDI_XMM_XMM = 2509;
	/**
	 * {@code PSUBB mm, mm/m64}<p>
	 * {@code NP 0F F8 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBB_MM_MMM64 = 2510;
	/**
	 * {@code PSUBB xmm1, xmm2/m128}<p>
	 * {@code 66 0F F8 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBB_XMM_XMMM128 = 2511;
	/**
	 * {@code VPSUBB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG F8 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBB_XMM_XMM_XMMM128 = 2512;
	/**
	 * {@code VPSUBB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG F8 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBB_YMM_YMM_YMMM256 = 2513;
	/**
	 * {@code VPSUBB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG F8 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBB_XMM_K1Z_XMM_XMMM128 = 2514;
	/**
	 * {@code VPSUBB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG F8 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBB_YMM_K1Z_YMM_YMMM256 = 2515;
	/**
	 * {@code VPSUBB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG F8 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBB_ZMM_K1Z_ZMM_ZMMM512 = 2516;
	/**
	 * {@code PSUBW mm, mm/m64}<p>
	 * {@code NP 0F F9 /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBW_MM_MMM64 = 2517;
	/**
	 * {@code PSUBW xmm1, xmm2/m128}<p>
	 * {@code 66 0F F9 /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBW_XMM_XMMM128 = 2518;
	/**
	 * {@code VPSUBW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG F9 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBW_XMM_XMM_XMMM128 = 2519;
	/**
	 * {@code VPSUBW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG F9 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBW_YMM_YMM_YMMM256 = 2520;
	/**
	 * {@code VPSUBW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG F9 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBW_XMM_K1Z_XMM_XMMM128 = 2521;
	/**
	 * {@code VPSUBW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG F9 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBW_YMM_K1Z_YMM_YMMM256 = 2522;
	/**
	 * {@code VPSUBW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG F9 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBW_ZMM_K1Z_ZMM_ZMMM512 = 2523;
	/**
	 * {@code PSUBD mm, mm/m64}<p>
	 * {@code NP 0F FA /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBD_MM_MMM64 = 2524;
	/**
	 * {@code PSUBD xmm1, xmm2/m128}<p>
	 * {@code 66 0F FA /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBD_XMM_XMMM128 = 2525;
	/**
	 * {@code VPSUBD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG FA /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBD_XMM_XMM_XMMM128 = 2526;
	/**
	 * {@code VPSUBD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG FA /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBD_YMM_YMM_YMMM256 = 2527;
	/**
	 * {@code VPSUBD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 FA /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBD_XMM_K1Z_XMM_XMMM128B32 = 2528;
	/**
	 * {@code VPSUBD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 FA /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBD_YMM_K1Z_YMM_YMMM256B32 = 2529;
	/**
	 * {@code VPSUBD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F.W0 FA /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBD_ZMM_K1Z_ZMM_ZMMM512B32 = 2530;
	/**
	 * {@code PSUBQ mm1, mm2/m64}<p>
	 * {@code NP 0F FB /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBQ_MM_MMM64 = 2531;
	/**
	 * {@code PSUBQ xmm1, xmm2/m128}<p>
	 * {@code 66 0F FB /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSUBQ_XMM_XMMM128 = 2532;
	/**
	 * {@code VPSUBQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG FB /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBQ_XMM_XMM_XMMM128 = 2533;
	/**
	 * {@code VPSUBQ ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG FB /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSUBQ_YMM_YMM_YMMM256 = 2534;
	/**
	 * {@code VPSUBQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F.W1 FB /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBQ_XMM_K1Z_XMM_XMMM128B64 = 2535;
	/**
	 * {@code VPSUBQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F.W1 FB /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBQ_YMM_K1Z_YMM_YMMM256B64 = 2536;
	/**
	 * {@code VPSUBQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F.W1 FB /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSUBQ_ZMM_K1Z_ZMM_ZMMM512B64 = 2537;
	/**
	 * {@code PADDB mm, mm/m64}<p>
	 * {@code NP 0F FC /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDB_MM_MMM64 = 2538;
	/**
	 * {@code PADDB xmm1, xmm2/m128}<p>
	 * {@code 66 0F FC /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDB_XMM_XMMM128 = 2539;
	/**
	 * {@code VPADDB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG FC /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDB_XMM_XMM_XMMM128 = 2540;
	/**
	 * {@code VPADDB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG FC /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDB_YMM_YMM_YMMM256 = 2541;
	/**
	 * {@code VPADDB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG FC /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDB_XMM_K1Z_XMM_XMMM128 = 2542;
	/**
	 * {@code VPADDB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG FC /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDB_YMM_K1Z_YMM_YMMM256 = 2543;
	/**
	 * {@code VPADDB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG FC /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDB_ZMM_K1Z_ZMM_ZMMM512 = 2544;
	/**
	 * {@code PADDW mm, mm/m64}<p>
	 * {@code NP 0F FD /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDW_MM_MMM64 = 2545;
	/**
	 * {@code PADDW xmm1, xmm2/m128}<p>
	 * {@code 66 0F FD /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDW_XMM_XMMM128 = 2546;
	/**
	 * {@code VPADDW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG FD /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDW_XMM_XMM_XMMM128 = 2547;
	/**
	 * {@code VPADDW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG FD /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDW_YMM_YMM_YMMM256 = 2548;
	/**
	 * {@code VPADDW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F.WIG FD /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDW_XMM_K1Z_XMM_XMMM128 = 2549;
	/**
	 * {@code VPADDW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F.WIG FD /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDW_YMM_K1Z_YMM_YMMM256 = 2550;
	/**
	 * {@code VPADDW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F.WIG FD /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDW_ZMM_K1Z_ZMM_ZMMM512 = 2551;
	/**
	 * {@code PADDD mm, mm/m64}<p>
	 * {@code NP 0F FE /r}<p>
	 * {@code MMX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDD_MM_MMM64 = 2552;
	/**
	 * {@code PADDD xmm1, xmm2/m128}<p>
	 * {@code 66 0F FE /r}<p>
	 * {@code SSE2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PADDD_XMM_XMMM128 = 2553;
	/**
	 * {@code VPADDD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F.WIG FE /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDD_XMM_XMM_XMMM128 = 2554;
	/**
	 * {@code VPADDD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F.WIG FE /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPADDD_YMM_YMM_YMMM256 = 2555;
	/**
	 * {@code VPADDD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F.W0 FE /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDD_XMM_K1Z_XMM_XMMM128B32 = 2556;
	/**
	 * {@code VPADDD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F.W0 FE /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDD_YMM_K1Z_YMM_YMMM256B32 = 2557;
	/**
	 * {@code VPADDD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F.W0 FE /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPADDD_ZMM_K1Z_ZMM_ZMMM512B32 = 2558;
	/**
	 * {@code UD0 r16, r/m16}<p>
	 * {@code o16 0F FF /r}<p>
	 * {@code 286+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int UD0_R16_RM16 = 2559;
	/**
	 * {@code UD0 r32, r/m32}<p>
	 * {@code o32 0F FF /r}<p>
	 * {@code 386+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int UD0_R32_RM32 = 2560;
	/**
	 * {@code UD0 r64, r/m64}<p>
	 * {@code o64 0F FF /r}<p>
	 * {@code X64}<p>
	 * {@code 64-bit}
	 */
	public static final int UD0_R64_RM64 = 2561;
	/**
	 * {@code PSHUFB mm1, mm2/m64}<p>
	 * {@code NP 0F 38 00 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSHUFB_MM_MMM64 = 2562;
	/**
	 * {@code PSHUFB xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 00 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSHUFB_XMM_XMMM128 = 2563;
	/**
	 * {@code VPSHUFB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 00 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSHUFB_XMM_XMM_XMMM128 = 2564;
	/**
	 * {@code VPSHUFB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 00 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSHUFB_YMM_YMM_YMMM256 = 2565;
	/**
	 * {@code VPSHUFB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.WIG 00 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHUFB_XMM_K1Z_XMM_XMMM128 = 2566;
	/**
	 * {@code VPSHUFB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.WIG 00 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHUFB_YMM_K1Z_YMM_YMMM256 = 2567;
	/**
	 * {@code VPSHUFB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.WIG 00 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHUFB_ZMM_K1Z_ZMM_ZMMM512 = 2568;
	/**
	 * {@code PHADDW mm1, mm2/m64}<p>
	 * {@code NP 0F 38 01 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PHADDW_MM_MMM64 = 2569;
	/**
	 * {@code PHADDW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 01 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PHADDW_XMM_XMMM128 = 2570;
	/**
	 * {@code VPHADDW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 01 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPHADDW_XMM_XMM_XMMM128 = 2571;
	/**
	 * {@code VPHADDW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 01 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPHADDW_YMM_YMM_YMMM256 = 2572;
	/**
	 * {@code PHADDD mm1, mm2/m64}<p>
	 * {@code NP 0F 38 02 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PHADDD_MM_MMM64 = 2573;
	/**
	 * {@code PHADDD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 02 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PHADDD_XMM_XMMM128 = 2574;
	/**
	 * {@code VPHADDD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 02 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPHADDD_XMM_XMM_XMMM128 = 2575;
	/**
	 * {@code VPHADDD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 02 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPHADDD_YMM_YMM_YMMM256 = 2576;
	/**
	 * {@code PHADDSW mm1, mm2/m64}<p>
	 * {@code NP 0F 38 03 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PHADDSW_MM_MMM64 = 2577;
	/**
	 * {@code PHADDSW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 03 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PHADDSW_XMM_XMMM128 = 2578;
	/**
	 * {@code VPHADDSW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 03 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPHADDSW_XMM_XMM_XMMM128 = 2579;
	/**
	 * {@code VPHADDSW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 03 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPHADDSW_YMM_YMM_YMMM256 = 2580;
	/**
	 * {@code PMADDUBSW mm1, mm2/m64}<p>
	 * {@code NP 0F 38 04 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMADDUBSW_MM_MMM64 = 2581;
	/**
	 * {@code PMADDUBSW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 04 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMADDUBSW_XMM_XMMM128 = 2582;
	/**
	 * {@code VPMADDUBSW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 04 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMADDUBSW_XMM_XMM_XMMM128 = 2583;
	/**
	 * {@code VPMADDUBSW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 04 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMADDUBSW_YMM_YMM_YMMM256 = 2584;
	/**
	 * {@code VPMADDUBSW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.WIG 04 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMADDUBSW_XMM_K1Z_XMM_XMMM128 = 2585;
	/**
	 * {@code VPMADDUBSW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.WIG 04 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMADDUBSW_YMM_K1Z_YMM_YMMM256 = 2586;
	/**
	 * {@code VPMADDUBSW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.WIG 04 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMADDUBSW_ZMM_K1Z_ZMM_ZMMM512 = 2587;
	/**
	 * {@code PHSUBW mm1, mm2/m64}<p>
	 * {@code NP 0F 38 05 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PHSUBW_MM_MMM64 = 2588;
	/**
	 * {@code PHSUBW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 05 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PHSUBW_XMM_XMMM128 = 2589;
	/**
	 * {@code VPHSUBW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 05 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPHSUBW_XMM_XMM_XMMM128 = 2590;
	/**
	 * {@code VPHSUBW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 05 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPHSUBW_YMM_YMM_YMMM256 = 2591;
	/**
	 * {@code PHSUBD mm1, mm2/m64}<p>
	 * {@code NP 0F 38 06 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PHSUBD_MM_MMM64 = 2592;
	/**
	 * {@code PHSUBD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 06 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PHSUBD_XMM_XMMM128 = 2593;
	/**
	 * {@code VPHSUBD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 06 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPHSUBD_XMM_XMM_XMMM128 = 2594;
	/**
	 * {@code VPHSUBD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 06 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPHSUBD_YMM_YMM_YMMM256 = 2595;
	/**
	 * {@code PHSUBSW mm1, mm2/m64}<p>
	 * {@code NP 0F 38 07 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PHSUBSW_MM_MMM64 = 2596;
	/**
	 * {@code PHSUBSW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 07 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PHSUBSW_XMM_XMMM128 = 2597;
	/**
	 * {@code VPHSUBSW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 07 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPHSUBSW_XMM_XMM_XMMM128 = 2598;
	/**
	 * {@code VPHSUBSW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 07 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPHSUBSW_YMM_YMM_YMMM256 = 2599;
	/**
	 * {@code PSIGNB mm1, mm2/m64}<p>
	 * {@code NP 0F 38 08 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSIGNB_MM_MMM64 = 2600;
	/**
	 * {@code PSIGNB xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 08 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSIGNB_XMM_XMMM128 = 2601;
	/**
	 * {@code VPSIGNB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 08 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSIGNB_XMM_XMM_XMMM128 = 2602;
	/**
	 * {@code VPSIGNB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 08 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSIGNB_YMM_YMM_YMMM256 = 2603;
	/**
	 * {@code PSIGNW mm1, mm2/m64}<p>
	 * {@code NP 0F 38 09 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSIGNW_MM_MMM64 = 2604;
	/**
	 * {@code PSIGNW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 09 /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSIGNW_XMM_XMMM128 = 2605;
	/**
	 * {@code VPSIGNW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 09 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSIGNW_XMM_XMM_XMMM128 = 2606;
	/**
	 * {@code VPSIGNW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 09 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSIGNW_YMM_YMM_YMMM256 = 2607;
	/**
	 * {@code PSIGND mm1, mm2/m64}<p>
	 * {@code NP 0F 38 0A /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSIGND_MM_MMM64 = 2608;
	/**
	 * {@code PSIGND xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 0A /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PSIGND_XMM_XMMM128 = 2609;
	/**
	 * {@code VPSIGND xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 0A /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSIGND_XMM_XMM_XMMM128 = 2610;
	/**
	 * {@code VPSIGND ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 0A /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSIGND_YMM_YMM_YMMM256 = 2611;
	/**
	 * {@code PMULHRSW mm1, mm2/m64}<p>
	 * {@code NP 0F 38 0B /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMULHRSW_MM_MMM64 = 2612;
	/**
	 * {@code PMULHRSW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 0B /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMULHRSW_XMM_XMMM128 = 2613;
	/**
	 * {@code VPMULHRSW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 0B /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMULHRSW_XMM_XMM_XMMM128 = 2614;
	/**
	 * {@code VPMULHRSW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 0B /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMULHRSW_YMM_YMM_YMMM256 = 2615;
	/**
	 * {@code VPMULHRSW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.WIG 0B /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULHRSW_XMM_K1Z_XMM_XMMM128 = 2616;
	/**
	 * {@code VPMULHRSW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.WIG 0B /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULHRSW_YMM_K1Z_YMM_YMMM256 = 2617;
	/**
	 * {@code VPMULHRSW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.WIG 0B /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULHRSW_ZMM_K1Z_ZMM_ZMMM512 = 2618;
	/**
	 * {@code VPERMILPS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 0C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMILPS_XMM_XMM_XMMM128 = 2619;
	/**
	 * {@code VPERMILPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 0C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMILPS_YMM_YMM_YMMM256 = 2620;
	/**
	 * {@code VPERMILPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 0C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMILPS_XMM_K1Z_XMM_XMMM128B32 = 2621;
	/**
	 * {@code VPERMILPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 0C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMILPS_YMM_K1Z_YMM_YMMM256B32 = 2622;
	/**
	 * {@code VPERMILPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 0C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMILPS_ZMM_K1Z_ZMM_ZMMM512B32 = 2623;
	/**
	 * {@code VPERMILPD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 0D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMILPD_XMM_XMM_XMMM128 = 2624;
	/**
	 * {@code VPERMILPD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 0D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMILPD_YMM_YMM_YMMM256 = 2625;
	/**
	 * {@code VPERMILPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 0D /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMILPD_XMM_K1Z_XMM_XMMM128B64 = 2626;
	/**
	 * {@code VPERMILPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 0D /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMILPD_YMM_K1Z_YMM_YMMM256B64 = 2627;
	/**
	 * {@code VPERMILPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 0D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMILPD_ZMM_K1Z_ZMM_ZMMM512B64 = 2628;
	/**
	 * {@code VTESTPS xmm1, xmm2/m128}<p>
	 * {@code VEX.128.66.0F38.W0 0E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VTESTPS_XMM_XMMM128 = 2629;
	/**
	 * {@code VTESTPS ymm1, ymm2/m256}<p>
	 * {@code VEX.256.66.0F38.W0 0E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VTESTPS_YMM_YMMM256 = 2630;
	/**
	 * {@code VTESTPD xmm1, xmm2/m128}<p>
	 * {@code VEX.128.66.0F38.W0 0F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VTESTPD_XMM_XMMM128 = 2631;
	/**
	 * {@code VTESTPD ymm1, ymm2/m256}<p>
	 * {@code VEX.256.66.0F38.W0 0F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VTESTPD_YMM_YMMM256 = 2632;
	/**
	 * {@code PBLENDVB xmm1, xmm2/m128, &lt;XMM0&gt;}<p>
	 * {@code 66 0F 38 10 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PBLENDVB_XMM_XMMM128 = 2633;
	/**
	 * {@code VPSRLVW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W1 10 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLVW_XMM_K1Z_XMM_XMMM128 = 2634;
	/**
	 * {@code VPSRLVW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W1 10 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLVW_YMM_K1Z_YMM_YMMM256 = 2635;
	/**
	 * {@code VPSRLVW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W1 10 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLVW_ZMM_K1Z_ZMM_ZMMM512 = 2636;
	/**
	 * {@code VPMOVUSWB xmm1/m64 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 10 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSWB_XMMM64_K1Z_XMM = 2637;
	/**
	 * {@code VPMOVUSWB xmm1/m128 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 10 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSWB_XMMM128_K1Z_YMM = 2638;
	/**
	 * {@code VPMOVUSWB ymm1/m256 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 10 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSWB_YMMM256_K1Z_ZMM = 2639;
	/**
	 * {@code VPSRAVW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W1 11 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAVW_XMM_K1Z_XMM_XMMM128 = 2640;
	/**
	 * {@code VPSRAVW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W1 11 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAVW_YMM_K1Z_YMM_YMMM256 = 2641;
	/**
	 * {@code VPSRAVW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W1 11 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAVW_ZMM_K1Z_ZMM_ZMMM512 = 2642;
	/**
	 * {@code VPMOVUSDB xmm1/m32 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 11 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSDB_XMMM32_K1Z_XMM = 2643;
	/**
	 * {@code VPMOVUSDB xmm1/m64 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 11 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSDB_XMMM64_K1Z_YMM = 2644;
	/**
	 * {@code VPMOVUSDB xmm1/m128 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 11 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSDB_XMMM128_K1Z_ZMM = 2645;
	/**
	 * {@code VPSLLVW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W1 12 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLVW_XMM_K1Z_XMM_XMMM128 = 2646;
	/**
	 * {@code VPSLLVW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W1 12 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLVW_YMM_K1Z_YMM_YMMM256 = 2647;
	/**
	 * {@code VPSLLVW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W1 12 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLVW_ZMM_K1Z_ZMM_ZMMM512 = 2648;
	/**
	 * {@code VPMOVUSQB xmm1/m16 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 12 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSQB_XMMM16_K1Z_XMM = 2649;
	/**
	 * {@code VPMOVUSQB xmm1/m32 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 12 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSQB_XMMM32_K1Z_YMM = 2650;
	/**
	 * {@code VPMOVUSQB xmm1/m64 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 12 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSQB_XMMM64_K1Z_ZMM = 2651;
	/**
	 * {@code VCVTPH2PS xmm1, xmm2/m64}<p>
	 * {@code VEX.128.66.0F38.W0 13 /r}<p>
	 * {@code F16C}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTPH2PS_XMM_XMMM64 = 2652;
	/**
	 * {@code VCVTPH2PS ymm1, xmm2/m128}<p>
	 * {@code VEX.256.66.0F38.W0 13 /r}<p>
	 * {@code F16C}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTPH2PS_YMM_XMMM128 = 2653;
	/**
	 * {@code VCVTPH2PS xmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.128.66.0F38.W0 13 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2PS_XMM_K1Z_XMMM64 = 2654;
	/**
	 * {@code VCVTPH2PS ymm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.256.66.0F38.W0 13 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2PS_YMM_K1Z_XMMM128 = 2655;
	/**
	 * {@code VCVTPH2PS zmm1 {k1}{z}, ymm2/m256{sae}}<p>
	 * {@code EVEX.512.66.0F38.W0 13 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2PS_ZMM_K1Z_YMMM256_SAE = 2656;
	/**
	 * {@code VPMOVUSDW xmm1/m64 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 13 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSDW_XMMM64_K1Z_XMM = 2657;
	/**
	 * {@code VPMOVUSDW xmm1/m128 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 13 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSDW_XMMM128_K1Z_YMM = 2658;
	/**
	 * {@code VPMOVUSDW ymm1/m256 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 13 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSDW_YMMM256_K1Z_ZMM = 2659;
	/**
	 * {@code BLENDVPS xmm1, xmm2/m128, &lt;XMM0&gt;}<p>
	 * {@code 66 0F 38 14 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BLENDVPS_XMM_XMMM128 = 2660;
	/**
	 * {@code VPRORVD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 14 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPRORVD_XMM_K1Z_XMM_XMMM128B32 = 2661;
	/**
	 * {@code VPRORVD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 14 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPRORVD_YMM_K1Z_YMM_YMMM256B32 = 2662;
	/**
	 * {@code VPRORVD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 14 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPRORVD_ZMM_K1Z_ZMM_ZMMM512B32 = 2663;
	/**
	 * {@code VPRORVQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 14 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPRORVQ_XMM_K1Z_XMM_XMMM128B64 = 2664;
	/**
	 * {@code VPRORVQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 14 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPRORVQ_YMM_K1Z_YMM_YMMM256B64 = 2665;
	/**
	 * {@code VPRORVQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 14 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPRORVQ_ZMM_K1Z_ZMM_ZMMM512B64 = 2666;
	/**
	 * {@code VPMOVUSQW xmm1/m32 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 14 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSQW_XMMM32_K1Z_XMM = 2667;
	/**
	 * {@code VPMOVUSQW xmm1/m64 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 14 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSQW_XMMM64_K1Z_YMM = 2668;
	/**
	 * {@code VPMOVUSQW xmm1/m128 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 14 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSQW_XMMM128_K1Z_ZMM = 2669;
	/**
	 * {@code BLENDVPD xmm1, xmm2/m128, &lt;XMM0&gt;}<p>
	 * {@code 66 0F 38 15 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BLENDVPD_XMM_XMMM128 = 2670;
	/**
	 * {@code VPROLVD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 15 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPROLVD_XMM_K1Z_XMM_XMMM128B32 = 2671;
	/**
	 * {@code VPROLVD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 15 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPROLVD_YMM_K1Z_YMM_YMMM256B32 = 2672;
	/**
	 * {@code VPROLVD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 15 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPROLVD_ZMM_K1Z_ZMM_ZMMM512B32 = 2673;
	/**
	 * {@code VPROLVQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 15 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPROLVQ_XMM_K1Z_XMM_XMMM128B64 = 2674;
	/**
	 * {@code VPROLVQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 15 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPROLVQ_YMM_K1Z_YMM_YMMM256B64 = 2675;
	/**
	 * {@code VPROLVQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 15 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPROLVQ_ZMM_K1Z_ZMM_ZMMM512B64 = 2676;
	/**
	 * {@code VPMOVUSQD xmm1/m64 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 15 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSQD_XMMM64_K1Z_XMM = 2677;
	/**
	 * {@code VPMOVUSQD xmm1/m128 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 15 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSQD_XMMM128_K1Z_YMM = 2678;
	/**
	 * {@code VPMOVUSQD ymm1/m256 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 15 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVUSQD_YMMM256_K1Z_ZMM = 2679;
	/**
	 * {@code VPERMPS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 16 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMPS_YMM_YMM_YMMM256 = 2680;
	/**
	 * {@code VPERMPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 16 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMPS_YMM_K1Z_YMM_YMMM256B32 = 2681;
	/**
	 * {@code VPERMPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 16 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMPS_ZMM_K1Z_ZMM_ZMMM512B32 = 2682;
	/**
	 * {@code VPERMPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 16 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMPD_YMM_K1Z_YMM_YMMM256B64 = 2683;
	/**
	 * {@code VPERMPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 16 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMPD_ZMM_K1Z_ZMM_ZMMM512B64 = 2684;
	/**
	 * {@code PTEST xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 17 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PTEST_XMM_XMMM128 = 2685;
	/**
	 * {@code VPTEST xmm1, xmm2/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 17 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPTEST_XMM_XMMM128 = 2686;
	/**
	 * {@code VPTEST ymm1, ymm2/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 17 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPTEST_YMM_YMMM256 = 2687;
	/**
	 * {@code VBROADCASTSS xmm1, m32}<p>
	 * {@code VEX.128.66.0F38.W0 18 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBROADCASTSS_XMM_M32 = 2688;
	/**
	 * {@code VBROADCASTSS ymm1, m32}<p>
	 * {@code VEX.256.66.0F38.W0 18 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBROADCASTSS_YMM_M32 = 2689;
	/**
	 * {@code VBROADCASTSS xmm1 {k1}{z}, xmm2/m32}<p>
	 * {@code EVEX.128.66.0F38.W0 18 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTSS_XMM_K1Z_XMMM32 = 2690;
	/**
	 * {@code VBROADCASTSS ymm1 {k1}{z}, xmm2/m32}<p>
	 * {@code EVEX.256.66.0F38.W0 18 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTSS_YMM_K1Z_XMMM32 = 2691;
	/**
	 * {@code VBROADCASTSS zmm1 {k1}{z}, xmm2/m32}<p>
	 * {@code EVEX.512.66.0F38.W0 18 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTSS_ZMM_K1Z_XMMM32 = 2692;
	/**
	 * {@code VBROADCASTSD ymm1, m64}<p>
	 * {@code VEX.256.66.0F38.W0 19 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBROADCASTSD_YMM_M64 = 2693;
	/**
	 * {@code VBROADCASTF32X2 ymm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.256.66.0F38.W0 19 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTF32X2_YMM_K1Z_XMMM64 = 2694;
	/**
	 * {@code VBROADCASTF32X2 zmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.512.66.0F38.W0 19 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTF32X2_ZMM_K1Z_XMMM64 = 2695;
	/**
	 * {@code VBROADCASTSD ymm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.256.66.0F38.W1 19 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTSD_YMM_K1Z_XMMM64 = 2696;
	/**
	 * {@code VBROADCASTSD zmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.512.66.0F38.W1 19 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTSD_ZMM_K1Z_XMMM64 = 2697;
	/**
	 * {@code VBROADCASTF128 ymm1, m128}<p>
	 * {@code VEX.256.66.0F38.W0 1A /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBROADCASTF128_YMM_M128 = 2698;
	/**
	 * {@code VBROADCASTF32X4 ymm1 {k1}{z}, m128}<p>
	 * {@code EVEX.256.66.0F38.W0 1A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTF32X4_YMM_K1Z_M128 = 2699;
	/**
	 * {@code VBROADCASTF32X4 zmm1 {k1}{z}, m128}<p>
	 * {@code EVEX.512.66.0F38.W0 1A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTF32X4_ZMM_K1Z_M128 = 2700;
	/**
	 * {@code VBROADCASTF64X2 ymm1 {k1}{z}, m128}<p>
	 * {@code EVEX.256.66.0F38.W1 1A /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTF64X2_YMM_K1Z_M128 = 2701;
	/**
	 * {@code VBROADCASTF64X2 zmm1 {k1}{z}, m128}<p>
	 * {@code EVEX.512.66.0F38.W1 1A /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTF64X2_ZMM_K1Z_M128 = 2702;
	/**
	 * {@code VBROADCASTF32X8 zmm1 {k1}{z}, m256}<p>
	 * {@code EVEX.512.66.0F38.W0 1B /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTF32X8_ZMM_K1Z_M256 = 2703;
	/**
	 * {@code VBROADCASTF64X4 zmm1 {k1}{z}, m256}<p>
	 * {@code EVEX.512.66.0F38.W1 1B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTF64X4_ZMM_K1Z_M256 = 2704;
	/**
	 * {@code PABSB mm1, mm2/m64}<p>
	 * {@code NP 0F 38 1C /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PABSB_MM_MMM64 = 2705;
	/**
	 * {@code PABSB xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 1C /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PABSB_XMM_XMMM128 = 2706;
	/**
	 * {@code VPABSB xmm1, xmm2/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 1C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPABSB_XMM_XMMM128 = 2707;
	/**
	 * {@code VPABSB ymm1, ymm2/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 1C /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPABSB_YMM_YMMM256 = 2708;
	/**
	 * {@code VPABSB xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.66.0F38.WIG 1C /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPABSB_XMM_K1Z_XMMM128 = 2709;
	/**
	 * {@code VPABSB ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.66.0F38.WIG 1C /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPABSB_YMM_K1Z_YMMM256 = 2710;
	/**
	 * {@code VPABSB zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.66.0F38.WIG 1C /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPABSB_ZMM_K1Z_ZMMM512 = 2711;
	/**
	 * {@code PABSW mm1, mm2/m64}<p>
	 * {@code NP 0F 38 1D /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PABSW_MM_MMM64 = 2712;
	/**
	 * {@code PABSW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 1D /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PABSW_XMM_XMMM128 = 2713;
	/**
	 * {@code VPABSW xmm1, xmm2/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 1D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPABSW_XMM_XMMM128 = 2714;
	/**
	 * {@code VPABSW ymm1, ymm2/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 1D /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPABSW_YMM_YMMM256 = 2715;
	/**
	 * {@code VPABSW xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.66.0F38.WIG 1D /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPABSW_XMM_K1Z_XMMM128 = 2716;
	/**
	 * {@code VPABSW ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.66.0F38.WIG 1D /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPABSW_YMM_K1Z_YMMM256 = 2717;
	/**
	 * {@code VPABSW zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.66.0F38.WIG 1D /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPABSW_ZMM_K1Z_ZMMM512 = 2718;
	/**
	 * {@code PABSD mm1, mm2/m64}<p>
	 * {@code NP 0F 38 1E /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PABSD_MM_MMM64 = 2719;
	/**
	 * {@code PABSD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 1E /r}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PABSD_XMM_XMMM128 = 2720;
	/**
	 * {@code VPABSD xmm1, xmm2/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 1E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPABSD_XMM_XMMM128 = 2721;
	/**
	 * {@code VPABSD ymm1, ymm2/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 1E /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPABSD_YMM_YMMM256 = 2722;
	/**
	 * {@code VPABSD xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 1E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPABSD_XMM_K1Z_XMMM128B32 = 2723;
	/**
	 * {@code VPABSD ymm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 1E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPABSD_YMM_K1Z_YMMM256B32 = 2724;
	/**
	 * {@code VPABSD zmm1 {k1}{z}, zmm2/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 1E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPABSD_ZMM_K1Z_ZMMM512B32 = 2725;
	/**
	 * {@code VPABSQ xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 1F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPABSQ_XMM_K1Z_XMMM128B64 = 2726;
	/**
	 * {@code VPABSQ ymm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 1F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPABSQ_YMM_K1Z_YMMM256B64 = 2727;
	/**
	 * {@code VPABSQ zmm1 {k1}{z}, zmm2/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 1F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPABSQ_ZMM_K1Z_ZMMM512B64 = 2728;
	/**
	 * {@code PMOVSXBW xmm1, xmm2/m64}<p>
	 * {@code 66 0F 38 20 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMOVSXBW_XMM_XMMM64 = 2729;
	/**
	 * {@code VPMOVSXBW xmm1, xmm2/m64}<p>
	 * {@code VEX.128.66.0F38.WIG 20 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVSXBW_XMM_XMMM64 = 2730;
	/**
	 * {@code VPMOVSXBW ymm1, xmm2/m128}<p>
	 * {@code VEX.256.66.0F38.WIG 20 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVSXBW_YMM_XMMM128 = 2731;
	/**
	 * {@code VPMOVSXBW xmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.128.66.0F38.WIG 20 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXBW_XMM_K1Z_XMMM64 = 2732;
	/**
	 * {@code VPMOVSXBW ymm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.256.66.0F38.WIG 20 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXBW_YMM_K1Z_XMMM128 = 2733;
	/**
	 * {@code VPMOVSXBW zmm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.512.66.0F38.WIG 20 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXBW_ZMM_K1Z_YMMM256 = 2734;
	/**
	 * {@code VPMOVSWB xmm1/m64 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 20 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSWB_XMMM64_K1Z_XMM = 2735;
	/**
	 * {@code VPMOVSWB xmm1/m128 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 20 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSWB_XMMM128_K1Z_YMM = 2736;
	/**
	 * {@code VPMOVSWB ymm1/m256 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 20 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSWB_YMMM256_K1Z_ZMM = 2737;
	/**
	 * {@code PMOVSXBD xmm1, xmm2/m32}<p>
	 * {@code 66 0F 38 21 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMOVSXBD_XMM_XMMM32 = 2738;
	/**
	 * {@code VPMOVSXBD xmm1, xmm2/m32}<p>
	 * {@code VEX.128.66.0F38.WIG 21 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVSXBD_XMM_XMMM32 = 2739;
	/**
	 * {@code VPMOVSXBD ymm1, xmm2/m64}<p>
	 * {@code VEX.256.66.0F38.WIG 21 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVSXBD_YMM_XMMM64 = 2740;
	/**
	 * {@code VPMOVSXBD xmm1 {k1}{z}, xmm2/m32}<p>
	 * {@code EVEX.128.66.0F38.WIG 21 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXBD_XMM_K1Z_XMMM32 = 2741;
	/**
	 * {@code VPMOVSXBD ymm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.256.66.0F38.WIG 21 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXBD_YMM_K1Z_XMMM64 = 2742;
	/**
	 * {@code VPMOVSXBD zmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.512.66.0F38.WIG 21 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXBD_ZMM_K1Z_XMMM128 = 2743;
	/**
	 * {@code VPMOVSDB xmm1/m32 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 21 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSDB_XMMM32_K1Z_XMM = 2744;
	/**
	 * {@code VPMOVSDB xmm1/m64 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 21 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSDB_XMMM64_K1Z_YMM = 2745;
	/**
	 * {@code VPMOVSDB xmm1/m128 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 21 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSDB_XMMM128_K1Z_ZMM = 2746;
	/**
	 * {@code PMOVSXBQ xmm1, xmm2/m16}<p>
	 * {@code 66 0F 38 22 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMOVSXBQ_XMM_XMMM16 = 2747;
	/**
	 * {@code VPMOVSXBQ xmm1, xmm2/m16}<p>
	 * {@code VEX.128.66.0F38.WIG 22 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVSXBQ_XMM_XMMM16 = 2748;
	/**
	 * {@code VPMOVSXBQ ymm1, xmm2/m32}<p>
	 * {@code VEX.256.66.0F38.WIG 22 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVSXBQ_YMM_XMMM32 = 2749;
	/**
	 * {@code VPMOVSXBQ xmm1 {k1}{z}, xmm2/m16}<p>
	 * {@code EVEX.128.66.0F38.WIG 22 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXBQ_XMM_K1Z_XMMM16 = 2750;
	/**
	 * {@code VPMOVSXBQ ymm1 {k1}{z}, xmm2/m32}<p>
	 * {@code EVEX.256.66.0F38.WIG 22 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXBQ_YMM_K1Z_XMMM32 = 2751;
	/**
	 * {@code VPMOVSXBQ zmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.512.66.0F38.WIG 22 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXBQ_ZMM_K1Z_XMMM64 = 2752;
	/**
	 * {@code VPMOVSQB xmm1/m16 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 22 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSQB_XMMM16_K1Z_XMM = 2753;
	/**
	 * {@code VPMOVSQB xmm1/m32 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 22 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSQB_XMMM32_K1Z_YMM = 2754;
	/**
	 * {@code VPMOVSQB xmm1/m64 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 22 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSQB_XMMM64_K1Z_ZMM = 2755;
	/**
	 * {@code PMOVSXWD xmm1, xmm2/m64}<p>
	 * {@code 66 0F 38 23 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMOVSXWD_XMM_XMMM64 = 2756;
	/**
	 * {@code VPMOVSXWD xmm1, xmm2/m64}<p>
	 * {@code VEX.128.66.0F38.WIG 23 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVSXWD_XMM_XMMM64 = 2757;
	/**
	 * {@code VPMOVSXWD ymm1, xmm2/m128}<p>
	 * {@code VEX.256.66.0F38.WIG 23 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVSXWD_YMM_XMMM128 = 2758;
	/**
	 * {@code VPMOVSXWD xmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.128.66.0F38.WIG 23 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXWD_XMM_K1Z_XMMM64 = 2759;
	/**
	 * {@code VPMOVSXWD ymm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.256.66.0F38.WIG 23 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXWD_YMM_K1Z_XMMM128 = 2760;
	/**
	 * {@code VPMOVSXWD zmm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.512.66.0F38.WIG 23 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXWD_ZMM_K1Z_YMMM256 = 2761;
	/**
	 * {@code VPMOVSDW xmm1/m64 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 23 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSDW_XMMM64_K1Z_XMM = 2762;
	/**
	 * {@code VPMOVSDW xmm1/m128 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 23 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSDW_XMMM128_K1Z_YMM = 2763;
	/**
	 * {@code VPMOVSDW ymm1/m256 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 23 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSDW_YMMM256_K1Z_ZMM = 2764;
	/**
	 * {@code PMOVSXWQ xmm1, xmm2/m32}<p>
	 * {@code 66 0F 38 24 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMOVSXWQ_XMM_XMMM32 = 2765;
	/**
	 * {@code VPMOVSXWQ xmm1, xmm2/m32}<p>
	 * {@code VEX.128.66.0F38.WIG 24 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVSXWQ_XMM_XMMM32 = 2766;
	/**
	 * {@code VPMOVSXWQ ymm1, xmm2/m64}<p>
	 * {@code VEX.256.66.0F38.WIG 24 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVSXWQ_YMM_XMMM64 = 2767;
	/**
	 * {@code VPMOVSXWQ xmm1 {k1}{z}, xmm2/m32}<p>
	 * {@code EVEX.128.66.0F38.WIG 24 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXWQ_XMM_K1Z_XMMM32 = 2768;
	/**
	 * {@code VPMOVSXWQ ymm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.256.66.0F38.WIG 24 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXWQ_YMM_K1Z_XMMM64 = 2769;
	/**
	 * {@code VPMOVSXWQ zmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.512.66.0F38.WIG 24 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXWQ_ZMM_K1Z_XMMM128 = 2770;
	/**
	 * {@code VPMOVSQW xmm1/m32 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 24 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSQW_XMMM32_K1Z_XMM = 2771;
	/**
	 * {@code VPMOVSQW xmm1/m64 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 24 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSQW_XMMM64_K1Z_YMM = 2772;
	/**
	 * {@code VPMOVSQW xmm1/m128 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 24 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSQW_XMMM128_K1Z_ZMM = 2773;
	/**
	 * {@code PMOVSXDQ xmm1, xmm2/m64}<p>
	 * {@code 66 0F 38 25 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMOVSXDQ_XMM_XMMM64 = 2774;
	/**
	 * {@code VPMOVSXDQ xmm1, xmm2/m64}<p>
	 * {@code VEX.128.66.0F38.WIG 25 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVSXDQ_XMM_XMMM64 = 2775;
	/**
	 * {@code VPMOVSXDQ ymm1, xmm2/m128}<p>
	 * {@code VEX.256.66.0F38.WIG 25 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVSXDQ_YMM_XMMM128 = 2776;
	/**
	 * {@code VPMOVSXDQ xmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.128.66.0F38.W0 25 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXDQ_XMM_K1Z_XMMM64 = 2777;
	/**
	 * {@code VPMOVSXDQ ymm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.256.66.0F38.W0 25 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXDQ_YMM_K1Z_XMMM128 = 2778;
	/**
	 * {@code VPMOVSXDQ zmm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.512.66.0F38.W0 25 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSXDQ_ZMM_K1Z_YMMM256 = 2779;
	/**
	 * {@code VPMOVSQD xmm1/m64 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 25 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSQD_XMMM64_K1Z_XMM = 2780;
	/**
	 * {@code VPMOVSQD xmm1/m128 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 25 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSQD_XMMM128_K1Z_YMM = 2781;
	/**
	 * {@code VPMOVSQD ymm1/m256 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 25 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVSQD_YMMM256_K1Z_ZMM = 2782;
	/**
	 * {@code VPTESTMB k2 {k1}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W0 26 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTMB_KR_K1_XMM_XMMM128 = 2783;
	/**
	 * {@code VPTESTMB k2 {k1}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W0 26 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTMB_KR_K1_YMM_YMMM256 = 2784;
	/**
	 * {@code VPTESTMB k2 {k1}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W0 26 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTMB_KR_K1_ZMM_ZMMM512 = 2785;
	/**
	 * {@code VPTESTMW k2 {k1}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W1 26 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTMW_KR_K1_XMM_XMMM128 = 2786;
	/**
	 * {@code VPTESTMW k2 {k1}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W1 26 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTMW_KR_K1_YMM_YMMM256 = 2787;
	/**
	 * {@code VPTESTMW k2 {k1}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W1 26 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTMW_KR_K1_ZMM_ZMMM512 = 2788;
	/**
	 * {@code VPTESTNMB k2 {k1}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.F3.0F38.W0 26 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTNMB_KR_K1_XMM_XMMM128 = 2789;
	/**
	 * {@code VPTESTNMB k2 {k1}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.F3.0F38.W0 26 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTNMB_KR_K1_YMM_YMMM256 = 2790;
	/**
	 * {@code VPTESTNMB k2 {k1}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.F3.0F38.W0 26 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTNMB_KR_K1_ZMM_ZMMM512 = 2791;
	/**
	 * {@code VPTESTNMW k2 {k1}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.F3.0F38.W1 26 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTNMW_KR_K1_XMM_XMMM128 = 2792;
	/**
	 * {@code VPTESTNMW k2 {k1}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.F3.0F38.W1 26 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTNMW_KR_K1_YMM_YMMM256 = 2793;
	/**
	 * {@code VPTESTNMW k2 {k1}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.F3.0F38.W1 26 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTNMW_KR_K1_ZMM_ZMMM512 = 2794;
	/**
	 * {@code VPTESTMD k2 {k1}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 27 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTMD_KR_K1_XMM_XMMM128B32 = 2795;
	/**
	 * {@code VPTESTMD k2 {k1}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 27 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTMD_KR_K1_YMM_YMMM256B32 = 2796;
	/**
	 * {@code VPTESTMD k2 {k1}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 27 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTMD_KR_K1_ZMM_ZMMM512B32 = 2797;
	/**
	 * {@code VPTESTMQ k2 {k1}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 27 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTMQ_KR_K1_XMM_XMMM128B64 = 2798;
	/**
	 * {@code VPTESTMQ k2 {k1}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 27 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTMQ_KR_K1_YMM_YMMM256B64 = 2799;
	/**
	 * {@code VPTESTMQ k2 {k1}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 27 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTMQ_KR_K1_ZMM_ZMMM512B64 = 2800;
	/**
	 * {@code VPTESTNMD k2 {k1}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.F3.0F38.W0 27 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTNMD_KR_K1_XMM_XMMM128B32 = 2801;
	/**
	 * {@code VPTESTNMD k2 {k1}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.F3.0F38.W0 27 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTNMD_KR_K1_YMM_YMMM256B32 = 2802;
	/**
	 * {@code VPTESTNMD k2 {k1}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.F3.0F38.W0 27 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTNMD_KR_K1_ZMM_ZMMM512B32 = 2803;
	/**
	 * {@code VPTESTNMQ k2 {k1}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.F3.0F38.W1 27 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTNMQ_KR_K1_XMM_XMMM128B64 = 2804;
	/**
	 * {@code VPTESTNMQ k2 {k1}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.F3.0F38.W1 27 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTNMQ_KR_K1_YMM_YMMM256B64 = 2805;
	/**
	 * {@code VPTESTNMQ k2 {k1}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.F3.0F38.W1 27 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTESTNMQ_KR_K1_ZMM_ZMMM512B64 = 2806;
	/**
	 * {@code PMULDQ xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 28 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMULDQ_XMM_XMMM128 = 2807;
	/**
	 * {@code VPMULDQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 28 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMULDQ_XMM_XMM_XMMM128 = 2808;
	/**
	 * {@code VPMULDQ ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 28 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMULDQ_YMM_YMM_YMMM256 = 2809;
	/**
	 * {@code VPMULDQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 28 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULDQ_XMM_K1Z_XMM_XMMM128B64 = 2810;
	/**
	 * {@code VPMULDQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 28 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULDQ_YMM_K1Z_YMM_YMMM256B64 = 2811;
	/**
	 * {@code VPMULDQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 28 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULDQ_ZMM_K1Z_ZMM_ZMMM512B64 = 2812;
	/**
	 * {@code VPMOVM2B xmm1, k1}<p>
	 * {@code EVEX.128.F3.0F38.W0 28 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVM2B_XMM_KR = 2813;
	/**
	 * {@code VPMOVM2B ymm1, k1}<p>
	 * {@code EVEX.256.F3.0F38.W0 28 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVM2B_YMM_KR = 2814;
	/**
	 * {@code VPMOVM2B zmm1, k1}<p>
	 * {@code EVEX.512.F3.0F38.W0 28 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVM2B_ZMM_KR = 2815;
	/**
	 * {@code VPMOVM2W xmm1, k1}<p>
	 * {@code EVEX.128.F3.0F38.W1 28 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVM2W_XMM_KR = 2816;
	/**
	 * {@code VPMOVM2W ymm1, k1}<p>
	 * {@code EVEX.256.F3.0F38.W1 28 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVM2W_YMM_KR = 2817;
	/**
	 * {@code VPMOVM2W zmm1, k1}<p>
	 * {@code EVEX.512.F3.0F38.W1 28 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVM2W_ZMM_KR = 2818;
	/**
	 * {@code PCMPEQQ xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 29 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPEQQ_XMM_XMMM128 = 2819;
	/**
	 * {@code VPCMPEQQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 29 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPEQQ_XMM_XMM_XMMM128 = 2820;
	/**
	 * {@code VPCMPEQQ ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 29 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPEQQ_YMM_YMM_YMMM256 = 2821;
	/**
	 * {@code VPCMPEQQ k1 {k2}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 29 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPEQQ_KR_K1_XMM_XMMM128B64 = 2822;
	/**
	 * {@code VPCMPEQQ k1 {k2}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 29 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPEQQ_KR_K1_YMM_YMMM256B64 = 2823;
	/**
	 * {@code VPCMPEQQ k1 {k2}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 29 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPEQQ_KR_K1_ZMM_ZMMM512B64 = 2824;
	/**
	 * {@code VPMOVB2M k1, xmm1}<p>
	 * {@code EVEX.128.F3.0F38.W0 29 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVB2M_KR_XMM = 2825;
	/**
	 * {@code VPMOVB2M k1, ymm1}<p>
	 * {@code EVEX.256.F3.0F38.W0 29 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVB2M_KR_YMM = 2826;
	/**
	 * {@code VPMOVB2M k1, zmm1}<p>
	 * {@code EVEX.512.F3.0F38.W0 29 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVB2M_KR_ZMM = 2827;
	/**
	 * {@code VPMOVW2M k1, xmm1}<p>
	 * {@code EVEX.128.F3.0F38.W1 29 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVW2M_KR_XMM = 2828;
	/**
	 * {@code VPMOVW2M k1, ymm1}<p>
	 * {@code EVEX.256.F3.0F38.W1 29 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVW2M_KR_YMM = 2829;
	/**
	 * {@code VPMOVW2M k1, zmm1}<p>
	 * {@code EVEX.512.F3.0F38.W1 29 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVW2M_KR_ZMM = 2830;
	/**
	 * {@code MOVNTDQA xmm1, m128}<p>
	 * {@code 66 0F 38 2A /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVNTDQA_XMM_M128 = 2831;
	/**
	 * {@code VMOVNTDQA xmm1, m128}<p>
	 * {@code VEX.128.66.0F38.WIG 2A /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVNTDQA_XMM_M128 = 2832;
	/**
	 * {@code VMOVNTDQA ymm1, m256}<p>
	 * {@code VEX.256.66.0F38.WIG 2A /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMOVNTDQA_YMM_M256 = 2833;
	/**
	 * {@code VMOVNTDQA xmm1, m128}<p>
	 * {@code EVEX.128.66.0F38.W0 2A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVNTDQA_XMM_M128 = 2834;
	/**
	 * {@code VMOVNTDQA ymm1, m256}<p>
	 * {@code EVEX.256.66.0F38.W0 2A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVNTDQA_YMM_M256 = 2835;
	/**
	 * {@code VMOVNTDQA zmm1, m512}<p>
	 * {@code EVEX.512.66.0F38.W0 2A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVNTDQA_ZMM_M512 = 2836;
	/**
	 * {@code VPBROADCASTMB2Q xmm1, k1}<p>
	 * {@code EVEX.128.F3.0F38.W1 2A /r}<p>
	 * {@code AVX512VL and AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTMB2Q_XMM_KR = 2837;
	/**
	 * {@code VPBROADCASTMB2Q ymm1, k1}<p>
	 * {@code EVEX.256.F3.0F38.W1 2A /r}<p>
	 * {@code AVX512VL and AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTMB2Q_YMM_KR = 2838;
	/**
	 * {@code VPBROADCASTMB2Q zmm1, k1}<p>
	 * {@code EVEX.512.F3.0F38.W1 2A /r}<p>
	 * {@code AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTMB2Q_ZMM_KR = 2839;
	/**
	 * {@code PACKUSDW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 2B /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PACKUSDW_XMM_XMMM128 = 2840;
	/**
	 * {@code VPACKUSDW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 2B /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPACKUSDW_XMM_XMM_XMMM128 = 2841;
	/**
	 * {@code VPACKUSDW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 2B /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPACKUSDW_YMM_YMM_YMMM256 = 2842;
	/**
	 * {@code VPACKUSDW xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 2B /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPACKUSDW_XMM_K1Z_XMM_XMMM128B32 = 2843;
	/**
	 * {@code VPACKUSDW ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 2B /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPACKUSDW_YMM_K1Z_YMM_YMMM256B32 = 2844;
	/**
	 * {@code VPACKUSDW zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 2B /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPACKUSDW_ZMM_K1Z_ZMM_ZMMM512B32 = 2845;
	/**
	 * {@code VMASKMOVPS xmm1, xmm2, m128}<p>
	 * {@code VEX.128.66.0F38.W0 2C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMASKMOVPS_XMM_XMM_M128 = 2846;
	/**
	 * {@code VMASKMOVPS ymm1, ymm2, m256}<p>
	 * {@code VEX.256.66.0F38.W0 2C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMASKMOVPS_YMM_YMM_M256 = 2847;
	/**
	 * {@code VSCALEFPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 2C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCALEFPS_XMM_K1Z_XMM_XMMM128B32 = 2848;
	/**
	 * {@code VSCALEFPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 2C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCALEFPS_YMM_K1Z_YMM_YMMM256B32 = 2849;
	/**
	 * {@code VSCALEFPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 2C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCALEFPS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 2850;
	/**
	 * {@code VSCALEFPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 2C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCALEFPD_XMM_K1Z_XMM_XMMM128B64 = 2851;
	/**
	 * {@code VSCALEFPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 2C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCALEFPD_YMM_K1Z_YMM_YMMM256B64 = 2852;
	/**
	 * {@code VSCALEFPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 2C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCALEFPD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 2853;
	/**
	 * {@code VMASKMOVPD xmm1, xmm2, m128}<p>
	 * {@code VEX.128.66.0F38.W0 2D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMASKMOVPD_XMM_XMM_M128 = 2854;
	/**
	 * {@code VMASKMOVPD ymm1, ymm2, m256}<p>
	 * {@code VEX.256.66.0F38.W0 2D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMASKMOVPD_YMM_YMM_M256 = 2855;
	/**
	 * {@code VSCALEFSS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 2D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCALEFSS_XMM_K1Z_XMM_XMMM32_ER = 2856;
	/**
	 * {@code VSCALEFSD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 2D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCALEFSD_XMM_K1Z_XMM_XMMM64_ER = 2857;
	/**
	 * {@code VMASKMOVPS m128, xmm1, xmm2}<p>
	 * {@code VEX.128.66.0F38.W0 2E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMASKMOVPS_M128_XMM_XMM = 2858;
	/**
	 * {@code VMASKMOVPS m256, ymm1, ymm2}<p>
	 * {@code VEX.256.66.0F38.W0 2E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMASKMOVPS_M256_YMM_YMM = 2859;
	/**
	 * {@code VMASKMOVPD m128, xmm1, xmm2}<p>
	 * {@code VEX.128.66.0F38.W0 2F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMASKMOVPD_M128_XMM_XMM = 2860;
	/**
	 * {@code VMASKMOVPD m256, ymm1, ymm2}<p>
	 * {@code VEX.256.66.0F38.W0 2F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMASKMOVPD_M256_YMM_YMM = 2861;
	/**
	 * {@code PMOVZXBW xmm1, xmm2/m64}<p>
	 * {@code 66 0F 38 30 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMOVZXBW_XMM_XMMM64 = 2862;
	/**
	 * {@code VPMOVZXBW xmm1, xmm2/m64}<p>
	 * {@code VEX.128.66.0F38.WIG 30 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVZXBW_XMM_XMMM64 = 2863;
	/**
	 * {@code VPMOVZXBW ymm1, xmm2/m128}<p>
	 * {@code VEX.256.66.0F38.WIG 30 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVZXBW_YMM_XMMM128 = 2864;
	/**
	 * {@code VPMOVZXBW xmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.128.66.0F38.WIG 30 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXBW_XMM_K1Z_XMMM64 = 2865;
	/**
	 * {@code VPMOVZXBW ymm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.256.66.0F38.WIG 30 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXBW_YMM_K1Z_XMMM128 = 2866;
	/**
	 * {@code VPMOVZXBW zmm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.512.66.0F38.WIG 30 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXBW_ZMM_K1Z_YMMM256 = 2867;
	/**
	 * {@code VPMOVWB xmm1/m64 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 30 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVWB_XMMM64_K1Z_XMM = 2868;
	/**
	 * {@code VPMOVWB xmm1/m128 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 30 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVWB_XMMM128_K1Z_YMM = 2869;
	/**
	 * {@code VPMOVWB ymm1/m256 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 30 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVWB_YMMM256_K1Z_ZMM = 2870;
	/**
	 * {@code PMOVZXBD xmm1, xmm2/m32}<p>
	 * {@code 66 0F 38 31 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMOVZXBD_XMM_XMMM32 = 2871;
	/**
	 * {@code VPMOVZXBD xmm1, xmm2/m32}<p>
	 * {@code VEX.128.66.0F38.WIG 31 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVZXBD_XMM_XMMM32 = 2872;
	/**
	 * {@code VPMOVZXBD ymm1, xmm2/m64}<p>
	 * {@code VEX.256.66.0F38.WIG 31 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVZXBD_YMM_XMMM64 = 2873;
	/**
	 * {@code VPMOVZXBD xmm1 {k1}{z}, xmm2/m32}<p>
	 * {@code EVEX.128.66.0F38.WIG 31 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXBD_XMM_K1Z_XMMM32 = 2874;
	/**
	 * {@code VPMOVZXBD ymm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.256.66.0F38.WIG 31 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXBD_YMM_K1Z_XMMM64 = 2875;
	/**
	 * {@code VPMOVZXBD zmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.512.66.0F38.WIG 31 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXBD_ZMM_K1Z_XMMM128 = 2876;
	/**
	 * {@code VPMOVDB xmm1/m32 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 31 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVDB_XMMM32_K1Z_XMM = 2877;
	/**
	 * {@code VPMOVDB xmm1/m64 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 31 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVDB_XMMM64_K1Z_YMM = 2878;
	/**
	 * {@code VPMOVDB xmm1/m128 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 31 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVDB_XMMM128_K1Z_ZMM = 2879;
	/**
	 * {@code PMOVZXBQ xmm1, xmm2/m16}<p>
	 * {@code 66 0F 38 32 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMOVZXBQ_XMM_XMMM16 = 2880;
	/**
	 * {@code VPMOVZXBQ xmm1, xmm2/m16}<p>
	 * {@code VEX.128.66.0F38.WIG 32 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVZXBQ_XMM_XMMM16 = 2881;
	/**
	 * {@code VPMOVZXBQ ymm1, xmm2/m32}<p>
	 * {@code VEX.256.66.0F38.WIG 32 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVZXBQ_YMM_XMMM32 = 2882;
	/**
	 * {@code VPMOVZXBQ xmm1 {k1}{z}, xmm2/m16}<p>
	 * {@code EVEX.128.66.0F38.WIG 32 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXBQ_XMM_K1Z_XMMM16 = 2883;
	/**
	 * {@code VPMOVZXBQ ymm1 {k1}{z}, xmm2/m32}<p>
	 * {@code EVEX.256.66.0F38.WIG 32 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXBQ_YMM_K1Z_XMMM32 = 2884;
	/**
	 * {@code VPMOVZXBQ zmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.512.66.0F38.WIG 32 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXBQ_ZMM_K1Z_XMMM64 = 2885;
	/**
	 * {@code VPMOVQB xmm1/m16 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 32 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVQB_XMMM16_K1Z_XMM = 2886;
	/**
	 * {@code VPMOVQB xmm1/m32 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 32 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVQB_XMMM32_K1Z_YMM = 2887;
	/**
	 * {@code VPMOVQB xmm1/m64 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 32 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVQB_XMMM64_K1Z_ZMM = 2888;
	/**
	 * {@code PMOVZXWD xmm1, xmm2/m64}<p>
	 * {@code 66 0F 38 33 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMOVZXWD_XMM_XMMM64 = 2889;
	/**
	 * {@code VPMOVZXWD xmm1, xmm2/m64}<p>
	 * {@code VEX.128.66.0F38.WIG 33 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVZXWD_XMM_XMMM64 = 2890;
	/**
	 * {@code VPMOVZXWD ymm1, xmm2/m128}<p>
	 * {@code VEX.256.66.0F38.WIG 33 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVZXWD_YMM_XMMM128 = 2891;
	/**
	 * {@code VPMOVZXWD xmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.128.66.0F38.WIG 33 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXWD_XMM_K1Z_XMMM64 = 2892;
	/**
	 * {@code VPMOVZXWD ymm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.256.66.0F38.WIG 33 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXWD_YMM_K1Z_XMMM128 = 2893;
	/**
	 * {@code VPMOVZXWD zmm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.512.66.0F38.WIG 33 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXWD_ZMM_K1Z_YMMM256 = 2894;
	/**
	 * {@code VPMOVDW xmm1/m64 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 33 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVDW_XMMM64_K1Z_XMM = 2895;
	/**
	 * {@code VPMOVDW xmm1/m128 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 33 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVDW_XMMM128_K1Z_YMM = 2896;
	/**
	 * {@code VPMOVDW ymm1/m256 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 33 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVDW_YMMM256_K1Z_ZMM = 2897;
	/**
	 * {@code PMOVZXWQ xmm1, xmm2/m32}<p>
	 * {@code 66 0F 38 34 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMOVZXWQ_XMM_XMMM32 = 2898;
	/**
	 * {@code VPMOVZXWQ xmm1, xmm2/m32}<p>
	 * {@code VEX.128.66.0F38.WIG 34 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVZXWQ_XMM_XMMM32 = 2899;
	/**
	 * {@code VPMOVZXWQ ymm1, xmm2/m64}<p>
	 * {@code VEX.256.66.0F38.WIG 34 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVZXWQ_YMM_XMMM64 = 2900;
	/**
	 * {@code VPMOVZXWQ xmm1 {k1}{z}, xmm2/m32}<p>
	 * {@code EVEX.128.66.0F38.WIG 34 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXWQ_XMM_K1Z_XMMM32 = 2901;
	/**
	 * {@code VPMOVZXWQ ymm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.256.66.0F38.WIG 34 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXWQ_YMM_K1Z_XMMM64 = 2902;
	/**
	 * {@code VPMOVZXWQ zmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.512.66.0F38.WIG 34 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXWQ_ZMM_K1Z_XMMM128 = 2903;
	/**
	 * {@code VPMOVQW xmm1/m32 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 34 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVQW_XMMM32_K1Z_XMM = 2904;
	/**
	 * {@code VPMOVQW xmm1/m64 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 34 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVQW_XMMM64_K1Z_YMM = 2905;
	/**
	 * {@code VPMOVQW xmm1/m128 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 34 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVQW_XMMM128_K1Z_ZMM = 2906;
	/**
	 * {@code PMOVZXDQ xmm1, xmm2/m64}<p>
	 * {@code 66 0F 38 35 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMOVZXDQ_XMM_XMMM64 = 2907;
	/**
	 * {@code VPMOVZXDQ xmm1, xmm2/m64}<p>
	 * {@code VEX.128.66.0F38.WIG 35 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVZXDQ_XMM_XMMM64 = 2908;
	/**
	 * {@code VPMOVZXDQ ymm1, xmm2/m128}<p>
	 * {@code VEX.256.66.0F38.WIG 35 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMOVZXDQ_YMM_XMMM128 = 2909;
	/**
	 * {@code VPMOVZXDQ xmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.128.66.0F38.W0 35 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXDQ_XMM_K1Z_XMMM64 = 2910;
	/**
	 * {@code VPMOVZXDQ ymm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.256.66.0F38.W0 35 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXDQ_YMM_K1Z_XMMM128 = 2911;
	/**
	 * {@code VPMOVZXDQ zmm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.512.66.0F38.W0 35 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVZXDQ_ZMM_K1Z_YMMM256 = 2912;
	/**
	 * {@code VPMOVQD xmm1/m64 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.F3.0F38.W0 35 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVQD_XMMM64_K1Z_XMM = 2913;
	/**
	 * {@code VPMOVQD xmm1/m128 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.F3.0F38.W0 35 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVQD_XMMM128_K1Z_YMM = 2914;
	/**
	 * {@code VPMOVQD ymm1/m256 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.F3.0F38.W0 35 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVQD_YMMM256_K1Z_ZMM = 2915;
	/**
	 * {@code VPERMD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 36 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMD_YMM_YMM_YMMM256 = 2916;
	/**
	 * {@code VPERMD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 36 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMD_YMM_K1Z_YMM_YMMM256B32 = 2917;
	/**
	 * {@code VPERMD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 36 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMD_ZMM_K1Z_ZMM_ZMMM512B32 = 2918;
	/**
	 * {@code VPERMQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 36 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMQ_YMM_K1Z_YMM_YMMM256B64 = 2919;
	/**
	 * {@code VPERMQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 36 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMQ_ZMM_K1Z_ZMM_ZMMM512B64 = 2920;
	/**
	 * {@code PCMPGTQ xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 37 /r}<p>
	 * {@code SSE4.2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPGTQ_XMM_XMMM128 = 2921;
	/**
	 * {@code VPCMPGTQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 37 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPGTQ_XMM_XMM_XMMM128 = 2922;
	/**
	 * {@code VPCMPGTQ ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 37 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPGTQ_YMM_YMM_YMMM256 = 2923;
	/**
	 * {@code VPCMPGTQ k1 {k2}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 37 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPGTQ_KR_K1_XMM_XMMM128B64 = 2924;
	/**
	 * {@code VPCMPGTQ k1 {k2}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 37 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPGTQ_KR_K1_YMM_YMMM256B64 = 2925;
	/**
	 * {@code VPCMPGTQ k1 {k2}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 37 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPGTQ_KR_K1_ZMM_ZMMM512B64 = 2926;
	/**
	 * {@code PMINSB xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 38 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMINSB_XMM_XMMM128 = 2927;
	/**
	 * {@code VPMINSB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 38 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMINSB_XMM_XMM_XMMM128 = 2928;
	/**
	 * {@code VPMINSB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 38 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMINSB_YMM_YMM_YMMM256 = 2929;
	/**
	 * {@code VPMINSB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.WIG 38 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINSB_XMM_K1Z_XMM_XMMM128 = 2930;
	/**
	 * {@code VPMINSB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.WIG 38 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINSB_YMM_K1Z_YMM_YMMM256 = 2931;
	/**
	 * {@code VPMINSB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.WIG 38 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINSB_ZMM_K1Z_ZMM_ZMMM512 = 2932;
	/**
	 * {@code VPMOVM2D xmm1, k1}<p>
	 * {@code EVEX.128.F3.0F38.W0 38 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVM2D_XMM_KR = 2933;
	/**
	 * {@code VPMOVM2D ymm1, k1}<p>
	 * {@code EVEX.256.F3.0F38.W0 38 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVM2D_YMM_KR = 2934;
	/**
	 * {@code VPMOVM2D zmm1, k1}<p>
	 * {@code EVEX.512.F3.0F38.W0 38 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVM2D_ZMM_KR = 2935;
	/**
	 * {@code VPMOVM2Q xmm1, k1}<p>
	 * {@code EVEX.128.F3.0F38.W1 38 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVM2Q_XMM_KR = 2936;
	/**
	 * {@code VPMOVM2Q ymm1, k1}<p>
	 * {@code EVEX.256.F3.0F38.W1 38 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVM2Q_YMM_KR = 2937;
	/**
	 * {@code VPMOVM2Q zmm1, k1}<p>
	 * {@code EVEX.512.F3.0F38.W1 38 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVM2Q_ZMM_KR = 2938;
	/**
	 * {@code PMINSD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 39 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMINSD_XMM_XMMM128 = 2939;
	/**
	 * {@code VPMINSD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 39 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMINSD_XMM_XMM_XMMM128 = 2940;
	/**
	 * {@code VPMINSD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 39 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMINSD_YMM_YMM_YMMM256 = 2941;
	/**
	 * {@code VPMINSD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 39 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINSD_XMM_K1Z_XMM_XMMM128B32 = 2942;
	/**
	 * {@code VPMINSD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 39 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINSD_YMM_K1Z_YMM_YMMM256B32 = 2943;
	/**
	 * {@code VPMINSD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 39 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINSD_ZMM_K1Z_ZMM_ZMMM512B32 = 2944;
	/**
	 * {@code VPMINSQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 39 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINSQ_XMM_K1Z_XMM_XMMM128B64 = 2945;
	/**
	 * {@code VPMINSQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 39 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINSQ_YMM_K1Z_YMM_YMMM256B64 = 2946;
	/**
	 * {@code VPMINSQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 39 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINSQ_ZMM_K1Z_ZMM_ZMMM512B64 = 2947;
	/**
	 * {@code VPMOVD2M k1, xmm1}<p>
	 * {@code EVEX.128.F3.0F38.W0 39 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVD2M_KR_XMM = 2948;
	/**
	 * {@code VPMOVD2M k1, ymm1}<p>
	 * {@code EVEX.256.F3.0F38.W0 39 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVD2M_KR_YMM = 2949;
	/**
	 * {@code VPMOVD2M k1, zmm1}<p>
	 * {@code EVEX.512.F3.0F38.W0 39 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVD2M_KR_ZMM = 2950;
	/**
	 * {@code VPMOVQ2M k1, xmm1}<p>
	 * {@code EVEX.128.F3.0F38.W1 39 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVQ2M_KR_XMM = 2951;
	/**
	 * {@code VPMOVQ2M k1, ymm1}<p>
	 * {@code EVEX.256.F3.0F38.W1 39 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVQ2M_KR_YMM = 2952;
	/**
	 * {@code VPMOVQ2M k1, zmm1}<p>
	 * {@code EVEX.512.F3.0F38.W1 39 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMOVQ2M_KR_ZMM = 2953;
	/**
	 * {@code PMINUW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 3A /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMINUW_XMM_XMMM128 = 2954;
	/**
	 * {@code VPMINUW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 3A /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMINUW_XMM_XMM_XMMM128 = 2955;
	/**
	 * {@code VPMINUW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 3A /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMINUW_YMM_YMM_YMMM256 = 2956;
	/**
	 * {@code VPMINUW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.WIG 3A /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINUW_XMM_K1Z_XMM_XMMM128 = 2957;
	/**
	 * {@code VPMINUW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.WIG 3A /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINUW_YMM_K1Z_YMM_YMMM256 = 2958;
	/**
	 * {@code VPMINUW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.WIG 3A /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINUW_ZMM_K1Z_ZMM_ZMMM512 = 2959;
	/**
	 * {@code VPBROADCASTMW2D xmm1, k1}<p>
	 * {@code EVEX.128.F3.0F38.W0 3A /r}<p>
	 * {@code AVX512VL and AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTMW2D_XMM_KR = 2960;
	/**
	 * {@code VPBROADCASTMW2D ymm1, k1}<p>
	 * {@code EVEX.256.F3.0F38.W0 3A /r}<p>
	 * {@code AVX512VL and AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTMW2D_YMM_KR = 2961;
	/**
	 * {@code VPBROADCASTMW2D zmm1, k1}<p>
	 * {@code EVEX.512.F3.0F38.W0 3A /r}<p>
	 * {@code AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTMW2D_ZMM_KR = 2962;
	/**
	 * {@code PMINUD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 3B /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMINUD_XMM_XMMM128 = 2963;
	/**
	 * {@code VPMINUD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 3B /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMINUD_XMM_XMM_XMMM128 = 2964;
	/**
	 * {@code VPMINUD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 3B /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMINUD_YMM_YMM_YMMM256 = 2965;
	/**
	 * {@code VPMINUD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 3B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINUD_XMM_K1Z_XMM_XMMM128B32 = 2966;
	/**
	 * {@code VPMINUD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 3B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINUD_YMM_K1Z_YMM_YMMM256B32 = 2967;
	/**
	 * {@code VPMINUD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 3B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINUD_ZMM_K1Z_ZMM_ZMMM512B32 = 2968;
	/**
	 * {@code VPMINUQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 3B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINUQ_XMM_K1Z_XMM_XMMM128B64 = 2969;
	/**
	 * {@code VPMINUQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 3B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINUQ_YMM_K1Z_YMM_YMMM256B64 = 2970;
	/**
	 * {@code VPMINUQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 3B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMINUQ_ZMM_K1Z_ZMM_ZMMM512B64 = 2971;
	/**
	 * {@code PMAXSB xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 3C /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMAXSB_XMM_XMMM128 = 2972;
	/**
	 * {@code VPMAXSB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 3C /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMAXSB_XMM_XMM_XMMM128 = 2973;
	/**
	 * {@code VPMAXSB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 3C /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMAXSB_YMM_YMM_YMMM256 = 2974;
	/**
	 * {@code VPMAXSB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.WIG 3C /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXSB_XMM_K1Z_XMM_XMMM128 = 2975;
	/**
	 * {@code VPMAXSB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.WIG 3C /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXSB_YMM_K1Z_YMM_YMMM256 = 2976;
	/**
	 * {@code VPMAXSB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.WIG 3C /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXSB_ZMM_K1Z_ZMM_ZMMM512 = 2977;
	/**
	 * {@code PMAXSD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 3D /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMAXSD_XMM_XMMM128 = 2978;
	/**
	 * {@code VPMAXSD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 3D /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMAXSD_XMM_XMM_XMMM128 = 2979;
	/**
	 * {@code VPMAXSD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 3D /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMAXSD_YMM_YMM_YMMM256 = 2980;
	/**
	 * {@code VPMAXSD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 3D /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXSD_XMM_K1Z_XMM_XMMM128B32 = 2981;
	/**
	 * {@code VPMAXSD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 3D /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXSD_YMM_K1Z_YMM_YMMM256B32 = 2982;
	/**
	 * {@code VPMAXSD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 3D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXSD_ZMM_K1Z_ZMM_ZMMM512B32 = 2983;
	/**
	 * {@code VPMAXSQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 3D /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXSQ_XMM_K1Z_XMM_XMMM128B64 = 2984;
	/**
	 * {@code VPMAXSQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 3D /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXSQ_YMM_K1Z_YMM_YMMM256B64 = 2985;
	/**
	 * {@code VPMAXSQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 3D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXSQ_ZMM_K1Z_ZMM_ZMMM512B64 = 2986;
	/**
	 * {@code PMAXUW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 3E /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMAXUW_XMM_XMMM128 = 2987;
	/**
	 * {@code VPMAXUW xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 3E /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMAXUW_XMM_XMM_XMMM128 = 2988;
	/**
	 * {@code VPMAXUW ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 3E /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMAXUW_YMM_YMM_YMMM256 = 2989;
	/**
	 * {@code VPMAXUW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.WIG 3E /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXUW_XMM_K1Z_XMM_XMMM128 = 2990;
	/**
	 * {@code VPMAXUW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.WIG 3E /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXUW_YMM_K1Z_YMM_YMMM256 = 2991;
	/**
	 * {@code VPMAXUW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.WIG 3E /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXUW_ZMM_K1Z_ZMM_ZMMM512 = 2992;
	/**
	 * {@code PMAXUD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 3F /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMAXUD_XMM_XMMM128 = 2993;
	/**
	 * {@code VPMAXUD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 3F /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMAXUD_XMM_XMM_XMMM128 = 2994;
	/**
	 * {@code VPMAXUD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 3F /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMAXUD_YMM_YMM_YMMM256 = 2995;
	/**
	 * {@code VPMAXUD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 3F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXUD_XMM_K1Z_XMM_XMMM128B32 = 2996;
	/**
	 * {@code VPMAXUD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 3F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXUD_YMM_K1Z_YMM_YMMM256B32 = 2997;
	/**
	 * {@code VPMAXUD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 3F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXUD_ZMM_K1Z_ZMM_ZMMM512B32 = 2998;
	/**
	 * {@code VPMAXUQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 3F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXUQ_XMM_K1Z_XMM_XMMM128B64 = 2999;
	/**
	 * {@code VPMAXUQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 3F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXUQ_YMM_K1Z_YMM_YMMM256B64 = 3000;
	/**
	 * {@code VPMAXUQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 3F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMAXUQ_ZMM_K1Z_ZMM_ZMMM512B64 = 3001;
	/**
	 * {@code PMULLD xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 40 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PMULLD_XMM_XMMM128 = 3002;
	/**
	 * {@code VPMULLD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 40 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMULLD_XMM_XMM_XMMM128 = 3003;
	/**
	 * {@code VPMULLD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG 40 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMULLD_YMM_YMM_YMMM256 = 3004;
	/**
	 * {@code VPMULLD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 40 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULLD_XMM_K1Z_XMM_XMMM128B32 = 3005;
	/**
	 * {@code VPMULLD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 40 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULLD_YMM_K1Z_YMM_YMMM256B32 = 3006;
	/**
	 * {@code VPMULLD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 40 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULLD_ZMM_K1Z_ZMM_ZMMM512B32 = 3007;
	/**
	 * {@code VPMULLQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 40 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULLQ_XMM_K1Z_XMM_XMMM128B64 = 3008;
	/**
	 * {@code VPMULLQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 40 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULLQ_YMM_K1Z_YMM_YMMM256B64 = 3009;
	/**
	 * {@code VPMULLQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 40 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULLQ_ZMM_K1Z_ZMM_ZMMM512B64 = 3010;
	/**
	 * {@code PHMINPOSUW xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 41 /r}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PHMINPOSUW_XMM_XMMM128 = 3011;
	/**
	 * {@code VPHMINPOSUW xmm1, xmm2/m128}<p>
	 * {@code VEX.128.66.0F38.WIG 41 /r}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPHMINPOSUW_XMM_XMMM128 = 3012;
	/**
	 * {@code VGETEXPPS xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 42 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETEXPPS_XMM_K1Z_XMMM128B32 = 3013;
	/**
	 * {@code VGETEXPPS ymm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 42 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETEXPPS_YMM_K1Z_YMMM256B32 = 3014;
	/**
	 * {@code VGETEXPPS zmm1 {k1}{z}, zmm2/m512/m32bcst{sae}}<p>
	 * {@code EVEX.512.66.0F38.W0 42 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETEXPPS_ZMM_K1Z_ZMMM512B32_SAE = 3015;
	/**
	 * {@code VGETEXPPD xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 42 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETEXPPD_XMM_K1Z_XMMM128B64 = 3016;
	/**
	 * {@code VGETEXPPD ymm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 42 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETEXPPD_YMM_K1Z_YMMM256B64 = 3017;
	/**
	 * {@code VGETEXPPD zmm1 {k1}{z}, zmm2/m512/m64bcst{sae}}<p>
	 * {@code EVEX.512.66.0F38.W1 42 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETEXPPD_ZMM_K1Z_ZMMM512B64_SAE = 3018;
	/**
	 * {@code VGETEXPSS xmm1 {k1}{z}, xmm2, xmm3/m32{sae}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 43 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETEXPSS_XMM_K1Z_XMM_XMMM32_SAE = 3019;
	/**
	 * {@code VGETEXPSD xmm1 {k1}{z}, xmm2, xmm3/m64{sae}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 43 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETEXPSD_XMM_K1Z_XMM_XMMM64_SAE = 3020;
	/**
	 * {@code VPLZCNTD xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 44 /r}<p>
	 * {@code AVX512VL and AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPLZCNTD_XMM_K1Z_XMMM128B32 = 3021;
	/**
	 * {@code VPLZCNTD ymm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 44 /r}<p>
	 * {@code AVX512VL and AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPLZCNTD_YMM_K1Z_YMMM256B32 = 3022;
	/**
	 * {@code VPLZCNTD zmm1 {k1}{z}, zmm2/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 44 /r}<p>
	 * {@code AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPLZCNTD_ZMM_K1Z_ZMMM512B32 = 3023;
	/**
	 * {@code VPLZCNTQ xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 44 /r}<p>
	 * {@code AVX512VL and AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPLZCNTQ_XMM_K1Z_XMMM128B64 = 3024;
	/**
	 * {@code VPLZCNTQ ymm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 44 /r}<p>
	 * {@code AVX512VL and AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPLZCNTQ_YMM_K1Z_YMMM256B64 = 3025;
	/**
	 * {@code VPLZCNTQ zmm1 {k1}{z}, zmm2/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 44 /r}<p>
	 * {@code AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPLZCNTQ_ZMM_K1Z_ZMMM512B64 = 3026;
	/**
	 * {@code VPSRLVD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 45 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLVD_XMM_XMM_XMMM128 = 3027;
	/**
	 * {@code VPSRLVD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 45 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLVD_YMM_YMM_YMMM256 = 3028;
	/**
	 * {@code VPSRLVQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 45 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLVQ_XMM_XMM_XMMM128 = 3029;
	/**
	 * {@code VPSRLVQ ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 45 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRLVQ_YMM_YMM_YMMM256 = 3030;
	/**
	 * {@code VPSRLVD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 45 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLVD_XMM_K1Z_XMM_XMMM128B32 = 3031;
	/**
	 * {@code VPSRLVD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 45 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLVD_YMM_K1Z_YMM_YMMM256B32 = 3032;
	/**
	 * {@code VPSRLVD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 45 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLVD_ZMM_K1Z_ZMM_ZMMM512B32 = 3033;
	/**
	 * {@code VPSRLVQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 45 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLVQ_XMM_K1Z_XMM_XMMM128B64 = 3034;
	/**
	 * {@code VPSRLVQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 45 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLVQ_YMM_K1Z_YMM_YMMM256B64 = 3035;
	/**
	 * {@code VPSRLVQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 45 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRLVQ_ZMM_K1Z_ZMM_ZMMM512B64 = 3036;
	/**
	 * {@code VPSRAVD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 46 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRAVD_XMM_XMM_XMMM128 = 3037;
	/**
	 * {@code VPSRAVD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 46 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSRAVD_YMM_YMM_YMMM256 = 3038;
	/**
	 * {@code VPSRAVD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 46 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAVD_XMM_K1Z_XMM_XMMM128B32 = 3039;
	/**
	 * {@code VPSRAVD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 46 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAVD_YMM_K1Z_YMM_YMMM256B32 = 3040;
	/**
	 * {@code VPSRAVD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 46 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAVD_ZMM_K1Z_ZMM_ZMMM512B32 = 3041;
	/**
	 * {@code VPSRAVQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 46 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAVQ_XMM_K1Z_XMM_XMMM128B64 = 3042;
	/**
	 * {@code VPSRAVQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 46 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAVQ_YMM_K1Z_YMM_YMMM256B64 = 3043;
	/**
	 * {@code VPSRAVQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 46 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSRAVQ_ZMM_K1Z_ZMM_ZMMM512B64 = 3044;
	/**
	 * {@code VPSLLVD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 47 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLVD_XMM_XMM_XMMM128 = 3045;
	/**
	 * {@code VPSLLVD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 47 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLVD_YMM_YMM_YMMM256 = 3046;
	/**
	 * {@code VPSLLVQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 47 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLVQ_XMM_XMM_XMMM128 = 3047;
	/**
	 * {@code VPSLLVQ ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 47 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPSLLVQ_YMM_YMM_YMMM256 = 3048;
	/**
	 * {@code VPSLLVD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 47 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLVD_XMM_K1Z_XMM_XMMM128B32 = 3049;
	/**
	 * {@code VPSLLVD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 47 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLVD_YMM_K1Z_YMM_YMMM256B32 = 3050;
	/**
	 * {@code VPSLLVD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 47 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLVD_ZMM_K1Z_ZMM_ZMMM512B32 = 3051;
	/**
	 * {@code VPSLLVQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 47 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLVQ_XMM_K1Z_XMM_XMMM128B64 = 3052;
	/**
	 * {@code VPSLLVQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 47 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLVQ_YMM_K1Z_YMM_YMMM256B64 = 3053;
	/**
	 * {@code VPSLLVQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 47 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSLLVQ_ZMM_K1Z_ZMM_ZMMM512B64 = 3054;
	/**
	 * {@code VRCP14PS xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 4C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCP14PS_XMM_K1Z_XMMM128B32 = 3055;
	/**
	 * {@code VRCP14PS ymm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 4C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCP14PS_YMM_K1Z_YMMM256B32 = 3056;
	/**
	 * {@code VRCP14PS zmm1 {k1}{z}, zmm2/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 4C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCP14PS_ZMM_K1Z_ZMMM512B32 = 3057;
	/**
	 * {@code VRCP14PD xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 4C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCP14PD_XMM_K1Z_XMMM128B64 = 3058;
	/**
	 * {@code VRCP14PD ymm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 4C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCP14PD_YMM_K1Z_YMMM256B64 = 3059;
	/**
	 * {@code VRCP14PD zmm1 {k1}{z}, zmm2/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 4C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCP14PD_ZMM_K1Z_ZMMM512B64 = 3060;
	/**
	 * {@code VRCP14SS xmm1 {k1}{z}, xmm2, xmm3/m32}<p>
	 * {@code EVEX.LIG.66.0F38.W0 4D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCP14SS_XMM_K1Z_XMM_XMMM32 = 3061;
	/**
	 * {@code VRCP14SD xmm1 {k1}{z}, xmm2, xmm3/m64}<p>
	 * {@code EVEX.LIG.66.0F38.W1 4D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCP14SD_XMM_K1Z_XMM_XMMM64 = 3062;
	/**
	 * {@code VRSQRT14PS xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 4E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRT14PS_XMM_K1Z_XMMM128B32 = 3063;
	/**
	 * {@code VRSQRT14PS ymm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 4E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRT14PS_YMM_K1Z_YMMM256B32 = 3064;
	/**
	 * {@code VRSQRT14PS zmm1 {k1}{z}, zmm2/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 4E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRT14PS_ZMM_K1Z_ZMMM512B32 = 3065;
	/**
	 * {@code VRSQRT14PD xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 4E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRT14PD_XMM_K1Z_XMMM128B64 = 3066;
	/**
	 * {@code VRSQRT14PD ymm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 4E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRT14PD_YMM_K1Z_YMMM256B64 = 3067;
	/**
	 * {@code VRSQRT14PD zmm1 {k1}{z}, zmm2/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 4E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRT14PD_ZMM_K1Z_ZMMM512B64 = 3068;
	/**
	 * {@code VRSQRT14SS xmm1 {k1}{z}, xmm2, xmm3/m32}<p>
	 * {@code EVEX.LIG.66.0F38.W0 4F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRT14SS_XMM_K1Z_XMM_XMMM32 = 3069;
	/**
	 * {@code VRSQRT14SD xmm1 {k1}{z}, xmm2, xmm3/m64}<p>
	 * {@code EVEX.LIG.66.0F38.W1 4F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRT14SD_XMM_K1Z_XMM_XMMM64 = 3070;
	/**
	 * {@code VPDPBUSD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 50 /r}<p>
	 * {@code AVX512VL and AVX512_VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPDPBUSD_XMM_K1Z_XMM_XMMM128B32 = 3071;
	/**
	 * {@code VPDPBUSD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 50 /r}<p>
	 * {@code AVX512VL and AVX512_VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPDPBUSD_YMM_K1Z_YMM_YMMM256B32 = 3072;
	/**
	 * {@code VPDPBUSD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 50 /r}<p>
	 * {@code AVX512_VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPDPBUSD_ZMM_K1Z_ZMM_ZMMM512B32 = 3073;
	/**
	 * {@code VPDPBUSDS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 51 /r}<p>
	 * {@code AVX512VL and AVX512_VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPDPBUSDS_XMM_K1Z_XMM_XMMM128B32 = 3074;
	/**
	 * {@code VPDPBUSDS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 51 /r}<p>
	 * {@code AVX512VL and AVX512_VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPDPBUSDS_YMM_K1Z_YMM_YMMM256B32 = 3075;
	/**
	 * {@code VPDPBUSDS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 51 /r}<p>
	 * {@code AVX512_VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPDPBUSDS_ZMM_K1Z_ZMM_ZMMM512B32 = 3076;
	/**
	 * {@code VPDPWSSD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 52 /r}<p>
	 * {@code AVX512VL and AVX512_VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPDPWSSD_XMM_K1Z_XMM_XMMM128B32 = 3077;
	/**
	 * {@code VPDPWSSD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 52 /r}<p>
	 * {@code AVX512VL and AVX512_VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPDPWSSD_YMM_K1Z_YMM_YMMM256B32 = 3078;
	/**
	 * {@code VPDPWSSD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 52 /r}<p>
	 * {@code AVX512_VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPDPWSSD_ZMM_K1Z_ZMM_ZMMM512B32 = 3079;
	/**
	 * {@code VDPBF16PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.F3.0F38.W0 52 /r}<p>
	 * {@code AVX512VL and AVX512_BF16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDPBF16PS_XMM_K1Z_XMM_XMMM128B32 = 3080;
	/**
	 * {@code VDPBF16PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.F3.0F38.W0 52 /r}<p>
	 * {@code AVX512VL and AVX512_BF16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDPBF16PS_YMM_K1Z_YMM_YMMM256B32 = 3081;
	/**
	 * {@code VDPBF16PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.F3.0F38.W0 52 /r}<p>
	 * {@code AVX512F and AVX512_BF16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDPBF16PS_ZMM_K1Z_ZMM_ZMMM512B32 = 3082;
	/**
	 * {@code VP4DPWSSD zmm1 {k1}{z}, zmm2+3, m128}<p>
	 * {@code EVEX.512.F2.0F38.W0 52 /r}<p>
	 * {@code AVX512_4VNNIW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VP4DPWSSD_ZMM_K1Z_ZMMP3_M128 = 3083;
	/**
	 * {@code VPDPWSSDS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 53 /r}<p>
	 * {@code AVX512VL and AVX512_VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPDPWSSDS_XMM_K1Z_XMM_XMMM128B32 = 3084;
	/**
	 * {@code VPDPWSSDS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 53 /r}<p>
	 * {@code AVX512VL and AVX512_VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPDPWSSDS_YMM_K1Z_YMM_YMMM256B32 = 3085;
	/**
	 * {@code VPDPWSSDS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 53 /r}<p>
	 * {@code AVX512_VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPDPWSSDS_ZMM_K1Z_ZMM_ZMMM512B32 = 3086;
	/**
	 * {@code VP4DPWSSDS zmm1 {k1}{z}, zmm2+3, m128}<p>
	 * {@code EVEX.512.F2.0F38.W0 53 /r}<p>
	 * {@code AVX512_4VNNIW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VP4DPWSSDS_ZMM_K1Z_ZMMP3_M128 = 3087;
	/**
	 * {@code VPOPCNTB xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.66.0F38.W0 54 /r}<p>
	 * {@code AVX512VL and AVX512_BITALG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPOPCNTB_XMM_K1Z_XMMM128 = 3088;
	/**
	 * {@code VPOPCNTB ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.66.0F38.W0 54 /r}<p>
	 * {@code AVX512VL and AVX512_BITALG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPOPCNTB_YMM_K1Z_YMMM256 = 3089;
	/**
	 * {@code VPOPCNTB zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.66.0F38.W0 54 /r}<p>
	 * {@code AVX512_BITALG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPOPCNTB_ZMM_K1Z_ZMMM512 = 3090;
	/**
	 * {@code VPOPCNTW xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.66.0F38.W1 54 /r}<p>
	 * {@code AVX512VL and AVX512_BITALG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPOPCNTW_XMM_K1Z_XMMM128 = 3091;
	/**
	 * {@code VPOPCNTW ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.66.0F38.W1 54 /r}<p>
	 * {@code AVX512VL and AVX512_BITALG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPOPCNTW_YMM_K1Z_YMMM256 = 3092;
	/**
	 * {@code VPOPCNTW zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.66.0F38.W1 54 /r}<p>
	 * {@code AVX512_BITALG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPOPCNTW_ZMM_K1Z_ZMMM512 = 3093;
	/**
	 * {@code VPOPCNTD xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 55 /r}<p>
	 * {@code AVX512VL and AVX512_VPOPCNTDQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPOPCNTD_XMM_K1Z_XMMM128B32 = 3094;
	/**
	 * {@code VPOPCNTD ymm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 55 /r}<p>
	 * {@code AVX512VL and AVX512_VPOPCNTDQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPOPCNTD_YMM_K1Z_YMMM256B32 = 3095;
	/**
	 * {@code VPOPCNTD zmm1 {k1}{z}, zmm2/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 55 /r}<p>
	 * {@code AVX512_VPOPCNTDQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPOPCNTD_ZMM_K1Z_ZMMM512B32 = 3096;
	/**
	 * {@code VPOPCNTQ xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 55 /r}<p>
	 * {@code AVX512VL and AVX512_VPOPCNTDQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPOPCNTQ_XMM_K1Z_XMMM128B64 = 3097;
	/**
	 * {@code VPOPCNTQ ymm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 55 /r}<p>
	 * {@code AVX512VL and AVX512_VPOPCNTDQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPOPCNTQ_YMM_K1Z_YMMM256B64 = 3098;
	/**
	 * {@code VPOPCNTQ zmm1 {k1}{z}, zmm2/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 55 /r}<p>
	 * {@code AVX512_VPOPCNTDQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPOPCNTQ_ZMM_K1Z_ZMMM512B64 = 3099;
	/**
	 * {@code VPBROADCASTD xmm1, xmm2/m32}<p>
	 * {@code VEX.128.66.0F38.W0 58 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPBROADCASTD_XMM_XMMM32 = 3100;
	/**
	 * {@code VPBROADCASTD ymm1, xmm2/m32}<p>
	 * {@code VEX.256.66.0F38.W0 58 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPBROADCASTD_YMM_XMMM32 = 3101;
	/**
	 * {@code VPBROADCASTD xmm1 {k1}{z}, xmm2/m32}<p>
	 * {@code EVEX.128.66.0F38.W0 58 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTD_XMM_K1Z_XMMM32 = 3102;
	/**
	 * {@code VPBROADCASTD ymm1 {k1}{z}, xmm2/m32}<p>
	 * {@code EVEX.256.66.0F38.W0 58 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTD_YMM_K1Z_XMMM32 = 3103;
	/**
	 * {@code VPBROADCASTD zmm1 {k1}{z}, xmm2/m32}<p>
	 * {@code EVEX.512.66.0F38.W0 58 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTD_ZMM_K1Z_XMMM32 = 3104;
	/**
	 * {@code VPBROADCASTQ xmm1, xmm2/m64}<p>
	 * {@code VEX.128.66.0F38.W0 59 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPBROADCASTQ_XMM_XMMM64 = 3105;
	/**
	 * {@code VPBROADCASTQ ymm1, xmm2/m64}<p>
	 * {@code VEX.256.66.0F38.W0 59 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPBROADCASTQ_YMM_XMMM64 = 3106;
	/**
	 * {@code VBROADCASTI32X2 xmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.128.66.0F38.W0 59 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTI32X2_XMM_K1Z_XMMM64 = 3107;
	/**
	 * {@code VBROADCASTI32X2 ymm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.256.66.0F38.W0 59 /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTI32X2_YMM_K1Z_XMMM64 = 3108;
	/**
	 * {@code VBROADCASTI32X2 zmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.512.66.0F38.W0 59 /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTI32X2_ZMM_K1Z_XMMM64 = 3109;
	/**
	 * {@code VPBROADCASTQ xmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.128.66.0F38.W1 59 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTQ_XMM_K1Z_XMMM64 = 3110;
	/**
	 * {@code VPBROADCASTQ ymm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.256.66.0F38.W1 59 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTQ_YMM_K1Z_XMMM64 = 3111;
	/**
	 * {@code VPBROADCASTQ zmm1 {k1}{z}, xmm2/m64}<p>
	 * {@code EVEX.512.66.0F38.W1 59 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTQ_ZMM_K1Z_XMMM64 = 3112;
	/**
	 * {@code VBROADCASTI128 ymm1, m128}<p>
	 * {@code VEX.256.66.0F38.W0 5A /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBROADCASTI128_YMM_M128 = 3113;
	/**
	 * {@code VBROADCASTI32X4 ymm1 {k1}{z}, m128}<p>
	 * {@code EVEX.256.66.0F38.W0 5A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTI32X4_YMM_K1Z_M128 = 3114;
	/**
	 * {@code VBROADCASTI32X4 zmm1 {k1}{z}, m128}<p>
	 * {@code EVEX.512.66.0F38.W0 5A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTI32X4_ZMM_K1Z_M128 = 3115;
	/**
	 * {@code VBROADCASTI64X2 ymm1 {k1}{z}, m128}<p>
	 * {@code EVEX.256.66.0F38.W1 5A /r}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTI64X2_YMM_K1Z_M128 = 3116;
	/**
	 * {@code VBROADCASTI64X2 zmm1 {k1}{z}, m128}<p>
	 * {@code EVEX.512.66.0F38.W1 5A /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTI64X2_ZMM_K1Z_M128 = 3117;
	/**
	 * {@code VBROADCASTI32X8 zmm1 {k1}{z}, m256}<p>
	 * {@code EVEX.512.66.0F38.W0 5B /r}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTI32X8_ZMM_K1Z_M256 = 3118;
	/**
	 * {@code VBROADCASTI64X4 zmm1 {k1}{z}, m256}<p>
	 * {@code EVEX.512.66.0F38.W1 5B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBROADCASTI64X4_ZMM_K1Z_M256 = 3119;
	/**
	 * {@code VPEXPANDB xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.66.0F38.W0 62 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXPANDB_XMM_K1Z_XMMM128 = 3120;
	/**
	 * {@code VPEXPANDB ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.66.0F38.W0 62 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXPANDB_YMM_K1Z_YMMM256 = 3121;
	/**
	 * {@code VPEXPANDB zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.66.0F38.W0 62 /r}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXPANDB_ZMM_K1Z_ZMMM512 = 3122;
	/**
	 * {@code VPEXPANDW xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.66.0F38.W1 62 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXPANDW_XMM_K1Z_XMMM128 = 3123;
	/**
	 * {@code VPEXPANDW ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.66.0F38.W1 62 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXPANDW_YMM_K1Z_YMMM256 = 3124;
	/**
	 * {@code VPEXPANDW zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.66.0F38.W1 62 /r}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXPANDW_ZMM_K1Z_ZMMM512 = 3125;
	/**
	 * {@code VPCOMPRESSB xmm1/m128 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.66.0F38.W0 63 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCOMPRESSB_XMMM128_K1Z_XMM = 3126;
	/**
	 * {@code VPCOMPRESSB ymm1/m256 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.66.0F38.W0 63 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCOMPRESSB_YMMM256_K1Z_YMM = 3127;
	/**
	 * {@code VPCOMPRESSB zmm1/m512 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.66.0F38.W0 63 /r}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCOMPRESSB_ZMMM512_K1Z_ZMM = 3128;
	/**
	 * {@code VPCOMPRESSW xmm1/m128 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.66.0F38.W1 63 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCOMPRESSW_XMMM128_K1Z_XMM = 3129;
	/**
	 * {@code VPCOMPRESSW ymm1/m256 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.66.0F38.W1 63 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCOMPRESSW_YMMM256_K1Z_YMM = 3130;
	/**
	 * {@code VPCOMPRESSW zmm1/m512 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.66.0F38.W1 63 /r}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCOMPRESSW_ZMMM512_K1Z_ZMM = 3131;
	/**
	 * {@code VPBLENDMD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 64 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBLENDMD_XMM_K1Z_XMM_XMMM128B32 = 3132;
	/**
	 * {@code VPBLENDMD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 64 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBLENDMD_YMM_K1Z_YMM_YMMM256B32 = 3133;
	/**
	 * {@code VPBLENDMD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 64 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBLENDMD_ZMM_K1Z_ZMM_ZMMM512B32 = 3134;
	/**
	 * {@code VPBLENDMQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 64 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBLENDMQ_XMM_K1Z_XMM_XMMM128B64 = 3135;
	/**
	 * {@code VPBLENDMQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 64 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBLENDMQ_YMM_K1Z_YMM_YMMM256B64 = 3136;
	/**
	 * {@code VPBLENDMQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 64 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBLENDMQ_ZMM_K1Z_ZMM_ZMMM512B64 = 3137;
	/**
	 * {@code VBLENDMPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 65 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBLENDMPS_XMM_K1Z_XMM_XMMM128B32 = 3138;
	/**
	 * {@code VBLENDMPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 65 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBLENDMPS_YMM_K1Z_YMM_YMMM256B32 = 3139;
	/**
	 * {@code VBLENDMPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 65 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBLENDMPS_ZMM_K1Z_ZMM_ZMMM512B32 = 3140;
	/**
	 * {@code VBLENDMPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 65 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBLENDMPD_XMM_K1Z_XMM_XMMM128B64 = 3141;
	/**
	 * {@code VBLENDMPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 65 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBLENDMPD_YMM_K1Z_YMM_YMMM256B64 = 3142;
	/**
	 * {@code VBLENDMPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 65 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VBLENDMPD_ZMM_K1Z_ZMM_ZMMM512B64 = 3143;
	/**
	 * {@code VPBLENDMB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W0 66 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBLENDMB_XMM_K1Z_XMM_XMMM128 = 3144;
	/**
	 * {@code VPBLENDMB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W0 66 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBLENDMB_YMM_K1Z_YMM_YMMM256 = 3145;
	/**
	 * {@code VPBLENDMB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W0 66 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBLENDMB_ZMM_K1Z_ZMM_ZMMM512 = 3146;
	/**
	 * {@code VPBLENDMW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W1 66 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBLENDMW_XMM_K1Z_XMM_XMMM128 = 3147;
	/**
	 * {@code VPBLENDMW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W1 66 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBLENDMW_YMM_K1Z_YMM_YMMM256 = 3148;
	/**
	 * {@code VPBLENDMW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W1 66 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBLENDMW_ZMM_K1Z_ZMM_ZMMM512 = 3149;
	/**
	 * {@code VP2INTERSECTD k1+1, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.F2.0F38.W0 68 /r}<p>
	 * {@code AVX512VL and AVX512_VP2INTERSECT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VP2INTERSECTD_KP1_XMM_XMMM128B32 = 3150;
	/**
	 * {@code VP2INTERSECTD k1+1, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.F2.0F38.W0 68 /r}<p>
	 * {@code AVX512VL and AVX512_VP2INTERSECT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VP2INTERSECTD_KP1_YMM_YMMM256B32 = 3151;
	/**
	 * {@code VP2INTERSECTD k1+1, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.F2.0F38.W0 68 /r}<p>
	 * {@code AVX512F and AVX512_VP2INTERSECT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VP2INTERSECTD_KP1_ZMM_ZMMM512B32 = 3152;
	/**
	 * {@code VP2INTERSECTQ k1+1, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.F2.0F38.W1 68 /r}<p>
	 * {@code AVX512VL and AVX512_VP2INTERSECT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VP2INTERSECTQ_KP1_XMM_XMMM128B64 = 3153;
	/**
	 * {@code VP2INTERSECTQ k1+1, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.F2.0F38.W1 68 /r}<p>
	 * {@code AVX512VL and AVX512_VP2INTERSECT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VP2INTERSECTQ_KP1_YMM_YMMM256B64 = 3154;
	/**
	 * {@code VP2INTERSECTQ k1+1, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.F2.0F38.W1 68 /r}<p>
	 * {@code AVX512F and AVX512_VP2INTERSECT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VP2INTERSECTQ_KP1_ZMM_ZMMM512B64 = 3155;
	/**
	 * {@code VPSHLDVW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W1 70 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDVW_XMM_K1Z_XMM_XMMM128 = 3156;
	/**
	 * {@code VPSHLDVW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W1 70 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDVW_YMM_K1Z_YMM_YMMM256 = 3157;
	/**
	 * {@code VPSHLDVW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W1 70 /r}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDVW_ZMM_K1Z_ZMM_ZMMM512 = 3158;
	/**
	 * {@code VPSHLDVD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 71 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDVD_XMM_K1Z_XMM_XMMM128B32 = 3159;
	/**
	 * {@code VPSHLDVD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 71 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDVD_YMM_K1Z_YMM_YMMM256B32 = 3160;
	/**
	 * {@code VPSHLDVD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 71 /r}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDVD_ZMM_K1Z_ZMM_ZMMM512B32 = 3161;
	/**
	 * {@code VPSHLDVQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 71 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDVQ_XMM_K1Z_XMM_XMMM128B64 = 3162;
	/**
	 * {@code VPSHLDVQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 71 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDVQ_YMM_K1Z_YMM_YMMM256B64 = 3163;
	/**
	 * {@code VPSHLDVQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 71 /r}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDVQ_ZMM_K1Z_ZMM_ZMMM512B64 = 3164;
	/**
	 * {@code VPSHRDVW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W1 72 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDVW_XMM_K1Z_XMM_XMMM128 = 3165;
	/**
	 * {@code VPSHRDVW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W1 72 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDVW_YMM_K1Z_YMM_YMMM256 = 3166;
	/**
	 * {@code VPSHRDVW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W1 72 /r}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDVW_ZMM_K1Z_ZMM_ZMMM512 = 3167;
	/**
	 * {@code VCVTNEPS2BF16 xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.F3.0F38.W0 72 /r}<p>
	 * {@code AVX512VL and AVX512_BF16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTNEPS2BF16_XMM_K1Z_XMMM128B32 = 3168;
	/**
	 * {@code VCVTNEPS2BF16 xmm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.F3.0F38.W0 72 /r}<p>
	 * {@code AVX512VL and AVX512_BF16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTNEPS2BF16_XMM_K1Z_YMMM256B32 = 3169;
	/**
	 * {@code VCVTNEPS2BF16 ymm1 {k1}{z}, zmm2/m512/m32bcst}<p>
	 * {@code EVEX.512.F3.0F38.W0 72 /r}<p>
	 * {@code AVX512F and AVX512_BF16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTNEPS2BF16_YMM_K1Z_ZMMM512B32 = 3170;
	/**
	 * {@code VCVTNE2PS2BF16 xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.F2.0F38.W0 72 /r}<p>
	 * {@code AVX512VL and AVX512_BF16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTNE2PS2BF16_XMM_K1Z_XMM_XMMM128B32 = 3171;
	/**
	 * {@code VCVTNE2PS2BF16 ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.F2.0F38.W0 72 /r}<p>
	 * {@code AVX512VL and AVX512_BF16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTNE2PS2BF16_YMM_K1Z_YMM_YMMM256B32 = 3172;
	/**
	 * {@code VCVTNE2PS2BF16 zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.F2.0F38.W0 72 /r}<p>
	 * {@code AVX512F and AVX512_BF16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTNE2PS2BF16_ZMM_K1Z_ZMM_ZMMM512B32 = 3173;
	/**
	 * {@code VPSHRDVD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 73 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDVD_XMM_K1Z_XMM_XMMM128B32 = 3174;
	/**
	 * {@code VPSHRDVD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 73 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDVD_YMM_K1Z_YMM_YMMM256B32 = 3175;
	/**
	 * {@code VPSHRDVD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 73 /r}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDVD_ZMM_K1Z_ZMM_ZMMM512B32 = 3176;
	/**
	 * {@code VPSHRDVQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 73 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDVQ_XMM_K1Z_XMM_XMMM128B64 = 3177;
	/**
	 * {@code VPSHRDVQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 73 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDVQ_YMM_K1Z_YMM_YMMM256B64 = 3178;
	/**
	 * {@code VPSHRDVQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 73 /r}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDVQ_ZMM_K1Z_ZMM_ZMMM512B64 = 3179;
	/**
	 * {@code VPERMI2B xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W0 75 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2B_XMM_K1Z_XMM_XMMM128 = 3180;
	/**
	 * {@code VPERMI2B ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W0 75 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2B_YMM_K1Z_YMM_YMMM256 = 3181;
	/**
	 * {@code VPERMI2B zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W0 75 /r}<p>
	 * {@code AVX512_VBMI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2B_ZMM_K1Z_ZMM_ZMMM512 = 3182;
	/**
	 * {@code VPERMI2W xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W1 75 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2W_XMM_K1Z_XMM_XMMM128 = 3183;
	/**
	 * {@code VPERMI2W ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W1 75 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2W_YMM_K1Z_YMM_YMMM256 = 3184;
	/**
	 * {@code VPERMI2W zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W1 75 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2W_ZMM_K1Z_ZMM_ZMMM512 = 3185;
	/**
	 * {@code VPERMI2D xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 76 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2D_XMM_K1Z_XMM_XMMM128B32 = 3186;
	/**
	 * {@code VPERMI2D ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 76 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2D_YMM_K1Z_YMM_YMMM256B32 = 3187;
	/**
	 * {@code VPERMI2D zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 76 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2D_ZMM_K1Z_ZMM_ZMMM512B32 = 3188;
	/**
	 * {@code VPERMI2Q xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 76 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2Q_XMM_K1Z_XMM_XMMM128B64 = 3189;
	/**
	 * {@code VPERMI2Q ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 76 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2Q_YMM_K1Z_YMM_YMMM256B64 = 3190;
	/**
	 * {@code VPERMI2Q zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 76 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2Q_ZMM_K1Z_ZMM_ZMMM512B64 = 3191;
	/**
	 * {@code VPERMI2PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 77 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2PS_XMM_K1Z_XMM_XMMM128B32 = 3192;
	/**
	 * {@code VPERMI2PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 77 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2PS_YMM_K1Z_YMM_YMMM256B32 = 3193;
	/**
	 * {@code VPERMI2PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 77 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2PS_ZMM_K1Z_ZMM_ZMMM512B32 = 3194;
	/**
	 * {@code VPERMI2PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 77 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2PD_XMM_K1Z_XMM_XMMM128B64 = 3195;
	/**
	 * {@code VPERMI2PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 77 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2PD_YMM_K1Z_YMM_YMMM256B64 = 3196;
	/**
	 * {@code VPERMI2PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 77 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMI2PD_ZMM_K1Z_ZMM_ZMMM512B64 = 3197;
	/**
	 * {@code VPBROADCASTB xmm1, xmm2/m8}<p>
	 * {@code VEX.128.66.0F38.W0 78 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPBROADCASTB_XMM_XMMM8 = 3198;
	/**
	 * {@code VPBROADCASTB ymm1, xmm2/m8}<p>
	 * {@code VEX.256.66.0F38.W0 78 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPBROADCASTB_YMM_XMMM8 = 3199;
	/**
	 * {@code VPBROADCASTB xmm1 {k1}{z}, xmm2/m8}<p>
	 * {@code EVEX.128.66.0F38.W0 78 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTB_XMM_K1Z_XMMM8 = 3200;
	/**
	 * {@code VPBROADCASTB ymm1 {k1}{z}, xmm2/m8}<p>
	 * {@code EVEX.256.66.0F38.W0 78 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTB_YMM_K1Z_XMMM8 = 3201;
	/**
	 * {@code VPBROADCASTB zmm1 {k1}{z}, xmm2/m8}<p>
	 * {@code EVEX.512.66.0F38.W0 78 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTB_ZMM_K1Z_XMMM8 = 3202;
	/**
	 * {@code VPBROADCASTW xmm1, xmm2/m16}<p>
	 * {@code VEX.128.66.0F38.W0 79 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPBROADCASTW_XMM_XMMM16 = 3203;
	/**
	 * {@code VPBROADCASTW ymm1, xmm2/m16}<p>
	 * {@code VEX.256.66.0F38.W0 79 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPBROADCASTW_YMM_XMMM16 = 3204;
	/**
	 * {@code VPBROADCASTW xmm1 {k1}{z}, xmm2/m16}<p>
	 * {@code EVEX.128.66.0F38.W0 79 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTW_XMM_K1Z_XMMM16 = 3205;
	/**
	 * {@code VPBROADCASTW ymm1 {k1}{z}, xmm2/m16}<p>
	 * {@code EVEX.256.66.0F38.W0 79 /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTW_YMM_K1Z_XMMM16 = 3206;
	/**
	 * {@code VPBROADCASTW zmm1 {k1}{z}, xmm2/m16}<p>
	 * {@code EVEX.512.66.0F38.W0 79 /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTW_ZMM_K1Z_XMMM16 = 3207;
	/**
	 * {@code VPBROADCASTB xmm1 {k1}{z}, r32}<p>
	 * {@code EVEX.128.66.0F38.W0 7A /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTB_XMM_K1Z_R32 = 3208;
	/**
	 * {@code VPBROADCASTB ymm1 {k1}{z}, r32}<p>
	 * {@code EVEX.256.66.0F38.W0 7A /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTB_YMM_K1Z_R32 = 3209;
	/**
	 * {@code VPBROADCASTB zmm1 {k1}{z}, r32}<p>
	 * {@code EVEX.512.66.0F38.W0 7A /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTB_ZMM_K1Z_R32 = 3210;
	/**
	 * {@code VPBROADCASTW xmm1 {k1}{z}, r32}<p>
	 * {@code EVEX.128.66.0F38.W0 7B /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTW_XMM_K1Z_R32 = 3211;
	/**
	 * {@code VPBROADCASTW ymm1 {k1}{z}, r32}<p>
	 * {@code EVEX.256.66.0F38.W0 7B /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTW_YMM_K1Z_R32 = 3212;
	/**
	 * {@code VPBROADCASTW zmm1 {k1}{z}, r32}<p>
	 * {@code EVEX.512.66.0F38.W0 7B /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTW_ZMM_K1Z_R32 = 3213;
	/**
	 * {@code VPBROADCASTD xmm1 {k1}{z}, r32}<p>
	 * {@code EVEX.128.66.0F38.W0 7C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTD_XMM_K1Z_R32 = 3214;
	/**
	 * {@code VPBROADCASTD ymm1 {k1}{z}, r32}<p>
	 * {@code EVEX.256.66.0F38.W0 7C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTD_YMM_K1Z_R32 = 3215;
	/**
	 * {@code VPBROADCASTD zmm1 {k1}{z}, r32}<p>
	 * {@code EVEX.512.66.0F38.W0 7C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPBROADCASTD_ZMM_K1Z_R32 = 3216;
	/**
	 * {@code VPBROADCASTQ xmm1 {k1}{z}, r64}<p>
	 * {@code EVEX.128.66.0F38.W1 7C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VPBROADCASTQ_XMM_K1Z_R64 = 3217;
	/**
	 * {@code VPBROADCASTQ ymm1 {k1}{z}, r64}<p>
	 * {@code EVEX.256.66.0F38.W1 7C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VPBROADCASTQ_YMM_K1Z_R64 = 3218;
	/**
	 * {@code VPBROADCASTQ zmm1 {k1}{z}, r64}<p>
	 * {@code EVEX.512.66.0F38.W1 7C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VPBROADCASTQ_ZMM_K1Z_R64 = 3219;
	/**
	 * {@code VPERMT2B xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W0 7D /r}<p>
	 * {@code AVX512VL and AVX512_VBMI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2B_XMM_K1Z_XMM_XMMM128 = 3220;
	/**
	 * {@code VPERMT2B ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W0 7D /r}<p>
	 * {@code AVX512VL and AVX512_VBMI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2B_YMM_K1Z_YMM_YMMM256 = 3221;
	/**
	 * {@code VPERMT2B zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W0 7D /r}<p>
	 * {@code AVX512_VBMI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2B_ZMM_K1Z_ZMM_ZMMM512 = 3222;
	/**
	 * {@code VPERMT2W xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W1 7D /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2W_XMM_K1Z_XMM_XMMM128 = 3223;
	/**
	 * {@code VPERMT2W ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W1 7D /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2W_YMM_K1Z_YMM_YMMM256 = 3224;
	/**
	 * {@code VPERMT2W zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W1 7D /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2W_ZMM_K1Z_ZMM_ZMMM512 = 3225;
	/**
	 * {@code VPERMT2D xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 7E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2D_XMM_K1Z_XMM_XMMM128B32 = 3226;
	/**
	 * {@code VPERMT2D ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 7E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2D_YMM_K1Z_YMM_YMMM256B32 = 3227;
	/**
	 * {@code VPERMT2D zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 7E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2D_ZMM_K1Z_ZMM_ZMMM512B32 = 3228;
	/**
	 * {@code VPERMT2Q xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 7E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2Q_XMM_K1Z_XMM_XMMM128B64 = 3229;
	/**
	 * {@code VPERMT2Q ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 7E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2Q_YMM_K1Z_YMM_YMMM256B64 = 3230;
	/**
	 * {@code VPERMT2Q zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 7E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2Q_ZMM_K1Z_ZMM_ZMMM512B64 = 3231;
	/**
	 * {@code VPERMT2PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 7F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2PS_XMM_K1Z_XMM_XMMM128B32 = 3232;
	/**
	 * {@code VPERMT2PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 7F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2PS_YMM_K1Z_YMM_YMMM256B32 = 3233;
	/**
	 * {@code VPERMT2PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 7F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2PS_ZMM_K1Z_ZMM_ZMMM512B32 = 3234;
	/**
	 * {@code VPERMT2PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 7F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2PD_XMM_K1Z_XMM_XMMM128B64 = 3235;
	/**
	 * {@code VPERMT2PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 7F /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2PD_YMM_K1Z_YMM_YMMM256B64 = 3236;
	/**
	 * {@code VPERMT2PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 7F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMT2PD_ZMM_K1Z_ZMM_ZMMM512B64 = 3237;
	/**
	 * {@code INVEPT r32, m128}<p>
	 * {@code 66 0F 38 80 /r}<p>
	 * {@code VMX and IA32_VMX_EPT_VPID_CAP[bit 20]}<p>
	 * {@code 16/32-bit}
	 */
	public static final int INVEPT_R32_M128 = 3238;
	/**
	 * {@code INVEPT r64, m128}<p>
	 * {@code 66 0F 38 80 /r}<p>
	 * {@code VMX and IA32_VMX_EPT_VPID_CAP[bit 20]}<p>
	 * {@code 64-bit}
	 */
	public static final int INVEPT_R64_M128 = 3239;
	/**
	 * {@code INVVPID r32, m128}<p>
	 * {@code 66 0F 38 81 /r}<p>
	 * {@code VMX and IA32_VMX_EPT_VPID_CAP[bit 32]}<p>
	 * {@code 16/32-bit}
	 */
	public static final int INVVPID_R32_M128 = 3240;
	/**
	 * {@code INVVPID r64, m128}<p>
	 * {@code 66 0F 38 81 /r}<p>
	 * {@code VMX and IA32_VMX_EPT_VPID_CAP[bit 32]}<p>
	 * {@code 64-bit}
	 */
	public static final int INVVPID_R64_M128 = 3241;
	/**
	 * {@code INVPCID r32, m128}<p>
	 * {@code 66 0F 38 82 /r}<p>
	 * {@code INVPCID}<p>
	 * {@code 16/32-bit}
	 */
	public static final int INVPCID_R32_M128 = 3242;
	/**
	 * {@code INVPCID r64, m128}<p>
	 * {@code 66 0F 38 82 /r}<p>
	 * {@code INVPCID}<p>
	 * {@code 64-bit}
	 */
	public static final int INVPCID_R64_M128 = 3243;
	/**
	 * {@code VPMULTISHIFTQB xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 83 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULTISHIFTQB_XMM_K1Z_XMM_XMMM128B64 = 3244;
	/**
	 * {@code VPMULTISHIFTQB ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 83 /r}<p>
	 * {@code AVX512VL and AVX512_VBMI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULTISHIFTQB_YMM_K1Z_YMM_YMMM256B64 = 3245;
	/**
	 * {@code VPMULTISHIFTQB zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 83 /r}<p>
	 * {@code AVX512_VBMI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMULTISHIFTQB_ZMM_K1Z_ZMM_ZMMM512B64 = 3246;
	/**
	 * {@code VEXPANDPS xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.66.0F38.W0 88 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXPANDPS_XMM_K1Z_XMMM128 = 3247;
	/**
	 * {@code VEXPANDPS ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.66.0F38.W0 88 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXPANDPS_YMM_K1Z_YMMM256 = 3248;
	/**
	 * {@code VEXPANDPS zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.66.0F38.W0 88 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXPANDPS_ZMM_K1Z_ZMMM512 = 3249;
	/**
	 * {@code VEXPANDPD xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.66.0F38.W1 88 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXPANDPD_XMM_K1Z_XMMM128 = 3250;
	/**
	 * {@code VEXPANDPD ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.66.0F38.W1 88 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXPANDPD_YMM_K1Z_YMMM256 = 3251;
	/**
	 * {@code VEXPANDPD zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.66.0F38.W1 88 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXPANDPD_ZMM_K1Z_ZMMM512 = 3252;
	/**
	 * {@code VPEXPANDD xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.66.0F38.W0 89 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXPANDD_XMM_K1Z_XMMM128 = 3253;
	/**
	 * {@code VPEXPANDD ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.66.0F38.W0 89 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXPANDD_YMM_K1Z_YMMM256 = 3254;
	/**
	 * {@code VPEXPANDD zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.66.0F38.W0 89 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXPANDD_ZMM_K1Z_ZMMM512 = 3255;
	/**
	 * {@code VPEXPANDQ xmm1 {k1}{z}, xmm2/m128}<p>
	 * {@code EVEX.128.66.0F38.W1 89 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXPANDQ_XMM_K1Z_XMMM128 = 3256;
	/**
	 * {@code VPEXPANDQ ymm1 {k1}{z}, ymm2/m256}<p>
	 * {@code EVEX.256.66.0F38.W1 89 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXPANDQ_YMM_K1Z_YMMM256 = 3257;
	/**
	 * {@code VPEXPANDQ zmm1 {k1}{z}, zmm2/m512}<p>
	 * {@code EVEX.512.66.0F38.W1 89 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXPANDQ_ZMM_K1Z_ZMMM512 = 3258;
	/**
	 * {@code VCOMPRESSPS xmm1/m128 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.66.0F38.W0 8A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCOMPRESSPS_XMMM128_K1Z_XMM = 3259;
	/**
	 * {@code VCOMPRESSPS ymm1/m256 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.66.0F38.W0 8A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCOMPRESSPS_YMMM256_K1Z_YMM = 3260;
	/**
	 * {@code VCOMPRESSPS zmm1/m512 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.66.0F38.W0 8A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCOMPRESSPS_ZMMM512_K1Z_ZMM = 3261;
	/**
	 * {@code VCOMPRESSPD xmm1/m128 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.66.0F38.W1 8A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCOMPRESSPD_XMMM128_K1Z_XMM = 3262;
	/**
	 * {@code VCOMPRESSPD ymm1/m256 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.66.0F38.W1 8A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCOMPRESSPD_YMMM256_K1Z_YMM = 3263;
	/**
	 * {@code VCOMPRESSPD zmm1/m512 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.66.0F38.W1 8A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCOMPRESSPD_ZMMM512_K1Z_ZMM = 3264;
	/**
	 * {@code VPCOMPRESSD xmm1/m128 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.66.0F38.W0 8B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCOMPRESSD_XMMM128_K1Z_XMM = 3265;
	/**
	 * {@code VPCOMPRESSD ymm1/m256 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.66.0F38.W0 8B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCOMPRESSD_YMMM256_K1Z_YMM = 3266;
	/**
	 * {@code VPCOMPRESSD zmm1/m512 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.66.0F38.W0 8B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCOMPRESSD_ZMMM512_K1Z_ZMM = 3267;
	/**
	 * {@code VPCOMPRESSQ xmm1/m128 {k1}{z}, xmm2}<p>
	 * {@code EVEX.128.66.0F38.W1 8B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCOMPRESSQ_XMMM128_K1Z_XMM = 3268;
	/**
	 * {@code VPCOMPRESSQ ymm1/m256 {k1}{z}, ymm2}<p>
	 * {@code EVEX.256.66.0F38.W1 8B /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCOMPRESSQ_YMMM256_K1Z_YMM = 3269;
	/**
	 * {@code VPCOMPRESSQ zmm1/m512 {k1}{z}, zmm2}<p>
	 * {@code EVEX.512.66.0F38.W1 8B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCOMPRESSQ_ZMMM512_K1Z_ZMM = 3270;
	/**
	 * {@code VPMASKMOVD xmm1, xmm2, m128}<p>
	 * {@code VEX.128.66.0F38.W0 8C /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMASKMOVD_XMM_XMM_M128 = 3271;
	/**
	 * {@code VPMASKMOVD ymm1, ymm2, m256}<p>
	 * {@code VEX.256.66.0F38.W0 8C /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMASKMOVD_YMM_YMM_M256 = 3272;
	/**
	 * {@code VPMASKMOVQ xmm1, xmm2, m128}<p>
	 * {@code VEX.128.66.0F38.W1 8C /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMASKMOVQ_XMM_XMM_M128 = 3273;
	/**
	 * {@code VPMASKMOVQ ymm1, ymm2, m256}<p>
	 * {@code VEX.256.66.0F38.W1 8C /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMASKMOVQ_YMM_YMM_M256 = 3274;
	/**
	 * {@code VPERMB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W0 8D /r}<p>
	 * {@code AVX512VL and AVX512_VBMI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMB_XMM_K1Z_XMM_XMMM128 = 3275;
	/**
	 * {@code VPERMB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W0 8D /r}<p>
	 * {@code AVX512VL and AVX512_VBMI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMB_YMM_K1Z_YMM_YMMM256 = 3276;
	/**
	 * {@code VPERMB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W0 8D /r}<p>
	 * {@code AVX512_VBMI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMB_ZMM_K1Z_ZMM_ZMMM512 = 3277;
	/**
	 * {@code VPERMW xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W1 8D /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMW_XMM_K1Z_XMM_XMMM128 = 3278;
	/**
	 * {@code VPERMW ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W1 8D /r}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMW_YMM_K1Z_YMM_YMMM256 = 3279;
	/**
	 * {@code VPERMW zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W1 8D /r}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMW_ZMM_K1Z_ZMM_ZMMM512 = 3280;
	/**
	 * {@code VPMASKMOVD m128, xmm1, xmm2}<p>
	 * {@code VEX.128.66.0F38.W0 8E /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMASKMOVD_M128_XMM_XMM = 3281;
	/**
	 * {@code VPMASKMOVD m256, ymm1, ymm2}<p>
	 * {@code VEX.256.66.0F38.W0 8E /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMASKMOVD_M256_YMM_YMM = 3282;
	/**
	 * {@code VPMASKMOVQ m128, xmm1, xmm2}<p>
	 * {@code VEX.128.66.0F38.W1 8E /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMASKMOVQ_M128_XMM_XMM = 3283;
	/**
	 * {@code VPMASKMOVQ m256, ymm1, ymm2}<p>
	 * {@code VEX.256.66.0F38.W1 8E /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMASKMOVQ_M256_YMM_YMM = 3284;
	/**
	 * {@code VPSHUFBITQMB k1 {k2}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W0 8F /r}<p>
	 * {@code AVX512VL and AVX512_BITALG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHUFBITQMB_KR_K1_XMM_XMMM128 = 3285;
	/**
	 * {@code VPSHUFBITQMB k1 {k2}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W0 8F /r}<p>
	 * {@code AVX512VL and AVX512_BITALG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHUFBITQMB_KR_K1_YMM_YMMM256 = 3286;
	/**
	 * {@code VPSHUFBITQMB k1 {k2}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W0 8F /r}<p>
	 * {@code AVX512_BITALG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHUFBITQMB_KR_K1_ZMM_ZMMM512 = 3287;
	/**
	 * {@code VPGATHERDD xmm1, vm32x, xmm2}<p>
	 * {@code VEX.128.66.0F38.W0 90 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPGATHERDD_XMM_VM32X_XMM = 3288;
	/**
	 * {@code VPGATHERDD ymm1, vm32y, ymm2}<p>
	 * {@code VEX.256.66.0F38.W0 90 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPGATHERDD_YMM_VM32Y_YMM = 3289;
	/**
	 * {@code VPGATHERDQ xmm1, vm32x, xmm2}<p>
	 * {@code VEX.128.66.0F38.W1 90 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPGATHERDQ_XMM_VM32X_XMM = 3290;
	/**
	 * {@code VPGATHERDQ ymm1, vm32x, ymm2}<p>
	 * {@code VEX.256.66.0F38.W1 90 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPGATHERDQ_YMM_VM32X_YMM = 3291;
	/**
	 * {@code VPGATHERDD xmm1 {k1}, vm32x}<p>
	 * {@code EVEX.128.66.0F38.W0 90 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPGATHERDD_XMM_K1_VM32X = 3292;
	/**
	 * {@code VPGATHERDD ymm1 {k1}, vm32y}<p>
	 * {@code EVEX.256.66.0F38.W0 90 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPGATHERDD_YMM_K1_VM32Y = 3293;
	/**
	 * {@code VPGATHERDD zmm1 {k1}, vm32z}<p>
	 * {@code EVEX.512.66.0F38.W0 90 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPGATHERDD_ZMM_K1_VM32Z = 3294;
	/**
	 * {@code VPGATHERDQ xmm1 {k1}, vm32x}<p>
	 * {@code EVEX.128.66.0F38.W1 90 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPGATHERDQ_XMM_K1_VM32X = 3295;
	/**
	 * {@code VPGATHERDQ ymm1 {k1}, vm32x}<p>
	 * {@code EVEX.256.66.0F38.W1 90 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPGATHERDQ_YMM_K1_VM32X = 3296;
	/**
	 * {@code VPGATHERDQ zmm1 {k1}, vm32y}<p>
	 * {@code EVEX.512.66.0F38.W1 90 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPGATHERDQ_ZMM_K1_VM32Y = 3297;
	/**
	 * {@code VPGATHERQD xmm1, vm64x, xmm2}<p>
	 * {@code VEX.128.66.0F38.W0 91 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPGATHERQD_XMM_VM64X_XMM = 3298;
	/**
	 * {@code VPGATHERQD xmm1, vm64y, xmm2}<p>
	 * {@code VEX.256.66.0F38.W0 91 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPGATHERQD_XMM_VM64Y_XMM = 3299;
	/**
	 * {@code VPGATHERQQ xmm1, vm64x, xmm2}<p>
	 * {@code VEX.128.66.0F38.W1 91 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPGATHERQQ_XMM_VM64X_XMM = 3300;
	/**
	 * {@code VPGATHERQQ ymm1, vm64y, ymm2}<p>
	 * {@code VEX.256.66.0F38.W1 91 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPGATHERQQ_YMM_VM64Y_YMM = 3301;
	/**
	 * {@code VPGATHERQD xmm1 {k1}, vm64x}<p>
	 * {@code EVEX.128.66.0F38.W0 91 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPGATHERQD_XMM_K1_VM64X = 3302;
	/**
	 * {@code VPGATHERQD xmm1 {k1}, vm64y}<p>
	 * {@code EVEX.256.66.0F38.W0 91 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPGATHERQD_XMM_K1_VM64Y = 3303;
	/**
	 * {@code VPGATHERQD ymm1 {k1}, vm64z}<p>
	 * {@code EVEX.512.66.0F38.W0 91 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPGATHERQD_YMM_K1_VM64Z = 3304;
	/**
	 * {@code VPGATHERQQ xmm1 {k1}, vm64x}<p>
	 * {@code EVEX.128.66.0F38.W1 91 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPGATHERQQ_XMM_K1_VM64X = 3305;
	/**
	 * {@code VPGATHERQQ ymm1 {k1}, vm64y}<p>
	 * {@code EVEX.256.66.0F38.W1 91 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPGATHERQQ_YMM_K1_VM64Y = 3306;
	/**
	 * {@code VPGATHERQQ zmm1 {k1}, vm64z}<p>
	 * {@code EVEX.512.66.0F38.W1 91 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPGATHERQQ_ZMM_K1_VM64Z = 3307;
	/**
	 * {@code VGATHERDPS xmm1, vm32x, xmm2}<p>
	 * {@code VEX.128.66.0F38.W0 92 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VGATHERDPS_XMM_VM32X_XMM = 3308;
	/**
	 * {@code VGATHERDPS ymm1, vm32y, ymm2}<p>
	 * {@code VEX.256.66.0F38.W0 92 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VGATHERDPS_YMM_VM32Y_YMM = 3309;
	/**
	 * {@code VGATHERDPD xmm1, vm32x, xmm2}<p>
	 * {@code VEX.128.66.0F38.W1 92 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VGATHERDPD_XMM_VM32X_XMM = 3310;
	/**
	 * {@code VGATHERDPD ymm1, vm32x, ymm2}<p>
	 * {@code VEX.256.66.0F38.W1 92 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VGATHERDPD_YMM_VM32X_YMM = 3311;
	/**
	 * {@code VGATHERDPS xmm1 {k1}, vm32x}<p>
	 * {@code EVEX.128.66.0F38.W0 92 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERDPS_XMM_K1_VM32X = 3312;
	/**
	 * {@code VGATHERDPS ymm1 {k1}, vm32y}<p>
	 * {@code EVEX.256.66.0F38.W0 92 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERDPS_YMM_K1_VM32Y = 3313;
	/**
	 * {@code VGATHERDPS zmm1 {k1}, vm32z}<p>
	 * {@code EVEX.512.66.0F38.W0 92 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERDPS_ZMM_K1_VM32Z = 3314;
	/**
	 * {@code VGATHERDPD xmm1 {k1}, vm32x}<p>
	 * {@code EVEX.128.66.0F38.W1 92 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERDPD_XMM_K1_VM32X = 3315;
	/**
	 * {@code VGATHERDPD ymm1 {k1}, vm32x}<p>
	 * {@code EVEX.256.66.0F38.W1 92 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERDPD_YMM_K1_VM32X = 3316;
	/**
	 * {@code VGATHERDPD zmm1 {k1}, vm32y}<p>
	 * {@code EVEX.512.66.0F38.W1 92 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERDPD_ZMM_K1_VM32Y = 3317;
	/**
	 * {@code VGATHERQPS xmm1, vm64x, xmm2}<p>
	 * {@code VEX.128.66.0F38.W0 93 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VGATHERQPS_XMM_VM64X_XMM = 3318;
	/**
	 * {@code VGATHERQPS xmm1, vm64y, xmm2}<p>
	 * {@code VEX.256.66.0F38.W0 93 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VGATHERQPS_XMM_VM64Y_XMM = 3319;
	/**
	 * {@code VGATHERQPD xmm1, vm64x, xmm2}<p>
	 * {@code VEX.128.66.0F38.W1 93 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VGATHERQPD_XMM_VM64X_XMM = 3320;
	/**
	 * {@code VGATHERQPD ymm1, vm64y, ymm2}<p>
	 * {@code VEX.256.66.0F38.W1 93 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VGATHERQPD_YMM_VM64Y_YMM = 3321;
	/**
	 * {@code VGATHERQPS xmm1 {k1}, vm64x}<p>
	 * {@code EVEX.128.66.0F38.W0 93 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERQPS_XMM_K1_VM64X = 3322;
	/**
	 * {@code VGATHERQPS xmm1 {k1}, vm64y}<p>
	 * {@code EVEX.256.66.0F38.W0 93 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERQPS_XMM_K1_VM64Y = 3323;
	/**
	 * {@code VGATHERQPS ymm1 {k1}, vm64z}<p>
	 * {@code EVEX.512.66.0F38.W0 93 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERQPS_YMM_K1_VM64Z = 3324;
	/**
	 * {@code VGATHERQPD xmm1 {k1}, vm64x}<p>
	 * {@code EVEX.128.66.0F38.W1 93 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERQPD_XMM_K1_VM64X = 3325;
	/**
	 * {@code VGATHERQPD ymm1 {k1}, vm64y}<p>
	 * {@code EVEX.256.66.0F38.W1 93 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERQPD_YMM_K1_VM64Y = 3326;
	/**
	 * {@code VGATHERQPD zmm1 {k1}, vm64z}<p>
	 * {@code EVEX.512.66.0F38.W1 93 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERQPD_ZMM_K1_VM64Z = 3327;
	/**
	 * {@code VFMADDSUB132PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 96 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUB132PS_XMM_XMM_XMMM128 = 3328;
	/**
	 * {@code VFMADDSUB132PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 96 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUB132PS_YMM_YMM_YMMM256 = 3329;
	/**
	 * {@code VFMADDSUB132PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 96 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUB132PD_XMM_XMM_XMMM128 = 3330;
	/**
	 * {@code VFMADDSUB132PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 96 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUB132PD_YMM_YMM_YMMM256 = 3331;
	/**
	 * {@code VFMADDSUB132PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 96 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB132PS_XMM_K1Z_XMM_XMMM128B32 = 3332;
	/**
	 * {@code VFMADDSUB132PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 96 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB132PS_YMM_K1Z_YMM_YMMM256B32 = 3333;
	/**
	 * {@code VFMADDSUB132PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 96 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB132PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3334;
	/**
	 * {@code VFMADDSUB132PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 96 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB132PD_XMM_K1Z_XMM_XMMM128B64 = 3335;
	/**
	 * {@code VFMADDSUB132PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 96 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB132PD_YMM_K1Z_YMM_YMMM256B64 = 3336;
	/**
	 * {@code VFMADDSUB132PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 96 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB132PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3337;
	/**
	 * {@code VFMSUBADD132PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 97 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADD132PS_XMM_XMM_XMMM128 = 3338;
	/**
	 * {@code VFMSUBADD132PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 97 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADD132PS_YMM_YMM_YMMM256 = 3339;
	/**
	 * {@code VFMSUBADD132PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 97 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADD132PD_XMM_XMM_XMMM128 = 3340;
	/**
	 * {@code VFMSUBADD132PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 97 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADD132PD_YMM_YMM_YMMM256 = 3341;
	/**
	 * {@code VFMSUBADD132PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 97 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD132PS_XMM_K1Z_XMM_XMMM128B32 = 3342;
	/**
	 * {@code VFMSUBADD132PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 97 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD132PS_YMM_K1Z_YMM_YMMM256B32 = 3343;
	/**
	 * {@code VFMSUBADD132PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 97 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD132PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3344;
	/**
	 * {@code VFMSUBADD132PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 97 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD132PD_XMM_K1Z_XMM_XMMM128B64 = 3345;
	/**
	 * {@code VFMSUBADD132PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 97 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD132PD_YMM_K1Z_YMM_YMMM256B64 = 3346;
	/**
	 * {@code VFMSUBADD132PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 97 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD132PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3347;
	/**
	 * {@code VFMADD132PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 98 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD132PS_XMM_XMM_XMMM128 = 3348;
	/**
	 * {@code VFMADD132PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 98 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD132PS_YMM_YMM_YMMM256 = 3349;
	/**
	 * {@code VFMADD132PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 98 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD132PD_XMM_XMM_XMMM128 = 3350;
	/**
	 * {@code VFMADD132PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 98 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD132PD_YMM_YMM_YMMM256 = 3351;
	/**
	 * {@code VFMADD132PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 98 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD132PS_XMM_K1Z_XMM_XMMM128B32 = 3352;
	/**
	 * {@code VFMADD132PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 98 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD132PS_YMM_K1Z_YMM_YMMM256B32 = 3353;
	/**
	 * {@code VFMADD132PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 98 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD132PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3354;
	/**
	 * {@code VFMADD132PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 98 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD132PD_XMM_K1Z_XMM_XMMM128B64 = 3355;
	/**
	 * {@code VFMADD132PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 98 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD132PD_YMM_K1Z_YMM_YMMM256B64 = 3356;
	/**
	 * {@code VFMADD132PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 98 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD132PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3357;
	/**
	 * {@code VFMADD132SS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.66.0F38.W0 99 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD132SS_XMM_XMM_XMMM32 = 3358;
	/**
	 * {@code VFMADD132SD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.66.0F38.W1 99 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD132SD_XMM_XMM_XMMM64 = 3359;
	/**
	 * {@code VFMADD132SS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 99 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD132SS_XMM_K1Z_XMM_XMMM32_ER = 3360;
	/**
	 * {@code VFMADD132SD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 99 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD132SD_XMM_K1Z_XMM_XMMM64_ER = 3361;
	/**
	 * {@code VFMSUB132PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 9A /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB132PS_XMM_XMM_XMMM128 = 3362;
	/**
	 * {@code VFMSUB132PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 9A /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB132PS_YMM_YMM_YMMM256 = 3363;
	/**
	 * {@code VFMSUB132PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 9A /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB132PD_XMM_XMM_XMMM128 = 3364;
	/**
	 * {@code VFMSUB132PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 9A /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB132PD_YMM_YMM_YMMM256 = 3365;
	/**
	 * {@code VFMSUB132PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 9A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB132PS_XMM_K1Z_XMM_XMMM128B32 = 3366;
	/**
	 * {@code VFMSUB132PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 9A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB132PS_YMM_K1Z_YMM_YMMM256B32 = 3367;
	/**
	 * {@code VFMSUB132PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 9A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB132PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3368;
	/**
	 * {@code VFMSUB132PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 9A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB132PD_XMM_K1Z_XMM_XMMM128B64 = 3369;
	/**
	 * {@code VFMSUB132PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 9A /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB132PD_YMM_K1Z_YMM_YMMM256B64 = 3370;
	/**
	 * {@code VFMSUB132PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 9A /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB132PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3371;
	/**
	 * {@code V4FMADDPS zmm1 {k1}{z}, zmm2+3, m128}<p>
	 * {@code EVEX.512.F2.0F38.W0 9A /r}<p>
	 * {@code AVX512_4FMAPS}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_V4FMADDPS_ZMM_K1Z_ZMMP3_M128 = 3372;
	/**
	 * {@code VFMSUB132SS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.66.0F38.W0 9B /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB132SS_XMM_XMM_XMMM32 = 3373;
	/**
	 * {@code VFMSUB132SD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.66.0F38.W1 9B /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB132SD_XMM_XMM_XMMM64 = 3374;
	/**
	 * {@code VFMSUB132SS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 9B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB132SS_XMM_K1Z_XMM_XMMM32_ER = 3375;
	/**
	 * {@code VFMSUB132SD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 9B /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB132SD_XMM_K1Z_XMM_XMMM64_ER = 3376;
	/**
	 * {@code V4FMADDSS xmm1 {k1}{z}, xmm2+3, m128}<p>
	 * {@code EVEX.LIG.F2.0F38.W0 9B /r}<p>
	 * {@code AVX512_4FMAPS}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_V4FMADDSS_XMM_K1Z_XMMP3_M128 = 3377;
	/**
	 * {@code VFNMADD132PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 9C /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD132PS_XMM_XMM_XMMM128 = 3378;
	/**
	 * {@code VFNMADD132PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 9C /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD132PS_YMM_YMM_YMMM256 = 3379;
	/**
	 * {@code VFNMADD132PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 9C /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD132PD_XMM_XMM_XMMM128 = 3380;
	/**
	 * {@code VFNMADD132PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 9C /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD132PD_YMM_YMM_YMMM256 = 3381;
	/**
	 * {@code VFNMADD132PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 9C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD132PS_XMM_K1Z_XMM_XMMM128B32 = 3382;
	/**
	 * {@code VFNMADD132PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 9C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD132PS_YMM_K1Z_YMM_YMMM256B32 = 3383;
	/**
	 * {@code VFNMADD132PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 9C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD132PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3384;
	/**
	 * {@code VFNMADD132PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 9C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD132PD_XMM_K1Z_XMM_XMMM128B64 = 3385;
	/**
	 * {@code VFNMADD132PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 9C /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD132PD_YMM_K1Z_YMM_YMMM256B64 = 3386;
	/**
	 * {@code VFNMADD132PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 9C /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD132PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3387;
	/**
	 * {@code VFNMADD132SS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.66.0F38.W0 9D /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD132SS_XMM_XMM_XMMM32 = 3388;
	/**
	 * {@code VFNMADD132SD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.66.0F38.W1 9D /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD132SD_XMM_XMM_XMMM64 = 3389;
	/**
	 * {@code VFNMADD132SS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 9D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD132SS_XMM_K1Z_XMM_XMMM32_ER = 3390;
	/**
	 * {@code VFNMADD132SD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 9D /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD132SD_XMM_K1Z_XMM_XMMM64_ER = 3391;
	/**
	 * {@code VFNMSUB132PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 9E /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB132PS_XMM_XMM_XMMM128 = 3392;
	/**
	 * {@code VFNMSUB132PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 9E /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB132PS_YMM_YMM_YMMM256 = 3393;
	/**
	 * {@code VFNMSUB132PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 9E /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB132PD_XMM_XMM_XMMM128 = 3394;
	/**
	 * {@code VFNMSUB132PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 9E /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB132PD_YMM_YMM_YMMM256 = 3395;
	/**
	 * {@code VFNMSUB132PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 9E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB132PS_XMM_K1Z_XMM_XMMM128B32 = 3396;
	/**
	 * {@code VFNMSUB132PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 9E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB132PS_YMM_K1Z_YMM_YMMM256B32 = 3397;
	/**
	 * {@code VFNMSUB132PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 9E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB132PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3398;
	/**
	 * {@code VFNMSUB132PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 9E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB132PD_XMM_K1Z_XMM_XMMM128B64 = 3399;
	/**
	 * {@code VFNMSUB132PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 9E /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB132PD_YMM_K1Z_YMM_YMMM256B64 = 3400;
	/**
	 * {@code VFNMSUB132PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 9E /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB132PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3401;
	/**
	 * {@code VFNMSUB132SS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.66.0F38.W0 9F /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB132SS_XMM_XMM_XMMM32 = 3402;
	/**
	 * {@code VFNMSUB132SD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.66.0F38.W1 9F /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB132SD_XMM_XMM_XMMM64 = 3403;
	/**
	 * {@code VFNMSUB132SS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 9F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB132SS_XMM_K1Z_XMM_XMMM32_ER = 3404;
	/**
	 * {@code VFNMSUB132SD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 9F /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB132SD_XMM_K1Z_XMM_XMMM64_ER = 3405;
	/**
	 * {@code VPSCATTERDD vm32x {k1}, xmm1}<p>
	 * {@code EVEX.128.66.0F38.W0 A0 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSCATTERDD_VM32X_K1_XMM = 3406;
	/**
	 * {@code VPSCATTERDD vm32y {k1}, ymm1}<p>
	 * {@code EVEX.256.66.0F38.W0 A0 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSCATTERDD_VM32Y_K1_YMM = 3407;
	/**
	 * {@code VPSCATTERDD vm32z {k1}, zmm1}<p>
	 * {@code EVEX.512.66.0F38.W0 A0 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSCATTERDD_VM32Z_K1_ZMM = 3408;
	/**
	 * {@code VPSCATTERDQ vm32x {k1}, xmm1}<p>
	 * {@code EVEX.128.66.0F38.W1 A0 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSCATTERDQ_VM32X_K1_XMM = 3409;
	/**
	 * {@code VPSCATTERDQ vm32x {k1}, ymm1}<p>
	 * {@code EVEX.256.66.0F38.W1 A0 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSCATTERDQ_VM32X_K1_YMM = 3410;
	/**
	 * {@code VPSCATTERDQ vm32y {k1}, zmm1}<p>
	 * {@code EVEX.512.66.0F38.W1 A0 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSCATTERDQ_VM32Y_K1_ZMM = 3411;
	/**
	 * {@code VPSCATTERQD vm64x {k1}, xmm1}<p>
	 * {@code EVEX.128.66.0F38.W0 A1 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSCATTERQD_VM64X_K1_XMM = 3412;
	/**
	 * {@code VPSCATTERQD vm64y {k1}, xmm1}<p>
	 * {@code EVEX.256.66.0F38.W0 A1 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSCATTERQD_VM64Y_K1_XMM = 3413;
	/**
	 * {@code VPSCATTERQD vm64z {k1}, ymm1}<p>
	 * {@code EVEX.512.66.0F38.W0 A1 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSCATTERQD_VM64Z_K1_YMM = 3414;
	/**
	 * {@code VPSCATTERQQ vm64x {k1}, xmm1}<p>
	 * {@code EVEX.128.66.0F38.W1 A1 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSCATTERQQ_VM64X_K1_XMM = 3415;
	/**
	 * {@code VPSCATTERQQ vm64y {k1}, ymm1}<p>
	 * {@code EVEX.256.66.0F38.W1 A1 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSCATTERQQ_VM64Y_K1_YMM = 3416;
	/**
	 * {@code VPSCATTERQQ vm64z {k1}, zmm1}<p>
	 * {@code EVEX.512.66.0F38.W1 A1 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSCATTERQQ_VM64Z_K1_ZMM = 3417;
	/**
	 * {@code VSCATTERDPS vm32x {k1}, xmm1}<p>
	 * {@code EVEX.128.66.0F38.W0 A2 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERDPS_VM32X_K1_XMM = 3418;
	/**
	 * {@code VSCATTERDPS vm32y {k1}, ymm1}<p>
	 * {@code EVEX.256.66.0F38.W0 A2 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERDPS_VM32Y_K1_YMM = 3419;
	/**
	 * {@code VSCATTERDPS vm32z {k1}, zmm1}<p>
	 * {@code EVEX.512.66.0F38.W0 A2 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERDPS_VM32Z_K1_ZMM = 3420;
	/**
	 * {@code VSCATTERDPD vm32x {k1}, xmm1}<p>
	 * {@code EVEX.128.66.0F38.W1 A2 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERDPD_VM32X_K1_XMM = 3421;
	/**
	 * {@code VSCATTERDPD vm32x {k1}, ymm1}<p>
	 * {@code EVEX.256.66.0F38.W1 A2 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERDPD_VM32X_K1_YMM = 3422;
	/**
	 * {@code VSCATTERDPD vm32y {k1}, zmm1}<p>
	 * {@code EVEX.512.66.0F38.W1 A2 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERDPD_VM32Y_K1_ZMM = 3423;
	/**
	 * {@code VSCATTERQPS vm64x {k1}, xmm1}<p>
	 * {@code EVEX.128.66.0F38.W0 A3 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERQPS_VM64X_K1_XMM = 3424;
	/**
	 * {@code VSCATTERQPS vm64y {k1}, xmm1}<p>
	 * {@code EVEX.256.66.0F38.W0 A3 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERQPS_VM64Y_K1_XMM = 3425;
	/**
	 * {@code VSCATTERQPS vm64z {k1}, ymm1}<p>
	 * {@code EVEX.512.66.0F38.W0 A3 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERQPS_VM64Z_K1_YMM = 3426;
	/**
	 * {@code VSCATTERQPD vm64x {k1}, xmm1}<p>
	 * {@code EVEX.128.66.0F38.W1 A3 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERQPD_VM64X_K1_XMM = 3427;
	/**
	 * {@code VSCATTERQPD vm64y {k1}, ymm1}<p>
	 * {@code EVEX.256.66.0F38.W1 A3 /vsib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERQPD_VM64Y_K1_YMM = 3428;
	/**
	 * {@code VSCATTERQPD vm64z {k1}, zmm1}<p>
	 * {@code EVEX.512.66.0F38.W1 A3 /vsib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERQPD_VM64Z_K1_ZMM = 3429;
	/**
	 * {@code VFMADDSUB213PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 A6 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUB213PS_XMM_XMM_XMMM128 = 3430;
	/**
	 * {@code VFMADDSUB213PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 A6 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUB213PS_YMM_YMM_YMMM256 = 3431;
	/**
	 * {@code VFMADDSUB213PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 A6 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUB213PD_XMM_XMM_XMMM128 = 3432;
	/**
	 * {@code VFMADDSUB213PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 A6 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUB213PD_YMM_YMM_YMMM256 = 3433;
	/**
	 * {@code VFMADDSUB213PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 A6 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB213PS_XMM_K1Z_XMM_XMMM128B32 = 3434;
	/**
	 * {@code VFMADDSUB213PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 A6 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB213PS_YMM_K1Z_YMM_YMMM256B32 = 3435;
	/**
	 * {@code VFMADDSUB213PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 A6 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB213PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3436;
	/**
	 * {@code VFMADDSUB213PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 A6 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB213PD_XMM_K1Z_XMM_XMMM128B64 = 3437;
	/**
	 * {@code VFMADDSUB213PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 A6 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB213PD_YMM_K1Z_YMM_YMMM256B64 = 3438;
	/**
	 * {@code VFMADDSUB213PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 A6 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB213PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3439;
	/**
	 * {@code VFMSUBADD213PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 A7 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADD213PS_XMM_XMM_XMMM128 = 3440;
	/**
	 * {@code VFMSUBADD213PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 A7 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADD213PS_YMM_YMM_YMMM256 = 3441;
	/**
	 * {@code VFMSUBADD213PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 A7 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADD213PD_XMM_XMM_XMMM128 = 3442;
	/**
	 * {@code VFMSUBADD213PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 A7 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADD213PD_YMM_YMM_YMMM256 = 3443;
	/**
	 * {@code VFMSUBADD213PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 A7 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD213PS_XMM_K1Z_XMM_XMMM128B32 = 3444;
	/**
	 * {@code VFMSUBADD213PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 A7 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD213PS_YMM_K1Z_YMM_YMMM256B32 = 3445;
	/**
	 * {@code VFMSUBADD213PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 A7 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD213PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3446;
	/**
	 * {@code VFMSUBADD213PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 A7 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD213PD_XMM_K1Z_XMM_XMMM128B64 = 3447;
	/**
	 * {@code VFMSUBADD213PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 A7 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD213PD_YMM_K1Z_YMM_YMMM256B64 = 3448;
	/**
	 * {@code VFMSUBADD213PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 A7 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD213PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3449;
	/**
	 * {@code VFMADD213PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 A8 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD213PS_XMM_XMM_XMMM128 = 3450;
	/**
	 * {@code VFMADD213PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 A8 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD213PS_YMM_YMM_YMMM256 = 3451;
	/**
	 * {@code VFMADD213PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 A8 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD213PD_XMM_XMM_XMMM128 = 3452;
	/**
	 * {@code VFMADD213PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 A8 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD213PD_YMM_YMM_YMMM256 = 3453;
	/**
	 * {@code VFMADD213PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 A8 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD213PS_XMM_K1Z_XMM_XMMM128B32 = 3454;
	/**
	 * {@code VFMADD213PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 A8 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD213PS_YMM_K1Z_YMM_YMMM256B32 = 3455;
	/**
	 * {@code VFMADD213PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 A8 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD213PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3456;
	/**
	 * {@code VFMADD213PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 A8 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD213PD_XMM_K1Z_XMM_XMMM128B64 = 3457;
	/**
	 * {@code VFMADD213PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 A8 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD213PD_YMM_K1Z_YMM_YMMM256B64 = 3458;
	/**
	 * {@code VFMADD213PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 A8 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD213PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3459;
	/**
	 * {@code VFMADD213SS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.66.0F38.W0 A9 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD213SS_XMM_XMM_XMMM32 = 3460;
	/**
	 * {@code VFMADD213SD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.66.0F38.W1 A9 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD213SD_XMM_XMM_XMMM64 = 3461;
	/**
	 * {@code VFMADD213SS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 A9 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD213SS_XMM_K1Z_XMM_XMMM32_ER = 3462;
	/**
	 * {@code VFMADD213SD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 A9 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD213SD_XMM_K1Z_XMM_XMMM64_ER = 3463;
	/**
	 * {@code VFMSUB213PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 AA /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB213PS_XMM_XMM_XMMM128 = 3464;
	/**
	 * {@code VFMSUB213PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 AA /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB213PS_YMM_YMM_YMMM256 = 3465;
	/**
	 * {@code VFMSUB213PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 AA /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB213PD_XMM_XMM_XMMM128 = 3466;
	/**
	 * {@code VFMSUB213PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 AA /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB213PD_YMM_YMM_YMMM256 = 3467;
	/**
	 * {@code VFMSUB213PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 AA /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB213PS_XMM_K1Z_XMM_XMMM128B32 = 3468;
	/**
	 * {@code VFMSUB213PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 AA /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB213PS_YMM_K1Z_YMM_YMMM256B32 = 3469;
	/**
	 * {@code VFMSUB213PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 AA /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB213PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3470;
	/**
	 * {@code VFMSUB213PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 AA /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB213PD_XMM_K1Z_XMM_XMMM128B64 = 3471;
	/**
	 * {@code VFMSUB213PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 AA /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB213PD_YMM_K1Z_YMM_YMMM256B64 = 3472;
	/**
	 * {@code VFMSUB213PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 AA /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB213PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3473;
	/**
	 * {@code V4FNMADDPS zmm1 {k1}{z}, zmm2+3, m128}<p>
	 * {@code EVEX.512.F2.0F38.W0 AA /r}<p>
	 * {@code AVX512_4FMAPS}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_V4FNMADDPS_ZMM_K1Z_ZMMP3_M128 = 3474;
	/**
	 * {@code VFMSUB213SS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.66.0F38.W0 AB /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB213SS_XMM_XMM_XMMM32 = 3475;
	/**
	 * {@code VFMSUB213SD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.66.0F38.W1 AB /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB213SD_XMM_XMM_XMMM64 = 3476;
	/**
	 * {@code VFMSUB213SS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 AB /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB213SS_XMM_K1Z_XMM_XMMM32_ER = 3477;
	/**
	 * {@code VFMSUB213SD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 AB /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB213SD_XMM_K1Z_XMM_XMMM64_ER = 3478;
	/**
	 * {@code V4FNMADDSS xmm1 {k1}{z}, xmm2+3, m128}<p>
	 * {@code EVEX.LIG.F2.0F38.W0 AB /r}<p>
	 * {@code AVX512_4FMAPS}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_V4FNMADDSS_XMM_K1Z_XMMP3_M128 = 3479;
	/**
	 * {@code VFNMADD213PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 AC /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD213PS_XMM_XMM_XMMM128 = 3480;
	/**
	 * {@code VFNMADD213PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 AC /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD213PS_YMM_YMM_YMMM256 = 3481;
	/**
	 * {@code VFNMADD213PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 AC /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD213PD_XMM_XMM_XMMM128 = 3482;
	/**
	 * {@code VFNMADD213PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 AC /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD213PD_YMM_YMM_YMMM256 = 3483;
	/**
	 * {@code VFNMADD213PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 AC /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD213PS_XMM_K1Z_XMM_XMMM128B32 = 3484;
	/**
	 * {@code VFNMADD213PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 AC /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD213PS_YMM_K1Z_YMM_YMMM256B32 = 3485;
	/**
	 * {@code VFNMADD213PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 AC /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD213PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3486;
	/**
	 * {@code VFNMADD213PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 AC /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD213PD_XMM_K1Z_XMM_XMMM128B64 = 3487;
	/**
	 * {@code VFNMADD213PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 AC /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD213PD_YMM_K1Z_YMM_YMMM256B64 = 3488;
	/**
	 * {@code VFNMADD213PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 AC /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD213PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3489;
	/**
	 * {@code VFNMADD213SS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.66.0F38.W0 AD /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD213SS_XMM_XMM_XMMM32 = 3490;
	/**
	 * {@code VFNMADD213SD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.66.0F38.W1 AD /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD213SD_XMM_XMM_XMMM64 = 3491;
	/**
	 * {@code VFNMADD213SS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 AD /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD213SS_XMM_K1Z_XMM_XMMM32_ER = 3492;
	/**
	 * {@code VFNMADD213SD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 AD /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD213SD_XMM_K1Z_XMM_XMMM64_ER = 3493;
	/**
	 * {@code VFNMSUB213PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 AE /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB213PS_XMM_XMM_XMMM128 = 3494;
	/**
	 * {@code VFNMSUB213PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 AE /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB213PS_YMM_YMM_YMMM256 = 3495;
	/**
	 * {@code VFNMSUB213PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 AE /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB213PD_XMM_XMM_XMMM128 = 3496;
	/**
	 * {@code VFNMSUB213PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 AE /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB213PD_YMM_YMM_YMMM256 = 3497;
	/**
	 * {@code VFNMSUB213PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 AE /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB213PS_XMM_K1Z_XMM_XMMM128B32 = 3498;
	/**
	 * {@code VFNMSUB213PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 AE /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB213PS_YMM_K1Z_YMM_YMMM256B32 = 3499;
	/**
	 * {@code VFNMSUB213PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 AE /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB213PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3500;
	/**
	 * {@code VFNMSUB213PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 AE /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB213PD_XMM_K1Z_XMM_XMMM128B64 = 3501;
	/**
	 * {@code VFNMSUB213PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 AE /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB213PD_YMM_K1Z_YMM_YMMM256B64 = 3502;
	/**
	 * {@code VFNMSUB213PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 AE /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB213PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3503;
	/**
	 * {@code VFNMSUB213SS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.66.0F38.W0 AF /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB213SS_XMM_XMM_XMMM32 = 3504;
	/**
	 * {@code VFNMSUB213SD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.66.0F38.W1 AF /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB213SD_XMM_XMM_XMMM64 = 3505;
	/**
	 * {@code VFNMSUB213SS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 AF /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB213SS_XMM_K1Z_XMM_XMMM32_ER = 3506;
	/**
	 * {@code VFNMSUB213SD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 AF /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB213SD_XMM_K1Z_XMM_XMMM64_ER = 3507;
	/**
	 * {@code VPMADD52LUQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 B4 /r}<p>
	 * {@code AVX512VL and AVX512_IFMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMADD52LUQ_XMM_K1Z_XMM_XMMM128B64 = 3508;
	/**
	 * {@code VPMADD52LUQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 B4 /r}<p>
	 * {@code AVX512VL and AVX512_IFMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMADD52LUQ_YMM_K1Z_YMM_YMMM256B64 = 3509;
	/**
	 * {@code VPMADD52LUQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 B4 /r}<p>
	 * {@code AVX512_IFMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMADD52LUQ_ZMM_K1Z_ZMM_ZMMM512B64 = 3510;
	/**
	 * {@code VPMADD52HUQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 B5 /r}<p>
	 * {@code AVX512VL and AVX512_IFMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMADD52HUQ_XMM_K1Z_XMM_XMMM128B64 = 3511;
	/**
	 * {@code VPMADD52HUQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 B5 /r}<p>
	 * {@code AVX512VL and AVX512_IFMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMADD52HUQ_YMM_K1Z_YMM_YMMM256B64 = 3512;
	/**
	 * {@code VPMADD52HUQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 B5 /r}<p>
	 * {@code AVX512_IFMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPMADD52HUQ_ZMM_K1Z_ZMM_ZMMM512B64 = 3513;
	/**
	 * {@code VFMADDSUB231PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 B6 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUB231PS_XMM_XMM_XMMM128 = 3514;
	/**
	 * {@code VFMADDSUB231PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 B6 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUB231PS_YMM_YMM_YMMM256 = 3515;
	/**
	 * {@code VFMADDSUB231PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 B6 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUB231PD_XMM_XMM_XMMM128 = 3516;
	/**
	 * {@code VFMADDSUB231PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 B6 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUB231PD_YMM_YMM_YMMM256 = 3517;
	/**
	 * {@code VFMADDSUB231PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 B6 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB231PS_XMM_K1Z_XMM_XMMM128B32 = 3518;
	/**
	 * {@code VFMADDSUB231PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 B6 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB231PS_YMM_K1Z_YMM_YMMM256B32 = 3519;
	/**
	 * {@code VFMADDSUB231PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 B6 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB231PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3520;
	/**
	 * {@code VFMADDSUB231PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 B6 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB231PD_XMM_K1Z_XMM_XMMM128B64 = 3521;
	/**
	 * {@code VFMADDSUB231PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 B6 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB231PD_YMM_K1Z_YMM_YMMM256B64 = 3522;
	/**
	 * {@code VFMADDSUB231PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 B6 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB231PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3523;
	/**
	 * {@code VFMSUBADD231PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 B7 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADD231PS_XMM_XMM_XMMM128 = 3524;
	/**
	 * {@code VFMSUBADD231PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 B7 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADD231PS_YMM_YMM_YMMM256 = 3525;
	/**
	 * {@code VFMSUBADD231PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 B7 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADD231PD_XMM_XMM_XMMM128 = 3526;
	/**
	 * {@code VFMSUBADD231PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 B7 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADD231PD_YMM_YMM_YMMM256 = 3527;
	/**
	 * {@code VFMSUBADD231PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 B7 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD231PS_XMM_K1Z_XMM_XMMM128B32 = 3528;
	/**
	 * {@code VFMSUBADD231PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 B7 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD231PS_YMM_K1Z_YMM_YMMM256B32 = 3529;
	/**
	 * {@code VFMSUBADD231PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 B7 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD231PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3530;
	/**
	 * {@code VFMSUBADD231PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 B7 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD231PD_XMM_K1Z_XMM_XMMM128B64 = 3531;
	/**
	 * {@code VFMSUBADD231PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 B7 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD231PD_YMM_K1Z_YMM_YMMM256B64 = 3532;
	/**
	 * {@code VFMSUBADD231PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 B7 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD231PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3533;
	/**
	 * {@code VFMADD231PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 B8 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD231PS_XMM_XMM_XMMM128 = 3534;
	/**
	 * {@code VFMADD231PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 B8 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD231PS_YMM_YMM_YMMM256 = 3535;
	/**
	 * {@code VFMADD231PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 B8 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD231PD_XMM_XMM_XMMM128 = 3536;
	/**
	 * {@code VFMADD231PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 B8 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD231PD_YMM_YMM_YMMM256 = 3537;
	/**
	 * {@code VFMADD231PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 B8 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD231PS_XMM_K1Z_XMM_XMMM128B32 = 3538;
	/**
	 * {@code VFMADD231PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 B8 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD231PS_YMM_K1Z_YMM_YMMM256B32 = 3539;
	/**
	 * {@code VFMADD231PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 B8 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD231PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3540;
	/**
	 * {@code VFMADD231PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 B8 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD231PD_XMM_K1Z_XMM_XMMM128B64 = 3541;
	/**
	 * {@code VFMADD231PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 B8 /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD231PD_YMM_K1Z_YMM_YMMM256B64 = 3542;
	/**
	 * {@code VFMADD231PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 B8 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD231PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3543;
	/**
	 * {@code VFMADD231SS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.66.0F38.W0 B9 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD231SS_XMM_XMM_XMMM32 = 3544;
	/**
	 * {@code VFMADD231SD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.66.0F38.W1 B9 /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADD231SD_XMM_XMM_XMMM64 = 3545;
	/**
	 * {@code VFMADD231SS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 B9 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD231SS_XMM_K1Z_XMM_XMMM32_ER = 3546;
	/**
	 * {@code VFMADD231SD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 B9 /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD231SD_XMM_K1Z_XMM_XMMM64_ER = 3547;
	/**
	 * {@code VFMSUB231PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 BA /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB231PS_XMM_XMM_XMMM128 = 3548;
	/**
	 * {@code VFMSUB231PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 BA /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB231PS_YMM_YMM_YMMM256 = 3549;
	/**
	 * {@code VFMSUB231PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 BA /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB231PD_XMM_XMM_XMMM128 = 3550;
	/**
	 * {@code VFMSUB231PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 BA /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB231PD_YMM_YMM_YMMM256 = 3551;
	/**
	 * {@code VFMSUB231PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 BA /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB231PS_XMM_K1Z_XMM_XMMM128B32 = 3552;
	/**
	 * {@code VFMSUB231PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 BA /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB231PS_YMM_K1Z_YMM_YMMM256B32 = 3553;
	/**
	 * {@code VFMSUB231PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 BA /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB231PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3554;
	/**
	 * {@code VFMSUB231PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 BA /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB231PD_XMM_K1Z_XMM_XMMM128B64 = 3555;
	/**
	 * {@code VFMSUB231PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 BA /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB231PD_YMM_K1Z_YMM_YMMM256B64 = 3556;
	/**
	 * {@code VFMSUB231PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 BA /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB231PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3557;
	/**
	 * {@code VFMSUB231SS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.66.0F38.W0 BB /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB231SS_XMM_XMM_XMMM32 = 3558;
	/**
	 * {@code VFMSUB231SD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.66.0F38.W1 BB /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUB231SD_XMM_XMM_XMMM64 = 3559;
	/**
	 * {@code VFMSUB231SS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 BB /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB231SS_XMM_K1Z_XMM_XMMM32_ER = 3560;
	/**
	 * {@code VFMSUB231SD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 BB /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB231SD_XMM_K1Z_XMM_XMMM64_ER = 3561;
	/**
	 * {@code VFNMADD231PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 BC /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD231PS_XMM_XMM_XMMM128 = 3562;
	/**
	 * {@code VFNMADD231PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 BC /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD231PS_YMM_YMM_YMMM256 = 3563;
	/**
	 * {@code VFNMADD231PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 BC /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD231PD_XMM_XMM_XMMM128 = 3564;
	/**
	 * {@code VFNMADD231PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 BC /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD231PD_YMM_YMM_YMMM256 = 3565;
	/**
	 * {@code VFNMADD231PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 BC /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD231PS_XMM_K1Z_XMM_XMMM128B32 = 3566;
	/**
	 * {@code VFNMADD231PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 BC /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD231PS_YMM_K1Z_YMM_YMMM256B32 = 3567;
	/**
	 * {@code VFNMADD231PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 BC /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD231PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3568;
	/**
	 * {@code VFNMADD231PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 BC /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD231PD_XMM_K1Z_XMM_XMMM128B64 = 3569;
	/**
	 * {@code VFNMADD231PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 BC /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD231PD_YMM_K1Z_YMM_YMMM256B64 = 3570;
	/**
	 * {@code VFNMADD231PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 BC /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD231PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3571;
	/**
	 * {@code VFNMADD231SS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.66.0F38.W0 BD /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD231SS_XMM_XMM_XMMM32 = 3572;
	/**
	 * {@code VFNMADD231SD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.66.0F38.W1 BD /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADD231SD_XMM_XMM_XMMM64 = 3573;
	/**
	 * {@code VFNMADD231SS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 BD /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD231SS_XMM_K1Z_XMM_XMMM32_ER = 3574;
	/**
	 * {@code VFNMADD231SD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 BD /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD231SD_XMM_K1Z_XMM_XMMM64_ER = 3575;
	/**
	 * {@code VFNMSUB231PS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 BE /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB231PS_XMM_XMM_XMMM128 = 3576;
	/**
	 * {@code VFNMSUB231PS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 BE /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB231PS_YMM_YMM_YMMM256 = 3577;
	/**
	 * {@code VFNMSUB231PD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 BE /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB231PD_XMM_XMM_XMMM128 = 3578;
	/**
	 * {@code VFNMSUB231PD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 BE /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB231PD_YMM_YMM_YMMM256 = 3579;
	/**
	 * {@code VFNMSUB231PS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 BE /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB231PS_XMM_K1Z_XMM_XMMM128B32 = 3580;
	/**
	 * {@code VFNMSUB231PS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 BE /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB231PS_YMM_K1Z_YMM_YMMM256B32 = 3581;
	/**
	 * {@code VFNMSUB231PS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W0 BE /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB231PS_ZMM_K1Z_ZMM_ZMMM512B32_ER = 3582;
	/**
	 * {@code VFNMSUB231PD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 BE /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB231PD_XMM_K1Z_XMM_XMMM128B64 = 3583;
	/**
	 * {@code VFNMSUB231PD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 BE /r}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB231PD_YMM_K1Z_YMM_YMMM256B64 = 3584;
	/**
	 * {@code VFNMSUB231PD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.0F38.W1 BE /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB231PD_ZMM_K1Z_ZMM_ZMMM512B64_ER = 3585;
	/**
	 * {@code VFNMSUB231SS xmm1, xmm2, xmm3/m32}<p>
	 * {@code VEX.LIG.66.0F38.W0 BF /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB231SS_XMM_XMM_XMMM32 = 3586;
	/**
	 * {@code VFNMSUB231SD xmm1, xmm2, xmm3/m64}<p>
	 * {@code VEX.LIG.66.0F38.W1 BF /r}<p>
	 * {@code FMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUB231SD_XMM_XMM_XMMM64 = 3587;
	/**
	 * {@code VFNMSUB231SS xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 BF /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB231SS_XMM_K1Z_XMM_XMMM32_ER = 3588;
	/**
	 * {@code VFNMSUB231SD xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 BF /r}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB231SD_XMM_K1Z_XMM_XMMM64_ER = 3589;
	/**
	 * {@code VPCONFLICTD xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.66.0F38.W0 C4 /r}<p>
	 * {@code AVX512VL and AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCONFLICTD_XMM_K1Z_XMMM128B32 = 3590;
	/**
	 * {@code VPCONFLICTD ymm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.66.0F38.W0 C4 /r}<p>
	 * {@code AVX512VL and AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCONFLICTD_YMM_K1Z_YMMM256B32 = 3591;
	/**
	 * {@code VPCONFLICTD zmm1 {k1}{z}, zmm2/m512/m32bcst}<p>
	 * {@code EVEX.512.66.0F38.W0 C4 /r}<p>
	 * {@code AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCONFLICTD_ZMM_K1Z_ZMMM512B32 = 3592;
	/**
	 * {@code VPCONFLICTQ xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.66.0F38.W1 C4 /r}<p>
	 * {@code AVX512VL and AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCONFLICTQ_XMM_K1Z_XMMM128B64 = 3593;
	/**
	 * {@code VPCONFLICTQ ymm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.66.0F38.W1 C4 /r}<p>
	 * {@code AVX512VL and AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCONFLICTQ_YMM_K1Z_YMMM256B64 = 3594;
	/**
	 * {@code VPCONFLICTQ zmm1 {k1}{z}, zmm2/m512/m64bcst}<p>
	 * {@code EVEX.512.66.0F38.W1 C4 /r}<p>
	 * {@code AVX512CD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCONFLICTQ_ZMM_K1Z_ZMMM512B64 = 3595;
	/**
	 * {@code VGATHERPF0DPS vm32z {k1}}<p>
	 * {@code EVEX.512.66.0F38.W0 C6 /1 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERPF0DPS_VM32Z_K1 = 3596;
	/**
	 * {@code VGATHERPF0DPD vm32y {k1}}<p>
	 * {@code EVEX.512.66.0F38.W1 C6 /1 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERPF0DPD_VM32Y_K1 = 3597;
	/**
	 * {@code VGATHERPF1DPS vm32z {k1}}<p>
	 * {@code EVEX.512.66.0F38.W0 C6 /2 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERPF1DPS_VM32Z_K1 = 3598;
	/**
	 * {@code VGATHERPF1DPD vm32y {k1}}<p>
	 * {@code EVEX.512.66.0F38.W1 C6 /2 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERPF1DPD_VM32Y_K1 = 3599;
	/**
	 * {@code VSCATTERPF0DPS vm32z {k1}}<p>
	 * {@code EVEX.512.66.0F38.W0 C6 /5 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERPF0DPS_VM32Z_K1 = 3600;
	/**
	 * {@code VSCATTERPF0DPD vm32y {k1}}<p>
	 * {@code EVEX.512.66.0F38.W1 C6 /5 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERPF0DPD_VM32Y_K1 = 3601;
	/**
	 * {@code VSCATTERPF1DPS vm32z {k1}}<p>
	 * {@code EVEX.512.66.0F38.W0 C6 /6 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERPF1DPS_VM32Z_K1 = 3602;
	/**
	 * {@code VSCATTERPF1DPD vm32y {k1}}<p>
	 * {@code EVEX.512.66.0F38.W1 C6 /6 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERPF1DPD_VM32Y_K1 = 3603;
	/**
	 * {@code VGATHERPF0QPS vm64z {k1}}<p>
	 * {@code EVEX.512.66.0F38.W0 C7 /1 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERPF0QPS_VM64Z_K1 = 3604;
	/**
	 * {@code VGATHERPF0QPD vm64z {k1}}<p>
	 * {@code EVEX.512.66.0F38.W1 C7 /1 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERPF0QPD_VM64Z_K1 = 3605;
	/**
	 * {@code VGATHERPF1QPS vm64z {k1}}<p>
	 * {@code EVEX.512.66.0F38.W0 C7 /2 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERPF1QPS_VM64Z_K1 = 3606;
	/**
	 * {@code VGATHERPF1QPD vm64z {k1}}<p>
	 * {@code EVEX.512.66.0F38.W1 C7 /2 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGATHERPF1QPD_VM64Z_K1 = 3607;
	/**
	 * {@code VSCATTERPF0QPS vm64z {k1}}<p>
	 * {@code EVEX.512.66.0F38.W0 C7 /5 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERPF0QPS_VM64Z_K1 = 3608;
	/**
	 * {@code VSCATTERPF0QPD vm64z {k1}}<p>
	 * {@code EVEX.512.66.0F38.W1 C7 /5 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERPF0QPD_VM64Z_K1 = 3609;
	/**
	 * {@code VSCATTERPF1QPS vm64z {k1}}<p>
	 * {@code EVEX.512.66.0F38.W0 C7 /6 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERPF1QPS_VM64Z_K1 = 3610;
	/**
	 * {@code VSCATTERPF1QPD vm64z {k1}}<p>
	 * {@code EVEX.512.66.0F38.W1 C7 /6 /vsib}<p>
	 * {@code AVX512PF}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCATTERPF1QPD_VM64Z_K1 = 3611;
	/**
	 * {@code SHA1NEXTE xmm1, xmm2/m128}<p>
	 * {@code NP 0F 38 C8 /r}<p>
	 * {@code SHA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHA1NEXTE_XMM_XMMM128 = 3612;
	/**
	 * {@code VEXP2PS zmm1 {k1}{z}, zmm2/m512/m32bcst{sae}}<p>
	 * {@code EVEX.512.66.0F38.W0 C8 /r}<p>
	 * {@code AVX512ER}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXP2PS_ZMM_K1Z_ZMMM512B32_SAE = 3613;
	/**
	 * {@code VEXP2PD zmm1 {k1}{z}, zmm2/m512/m64bcst{sae}}<p>
	 * {@code EVEX.512.66.0F38.W1 C8 /r}<p>
	 * {@code AVX512ER}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXP2PD_ZMM_K1Z_ZMMM512B64_SAE = 3614;
	/**
	 * {@code SHA1MSG1 xmm1, xmm2/m128}<p>
	 * {@code NP 0F 38 C9 /r}<p>
	 * {@code SHA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHA1MSG1_XMM_XMMM128 = 3615;
	/**
	 * {@code SHA1MSG2 xmm1, xmm2/m128}<p>
	 * {@code NP 0F 38 CA /r}<p>
	 * {@code SHA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHA1MSG2_XMM_XMMM128 = 3616;
	/**
	 * {@code VRCP28PS zmm1 {k1}{z}, zmm2/m512/m32bcst{sae}}<p>
	 * {@code EVEX.512.66.0F38.W0 CA /r}<p>
	 * {@code AVX512ER}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCP28PS_ZMM_K1Z_ZMMM512B32_SAE = 3617;
	/**
	 * {@code VRCP28PD zmm1 {k1}{z}, zmm2/m512/m64bcst{sae}}<p>
	 * {@code EVEX.512.66.0F38.W1 CA /r}<p>
	 * {@code AVX512ER}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCP28PD_ZMM_K1Z_ZMMM512B64_SAE = 3618;
	/**
	 * {@code SHA256RNDS2 xmm1, xmm2/m128, &lt;XMM0&gt;}<p>
	 * {@code NP 0F 38 CB /r}<p>
	 * {@code SHA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHA256RNDS2_XMM_XMMM128 = 3619;
	/**
	 * {@code VRCP28SS xmm1 {k1}{z}, xmm2, xmm3/m32{sae}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 CB /r}<p>
	 * {@code AVX512ER}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCP28SS_XMM_K1Z_XMM_XMMM32_SAE = 3620;
	/**
	 * {@code VRCP28SD xmm1 {k1}{z}, xmm2, xmm3/m64{sae}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 CB /r}<p>
	 * {@code AVX512ER}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCP28SD_XMM_K1Z_XMM_XMMM64_SAE = 3621;
	/**
	 * {@code SHA256MSG1 xmm1, xmm2/m128}<p>
	 * {@code NP 0F 38 CC /r}<p>
	 * {@code SHA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHA256MSG1_XMM_XMMM128 = 3622;
	/**
	 * {@code VRSQRT28PS zmm1 {k1}{z}, zmm2/m512/m32bcst{sae}}<p>
	 * {@code EVEX.512.66.0F38.W0 CC /r}<p>
	 * {@code AVX512ER}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRT28PS_ZMM_K1Z_ZMMM512B32_SAE = 3623;
	/**
	 * {@code VRSQRT28PD zmm1 {k1}{z}, zmm2/m512/m64bcst{sae}}<p>
	 * {@code EVEX.512.66.0F38.W1 CC /r}<p>
	 * {@code AVX512ER}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRT28PD_ZMM_K1Z_ZMMM512B64_SAE = 3624;
	/**
	 * {@code SHA256MSG2 xmm1, xmm2/m128}<p>
	 * {@code NP 0F 38 CD /r}<p>
	 * {@code SHA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHA256MSG2_XMM_XMMM128 = 3625;
	/**
	 * {@code VRSQRT28SS xmm1 {k1}{z}, xmm2, xmm3/m32{sae}}<p>
	 * {@code EVEX.LIG.66.0F38.W0 CD /r}<p>
	 * {@code AVX512ER}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRT28SS_XMM_K1Z_XMM_XMMM32_SAE = 3626;
	/**
	 * {@code VRSQRT28SD xmm1 {k1}{z}, xmm2, xmm3/m64{sae}}<p>
	 * {@code EVEX.LIG.66.0F38.W1 CD /r}<p>
	 * {@code AVX512ER}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRT28SD_XMM_K1Z_XMM_XMMM64_SAE = 3627;
	/**
	 * {@code GF2P8MULB xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 CF /r}<p>
	 * {@code GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int GF2P8MULB_XMM_XMMM128 = 3628;
	/**
	 * {@code VGF2P8MULB xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 CF /r}<p>
	 * {@code AVX and GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VGF2P8MULB_XMM_XMM_XMMM128 = 3629;
	/**
	 * {@code VGF2P8MULB ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 CF /r}<p>
	 * {@code AVX and GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VGF2P8MULB_YMM_YMM_YMMM256 = 3630;
	/**
	 * {@code VGF2P8MULB xmm1 {k1}{z}, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.W0 CF /r}<p>
	 * {@code AVX512VL and GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGF2P8MULB_XMM_K1Z_XMM_XMMM128 = 3631;
	/**
	 * {@code VGF2P8MULB ymm1 {k1}{z}, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.W0 CF /r}<p>
	 * {@code AVX512VL and GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGF2P8MULB_YMM_K1Z_YMM_YMMM256 = 3632;
	/**
	 * {@code VGF2P8MULB zmm1 {k1}{z}, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.W0 CF /r}<p>
	 * {@code AVX512F and GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGF2P8MULB_ZMM_K1Z_ZMM_ZMMM512 = 3633;
	/**
	 * {@code AESIMC xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 DB /r}<p>
	 * {@code AES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AESIMC_XMM_XMMM128 = 3634;
	/**
	 * {@code VAESIMC xmm1, xmm2/m128}<p>
	 * {@code VEX.128.66.0F38.WIG DB /r}<p>
	 * {@code AES and AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VAESIMC_XMM_XMMM128 = 3635;
	/**
	 * {@code AESENC xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 DC /r}<p>
	 * {@code AES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AESENC_XMM_XMMM128 = 3636;
	/**
	 * {@code VAESENC xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG DC /r}<p>
	 * {@code AES and AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VAESENC_XMM_XMM_XMMM128 = 3637;
	/**
	 * {@code VAESENC ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG DC /r}<p>
	 * {@code VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VAESENC_YMM_YMM_YMMM256 = 3638;
	/**
	 * {@code VAESENC xmm1, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.WIG DC /r}<p>
	 * {@code AVX512VL and VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VAESENC_XMM_XMM_XMMM128 = 3639;
	/**
	 * {@code VAESENC ymm1, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.WIG DC /r}<p>
	 * {@code AVX512VL and VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VAESENC_YMM_YMM_YMMM256 = 3640;
	/**
	 * {@code VAESENC zmm1, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.WIG DC /r}<p>
	 * {@code AVX512F and VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VAESENC_ZMM_ZMM_ZMMM512 = 3641;
	/**
	 * {@code AESENCLAST xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 DD /r}<p>
	 * {@code AES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AESENCLAST_XMM_XMMM128 = 3642;
	/**
	 * {@code VAESENCLAST xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG DD /r}<p>
	 * {@code AES and AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VAESENCLAST_XMM_XMM_XMMM128 = 3643;
	/**
	 * {@code VAESENCLAST ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG DD /r}<p>
	 * {@code VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VAESENCLAST_YMM_YMM_YMMM256 = 3644;
	/**
	 * {@code VAESENCLAST xmm1, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.WIG DD /r}<p>
	 * {@code AVX512VL and VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VAESENCLAST_XMM_XMM_XMMM128 = 3645;
	/**
	 * {@code VAESENCLAST ymm1, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.WIG DD /r}<p>
	 * {@code AVX512VL and VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VAESENCLAST_YMM_YMM_YMMM256 = 3646;
	/**
	 * {@code VAESENCLAST zmm1, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.WIG DD /r}<p>
	 * {@code AVX512F and VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VAESENCLAST_ZMM_ZMM_ZMMM512 = 3647;
	/**
	 * {@code AESDEC xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 DE /r}<p>
	 * {@code AES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AESDEC_XMM_XMMM128 = 3648;
	/**
	 * {@code VAESDEC xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG DE /r}<p>
	 * {@code AES and AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VAESDEC_XMM_XMM_XMMM128 = 3649;
	/**
	 * {@code VAESDEC ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG DE /r}<p>
	 * {@code VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VAESDEC_YMM_YMM_YMMM256 = 3650;
	/**
	 * {@code VAESDEC xmm1, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.WIG DE /r}<p>
	 * {@code AVX512VL and VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VAESDEC_XMM_XMM_XMMM128 = 3651;
	/**
	 * {@code VAESDEC ymm1, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.WIG DE /r}<p>
	 * {@code AVX512VL and VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VAESDEC_YMM_YMM_YMMM256 = 3652;
	/**
	 * {@code VAESDEC zmm1, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.WIG DE /r}<p>
	 * {@code AVX512F and VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VAESDEC_ZMM_ZMM_ZMMM512 = 3653;
	/**
	 * {@code AESDECLAST xmm1, xmm2/m128}<p>
	 * {@code 66 0F 38 DF /r}<p>
	 * {@code AES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AESDECLAST_XMM_XMMM128 = 3654;
	/**
	 * {@code VAESDECLAST xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.WIG DF /r}<p>
	 * {@code AES and AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VAESDECLAST_XMM_XMM_XMMM128 = 3655;
	/**
	 * {@code VAESDECLAST ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.WIG DF /r}<p>
	 * {@code VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VAESDECLAST_YMM_YMM_YMMM256 = 3656;
	/**
	 * {@code VAESDECLAST xmm1, xmm2, xmm3/m128}<p>
	 * {@code EVEX.128.66.0F38.WIG DF /r}<p>
	 * {@code AVX512VL and VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VAESDECLAST_XMM_XMM_XMMM128 = 3657;
	/**
	 * {@code VAESDECLAST ymm1, ymm2, ymm3/m256}<p>
	 * {@code EVEX.256.66.0F38.WIG DF /r}<p>
	 * {@code AVX512VL and VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VAESDECLAST_YMM_YMM_YMMM256 = 3658;
	/**
	 * {@code VAESDECLAST zmm1, zmm2, zmm3/m512}<p>
	 * {@code EVEX.512.66.0F38.WIG DF /r}<p>
	 * {@code AVX512F and VAES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VAESDECLAST_ZMM_ZMM_ZMMM512 = 3659;
	/**
	 * {@code MOVBE r16, m16}<p>
	 * {@code o16 0F 38 F0 /r}<p>
	 * {@code MOVBE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVBE_R16_M16 = 3660;
	/**
	 * {@code MOVBE r32, m32}<p>
	 * {@code o32 0F 38 F0 /r}<p>
	 * {@code MOVBE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVBE_R32_M32 = 3661;
	/**
	 * {@code MOVBE r64, m64}<p>
	 * {@code o64 0F 38 F0 /r}<p>
	 * {@code MOVBE}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVBE_R64_M64 = 3662;
	/**
	 * {@code CRC32 r32, r/m8}<p>
	 * {@code F2 0F 38 F0 /r}<p>
	 * {@code SSE4.2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CRC32_R32_RM8 = 3663;
	/**
	 * {@code CRC32 r64, r/m8}<p>
	 * {@code F2 o64 0F 38 F0 /r}<p>
	 * {@code SSE4.2}<p>
	 * {@code 64-bit}
	 */
	public static final int CRC32_R64_RM8 = 3664;
	/**
	 * {@code MOVBE m16, r16}<p>
	 * {@code o16 0F 38 F1 /r}<p>
	 * {@code MOVBE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVBE_M16_R16 = 3665;
	/**
	 * {@code MOVBE m32, r32}<p>
	 * {@code o32 0F 38 F1 /r}<p>
	 * {@code MOVBE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVBE_M32_R32 = 3666;
	/**
	 * {@code MOVBE m64, r64}<p>
	 * {@code o64 0F 38 F1 /r}<p>
	 * {@code MOVBE}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVBE_M64_R64 = 3667;
	/**
	 * {@code CRC32 r32, r/m16}<p>
	 * {@code o16 F2 0F 38 F1 /r}<p>
	 * {@code SSE4.2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CRC32_R32_RM16 = 3668;
	/**
	 * {@code CRC32 r32, r/m32}<p>
	 * {@code o32 F2 0F 38 F1 /r}<p>
	 * {@code SSE4.2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CRC32_R32_RM32 = 3669;
	/**
	 * {@code CRC32 r64, r/m64}<p>
	 * {@code F2 o64 0F 38 F1 /r}<p>
	 * {@code SSE4.2}<p>
	 * {@code 64-bit}
	 */
	public static final int CRC32_R64_RM64 = 3670;
	/**
	 * {@code ANDN r32a, r32b, r/m32}<p>
	 * {@code VEX.LZ.0F38.W0 F2 /r}<p>
	 * {@code BMI1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_ANDN_R32_R32_RM32 = 3671;
	/**
	 * {@code ANDN r64a, r64b, r/m64}<p>
	 * {@code VEX.LZ.0F38.W1 F2 /r}<p>
	 * {@code BMI1}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_ANDN_R64_R64_RM64 = 3672;
	/**
	 * {@code BLSR r32, r/m32}<p>
	 * {@code VEX.LZ.0F38.W0 F3 /1}<p>
	 * {@code BMI1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_BLSR_R32_RM32 = 3673;
	/**
	 * {@code BLSR r64, r/m64}<p>
	 * {@code VEX.LZ.0F38.W1 F3 /1}<p>
	 * {@code BMI1}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_BLSR_R64_RM64 = 3674;
	/**
	 * {@code BLSMSK r32, r/m32}<p>
	 * {@code VEX.LZ.0F38.W0 F3 /2}<p>
	 * {@code BMI1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_BLSMSK_R32_RM32 = 3675;
	/**
	 * {@code BLSMSK r64, r/m64}<p>
	 * {@code VEX.LZ.0F38.W1 F3 /2}<p>
	 * {@code BMI1}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_BLSMSK_R64_RM64 = 3676;
	/**
	 * {@code BLSI r32, r/m32}<p>
	 * {@code VEX.LZ.0F38.W0 F3 /3}<p>
	 * {@code BMI1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_BLSI_R32_RM32 = 3677;
	/**
	 * {@code BLSI r64, r/m64}<p>
	 * {@code VEX.LZ.0F38.W1 F3 /3}<p>
	 * {@code BMI1}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_BLSI_R64_RM64 = 3678;
	/**
	 * {@code BZHI r32a, r/m32, r32b}<p>
	 * {@code VEX.LZ.0F38.W0 F5 /r}<p>
	 * {@code BMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_BZHI_R32_RM32_R32 = 3679;
	/**
	 * {@code BZHI r64a, r/m64, r64b}<p>
	 * {@code VEX.LZ.0F38.W1 F5 /r}<p>
	 * {@code BMI2}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_BZHI_R64_RM64_R64 = 3680;
	/**
	 * {@code WRUSSD m32, r32}<p>
	 * {@code 66 0F 38 F5 /r}<p>
	 * {@code CET_SS}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int WRUSSD_M32_R32 = 3681;
	/**
	 * {@code WRUSSQ m64, r64}<p>
	 * {@code 66 o64 0F 38 F5 /r}<p>
	 * {@code CET_SS}<p>
	 * {@code 64-bit}
	 */
	public static final int WRUSSQ_M64_R64 = 3682;
	/**
	 * {@code PEXT r32a, r32b, r/m32}<p>
	 * {@code VEX.LZ.F3.0F38.W0 F5 /r}<p>
	 * {@code BMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_PEXT_R32_R32_RM32 = 3683;
	/**
	 * {@code PEXT r64a, r64b, r/m64}<p>
	 * {@code VEX.LZ.F3.0F38.W1 F5 /r}<p>
	 * {@code BMI2}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_PEXT_R64_R64_RM64 = 3684;
	/**
	 * {@code PDEP r32a, r32b, r/m32}<p>
	 * {@code VEX.LZ.F2.0F38.W0 F5 /r}<p>
	 * {@code BMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_PDEP_R32_R32_RM32 = 3685;
	/**
	 * {@code PDEP r64a, r64b, r/m64}<p>
	 * {@code VEX.LZ.F2.0F38.W1 F5 /r}<p>
	 * {@code BMI2}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_PDEP_R64_R64_RM64 = 3686;
	/**
	 * {@code WRSSD m32, r32}<p>
	 * {@code NP 0F 38 F6 /r}<p>
	 * {@code CET_SS}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int WRSSD_M32_R32 = 3687;
	/**
	 * {@code WRSSQ m64, r64}<p>
	 * {@code NP o64 0F 38 F6 /r}<p>
	 * {@code CET_SS}<p>
	 * {@code 64-bit}
	 */
	public static final int WRSSQ_M64_R64 = 3688;
	/**
	 * {@code ADCX r32, r/m32}<p>
	 * {@code 66 0F 38 F6 /r}<p>
	 * {@code ADX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADCX_R32_RM32 = 3689;
	/**
	 * {@code ADCX r64, r/m64}<p>
	 * {@code 66 o64 0F 38 F6 /r}<p>
	 * {@code ADX}<p>
	 * {@code 64-bit}
	 */
	public static final int ADCX_R64_RM64 = 3690;
	/**
	 * {@code ADOX r32, r/m32}<p>
	 * {@code F3 0F 38 F6 /r}<p>
	 * {@code ADX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ADOX_R32_RM32 = 3691;
	/**
	 * {@code ADOX r64, r/m64}<p>
	 * {@code F3 o64 0F 38 F6 /r}<p>
	 * {@code ADX}<p>
	 * {@code 64-bit}
	 */
	public static final int ADOX_R64_RM64 = 3692;
	/**
	 * {@code MULX r32a, r32b, r/m32}<p>
	 * {@code VEX.LZ.F2.0F38.W0 F6 /r}<p>
	 * {@code BMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_MULX_R32_R32_RM32 = 3693;
	/**
	 * {@code MULX r64a, r64b, r/m64}<p>
	 * {@code VEX.LZ.F2.0F38.W1 F6 /r}<p>
	 * {@code BMI2}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_MULX_R64_R64_RM64 = 3694;
	/**
	 * {@code BEXTR r32a, r/m32, r32b}<p>
	 * {@code VEX.LZ.0F38.W0 F7 /r}<p>
	 * {@code BMI1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_BEXTR_R32_RM32_R32 = 3695;
	/**
	 * {@code BEXTR r64a, r/m64, r64b}<p>
	 * {@code VEX.LZ.0F38.W1 F7 /r}<p>
	 * {@code BMI1}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_BEXTR_R64_RM64_R64 = 3696;
	/**
	 * {@code SHLX r32a, r/m32, r32b}<p>
	 * {@code VEX.LZ.66.0F38.W0 F7 /r}<p>
	 * {@code BMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_SHLX_R32_RM32_R32 = 3697;
	/**
	 * {@code SHLX r64a, r/m64, r64b}<p>
	 * {@code VEX.LZ.66.0F38.W1 F7 /r}<p>
	 * {@code BMI2}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_SHLX_R64_RM64_R64 = 3698;
	/**
	 * {@code SARX r32a, r/m32, r32b}<p>
	 * {@code VEX.LZ.F3.0F38.W0 F7 /r}<p>
	 * {@code BMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_SARX_R32_RM32_R32 = 3699;
	/**
	 * {@code SARX r64a, r/m64, r64b}<p>
	 * {@code VEX.LZ.F3.0F38.W1 F7 /r}<p>
	 * {@code BMI2}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_SARX_R64_RM64_R64 = 3700;
	/**
	 * {@code SHRX r32a, r/m32, r32b}<p>
	 * {@code VEX.LZ.F2.0F38.W0 F7 /r}<p>
	 * {@code BMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_SHRX_R32_RM32_R32 = 3701;
	/**
	 * {@code SHRX r64a, r/m64, r64b}<p>
	 * {@code VEX.LZ.F2.0F38.W1 F7 /r}<p>
	 * {@code BMI2}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_SHRX_R64_RM64_R64 = 3702;
	/**
	 * {@code MOVDIR64B r16, m512}<p>
	 * {@code a16 66 0F 38 F8 /r}<p>
	 * {@code MOVDIR64B}<p>
	 * {@code 16/32-bit}
	 */
	public static final int MOVDIR64B_R16_M512 = 3703;
	/**
	 * {@code MOVDIR64B r32, m512}<p>
	 * {@code a32 66 0F 38 F8 /r}<p>
	 * {@code MOVDIR64B}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVDIR64B_R32_M512 = 3704;
	/**
	 * {@code MOVDIR64B r64, m512}<p>
	 * {@code a64 66 0F 38 F8 /r}<p>
	 * {@code MOVDIR64B}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVDIR64B_R64_M512 = 3705;
	/**
	 * {@code ENQCMDS r16, m512}<p>
	 * {@code a16 F3 0F 38 F8 !(11):rrr:bbb}<p>
	 * {@code ENQCMD}<p>
	 * {@code 16/32-bit}
	 */
	public static final int ENQCMDS_R16_M512 = 3706;
	/**
	 * {@code ENQCMDS r32, m512}<p>
	 * {@code a32 F3 0F 38 F8 !(11):rrr:bbb}<p>
	 * {@code ENQCMD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ENQCMDS_R32_M512 = 3707;
	/**
	 * {@code ENQCMDS r64, m512}<p>
	 * {@code a64 F3 0F 38 F8 !(11):rrr:bbb}<p>
	 * {@code ENQCMD}<p>
	 * {@code 64-bit}
	 */
	public static final int ENQCMDS_R64_M512 = 3708;
	/**
	 * {@code ENQCMD r16, m512}<p>
	 * {@code a16 F2 0F 38 F8 !(11):rrr:bbb}<p>
	 * {@code ENQCMD}<p>
	 * {@code 16/32-bit}
	 */
	public static final int ENQCMD_R16_M512 = 3709;
	/**
	 * {@code ENQCMD r32, m512}<p>
	 * {@code a32 F2 0F 38 F8 !(11):rrr:bbb}<p>
	 * {@code ENQCMD}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ENQCMD_R32_M512 = 3710;
	/**
	 * {@code ENQCMD r64, m512}<p>
	 * {@code a64 F2 0F 38 F8 !(11):rrr:bbb}<p>
	 * {@code ENQCMD}<p>
	 * {@code 64-bit}
	 */
	public static final int ENQCMD_R64_M512 = 3711;
	/**
	 * {@code MOVDIRI m32, r32}<p>
	 * {@code NP 0F 38 F9 /r}<p>
	 * {@code MOVDIRI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MOVDIRI_M32_R32 = 3712;
	/**
	 * {@code MOVDIRI m64, r64}<p>
	 * {@code NP o64 0F 38 F9 /r}<p>
	 * {@code MOVDIRI}<p>
	 * {@code 64-bit}
	 */
	public static final int MOVDIRI_M64_R64 = 3713;
	/**
	 * {@code VPERMQ ymm1, ymm2/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.W1 00 /r ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMQ_YMM_YMMM256_IMM8 = 3714;
	/**
	 * {@code VPERMQ ymm1 {k1}{z}, ymm2/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 00 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMQ_YMM_K1Z_YMMM256B64_IMM8 = 3715;
	/**
	 * {@code VPERMQ zmm1 {k1}{z}, zmm2/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 00 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMQ_ZMM_K1Z_ZMMM512B64_IMM8 = 3716;
	/**
	 * {@code VPERMPD ymm1, ymm2/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.W1 01 /r ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMPD_YMM_YMMM256_IMM8 = 3717;
	/**
	 * {@code VPERMPD ymm1 {k1}{z}, ymm2/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 01 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMPD_YMM_K1Z_YMMM256B64_IMM8 = 3718;
	/**
	 * {@code VPERMPD zmm1 {k1}{z}, zmm2/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 01 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMPD_ZMM_K1Z_ZMMM512B64_IMM8 = 3719;
	/**
	 * {@code VPBLENDD xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.W0 02 /r ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPBLENDD_XMM_XMM_XMMM128_IMM8 = 3720;
	/**
	 * {@code VPBLENDD ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.W0 02 /r ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPBLENDD_YMM_YMM_YMMM256_IMM8 = 3721;
	/**
	 * {@code VALIGND xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 03 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VALIGND_XMM_K1Z_XMM_XMMM128B32_IMM8 = 3722;
	/**
	 * {@code VALIGND ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 03 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VALIGND_YMM_K1Z_YMM_YMMM256B32_IMM8 = 3723;
	/**
	 * {@code VALIGND zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 03 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VALIGND_ZMM_K1Z_ZMM_ZMMM512B32_IMM8 = 3724;
	/**
	 * {@code VALIGNQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 03 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VALIGNQ_XMM_K1Z_XMM_XMMM128B64_IMM8 = 3725;
	/**
	 * {@code VALIGNQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 03 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VALIGNQ_YMM_K1Z_YMM_YMMM256B64_IMM8 = 3726;
	/**
	 * {@code VALIGNQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 03 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VALIGNQ_ZMM_K1Z_ZMM_ZMMM512B64_IMM8 = 3727;
	/**
	 * {@code VPERMILPS xmm1, xmm2/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.W0 04 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMILPS_XMM_XMMM128_IMM8 = 3728;
	/**
	 * {@code VPERMILPS ymm1, ymm2/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.W0 04 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMILPS_YMM_YMMM256_IMM8 = 3729;
	/**
	 * {@code VPERMILPS xmm1 {k1}{z}, xmm2/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 04 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMILPS_XMM_K1Z_XMMM128B32_IMM8 = 3730;
	/**
	 * {@code VPERMILPS ymm1 {k1}{z}, ymm2/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 04 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMILPS_YMM_K1Z_YMMM256B32_IMM8 = 3731;
	/**
	 * {@code VPERMILPS zmm1 {k1}{z}, zmm2/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 04 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMILPS_ZMM_K1Z_ZMMM512B32_IMM8 = 3732;
	/**
	 * {@code VPERMILPD xmm1, xmm2/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.W0 05 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMILPD_XMM_XMMM128_IMM8 = 3733;
	/**
	 * {@code VPERMILPD ymm1, ymm2/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.W0 05 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMILPD_YMM_YMMM256_IMM8 = 3734;
	/**
	 * {@code VPERMILPD xmm1 {k1}{z}, xmm2/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 05 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMILPD_XMM_K1Z_XMMM128B64_IMM8 = 3735;
	/**
	 * {@code VPERMILPD ymm1 {k1}{z}, ymm2/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 05 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMILPD_YMM_K1Z_YMMM256B64_IMM8 = 3736;
	/**
	 * {@code VPERMILPD zmm1 {k1}{z}, zmm2/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 05 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPERMILPD_ZMM_K1Z_ZMMM512B64_IMM8 = 3737;
	/**
	 * {@code VPERM2F128 ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.W0 06 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERM2F128_YMM_YMM_YMMM256_IMM8 = 3738;
	/**
	 * {@code ROUNDPS xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A 08 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROUNDPS_XMM_XMMM128_IMM8 = 3739;
	/**
	 * {@code VROUNDPS xmm1, xmm2/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.WIG 08 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VROUNDPS_XMM_XMMM128_IMM8 = 3740;
	/**
	 * {@code VROUNDPS ymm1, ymm2/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.WIG 08 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VROUNDPS_YMM_YMMM256_IMM8 = 3741;
	/**
	 * {@code VRNDSCALEPS xmm1 {k1}{z}, xmm2/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 08 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRNDSCALEPS_XMM_K1Z_XMMM128B32_IMM8 = 3742;
	/**
	 * {@code VRNDSCALEPS ymm1 {k1}{z}, ymm2/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 08 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRNDSCALEPS_YMM_K1Z_YMMM256B32_IMM8 = 3743;
	/**
	 * {@code VRNDSCALEPS zmm1 {k1}{z}, zmm2/m512/m32bcst{sae}, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 08 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRNDSCALEPS_ZMM_K1Z_ZMMM512B32_IMM8_SAE = 3744;
	/**
	 * {@code ROUNDPD xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A 09 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROUNDPD_XMM_XMMM128_IMM8 = 3745;
	/**
	 * {@code VROUNDPD xmm1, xmm2/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.WIG 09 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VROUNDPD_XMM_XMMM128_IMM8 = 3746;
	/**
	 * {@code VROUNDPD ymm1, ymm2/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.WIG 09 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VROUNDPD_YMM_YMMM256_IMM8 = 3747;
	/**
	 * {@code VRNDSCALEPD xmm1 {k1}{z}, xmm2/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 09 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRNDSCALEPD_XMM_K1Z_XMMM128B64_IMM8 = 3748;
	/**
	 * {@code VRNDSCALEPD ymm1 {k1}{z}, ymm2/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 09 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRNDSCALEPD_YMM_K1Z_YMMM256B64_IMM8 = 3749;
	/**
	 * {@code VRNDSCALEPD zmm1 {k1}{z}, zmm2/m512/m64bcst{sae}, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 09 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRNDSCALEPD_ZMM_K1Z_ZMMM512B64_IMM8_SAE = 3750;
	/**
	 * {@code ROUNDSS xmm1, xmm2/m32, imm8}<p>
	 * {@code 66 0F 3A 0A /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROUNDSS_XMM_XMMM32_IMM8 = 3751;
	/**
	 * {@code VROUNDSS xmm1, xmm2, xmm3/m32, imm8}<p>
	 * {@code VEX.LIG.66.0F3A.WIG 0A /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VROUNDSS_XMM_XMM_XMMM32_IMM8 = 3752;
	/**
	 * {@code VRNDSCALESS xmm1 {k1}{z}, xmm2, xmm3/m32{sae}, imm8}<p>
	 * {@code EVEX.LIG.66.0F3A.W0 0A /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRNDSCALESS_XMM_K1Z_XMM_XMMM32_IMM8_SAE = 3753;
	/**
	 * {@code ROUNDSD xmm1, xmm2/m64, imm8}<p>
	 * {@code 66 0F 3A 0B /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ROUNDSD_XMM_XMMM64_IMM8 = 3754;
	/**
	 * {@code VROUNDSD xmm1, xmm2, xmm3/m64, imm8}<p>
	 * {@code VEX.LIG.66.0F3A.WIG 0B /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VROUNDSD_XMM_XMM_XMMM64_IMM8 = 3755;
	/**
	 * {@code VRNDSCALESD xmm1 {k1}{z}, xmm2, xmm3/m64{sae}, imm8}<p>
	 * {@code EVEX.LIG.66.0F3A.W1 0B /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRNDSCALESD_XMM_K1Z_XMM_XMMM64_IMM8_SAE = 3756;
	/**
	 * {@code BLENDPS xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A 0C /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BLENDPS_XMM_XMMM128_IMM8 = 3757;
	/**
	 * {@code VBLENDPS xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.WIG 0C /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBLENDPS_XMM_XMM_XMMM128_IMM8 = 3758;
	/**
	 * {@code VBLENDPS ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.WIG 0C /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBLENDPS_YMM_YMM_YMMM256_IMM8 = 3759;
	/**
	 * {@code BLENDPD xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A 0D /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int BLENDPD_XMM_XMMM128_IMM8 = 3760;
	/**
	 * {@code VBLENDPD xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.WIG 0D /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBLENDPD_XMM_XMM_XMMM128_IMM8 = 3761;
	/**
	 * {@code VBLENDPD ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.WIG 0D /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBLENDPD_YMM_YMM_YMMM256_IMM8 = 3762;
	/**
	 * {@code PBLENDW xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A 0E /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PBLENDW_XMM_XMMM128_IMM8 = 3763;
	/**
	 * {@code VPBLENDW xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.WIG 0E /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPBLENDW_XMM_XMM_XMMM128_IMM8 = 3764;
	/**
	 * {@code VPBLENDW ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.WIG 0E /r ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPBLENDW_YMM_YMM_YMMM256_IMM8 = 3765;
	/**
	 * {@code PALIGNR mm1, mm2/m64, imm8}<p>
	 * {@code NP 0F 3A 0F /r ib}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PALIGNR_MM_MMM64_IMM8 = 3766;
	/**
	 * {@code PALIGNR xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A 0F /r ib}<p>
	 * {@code SSSE3}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PALIGNR_XMM_XMMM128_IMM8 = 3767;
	/**
	 * {@code VPALIGNR xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.WIG 0F /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPALIGNR_XMM_XMM_XMMM128_IMM8 = 3768;
	/**
	 * {@code VPALIGNR ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.WIG 0F /r ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPALIGNR_YMM_YMM_YMMM256_IMM8 = 3769;
	/**
	 * {@code VPALIGNR xmm1 {k1}{z}, xmm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.128.66.0F3A.WIG 0F /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPALIGNR_XMM_K1Z_XMM_XMMM128_IMM8 = 3770;
	/**
	 * {@code VPALIGNR ymm1 {k1}{z}, ymm2, ymm3/m256, imm8}<p>
	 * {@code EVEX.256.66.0F3A.WIG 0F /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPALIGNR_YMM_K1Z_YMM_YMMM256_IMM8 = 3771;
	/**
	 * {@code VPALIGNR zmm1 {k1}{z}, zmm2, zmm3/m512, imm8}<p>
	 * {@code EVEX.512.66.0F3A.WIG 0F /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPALIGNR_ZMM_K1Z_ZMM_ZMMM512_IMM8 = 3772;
	/**
	 * {@code PEXTRB r32/m8, xmm2, imm8}<p>
	 * {@code 66 0F 3A 14 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PEXTRB_R32M8_XMM_IMM8 = 3773;
	/**
	 * {@code PEXTRB r64/m8, xmm2, imm8}<p>
	 * {@code 66 o64 0F 3A 14 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 64-bit}
	 */
	public static final int PEXTRB_R64M8_XMM_IMM8 = 3774;
	/**
	 * {@code VPEXTRB r32/m8, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F3A.W0 14 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPEXTRB_R32M8_XMM_IMM8 = 3775;
	/**
	 * {@code VPEXTRB r64/m8, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F3A.W1 14 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VPEXTRB_R64M8_XMM_IMM8 = 3776;
	/**
	 * {@code VPEXTRB r32/m8, xmm2, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 14 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXTRB_R32M8_XMM_IMM8 = 3777;
	/**
	 * {@code VPEXTRB r64/m8, xmm2, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 14 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VPEXTRB_R64M8_XMM_IMM8 = 3778;
	/**
	 * {@code PEXTRW r32/m16, xmm, imm8}<p>
	 * {@code 66 0F 3A 15 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PEXTRW_R32M16_XMM_IMM8 = 3779;
	/**
	 * {@code PEXTRW r64/m16, xmm, imm8}<p>
	 * {@code 66 o64 0F 3A 15 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 64-bit}
	 */
	public static final int PEXTRW_R64M16_XMM_IMM8 = 3780;
	/**
	 * {@code VPEXTRW r32/m16, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F3A.W0 15 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPEXTRW_R32M16_XMM_IMM8 = 3781;
	/**
	 * {@code VPEXTRW r64/m16, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F3A.W1 15 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VPEXTRW_R64M16_XMM_IMM8 = 3782;
	/**
	 * {@code VPEXTRW r32/m16, xmm2, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 15 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXTRW_R32M16_XMM_IMM8 = 3783;
	/**
	 * {@code VPEXTRW r64/m16, xmm2, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 15 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VPEXTRW_R64M16_XMM_IMM8 = 3784;
	/**
	 * {@code PEXTRD r/m32, xmm2, imm8}<p>
	 * {@code 66 0F 3A 16 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PEXTRD_RM32_XMM_IMM8 = 3785;
	/**
	 * {@code PEXTRQ r/m64, xmm2, imm8}<p>
	 * {@code 66 o64 0F 3A 16 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 64-bit}
	 */
	public static final int PEXTRQ_RM64_XMM_IMM8 = 3786;
	/**
	 * {@code VPEXTRD r/m32, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F3A.W0 16 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPEXTRD_RM32_XMM_IMM8 = 3787;
	/**
	 * {@code VPEXTRQ r/m64, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F3A.W1 16 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VPEXTRQ_RM64_XMM_IMM8 = 3788;
	/**
	 * {@code VPEXTRD r/m32, xmm2, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 16 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPEXTRD_RM32_XMM_IMM8 = 3789;
	/**
	 * {@code VPEXTRQ r/m64, xmm2, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 16 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VPEXTRQ_RM64_XMM_IMM8 = 3790;
	/**
	 * {@code EXTRACTPS r/m32, xmm1, imm8}<p>
	 * {@code 66 0F 3A 17 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EXTRACTPS_RM32_XMM_IMM8 = 3791;
	/**
	 * {@code EXTRACTPS r64/m32, xmm1, imm8}<p>
	 * {@code 66 o64 0F 3A 17 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 64-bit}
	 */
	public static final int EXTRACTPS_R64M32_XMM_IMM8 = 3792;
	/**
	 * {@code VEXTRACTPS r/m32, xmm1, imm8}<p>
	 * {@code VEX.128.66.0F3A.W0 17 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VEXTRACTPS_RM32_XMM_IMM8 = 3793;
	/**
	 * {@code VEXTRACTPS r64/m32, xmm1, imm8}<p>
	 * {@code VEX.128.66.0F3A.W1 17 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VEXTRACTPS_R64M32_XMM_IMM8 = 3794;
	/**
	 * {@code VEXTRACTPS r/m32, xmm1, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 17 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXTRACTPS_RM32_XMM_IMM8 = 3795;
	/**
	 * {@code VEXTRACTPS r64/m32, xmm1, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 17 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VEXTRACTPS_R64M32_XMM_IMM8 = 3796;
	/**
	 * {@code VINSERTF128 ymm1, ymm2, xmm3/m128, imm8}<p>
	 * {@code VEX.256.66.0F3A.W0 18 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VINSERTF128_YMM_YMM_XMMM128_IMM8 = 3797;
	/**
	 * {@code VINSERTF32X4 ymm1 {k1}{z}, ymm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 18 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VINSERTF32X4_YMM_K1Z_YMM_XMMM128_IMM8 = 3798;
	/**
	 * {@code VINSERTF32X4 zmm1 {k1}{z}, zmm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 18 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VINSERTF32X4_ZMM_K1Z_ZMM_XMMM128_IMM8 = 3799;
	/**
	 * {@code VINSERTF64X2 ymm1 {k1}{z}, ymm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 18 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VINSERTF64X2_YMM_K1Z_YMM_XMMM128_IMM8 = 3800;
	/**
	 * {@code VINSERTF64X2 zmm1 {k1}{z}, zmm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 18 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VINSERTF64X2_ZMM_K1Z_ZMM_XMMM128_IMM8 = 3801;
	/**
	 * {@code VEXTRACTF128 xmm1/m128, ymm2, imm8}<p>
	 * {@code VEX.256.66.0F3A.W0 19 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VEXTRACTF128_XMMM128_YMM_IMM8 = 3802;
	/**
	 * {@code VEXTRACTF32X4 xmm1/m128 {k1}{z}, ymm2, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 19 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXTRACTF32X4_XMMM128_K1Z_YMM_IMM8 = 3803;
	/**
	 * {@code VEXTRACTF32X4 xmm1/m128 {k1}{z}, zmm2, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 19 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXTRACTF32X4_XMMM128_K1Z_ZMM_IMM8 = 3804;
	/**
	 * {@code VEXTRACTF64X2 xmm1/m128 {k1}{z}, ymm2, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 19 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXTRACTF64X2_XMMM128_K1Z_YMM_IMM8 = 3805;
	/**
	 * {@code VEXTRACTF64X2 xmm1/m128 {k1}{z}, zmm2, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 19 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXTRACTF64X2_XMMM128_K1Z_ZMM_IMM8 = 3806;
	/**
	 * {@code VINSERTF32X8 zmm1 {k1}{z}, zmm2, ymm3/m256, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 1A /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VINSERTF32X8_ZMM_K1Z_ZMM_YMMM256_IMM8 = 3807;
	/**
	 * {@code VINSERTF64X4 zmm1 {k1}{z}, zmm2, ymm3/m256, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 1A /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VINSERTF64X4_ZMM_K1Z_ZMM_YMMM256_IMM8 = 3808;
	/**
	 * {@code VEXTRACTF32X8 ymm1/m256 {k1}{z}, zmm2, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 1B /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXTRACTF32X8_YMMM256_K1Z_ZMM_IMM8 = 3809;
	/**
	 * {@code VEXTRACTF64X4 ymm1/m256 {k1}{z}, zmm2, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 1B /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXTRACTF64X4_YMMM256_K1Z_ZMM_IMM8 = 3810;
	/**
	 * {@code VCVTPS2PH xmm1/m64, xmm2, imm8}<p>
	 * {@code VEX.128.66.0F3A.W0 1D /r ib}<p>
	 * {@code F16C}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTPS2PH_XMMM64_XMM_IMM8 = 3811;
	/**
	 * {@code VCVTPS2PH xmm1/m128, ymm2, imm8}<p>
	 * {@code VEX.256.66.0F3A.W0 1D /r ib}<p>
	 * {@code F16C}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTPS2PH_XMMM128_YMM_IMM8 = 3812;
	/**
	 * {@code VCVTPS2PH xmm1/m64 {k1}{z}, xmm2, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 1D /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2PH_XMMM64_K1Z_XMM_IMM8 = 3813;
	/**
	 * {@code VCVTPS2PH xmm1/m128 {k1}{z}, ymm2, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 1D /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2PH_XMMM128_K1Z_YMM_IMM8 = 3814;
	/**
	 * {@code VCVTPS2PH ymm1/m256 {k1}{z}, zmm2{sae}, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 1D /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2PH_YMMM256_K1Z_ZMM_IMM8_SAE = 3815;
	/**
	 * {@code VPCMPUD k1 {k2}, xmm2, xmm3/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 1E /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPUD_KR_K1_XMM_XMMM128B32_IMM8 = 3816;
	/**
	 * {@code VPCMPUD k1 {k2}, ymm2, ymm3/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 1E /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPUD_KR_K1_YMM_YMMM256B32_IMM8 = 3817;
	/**
	 * {@code VPCMPUD k1 {k2}, zmm2, zmm3/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 1E /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPUD_KR_K1_ZMM_ZMMM512B32_IMM8 = 3818;
	/**
	 * {@code VPCMPUQ k1 {k2}, xmm2, xmm3/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 1E /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPUQ_KR_K1_XMM_XMMM128B64_IMM8 = 3819;
	/**
	 * {@code VPCMPUQ k1 {k2}, ymm2, ymm3/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 1E /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPUQ_KR_K1_YMM_YMMM256B64_IMM8 = 3820;
	/**
	 * {@code VPCMPUQ k1 {k2}, zmm2, zmm3/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 1E /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPUQ_KR_K1_ZMM_ZMMM512B64_IMM8 = 3821;
	/**
	 * {@code VPCMPD k1 {k2}, xmm2, xmm3/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 1F /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPD_KR_K1_XMM_XMMM128B32_IMM8 = 3822;
	/**
	 * {@code VPCMPD k1 {k2}, ymm2, ymm3/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 1F /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPD_KR_K1_YMM_YMMM256B32_IMM8 = 3823;
	/**
	 * {@code VPCMPD k1 {k2}, zmm2, zmm3/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 1F /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPD_KR_K1_ZMM_ZMMM512B32_IMM8 = 3824;
	/**
	 * {@code VPCMPQ k1 {k2}, xmm2, xmm3/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 1F /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPQ_KR_K1_XMM_XMMM128B64_IMM8 = 3825;
	/**
	 * {@code VPCMPQ k1 {k2}, ymm2, ymm3/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 1F /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPQ_KR_K1_YMM_YMMM256B64_IMM8 = 3826;
	/**
	 * {@code VPCMPQ k1 {k2}, zmm2, zmm3/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 1F /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPQ_KR_K1_ZMM_ZMMM512B64_IMM8 = 3827;
	/**
	 * {@code PINSRB xmm1, r32/m8, imm8}<p>
	 * {@code 66 0F 3A 20 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PINSRB_XMM_R32M8_IMM8 = 3828;
	/**
	 * {@code PINSRB xmm1, r64/m8, imm8}<p>
	 * {@code 66 o64 0F 3A 20 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 64-bit}
	 */
	public static final int PINSRB_XMM_R64M8_IMM8 = 3829;
	/**
	 * {@code VPINSRB xmm1, xmm2, r32/m8, imm8}<p>
	 * {@code VEX.128.66.0F3A.W0 20 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPINSRB_XMM_XMM_R32M8_IMM8 = 3830;
	/**
	 * {@code VPINSRB xmm1, xmm2, r64/m8, imm8}<p>
	 * {@code VEX.128.66.0F3A.W1 20 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VPINSRB_XMM_XMM_R64M8_IMM8 = 3831;
	/**
	 * {@code VPINSRB xmm1, xmm2, r32/m8, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 20 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPINSRB_XMM_XMM_R32M8_IMM8 = 3832;
	/**
	 * {@code VPINSRB xmm1, xmm2, r64/m8, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 20 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VPINSRB_XMM_XMM_R64M8_IMM8 = 3833;
	/**
	 * {@code INSERTPS xmm1, xmm2/m32, imm8}<p>
	 * {@code 66 0F 3A 21 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INSERTPS_XMM_XMMM32_IMM8 = 3834;
	/**
	 * {@code VINSERTPS xmm1, xmm2, xmm3/m32, imm8}<p>
	 * {@code VEX.128.66.0F3A.WIG 21 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VINSERTPS_XMM_XMM_XMMM32_IMM8 = 3835;
	/**
	 * {@code VINSERTPS xmm1, xmm2, xmm3/m32, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 21 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VINSERTPS_XMM_XMM_XMMM32_IMM8 = 3836;
	/**
	 * {@code PINSRD xmm1, r/m32, imm8}<p>
	 * {@code 66 0F 3A 22 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PINSRD_XMM_RM32_IMM8 = 3837;
	/**
	 * {@code PINSRQ xmm1, r/m64, imm8}<p>
	 * {@code 66 o64 0F 3A 22 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 64-bit}
	 */
	public static final int PINSRQ_XMM_RM64_IMM8 = 3838;
	/**
	 * {@code VPINSRD xmm1, xmm2, r/m32, imm8}<p>
	 * {@code VEX.128.66.0F3A.W0 22 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPINSRD_XMM_XMM_RM32_IMM8 = 3839;
	/**
	 * {@code VPINSRQ xmm1, xmm2, r/m64, imm8}<p>
	 * {@code VEX.128.66.0F3A.W1 22 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VPINSRQ_XMM_XMM_RM64_IMM8 = 3840;
	/**
	 * {@code VPINSRD xmm1, xmm2, r/m32, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 22 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPINSRD_XMM_XMM_RM32_IMM8 = 3841;
	/**
	 * {@code VPINSRQ xmm1, xmm2, r/m64, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 22 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VPINSRQ_XMM_XMM_RM64_IMM8 = 3842;
	/**
	 * {@code VSHUFF32X4 ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 23 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSHUFF32X4_YMM_K1Z_YMM_YMMM256B32_IMM8 = 3843;
	/**
	 * {@code VSHUFF32X4 zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 23 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSHUFF32X4_ZMM_K1Z_ZMM_ZMMM512B32_IMM8 = 3844;
	/**
	 * {@code VSHUFF64X2 ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 23 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSHUFF64X2_YMM_K1Z_YMM_YMMM256B64_IMM8 = 3845;
	/**
	 * {@code VSHUFF64X2 zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 23 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSHUFF64X2_ZMM_K1Z_ZMM_ZMMM512B64_IMM8 = 3846;
	/**
	 * {@code VPTERNLOGD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 25 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTERNLOGD_XMM_K1Z_XMM_XMMM128B32_IMM8 = 3847;
	/**
	 * {@code VPTERNLOGD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 25 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTERNLOGD_YMM_K1Z_YMM_YMMM256B32_IMM8 = 3848;
	/**
	 * {@code VPTERNLOGD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 25 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTERNLOGD_ZMM_K1Z_ZMM_ZMMM512B32_IMM8 = 3849;
	/**
	 * {@code VPTERNLOGQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 25 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTERNLOGQ_XMM_K1Z_XMM_XMMM128B64_IMM8 = 3850;
	/**
	 * {@code VPTERNLOGQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 25 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTERNLOGQ_YMM_K1Z_YMM_YMMM256B64_IMM8 = 3851;
	/**
	 * {@code VPTERNLOGQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 25 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPTERNLOGQ_ZMM_K1Z_ZMM_ZMMM512B64_IMM8 = 3852;
	/**
	 * {@code VGETMANTPS xmm1 {k1}{z}, xmm2/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 26 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETMANTPS_XMM_K1Z_XMMM128B32_IMM8 = 3853;
	/**
	 * {@code VGETMANTPS ymm1 {k1}{z}, ymm2/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 26 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETMANTPS_YMM_K1Z_YMMM256B32_IMM8 = 3854;
	/**
	 * {@code VGETMANTPS zmm1 {k1}{z}, zmm2/m512/m32bcst{sae}, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 26 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETMANTPS_ZMM_K1Z_ZMMM512B32_IMM8_SAE = 3855;
	/**
	 * {@code VGETMANTPD xmm1 {k1}{z}, xmm2/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 26 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETMANTPD_XMM_K1Z_XMMM128B64_IMM8 = 3856;
	/**
	 * {@code VGETMANTPD ymm1 {k1}{z}, ymm2/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 26 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETMANTPD_YMM_K1Z_YMMM256B64_IMM8 = 3857;
	/**
	 * {@code VGETMANTPD zmm1 {k1}{z}, zmm2/m512/m64bcst{sae}, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 26 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETMANTPD_ZMM_K1Z_ZMMM512B64_IMM8_SAE = 3858;
	/**
	 * {@code VGETMANTSS xmm1 {k1}{z}, xmm2, xmm3/m32{sae}, imm8}<p>
	 * {@code EVEX.LIG.66.0F3A.W0 27 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETMANTSS_XMM_K1Z_XMM_XMMM32_IMM8_SAE = 3859;
	/**
	 * {@code VGETMANTSD xmm1 {k1}{z}, xmm2, xmm3/m64{sae}, imm8}<p>
	 * {@code EVEX.LIG.66.0F3A.W1 27 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETMANTSD_XMM_K1Z_XMM_XMMM64_IMM8_SAE = 3860;
	/**
	 * {@code KSHIFTRB k1, k2, imm8}<p>
	 * {@code VEX.L0.66.0F3A.W0 30 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KSHIFTRB_KR_KR_IMM8 = 3861;
	/**
	 * {@code KSHIFTRW k1, k2, imm8}<p>
	 * {@code VEX.L0.66.0F3A.W1 30 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KSHIFTRW_KR_KR_IMM8 = 3862;
	/**
	 * {@code KSHIFTRD k1, k2, imm8}<p>
	 * {@code VEX.L0.66.0F3A.W0 31 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KSHIFTRD_KR_KR_IMM8 = 3863;
	/**
	 * {@code KSHIFTRQ k1, k2, imm8}<p>
	 * {@code VEX.L0.66.0F3A.W1 31 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KSHIFTRQ_KR_KR_IMM8 = 3864;
	/**
	 * {@code KSHIFTLB k1, k2, imm8}<p>
	 * {@code VEX.L0.66.0F3A.W0 32 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KSHIFTLB_KR_KR_IMM8 = 3865;
	/**
	 * {@code KSHIFTLW k1, k2, imm8}<p>
	 * {@code VEX.L0.66.0F3A.W1 32 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KSHIFTLW_KR_KR_IMM8 = 3866;
	/**
	 * {@code KSHIFTLD k1, k2, imm8}<p>
	 * {@code VEX.L0.66.0F3A.W0 33 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KSHIFTLD_KR_KR_IMM8 = 3867;
	/**
	 * {@code KSHIFTLQ k1, k2, imm8}<p>
	 * {@code VEX.L0.66.0F3A.W1 33 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_KSHIFTLQ_KR_KR_IMM8 = 3868;
	/**
	 * {@code VINSERTI128 ymm1, ymm2, xmm3/m128, imm8}<p>
	 * {@code VEX.256.66.0F3A.W0 38 /r ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VINSERTI128_YMM_YMM_XMMM128_IMM8 = 3869;
	/**
	 * {@code VINSERTI32X4 ymm1 {k1}{z}, ymm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 38 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VINSERTI32X4_YMM_K1Z_YMM_XMMM128_IMM8 = 3870;
	/**
	 * {@code VINSERTI32X4 zmm1 {k1}{z}, zmm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 38 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VINSERTI32X4_ZMM_K1Z_ZMM_XMMM128_IMM8 = 3871;
	/**
	 * {@code VINSERTI64X2 ymm1 {k1}{z}, ymm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 38 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VINSERTI64X2_YMM_K1Z_YMM_XMMM128_IMM8 = 3872;
	/**
	 * {@code VINSERTI64X2 zmm1 {k1}{z}, zmm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 38 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VINSERTI64X2_ZMM_K1Z_ZMM_XMMM128_IMM8 = 3873;
	/**
	 * {@code VEXTRACTI128 xmm1/m128, ymm2, imm8}<p>
	 * {@code VEX.256.66.0F3A.W0 39 /r ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VEXTRACTI128_XMMM128_YMM_IMM8 = 3874;
	/**
	 * {@code VEXTRACTI32X4 xmm1/m128 {k1}{z}, ymm2, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 39 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXTRACTI32X4_XMMM128_K1Z_YMM_IMM8 = 3875;
	/**
	 * {@code VEXTRACTI32X4 xmm1/m128 {k1}{z}, zmm2, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 39 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXTRACTI32X4_XMMM128_K1Z_ZMM_IMM8 = 3876;
	/**
	 * {@code VEXTRACTI64X2 xmm1/m128 {k1}{z}, ymm2, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 39 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXTRACTI64X2_XMMM128_K1Z_YMM_IMM8 = 3877;
	/**
	 * {@code VEXTRACTI64X2 xmm1/m128 {k1}{z}, zmm2, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 39 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXTRACTI64X2_XMMM128_K1Z_ZMM_IMM8 = 3878;
	/**
	 * {@code VINSERTI32X8 zmm1 {k1}{z}, zmm2, ymm3/m256, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 3A /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VINSERTI32X8_ZMM_K1Z_ZMM_YMMM256_IMM8 = 3879;
	/**
	 * {@code VINSERTI64X4 zmm1 {k1}{z}, zmm2, ymm3/m256, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 3A /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VINSERTI64X4_ZMM_K1Z_ZMM_YMMM256_IMM8 = 3880;
	/**
	 * {@code VEXTRACTI32X8 ymm1/m256 {k1}{z}, zmm2, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 3B /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXTRACTI32X8_YMMM256_K1Z_ZMM_IMM8 = 3881;
	/**
	 * {@code VEXTRACTI64X4 ymm1/m256 {k1}{z}, zmm2, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 3B /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VEXTRACTI64X4_YMMM256_K1Z_ZMM_IMM8 = 3882;
	/**
	 * {@code VPCMPUB k1 {k2}, xmm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 3E /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPUB_KR_K1_XMM_XMMM128_IMM8 = 3883;
	/**
	 * {@code VPCMPUB k1 {k2}, ymm2, ymm3/m256, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 3E /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPUB_KR_K1_YMM_YMMM256_IMM8 = 3884;
	/**
	 * {@code VPCMPUB k1 {k2}, zmm2, zmm3/m512, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 3E /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPUB_KR_K1_ZMM_ZMMM512_IMM8 = 3885;
	/**
	 * {@code VPCMPUW k1 {k2}, xmm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 3E /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPUW_KR_K1_XMM_XMMM128_IMM8 = 3886;
	/**
	 * {@code VPCMPUW k1 {k2}, ymm2, ymm3/m256, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 3E /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPUW_KR_K1_YMM_YMMM256_IMM8 = 3887;
	/**
	 * {@code VPCMPUW k1 {k2}, zmm2, zmm3/m512, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 3E /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPUW_KR_K1_ZMM_ZMMM512_IMM8 = 3888;
	/**
	 * {@code VPCMPB k1 {k2}, xmm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 3F /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPB_KR_K1_XMM_XMMM128_IMM8 = 3889;
	/**
	 * {@code VPCMPB k1 {k2}, ymm2, ymm3/m256, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 3F /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPB_KR_K1_YMM_YMMM256_IMM8 = 3890;
	/**
	 * {@code VPCMPB k1 {k2}, zmm2, zmm3/m512, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 3F /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPB_KR_K1_ZMM_ZMMM512_IMM8 = 3891;
	/**
	 * {@code VPCMPW k1 {k2}, xmm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 3F /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPW_KR_K1_XMM_XMMM128_IMM8 = 3892;
	/**
	 * {@code VPCMPW k1 {k2}, ymm2, ymm3/m256, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 3F /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPW_KR_K1_YMM_YMMM256_IMM8 = 3893;
	/**
	 * {@code VPCMPW k1 {k2}, zmm2, zmm3/m512, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 3F /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCMPW_KR_K1_ZMM_ZMMM512_IMM8 = 3894;
	/**
	 * {@code DPPS xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A 40 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int DPPS_XMM_XMMM128_IMM8 = 3895;
	/**
	 * {@code VDPPS xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.WIG 40 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VDPPS_XMM_XMM_XMMM128_IMM8 = 3896;
	/**
	 * {@code VDPPS ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.WIG 40 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VDPPS_YMM_YMM_YMMM256_IMM8 = 3897;
	/**
	 * {@code DPPD xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A 41 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int DPPD_XMM_XMMM128_IMM8 = 3898;
	/**
	 * {@code VDPPD xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.WIG 41 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VDPPD_XMM_XMM_XMMM128_IMM8 = 3899;
	/**
	 * {@code MPSADBW xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A 42 /r ib}<p>
	 * {@code SSE4.1}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int MPSADBW_XMM_XMMM128_IMM8 = 3900;
	/**
	 * {@code VMPSADBW xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.WIG 42 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMPSADBW_XMM_XMM_XMMM128_IMM8 = 3901;
	/**
	 * {@code VMPSADBW ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.WIG 42 /r ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VMPSADBW_YMM_YMM_YMMM256_IMM8 = 3902;
	/**
	 * {@code VDBPSADBW xmm1 {k1}{z}, xmm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 42 /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDBPSADBW_XMM_K1Z_XMM_XMMM128_IMM8 = 3903;
	/**
	 * {@code VDBPSADBW ymm1 {k1}{z}, ymm2, ymm3/m256, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 42 /r ib}<p>
	 * {@code AVX512VL and AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDBPSADBW_YMM_K1Z_YMM_YMMM256_IMM8 = 3904;
	/**
	 * {@code VDBPSADBW zmm1 {k1}{z}, zmm2, zmm3/m512, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 42 /r ib}<p>
	 * {@code AVX512BW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDBPSADBW_ZMM_K1Z_ZMM_ZMMM512_IMM8 = 3905;
	/**
	 * {@code VSHUFI32X4 ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 43 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSHUFI32X4_YMM_K1Z_YMM_YMMM256B32_IMM8 = 3906;
	/**
	 * {@code VSHUFI32X4 zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 43 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSHUFI32X4_ZMM_K1Z_ZMM_ZMMM512B32_IMM8 = 3907;
	/**
	 * {@code VSHUFI64X2 ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 43 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSHUFI64X2_YMM_K1Z_YMM_YMMM256B64_IMM8 = 3908;
	/**
	 * {@code VSHUFI64X2 zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 43 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSHUFI64X2_ZMM_K1Z_ZMM_ZMMM512B64_IMM8 = 3909;
	/**
	 * {@code PCLMULQDQ xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A 44 /r ib}<p>
	 * {@code PCLMULQDQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCLMULQDQ_XMM_XMMM128_IMM8 = 3910;
	/**
	 * {@code VPCLMULQDQ xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.WIG 44 /r ib}<p>
	 * {@code PCLMULQDQ and AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCLMULQDQ_XMM_XMM_XMMM128_IMM8 = 3911;
	/**
	 * {@code VPCLMULQDQ ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.WIG 44 /r ib}<p>
	 * {@code VPCLMULQDQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCLMULQDQ_YMM_YMM_YMMM256_IMM8 = 3912;
	/**
	 * {@code VPCLMULQDQ xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.128.66.0F3A.WIG 44 /r ib}<p>
	 * {@code AVX512VL and VPCLMULQDQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCLMULQDQ_XMM_XMM_XMMM128_IMM8 = 3913;
	/**
	 * {@code VPCLMULQDQ ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code EVEX.256.66.0F3A.WIG 44 /r ib}<p>
	 * {@code AVX512VL and VPCLMULQDQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCLMULQDQ_YMM_YMM_YMMM256_IMM8 = 3914;
	/**
	 * {@code VPCLMULQDQ zmm1, zmm2, zmm3/m512, imm8}<p>
	 * {@code EVEX.512.66.0F3A.WIG 44 /r ib}<p>
	 * {@code AVX512F and VPCLMULQDQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPCLMULQDQ_ZMM_ZMM_ZMMM512_IMM8 = 3915;
	/**
	 * {@code VPERM2I128 ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.W0 46 /r ib}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERM2I128_YMM_YMM_YMMM256_IMM8 = 3916;
	/**
	 * {@code VPERMIL2PS xmm1, xmm2, xmm3/m128, xmm4, imm4}<p>
	 * {@code VEX.128.66.0F3A.W0 48 /r /is5}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMIL2PS_XMM_XMM_XMMM128_XMM_IMM4 = 3917;
	/**
	 * {@code VPERMIL2PS ymm1, ymm2, ymm3/m256, ymm4, imm4}<p>
	 * {@code VEX.256.66.0F3A.W0 48 /r /is5}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMIL2PS_YMM_YMM_YMMM256_YMM_IMM4 = 3918;
	/**
	 * {@code VPERMIL2PS xmm1, xmm2, xmm3, xmm4/m128, imm4}<p>
	 * {@code VEX.128.66.0F3A.W1 48 /r /is5}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMIL2PS_XMM_XMM_XMM_XMMM128_IMM4 = 3919;
	/**
	 * {@code VPERMIL2PS ymm1, ymm2, ymm3, ymm4/m256, imm4}<p>
	 * {@code VEX.256.66.0F3A.W1 48 /r /is5}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMIL2PS_YMM_YMM_YMM_YMMM256_IMM4 = 3920;
	/**
	 * {@code VPERMIL2PD xmm1, xmm2, xmm3/m128, xmm4, imm4}<p>
	 * {@code VEX.128.66.0F3A.W0 49 /r /is5}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMIL2PD_XMM_XMM_XMMM128_XMM_IMM4 = 3921;
	/**
	 * {@code VPERMIL2PD ymm1, ymm2, ymm3/m256, ymm4, imm4}<p>
	 * {@code VEX.256.66.0F3A.W0 49 /r /is5}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMIL2PD_YMM_YMM_YMMM256_YMM_IMM4 = 3922;
	/**
	 * {@code VPERMIL2PD xmm1, xmm2, xmm3, xmm4/m128, imm4}<p>
	 * {@code VEX.128.66.0F3A.W1 49 /r /is5}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMIL2PD_XMM_XMM_XMM_XMMM128_IMM4 = 3923;
	/**
	 * {@code VPERMIL2PD ymm1, ymm2, ymm3, ymm4/m256, imm4}<p>
	 * {@code VEX.256.66.0F3A.W1 49 /r /is5}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPERMIL2PD_YMM_YMM_YMM_YMMM256_IMM4 = 3924;
	/**
	 * {@code VBLENDVPS xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code VEX.128.66.0F3A.W0 4A /r /is4}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBLENDVPS_XMM_XMM_XMMM128_XMM = 3925;
	/**
	 * {@code VBLENDVPS ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code VEX.256.66.0F3A.W0 4A /r /is4}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBLENDVPS_YMM_YMM_YMMM256_YMM = 3926;
	/**
	 * {@code VBLENDVPD xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code VEX.128.66.0F3A.W0 4B /r /is4}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBLENDVPD_XMM_XMM_XMMM128_XMM = 3927;
	/**
	 * {@code VBLENDVPD ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code VEX.256.66.0F3A.W0 4B /r /is4}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBLENDVPD_YMM_YMM_YMMM256_YMM = 3928;
	/**
	 * {@code VPBLENDVB xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code VEX.128.66.0F3A.W0 4C /r /is4}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPBLENDVB_XMM_XMM_XMMM128_XMM = 3929;
	/**
	 * {@code VPBLENDVB ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code VEX.256.66.0F3A.W0 4C /r /is4}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPBLENDVB_YMM_YMM_YMMM256_YMM = 3930;
	/**
	 * {@code VRANGEPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 50 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRANGEPS_XMM_K1Z_XMM_XMMM128B32_IMM8 = 3931;
	/**
	 * {@code VRANGEPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 50 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRANGEPS_YMM_K1Z_YMM_YMMM256B32_IMM8 = 3932;
	/**
	 * {@code VRANGEPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{sae}, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 50 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRANGEPS_ZMM_K1Z_ZMM_ZMMM512B32_IMM8_SAE = 3933;
	/**
	 * {@code VRANGEPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 50 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRANGEPD_XMM_K1Z_XMM_XMMM128B64_IMM8 = 3934;
	/**
	 * {@code VRANGEPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 50 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRANGEPD_YMM_K1Z_YMM_YMMM256B64_IMM8 = 3935;
	/**
	 * {@code VRANGEPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{sae}, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 50 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRANGEPD_ZMM_K1Z_ZMM_ZMMM512B64_IMM8_SAE = 3936;
	/**
	 * {@code VRANGESS xmm1 {k1}{z}, xmm2, xmm3/m32{sae}, imm8}<p>
	 * {@code EVEX.LIG.66.0F3A.W0 51 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRANGESS_XMM_K1Z_XMM_XMMM32_IMM8_SAE = 3937;
	/**
	 * {@code VRANGESD xmm1 {k1}{z}, xmm2, xmm3/m64{sae}, imm8}<p>
	 * {@code EVEX.LIG.66.0F3A.W1 51 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRANGESD_XMM_K1Z_XMM_XMMM64_IMM8_SAE = 3938;
	/**
	 * {@code VFIXUPIMMPS xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 54 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFIXUPIMMPS_XMM_K1Z_XMM_XMMM128B32_IMM8 = 3939;
	/**
	 * {@code VFIXUPIMMPS ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 54 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFIXUPIMMPS_YMM_K1Z_YMM_YMMM256B32_IMM8 = 3940;
	/**
	 * {@code VFIXUPIMMPS zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{sae}, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 54 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFIXUPIMMPS_ZMM_K1Z_ZMM_ZMMM512B32_IMM8_SAE = 3941;
	/**
	 * {@code VFIXUPIMMPD xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 54 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFIXUPIMMPD_XMM_K1Z_XMM_XMMM128B64_IMM8 = 3942;
	/**
	 * {@code VFIXUPIMMPD ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 54 /r ib}<p>
	 * {@code AVX512VL and AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFIXUPIMMPD_YMM_K1Z_YMM_YMMM256B64_IMM8 = 3943;
	/**
	 * {@code VFIXUPIMMPD zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst{sae}, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 54 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFIXUPIMMPD_ZMM_K1Z_ZMM_ZMMM512B64_IMM8_SAE = 3944;
	/**
	 * {@code VFIXUPIMMSS xmm1 {k1}{z}, xmm2, xmm3/m32{sae}, imm8}<p>
	 * {@code EVEX.LIG.66.0F3A.W0 55 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFIXUPIMMSS_XMM_K1Z_XMM_XMMM32_IMM8_SAE = 3945;
	/**
	 * {@code VFIXUPIMMSD xmm1 {k1}{z}, xmm2, xmm3/m64{sae}, imm8}<p>
	 * {@code EVEX.LIG.66.0F3A.W1 55 /r ib}<p>
	 * {@code AVX512F}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFIXUPIMMSD_XMM_K1Z_XMM_XMMM64_IMM8_SAE = 3946;
	/**
	 * {@code VREDUCEPS xmm1 {k1}{z}, xmm2/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 56 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VREDUCEPS_XMM_K1Z_XMMM128B32_IMM8 = 3947;
	/**
	 * {@code VREDUCEPS ymm1 {k1}{z}, ymm2/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 56 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VREDUCEPS_YMM_K1Z_YMMM256B32_IMM8 = 3948;
	/**
	 * {@code VREDUCEPS zmm1 {k1}{z}, zmm2/m512/m32bcst{sae}, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 56 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VREDUCEPS_ZMM_K1Z_ZMMM512B32_IMM8_SAE = 3949;
	/**
	 * {@code VREDUCEPD xmm1 {k1}{z}, xmm2/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 56 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VREDUCEPD_XMM_K1Z_XMMM128B64_IMM8 = 3950;
	/**
	 * {@code VREDUCEPD ymm1 {k1}{z}, ymm2/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 56 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VREDUCEPD_YMM_K1Z_YMMM256B64_IMM8 = 3951;
	/**
	 * {@code VREDUCEPD zmm1 {k1}{z}, zmm2/m512/m64bcst{sae}, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 56 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VREDUCEPD_ZMM_K1Z_ZMMM512B64_IMM8_SAE = 3952;
	/**
	 * {@code VREDUCESS xmm1 {k1}{z}, xmm2, xmm3/m32{sae}, imm8}<p>
	 * {@code EVEX.LIG.66.0F3A.W0 57 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VREDUCESS_XMM_K1Z_XMM_XMMM32_IMM8_SAE = 3953;
	/**
	 * {@code VREDUCESD xmm1 {k1}{z}, xmm2, xmm3/m64{sae}, imm8}<p>
	 * {@code EVEX.LIG.66.0F3A.W1 57 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VREDUCESD_XMM_K1Z_XMM_XMMM64_IMM8_SAE = 3954;
	/**
	 * {@code VFMADDSUBPS xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code VEX.128.66.0F3A.W0 5C /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUBPS_XMM_XMM_XMMM128_XMM = 3955;
	/**
	 * {@code VFMADDSUBPS ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code VEX.256.66.0F3A.W0 5C /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUBPS_YMM_YMM_YMMM256_YMM = 3956;
	/**
	 * {@code VFMADDSUBPS xmm1, xmm2, xmm3, xmm4/m128}<p>
	 * {@code VEX.128.66.0F3A.W1 5C /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUBPS_XMM_XMM_XMM_XMMM128 = 3957;
	/**
	 * {@code VFMADDSUBPS ymm1, ymm2, ymm3, ymm4/m256}<p>
	 * {@code VEX.256.66.0F3A.W1 5C /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUBPS_YMM_YMM_YMM_YMMM256 = 3958;
	/**
	 * {@code VFMADDSUBPD xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code VEX.128.66.0F3A.W0 5D /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUBPD_XMM_XMM_XMMM128_XMM = 3959;
	/**
	 * {@code VFMADDSUBPD ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code VEX.256.66.0F3A.W0 5D /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUBPD_YMM_YMM_YMMM256_YMM = 3960;
	/**
	 * {@code VFMADDSUBPD xmm1, xmm2, xmm3, xmm4/m128}<p>
	 * {@code VEX.128.66.0F3A.W1 5D /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUBPD_XMM_XMM_XMM_XMMM128 = 3961;
	/**
	 * {@code VFMADDSUBPD ymm1, ymm2, ymm3, ymm4/m256}<p>
	 * {@code VEX.256.66.0F3A.W1 5D /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSUBPD_YMM_YMM_YMM_YMMM256 = 3962;
	/**
	 * {@code VFMSUBADDPS xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code VEX.128.66.0F3A.W0 5E /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADDPS_XMM_XMM_XMMM128_XMM = 3963;
	/**
	 * {@code VFMSUBADDPS ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code VEX.256.66.0F3A.W0 5E /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADDPS_YMM_YMM_YMMM256_YMM = 3964;
	/**
	 * {@code VFMSUBADDPS xmm1, xmm2, xmm3, xmm4/m128}<p>
	 * {@code VEX.128.66.0F3A.W1 5E /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADDPS_XMM_XMM_XMM_XMMM128 = 3965;
	/**
	 * {@code VFMSUBADDPS ymm1, ymm2, ymm3, ymm4/m256}<p>
	 * {@code VEX.256.66.0F3A.W1 5E /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADDPS_YMM_YMM_YMM_YMMM256 = 3966;
	/**
	 * {@code VFMSUBADDPD xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code VEX.128.66.0F3A.W0 5F /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADDPD_XMM_XMM_XMMM128_XMM = 3967;
	/**
	 * {@code VFMSUBADDPD ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code VEX.256.66.0F3A.W0 5F /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADDPD_YMM_YMM_YMMM256_YMM = 3968;
	/**
	 * {@code VFMSUBADDPD xmm1, xmm2, xmm3, xmm4/m128}<p>
	 * {@code VEX.128.66.0F3A.W1 5F /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADDPD_XMM_XMM_XMM_XMMM128 = 3969;
	/**
	 * {@code VFMSUBADDPD ymm1, ymm2, ymm3, ymm4/m256}<p>
	 * {@code VEX.256.66.0F3A.W1 5F /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBADDPD_YMM_YMM_YMM_YMMM256 = 3970;
	/**
	 * {@code PCMPESTRM xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A 60 /r ib}<p>
	 * {@code SSE4.2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPESTRM_XMM_XMMM128_IMM8 = 3971;
	/**
	 * {@code PCMPESTRM64 xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 o64 0F 3A 60 /r ib}<p>
	 * {@code SSE4.2}<p>
	 * {@code 64-bit}
	 */
	public static final int PCMPESTRM64_XMM_XMMM128_IMM8 = 3972;
	/**
	 * {@code VPCMPESTRM xmm1, xmm2/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.W0 60 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPESTRM_XMM_XMMM128_IMM8 = 3973;
	/**
	 * {@code VPCMPESTRM64 xmm1, xmm2/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.W1 60 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VPCMPESTRM64_XMM_XMMM128_IMM8 = 3974;
	/**
	 * {@code PCMPESTRI xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A 61 /r ib}<p>
	 * {@code SSE4.2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPESTRI_XMM_XMMM128_IMM8 = 3975;
	/**
	 * {@code PCMPESTRI64 xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 o64 0F 3A 61 /r ib}<p>
	 * {@code SSE4.2}<p>
	 * {@code 64-bit}
	 */
	public static final int PCMPESTRI64_XMM_XMMM128_IMM8 = 3976;
	/**
	 * {@code VPCMPESTRI xmm1, xmm2/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.W0 61 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPESTRI_XMM_XMMM128_IMM8 = 3977;
	/**
	 * {@code VPCMPESTRI64 xmm1, xmm2/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.W1 61 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_VPCMPESTRI64_XMM_XMMM128_IMM8 = 3978;
	/**
	 * {@code PCMPISTRM xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A 62 /r ib}<p>
	 * {@code SSE4.2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPISTRM_XMM_XMMM128_IMM8 = 3979;
	/**
	 * {@code VPCMPISTRM xmm1, xmm2/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.WIG 62 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPISTRM_XMM_XMMM128_IMM8 = 3980;
	/**
	 * {@code PCMPISTRI xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A 63 /r ib}<p>
	 * {@code SSE4.2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PCMPISTRI_XMM_XMMM128_IMM8 = 3981;
	/**
	 * {@code VPCMPISTRI xmm1, xmm2/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.WIG 63 /r ib}<p>
	 * {@code AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPCMPISTRI_XMM_XMMM128_IMM8 = 3982;
	/**
	 * {@code VFPCLASSPS k2 {k1}, xmm2/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 66 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFPCLASSPS_KR_K1_XMMM128B32_IMM8 = 3983;
	/**
	 * {@code VFPCLASSPS k2 {k1}, ymm2/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 66 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFPCLASSPS_KR_K1_YMMM256B32_IMM8 = 3984;
	/**
	 * {@code VFPCLASSPS k2 {k1}, zmm2/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 66 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFPCLASSPS_KR_K1_ZMMM512B32_IMM8 = 3985;
	/**
	 * {@code VFPCLASSPD k2 {k1}, xmm2/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 66 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFPCLASSPD_KR_K1_XMMM128B64_IMM8 = 3986;
	/**
	 * {@code VFPCLASSPD k2 {k1}, ymm2/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 66 /r ib}<p>
	 * {@code AVX512VL and AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFPCLASSPD_KR_K1_YMMM256B64_IMM8 = 3987;
	/**
	 * {@code VFPCLASSPD k2 {k1}, zmm2/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 66 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFPCLASSPD_KR_K1_ZMMM512B64_IMM8 = 3988;
	/**
	 * {@code VFPCLASSSS k2 {k1}, xmm2/m32, imm8}<p>
	 * {@code EVEX.LIG.66.0F3A.W0 67 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFPCLASSSS_KR_K1_XMMM32_IMM8 = 3989;
	/**
	 * {@code VFPCLASSSD k2 {k1}, xmm2/m64, imm8}<p>
	 * {@code EVEX.LIG.66.0F3A.W1 67 /r ib}<p>
	 * {@code AVX512DQ}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFPCLASSSD_KR_K1_XMMM64_IMM8 = 3990;
	/**
	 * {@code VFMADDPS xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code VEX.128.66.0F3A.W0 68 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDPS_XMM_XMM_XMMM128_XMM = 3991;
	/**
	 * {@code VFMADDPS ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code VEX.256.66.0F3A.W0 68 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDPS_YMM_YMM_YMMM256_YMM = 3992;
	/**
	 * {@code VFMADDPS xmm1, xmm2, xmm3, xmm4/m128}<p>
	 * {@code VEX.128.66.0F3A.W1 68 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDPS_XMM_XMM_XMM_XMMM128 = 3993;
	/**
	 * {@code VFMADDPS ymm1, ymm2, ymm3, ymm4/m256}<p>
	 * {@code VEX.256.66.0F3A.W1 68 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDPS_YMM_YMM_YMM_YMMM256 = 3994;
	/**
	 * {@code VFMADDPD xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code VEX.128.66.0F3A.W0 69 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDPD_XMM_XMM_XMMM128_XMM = 3995;
	/**
	 * {@code VFMADDPD ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code VEX.256.66.0F3A.W0 69 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDPD_YMM_YMM_YMMM256_YMM = 3996;
	/**
	 * {@code VFMADDPD xmm1, xmm2, xmm3, xmm4/m128}<p>
	 * {@code VEX.128.66.0F3A.W1 69 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDPD_XMM_XMM_XMM_XMMM128 = 3997;
	/**
	 * {@code VFMADDPD ymm1, ymm2, ymm3, ymm4/m256}<p>
	 * {@code VEX.256.66.0F3A.W1 69 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDPD_YMM_YMM_YMM_YMMM256 = 3998;
	/**
	 * {@code VFMADDSS xmm1, xmm2, xmm3/m32, xmm4}<p>
	 * {@code VEX.LIG.66.0F3A.W0 6A /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSS_XMM_XMM_XMMM32_XMM = 3999;
	/**
	 * {@code VFMADDSS xmm1, xmm2, xmm3, xmm4/m32}<p>
	 * {@code VEX.LIG.66.0F3A.W1 6A /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSS_XMM_XMM_XMM_XMMM32 = 4000;
	/**
	 * {@code VFMADDSD xmm1, xmm2, xmm3/m64, xmm4}<p>
	 * {@code VEX.LIG.66.0F3A.W0 6B /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSD_XMM_XMM_XMMM64_XMM = 4001;
	/**
	 * {@code VFMADDSD xmm1, xmm2, xmm3, xmm4/m64}<p>
	 * {@code VEX.LIG.66.0F3A.W1 6B /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMADDSD_XMM_XMM_XMM_XMMM64 = 4002;
	/**
	 * {@code VFMSUBPS xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code VEX.128.66.0F3A.W0 6C /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBPS_XMM_XMM_XMMM128_XMM = 4003;
	/**
	 * {@code VFMSUBPS ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code VEX.256.66.0F3A.W0 6C /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBPS_YMM_YMM_YMMM256_YMM = 4004;
	/**
	 * {@code VFMSUBPS xmm1, xmm2, xmm3, xmm4/m128}<p>
	 * {@code VEX.128.66.0F3A.W1 6C /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBPS_XMM_XMM_XMM_XMMM128 = 4005;
	/**
	 * {@code VFMSUBPS ymm1, ymm2, ymm3, ymm4/m256}<p>
	 * {@code VEX.256.66.0F3A.W1 6C /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBPS_YMM_YMM_YMM_YMMM256 = 4006;
	/**
	 * {@code VFMSUBPD xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code VEX.128.66.0F3A.W0 6D /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBPD_XMM_XMM_XMMM128_XMM = 4007;
	/**
	 * {@code VFMSUBPD ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code VEX.256.66.0F3A.W0 6D /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBPD_YMM_YMM_YMMM256_YMM = 4008;
	/**
	 * {@code VFMSUBPD xmm1, xmm2, xmm3, xmm4/m128}<p>
	 * {@code VEX.128.66.0F3A.W1 6D /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBPD_XMM_XMM_XMM_XMMM128 = 4009;
	/**
	 * {@code VFMSUBPD ymm1, ymm2, ymm3, ymm4/m256}<p>
	 * {@code VEX.256.66.0F3A.W1 6D /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBPD_YMM_YMM_YMM_YMMM256 = 4010;
	/**
	 * {@code VFMSUBSS xmm1, xmm2, xmm3/m32, xmm4}<p>
	 * {@code VEX.LIG.66.0F3A.W0 6E /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBSS_XMM_XMM_XMMM32_XMM = 4011;
	/**
	 * {@code VFMSUBSS xmm1, xmm2, xmm3, xmm4/m32}<p>
	 * {@code VEX.LIG.66.0F3A.W1 6E /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBSS_XMM_XMM_XMM_XMMM32 = 4012;
	/**
	 * {@code VFMSUBSD xmm1, xmm2, xmm3/m64, xmm4}<p>
	 * {@code VEX.LIG.66.0F3A.W0 6F /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBSD_XMM_XMM_XMMM64_XMM = 4013;
	/**
	 * {@code VFMSUBSD xmm1, xmm2, xmm3, xmm4/m64}<p>
	 * {@code VEX.LIG.66.0F3A.W1 6F /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFMSUBSD_XMM_XMM_XMM_XMMM64 = 4014;
	/**
	 * {@code VPSHLDW xmm1 {k1}{z}, xmm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 70 /r ib}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDW_XMM_K1Z_XMM_XMMM128_IMM8 = 4015;
	/**
	 * {@code VPSHLDW ymm1 {k1}{z}, ymm2, ymm3/m256, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 70 /r ib}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDW_YMM_K1Z_YMM_YMMM256_IMM8 = 4016;
	/**
	 * {@code VPSHLDW zmm1 {k1}{z}, zmm2, zmm3/m512, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 70 /r ib}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDW_ZMM_K1Z_ZMM_ZMMM512_IMM8 = 4017;
	/**
	 * {@code VPSHLDD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 71 /r ib}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDD_XMM_K1Z_XMM_XMMM128B32_IMM8 = 4018;
	/**
	 * {@code VPSHLDD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 71 /r ib}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDD_YMM_K1Z_YMM_YMMM256B32_IMM8 = 4019;
	/**
	 * {@code VPSHLDD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 71 /r ib}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDD_ZMM_K1Z_ZMM_ZMMM512B32_IMM8 = 4020;
	/**
	 * {@code VPSHLDQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 71 /r ib}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDQ_XMM_K1Z_XMM_XMMM128B64_IMM8 = 4021;
	/**
	 * {@code VPSHLDQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 71 /r ib}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDQ_YMM_K1Z_YMM_YMMM256B64_IMM8 = 4022;
	/**
	 * {@code VPSHLDQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 71 /r ib}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHLDQ_ZMM_K1Z_ZMM_ZMMM512B64_IMM8 = 4023;
	/**
	 * {@code VPSHRDW xmm1 {k1}{z}, xmm2, xmm3/m128, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 72 /r ib}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDW_XMM_K1Z_XMM_XMMM128_IMM8 = 4024;
	/**
	 * {@code VPSHRDW ymm1 {k1}{z}, ymm2, ymm3/m256, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 72 /r ib}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDW_YMM_K1Z_YMM_YMMM256_IMM8 = 4025;
	/**
	 * {@code VPSHRDW zmm1 {k1}{z}, zmm2, zmm3/m512, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 72 /r ib}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDW_ZMM_K1Z_ZMM_ZMMM512_IMM8 = 4026;
	/**
	 * {@code VPSHRDD xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W0 73 /r ib}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDD_XMM_K1Z_XMM_XMMM128B32_IMM8 = 4027;
	/**
	 * {@code VPSHRDD ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W0 73 /r ib}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDD_YMM_K1Z_YMM_YMMM256B32_IMM8 = 4028;
	/**
	 * {@code VPSHRDD zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W0 73 /r ib}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDD_ZMM_K1Z_ZMM_ZMMM512B32_IMM8 = 4029;
	/**
	 * {@code VPSHRDQ xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 73 /r ib}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDQ_XMM_K1Z_XMM_XMMM128B64_IMM8 = 4030;
	/**
	 * {@code VPSHRDQ ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 73 /r ib}<p>
	 * {@code AVX512VL and AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDQ_YMM_K1Z_YMM_YMMM256B64_IMM8 = 4031;
	/**
	 * {@code VPSHRDQ zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 73 /r ib}<p>
	 * {@code AVX512_VBMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VPSHRDQ_ZMM_K1Z_ZMM_ZMMM512B64_IMM8 = 4032;
	/**
	 * {@code VFNMADDPS xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code VEX.128.66.0F3A.W0 78 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADDPS_XMM_XMM_XMMM128_XMM = 4033;
	/**
	 * {@code VFNMADDPS ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code VEX.256.66.0F3A.W0 78 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADDPS_YMM_YMM_YMMM256_YMM = 4034;
	/**
	 * {@code VFNMADDPS xmm1, xmm2, xmm3, xmm4/m128}<p>
	 * {@code VEX.128.66.0F3A.W1 78 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADDPS_XMM_XMM_XMM_XMMM128 = 4035;
	/**
	 * {@code VFNMADDPS ymm1, ymm2, ymm3, ymm4/m256}<p>
	 * {@code VEX.256.66.0F3A.W1 78 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADDPS_YMM_YMM_YMM_YMMM256 = 4036;
	/**
	 * {@code VFNMADDPD xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code VEX.128.66.0F3A.W0 79 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADDPD_XMM_XMM_XMMM128_XMM = 4037;
	/**
	 * {@code VFNMADDPD ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code VEX.256.66.0F3A.W0 79 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADDPD_YMM_YMM_YMMM256_YMM = 4038;
	/**
	 * {@code VFNMADDPD xmm1, xmm2, xmm3, xmm4/m128}<p>
	 * {@code VEX.128.66.0F3A.W1 79 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADDPD_XMM_XMM_XMM_XMMM128 = 4039;
	/**
	 * {@code VFNMADDPD ymm1, ymm2, ymm3, ymm4/m256}<p>
	 * {@code VEX.256.66.0F3A.W1 79 /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADDPD_YMM_YMM_YMM_YMMM256 = 4040;
	/**
	 * {@code VFNMADDSS xmm1, xmm2, xmm3/m32, xmm4}<p>
	 * {@code VEX.LIG.66.0F3A.W0 7A /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADDSS_XMM_XMM_XMMM32_XMM = 4041;
	/**
	 * {@code VFNMADDSS xmm1, xmm2, xmm3, xmm4/m32}<p>
	 * {@code VEX.LIG.66.0F3A.W1 7A /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADDSS_XMM_XMM_XMM_XMMM32 = 4042;
	/**
	 * {@code VFNMADDSD xmm1, xmm2, xmm3/m64, xmm4}<p>
	 * {@code VEX.LIG.66.0F3A.W0 7B /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADDSD_XMM_XMM_XMMM64_XMM = 4043;
	/**
	 * {@code VFNMADDSD xmm1, xmm2, xmm3, xmm4/m64}<p>
	 * {@code VEX.LIG.66.0F3A.W1 7B /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMADDSD_XMM_XMM_XMM_XMMM64 = 4044;
	/**
	 * {@code VFNMSUBPS xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code VEX.128.66.0F3A.W0 7C /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUBPS_XMM_XMM_XMMM128_XMM = 4045;
	/**
	 * {@code VFNMSUBPS ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code VEX.256.66.0F3A.W0 7C /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUBPS_YMM_YMM_YMMM256_YMM = 4046;
	/**
	 * {@code VFNMSUBPS xmm1, xmm2, xmm3, xmm4/m128}<p>
	 * {@code VEX.128.66.0F3A.W1 7C /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUBPS_XMM_XMM_XMM_XMMM128 = 4047;
	/**
	 * {@code VFNMSUBPS ymm1, ymm2, ymm3, ymm4/m256}<p>
	 * {@code VEX.256.66.0F3A.W1 7C /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUBPS_YMM_YMM_YMM_YMMM256 = 4048;
	/**
	 * {@code VFNMSUBPD xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code VEX.128.66.0F3A.W0 7D /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUBPD_XMM_XMM_XMMM128_XMM = 4049;
	/**
	 * {@code VFNMSUBPD ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code VEX.256.66.0F3A.W0 7D /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUBPD_YMM_YMM_YMMM256_YMM = 4050;
	/**
	 * {@code VFNMSUBPD xmm1, xmm2, xmm3, xmm4/m128}<p>
	 * {@code VEX.128.66.0F3A.W1 7D /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUBPD_XMM_XMM_XMM_XMMM128 = 4051;
	/**
	 * {@code VFNMSUBPD ymm1, ymm2, ymm3, ymm4/m256}<p>
	 * {@code VEX.256.66.0F3A.W1 7D /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUBPD_YMM_YMM_YMM_YMMM256 = 4052;
	/**
	 * {@code VFNMSUBSS xmm1, xmm2, xmm3/m32, xmm4}<p>
	 * {@code VEX.LIG.66.0F3A.W0 7E /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUBSS_XMM_XMM_XMMM32_XMM = 4053;
	/**
	 * {@code VFNMSUBSS xmm1, xmm2, xmm3, xmm4/m32}<p>
	 * {@code VEX.LIG.66.0F3A.W1 7E /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUBSS_XMM_XMM_XMM_XMMM32 = 4054;
	/**
	 * {@code VFNMSUBSD xmm1, xmm2, xmm3/m64, xmm4}<p>
	 * {@code VEX.LIG.66.0F3A.W0 7F /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUBSD_XMM_XMM_XMMM64_XMM = 4055;
	/**
	 * {@code VFNMSUBSD xmm1, xmm2, xmm3, xmm4/m64}<p>
	 * {@code VEX.LIG.66.0F3A.W1 7F /r /is4}<p>
	 * {@code FMA4}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VFNMSUBSD_XMM_XMM_XMM_XMMM64 = 4056;
	/**
	 * {@code SHA1RNDS4 xmm1, xmm2/m128, imm8}<p>
	 * {@code NP 0F 3A CC /r ib}<p>
	 * {@code SHA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SHA1RNDS4_XMM_XMMM128_IMM8 = 4057;
	/**
	 * {@code GF2P8AFFINEQB xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A CE /r ib}<p>
	 * {@code GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int GF2P8AFFINEQB_XMM_XMMM128_IMM8 = 4058;
	/**
	 * {@code VGF2P8AFFINEQB xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.W1 CE /r ib}<p>
	 * {@code AVX and GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VGF2P8AFFINEQB_XMM_XMM_XMMM128_IMM8 = 4059;
	/**
	 * {@code VGF2P8AFFINEQB ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.W1 CE /r ib}<p>
	 * {@code AVX and GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VGF2P8AFFINEQB_YMM_YMM_YMMM256_IMM8 = 4060;
	/**
	 * {@code VGF2P8AFFINEQB xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 CE /r ib}<p>
	 * {@code AVX512VL and GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGF2P8AFFINEQB_XMM_K1Z_XMM_XMMM128B64_IMM8 = 4061;
	/**
	 * {@code VGF2P8AFFINEQB ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 CE /r ib}<p>
	 * {@code AVX512VL and GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGF2P8AFFINEQB_YMM_K1Z_YMM_YMMM256B64_IMM8 = 4062;
	/**
	 * {@code VGF2P8AFFINEQB zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 CE /r ib}<p>
	 * {@code AVX512F and GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGF2P8AFFINEQB_ZMM_K1Z_ZMM_ZMMM512B64_IMM8 = 4063;
	/**
	 * {@code GF2P8AFFINEINVQB xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A CF /r ib}<p>
	 * {@code GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int GF2P8AFFINEINVQB_XMM_XMMM128_IMM8 = 4064;
	/**
	 * {@code VGF2P8AFFINEINVQB xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.W1 CF /r ib}<p>
	 * {@code AVX and GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VGF2P8AFFINEINVQB_XMM_XMM_XMMM128_IMM8 = 4065;
	/**
	 * {@code VGF2P8AFFINEINVQB ymm1, ymm2, ymm3/m256, imm8}<p>
	 * {@code VEX.256.66.0F3A.W1 CF /r ib}<p>
	 * {@code AVX and GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VGF2P8AFFINEINVQB_YMM_YMM_YMMM256_IMM8 = 4066;
	/**
	 * {@code VGF2P8AFFINEINVQB xmm1 {k1}{z}, xmm2, xmm3/m128/m64bcst, imm8}<p>
	 * {@code EVEX.128.66.0F3A.W1 CF /r ib}<p>
	 * {@code AVX512VL and GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGF2P8AFFINEINVQB_XMM_K1Z_XMM_XMMM128B64_IMM8 = 4067;
	/**
	 * {@code VGF2P8AFFINEINVQB ymm1 {k1}{z}, ymm2, ymm3/m256/m64bcst, imm8}<p>
	 * {@code EVEX.256.66.0F3A.W1 CF /r ib}<p>
	 * {@code AVX512VL and GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGF2P8AFFINEINVQB_YMM_K1Z_YMM_YMMM256B64_IMM8 = 4068;
	/**
	 * {@code VGF2P8AFFINEINVQB zmm1 {k1}{z}, zmm2, zmm3/m512/m64bcst, imm8}<p>
	 * {@code EVEX.512.66.0F3A.W1 CF /r ib}<p>
	 * {@code AVX512F and GFNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGF2P8AFFINEINVQB_ZMM_K1Z_ZMM_ZMMM512B64_IMM8 = 4069;
	/**
	 * {@code AESKEYGENASSIST xmm1, xmm2/m128, imm8}<p>
	 * {@code 66 0F 3A DF /r ib}<p>
	 * {@code AES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AESKEYGENASSIST_XMM_XMMM128_IMM8 = 4070;
	/**
	 * {@code VAESKEYGENASSIST xmm1, xmm2/m128, imm8}<p>
	 * {@code VEX.128.66.0F3A.WIG DF /r ib}<p>
	 * {@code AES and AVX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VAESKEYGENASSIST_XMM_XMMM128_IMM8 = 4071;
	/**
	 * {@code RORX r32, r/m32, imm8}<p>
	 * {@code VEX.LZ.F2.0F3A.W0 F0 /r ib}<p>
	 * {@code BMI2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_RORX_R32_RM32_IMM8 = 4072;
	/**
	 * {@code RORX r64, r/m64, imm8}<p>
	 * {@code VEX.LZ.F2.0F3A.W1 F0 /r ib}<p>
	 * {@code BMI2}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_RORX_R64_RM64_IMM8 = 4073;
	/**
	 * {@code VPMACSSWW xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code XOP.128.X8.W0 85 /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPMACSSWW_XMM_XMM_XMMM128_XMM = 4074;
	/**
	 * {@code VPMACSSWD xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code XOP.128.X8.W0 86 /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPMACSSWD_XMM_XMM_XMMM128_XMM = 4075;
	/**
	 * {@code VPMACSSDQL xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code XOP.128.X8.W0 87 /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPMACSSDQL_XMM_XMM_XMMM128_XMM = 4076;
	/**
	 * {@code VPMACSSDD xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code XOP.128.X8.W0 8E /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPMACSSDD_XMM_XMM_XMMM128_XMM = 4077;
	/**
	 * {@code VPMACSSDQH xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code XOP.128.X8.W0 8F /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPMACSSDQH_XMM_XMM_XMMM128_XMM = 4078;
	/**
	 * {@code VPMACSWW xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code XOP.128.X8.W0 95 /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPMACSWW_XMM_XMM_XMMM128_XMM = 4079;
	/**
	 * {@code VPMACSWD xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code XOP.128.X8.W0 96 /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPMACSWD_XMM_XMM_XMMM128_XMM = 4080;
	/**
	 * {@code VPMACSDQL xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code XOP.128.X8.W0 97 /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPMACSDQL_XMM_XMM_XMMM128_XMM = 4081;
	/**
	 * {@code VPMACSDD xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code XOP.128.X8.W0 9E /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPMACSDD_XMM_XMM_XMMM128_XMM = 4082;
	/**
	 * {@code VPMACSDQH xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code XOP.128.X8.W0 9F /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPMACSDQH_XMM_XMM_XMMM128_XMM = 4083;
	/**
	 * {@code VPCMOV xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code XOP.128.X8.W0 A2 /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPCMOV_XMM_XMM_XMMM128_XMM = 4084;
	/**
	 * {@code VPCMOV ymm1, ymm2, ymm3/m256, ymm4}<p>
	 * {@code XOP.256.X8.W0 A2 /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPCMOV_YMM_YMM_YMMM256_YMM = 4085;
	/**
	 * {@code VPCMOV xmm1, xmm2, xmm3, xmm4/m128}<p>
	 * {@code XOP.128.X8.W1 A2 /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPCMOV_XMM_XMM_XMM_XMMM128 = 4086;
	/**
	 * {@code VPCMOV ymm1, ymm2, ymm3, ymm4/m256}<p>
	 * {@code XOP.256.X8.W1 A2 /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPCMOV_YMM_YMM_YMM_YMMM256 = 4087;
	/**
	 * {@code VPPERM xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code XOP.128.X8.W0 A3 /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPPERM_XMM_XMM_XMMM128_XMM = 4088;
	/**
	 * {@code VPPERM xmm1, xmm2, xmm3, xmm4/m128}<p>
	 * {@code XOP.128.X8.W1 A3 /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPPERM_XMM_XMM_XMM_XMMM128 = 4089;
	/**
	 * {@code VPMADCSSWD xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code XOP.128.X8.W0 A6 /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPMADCSSWD_XMM_XMM_XMMM128_XMM = 4090;
	/**
	 * {@code VPMADCSWD xmm1, xmm2, xmm3/m128, xmm4}<p>
	 * {@code XOP.128.X8.W0 B6 /r /is4}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPMADCSWD_XMM_XMM_XMMM128_XMM = 4091;
	/**
	 * {@code VPROTB xmm1, xmm2/m128, imm8}<p>
	 * {@code XOP.128.X8.W0 C0 /r ib}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPROTB_XMM_XMMM128_IMM8 = 4092;
	/**
	 * {@code VPROTW xmm1, xmm2/m128, imm8}<p>
	 * {@code XOP.128.X8.W0 C1 /r ib}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPROTW_XMM_XMMM128_IMM8 = 4093;
	/**
	 * {@code VPROTD xmm1, xmm2/m128, imm8}<p>
	 * {@code XOP.128.X8.W0 C2 /r ib}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPROTD_XMM_XMMM128_IMM8 = 4094;
	/**
	 * {@code VPROTQ xmm1, xmm2/m128, imm8}<p>
	 * {@code XOP.128.X8.W0 C3 /r ib}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPROTQ_XMM_XMMM128_IMM8 = 4095;
	/**
	 * {@code VPCOMB xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code XOP.128.X8.W0 CC /r ib}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPCOMB_XMM_XMM_XMMM128_IMM8 = 4096;
	/**
	 * {@code VPCOMW xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code XOP.128.X8.W0 CD /r ib}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPCOMW_XMM_XMM_XMMM128_IMM8 = 4097;
	/**
	 * {@code VPCOMD xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code XOP.128.X8.W0 CE /r ib}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPCOMD_XMM_XMM_XMMM128_IMM8 = 4098;
	/**
	 * {@code VPCOMQ xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code XOP.128.X8.W0 CF /r ib}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPCOMQ_XMM_XMM_XMMM128_IMM8 = 4099;
	/**
	 * {@code VPCOMUB xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code XOP.128.X8.W0 EC /r ib}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPCOMUB_XMM_XMM_XMMM128_IMM8 = 4100;
	/**
	 * {@code VPCOMUW xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code XOP.128.X8.W0 ED /r ib}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPCOMUW_XMM_XMM_XMMM128_IMM8 = 4101;
	/**
	 * {@code VPCOMUD xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code XOP.128.X8.W0 EE /r ib}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPCOMUD_XMM_XMM_XMMM128_IMM8 = 4102;
	/**
	 * {@code VPCOMUQ xmm1, xmm2, xmm3/m128, imm8}<p>
	 * {@code XOP.128.X8.W0 EF /r ib}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPCOMUQ_XMM_XMM_XMMM128_IMM8 = 4103;
	/**
	 * {@code BLCFILL r32, r/m32}<p>
	 * {@code XOP.L0.X9.W0 01 /1}<p>
	 * {@code TBM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_BLCFILL_R32_RM32 = 4104;
	/**
	 * {@code BLCFILL r64, r/m64}<p>
	 * {@code XOP.L0.X9.W1 01 /1}<p>
	 * {@code TBM}<p>
	 * {@code 64-bit}
	 */
	public static final int XOP_BLCFILL_R64_RM64 = 4105;
	/**
	 * {@code BLSFILL r32, r/m32}<p>
	 * {@code XOP.L0.X9.W0 01 /2}<p>
	 * {@code TBM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_BLSFILL_R32_RM32 = 4106;
	/**
	 * {@code BLSFILL r64, r/m64}<p>
	 * {@code XOP.L0.X9.W1 01 /2}<p>
	 * {@code TBM}<p>
	 * {@code 64-bit}
	 */
	public static final int XOP_BLSFILL_R64_RM64 = 4107;
	/**
	 * {@code BLCS r32, r/m32}<p>
	 * {@code XOP.L0.X9.W0 01 /3}<p>
	 * {@code TBM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_BLCS_R32_RM32 = 4108;
	/**
	 * {@code BLCS r64, r/m64}<p>
	 * {@code XOP.L0.X9.W1 01 /3}<p>
	 * {@code TBM}<p>
	 * {@code 64-bit}
	 */
	public static final int XOP_BLCS_R64_RM64 = 4109;
	/**
	 * {@code TZMSK r32, r/m32}<p>
	 * {@code XOP.L0.X9.W0 01 /4}<p>
	 * {@code TBM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_TZMSK_R32_RM32 = 4110;
	/**
	 * {@code TZMSK r64, r/m64}<p>
	 * {@code XOP.L0.X9.W1 01 /4}<p>
	 * {@code TBM}<p>
	 * {@code 64-bit}
	 */
	public static final int XOP_TZMSK_R64_RM64 = 4111;
	/**
	 * {@code BLCIC r32, r/m32}<p>
	 * {@code XOP.L0.X9.W0 01 /5}<p>
	 * {@code TBM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_BLCIC_R32_RM32 = 4112;
	/**
	 * {@code BLCIC r64, r/m64}<p>
	 * {@code XOP.L0.X9.W1 01 /5}<p>
	 * {@code TBM}<p>
	 * {@code 64-bit}
	 */
	public static final int XOP_BLCIC_R64_RM64 = 4113;
	/**
	 * {@code BLSIC r32, r/m32}<p>
	 * {@code XOP.L0.X9.W0 01 /6}<p>
	 * {@code TBM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_BLSIC_R32_RM32 = 4114;
	/**
	 * {@code BLSIC r64, r/m64}<p>
	 * {@code XOP.L0.X9.W1 01 /6}<p>
	 * {@code TBM}<p>
	 * {@code 64-bit}
	 */
	public static final int XOP_BLSIC_R64_RM64 = 4115;
	/**
	 * {@code T1MSKC r32, r/m32}<p>
	 * {@code XOP.L0.X9.W0 01 /7}<p>
	 * {@code TBM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_T1MSKC_R32_RM32 = 4116;
	/**
	 * {@code T1MSKC r64, r/m64}<p>
	 * {@code XOP.L0.X9.W1 01 /7}<p>
	 * {@code TBM}<p>
	 * {@code 64-bit}
	 */
	public static final int XOP_T1MSKC_R64_RM64 = 4117;
	/**
	 * {@code BLCMSK r32, r/m32}<p>
	 * {@code XOP.L0.X9.W0 02 /1}<p>
	 * {@code TBM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_BLCMSK_R32_RM32 = 4118;
	/**
	 * {@code BLCMSK r64, r/m64}<p>
	 * {@code XOP.L0.X9.W1 02 /1}<p>
	 * {@code TBM}<p>
	 * {@code 64-bit}
	 */
	public static final int XOP_BLCMSK_R64_RM64 = 4119;
	/**
	 * {@code BLCI r32, r/m32}<p>
	 * {@code XOP.L0.X9.W0 02 /6}<p>
	 * {@code TBM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_BLCI_R32_RM32 = 4120;
	/**
	 * {@code BLCI r64, r/m64}<p>
	 * {@code XOP.L0.X9.W1 02 /6}<p>
	 * {@code TBM}<p>
	 * {@code 64-bit}
	 */
	public static final int XOP_BLCI_R64_RM64 = 4121;
	/**
	 * {@code LLWPCB r32}<p>
	 * {@code XOP.L0.X9.W0 12 /0}<p>
	 * {@code LWP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_LLWPCB_R32 = 4122;
	/**
	 * {@code LLWPCB r64}<p>
	 * {@code XOP.L0.X9.W1 12 /0}<p>
	 * {@code LWP}<p>
	 * {@code 64-bit}
	 */
	public static final int XOP_LLWPCB_R64 = 4123;
	/**
	 * {@code SLWPCB r32}<p>
	 * {@code XOP.L0.X9.W0 12 /1}<p>
	 * {@code LWP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_SLWPCB_R32 = 4124;
	/**
	 * {@code SLWPCB r64}<p>
	 * {@code XOP.L0.X9.W1 12 /1}<p>
	 * {@code LWP}<p>
	 * {@code 64-bit}
	 */
	public static final int XOP_SLWPCB_R64 = 4125;
	/**
	 * {@code VFRCZPS xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 80 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VFRCZPS_XMM_XMMM128 = 4126;
	/**
	 * {@code VFRCZPS ymm1, ymm2/m256}<p>
	 * {@code XOP.256.X9.W0 80 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VFRCZPS_YMM_YMMM256 = 4127;
	/**
	 * {@code VFRCZPD xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 81 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VFRCZPD_XMM_XMMM128 = 4128;
	/**
	 * {@code VFRCZPD ymm1, ymm2/m256}<p>
	 * {@code XOP.256.X9.W0 81 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VFRCZPD_YMM_YMMM256 = 4129;
	/**
	 * {@code VFRCZSS xmm1, xmm2/m32}<p>
	 * {@code XOP.128.X9.W0 82 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VFRCZSS_XMM_XMMM32 = 4130;
	/**
	 * {@code VFRCZSD xmm1, xmm2/m64}<p>
	 * {@code XOP.128.X9.W0 83 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VFRCZSD_XMM_XMMM64 = 4131;
	/**
	 * {@code VPROTB xmm1, xmm2/m128, xmm3}<p>
	 * {@code XOP.128.X9.W0 90 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPROTB_XMM_XMMM128_XMM = 4132;
	/**
	 * {@code VPROTB xmm1, xmm2, xmm3/m128}<p>
	 * {@code XOP.128.X9.W1 90 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPROTB_XMM_XMM_XMMM128 = 4133;
	/**
	 * {@code VPROTW xmm1, xmm2/m128, xmm3}<p>
	 * {@code XOP.128.X9.W0 91 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPROTW_XMM_XMMM128_XMM = 4134;
	/**
	 * {@code VPROTW xmm1, xmm2, xmm3/m128}<p>
	 * {@code XOP.128.X9.W1 91 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPROTW_XMM_XMM_XMMM128 = 4135;
	/**
	 * {@code VPROTD xmm1, xmm2/m128, xmm3}<p>
	 * {@code XOP.128.X9.W0 92 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPROTD_XMM_XMMM128_XMM = 4136;
	/**
	 * {@code VPROTD xmm1, xmm2, xmm3/m128}<p>
	 * {@code XOP.128.X9.W1 92 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPROTD_XMM_XMM_XMMM128 = 4137;
	/**
	 * {@code VPROTQ xmm1, xmm2/m128, xmm3}<p>
	 * {@code XOP.128.X9.W0 93 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPROTQ_XMM_XMMM128_XMM = 4138;
	/**
	 * {@code VPROTQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code XOP.128.X9.W1 93 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPROTQ_XMM_XMM_XMMM128 = 4139;
	/**
	 * {@code VPSHLB xmm1, xmm2/m128, xmm3}<p>
	 * {@code XOP.128.X9.W0 94 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHLB_XMM_XMMM128_XMM = 4140;
	/**
	 * {@code VPSHLB xmm1, xmm2, xmm3/m128}<p>
	 * {@code XOP.128.X9.W1 94 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHLB_XMM_XMM_XMMM128 = 4141;
	/**
	 * {@code VPSHLW xmm1, xmm2/m128, xmm3}<p>
	 * {@code XOP.128.X9.W0 95 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHLW_XMM_XMMM128_XMM = 4142;
	/**
	 * {@code VPSHLW xmm1, xmm2, xmm3/m128}<p>
	 * {@code XOP.128.X9.W1 95 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHLW_XMM_XMM_XMMM128 = 4143;
	/**
	 * {@code VPSHLD xmm1, xmm2/m128, xmm3}<p>
	 * {@code XOP.128.X9.W0 96 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHLD_XMM_XMMM128_XMM = 4144;
	/**
	 * {@code VPSHLD xmm1, xmm2, xmm3/m128}<p>
	 * {@code XOP.128.X9.W1 96 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHLD_XMM_XMM_XMMM128 = 4145;
	/**
	 * {@code VPSHLQ xmm1, xmm2/m128, xmm3}<p>
	 * {@code XOP.128.X9.W0 97 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHLQ_XMM_XMMM128_XMM = 4146;
	/**
	 * {@code VPSHLQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code XOP.128.X9.W1 97 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHLQ_XMM_XMM_XMMM128 = 4147;
	/**
	 * {@code VPSHAB xmm1, xmm2/m128, xmm3}<p>
	 * {@code XOP.128.X9.W0 98 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHAB_XMM_XMMM128_XMM = 4148;
	/**
	 * {@code VPSHAB xmm1, xmm2, xmm3/m128}<p>
	 * {@code XOP.128.X9.W1 98 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHAB_XMM_XMM_XMMM128 = 4149;
	/**
	 * {@code VPSHAW xmm1, xmm2/m128, xmm3}<p>
	 * {@code XOP.128.X9.W0 99 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHAW_XMM_XMMM128_XMM = 4150;
	/**
	 * {@code VPSHAW xmm1, xmm2, xmm3/m128}<p>
	 * {@code XOP.128.X9.W1 99 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHAW_XMM_XMM_XMMM128 = 4151;
	/**
	 * {@code VPSHAD xmm1, xmm2/m128, xmm3}<p>
	 * {@code XOP.128.X9.W0 9A /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHAD_XMM_XMMM128_XMM = 4152;
	/**
	 * {@code VPSHAD xmm1, xmm2, xmm3/m128}<p>
	 * {@code XOP.128.X9.W1 9A /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHAD_XMM_XMM_XMMM128 = 4153;
	/**
	 * {@code VPSHAQ xmm1, xmm2/m128, xmm3}<p>
	 * {@code XOP.128.X9.W0 9B /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHAQ_XMM_XMMM128_XMM = 4154;
	/**
	 * {@code VPSHAQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code XOP.128.X9.W1 9B /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPSHAQ_XMM_XMM_XMMM128 = 4155;
	/**
	 * {@code VPHADDBW xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 C1 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPHADDBW_XMM_XMMM128 = 4156;
	/**
	 * {@code VPHADDBD xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 C2 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPHADDBD_XMM_XMMM128 = 4157;
	/**
	 * {@code VPHADDBQ xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 C3 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPHADDBQ_XMM_XMMM128 = 4158;
	/**
	 * {@code VPHADDWD xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 C6 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPHADDWD_XMM_XMMM128 = 4159;
	/**
	 * {@code VPHADDWQ xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 C7 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPHADDWQ_XMM_XMMM128 = 4160;
	/**
	 * {@code VPHADDDQ xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 CB /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPHADDDQ_XMM_XMMM128 = 4161;
	/**
	 * {@code VPHADDUBW xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 D1 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPHADDUBW_XMM_XMMM128 = 4162;
	/**
	 * {@code VPHADDUBD xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 D2 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPHADDUBD_XMM_XMMM128 = 4163;
	/**
	 * {@code VPHADDUBQ xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 D3 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPHADDUBQ_XMM_XMMM128 = 4164;
	/**
	 * {@code VPHADDUWD xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 D6 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPHADDUWD_XMM_XMMM128 = 4165;
	/**
	 * {@code VPHADDUWQ xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 D7 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPHADDUWQ_XMM_XMMM128 = 4166;
	/**
	 * {@code VPHADDUDQ xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 DB /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPHADDUDQ_XMM_XMMM128 = 4167;
	/**
	 * {@code VPHSUBBW xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 E1 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPHSUBBW_XMM_XMMM128 = 4168;
	/**
	 * {@code VPHSUBWD xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 E2 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPHSUBWD_XMM_XMMM128 = 4169;
	/**
	 * {@code VPHSUBDQ xmm1, xmm2/m128}<p>
	 * {@code XOP.128.X9.W0 E3 /r}<p>
	 * {@code XOP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_VPHSUBDQ_XMM_XMMM128 = 4170;
	/**
	 * {@code BEXTR r32, r/m32, imm32}<p>
	 * {@code XOP.L0.XA.W0 10 /r id}<p>
	 * {@code TBM}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_BEXTR_R32_RM32_IMM32 = 4171;
	/**
	 * {@code BEXTR r64, r/m64, imm32}<p>
	 * {@code XOP.L0.XA.W1 10 /r id}<p>
	 * {@code TBM}<p>
	 * {@code 64-bit}
	 */
	public static final int XOP_BEXTR_R64_RM64_IMM32 = 4172;
	/**
	 * {@code LWPINS r32, r/m32, imm32}<p>
	 * {@code XOP.L0.XA.W0 12 /0 id}<p>
	 * {@code LWP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_LWPINS_R32_RM32_IMM32 = 4173;
	/**
	 * {@code LWPINS r64, r/m32, imm32}<p>
	 * {@code XOP.L0.XA.W1 12 /0 id}<p>
	 * {@code LWP}<p>
	 * {@code 64-bit}
	 */
	public static final int XOP_LWPINS_R64_RM32_IMM32 = 4174;
	/**
	 * {@code LWPVAL r32, r/m32, imm32}<p>
	 * {@code XOP.L0.XA.W0 12 /1 id}<p>
	 * {@code LWP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XOP_LWPVAL_R32_RM32_IMM32 = 4175;
	/**
	 * {@code LWPVAL r64, r/m32, imm32}<p>
	 * {@code XOP.L0.XA.W1 12 /1 id}<p>
	 * {@code LWP}<p>
	 * {@code 64-bit}
	 */
	public static final int XOP_LWPVAL_R64_RM32_IMM32 = 4176;
	/**
	 * {@code PI2FW mm, mm/m64}<p>
	 * {@code 0F 0F /r 0C}<p>
	 * {@code 3DNOWEXT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PI2FW_MM_MMM64 = 4177;
	/**
	 * {@code PI2FD mm, mm/m64}<p>
	 * {@code 0F 0F /r 0D}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PI2FD_MM_MMM64 = 4178;
	/**
	 * {@code PF2IW mm, mm/m64}<p>
	 * {@code 0F 0F /r 1C}<p>
	 * {@code 3DNOWEXT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PF2IW_MM_MMM64 = 4179;
	/**
	 * {@code PF2ID mm, mm/m64}<p>
	 * {@code 0F 0F /r 1D}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PF2ID_MM_MMM64 = 4180;
	/**
	 * {@code PFRCPV mm, mm/m64}<p>
	 * {@code 0F 0F /r 86}<p>
	 * {@code AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int D3NOW_PFRCPV_MM_MMM64 = 4181;
	/**
	 * {@code PFRSQRTV mm, mm/m64}<p>
	 * {@code 0F 0F /r 87}<p>
	 * {@code AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int D3NOW_PFRSQRTV_MM_MMM64 = 4182;
	/**
	 * {@code PFNACC mm, mm/m64}<p>
	 * {@code 0F 0F /r 8A}<p>
	 * {@code 3DNOWEXT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFNACC_MM_MMM64 = 4183;
	/**
	 * {@code PFPNACC mm, mm/m64}<p>
	 * {@code 0F 0F /r 8E}<p>
	 * {@code 3DNOWEXT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFPNACC_MM_MMM64 = 4184;
	/**
	 * {@code PFCMPGE mm, mm/m64}<p>
	 * {@code 0F 0F /r 90}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFCMPGE_MM_MMM64 = 4185;
	/**
	 * {@code PFMIN mm, mm/m64}<p>
	 * {@code 0F 0F /r 94}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFMIN_MM_MMM64 = 4186;
	/**
	 * {@code PFRCP mm, mm/m64}<p>
	 * {@code 0F 0F /r 96}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFRCP_MM_MMM64 = 4187;
	/**
	 * {@code PFRSQRT mm, mm/m64}<p>
	 * {@code 0F 0F /r 97}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFRSQRT_MM_MMM64 = 4188;
	/**
	 * {@code PFSUB mm, mm/m64}<p>
	 * {@code 0F 0F /r 9A}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFSUB_MM_MMM64 = 4189;
	/**
	 * {@code PFADD mm, mm/m64}<p>
	 * {@code 0F 0F /r 9E}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFADD_MM_MMM64 = 4190;
	/**
	 * {@code PFCMPGT mm, mm/m64}<p>
	 * {@code 0F 0F /r A0}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFCMPGT_MM_MMM64 = 4191;
	/**
	 * {@code PFMAX mm, mm/m64}<p>
	 * {@code 0F 0F /r A4}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFMAX_MM_MMM64 = 4192;
	/**
	 * {@code PFRCPIT1 mm, mm/m64}<p>
	 * {@code 0F 0F /r A6}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFRCPIT1_MM_MMM64 = 4193;
	/**
	 * {@code PFRSQIT1 mm, mm/m64}<p>
	 * {@code 0F 0F /r A7}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFRSQIT1_MM_MMM64 = 4194;
	/**
	 * {@code PFSUBR mm, mm/m64}<p>
	 * {@code 0F 0F /r AA}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFSUBR_MM_MMM64 = 4195;
	/**
	 * {@code PFACC mm, mm/m64}<p>
	 * {@code 0F 0F /r AE}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFACC_MM_MMM64 = 4196;
	/**
	 * {@code PFCMPEQ mm, mm/m64}<p>
	 * {@code 0F 0F /r B0}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFCMPEQ_MM_MMM64 = 4197;
	/**
	 * {@code PFMUL mm, mm/m64}<p>
	 * {@code 0F 0F /r B4}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFMUL_MM_MMM64 = 4198;
	/**
	 * {@code PFRCPIT2 mm, mm/m64}<p>
	 * {@code 0F 0F /r B6}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PFRCPIT2_MM_MMM64 = 4199;
	/**
	 * {@code PMULHRW mm, mm/m64}<p>
	 * {@code 0F 0F /r B7}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PMULHRW_MM_MMM64 = 4200;
	/**
	 * {@code PSWAPD mm, mm/m64}<p>
	 * {@code 0F 0F /r BB}<p>
	 * {@code 3DNOWEXT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PSWAPD_MM_MMM64 = 4201;
	/**
	 * {@code PAVGUSB mm, mm/m64}<p>
	 * {@code 0F 0F /r BF}<p>
	 * {@code 3DNOW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int D3NOW_PAVGUSB_MM_MMM64 = 4202;
	/**
	 * {@code RMPADJUST}<p>
	 * {@code F3 0F 01 FE}<p>
	 * {@code SEV-SNP}<p>
	 * {@code 64-bit}
	 */
	public static final int RMPADJUST = 4203;
	/**
	 * {@code RMPUPDATE}<p>
	 * {@code F2 0F 01 FE}<p>
	 * {@code SEV-SNP}<p>
	 * {@code 64-bit}
	 */
	public static final int RMPUPDATE = 4204;
	/**
	 * {@code PSMASH}<p>
	 * {@code F3 0F 01 FF}<p>
	 * {@code SEV-SNP}<p>
	 * {@code 64-bit}
	 */
	public static final int PSMASH = 4205;
	/**
	 * {@code PVALIDATE}<p>
	 * {@code a16 F2 0F 01 FF}<p>
	 * {@code SEV-SNP}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PVALIDATEW = 4206;
	/**
	 * {@code PVALIDATE}<p>
	 * {@code a32 F2 0F 01 FF}<p>
	 * {@code SEV-SNP}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PVALIDATED = 4207;
	/**
	 * {@code PVALIDATE}<p>
	 * {@code a64 F2 0F 01 FF}<p>
	 * {@code SEV-SNP}<p>
	 * {@code 64-bit}
	 */
	public static final int PVALIDATEQ = 4208;
	/**
	 * {@code SERIALIZE}<p>
	 * {@code NP 0F 01 E8}<p>
	 * {@code SERIALIZE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int SERIALIZE = 4209;
	/**
	 * {@code XSUSLDTRK}<p>
	 * {@code F2 0F 01 E8}<p>
	 * {@code TSXLDTRK}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XSUSLDTRK = 4210;
	/**
	 * {@code XRESLDTRK}<p>
	 * {@code F2 0F 01 E9}<p>
	 * {@code TSXLDTRK}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XRESLDTRK = 4211;
	/**
	 * {@code INVLPGB}<p>
	 * {@code a16 NP 0F 01 FE}<p>
	 * {@code INVLPGB}<p>
	 * {@code 16/32-bit}
	 */
	public static final int INVLPGBW = 4212;
	/**
	 * {@code INVLPGB}<p>
	 * {@code a32 NP 0F 01 FE}<p>
	 * {@code INVLPGB}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int INVLPGBD = 4213;
	/**
	 * {@code INVLPGB}<p>
	 * {@code a64 NP 0F 01 FE}<p>
	 * {@code INVLPGB}<p>
	 * {@code 64-bit}
	 */
	public static final int INVLPGBQ = 4214;
	/**
	 * {@code TLBSYNC}<p>
	 * {@code NP 0F 01 FF}<p>
	 * {@code INVLPGB}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TLBSYNC = 4215;
	/**
	 * {@code PREFETCHW m8}<p>
	 * {@code 0F 0D /3}<p>
	 * {@code PREFETCHW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PREFETCHRESERVED3_M8 = 4216;
	/**
	 * {@code PREFETCH m8}<p>
	 * {@code 0F 0D /4}<p>
	 * {@code PREFETCHW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PREFETCHRESERVED4_M8 = 4217;
	/**
	 * {@code PREFETCH m8}<p>
	 * {@code 0F 0D /5}<p>
	 * {@code PREFETCHW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PREFETCHRESERVED5_M8 = 4218;
	/**
	 * {@code PREFETCH m8}<p>
	 * {@code 0F 0D /6}<p>
	 * {@code PREFETCHW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PREFETCHRESERVED6_M8 = 4219;
	/**
	 * {@code PREFETCH m8}<p>
	 * {@code 0F 0D /7}<p>
	 * {@code PREFETCHW}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PREFETCHRESERVED7_M8 = 4220;
	/**
	 * {@code UD0}<p>
	 * {@code 0F FF}<p>
	 * {@code 286+}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int UD0 = 4221;
	/**
	 * {@code VMGEXIT}<p>
	 * {@code F3 0F 01 D9}<p>
	 * {@code SEV-ES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VMGEXIT = 4222;
	/**
	 * {@code GETSECQ}<p>
	 * {@code NP o64 0F 37}<p>
	 * {@code SMX}<p>
	 * {@code 64-bit}
	 */
	public static final int GETSECQ = 4223;
	/**
	 * {@code LDTILECFG m512}<p>
	 * {@code VEX.128.0F38.W0 49 !(11):000:bbb}<p>
	 * {@code AMX-TILE}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_LDTILECFG_M512 = 4224;
	/**
	 * {@code TILERELEASE}<p>
	 * {@code VEX.128.0F38.W0 49 C0}<p>
	 * {@code AMX-TILE}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_TILERELEASE = 4225;
	/**
	 * {@code STTILECFG m512}<p>
	 * {@code VEX.128.66.0F38.W0 49 !(11):000:bbb}<p>
	 * {@code AMX-TILE}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_STTILECFG_M512 = 4226;
	/**
	 * {@code TILEZERO tmm1}<p>
	 * {@code VEX.128.F2.0F38.W0 49 11:rrr:000}<p>
	 * {@code AMX-TILE}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_TILEZERO_TMM = 4227;
	/**
	 * {@code TILELOADDT1 tmm1, sibmem}<p>
	 * {@code VEX.128.66.0F38.W0 4B !(11):rrr:100}<p>
	 * {@code AMX-TILE}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_TILELOADDT1_TMM_SIBMEM = 4228;
	/**
	 * {@code TILESTORED sibmem, tmm1}<p>
	 * {@code VEX.128.F3.0F38.W0 4B !(11):rrr:100}<p>
	 * {@code AMX-TILE}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_TILESTORED_SIBMEM_TMM = 4229;
	/**
	 * {@code TILELOADD tmm1, sibmem}<p>
	 * {@code VEX.128.F2.0F38.W0 4B !(11):rrr:100}<p>
	 * {@code AMX-TILE}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_TILELOADD_TMM_SIBMEM = 4230;
	/**
	 * {@code TDPBF16PS tmm1, tmm2, tmm3}<p>
	 * {@code VEX.128.F3.0F38.W0 5C 11:rrr:bbb}<p>
	 * {@code AMX-BF16}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_TDPBF16PS_TMM_TMM_TMM = 4231;
	/**
	 * {@code TDPBUUD tmm1, tmm2, tmm3}<p>
	 * {@code VEX.128.0F38.W0 5E 11:rrr:bbb}<p>
	 * {@code AMX-INT8}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_TDPBUUD_TMM_TMM_TMM = 4232;
	/**
	 * {@code TDPBUSD tmm1, tmm2, tmm3}<p>
	 * {@code VEX.128.66.0F38.W0 5E 11:rrr:bbb}<p>
	 * {@code AMX-INT8}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_TDPBUSD_TMM_TMM_TMM = 4233;
	/**
	 * {@code TDPBSUD tmm1, tmm2, tmm3}<p>
	 * {@code VEX.128.F3.0F38.W0 5E 11:rrr:bbb}<p>
	 * {@code AMX-INT8}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_TDPBSUD_TMM_TMM_TMM = 4234;
	/**
	 * {@code TDPBSSD tmm1, tmm2, tmm3}<p>
	 * {@code VEX.128.F2.0F38.W0 5E 11:rrr:bbb}<p>
	 * {@code AMX-INT8}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_TDPBSSD_TMM_TMM_TMM = 4235;
	/**
	 * {@code FNSTDW AX}<p>
	 * {@code DF E1}<p>
	 * {@code 387 SL}<p>
	 * {@code 16/32-bit}
	 */
	public static final int FNSTDW_AX = 4236;
	/**
	 * {@code FNSTSG AX}<p>
	 * {@code DF E2}<p>
	 * {@code 387 SL}<p>
	 * {@code 16/32-bit}
	 */
	public static final int FNSTSG_AX = 4237;
	/**
	 * {@code RDSHR r/m32}<p>
	 * {@code 0F 36 /0}<p>
	 * {@code Cyrix 6x86MX, M II, III}<p>
	 * {@code 16/32-bit}
	 */
	public static final int RDSHR_RM32 = 4238;
	/**
	 * {@code WRSHR r/m32}<p>
	 * {@code 0F 37 /0}<p>
	 * {@code Cyrix 6x86MX, M II, III}<p>
	 * {@code 16/32-bit}
	 */
	public static final int WRSHR_RM32 = 4239;
	/**
	 * {@code SMINT}<p>
	 * {@code 0F 38}<p>
	 * {@code Cyrix 6x86MX+, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int SMINT = 4240;
	/**
	 * {@code DMINT}<p>
	 * {@code 0F 39}<p>
	 * {@code AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int DMINT = 4241;
	/**
	 * {@code RDM}<p>
	 * {@code 0F 3A}<p>
	 * {@code AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int RDM = 4242;
	/**
	 * {@code SVDC m80, Sreg}<p>
	 * {@code 0F 78 /r}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int SVDC_M80_SREG = 4243;
	/**
	 * {@code RSDC Sreg, m80}<p>
	 * {@code 0F 79 /r}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int RSDC_SREG_M80 = 4244;
	/**
	 * {@code SVLDT m80}<p>
	 * {@code 0F 7A /0}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int SVLDT_M80 = 4245;
	/**
	 * {@code RSLDT m80}<p>
	 * {@code 0F 7B /0}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int RSLDT_M80 = 4246;
	/**
	 * {@code SVTS m80}<p>
	 * {@code 0F 7C /0}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int SVTS_M80 = 4247;
	/**
	 * {@code RSTS m80}<p>
	 * {@code 0F 7D /0}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int RSTS_M80 = 4248;
	/**
	 * {@code SMINT}<p>
	 * {@code 0F 7E}<p>
	 * {@code Cyrix 6x86 or earlier}<p>
	 * {@code 16/32-bit}
	 */
	public static final int SMINT_0F7E = 4249;
	/**
	 * {@code BB0_RESET}<p>
	 * {@code 0F 3A}<p>
	 * {@code Cyrix MediaGX, GXm, GXLV, GX1}<p>
	 * {@code 16/32-bit}
	 */
	public static final int BB0_RESET = 4250;
	/**
	 * {@code BB1_RESET}<p>
	 * {@code 0F 3B}<p>
	 * {@code Cyrix MediaGX, GXm, GXLV, GX1}<p>
	 * {@code 16/32-bit}
	 */
	public static final int BB1_RESET = 4251;
	/**
	 * {@code CPU_WRITE}<p>
	 * {@code 0F 3C}<p>
	 * {@code Cyrix MediaGX, GXm, GXLV, GX1}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CPU_WRITE = 4252;
	/**
	 * {@code CPU_READ}<p>
	 * {@code 0F 3D}<p>
	 * {@code Cyrix MediaGX, GXm, GXLV, GX1}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CPU_READ = 4253;
	/**
	 * {@code ALTINST}<p>
	 * {@code 0F 3F}<p>
	 * {@code Centaur AIS}<p>
	 * {@code 16/32-bit}
	 */
	public static final int ALTINST = 4254;
	/**
	 * {@code PAVEB mm, mm/m64}<p>
	 * {@code 0F 50 /r}<p>
	 * {@code CYRIX_EMMI}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PAVEB_MM_MMM64 = 4255;
	/**
	 * {@code PADDSIW mm, mm/m64}<p>
	 * {@code 0F 51 /r}<p>
	 * {@code CYRIX_EMMI}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PADDSIW_MM_MMM64 = 4256;
	/**
	 * {@code PMAGW mm, mm/m64}<p>
	 * {@code 0F 52 /r}<p>
	 * {@code CYRIX_EMMI}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PMAGW_MM_MMM64 = 4257;
	/**
	 * {@code PDISTIB mm, m64}<p>
	 * {@code 0F 54 /r}<p>
	 * {@code CYRIX_EMMI}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PDISTIB_MM_M64 = 4258;
	/**
	 * {@code PSUBSIW mm, mm/m64}<p>
	 * {@code 0F 55 /r}<p>
	 * {@code CYRIX_EMMI}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PSUBSIW_MM_MMM64 = 4259;
	/**
	 * {@code PMVZB mm, m64}<p>
	 * {@code 0F 58 /r}<p>
	 * {@code CYRIX_EMMI}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PMVZB_MM_M64 = 4260;
	/**
	 * {@code PMULHRW mm, mm/m64}<p>
	 * {@code 0F 59 /r}<p>
	 * {@code CYRIX_EMMI}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PMULHRW_MM_MMM64 = 4261;
	/**
	 * {@code PMVNZB mm, m64}<p>
	 * {@code 0F 5A /r}<p>
	 * {@code CYRIX_EMMI}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PMVNZB_MM_M64 = 4262;
	/**
	 * {@code PMVLZB mm, m64}<p>
	 * {@code 0F 5B /r}<p>
	 * {@code CYRIX_EMMI}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PMVLZB_MM_M64 = 4263;
	/**
	 * {@code PMVGEZB mm, m64}<p>
	 * {@code 0F 5C /r}<p>
	 * {@code CYRIX_EMMI}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PMVGEZB_MM_M64 = 4264;
	/**
	 * {@code PMULHRIW mm, mm/m64}<p>
	 * {@code 0F 5D /r}<p>
	 * {@code CYRIX_EMMI}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PMULHRIW_MM_MMM64 = 4265;
	/**
	 * {@code PMACHRIW mm, m64}<p>
	 * {@code 0F 5E /r}<p>
	 * {@code CYRIX_EMMI}<p>
	 * {@code 16/32-bit}
	 */
	public static final int PMACHRIW_MM_M64 = 4266;
	/**
	 * {@code UNDOC}<p>
	 * {@code D9 D7}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CYRIX_D9D7 = 4267;
	/**
	 * {@code UNDOC}<p>
	 * {@code D9 E2}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CYRIX_D9E2 = 4268;
	/**
	 * {@code FTSTP}<p>
	 * {@code D9 E6}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int FTSTP = 4269;
	/**
	 * {@code UNDOC}<p>
	 * {@code D9 E7}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CYRIX_D9E7 = 4270;
	/**
	 * {@code FRINT2}<p>
	 * {@code DB FC}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int FRINT2 = 4271;
	/**
	 * {@code FRICHOP}<p>
	 * {@code DD FC}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int FRICHOP = 4272;
	/**
	 * {@code UNDOC}<p>
	 * {@code DE D8}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CYRIX_DED8 = 4273;
	/**
	 * {@code UNDOC}<p>
	 * {@code DE DA}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CYRIX_DEDA = 4274;
	/**
	 * {@code UNDOC}<p>
	 * {@code DE DC}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CYRIX_DEDC = 4275;
	/**
	 * {@code UNDOC}<p>
	 * {@code DE DD}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CYRIX_DEDD = 4276;
	/**
	 * {@code UNDOC}<p>
	 * {@code DE DE}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CYRIX_DEDE = 4277;
	/**
	 * {@code FRINEAR}<p>
	 * {@code DF FC}<p>
	 * {@code Cyrix, AMD Geode GX/LX}<p>
	 * {@code 16/32-bit}
	 */
	public static final int FRINEAR = 4278;
	/**
	 * {@code TDCALL}<p>
	 * {@code 66 0F 01 CC}<p>
	 * {@code TDX}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int TDCALL = 4279;
	/**
	 * {@code SEAMRET}<p>
	 * {@code 66 0F 01 CD}<p>
	 * {@code TDX}<p>
	 * {@code 64-bit}
	 */
	public static final int SEAMRET = 4280;
	/**
	 * {@code SEAMOPS}<p>
	 * {@code 66 0F 01 CE}<p>
	 * {@code TDX}<p>
	 * {@code 64-bit}
	 */
	public static final int SEAMOPS = 4281;
	/**
	 * {@code SEAMCALL}<p>
	 * {@code 66 0F 01 CF}<p>
	 * {@code TDX}<p>
	 * {@code 64-bit}
	 */
	public static final int SEAMCALL = 4282;
	/**
	 * {@code AESENCWIDE128KL m384, &lt;XMM0-7&gt;}<p>
	 * {@code F3 0F 38 D8 !(11):000:bbb}<p>
	 * {@code AESKLE and WIDE_KL}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AESENCWIDE128KL_M384 = 4283;
	/**
	 * {@code AESDECWIDE128KL m384, &lt;XMM0-7&gt;}<p>
	 * {@code F3 0F 38 D8 !(11):001:bbb}<p>
	 * {@code AESKLE and WIDE_KL}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AESDECWIDE128KL_M384 = 4284;
	/**
	 * {@code AESENCWIDE256KL m512, &lt;XMM0-7&gt;}<p>
	 * {@code F3 0F 38 D8 !(11):010:bbb}<p>
	 * {@code AESKLE and WIDE_KL}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AESENCWIDE256KL_M512 = 4285;
	/**
	 * {@code AESDECWIDE256KL m512, &lt;XMM0-7&gt;}<p>
	 * {@code F3 0F 38 D8 !(11):011:bbb}<p>
	 * {@code AESKLE and WIDE_KL}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AESDECWIDE256KL_M512 = 4286;
	/**
	 * {@code LOADIWKEY xmm1, xmm2, &lt;EAX&gt;, &lt;XMM0&gt;}<p>
	 * {@code F3 0F 38 DC 11:rrr:bbb}<p>
	 * {@code KL}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int LOADIWKEY_XMM_XMM = 4287;
	/**
	 * {@code AESENC128KL xmm, m384}<p>
	 * {@code F3 0F 38 DC !(11):rrr:bbb}<p>
	 * {@code AESKLE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AESENC128KL_XMM_M384 = 4288;
	/**
	 * {@code AESDEC128KL xmm, m384}<p>
	 * {@code F3 0F 38 DD !(11):rrr:bbb}<p>
	 * {@code AESKLE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AESDEC128KL_XMM_M384 = 4289;
	/**
	 * {@code AESENC256KL xmm, m512}<p>
	 * {@code F3 0F 38 DE !(11):rrr:bbb}<p>
	 * {@code AESKLE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AESENC256KL_XMM_M512 = 4290;
	/**
	 * {@code AESDEC256KL xmm, m512}<p>
	 * {@code F3 0F 38 DF !(11):rrr:bbb}<p>
	 * {@code AESKLE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AESDEC256KL_XMM_M512 = 4291;
	/**
	 * {@code ENCODEKEY128 r32, r32, &lt;XMM0-2&gt;, &lt;XMM4-6&gt;}<p>
	 * {@code F3 0F 38 FA 11:rrr:bbb}<p>
	 * {@code AESKLE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ENCODEKEY128_R32_R32 = 4292;
	/**
	 * {@code ENCODEKEY256 r32, r32, &lt;XMM0-6&gt;}<p>
	 * {@code F3 0F 38 FB 11:rrr:bbb}<p>
	 * {@code AESKLE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int ENCODEKEY256_R32_R32 = 4293;
	/**
	 * {@code VBROADCASTSS xmm1, xmm2}<p>
	 * {@code VEX.128.66.0F38.W0 18 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBROADCASTSS_XMM_XMM = 4294;
	/**
	 * {@code VBROADCASTSS ymm1, xmm2}<p>
	 * {@code VEX.256.66.0F38.W0 18 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBROADCASTSS_YMM_XMM = 4295;
	/**
	 * {@code VBROADCASTSD ymm1, xmm2}<p>
	 * {@code VEX.256.66.0F38.W0 19 /r}<p>
	 * {@code AVX2}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBROADCASTSD_YMM_XMM = 4296;
	/**
	 * {@code VMGEXIT}<p>
	 * {@code F2 0F 01 D9}<p>
	 * {@code SEV-ES}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VMGEXIT_F2 = 4297;
	/**
	 * {@code UIRET}<p>
	 * {@code F3 0F 01 EC}<p>
	 * {@code UINTR}<p>
	 * {@code 64-bit}
	 */
	public static final int UIRET = 4298;
	/**
	 * {@code TESTUI}<p>
	 * {@code F3 0F 01 ED}<p>
	 * {@code UINTR}<p>
	 * {@code 64-bit}
	 */
	public static final int TESTUI = 4299;
	/**
	 * {@code CLUI}<p>
	 * {@code F3 0F 01 EE}<p>
	 * {@code UINTR}<p>
	 * {@code 64-bit}
	 */
	public static final int CLUI = 4300;
	/**
	 * {@code STUI}<p>
	 * {@code F3 0F 01 EF}<p>
	 * {@code UINTR}<p>
	 * {@code 64-bit}
	 */
	public static final int STUI = 4301;
	/**
	 * {@code SENDUIPI r64}<p>
	 * {@code F3 0F C7 /6}<p>
	 * {@code UINTR}<p>
	 * {@code 64-bit}
	 */
	public static final int SENDUIPI_R64 = 4302;
	/**
	 * {@code HRESET imm8, &lt;EAX&gt;}<p>
	 * {@code F3 0F 3A F0 C0 ib}<p>
	 * {@code HRESET}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int HRESET_IMM8 = 4303;
	/**
	 * {@code VPDPBUSD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 50 /r}<p>
	 * {@code AVX-VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBUSD_XMM_XMM_XMMM128 = 4304;
	/**
	 * {@code VPDPBUSD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 50 /r}<p>
	 * {@code AVX-VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBUSD_YMM_YMM_YMMM256 = 4305;
	/**
	 * {@code VPDPBUSDS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 51 /r}<p>
	 * {@code AVX-VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBUSDS_XMM_XMM_XMMM128 = 4306;
	/**
	 * {@code VPDPBUSDS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 51 /r}<p>
	 * {@code AVX-VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBUSDS_YMM_YMM_YMMM256 = 4307;
	/**
	 * {@code VPDPWSSD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 52 /r}<p>
	 * {@code AVX-VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPWSSD_XMM_XMM_XMMM128 = 4308;
	/**
	 * {@code VPDPWSSD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 52 /r}<p>
	 * {@code AVX-VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPWSSD_YMM_YMM_YMMM256 = 4309;
	/**
	 * {@code VPDPWSSDS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W0 53 /r}<p>
	 * {@code AVX-VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPWSSDS_XMM_XMM_XMMM128 = 4310;
	/**
	 * {@code VPDPWSSDS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W0 53 /r}<p>
	 * {@code AVX-VNNI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPWSSDS_YMM_YMM_YMMM256 = 4311;
	/**
	 * {@code CCS_HASH}<p>
	 * {@code a16 F3 0F A6 E8}<p>
	 * {@code PADLOCK_GMI}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CCS_HASH_16 = 4312;
	/**
	 * {@code CCS_HASH}<p>
	 * {@code a32 F3 0F A6 E8}<p>
	 * {@code PADLOCK_GMI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CCS_HASH_32 = 4313;
	/**
	 * {@code CCS_HASH}<p>
	 * {@code a64 F3 0F A6 E8}<p>
	 * {@code PADLOCK_GMI}<p>
	 * {@code 64-bit}
	 */
	public static final int CCS_HASH_64 = 4314;
	/**
	 * {@code CCS_ENCRYPT}<p>
	 * {@code a16 F3 0F A7 F0}<p>
	 * {@code PADLOCK_GMI}<p>
	 * {@code 16/32-bit}
	 */
	public static final int CCS_ENCRYPT_16 = 4315;
	/**
	 * {@code CCS_ENCRYPT}<p>
	 * {@code a32 F3 0F A7 F0}<p>
	 * {@code PADLOCK_GMI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int CCS_ENCRYPT_32 = 4316;
	/**
	 * {@code CCS_ENCRYPT}<p>
	 * {@code a64 F3 0F A7 F0}<p>
	 * {@code PADLOCK_GMI}<p>
	 * {@code 64-bit}
	 */
	public static final int CCS_ENCRYPT_64 = 4317;
	/**
	 * {@code LKGS r/m16}<p>
	 * {@code o16 F2 0F 00 /6}<p>
	 * {@code LKGS}<p>
	 * {@code 64-bit}
	 */
	public static final int LKGS_RM16 = 4318;
	/**
	 * {@code LKGS r32/m16}<p>
	 * {@code o32 F2 0F 00 /6}<p>
	 * {@code LKGS}<p>
	 * {@code 64-bit}
	 */
	public static final int LKGS_R32M16 = 4319;
	/**
	 * {@code LKGS r64/m16}<p>
	 * {@code F2 o64 0F 00 /6}<p>
	 * {@code LKGS}<p>
	 * {@code 64-bit}
	 */
	public static final int LKGS_R64M16 = 4320;
	/**
	 * {@code ERETU}<p>
	 * {@code F3 0F 01 CA}<p>
	 * {@code FRED}<p>
	 * {@code 64-bit}
	 */
	public static final int ERETU = 4321;
	/**
	 * {@code ERETS}<p>
	 * {@code F2 0F 01 CA}<p>
	 * {@code FRED}<p>
	 * {@code 64-bit}
	 */
	public static final int ERETS = 4322;
	/**
	 * {@code VADDPH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.MAP5.W0 58 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VADDPH_XMM_K1Z_XMM_XMMM128B16 = 4323;
	/**
	 * {@code VADDPH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.MAP5.W0 58 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VADDPH_YMM_K1Z_YMM_YMMM256B16 = 4324;
	/**
	 * {@code VADDPH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.MAP5.W0 58 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VADDPH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4325;
	/**
	 * {@code VADDSH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 58 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VADDSH_XMM_K1Z_XMM_XMMM16_ER = 4326;
	/**
	 * {@code VCMPPH k1 {k2}, xmm2, xmm3/m128/m16bcst, imm8}<p>
	 * {@code EVEX.128.0F3A.W0 C2 /r ib}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCMPPH_KR_K1_XMM_XMMM128B16_IMM8 = 4327;
	/**
	 * {@code VCMPPH k1 {k2}, ymm2, ymm3/m256/m16bcst, imm8}<p>
	 * {@code EVEX.256.0F3A.W0 C2 /r ib}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCMPPH_KR_K1_YMM_YMMM256B16_IMM8 = 4328;
	/**
	 * {@code VCMPPH k1 {k2}, zmm2, zmm3/m512/m16bcst{sae}, imm8}<p>
	 * {@code EVEX.512.0F3A.W0 C2 /r ib}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCMPPH_KR_K1_ZMM_ZMMM512B16_IMM8_SAE = 4329;
	/**
	 * {@code VCMPSH k1 {k2}, xmm2, xmm3/m16{sae}, imm8}<p>
	 * {@code EVEX.LIG.F3.0F3A.W0 C2 /r ib}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCMPSH_KR_K1_XMM_XMMM16_IMM8_SAE = 4330;
	/**
	 * {@code VCOMISH xmm1, xmm2/m16{sae}}<p>
	 * {@code EVEX.LIG.MAP5.W0 2F /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCOMISH_XMM_XMMM16_SAE = 4331;
	/**
	 * {@code VCVTDQ2PH xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.MAP5.W0 5B /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTDQ2PH_XMM_K1Z_XMMM128B32 = 4332;
	/**
	 * {@code VCVTDQ2PH xmm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.MAP5.W0 5B /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTDQ2PH_XMM_K1Z_YMMM256B32 = 4333;
	/**
	 * {@code VCVTDQ2PH ymm1 {k1}{z}, zmm2/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.MAP5.W0 5B /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTDQ2PH_YMM_K1Z_ZMMM512B32_ER = 4334;
	/**
	 * {@code VCVTPD2PH xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.66.MAP5.W1 5A /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2PH_XMM_K1Z_XMMM128B64 = 4335;
	/**
	 * {@code VCVTPD2PH xmm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.66.MAP5.W1 5A /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2PH_XMM_K1Z_YMMM256B64 = 4336;
	/**
	 * {@code VCVTPD2PH xmm1 {k1}{z}, zmm2/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.66.MAP5.W1 5A /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPD2PH_XMM_K1Z_ZMMM512B64_ER = 4337;
	/**
	 * {@code VCVTPH2DQ xmm1 {k1}{z}, xmm2/m64/m16bcst}<p>
	 * {@code EVEX.128.66.MAP5.W0 5B /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2DQ_XMM_K1Z_XMMM64B16 = 4338;
	/**
	 * {@code VCVTPH2DQ ymm1 {k1}{z}, xmm2/m128/m16bcst}<p>
	 * {@code EVEX.256.66.MAP5.W0 5B /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2DQ_YMM_K1Z_XMMM128B16 = 4339;
	/**
	 * {@code VCVTPH2DQ zmm1 {k1}{z}, ymm2/m256/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP5.W0 5B /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2DQ_ZMM_K1Z_YMMM256B16_ER = 4340;
	/**
	 * {@code VCVTPH2PD xmm1 {k1}{z}, xmm2/m32/m16bcst}<p>
	 * {@code EVEX.128.MAP5.W0 5A /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2PD_XMM_K1Z_XMMM32B16 = 4341;
	/**
	 * {@code VCVTPH2PD ymm1 {k1}{z}, xmm2/m64/m16bcst}<p>
	 * {@code EVEX.256.MAP5.W0 5A /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2PD_YMM_K1Z_XMMM64B16 = 4342;
	/**
	 * {@code VCVTPH2PD zmm1 {k1}{z}, xmm2/m128/m16bcst{sae}}<p>
	 * {@code EVEX.512.MAP5.W0 5A /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2PD_ZMM_K1Z_XMMM128B16_SAE = 4343;
	/**
	 * {@code VCVTPH2PSX xmm1 {k1}{z}, xmm2/m64/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 13 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2PSX_XMM_K1Z_XMMM64B16 = 4344;
	/**
	 * {@code VCVTPH2PSX ymm1 {k1}{z}, xmm2/m128/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 13 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2PSX_YMM_K1Z_XMMM128B16 = 4345;
	/**
	 * {@code VCVTPH2PSX zmm1 {k1}{z}, ymm2/m256/m16bcst{sae}}<p>
	 * {@code EVEX.512.66.MAP6.W0 13 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2PSX_ZMM_K1Z_YMMM256B16_SAE = 4346;
	/**
	 * {@code VCVTPH2QQ xmm1 {k1}{z}, xmm2/m32/m16bcst}<p>
	 * {@code EVEX.128.66.MAP5.W0 7B /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2QQ_XMM_K1Z_XMMM32B16 = 4347;
	/**
	 * {@code VCVTPH2QQ ymm1 {k1}{z}, xmm2/m64/m16bcst}<p>
	 * {@code EVEX.256.66.MAP5.W0 7B /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2QQ_YMM_K1Z_XMMM64B16 = 4348;
	/**
	 * {@code VCVTPH2QQ zmm1 {k1}{z}, xmm2/m128/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP5.W0 7B /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2QQ_ZMM_K1Z_XMMM128B16_ER = 4349;
	/**
	 * {@code VCVTPH2UDQ xmm1 {k1}{z}, xmm2/m64/m16bcst}<p>
	 * {@code EVEX.128.MAP5.W0 79 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2UDQ_XMM_K1Z_XMMM64B16 = 4350;
	/**
	 * {@code VCVTPH2UDQ ymm1 {k1}{z}, xmm2/m128/m16bcst}<p>
	 * {@code EVEX.256.MAP5.W0 79 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2UDQ_YMM_K1Z_XMMM128B16 = 4351;
	/**
	 * {@code VCVTPH2UDQ zmm1 {k1}{z}, ymm2/m256/m16bcst{er}}<p>
	 * {@code EVEX.512.MAP5.W0 79 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2UDQ_ZMM_K1Z_YMMM256B16_ER = 4352;
	/**
	 * {@code VCVTPH2UQQ xmm1 {k1}{z}, xmm2/m32/m16bcst}<p>
	 * {@code EVEX.128.66.MAP5.W0 79 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2UQQ_XMM_K1Z_XMMM32B16 = 4353;
	/**
	 * {@code VCVTPH2UQQ ymm1 {k1}{z}, xmm2/m64/m16bcst}<p>
	 * {@code EVEX.256.66.MAP5.W0 79 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2UQQ_YMM_K1Z_XMMM64B16 = 4354;
	/**
	 * {@code VCVTPH2UQQ zmm1 {k1}{z}, xmm2/m128/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP5.W0 79 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2UQQ_ZMM_K1Z_XMMM128B16_ER = 4355;
	/**
	 * {@code VCVTPH2UW xmm1 {k1}{z}, xmm2/m128/m16bcst}<p>
	 * {@code EVEX.128.MAP5.W0 7D /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2UW_XMM_K1Z_XMMM128B16 = 4356;
	/**
	 * {@code VCVTPH2UW ymm1 {k1}{z}, ymm2/m256/m16bcst}<p>
	 * {@code EVEX.256.MAP5.W0 7D /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2UW_YMM_K1Z_YMMM256B16 = 4357;
	/**
	 * {@code VCVTPH2UW zmm1 {k1}{z}, zmm2/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.MAP5.W0 7D /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2UW_ZMM_K1Z_ZMMM512B16_ER = 4358;
	/**
	 * {@code VCVTPH2W xmm1 {k1}{z}, xmm2/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP5.W0 7D /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2W_XMM_K1Z_XMMM128B16 = 4359;
	/**
	 * {@code VCVTPH2W ymm1 {k1}{z}, ymm2/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP5.W0 7D /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2W_YMM_K1Z_YMMM256B16 = 4360;
	/**
	 * {@code VCVTPH2W zmm1 {k1}{z}, zmm2/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP5.W0 7D /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPH2W_ZMM_K1Z_ZMMM512B16_ER = 4361;
	/**
	 * {@code VCVTPS2PHX xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.66.MAP5.W0 1D /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2PHX_XMM_K1Z_XMMM128B32 = 4362;
	/**
	 * {@code VCVTPS2PHX xmm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.66.MAP5.W0 1D /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2PHX_XMM_K1Z_YMMM256B32 = 4363;
	/**
	 * {@code VCVTPS2PHX ymm1 {k1}{z}, zmm2/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.66.MAP5.W0 1D /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTPS2PHX_YMM_K1Z_ZMMM512B32_ER = 4364;
	/**
	 * {@code VCVTQQ2PH xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.MAP5.W1 5B /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTQQ2PH_XMM_K1Z_XMMM128B64 = 4365;
	/**
	 * {@code VCVTQQ2PH xmm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.MAP5.W1 5B /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTQQ2PH_XMM_K1Z_YMMM256B64 = 4366;
	/**
	 * {@code VCVTQQ2PH xmm1 {k1}{z}, zmm2/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.MAP5.W1 5B /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTQQ2PH_XMM_K1Z_ZMMM512B64_ER = 4367;
	/**
	 * {@code VCVTSD2SH xmm1 {k1}{z}, xmm2, xmm3/m64{er}}<p>
	 * {@code EVEX.LIG.F2.MAP5.W1 5A /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTSD2SH_XMM_K1Z_XMM_XMMM64_ER = 4368;
	/**
	 * {@code VCVTSH2SD xmm1 {k1}{z}, xmm2, xmm3/m16{sae}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 5A /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTSH2SD_XMM_K1Z_XMM_XMMM16_SAE = 4369;
	/**
	 * {@code VCVTSH2SI r32, xmm1/m16{er}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 2D /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTSH2SI_R32_XMMM16_ER = 4370;
	/**
	 * {@code VCVTSH2SI r64, xmm1/m16{er}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W1 2D /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTSH2SI_R64_XMMM16_ER = 4371;
	/**
	 * {@code VCVTSH2SS xmm1 {k1}{z}, xmm2, xmm3/m16{sae}}<p>
	 * {@code EVEX.LIG.MAP6.W0 13 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTSH2SS_XMM_K1Z_XMM_XMMM16_SAE = 4372;
	/**
	 * {@code VCVTSH2USI r32, xmm1/m16{er}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 79 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTSH2USI_R32_XMMM16_ER = 4373;
	/**
	 * {@code VCVTSH2USI r64, xmm1/m16{er}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W1 79 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTSH2USI_R64_XMMM16_ER = 4374;
	/**
	 * {@code VCVTSI2SH xmm1, xmm2, r/m32{er}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 2A /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTSI2SH_XMM_XMM_RM32_ER = 4375;
	/**
	 * {@code VCVTSI2SH xmm1, xmm2, r/m64{er}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W1 2A /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTSI2SH_XMM_XMM_RM64_ER = 4376;
	/**
	 * {@code VCVTSS2SH xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.MAP5.W0 1D /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTSS2SH_XMM_K1Z_XMM_XMMM32_ER = 4377;
	/**
	 * {@code VCVTTPH2DQ xmm1 {k1}{z}, xmm2/m64/m16bcst}<p>
	 * {@code EVEX.128.F3.MAP5.W0 5B /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2DQ_XMM_K1Z_XMMM64B16 = 4378;
	/**
	 * {@code VCVTTPH2DQ ymm1 {k1}{z}, xmm2/m128/m16bcst}<p>
	 * {@code EVEX.256.F3.MAP5.W0 5B /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2DQ_YMM_K1Z_XMMM128B16 = 4379;
	/**
	 * {@code VCVTTPH2DQ zmm1 {k1}{z}, ymm2/m256/m16bcst{sae}}<p>
	 * {@code EVEX.512.F3.MAP5.W0 5B /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2DQ_ZMM_K1Z_YMMM256B16_SAE = 4380;
	/**
	 * {@code VCVTTPH2QQ xmm1 {k1}{z}, xmm2/m32/m16bcst}<p>
	 * {@code EVEX.128.66.MAP5.W0 7A /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2QQ_XMM_K1Z_XMMM32B16 = 4381;
	/**
	 * {@code VCVTTPH2QQ ymm1 {k1}{z}, xmm2/m64/m16bcst}<p>
	 * {@code EVEX.256.66.MAP5.W0 7A /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2QQ_YMM_K1Z_XMMM64B16 = 4382;
	/**
	 * {@code VCVTTPH2QQ zmm1 {k1}{z}, xmm2/m128/m16bcst{sae}}<p>
	 * {@code EVEX.512.66.MAP5.W0 7A /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2QQ_ZMM_K1Z_XMMM128B16_SAE = 4383;
	/**
	 * {@code VCVTTPH2UDQ xmm1 {k1}{z}, xmm2/m64/m16bcst}<p>
	 * {@code EVEX.128.MAP5.W0 78 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2UDQ_XMM_K1Z_XMMM64B16 = 4384;
	/**
	 * {@code VCVTTPH2UDQ ymm1 {k1}{z}, xmm2/m128/m16bcst}<p>
	 * {@code EVEX.256.MAP5.W0 78 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2UDQ_YMM_K1Z_XMMM128B16 = 4385;
	/**
	 * {@code VCVTTPH2UDQ zmm1 {k1}{z}, ymm2/m256/m16bcst{sae}}<p>
	 * {@code EVEX.512.MAP5.W0 78 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2UDQ_ZMM_K1Z_YMMM256B16_SAE = 4386;
	/**
	 * {@code VCVTTPH2UQQ xmm1 {k1}{z}, xmm2/m32/m16bcst}<p>
	 * {@code EVEX.128.66.MAP5.W0 78 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2UQQ_XMM_K1Z_XMMM32B16 = 4387;
	/**
	 * {@code VCVTTPH2UQQ ymm1 {k1}{z}, xmm2/m64/m16bcst}<p>
	 * {@code EVEX.256.66.MAP5.W0 78 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2UQQ_YMM_K1Z_XMMM64B16 = 4388;
	/**
	 * {@code VCVTTPH2UQQ zmm1 {k1}{z}, xmm2/m128/m16bcst{sae}}<p>
	 * {@code EVEX.512.66.MAP5.W0 78 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2UQQ_ZMM_K1Z_XMMM128B16_SAE = 4389;
	/**
	 * {@code VCVTTPH2UW xmm1 {k1}{z}, xmm2/m128/m16bcst}<p>
	 * {@code EVEX.128.MAP5.W0 7C /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2UW_XMM_K1Z_XMMM128B16 = 4390;
	/**
	 * {@code VCVTTPH2UW ymm1 {k1}{z}, ymm2/m256/m16bcst}<p>
	 * {@code EVEX.256.MAP5.W0 7C /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2UW_YMM_K1Z_YMMM256B16 = 4391;
	/**
	 * {@code VCVTTPH2UW zmm1 {k1}{z}, zmm2/m512/m16bcst{sae}}<p>
	 * {@code EVEX.512.MAP5.W0 7C /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2UW_ZMM_K1Z_ZMMM512B16_SAE = 4392;
	/**
	 * {@code VCVTTPH2W xmm1 {k1}{z}, xmm2/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP5.W0 7C /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2W_XMM_K1Z_XMMM128B16 = 4393;
	/**
	 * {@code VCVTTPH2W ymm1 {k1}{z}, ymm2/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP5.W0 7C /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2W_YMM_K1Z_YMMM256B16 = 4394;
	/**
	 * {@code VCVTTPH2W zmm1 {k1}{z}, zmm2/m512/m16bcst{sae}}<p>
	 * {@code EVEX.512.66.MAP5.W0 7C /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTPH2W_ZMM_K1Z_ZMMM512B16_SAE = 4395;
	/**
	 * {@code VCVTTSH2SI r32, xmm1/m16{sae}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 2C /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTSH2SI_R32_XMMM16_SAE = 4396;
	/**
	 * {@code VCVTTSH2SI r64, xmm1/m16{sae}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W1 2C /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTTSH2SI_R64_XMMM16_SAE = 4397;
	/**
	 * {@code VCVTTSH2USI r32, xmm1/m16{sae}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 78 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTTSH2USI_R32_XMMM16_SAE = 4398;
	/**
	 * {@code VCVTTSH2USI r64, xmm1/m16{sae}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W1 78 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTTSH2USI_R64_XMMM16_SAE = 4399;
	/**
	 * {@code VCVTUDQ2PH xmm1 {k1}{z}, xmm2/m128/m32bcst}<p>
	 * {@code EVEX.128.F2.MAP5.W0 7A /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUDQ2PH_XMM_K1Z_XMMM128B32 = 4400;
	/**
	 * {@code VCVTUDQ2PH xmm1 {k1}{z}, ymm2/m256/m32bcst}<p>
	 * {@code EVEX.256.F2.MAP5.W0 7A /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUDQ2PH_XMM_K1Z_YMMM256B32 = 4401;
	/**
	 * {@code VCVTUDQ2PH ymm1 {k1}{z}, zmm2/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.F2.MAP5.W0 7A /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUDQ2PH_YMM_K1Z_ZMMM512B32_ER = 4402;
	/**
	 * {@code VCVTUQQ2PH xmm1 {k1}{z}, xmm2/m128/m64bcst}<p>
	 * {@code EVEX.128.F2.MAP5.W1 7A /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUQQ2PH_XMM_K1Z_XMMM128B64 = 4403;
	/**
	 * {@code VCVTUQQ2PH xmm1 {k1}{z}, ymm2/m256/m64bcst}<p>
	 * {@code EVEX.256.F2.MAP5.W1 7A /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUQQ2PH_XMM_K1Z_YMMM256B64 = 4404;
	/**
	 * {@code VCVTUQQ2PH xmm1 {k1}{z}, zmm2/m512/m64bcst{er}}<p>
	 * {@code EVEX.512.F2.MAP5.W1 7A /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUQQ2PH_XMM_K1Z_ZMMM512B64_ER = 4405;
	/**
	 * {@code VCVTUSI2SH xmm1, xmm2, r/m32{er}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 7B /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUSI2SH_XMM_XMM_RM32_ER = 4406;
	/**
	 * {@code VCVTUSI2SH xmm1, xmm2, r/m64{er}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W1 7B /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VCVTUSI2SH_XMM_XMM_RM64_ER = 4407;
	/**
	 * {@code VCVTUW2PH xmm1 {k1}{z}, xmm2/m128/m16bcst}<p>
	 * {@code EVEX.128.F2.MAP5.W0 7D /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUW2PH_XMM_K1Z_XMMM128B16 = 4408;
	/**
	 * {@code VCVTUW2PH ymm1 {k1}{z}, ymm2/m256/m16bcst}<p>
	 * {@code EVEX.256.F2.MAP5.W0 7D /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUW2PH_YMM_K1Z_YMMM256B16 = 4409;
	/**
	 * {@code VCVTUW2PH zmm1 {k1}{z}, zmm2/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.F2.MAP5.W0 7D /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTUW2PH_ZMM_K1Z_ZMMM512B16_ER = 4410;
	/**
	 * {@code VCVTW2PH xmm1 {k1}{z}, xmm2/m128/m16bcst}<p>
	 * {@code EVEX.128.F3.MAP5.W0 7D /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTW2PH_XMM_K1Z_XMMM128B16 = 4411;
	/**
	 * {@code VCVTW2PH ymm1 {k1}{z}, ymm2/m256/m16bcst}<p>
	 * {@code EVEX.256.F3.MAP5.W0 7D /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTW2PH_YMM_K1Z_YMMM256B16 = 4412;
	/**
	 * {@code VCVTW2PH zmm1 {k1}{z}, zmm2/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.F3.MAP5.W0 7D /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VCVTW2PH_ZMM_K1Z_ZMMM512B16_ER = 4413;
	/**
	 * {@code VDIVPH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.MAP5.W0 5E /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDIVPH_XMM_K1Z_XMM_XMMM128B16 = 4414;
	/**
	 * {@code VDIVPH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.MAP5.W0 5E /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDIVPH_YMM_K1Z_YMM_YMMM256B16 = 4415;
	/**
	 * {@code VDIVPH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.MAP5.W0 5E /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDIVPH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4416;
	/**
	 * {@code VDIVSH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 5E /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VDIVSH_XMM_K1Z_XMM_XMMM16_ER = 4417;
	/**
	 * {@code VFCMADDCPH xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.F2.MAP6.W0 56 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFCMADDCPH_XMM_K1Z_XMM_XMMM128B32 = 4418;
	/**
	 * {@code VFCMADDCPH ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.F2.MAP6.W0 56 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFCMADDCPH_YMM_K1Z_YMM_YMMM256B32 = 4419;
	/**
	 * {@code VFCMADDCPH zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.F2.MAP6.W0 56 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFCMADDCPH_ZMM_K1Z_ZMM_ZMMM512B32_ER = 4420;
	/**
	 * {@code VFMADDCPH xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.F3.MAP6.W0 56 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDCPH_XMM_K1Z_XMM_XMMM128B32 = 4421;
	/**
	 * {@code VFMADDCPH ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.F3.MAP6.W0 56 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDCPH_YMM_K1Z_YMM_YMMM256B32 = 4422;
	/**
	 * {@code VFMADDCPH zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.F3.MAP6.W0 56 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDCPH_ZMM_K1Z_ZMM_ZMMM512B32_ER = 4423;
	/**
	 * {@code VFCMADDCSH xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.F2.MAP6.W0 57 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFCMADDCSH_XMM_K1Z_XMM_XMMM32_ER = 4424;
	/**
	 * {@code VFMADDCSH xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.F3.MAP6.W0 57 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDCSH_XMM_K1Z_XMM_XMMM32_ER = 4425;
	/**
	 * {@code VFCMULCPH xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.F2.MAP6.W0 D6 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFCMULCPH_XMM_K1Z_XMM_XMMM128B32 = 4426;
	/**
	 * {@code VFCMULCPH ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.F2.MAP6.W0 D6 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFCMULCPH_YMM_K1Z_YMM_YMMM256B32 = 4427;
	/**
	 * {@code VFCMULCPH zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.F2.MAP6.W0 D6 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFCMULCPH_ZMM_K1Z_ZMM_ZMMM512B32_ER = 4428;
	/**
	 * {@code VFMULCPH xmm1 {k1}{z}, xmm2, xmm3/m128/m32bcst}<p>
	 * {@code EVEX.128.F3.MAP6.W0 D6 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMULCPH_XMM_K1Z_XMM_XMMM128B32 = 4429;
	/**
	 * {@code VFMULCPH ymm1 {k1}{z}, ymm2, ymm3/m256/m32bcst}<p>
	 * {@code EVEX.256.F3.MAP6.W0 D6 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMULCPH_YMM_K1Z_YMM_YMMM256B32 = 4430;
	/**
	 * {@code VFMULCPH zmm1 {k1}{z}, zmm2, zmm3/m512/m32bcst{er}}<p>
	 * {@code EVEX.512.F3.MAP6.W0 D6 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMULCPH_ZMM_K1Z_ZMM_ZMMM512B32_ER = 4431;
	/**
	 * {@code VFCMULCSH xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.F2.MAP6.W0 D7 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFCMULCSH_XMM_K1Z_XMM_XMMM32_ER = 4432;
	/**
	 * {@code VFMULCSH xmm1 {k1}{z}, xmm2, xmm3/m32{er}}<p>
	 * {@code EVEX.LIG.F3.MAP6.W0 D7 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMULCSH_XMM_K1Z_XMM_XMMM32_ER = 4433;
	/**
	 * {@code VFMADDSUB132PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 96 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB132PH_XMM_K1Z_XMM_XMMM128B16 = 4434;
	/**
	 * {@code VFMADDSUB132PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 96 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB132PH_YMM_K1Z_YMM_YMMM256B16 = 4435;
	/**
	 * {@code VFMADDSUB132PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 96 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB132PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4436;
	/**
	 * {@code VFMADDSUB213PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 A6 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB213PH_XMM_K1Z_XMM_XMMM128B16 = 4437;
	/**
	 * {@code VFMADDSUB213PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 A6 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB213PH_YMM_K1Z_YMM_YMMM256B16 = 4438;
	/**
	 * {@code VFMADDSUB213PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 A6 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB213PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4439;
	/**
	 * {@code VFMADDSUB231PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 B6 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB231PH_XMM_K1Z_XMM_XMMM128B16 = 4440;
	/**
	 * {@code VFMADDSUB231PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 B6 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB231PH_YMM_K1Z_YMM_YMMM256B16 = 4441;
	/**
	 * {@code VFMADDSUB231PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 B6 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADDSUB231PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4442;
	/**
	 * {@code VFMSUBADD132PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 97 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD132PH_XMM_K1Z_XMM_XMMM128B16 = 4443;
	/**
	 * {@code VFMSUBADD132PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 97 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD132PH_YMM_K1Z_YMM_YMMM256B16 = 4444;
	/**
	 * {@code VFMSUBADD132PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 97 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD132PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4445;
	/**
	 * {@code VFMSUBADD213PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 A7 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD213PH_XMM_K1Z_XMM_XMMM128B16 = 4446;
	/**
	 * {@code VFMSUBADD213PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 A7 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD213PH_YMM_K1Z_YMM_YMMM256B16 = 4447;
	/**
	 * {@code VFMSUBADD213PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 A7 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD213PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4448;
	/**
	 * {@code VFMSUBADD231PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 B7 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD231PH_XMM_K1Z_XMM_XMMM128B16 = 4449;
	/**
	 * {@code VFMSUBADD231PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 B7 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD231PH_YMM_K1Z_YMM_YMMM256B16 = 4450;
	/**
	 * {@code VFMSUBADD231PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 B7 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUBADD231PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4451;
	/**
	 * {@code VFMADD132PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 98 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD132PH_XMM_K1Z_XMM_XMMM128B16 = 4452;
	/**
	 * {@code VFMADD132PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 98 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD132PH_YMM_K1Z_YMM_YMMM256B16 = 4453;
	/**
	 * {@code VFMADD132PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 98 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD132PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4454;
	/**
	 * {@code VFMADD213PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 A8 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD213PH_XMM_K1Z_XMM_XMMM128B16 = 4455;
	/**
	 * {@code VFMADD213PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 A8 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD213PH_YMM_K1Z_YMM_YMMM256B16 = 4456;
	/**
	 * {@code VFMADD213PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 A8 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD213PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4457;
	/**
	 * {@code VFMADD231PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 B8 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD231PH_XMM_K1Z_XMM_XMMM128B16 = 4458;
	/**
	 * {@code VFMADD231PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 B8 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD231PH_YMM_K1Z_YMM_YMMM256B16 = 4459;
	/**
	 * {@code VFMADD231PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 B8 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD231PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4460;
	/**
	 * {@code VFNMADD132PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 9C /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD132PH_XMM_K1Z_XMM_XMMM128B16 = 4461;
	/**
	 * {@code VFNMADD132PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 9C /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD132PH_YMM_K1Z_YMM_YMMM256B16 = 4462;
	/**
	 * {@code VFNMADD132PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 9C /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD132PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4463;
	/**
	 * {@code VFNMADD213PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 AC /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD213PH_XMM_K1Z_XMM_XMMM128B16 = 4464;
	/**
	 * {@code VFNMADD213PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 AC /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD213PH_YMM_K1Z_YMM_YMMM256B16 = 4465;
	/**
	 * {@code VFNMADD213PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 AC /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD213PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4466;
	/**
	 * {@code VFNMADD231PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 BC /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD231PH_XMM_K1Z_XMM_XMMM128B16 = 4467;
	/**
	 * {@code VFNMADD231PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 BC /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD231PH_YMM_K1Z_YMM_YMMM256B16 = 4468;
	/**
	 * {@code VFNMADD231PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 BC /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD231PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4469;
	/**
	 * {@code VFMADD132SH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 99 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD132SH_XMM_K1Z_XMM_XMMM16_ER = 4470;
	/**
	 * {@code VFMADD213SH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 A9 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD213SH_XMM_K1Z_XMM_XMMM16_ER = 4471;
	/**
	 * {@code VFMADD231SH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 B9 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMADD231SH_XMM_K1Z_XMM_XMMM16_ER = 4472;
	/**
	 * {@code VFNMADD132SH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 9D /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD132SH_XMM_K1Z_XMM_XMMM16_ER = 4473;
	/**
	 * {@code VFNMADD213SH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 AD /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD213SH_XMM_K1Z_XMM_XMMM16_ER = 4474;
	/**
	 * {@code VFNMADD231SH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 BD /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMADD231SH_XMM_K1Z_XMM_XMMM16_ER = 4475;
	/**
	 * {@code VFMSUB132PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 9A /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB132PH_XMM_K1Z_XMM_XMMM128B16 = 4476;
	/**
	 * {@code VFMSUB132PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 9A /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB132PH_YMM_K1Z_YMM_YMMM256B16 = 4477;
	/**
	 * {@code VFMSUB132PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 9A /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB132PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4478;
	/**
	 * {@code VFMSUB213PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 AA /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB213PH_XMM_K1Z_XMM_XMMM128B16 = 4479;
	/**
	 * {@code VFMSUB213PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 AA /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB213PH_YMM_K1Z_YMM_YMMM256B16 = 4480;
	/**
	 * {@code VFMSUB213PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 AA /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB213PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4481;
	/**
	 * {@code VFMSUB231PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 BA /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB231PH_XMM_K1Z_XMM_XMMM128B16 = 4482;
	/**
	 * {@code VFMSUB231PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 BA /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB231PH_YMM_K1Z_YMM_YMMM256B16 = 4483;
	/**
	 * {@code VFMSUB231PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 BA /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB231PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4484;
	/**
	 * {@code VFNMSUB132PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 9E /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB132PH_XMM_K1Z_XMM_XMMM128B16 = 4485;
	/**
	 * {@code VFNMSUB132PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 9E /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB132PH_YMM_K1Z_YMM_YMMM256B16 = 4486;
	/**
	 * {@code VFNMSUB132PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 9E /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB132PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4487;
	/**
	 * {@code VFNMSUB213PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 AE /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB213PH_XMM_K1Z_XMM_XMMM128B16 = 4488;
	/**
	 * {@code VFNMSUB213PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 AE /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB213PH_YMM_K1Z_YMM_YMMM256B16 = 4489;
	/**
	 * {@code VFNMSUB213PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 AE /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB213PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4490;
	/**
	 * {@code VFNMSUB231PH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 BE /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB231PH_XMM_K1Z_XMM_XMMM128B16 = 4491;
	/**
	 * {@code VFNMSUB231PH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 BE /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB231PH_YMM_K1Z_YMM_YMMM256B16 = 4492;
	/**
	 * {@code VFNMSUB231PH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 BE /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB231PH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4493;
	/**
	 * {@code VFMSUB132SH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 9B /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB132SH_XMM_K1Z_XMM_XMMM16_ER = 4494;
	/**
	 * {@code VFMSUB213SH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 AB /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB213SH_XMM_K1Z_XMM_XMMM16_ER = 4495;
	/**
	 * {@code VFMSUB231SH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 BB /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFMSUB231SH_XMM_K1Z_XMM_XMMM16_ER = 4496;
	/**
	 * {@code VFNMSUB132SH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 9F /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB132SH_XMM_K1Z_XMM_XMMM16_ER = 4497;
	/**
	 * {@code VFNMSUB213SH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 AF /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB213SH_XMM_K1Z_XMM_XMMM16_ER = 4498;
	/**
	 * {@code VFNMSUB231SH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 BF /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFNMSUB231SH_XMM_K1Z_XMM_XMMM16_ER = 4499;
	/**
	 * {@code VFPCLASSPH k1 {k2}, xmm2/m128/m16bcst, imm8}<p>
	 * {@code EVEX.128.0F3A.W0 66 /r ib}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFPCLASSPH_KR_K1_XMMM128B16_IMM8 = 4500;
	/**
	 * {@code VFPCLASSPH k1 {k2}, ymm2/m256/m16bcst, imm8}<p>
	 * {@code EVEX.256.0F3A.W0 66 /r ib}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFPCLASSPH_KR_K1_YMMM256B16_IMM8 = 4501;
	/**
	 * {@code VFPCLASSPH k1 {k2}, zmm2/m512/m16bcst, imm8}<p>
	 * {@code EVEX.512.0F3A.W0 66 /r ib}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFPCLASSPH_KR_K1_ZMMM512B16_IMM8 = 4502;
	/**
	 * {@code VFPCLASSSH k1 {k2}, xmm2/m16, imm8}<p>
	 * {@code EVEX.LIG.0F3A.W0 67 /r ib}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VFPCLASSSH_KR_K1_XMMM16_IMM8 = 4503;
	/**
	 * {@code VGETEXPPH xmm1 {k1}{z}, xmm2/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 42 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETEXPPH_XMM_K1Z_XMMM128B16 = 4504;
	/**
	 * {@code VGETEXPPH ymm1 {k1}{z}, ymm2/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 42 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETEXPPH_YMM_K1Z_YMMM256B16 = 4505;
	/**
	 * {@code VGETEXPPH zmm1 {k1}{z}, zmm2/m512/m16bcst{sae}}<p>
	 * {@code EVEX.512.66.MAP6.W0 42 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETEXPPH_ZMM_K1Z_ZMMM512B16_SAE = 4506;
	/**
	 * {@code VGETEXPSH xmm1 {k1}{z}, xmm2, xmm3/m16{sae}}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 43 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETEXPSH_XMM_K1Z_XMM_XMMM16_SAE = 4507;
	/**
	 * {@code VGETMANTPH xmm1 {k1}{z}, xmm2/m128/m16bcst, imm8}<p>
	 * {@code EVEX.128.0F3A.W0 26 /r ib}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETMANTPH_XMM_K1Z_XMMM128B16_IMM8 = 4508;
	/**
	 * {@code VGETMANTPH ymm1 {k1}{z}, ymm2/m256/m16bcst, imm8}<p>
	 * {@code EVEX.256.0F3A.W0 26 /r ib}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETMANTPH_YMM_K1Z_YMMM256B16_IMM8 = 4509;
	/**
	 * {@code VGETMANTPH zmm1 {k1}{z}, zmm2/m512/m16bcst{sae}, imm8}<p>
	 * {@code EVEX.512.0F3A.W0 26 /r ib}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETMANTPH_ZMM_K1Z_ZMMM512B16_IMM8_SAE = 4510;
	/**
	 * {@code VGETMANTSH xmm1 {k1}{z}, xmm2, xmm3/m16{sae}, imm8}<p>
	 * {@code EVEX.LIG.0F3A.W0 27 /r ib}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VGETMANTSH_XMM_K1Z_XMM_XMMM16_IMM8_SAE = 4511;
	/**
	 * {@code VMAXPH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.MAP5.W0 5F /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMAXPH_XMM_K1Z_XMM_XMMM128B16 = 4512;
	/**
	 * {@code VMAXPH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.MAP5.W0 5F /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMAXPH_YMM_K1Z_YMM_YMMM256B16 = 4513;
	/**
	 * {@code VMAXPH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{sae}}<p>
	 * {@code EVEX.512.MAP5.W0 5F /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMAXPH_ZMM_K1Z_ZMM_ZMMM512B16_SAE = 4514;
	/**
	 * {@code VMAXSH xmm1 {k1}{z}, xmm2, xmm3/m16{sae}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 5F /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMAXSH_XMM_K1Z_XMM_XMMM16_SAE = 4515;
	/**
	 * {@code VMINPH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.MAP5.W0 5D /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMINPH_XMM_K1Z_XMM_XMMM128B16 = 4516;
	/**
	 * {@code VMINPH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.MAP5.W0 5D /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMINPH_YMM_K1Z_YMM_YMMM256B16 = 4517;
	/**
	 * {@code VMINPH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{sae}}<p>
	 * {@code EVEX.512.MAP5.W0 5D /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMINPH_ZMM_K1Z_ZMM_ZMMM512B16_SAE = 4518;
	/**
	 * {@code VMINSH xmm1 {k1}{z}, xmm2, xmm3/m16{sae}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 5D /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMINSH_XMM_K1Z_XMM_XMMM16_SAE = 4519;
	/**
	 * {@code VMOVSH xmm1 {k1}{z}, m16}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 10 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSH_XMM_K1Z_M16 = 4520;
	/**
	 * {@code VMOVSH m16 {k1}, xmm1}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 11 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSH_M16_K1_XMM = 4521;
	/**
	 * {@code VMOVSH xmm1 {k1}{z}, xmm2, xmm3}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 10 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSH_XMM_K1Z_XMM_XMM = 4522;
	/**
	 * {@code VMOVSH xmm1 {k1}{z}, xmm2, xmm3}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 11 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVSH_XMM_K1Z_XMM_XMM_MAP5_11 = 4523;
	/**
	 * {@code VMOVW xmm1, r32/m16}<p>
	 * {@code EVEX.128.66.MAP5.W0 6E /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVW_XMM_R32M16 = 4524;
	/**
	 * {@code VMOVW xmm1, r64/m16}<p>
	 * {@code EVEX.128.66.MAP5.W1 6E /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VMOVW_XMM_R64M16 = 4525;
	/**
	 * {@code VMOVW r32/m16, xmm1}<p>
	 * {@code EVEX.128.66.MAP5.W0 7E /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMOVW_R32M16_XMM = 4526;
	/**
	 * {@code VMOVW r64/m16, xmm1}<p>
	 * {@code EVEX.128.66.MAP5.W1 7E /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 64-bit}
	 */
	public static final int EVEX_VMOVW_R64M16_XMM = 4527;
	/**
	 * {@code VMULPH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.MAP5.W0 59 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMULPH_XMM_K1Z_XMM_XMMM128B16 = 4528;
	/**
	 * {@code VMULPH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.MAP5.W0 59 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMULPH_YMM_K1Z_YMM_YMMM256B16 = 4529;
	/**
	 * {@code VMULPH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.MAP5.W0 59 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMULPH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4530;
	/**
	 * {@code VMULSH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 59 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VMULSH_XMM_K1Z_XMM_XMMM16_ER = 4531;
	/**
	 * {@code VRCPPH xmm1 {k1}{z}, xmm2/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 4C /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCPPH_XMM_K1Z_XMMM128B16 = 4532;
	/**
	 * {@code VRCPPH ymm1 {k1}{z}, ymm2/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 4C /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCPPH_YMM_K1Z_YMMM256B16 = 4533;
	/**
	 * {@code VRCPPH zmm1 {k1}{z}, zmm2/m512/m16bcst}<p>
	 * {@code EVEX.512.66.MAP6.W0 4C /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCPPH_ZMM_K1Z_ZMMM512B16 = 4534;
	/**
	 * {@code VRCPSH xmm1 {k1}{z}, xmm2, xmm3/m16}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 4D /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRCPSH_XMM_K1Z_XMM_XMMM16 = 4535;
	/**
	 * {@code VREDUCEPH xmm1 {k1}{z}, xmm2/m128/m16bcst, imm8}<p>
	 * {@code EVEX.128.0F3A.W0 56 /r ib}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VREDUCEPH_XMM_K1Z_XMMM128B16_IMM8 = 4536;
	/**
	 * {@code VREDUCEPH ymm1 {k1}{z}, ymm2/m256/m16bcst, imm8}<p>
	 * {@code EVEX.256.0F3A.W0 56 /r ib}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VREDUCEPH_YMM_K1Z_YMMM256B16_IMM8 = 4537;
	/**
	 * {@code VREDUCEPH zmm1 {k1}{z}, zmm2/m512/m16bcst{sae}, imm8}<p>
	 * {@code EVEX.512.0F3A.W0 56 /r ib}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VREDUCEPH_ZMM_K1Z_ZMMM512B16_IMM8_SAE = 4538;
	/**
	 * {@code VREDUCESH xmm1 {k1}{z}, xmm2, xmm3/m16{sae}, imm8}<p>
	 * {@code EVEX.LIG.0F3A.W0 57 /r ib}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VREDUCESH_XMM_K1Z_XMM_XMMM16_IMM8_SAE = 4539;
	/**
	 * {@code VRNDSCALEPH xmm1 {k1}{z}, xmm2/m128/m16bcst, imm8}<p>
	 * {@code EVEX.128.0F3A.W0 08 /r ib}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRNDSCALEPH_XMM_K1Z_XMMM128B16_IMM8 = 4540;
	/**
	 * {@code VRNDSCALEPH ymm1 {k1}{z}, ymm2/m256/m16bcst, imm8}<p>
	 * {@code EVEX.256.0F3A.W0 08 /r ib}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRNDSCALEPH_YMM_K1Z_YMMM256B16_IMM8 = 4541;
	/**
	 * {@code VRNDSCALEPH zmm1 {k1}{z}, zmm2/m512/m16bcst{sae}, imm8}<p>
	 * {@code EVEX.512.0F3A.W0 08 /r ib}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRNDSCALEPH_ZMM_K1Z_ZMMM512B16_IMM8_SAE = 4542;
	/**
	 * {@code VRNDSCALESH xmm1 {k1}{z}, xmm2, xmm3/m16{sae}, imm8}<p>
	 * {@code EVEX.LIG.0F3A.W0 0A /r ib}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRNDSCALESH_XMM_K1Z_XMM_XMMM16_IMM8_SAE = 4543;
	/**
	 * {@code VRSQRTPH xmm1 {k1}{z}, xmm2/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 4E /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRTPH_XMM_K1Z_XMMM128B16 = 4544;
	/**
	 * {@code VRSQRTPH ymm1 {k1}{z}, ymm2/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 4E /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRTPH_YMM_K1Z_YMMM256B16 = 4545;
	/**
	 * {@code VRSQRTPH zmm1 {k1}{z}, zmm2/m512/m16bcst}<p>
	 * {@code EVEX.512.66.MAP6.W0 4E /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRTPH_ZMM_K1Z_ZMMM512B16 = 4546;
	/**
	 * {@code VRSQRTSH xmm1 {k1}{z}, xmm2, xmm3/m16}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 4F /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VRSQRTSH_XMM_K1Z_XMM_XMMM16 = 4547;
	/**
	 * {@code VSCALEFPH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.66.MAP6.W0 2C /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCALEFPH_XMM_K1Z_XMM_XMMM128B16 = 4548;
	/**
	 * {@code VSCALEFPH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.66.MAP6.W0 2C /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCALEFPH_YMM_K1Z_YMM_YMMM256B16 = 4549;
	/**
	 * {@code VSCALEFPH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.66.MAP6.W0 2C /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCALEFPH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4550;
	/**
	 * {@code VSCALEFSH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.66.MAP6.W0 2D /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSCALEFSH_XMM_K1Z_XMM_XMMM16_ER = 4551;
	/**
	 * {@code VSQRTPH xmm1 {k1}{z}, xmm2/m128/m16bcst}<p>
	 * {@code EVEX.128.MAP5.W0 51 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSQRTPH_XMM_K1Z_XMMM128B16 = 4552;
	/**
	 * {@code VSQRTPH ymm1 {k1}{z}, ymm2/m256/m16bcst}<p>
	 * {@code EVEX.256.MAP5.W0 51 /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSQRTPH_YMM_K1Z_YMMM256B16 = 4553;
	/**
	 * {@code VSQRTPH zmm1 {k1}{z}, zmm2/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.MAP5.W0 51 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSQRTPH_ZMM_K1Z_ZMMM512B16_ER = 4554;
	/**
	 * {@code VSQRTSH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 51 /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSQRTSH_XMM_K1Z_XMM_XMMM16_ER = 4555;
	/**
	 * {@code VSUBPH xmm1 {k1}{z}, xmm2, xmm3/m128/m16bcst}<p>
	 * {@code EVEX.128.MAP5.W0 5C /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSUBPH_XMM_K1Z_XMM_XMMM128B16 = 4556;
	/**
	 * {@code VSUBPH ymm1 {k1}{z}, ymm2, ymm3/m256/m16bcst}<p>
	 * {@code EVEX.256.MAP5.W0 5C /r}<p>
	 * {@code AVX512VL and AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSUBPH_YMM_K1Z_YMM_YMMM256B16 = 4557;
	/**
	 * {@code VSUBPH zmm1 {k1}{z}, zmm2, zmm3/m512/m16bcst{er}}<p>
	 * {@code EVEX.512.MAP5.W0 5C /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSUBPH_ZMM_K1Z_ZMM_ZMMM512B16_ER = 4558;
	/**
	 * {@code VSUBSH xmm1 {k1}{z}, xmm2, xmm3/m16{er}}<p>
	 * {@code EVEX.LIG.F3.MAP5.W0 5C /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VSUBSH_XMM_K1Z_XMM_XMMM16_ER = 4559;
	/**
	 * {@code VUCOMISH xmm1, xmm2/m16{sae}}<p>
	 * {@code EVEX.LIG.MAP5.W0 2E /r}<p>
	 * {@code AVX512-FP16}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int EVEX_VUCOMISH_XMM_XMMM16_SAE = 4560;
	/**
	 * {@code RDUDBG}<p>
	 * {@code 0F 0E}<p>
	 * {@code UDBG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int RDUDBG = 4561;
	/**
	 * {@code WRUDBG}<p>
	 * {@code 0F 0F}<p>
	 * {@code UDBG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int WRUDBG = 4562;
	/**
	 * {@code JKZD k1, rel8}<p>
	 * {@code VEX.128.W0 74 cb}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_JKZD_KR_REL8_64 = 4563;
	/**
	 * {@code JKNZD k1, rel8}<p>
	 * {@code VEX.128.W0 75 cb}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_JKNZD_KR_REL8_64 = 4564;
	/**
	 * {@code VPREFETCHNTA m8}<p>
	 * {@code VEX.128.0F.WIG 18 /0}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_VPREFETCHNTA_M8 = 4565;
	/**
	 * {@code VPREFETCH0 m8}<p>
	 * {@code VEX.128.0F.WIG 18 /1}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_VPREFETCH0_M8 = 4566;
	/**
	 * {@code VPREFETCH1 m8}<p>
	 * {@code VEX.128.0F.WIG 18 /2}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_VPREFETCH1_M8 = 4567;
	/**
	 * {@code VPREFETCH2 m8}<p>
	 * {@code VEX.128.0F.WIG 18 /3}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_VPREFETCH2_M8 = 4568;
	/**
	 * {@code VPREFETCHENTA m8}<p>
	 * {@code VEX.128.0F.WIG 18 /4}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_VPREFETCHENTA_M8 = 4569;
	/**
	 * {@code VPREFETCHE0 m8}<p>
	 * {@code VEX.128.0F.WIG 18 /5}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_VPREFETCHE0_M8 = 4570;
	/**
	 * {@code VPREFETCHE1 m8}<p>
	 * {@code VEX.128.0F.WIG 18 /6}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_VPREFETCHE1_M8 = 4571;
	/**
	 * {@code VPREFETCHE2 m8}<p>
	 * {@code VEX.128.0F.WIG 18 /7}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_VPREFETCHE2_M8 = 4572;
	/**
	 * {@code KAND k1, k2}<p>
	 * {@code VEX.128.0F.W0 41 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KAND_KR_KR = 4573;
	/**
	 * {@code KANDN k1, k2}<p>
	 * {@code VEX.128.0F.W0 42 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KANDN_KR_KR = 4574;
	/**
	 * {@code KANDNR k1, k2}<p>
	 * {@code VEX.128.0F.W0 43 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KANDNR_KR_KR = 4575;
	/**
	 * {@code KNOT k1, k2}<p>
	 * {@code VEX.128.0F.W0 44 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KNOT_KR_KR = 4576;
	/**
	 * {@code KOR k1, k2}<p>
	 * {@code VEX.128.0F.W0 45 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KOR_KR_KR = 4577;
	/**
	 * {@code KXNOR k1, k2}<p>
	 * {@code VEX.128.0F.W0 46 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KXNOR_KR_KR = 4578;
	/**
	 * {@code KXOR k1, k2}<p>
	 * {@code VEX.128.0F.W0 47 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KXOR_KR_KR = 4579;
	/**
	 * {@code KMERGE2L1H k1, k2}<p>
	 * {@code VEX.128.0F.W0 48 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KMERGE2L1H_KR_KR = 4580;
	/**
	 * {@code KMERGE2L1L k1, k2}<p>
	 * {@code VEX.128.0F.W0 49 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KMERGE2L1L_KR_KR = 4581;
	/**
	 * {@code JKZD k1, rel32}<p>
	 * {@code VEX.128.0F.W0 84 cd}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_JKZD_KR_REL32_64 = 4582;
	/**
	 * {@code JKNZD k1, rel32}<p>
	 * {@code VEX.128.0F.W0 85 cd}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_JKNZD_KR_REL32_64 = 4583;
	/**
	 * {@code KMOV k1, k2}<p>
	 * {@code VEX.128.0F.W0 90 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KMOV_KR_KR = 4584;
	/**
	 * {@code KMOV k1, r32}<p>
	 * {@code VEX.128.0F.W0 92 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KMOV_KR_R32 = 4585;
	/**
	 * {@code KMOV r32, k1}<p>
	 * {@code VEX.128.0F.W0 93 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KMOV_R32_KR = 4586;
	/**
	 * {@code KCONCATH r64, k1, k2}<p>
	 * {@code VEX.128.0F.W0 95 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KCONCATH_R64_KR_KR = 4587;
	/**
	 * {@code KCONCATL r64, k1, k2}<p>
	 * {@code VEX.128.0F.W0 97 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KCONCATL_R64_KR_KR = 4588;
	/**
	 * {@code KORTEST k1, k2}<p>
	 * {@code VEX.128.0F.W0 98 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KORTEST_KR_KR = 4589;
	/**
	 * {@code DELAY r32}<p>
	 * {@code VEX.128.F3.0F.W0 AE /6}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_DELAY_R32 = 4590;
	/**
	 * {@code DELAY r64}<p>
	 * {@code VEX.128.F3.0F.W1 AE /6}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_DELAY_R64 = 4591;
	/**
	 * {@code SPFLT r32}<p>
	 * {@code VEX.128.F2.0F.W0 AE /6}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_SPFLT_R32 = 4592;
	/**
	 * {@code SPFLT r64}<p>
	 * {@code VEX.128.F2.0F.W1 AE /6}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_SPFLT_R64 = 4593;
	/**
	 * {@code CLEVICT1 m8}<p>
	 * {@code VEX.128.F3.0F.WIG AE /7}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_CLEVICT1_M8 = 4594;
	/**
	 * {@code CLEVICT0 m8}<p>
	 * {@code VEX.128.F2.0F.WIG AE /7}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_CLEVICT0_M8 = 4595;
	/**
	 * {@code POPCNT r32, r32}<p>
	 * {@code VEX.128.F3.0F.W0 B8 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_POPCNT_R32_R32 = 4596;
	/**
	 * {@code POPCNT r64, r64}<p>
	 * {@code VEX.128.F3.0F.W1 B8 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_POPCNT_R64_R64 = 4597;
	/**
	 * {@code TZCNT r32, r32}<p>
	 * {@code VEX.128.F3.0F.W0 BC /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_TZCNT_R32_R32 = 4598;
	/**
	 * {@code TZCNT r64, r64}<p>
	 * {@code VEX.128.F3.0F.W1 BC /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_TZCNT_R64_R64 = 4599;
	/**
	 * {@code TZCNTI r32, r32}<p>
	 * {@code VEX.128.F2.0F.W0 BC /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_TZCNTI_R32_R32 = 4600;
	/**
	 * {@code TZCNTI r64, r64}<p>
	 * {@code VEX.128.F2.0F.W1 BC /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_TZCNTI_R64_R64 = 4601;
	/**
	 * {@code LZCNT r32, r32}<p>
	 * {@code VEX.128.F3.0F.W0 BD /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_LZCNT_R32_R32 = 4602;
	/**
	 * {@code LZCNT r64, r64}<p>
	 * {@code VEX.128.F3.0F.W1 BD /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_LZCNT_R64_R64 = 4603;
	/**
	 * {@code UNDOC r32, r/m32}<p>
	 * {@code VEX.128.F3.0F38.W0 F0 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_UNDOC_R32_RM32_128_F3_0F38_W0_F0 = 4604;
	/**
	 * {@code UNDOC r64, r/m64}<p>
	 * {@code VEX.128.F3.0F38.W1 F0 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_UNDOC_R64_RM64_128_F3_0F38_W1_F0 = 4605;
	/**
	 * {@code UNDOC r32, r/m32}<p>
	 * {@code VEX.128.F2.0F38.W0 F0 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_UNDOC_R32_RM32_128_F2_0F38_W0_F0 = 4606;
	/**
	 * {@code UNDOC r64, r/m64}<p>
	 * {@code VEX.128.F2.0F38.W1 F0 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_UNDOC_R64_RM64_128_F2_0F38_W1_F0 = 4607;
	/**
	 * {@code UNDOC r32, r/m32}<p>
	 * {@code VEX.128.F2.0F38.W0 F1 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_UNDOC_R32_RM32_128_F2_0F38_W0_F1 = 4608;
	/**
	 * {@code UNDOC r64, r/m64}<p>
	 * {@code VEX.128.F2.0F38.W1 F1 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_UNDOC_R64_RM64_128_F2_0F38_W1_F1 = 4609;
	/**
	 * {@code KEXTRACT k1, r64, imm8}<p>
	 * {@code VEX.128.66.0F3A.W0 3E /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_KNC_KEXTRACT_KR_R64_IMM8 = 4610;
	/**
	 * {@code VPREFETCHNTA m}<p>
	 * {@code MVEX.512.0F.WIG 18 /0}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPREFETCHNTA_M = 4611;
	/**
	 * {@code VPREFETCH0 m}<p>
	 * {@code MVEX.512.0F.WIG 18 /1}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPREFETCH0_M = 4612;
	/**
	 * {@code VPREFETCH1 m}<p>
	 * {@code MVEX.512.0F.WIG 18 /2}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPREFETCH1_M = 4613;
	/**
	 * {@code VPREFETCH2 m}<p>
	 * {@code MVEX.512.0F.WIG 18 /3}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPREFETCH2_M = 4614;
	/**
	 * {@code VPREFETCHENTA m}<p>
	 * {@code MVEX.512.0F.WIG 18 /4}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPREFETCHENTA_M = 4615;
	/**
	 * {@code VPREFETCHE0 m}<p>
	 * {@code MVEX.512.0F.WIG 18 /5}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPREFETCHE0_M = 4616;
	/**
	 * {@code VPREFETCHE1 m}<p>
	 * {@code MVEX.512.0F.WIG 18 /6}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPREFETCHE1_M = 4617;
	/**
	 * {@code VPREFETCHE2 m}<p>
	 * {@code MVEX.512.0F.WIG 18 /7}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPREFETCHE2_M = 4618;
	/**
	 * {@code VMOVAPS zmm1 {k1}, Sf32(zmm2/mt)}<p>
	 * {@code MVEX.512.0F.W0 28 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VMOVAPS_ZMM_K1_ZMMMT = 4619;
	/**
	 * {@code VMOVAPD zmm1 {k1}, Sf64(zmm2/mt)}<p>
	 * {@code MVEX.512.66.0F.W1 28 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VMOVAPD_ZMM_K1_ZMMMT = 4620;
	/**
	 * {@code VMOVAPS mt {k1}, Df32(zmm1)}<p>
	 * {@code MVEX.512.0F.W0 29 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VMOVAPS_MT_K1_ZMM = 4621;
	/**
	 * {@code VMOVAPD mt {k1}, Df64(zmm1)}<p>
	 * {@code MVEX.512.66.0F.W1 29 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VMOVAPD_MT_K1_ZMM = 4622;
	/**
	 * {@code VMOVNRAPD m {k1}, Df64(zmm1)}<p>
	 * {@code MVEX.512.F3.0F.W1.EH0 29 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VMOVNRAPD_M_K1_ZMM = 4623;
	/**
	 * {@code VMOVNRNGOAPD m {k1}, Df64(zmm1)}<p>
	 * {@code MVEX.512.F3.0F.W1.EH1 29 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VMOVNRNGOAPD_M_K1_ZMM = 4624;
	/**
	 * {@code VMOVNRAPS m {k1}, Df32(zmm1)}<p>
	 * {@code MVEX.512.F2.0F.W0.EH0 29 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VMOVNRAPS_M_K1_ZMM = 4625;
	/**
	 * {@code VMOVNRNGOAPS m {k1}, Df32(zmm1)}<p>
	 * {@code MVEX.512.F2.0F.W0.EH1 29 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VMOVNRNGOAPS_M_K1_ZMM = 4626;
	/**
	 * {@code VADDPS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.0F.W0 58 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VADDPS_ZMM_K1_ZMM_ZMMMT = 4627;
	/**
	 * {@code VADDPD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F.W1 58 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VADDPD_ZMM_K1_ZMM_ZMMMT = 4628;
	/**
	 * {@code VMULPS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.0F.W0 59 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VMULPS_ZMM_K1_ZMM_ZMMMT = 4629;
	/**
	 * {@code VMULPD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F.W1 59 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VMULPD_ZMM_K1_ZMM_ZMMMT = 4630;
	/**
	 * {@code VCVTPS2PD zmm1 {k1}, Sf32(zmm2/mt)}<p>
	 * {@code MVEX.512.0F.W0 5A /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VCVTPS2PD_ZMM_K1_ZMMMT = 4631;
	/**
	 * {@code VCVTPD2PS zmm1 {k1}, Sf64(zmm2/mt)}<p>
	 * {@code MVEX.512.66.0F.W1 5A /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VCVTPD2PS_ZMM_K1_ZMMMT = 4632;
	/**
	 * {@code VSUBPS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.0F.W0 5C /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VSUBPS_ZMM_K1_ZMM_ZMMMT = 4633;
	/**
	 * {@code VSUBPD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F.W1 5C /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VSUBPD_ZMM_K1_ZMM_ZMMMT = 4634;
	/**
	 * {@code VPCMPGTD k2 {k1}, zmm1, Si32(zmm2/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F.W0 66 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPCMPGTD_KR_K1_ZMM_ZMMMT = 4635;
	/**
	 * {@code VMOVDQA32 zmm1 {k1}, Si32(zmm2/mt)}<p>
	 * {@code MVEX.512.66.0F.W0 6F /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VMOVDQA32_ZMM_K1_ZMMMT = 4636;
	/**
	 * {@code VMOVDQA64 zmm1 {k1}, Si64(zmm2/mt)}<p>
	 * {@code MVEX.512.66.0F.W1 6F /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VMOVDQA64_ZMM_K1_ZMMMT = 4637;
	/**
	 * {@code VPSHUFD zmm1 {k1}, zmm2/mt, imm8}<p>
	 * {@code MVEX.512.66.0F.W0 70 /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPSHUFD_ZMM_K1_ZMMMT_IMM8 = 4638;
	/**
	 * {@code VPSRLD zmm1 {k1}, Si32(zmm2/mt), imm8}<p>
	 * {@code MVEX.NDD.512.66.0F.W0 72 /2 ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPSRLD_ZMM_K1_ZMMMT_IMM8 = 4639;
	/**
	 * {@code VPSRAD zmm1 {k1}, Si32(zmm2/mt), imm8}<p>
	 * {@code MVEX.NDD.512.66.0F.W0 72 /4 ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPSRAD_ZMM_K1_ZMMMT_IMM8 = 4640;
	/**
	 * {@code VPSLLD zmm1 {k1}, Si32(zmm2/mt), imm8}<p>
	 * {@code MVEX.NDD.512.66.0F.W0 72 /6 ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPSLLD_ZMM_K1_ZMMMT_IMM8 = 4641;
	/**
	 * {@code VPCMPEQD k2 {k1}, zmm1, Si32(zmm2/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F.W0 76 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPCMPEQD_KR_K1_ZMM_ZMMMT = 4642;
	/**
	 * {@code VCVTUDQ2PD zmm1 {k1}, Si32(zmm2/mt)}<p>
	 * {@code MVEX.512.F3.0F.W0 7A /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VCVTUDQ2PD_ZMM_K1_ZMMMT = 4643;
	/**
	 * {@code VMOVDQA32 mt {k1}, Di32(zmm1)}<p>
	 * {@code MVEX.512.66.0F.W0 7F /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VMOVDQA32_MT_K1_ZMM = 4644;
	/**
	 * {@code VMOVDQA64 mt {k1}, Di64(zmm1)}<p>
	 * {@code MVEX.512.66.0F.W1 7F /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VMOVDQA64_MT_K1_ZMM = 4645;
	/**
	 * {@code CLEVICT1 m}<p>
	 * {@code MVEX.512.F3.0F.WIG AE /7}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_CLEVICT1_M = 4646;
	/**
	 * {@code CLEVICT0 m}<p>
	 * {@code MVEX.512.F2.0F.WIG AE /7}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_CLEVICT0_M = 4647;
	/**
	 * {@code VCMPPS k2 {k1}, zmm1, Sf32(zmm2/mt), imm8}<p>
	 * {@code MVEX.NDS.512.0F.W0 C2 /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VCMPPS_KR_K1_ZMM_ZMMMT_IMM8 = 4648;
	/**
	 * {@code VCMPPD k2 {k1}, zmm1, Sf64(zmm2/mt), imm8}<p>
	 * {@code MVEX.NDS.512.66.0F.W1 C2 /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VCMPPD_KR_K1_ZMM_ZMMMT_IMM8 = 4649;
	/**
	 * {@code VPANDD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F.W0 DB /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPANDD_ZMM_K1_ZMM_ZMMMT = 4650;
	/**
	 * {@code VPANDQ zmm1 {k1}, zmm2, Si64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F.W1 DB /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPANDQ_ZMM_K1_ZMM_ZMMMT = 4651;
	/**
	 * {@code VPANDND zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F.W0 DF /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPANDND_ZMM_K1_ZMM_ZMMMT = 4652;
	/**
	 * {@code VPANDNQ zmm1 {k1}, zmm2, Si64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F.W1 DF /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPANDNQ_ZMM_K1_ZMM_ZMMMT = 4653;
	/**
	 * {@code VCVTDQ2PD zmm1 {k1}, Si32(zmm2/mt)}<p>
	 * {@code MVEX.512.F3.0F.W0 E6 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VCVTDQ2PD_ZMM_K1_ZMMMT = 4654;
	/**
	 * {@code VPORD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F.W0 EB /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPORD_ZMM_K1_ZMM_ZMMMT = 4655;
	/**
	 * {@code VPORQ zmm1 {k1}, zmm2, Si64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F.W1 EB /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPORQ_ZMM_K1_ZMM_ZMMMT = 4656;
	/**
	 * {@code VPXORD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F.W0 EF /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPXORD_ZMM_K1_ZMM_ZMMMT = 4657;
	/**
	 * {@code VPXORQ zmm1 {k1}, zmm2, Si64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F.W1 EF /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPXORQ_ZMM_K1_ZMM_ZMMMT = 4658;
	/**
	 * {@code VPSUBD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F.W0 FA /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPSUBD_ZMM_K1_ZMM_ZMMMT = 4659;
	/**
	 * {@code VPADDD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F.W0 FE /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPADDD_ZMM_K1_ZMM_ZMMMT = 4660;
	/**
	 * {@code VBROADCASTSS zmm1 {k1}, Uf32(mt)}<p>
	 * {@code MVEX.512.66.0F38.W0 18 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VBROADCASTSS_ZMM_K1_MT = 4661;
	/**
	 * {@code VBROADCASTSD zmm1 {k1}, Uf64(mt)}<p>
	 * {@code MVEX.512.66.0F38.W1 19 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VBROADCASTSD_ZMM_K1_MT = 4662;
	/**
	 * {@code VBROADCASTF32X4 zmm1 {k1}, Uf32(mt)}<p>
	 * {@code MVEX.512.66.0F38.W0 1A /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VBROADCASTF32X4_ZMM_K1_MT = 4663;
	/**
	 * {@code VBROADCASTF64X4 zmm1 {k1}, Uf64(mt)}<p>
	 * {@code MVEX.512.66.0F38.W1 1B /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VBROADCASTF64X4_ZMM_K1_MT = 4664;
	/**
	 * {@code VPTESTMD k2 {k1}, zmm1, Si32(zmm2/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 27 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPTESTMD_KR_K1_ZMM_ZMMMT = 4665;
	/**
	 * {@code VPERMD zmm1 {k1}, zmm2, zmm3/mt}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 36 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPERMD_ZMM_K1_ZMM_ZMMMT = 4666;
	/**
	 * {@code VPMINSD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 39 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPMINSD_ZMM_K1_ZMM_ZMMMT = 4667;
	/**
	 * {@code VPMINUD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 3B /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPMINUD_ZMM_K1_ZMM_ZMMMT = 4668;
	/**
	 * {@code VPMAXSD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 3D /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPMAXSD_ZMM_K1_ZMM_ZMMMT = 4669;
	/**
	 * {@code VPMAXUD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 3F /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPMAXUD_ZMM_K1_ZMM_ZMMMT = 4670;
	/**
	 * {@code VPMULLD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 40 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPMULLD_ZMM_K1_ZMM_ZMMMT = 4671;
	/**
	 * {@code VGETEXPPS zmm1 {k1}, Sf32(zmm2/mt)}<p>
	 * {@code MVEX.512.66.0F38.W0 42 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VGETEXPPS_ZMM_K1_ZMMMT = 4672;
	/**
	 * {@code VGETEXPPD zmm1 {k1}, Sf64(zmm2/mt)}<p>
	 * {@code MVEX.512.66.0F38.W1 42 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VGETEXPPD_ZMM_K1_ZMMMT = 4673;
	/**
	 * {@code VPSRLVD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 45 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPSRLVD_ZMM_K1_ZMM_ZMMMT = 4674;
	/**
	 * {@code VPSRAVD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 46 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPSRAVD_ZMM_K1_ZMM_ZMMMT = 4675;
	/**
	 * {@code VPSLLVD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 47 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPSLLVD_ZMM_K1_ZMM_ZMMMT = 4676;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.66.0F38.W0 48 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_512_66_0F38_W0_48 = 4677;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.66.0F38.W0 49 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_512_66_0F38_W0_49 = 4678;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.66.0F38.W0 4A /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_512_66_0F38_W0_4A = 4679;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.66.0F38.W0 4B /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_512_66_0F38_W0_4B = 4680;
	/**
	 * {@code VADDNPS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 50 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VADDNPS_ZMM_K1_ZMM_ZMMMT = 4681;
	/**
	 * {@code VADDNPD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 50 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VADDNPD_ZMM_K1_ZMM_ZMMMT = 4682;
	/**
	 * {@code VGMAXABSPS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 51 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VGMAXABSPS_ZMM_K1_ZMM_ZMMMT = 4683;
	/**
	 * {@code VGMINPS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 52 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VGMINPS_ZMM_K1_ZMM_ZMMMT = 4684;
	/**
	 * {@code VGMINPD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 52 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VGMINPD_ZMM_K1_ZMM_ZMMMT = 4685;
	/**
	 * {@code VGMAXPS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 53 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VGMAXPS_ZMM_K1_ZMM_ZMMMT = 4686;
	/**
	 * {@code VGMAXPD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 53 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VGMAXPD_ZMM_K1_ZMM_ZMMMT = 4687;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2, zmm3/mt}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 54 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMM_ZMMMT_512_66_0F38_W0_54 = 4688;
	/**
	 * {@code VFIXUPNANPS zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 55 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFIXUPNANPS_ZMM_K1_ZMM_ZMMMT = 4689;
	/**
	 * {@code VFIXUPNANPD zmm1 {k1}, zmm2, Si64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 55 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFIXUPNANPD_ZMM_K1_ZMM_ZMMMT = 4690;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2, zmm3/mt}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 56 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMM_ZMMMT_512_66_0F38_W0_56 = 4691;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2, zmm3/mt}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 57 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMM_ZMMMT_512_66_0F38_W0_57 = 4692;
	/**
	 * {@code VPBROADCASTD zmm1 {k1}, Ui32(mt)}<p>
	 * {@code MVEX.512.66.0F38.W0 58 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPBROADCASTD_ZMM_K1_MT = 4693;
	/**
	 * {@code VPBROADCASTQ zmm1 {k1}, Ui64(mt)}<p>
	 * {@code MVEX.512.66.0F38.W1 59 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPBROADCASTQ_ZMM_K1_MT = 4694;
	/**
	 * {@code VBROADCASTI32X4 zmm1 {k1}, Ui32(mt)}<p>
	 * {@code MVEX.512.66.0F38.W0 5A /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VBROADCASTI32X4_ZMM_K1_MT = 4695;
	/**
	 * {@code VBROADCASTI64X4 zmm1 {k1}, Ui64(mt)}<p>
	 * {@code MVEX.512.66.0F38.W1 5B /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VBROADCASTI64X4_ZMM_K1_MT = 4696;
	/**
	 * {@code VPADCD zmm1 {k1}, k2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 5C /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPADCD_ZMM_K1_KR_ZMMMT = 4697;
	/**
	 * {@code VPADDSETCD zmm1 {k1}, k2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 5D /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPADDSETCD_ZMM_K1_KR_ZMMMT = 4698;
	/**
	 * {@code VPSBBD zmm1 {k1}, k2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 5E /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPSBBD_ZMM_K1_KR_ZMMMT = 4699;
	/**
	 * {@code VPSUBSETBD zmm1 {k1}, k2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 5F /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPSUBSETBD_ZMM_K1_KR_ZMMMT = 4700;
	/**
	 * {@code VPBLENDMD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 64 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPBLENDMD_ZMM_K1_ZMM_ZMMMT = 4701;
	/**
	 * {@code VPBLENDMQ zmm1 {k1}, zmm2, Si64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 64 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPBLENDMQ_ZMM_K1_ZMM_ZMMMT = 4702;
	/**
	 * {@code VBLENDMPS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 65 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VBLENDMPS_ZMM_K1_ZMM_ZMMMT = 4703;
	/**
	 * {@code VBLENDMPD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 65 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VBLENDMPD_ZMM_K1_ZMM_ZMMMT = 4704;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2, zmm3/mt}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 67 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMM_ZMMMT_512_66_0F38_W0_67 = 4705;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.66.0F38.W0 68 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_512_66_0F38_W0_68 = 4706;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.66.0F38.W0 69 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_512_66_0F38_W0_69 = 4707;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.66.0F38.W0 6A /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_512_66_0F38_W0_6A = 4708;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.66.0F38.W0 6B /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_512_66_0F38_W0_6B = 4709;
	/**
	 * {@code VPSUBRD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 6C /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPSUBRD_ZMM_K1_ZMM_ZMMMT = 4710;
	/**
	 * {@code VSUBRPS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 6D /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VSUBRPS_ZMM_K1_ZMM_ZMMMT = 4711;
	/**
	 * {@code VSUBRPD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 6D /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VSUBRPD_ZMM_K1_ZMM_ZMMMT = 4712;
	/**
	 * {@code VPSBBRD zmm1 {k1}, k2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 6E /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPSBBRD_ZMM_K1_KR_ZMMMT = 4713;
	/**
	 * {@code VPSUBRSETBD zmm1 {k1}, k2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 6F /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPSUBRSETBD_ZMM_K1_KR_ZMMMT = 4714;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2, zmm3/mt}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 70 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMM_ZMMMT_512_66_0F38_W0_70 = 4715;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2, zmm3/mt}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 71 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMM_ZMMMT_512_66_0F38_W0_71 = 4716;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2, zmm3/mt}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 72 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMM_ZMMMT_512_66_0F38_W0_72 = 4717;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2, zmm3/mt}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 73 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMM_ZMMMT_512_66_0F38_W0_73 = 4718;
	/**
	 * {@code VPCMPLTD k2 {k1}, zmm1, Si32(zmm2/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 74 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPCMPLTD_KR_K1_ZMM_ZMMMT = 4719;
	/**
	 * {@code VSCALEPS zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 84 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VSCALEPS_ZMM_K1_ZMM_ZMMMT = 4720;
	/**
	 * {@code VPMULHUD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 86 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPMULHUD_ZMM_K1_ZMM_ZMMMT = 4721;
	/**
	 * {@code VPMULHD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 87 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPMULHD_ZMM_K1_ZMM_ZMMMT = 4722;
	/**
	 * {@code VPGATHERDD zmm1 {k1}, Ui32(mvt)}<p>
	 * {@code MVEX.512.66.0F38.W0 90 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPGATHERDD_ZMM_K1_MVT = 4723;
	/**
	 * {@code VPGATHERDQ zmm1 {k1}, Ui64(mvt)}<p>
	 * {@code MVEX.512.66.0F38.W1 90 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPGATHERDQ_ZMM_K1_MVT = 4724;
	/**
	 * {@code VGATHERDPS zmm1 {k1}, Uf32(mvt)}<p>
	 * {@code MVEX.512.66.0F38.W0 92 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VGATHERDPS_ZMM_K1_MVT = 4725;
	/**
	 * {@code VGATHERDPD zmm1 {k1}, Uf64(mvt)}<p>
	 * {@code MVEX.512.66.0F38.W1 92 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VGATHERDPD_ZMM_K1_MVT = 4726;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2, zmm3/mt}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 94 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMM_ZMMMT_512_66_0F38_W0_94 = 4727;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2, zmm3/mt}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 94 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMM_ZMMMT_512_66_0F38_W1_94 = 4728;
	/**
	 * {@code VFMADD132PS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 98 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFMADD132PS_ZMM_K1_ZMM_ZMMMT = 4729;
	/**
	 * {@code VFMADD132PD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 98 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFMADD132PD_ZMM_K1_ZMM_ZMMMT = 4730;
	/**
	 * {@code VFMSUB132PS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 9A /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFMSUB132PS_ZMM_K1_ZMM_ZMMMT = 4731;
	/**
	 * {@code VFMSUB132PD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 9A /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFMSUB132PD_ZMM_K1_ZMM_ZMMMT = 4732;
	/**
	 * {@code VFNMADD132PS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 9C /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFNMADD132PS_ZMM_K1_ZMM_ZMMMT = 4733;
	/**
	 * {@code VFNMADD132PD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 9C /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFNMADD132PD_ZMM_K1_ZMM_ZMMMT = 4734;
	/**
	 * {@code VFNMSUB132PS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 9E /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFNMSUB132PS_ZMM_K1_ZMM_ZMMMT = 4735;
	/**
	 * {@code VFNMSUB132PD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 9E /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFNMSUB132PD_ZMM_K1_ZMM_ZMMMT = 4736;
	/**
	 * {@code VPSCATTERDD mvt {k1}, Di32(zmm1)}<p>
	 * {@code MVEX.512.66.0F38.W0 A0 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPSCATTERDD_MVT_K1_ZMM = 4737;
	/**
	 * {@code VPSCATTERDQ mvt {k1}, Di64(zmm1)}<p>
	 * {@code MVEX.512.66.0F38.W1 A0 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPSCATTERDQ_MVT_K1_ZMM = 4738;
	/**
	 * {@code VSCATTERDPS mvt {k1}, Df32(zmm1)}<p>
	 * {@code MVEX.512.66.0F38.W0 A2 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VSCATTERDPS_MVT_K1_ZMM = 4739;
	/**
	 * {@code VSCATTERDPD mvt {k1}, Df64(zmm1)}<p>
	 * {@code MVEX.512.66.0F38.W1 A2 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VSCATTERDPD_MVT_K1_ZMM = 4740;
	/**
	 * {@code VFMADD233PS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 A4 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFMADD233PS_ZMM_K1_ZMM_ZMMMT = 4741;
	/**
	 * {@code VFMADD213PS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 A8 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFMADD213PS_ZMM_K1_ZMM_ZMMMT = 4742;
	/**
	 * {@code VFMADD213PD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 A8 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFMADD213PD_ZMM_K1_ZMM_ZMMMT = 4743;
	/**
	 * {@code VFMSUB213PS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 AA /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFMSUB213PS_ZMM_K1_ZMM_ZMMMT = 4744;
	/**
	 * {@code VFMSUB213PD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 AA /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFMSUB213PD_ZMM_K1_ZMM_ZMMMT = 4745;
	/**
	 * {@code VFNMADD213PS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 AC /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFNMADD213PS_ZMM_K1_ZMM_ZMMMT = 4746;
	/**
	 * {@code VFNMADD213PD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 AC /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFNMADD213PD_ZMM_K1_ZMM_ZMMMT = 4747;
	/**
	 * {@code VFNMSUB213PS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 AE /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFNMSUB213PS_ZMM_K1_ZMM_ZMMMT = 4748;
	/**
	 * {@code VFNMSUB213PD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 AE /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFNMSUB213PD_ZMM_K1_ZMM_ZMMMT = 4749;
	/**
	 * {@code UNDOC zmm1 {k1}, mvt}<p>
	 * {@code MVEX.512.66.0F38.W0 B0 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_MVT_512_66_0F38_W0_B0 = 4750;
	/**
	 * {@code UNDOC zmm1 {k1}, mvt}<p>
	 * {@code MVEX.512.66.0F38.W0 B2 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_MVT_512_66_0F38_W0_B2 = 4751;
	/**
	 * {@code VPMADD233D zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 B4 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPMADD233D_ZMM_K1_ZMM_ZMMMT = 4752;
	/**
	 * {@code VPMADD231D zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 B5 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPMADD231D_ZMM_K1_ZMM_ZMMMT = 4753;
	/**
	 * {@code VFMADD231PS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 B8 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFMADD231PS_ZMM_K1_ZMM_ZMMMT = 4754;
	/**
	 * {@code VFMADD231PD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 B8 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFMADD231PD_ZMM_K1_ZMM_ZMMMT = 4755;
	/**
	 * {@code VFMSUB231PS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 BA /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFMSUB231PS_ZMM_K1_ZMM_ZMMMT = 4756;
	/**
	 * {@code VFMSUB231PD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 BA /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFMSUB231PD_ZMM_K1_ZMM_ZMMMT = 4757;
	/**
	 * {@code VFNMADD231PS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 BC /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFNMADD231PS_ZMM_K1_ZMM_ZMMMT = 4758;
	/**
	 * {@code VFNMADD231PD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 BC /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFNMADD231PD_ZMM_K1_ZMM_ZMMMT = 4759;
	/**
	 * {@code VFNMSUB231PS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 BE /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFNMSUB231PS_ZMM_K1_ZMM_ZMMMT = 4760;
	/**
	 * {@code VFNMSUB231PD zmm1 {k1}, zmm2, Sf64(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 BE /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VFNMSUB231PD_ZMM_K1_ZMM_ZMMMT = 4761;
	/**
	 * {@code UNDOC zmm1 {k1}, mvt}<p>
	 * {@code MVEX.512.66.0F38.W0 C0 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_MVT_512_66_0F38_W0_C0 = 4762;
	/**
	 * {@code VGATHERPF0HINTDPS Uf32(mvt) {k1}}<p>
	 * {@code MVEX.512.66.0F38.W0 C6 /0 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VGATHERPF0HINTDPS_MVT_K1 = 4763;
	/**
	 * {@code VGATHERPF0HINTDPD Uf64(mvt) {k1}}<p>
	 * {@code MVEX.512.66.0F38.W1 C6 /0 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VGATHERPF0HINTDPD_MVT_K1 = 4764;
	/**
	 * {@code VGATHERPF0DPS Uf32(mvt) {k1}}<p>
	 * {@code MVEX.512.66.0F38.W0 C6 /1 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VGATHERPF0DPS_MVT_K1 = 4765;
	/**
	 * {@code VGATHERPF1DPS Uf32(mvt) {k1}}<p>
	 * {@code MVEX.512.66.0F38.W0 C6 /2 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VGATHERPF1DPS_MVT_K1 = 4766;
	/**
	 * {@code VSCATTERPF0HINTDPS Uf32(mvt) {k1}}<p>
	 * {@code MVEX.512.66.0F38.W0 C6 /4 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VSCATTERPF0HINTDPS_MVT_K1 = 4767;
	/**
	 * {@code VSCATTERPF0HINTDPD Uf64(mvt) {k1}}<p>
	 * {@code MVEX.512.66.0F38.W1 C6 /4 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VSCATTERPF0HINTDPD_MVT_K1 = 4768;
	/**
	 * {@code VSCATTERPF0DPS Uf32(mvt) {k1}}<p>
	 * {@code MVEX.512.66.0F38.W0 C6 /5 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VSCATTERPF0DPS_MVT_K1 = 4769;
	/**
	 * {@code VSCATTERPF1DPS Uf32(mvt) {k1}}<p>
	 * {@code MVEX.512.66.0F38.W0 C6 /6 /vsib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VSCATTERPF1DPS_MVT_K1 = 4770;
	/**
	 * {@code VEXP223PS zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.66.0F38.W0 C8 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VEXP223PS_ZMM_K1_ZMMMT = 4771;
	/**
	 * {@code VLOG2PS zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.66.0F38.W0 C9 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VLOG2PS_ZMM_K1_ZMMMT = 4772;
	/**
	 * {@code VRCP23PS zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.66.0F38.W0 CA /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VRCP23PS_ZMM_K1_ZMMMT = 4773;
	/**
	 * {@code VRSQRT23PS zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.66.0F38.W0 CB /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VRSQRT23PS_ZMM_K1_ZMMMT = 4774;
	/**
	 * {@code VADDSETSPS zmm1 {k1}, zmm2, Sf32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 CC /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VADDSETSPS_ZMM_K1_ZMM_ZMMMT = 4775;
	/**
	 * {@code VPADDSETSD zmm1 {k1}, zmm2, Si32(zmm3/mt)}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 CD /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPADDSETSD_ZMM_K1_ZMM_ZMMMT = 4776;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2, zmm3/mt}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 CE /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMM_ZMMMT_512_66_0F38_W0_CE = 4777;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2, zmm3/mt}<p>
	 * {@code MVEX.NDS.512.66.0F38.W1 CE /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMM_ZMMMT_512_66_0F38_W1_CE = 4778;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2, zmm3/mt}<p>
	 * {@code MVEX.NDS.512.66.0F38.W0 CF /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMM_ZMMMT_512_66_0F38_W0_CF = 4779;
	/**
	 * {@code VLOADUNPACKLD zmm1 {k1}, Ui32(mt)}<p>
	 * {@code MVEX.512.0F38.W0 D0 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VLOADUNPACKLD_ZMM_K1_MT = 4780;
	/**
	 * {@code VLOADUNPACKLQ zmm1 {k1}, Ui64(mt)}<p>
	 * {@code MVEX.512.0F38.W1 D0 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VLOADUNPACKLQ_ZMM_K1_MT = 4781;
	/**
	 * {@code VPACKSTORELD mt {k1}, Di32(zmm1)}<p>
	 * {@code MVEX.512.66.0F38.W0 D0 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPACKSTORELD_MT_K1_ZMM = 4782;
	/**
	 * {@code VPACKSTORELQ mt {k1}, Di64(zmm1)}<p>
	 * {@code MVEX.512.66.0F38.W1 D0 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPACKSTORELQ_MT_K1_ZMM = 4783;
	/**
	 * {@code VLOADUNPACKLPS zmm1 {k1}, Uf32(mt)}<p>
	 * {@code MVEX.512.0F38.W0 D1 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VLOADUNPACKLPS_ZMM_K1_MT = 4784;
	/**
	 * {@code VLOADUNPACKLPD zmm1 {k1}, Uf64(mt)}<p>
	 * {@code MVEX.512.0F38.W1 D1 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VLOADUNPACKLPD_ZMM_K1_MT = 4785;
	/**
	 * {@code VPACKSTORELPS mt {k1}, Df32(zmm1)}<p>
	 * {@code MVEX.512.66.0F38.W0 D1 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPACKSTORELPS_MT_K1_ZMM = 4786;
	/**
	 * {@code VPACKSTORELPD mt {k1}, Df64(zmm1)}<p>
	 * {@code MVEX.512.66.0F38.W1 D1 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPACKSTORELPD_MT_K1_ZMM = 4787;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.0F38.W0 D2 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_512_0F38_W0_D2 = 4788;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.66.0F38.W0 D2 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_512_66_0F38_W0_D2 = 4789;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.0F38.W0 D3 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_512_0F38_W0_D3 = 4790;
	/**
	 * {@code VLOADUNPACKHD zmm1 {k1}, Ui32(mt)}<p>
	 * {@code MVEX.512.0F38.W0 D4 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VLOADUNPACKHD_ZMM_K1_MT = 4791;
	/**
	 * {@code VLOADUNPACKHQ zmm1 {k1}, Ui64(mt)}<p>
	 * {@code MVEX.512.0F38.W1 D4 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VLOADUNPACKHQ_ZMM_K1_MT = 4792;
	/**
	 * {@code VPACKSTOREHD mt {k1}, Di32(zmm1)}<p>
	 * {@code MVEX.512.66.0F38.W0 D4 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPACKSTOREHD_MT_K1_ZMM = 4793;
	/**
	 * {@code VPACKSTOREHQ mt {k1}, Di64(zmm1)}<p>
	 * {@code MVEX.512.66.0F38.W1 D4 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPACKSTOREHQ_MT_K1_ZMM = 4794;
	/**
	 * {@code VLOADUNPACKHPS zmm1 {k1}, Uf32(mt)}<p>
	 * {@code MVEX.512.0F38.W0 D5 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VLOADUNPACKHPS_ZMM_K1_MT = 4795;
	/**
	 * {@code VLOADUNPACKHPD zmm1 {k1}, Uf64(mt)}<p>
	 * {@code MVEX.512.0F38.W1 D5 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VLOADUNPACKHPD_ZMM_K1_MT = 4796;
	/**
	 * {@code VPACKSTOREHPS mt {k1}, Df32(zmm1)}<p>
	 * {@code MVEX.512.66.0F38.W0 D5 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPACKSTOREHPS_MT_K1_ZMM = 4797;
	/**
	 * {@code VPACKSTOREHPD mt {k1}, Df64(zmm1)}<p>
	 * {@code MVEX.512.66.0F38.W1 D5 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPACKSTOREHPD_MT_K1_ZMM = 4798;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.0F38.W0 D6 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_512_0F38_W0_D6 = 4799;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.66.0F38.W0 D6 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_512_66_0F38_W0_D6 = 4800;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt}<p>
	 * {@code MVEX.512.0F38.W0 D7 /r}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_512_0F38_W0_D7 = 4801;
	/**
	 * {@code VALIGND zmm1 {k1}, zmm2, zmm3/mt, imm8}<p>
	 * {@code MVEX.NDS.512.66.0F3A.W0 03 /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VALIGND_ZMM_K1_ZMM_ZMMMT_IMM8 = 4802;
	/**
	 * {@code VPERMF32X4 zmm1 {k1}, zmm2/mt, imm8}<p>
	 * {@code MVEX.512.66.0F3A.W0 07 /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPERMF32X4_ZMM_K1_ZMMMT_IMM8 = 4803;
	/**
	 * {@code VPCMPUD k2 {k1}, zmm1, Si32(zmm2/mt), imm8}<p>
	 * {@code MVEX.NDS.512.66.0F3A.W0 1E /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPCMPUD_KR_K1_ZMM_ZMMMT_IMM8 = 4804;
	/**
	 * {@code VPCMPD k2 {k1}, zmm1, Si32(zmm2/mt), imm8}<p>
	 * {@code MVEX.NDS.512.66.0F3A.W0 1F /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VPCMPD_KR_K1_ZMM_ZMMMT_IMM8 = 4805;
	/**
	 * {@code VGETMANTPS zmm1 {k1}, Sf32(zmm2/mt), imm8}<p>
	 * {@code MVEX.512.66.0F3A.W0 26 /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VGETMANTPS_ZMM_K1_ZMMMT_IMM8 = 4806;
	/**
	 * {@code VGETMANTPD zmm1 {k1}, Sf64(zmm2/mt), imm8}<p>
	 * {@code MVEX.512.66.0F3A.W1 26 /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VGETMANTPD_ZMM_K1_ZMMMT_IMM8 = 4807;
	/**
	 * {@code VRNDFXPNTPS zmm1 {k1}, Sf32(zmm2/mt), imm8}<p>
	 * {@code MVEX.512.66.0F3A.W0 52 /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VRNDFXPNTPS_ZMM_K1_ZMMMT_IMM8 = 4808;
	/**
	 * {@code VRNDFXPNTPD zmm1 {k1}, Sf64(zmm2/mt), imm8}<p>
	 * {@code MVEX.512.66.0F3A.W1 52 /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VRNDFXPNTPD_ZMM_K1_ZMMMT_IMM8 = 4809;
	/**
	 * {@code VCVTFXPNTUDQ2PS zmm1 {k1}, Si32(zmm2/mt), imm8}<p>
	 * {@code MVEX.512.0F3A.W0 CA /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VCVTFXPNTUDQ2PS_ZMM_K1_ZMMMT_IMM8 = 4810;
	/**
	 * {@code VCVTFXPNTPS2UDQ zmm1 {k1}, Sf32(zmm2/mt), imm8}<p>
	 * {@code MVEX.512.66.0F3A.W0 CA /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VCVTFXPNTPS2UDQ_ZMM_K1_ZMMMT_IMM8 = 4811;
	/**
	 * {@code VCVTFXPNTPD2UDQ zmm1 {k1}, Sf64(zmm2/mt), imm8}<p>
	 * {@code MVEX.512.F2.0F3A.W1 CA /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VCVTFXPNTPD2UDQ_ZMM_K1_ZMMMT_IMM8 = 4812;
	/**
	 * {@code VCVTFXPNTDQ2PS zmm1 {k1}, Si32(zmm2/mt), imm8}<p>
	 * {@code MVEX.512.0F3A.W0 CB /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VCVTFXPNTDQ2PS_ZMM_K1_ZMMMT_IMM8 = 4813;
	/**
	 * {@code VCVTFXPNTPS2DQ zmm1 {k1}, Sf32(zmm2/mt), imm8}<p>
	 * {@code MVEX.512.66.0F3A.W0 CB /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VCVTFXPNTPS2DQ_ZMM_K1_ZMMMT_IMM8 = 4814;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt, imm8}<p>
	 * {@code MVEX.512.66.0F3A.W0 D0 /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_IMM8_512_66_0F3A_W0_D0 = 4815;
	/**
	 * {@code UNDOC zmm1 {k1}, zmm2/mt, imm8}<p>
	 * {@code MVEX.512.66.0F3A.W0 D1 /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_UNDOC_ZMM_K1_ZMMMT_IMM8_512_66_0F3A_W0_D1 = 4816;
	/**
	 * {@code VCVTFXPNTPD2DQ zmm1 {k1}, Sf64(zmm2/mt), imm8}<p>
	 * {@code MVEX.512.F2.0F3A.W1 E6 /r ib}<p>
	 * {@code KNC}<p>
	 * {@code 64-bit}
	 */
	public static final int MVEX_VCVTFXPNTPD2DQ_ZMM_K1_ZMMMT_IMM8 = 4817;
	/**
	 * {@code UNDOC}<p>
	 * {@code a16 F3 0F A6 F0}<p>
	 * {@code PADLOCK_UNDOC}<p>
	 * {@code 16/32-bit}
	 */
	public static final int VIA_UNDOC_F30FA6F0_16 = 4818;
	/**
	 * {@code UNDOC}<p>
	 * {@code a32 F3 0F A6 F0}<p>
	 * {@code PADLOCK_UNDOC}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VIA_UNDOC_F30FA6F0_32 = 4819;
	/**
	 * {@code UNDOC}<p>
	 * {@code a64 F3 0F A6 F0}<p>
	 * {@code PADLOCK_UNDOC}<p>
	 * {@code 64-bit}
	 */
	public static final int VIA_UNDOC_F30FA6F0_64 = 4820;
	/**
	 * {@code UNDOC}<p>
	 * {@code a16 F3 0F A6 F8}<p>
	 * {@code PADLOCK_UNDOC}<p>
	 * {@code 16/32-bit}
	 */
	public static final int VIA_UNDOC_F30FA6F8_16 = 4821;
	/**
	 * {@code UNDOC}<p>
	 * {@code a32 F3 0F A6 F8}<p>
	 * {@code PADLOCK_UNDOC}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VIA_UNDOC_F30FA6F8_32 = 4822;
	/**
	 * {@code UNDOC}<p>
	 * {@code a64 F3 0F A6 F8}<p>
	 * {@code PADLOCK_UNDOC}<p>
	 * {@code 64-bit}
	 */
	public static final int VIA_UNDOC_F30FA6F8_64 = 4823;
	/**
	 * {@code XSHA512}<p>
	 * {@code a16 F3 0F A6 E0}<p>
	 * {@code PADLOCK_PHE}<p>
	 * {@code 16/32-bit}
	 */
	public static final int XSHA512_16 = 4824;
	/**
	 * {@code XSHA512}<p>
	 * {@code a32 F3 0F A6 E0}<p>
	 * {@code PADLOCK_PHE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XSHA512_32 = 4825;
	/**
	 * {@code XSHA512}<p>
	 * {@code a64 F3 0F A6 E0}<p>
	 * {@code PADLOCK_PHE}<p>
	 * {@code 64-bit}
	 */
	public static final int XSHA512_64 = 4826;
	/**
	 * {@code XSTORE_ALT}<p>
	 * {@code a16 F3 0F A7 F8}<p>
	 * {@code PADLOCK_RNG}<p>
	 * {@code 16/32-bit}
	 */
	public static final int XSTORE_ALT_16 = 4827;
	/**
	 * {@code XSTORE_ALT}<p>
	 * {@code a32 F3 0F A7 F8}<p>
	 * {@code PADLOCK_RNG}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XSTORE_ALT_32 = 4828;
	/**
	 * {@code XSTORE_ALT}<p>
	 * {@code a64 F3 0F A7 F8}<p>
	 * {@code PADLOCK_RNG}<p>
	 * {@code 64-bit}
	 */
	public static final int XSTORE_ALT_64 = 4829;
	/**
	 * {@code XSHA512_ALT}<p>
	 * {@code a16 F3 0F A6 D8}<p>
	 * {@code PADLOCK_PHE}<p>
	 * {@code 16/32-bit}
	 */
	public static final int XSHA512_ALT_16 = 4830;
	/**
	 * {@code XSHA512_ALT}<p>
	 * {@code a32 F3 0F A6 D8}<p>
	 * {@code PADLOCK_PHE}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int XSHA512_ALT_32 = 4831;
	/**
	 * {@code XSHA512_ALT}<p>
	 * {@code a64 F3 0F A6 D8}<p>
	 * {@code PADLOCK_PHE}<p>
	 * {@code 64-bit}
	 */
	public static final int XSHA512_ALT_64 = 4832;
	/**
	 * A zero-sized instruction.<!-- --> Can be used as a label.<!-- -->
	 */
	public static final int ZERO_BYTES = 4833;
	/**
	 * {@code WRMSRNS}<p>
	 * {@code NP 0F 01 C6}<p>
	 * {@code WRMSRNS}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int WRMSRNS = 4834;
	/**
	 * {@code WRMSRLIST}<p>
	 * {@code F3 0F 01 C6}<p>
	 * {@code MSRLIST}<p>
	 * {@code 64-bit}
	 */
	public static final int WRMSRLIST = 4835;
	/**
	 * {@code RDMSRLIST}<p>
	 * {@code F2 0F 01 C6}<p>
	 * {@code MSRLIST}<p>
	 * {@code 64-bit}
	 */
	public static final int RDMSRLIST = 4836;
	/**
	 * {@code RMPQUERY}<p>
	 * {@code F3 0F 01 FD}<p>
	 * {@code RMPQUERY}<p>
	 * {@code 64-bit}
	 */
	public static final int RMPQUERY = 4837;
	/**
	 * {@code PREFETCHIT1 m8}<p>
	 * {@code 0F 18 /6}<p>
	 * {@code PREFETCHITI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PREFETCHIT1_M8 = 4838;
	/**
	 * {@code PREFETCHIT0 m8}<p>
	 * {@code 0F 18 /7}<p>
	 * {@code PREFETCHITI}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int PREFETCHIT0_M8 = 4839;
	/**
	 * {@code AADD m32, r32}<p>
	 * {@code NP 0F 38 FC !(11):rrr:bbb}<p>
	 * {@code RAO-INT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AADD_M32_R32 = 4840;
	/**
	 * {@code AADD m64, r64}<p>
	 * {@code NP o64 0F 38 FC !(11):rrr:bbb}<p>
	 * {@code RAO-INT}<p>
	 * {@code 64-bit}
	 */
	public static final int AADD_M64_R64 = 4841;
	/**
	 * {@code AAND m32, r32}<p>
	 * {@code 66 0F 38 FC !(11):rrr:bbb}<p>
	 * {@code RAO-INT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AAND_M32_R32 = 4842;
	/**
	 * {@code AAND m64, r64}<p>
	 * {@code 66 o64 0F 38 FC !(11):rrr:bbb}<p>
	 * {@code RAO-INT}<p>
	 * {@code 64-bit}
	 */
	public static final int AAND_M64_R64 = 4843;
	/**
	 * {@code AXOR m32, r32}<p>
	 * {@code F3 0F 38 FC !(11):rrr:bbb}<p>
	 * {@code RAO-INT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AXOR_M32_R32 = 4844;
	/**
	 * {@code AXOR m64, r64}<p>
	 * {@code F3 o64 0F 38 FC !(11):rrr:bbb}<p>
	 * {@code RAO-INT}<p>
	 * {@code 64-bit}
	 */
	public static final int AXOR_M64_R64 = 4845;
	/**
	 * {@code AOR m32, r32}<p>
	 * {@code F2 0F 38 FC !(11):rrr:bbb}<p>
	 * {@code RAO-INT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int AOR_M32_R32 = 4846;
	/**
	 * {@code AOR m64, r64}<p>
	 * {@code F2 o64 0F 38 FC !(11):rrr:bbb}<p>
	 * {@code RAO-INT}<p>
	 * {@code 64-bit}
	 */
	public static final int AOR_M64_R64 = 4847;
	/**
	 * {@code VPDPBUUD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.0F38.W0 50 /r}<p>
	 * {@code AVX-VNNI-INT8}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBUUD_XMM_XMM_XMMM128 = 4848;
	/**
	 * {@code VPDPBUUD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.0F38.W0 50 /r}<p>
	 * {@code AVX-VNNI-INT8}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBUUD_YMM_YMM_YMMM256 = 4849;
	/**
	 * {@code VPDPBSUD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.F3.0F38.W0 50 /r}<p>
	 * {@code AVX-VNNI-INT8}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBSUD_XMM_XMM_XMMM128 = 4850;
	/**
	 * {@code VPDPBSUD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.F3.0F38.W0 50 /r}<p>
	 * {@code AVX-VNNI-INT8}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBSUD_YMM_YMM_YMMM256 = 4851;
	/**
	 * {@code VPDPBSSD xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.F2.0F38.W0 50 /r}<p>
	 * {@code AVX-VNNI-INT8}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBSSD_XMM_XMM_XMMM128 = 4852;
	/**
	 * {@code VPDPBSSD ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.F2.0F38.W0 50 /r}<p>
	 * {@code AVX-VNNI-INT8}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBSSD_YMM_YMM_YMMM256 = 4853;
	/**
	 * {@code VPDPBUUDS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.0F38.W0 51 /r}<p>
	 * {@code AVX-VNNI-INT8}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBUUDS_XMM_XMM_XMMM128 = 4854;
	/**
	 * {@code VPDPBUUDS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.0F38.W0 51 /r}<p>
	 * {@code AVX-VNNI-INT8}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBUUDS_YMM_YMM_YMMM256 = 4855;
	/**
	 * {@code VPDPBSUDS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.F3.0F38.W0 51 /r}<p>
	 * {@code AVX-VNNI-INT8}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBSUDS_XMM_XMM_XMMM128 = 4856;
	/**
	 * {@code VPDPBSUDS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.F3.0F38.W0 51 /r}<p>
	 * {@code AVX-VNNI-INT8}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBSUDS_YMM_YMM_YMMM256 = 4857;
	/**
	 * {@code VPDPBSSDS xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.F2.0F38.W0 51 /r}<p>
	 * {@code AVX-VNNI-INT8}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBSSDS_XMM_XMM_XMMM128 = 4858;
	/**
	 * {@code VPDPBSSDS ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.F2.0F38.W0 51 /r}<p>
	 * {@code AVX-VNNI-INT8}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPDPBSSDS_YMM_YMM_YMMM256 = 4859;
	/**
	 * {@code TDPFP16PS tmm1, tmm2, tmm3}<p>
	 * {@code VEX.128.F2.0F38.W0 5C 11:rrr:bbb}<p>
	 * {@code AMX-FP16}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_TDPFP16PS_TMM_TMM_TMM = 4860;
	/**
	 * {@code VCVTNEPS2BF16 xmm1, xmm2/m128}<p>
	 * {@code VEX.128.F3.0F38.W0 72 /r}<p>
	 * {@code AVX-NE-CONVERT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTNEPS2BF16_XMM_XMMM128 = 4861;
	/**
	 * {@code VCVTNEPS2BF16 xmm1, ymm2/m256}<p>
	 * {@code VEX.256.F3.0F38.W0 72 /r}<p>
	 * {@code AVX-NE-CONVERT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTNEPS2BF16_XMM_YMMM256 = 4862;
	/**
	 * {@code VCVTNEOPH2PS xmm1, m128}<p>
	 * {@code VEX.128.0F38.W0 B0 !(11):rrr:bbb}<p>
	 * {@code AVX-NE-CONVERT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTNEOPH2PS_XMM_M128 = 4863;
	/**
	 * {@code VCVTNEOPH2PS ymm1, m256}<p>
	 * {@code VEX.256.0F38.W0 B0 !(11):rrr:bbb}<p>
	 * {@code AVX-NE-CONVERT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTNEOPH2PS_YMM_M256 = 4864;
	/**
	 * {@code VCVTNEEPH2PS xmm1, m128}<p>
	 * {@code VEX.128.66.0F38.W0 B0 !(11):rrr:bbb}<p>
	 * {@code AVX-NE-CONVERT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTNEEPH2PS_XMM_M128 = 4865;
	/**
	 * {@code VCVTNEEPH2PS ymm1, m256}<p>
	 * {@code VEX.256.66.0F38.W0 B0 !(11):rrr:bbb}<p>
	 * {@code AVX-NE-CONVERT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTNEEPH2PS_YMM_M256 = 4866;
	/**
	 * {@code VCVTNEEBF162PS xmm1, m128}<p>
	 * {@code VEX.128.F3.0F38.W0 B0 !(11):rrr:bbb}<p>
	 * {@code AVX-NE-CONVERT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTNEEBF162PS_XMM_M128 = 4867;
	/**
	 * {@code VCVTNEEBF162PS ymm1, m256}<p>
	 * {@code VEX.256.F3.0F38.W0 B0 !(11):rrr:bbb}<p>
	 * {@code AVX-NE-CONVERT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTNEEBF162PS_YMM_M256 = 4868;
	/**
	 * {@code VCVTNEOBF162PS xmm1, m128}<p>
	 * {@code VEX.128.F2.0F38.W0 B0 !(11):rrr:bbb}<p>
	 * {@code AVX-NE-CONVERT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTNEOBF162PS_XMM_M128 = 4869;
	/**
	 * {@code VCVTNEOBF162PS ymm1, m256}<p>
	 * {@code VEX.256.F2.0F38.W0 B0 !(11):rrr:bbb}<p>
	 * {@code AVX-NE-CONVERT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VCVTNEOBF162PS_YMM_M256 = 4870;
	/**
	 * {@code VBCSTNESH2PS xmm1, m16}<p>
	 * {@code VEX.128.66.0F38.W0 B1 !(11):rrr:bbb}<p>
	 * {@code AVX-NE-CONVERT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBCSTNESH2PS_XMM_M16 = 4871;
	/**
	 * {@code VBCSTNESH2PS ymm1, m16}<p>
	 * {@code VEX.256.66.0F38.W0 B1 !(11):rrr:bbb}<p>
	 * {@code AVX-NE-CONVERT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBCSTNESH2PS_YMM_M16 = 4872;
	/**
	 * {@code VBCSTNEBF162PS xmm1, m16}<p>
	 * {@code VEX.128.F3.0F38.W0 B1 !(11):rrr:bbb}<p>
	 * {@code AVX-NE-CONVERT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBCSTNEBF162PS_XMM_M16 = 4873;
	/**
	 * {@code VBCSTNEBF162PS ymm1, m16}<p>
	 * {@code VEX.256.F3.0F38.W0 B1 !(11):rrr:bbb}<p>
	 * {@code AVX-NE-CONVERT}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VBCSTNEBF162PS_YMM_M16 = 4874;
	/**
	 * {@code VPMADD52LUQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 B4 /r}<p>
	 * {@code AVX-IFMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMADD52LUQ_XMM_XMM_XMMM128 = 4875;
	/**
	 * {@code VPMADD52LUQ ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 B4 /r}<p>
	 * {@code AVX-IFMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMADD52LUQ_YMM_YMM_YMMM256 = 4876;
	/**
	 * {@code VPMADD52HUQ xmm1, xmm2, xmm3/m128}<p>
	 * {@code VEX.128.66.0F38.W1 B5 /r}<p>
	 * {@code AVX-IFMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMADD52HUQ_XMM_XMM_XMMM128 = 4877;
	/**
	 * {@code VPMADD52HUQ ymm1, ymm2, ymm3/m256}<p>
	 * {@code VEX.256.66.0F38.W1 B5 /r}<p>
	 * {@code AVX-IFMA}<p>
	 * {@code 16/32/64-bit}
	 */
	public static final int VEX_VPMADD52HUQ_YMM_YMM_YMMM256 = 4878;
	/**
	 * {@code CMPOXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 E0 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPOXADD_M32_R32_R32 = 4879;
	/**
	 * {@code CMPOXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 E0 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPOXADD_M64_R64_R64 = 4880;
	/**
	 * {@code CMPNOXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 E1 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNOXADD_M32_R32_R32 = 4881;
	/**
	 * {@code CMPNOXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 E1 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNOXADD_M64_R64_R64 = 4882;
	/**
	 * {@code CMPBXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 E2 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPBXADD_M32_R32_R32 = 4883;
	/**
	 * {@code CMPBXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 E2 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPBXADD_M64_R64_R64 = 4884;
	/**
	 * {@code CMPNBXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 E3 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNBXADD_M32_R32_R32 = 4885;
	/**
	 * {@code CMPNBXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 E3 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNBXADD_M64_R64_R64 = 4886;
	/**
	 * {@code CMPZXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 E4 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPZXADD_M32_R32_R32 = 4887;
	/**
	 * {@code CMPZXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 E4 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPZXADD_M64_R64_R64 = 4888;
	/**
	 * {@code CMPNZXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 E5 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNZXADD_M32_R32_R32 = 4889;
	/**
	 * {@code CMPNZXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 E5 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNZXADD_M64_R64_R64 = 4890;
	/**
	 * {@code CMPBEXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 E6 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPBEXADD_M32_R32_R32 = 4891;
	/**
	 * {@code CMPBEXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 E6 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPBEXADD_M64_R64_R64 = 4892;
	/**
	 * {@code CMPNBEXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 E7 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNBEXADD_M32_R32_R32 = 4893;
	/**
	 * {@code CMPNBEXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 E7 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNBEXADD_M64_R64_R64 = 4894;
	/**
	 * {@code CMPSXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 E8 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPSXADD_M32_R32_R32 = 4895;
	/**
	 * {@code CMPSXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 E8 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPSXADD_M64_R64_R64 = 4896;
	/**
	 * {@code CMPNSXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 E9 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNSXADD_M32_R32_R32 = 4897;
	/**
	 * {@code CMPNSXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 E9 !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNSXADD_M64_R64_R64 = 4898;
	/**
	 * {@code CMPPXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 EA !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPPXADD_M32_R32_R32 = 4899;
	/**
	 * {@code CMPPXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 EA !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPPXADD_M64_R64_R64 = 4900;
	/**
	 * {@code CMPNPXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 EB !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNPXADD_M32_R32_R32 = 4901;
	/**
	 * {@code CMPNPXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 EB !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNPXADD_M64_R64_R64 = 4902;
	/**
	 * {@code CMPLXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 EC !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPLXADD_M32_R32_R32 = 4903;
	/**
	 * {@code CMPLXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 EC !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPLXADD_M64_R64_R64 = 4904;
	/**
	 * {@code CMPNLXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 ED !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNLXADD_M32_R32_R32 = 4905;
	/**
	 * {@code CMPNLXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 ED !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNLXADD_M64_R64_R64 = 4906;
	/**
	 * {@code CMPLEXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 EE !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPLEXADD_M32_R32_R32 = 4907;
	/**
	 * {@code CMPLEXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 EE !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPLEXADD_M64_R64_R64 = 4908;
	/**
	 * {@code CMPNLEXADD m32, r32, r32}<p>
	 * {@code VEX.128.66.0F38.W0 EF !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNLEXADD_M32_R32_R32 = 4909;
	/**
	 * {@code CMPNLEXADD m64, r64, r64}<p>
	 * {@code VEX.128.66.0F38.W1 EF !(11):rrr:bbb}<p>
	 * {@code CMPCCXADD}<p>
	 * {@code 64-bit}
	 */
	public static final int VEX_CMPNLEXADD_M64_R64_R64 = 4910;
	// GENERATOR-END: Variants
}
